/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef lint
static char rcsid[] = "$Header: sysv.c 1.4 90/01/04 $";
#endif

/*
 * Sequent additions for dual universe support.
 */

#include <sys/param.h>
#include <ufs/quota.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/file.h>
#include <sys/universe.h>

#include <sgtty.h>
#include <utmp.h>
#include <signal.h>
#include <pwd.h>
#include <stdio.h>
#include <lastlog.h>
#include <errno.h>

#include "pathnames.h"

extern int	univ;

extern  struct	utmp utmp;
#define NMAX	sizeof(utmp.ut_name)
struct	passwd nouser = {"", "nope", -1, -1, -1, "", "", "", "" };
char	rusername[NMAX+1], lusername[NMAX+1];
extern	char	term[64];
extern	struct	passwd *pwd;
char	*strcat(), *rindex(), *index(), *sprintf();

#ifdef SCGACCT
#include <local/scgacct.h>
/*
 * get the account for
 * this user, and set it.
 */
initaccount()
{
	register struct scgacct *ap;
	struct scgacct *getacuid();
	register int i, nacct, c;
	char acct;

	if((ap = getacuid(pwd->pw_uid)) == NULL)
	{
		acct = '0'; /* default */
		goto GOTACCT;
	}
	nacct = 0; 
	while(ap->a_id[nacct] != '\0' && nacct < N_SCGACCT) /* '\0' is marker */
	{
		nacct++;
	}
	if(nacct == 1) /* just one account, so skip the interrogation */
	{
		acct = ap->a_id[0];
		goto GOTACCT;
	}

	for(;;) 
	{
		printf("Account: ");
		fflush(stdout);

		for(acct = 0; (c = getchar()) != '\n'; ) 
		{
			if(c <= 0)	
				exit(0);
			if(acct == 0)
				acct = c;
		}

		for(i = 0; i < nacct; i++)
		{
			if(acct == ap->a_id[i])
				goto GOTACCT;
		}
		printf("Legal Accounts\n");
		fflush(stdout);
		for(i = 0; i < nacct; i++)
		{
			printf("%c - %s\n", ap->a_id[i], ap->a_gno[i]);
			fflush(stdout);
		}
	}
GOTACCT:
	if (setacct(acct) < 0)
		perror("setacct");
	utmp.ut_account = acct;  /* for session accounting */
}
#endif

/*
 * Is s1 a substring of s2?
 */
char *
substr(s1, s2)
	register char *s1, *s2;
{
	register n = strlen(s1);
	
	if (n == 0) 
		return (NULL);
	while (s2 && *s2) {
		if (*s1 == *s2 && strncmp(s1, s2, n) == 0)
			return (s2);
		++s2;
	}
	return (NULL);
}

/*
 * Look for universe specification in the GECOS
 * field of the passwd file entry.  Format requires
 * the string of the form "universe(UNIVERSE)"
 * where UNIVERSE is one of "att" or "ucb".
 */

#define	U_MAGIC "universe("

set_universe(p)
	struct passwd *p;
{
	register char *s, *q;
	int n;

	if ((s = substr(U_MAGIC, p->pw_gecos)) != NULL) {
		/* skip to actual universe name */
		s = q = s + strlen(U_MAGIC);
		while (*s && *s != ')')
			++s;
		if (*s == '\0' || s == q) return;
		*s = '\0';
		if (strcmp("ucb", q) == 0)
			n = universe(univ = U_UCB);
		else if (strcmp("att", q) == 0)
			n = universe(univ = U_ATT);
		else {
			printf("login: unknown universe '%s'.\n", q);
			fflush(stdout);
			return;
		}
		if (n < 0)
			perror("login: universe");
	}
}

set_tz(tz)
	char *tz;
{
	struct	timeval tval;
	struct	timezone tzone;
	char   *timezone(), tzh[4];

	gettimeofday(&tval, &tzone);
	if (DST_NONE != tzone.tz_dsttime) {
		(void)sprintf(tz, "%s%d%s", timezone(tzone.tz_minuteswest, 0),
					    tzone.tz_minuteswest/60,
					    timezone(tzone.tz_minuteswest, 1));
					    /* daylight saving */
	} else {
		(void)sprintf(tz, "%s%d", timezone(tzone.tz_minuteswest, 0),
					  tzone.tz_minuteswest/60);
	}
}

set_dual(pwd)
struct	passwd *pwd;
{
	char	timzone[20];
	char	mailbox[128];
	char	*ucbshell;
	char	*attshell;
	char	*path;

	set_tz(timzone);

	(void)setenv("TZ", timzone, 0);
	if (univ == U_UCB) {
		ucbshell = pwd->pw_shell;
		attshell = "/bin/sh";
		path = ":/usr/ucb:/bin:/usr/bin";
	} else {
		ucbshell = "/bin/sh";
		attshell = pwd->pw_shell;
		path = ":/bin:/usr/bin";
	}
	(void)setenv("UCBPATH", ":/usr/ucb:/bin:/usr/bin", 1);
	(void)setenv("ATTPATH", ":/bin:/usr/bin", 1);
	(void)setenv("UCBSHELL", ucbshell ,1);
	(void)setenv("ATTSHELL", attshell, 1);
	(void)setenv("LOGNAME", pwd->pw_name, 0);
	(void)sprintf(mailbox,"%s/%s", _PATH_MAILDIR, pwd->pw_name);
	(void)setenv("MAIL", mailbox, 0);
	(void)setenv("PATH", path, 0);
}


doremotelogin(host)
	char *host;
{
	FILE *hostf;
	int first = 1;
	char domain[256];

	if (getdomainname(domain, sizeof(domain)) < 0) {
		fprintf(stderr, "login: getdomainname system call missing\n");
		goto bad;
	}
	getstr(rusername, sizeof (rusername), "remuser");
	getstr(lusername, sizeof (lusername), "locuser");
	getstr(term+5, sizeof(term)-5, "Terminal type");
	if (getuid()) {
		pwd = &nouser;
		goto bad;
	}
	setpwent();
	pwd = getpwnam(lusername);
	endpwent();
	if (pwd == NULL) {
		pwd = &nouser;
		goto bad;
	}
	set_universe(pwd);
	hostf = pwd->pw_uid ? fopen("/etc/hosts.equiv", "r") : 0;
again:
	if (hostf) {
		char ahost[32];
		int hostmatch, usermatch;

		while (fgets(ahost, sizeof (ahost), hostf)) {
			register char *p;
			char *user;

			p = ahost;
			while (*p != '\n' && *p != ' ' && *p != '\t' && *p != '\0')
				p++;
			if (*p == ' ' || *p == '\t') {
				*p++ = '\0';
				while (*p == ' ' || *p == '\t')
					p++;
				user = p;
				while (*p != '\n' && *p != ' ' && *p != '\t' && *p != '\0')
					p++;
			} else
				user = p;
			*p = '\0';
			if (ahost[0] == '+' && ahost[1] == 0)
				hostmatch = 1;
			else if (ahost[0] == '+' && ahost[1] == '@')
				hostmatch = innetgr(ahost + 2, host,
				    NULL, domain);
			else if (ahost[0] == '-' && ahost[1] == '@') {
				if (innetgr(ahost+2, host, NULL, domain))
					break;
			}
			else if (ahost[0] == '-') {
				if (!strcmp(host, ahost+1))
					break;
			}
			else
				hostmatch = !strcmp(host, ahost);
			if (*user) {
				if (user[0] == '+' && user[1] == 0)
					usermatch = 1;
				else if (user[0] == '+' && user[1] == '@')
					usermatch = innetgr(user+2, NULL,
					    rusername, domain);
				else if (user[0] == '-' && user[1] == '@') {
					if (innetgr(user+2, NULL,
					    rusername, domain))
						break;
				}
				else if (user[0] == '-') {
					if (!strcmp(user+1, rusername))
						break;
				}
				else
					usermatch = !strcmp(user, rusername);
			}
			else
				usermatch = !strcmp(rusername, lusername);
			if (hostmatch && usermatch) {
				fclose(hostf);
				return (0);
			}
		}
		fclose(hostf);
	}
	if (first == 1) {
		char *rhosts = ".rhosts";
		struct stat sbuf;

		first = 0;
		if (chdir(pwd->pw_dir) < 0)
			goto again;
		if (lstat(rhosts, &sbuf) < 0)
			goto again;
		if ((sbuf.st_mode & S_IFMT) == S_IFLNK) {
			printf("login: .rhosts is a soft link.\r\n");
			goto bad;
		}
		hostf = fopen(rhosts, "r");
		fstat(fileno(hostf), &sbuf);
		if (sbuf.st_uid && sbuf.st_uid != pwd->pw_uid) {
			printf("login: Bad .rhosts ownership.\r\n");
			fclose(hostf);
			goto bad;
		}
		goto again;
	}
bad:
	fflush(stdout);
	return (-1);
}

char	*speeds[] =
    { "0", "50", "75", "110", "134", "150", "200", "300",
      "600", "1200", "1800", "2400", "4800", "9600", "19200", "38400" };
#define	NSPEEDS	(sizeof (speeds) / sizeof (speeds[0]))

doremoteterm(term, tp)
	char *term;
	struct sgttyb *tp;
{
	char *cp = index(term, '/');
	register int i;

	if (cp) {
		*cp++ = 0;
		for (i = 0; i < NSPEEDS; i++)
			if (!strcmp(speeds[i], cp)) {
				tp->sg_ispeed = tp->sg_ospeed = i;
				break;
			}
	}
	tp->sg_flags = ECHO|CRMOD|ANYP|XTABS;
}
