/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifdef RCS
static char rcsid[] = "$Header: boot.c 2.11 87/06/23 $";
#endif

#include <sys/param.h>
#include <sys/inode.h>
#include <sys/fs.h>
#include <sys/vm.h>
#include <sys/reboot.h>
#include <machine/cfg.h>
#include <a.out.h>
#include "saio.h"

caddr_t calloc();

/*
 * Boot program... flags/strings passed in configuration structures determine
 * whether boot stops to ask for system name and which device boot comes from.
 */

main()
{
	register struct config_desc *cd = CD_LOC;
	register int i;
	register char *r, *s;
	register int io;
	static	int retry;
	char	*line;

	/*
	 * Build new style configuration
	 * tables if SCED running old firmware.
	 * A nop if new firmware.
	 */
	buildcfg();

	printf("Boot\n");
	line = calloc(cd->c_bname_size);
	for (;;) {
		if (cd->c_boot_flag & RB_ASKNAME) {
			printf(": ");
			gets(s = line);
			r = cd->c_boot_name;
			/* skip leading white space */
			while (*s == ' ' || *s == '\t')
				s++;
			/* copy new line into config structure */
			for (i = cd->c_bname_size; i-- > 0; ) {
				switch (*s) {
				case '\0':	/* end of line */
					*r++ = 0;
					break;
				case '\t':	/* tab */
				case ' ':	/* space */
					*r++ = 0; s++;
					break;
				default:	/* character */
					*r++ = *s++;
					break;
				}
			}
			bcopy(cd->c_boot_name, 
			    ((struct cfg_ptr *)CFG_PTR)->head_cfg->b_boot_name, 
			    BNAMESIZE);
		} else
			printf(": %s\n", cd->c_boot_name);
		io = open(cd->c_boot_name, 0);
		if (io >= 0)
			copyunix(cd->c_boot_flag, io);
		else
			printf("Can't open %s\n", cd->c_boot_name);
		if (++retry > 2)
			cd->c_boot_flag = RB_SINGLE|RB_ASKNAME;
	}
}

/*ARGSUSED*/
copyunix(howto, io)
	int howto, io;
{
	register int n, offset;
	register struct exec *e;

	/*
	 * start reading file at 1K boundary past end of allocated memory.
	 */
	callocrnd(1024);
	offset = (int)calloc(0);
	e = (struct exec *)offset;

	/* read header and check magic number */
	n = roundup(sizeof(struct exec), DEV_BSIZE);
	if (read(io, offset, n) != n)
		goto sread;
	if (e->a_magic != SMAGIC)
		_stop("Bad a.out magic number");

	/* read text */
	printf("%d", e->a_text);
	if (read(io, offset + n, e->a_text - n) != e->a_text - n)
		goto sread;

	/* read data */
	printf("+%d", n = e->a_data);
	if (read(io, offset + e->a_text, n) != n)
		goto sread;
	
	/* NB: programs clear their own BSS */
	printf("+%d start 0x%x\n", e->a_bss, e->a_entry);

	(void) close(io);

#ifdef	ns32000
	/* detect if this is an old or new standalone binary */
	for (n = sizeof(struct exec); n < CD_STAND_ADDR; n += sizeof(int)) {
		if (*(int *)(offset + n) != 0)
			break;
	}
	if (n >= CD_STAND_ADDR) {
		/* new standalone so copy configuration tables */
		bcopy(CD_LOC, offset + (int)CD_LOC, CD_STAND_ADDR - (int)CD_LOC);
	} else { 
		int entry = e->a_entry;

		printf("warning: old standalone format file.\n");

		/* save old cfg pointer into new program */
		*(struct cfg_ptr *)(offset + CFG_PTR) = *(struct cfg_ptr *)CFG_PTR;

		/* exec new program over ourselves */
		gsp(offset, 0, e->a_text + e->a_data, entry);
		_stop("exec failed");
	}
#else	not ns32000
	/* copy configuration tables */
	bcopy(CD_LOC, offset + (int)CD_LOC, CD_STAND_ADDR - (int)CD_LOC);
#endif	ns32000

	/* exec new program over ourselves */
	gsp(offset, 0, e->a_text + e->a_data, e->a_entry);
sread:
	(void) close(io);
	_stop("short read");
}
