/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: clock.c 2.8 87/09/29 $";
#endif

/*
 * Machine-dependent clock routines.
 *
 * Included are the time-of-day clock initialization and
 * the per processor real-time clock initialization.
 */

/* $Log:	clock.c,v $
 */

#include "../h/param.h"
#include "../h/time.h"
#include "../h/kernel.h"
#include "../h/systm.h"

#include "../balance/cfg.h"
#include "../balance/clock.h"
#include "../balance/slic.h"

#include "../machine/ioconf.h"
#include "../machine/vmparam.h"
#include "../machine/hwparam.h"
#include "../machine/intctl.h"

#include "../sec/sec.h"

extern u_char	cons_scsi;		/* console scsi slic address */

/* For time-of-day handling */
struct	sec_cib *todcib;
struct	sec_gmode todgm;	/* getmodes command */
struct	sec_smode todsm;	/* setmodes command */

/*
 * startrtclock()
 *	Start the real-time clock.
 *
 * Startrtclock restarts the real-time clock, which provides
 * hardclock interrupts to kern_clock.c.  On Sequent HW, this
 * is one-time only per processor (eg, no restart, clock reprimes
 * itself).
 *
 * Called by localinit() during selfinit().
 * This turns on the processor-local SLIC timer.
 *
 * For testing/performance measurement convenience, enable_local_clock
 * allows the per-processor clock to be left OFF.  Need to patch the
 * kernel binary or system memory to effect this.
 */

static	bool_t	enable_local_clock = 1;		/* default ON */

startrtclock()
{
	register struct cpuslic *sl = va_slic;

	if (!enable_local_clock)
		return;

	sl->sl_trv = ((sys_clock_rate * 1000000) / (SL_TIMERDIV * hz)) - 1;
	/* clear prescaler, load reload value */
	sl->sl_tcont = 0;
	sl->sl_tctl = SL_TIMERINT | LCLKBIN;	/* timer on in given bin */
}

/*
 * Routines to manipulate the SCED based time-of-day register.
 * TOD clock interrupt handling done by todclock in kern_clock.c
 *
 *
 * Inittodr initializes the time-of-day hardware which provides
 * date functions. This starts the time-of-day clock.
 *
 * Base provides the time from the filesystem, and the time-of-day
 * clock provides the rest.
 */
inittodr(base)
	time_t base;
{
	register u_int todr;
	register struct	sec_gmode *todgmptr = &todgm;
	register int i;
	long deltat;
	spl_t s_ipl;

	/*
	 * Find console SCED and check if the TOD clock has
	 * failed powerup diagnostics.
	 */
	for (i = 0; i < NSEC; i++) {
		/* is SEC there? */
		if ((SECvec & (1 << i)) == 0)
			continue;

		if (SEC_desc[i].sec_is_cons)
			break;
	}
	if (SEC_desc[i].sec_diag_flags & CFG_S_TOD) {
		/*
		 * Clear todr if TOD failed powerup diagnostics.
		 */
		printf("WARNING: TOD failed powerup diagnostics\n");
		todr = 0;
	} else {
		/*
		 * get the current time-of-day from the SCED tod clock.
		 */
		todgmptr->gm_status = 0;
		todcib->cib_inst = SINST_GETMODE;
		todcib->cib_status = (int *)&todgm;
		s_ipl = splhi();
		mIntr(cons_scsi, TODCLKBIN, SDEV_TOD);
		splx(s_ipl);

		while ((todgmptr->gm_status & SINST_INSDONE) == 0)
			continue;

		if (todgmptr->gm_status != SINST_INSDONE)
			panic("Cannot get TOD value");

		todr = todgmptr->gm_un.gm_tod.tod_newtime;
	}

	if (todr < 13*SECYR) {
		printf("WARNING: TOD value bad -- Setting TOD to file system time\n");
		if (base < 13*SECYR) {
			printf("WARNING: preposterous time in file system -- Setting TOD to default time\n");
			time.tv_sec = 13*SECYR + 19*SECDAY + (2*SECDAY)/3;
			resettodr();
			goto check;
		}
		/*
		 * Believe the time in the file system for lack of
		 * anything better, resetting the TOD.
		 */
		time.tv_sec = base;
		resettodr();
		goto check;
	}

	time.tv_sec = todr;

	/*
	 * See if we gained/lost two or more days;
	 * if so, assume something is amiss.
	 */
	deltat = time.tv_sec - base;
	if (deltat < 0)
		deltat = -deltat;
	resettodr();			/* start the tod clock */
	if (deltat < 2*SECDAY) {
		return;
	}
	printf("WARNING: clock %s %d days",
	    time.tv_sec < base ? "lost" : "gained", deltat / SECDAY);
check:
	printf("\nCHECK AND RESET THE DATE!\n");
}

/*
 * Resettodr restores the time-of-day hardware after a time change.
 * Also, also called via inittodr to start todclock interrupts.
 *
 * Reset the TOD based on the time value; used when the TOD
 * has a preposterous value and also when the time is reset
 * by the stime system call.
 */
resettodr()
{
	register struct sec_smode *todsmptr = &todsm;
	spl_t s_ipl;

	todsmptr->sm_status = 0;
	todsmptr->sm_un.sm_tod.tod_freq = TODFREQ;
	todcib->cib_inst = SINST_SETMODE;
	todcib->cib_status = (int *)&todsm;
	s_ipl = splhi();
	todsmptr->sm_un.sm_tod.tod_newtime = time.tv_sec;
	/*
	 * Bin 3 is sufficient, helps avoid SLIC-bus lockup.
	 */
	mIntr(cons_scsi, 3, SDEV_TOD);
	splx(s_ipl);

	while ((todsmptr->sm_status & SINST_INSDONE) == 0)
		continue;

	if (todsmptr->sm_status != SINST_INSDONE)
		panic("Cannot set TOD value");
}
