/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: swapgeneric.c 2.4 87/01/23 $";
#endif

/*
 * swapgeneric.c
 *	GENERIC swap/root configuration.
 */

/* $Log:	swapgeneric.c,v $
 */

#include "../h/param.h"
#include "../h/conf.h"
#include "../h/buf.h"
#include "../h/vm.h"
#include "../h/systm.h"
#include "../h/reboot.h"

#include "../balance/cfg.h"

#include "../sec/sec.h"

#define	MINARGLEN	5

extern	dev_t rootdev;
extern	struct	swdevt swdevt[];
extern  struct genericconf {
	char	*gc_name;
	dev_t	gc_root;
} genericconf[];

/*
 * The generic rootdev and swapdev are passed into the kernel as the
 * 1st argument in the boot name. The argument specifies the device and
 * unit number to use for the rootdev and swapdev. The swapdev is typically
 * rootdev+1. The actual partitions to be used for a given device is 
 * specified in ../conf/conf_generic.c. An asterix at the end of the string
 * denotes that the root partition is the same as the swap partition.
 * The argument string is in the form:
 *		XXd[*]		(eg. sd0*)
 *
 *	where	XX is the device type (e.g. sd for scsi disk),
 *		d  is the Dynix device unit number,
 *		*  yes '*' is the flag denoting swapdev==rootdev.
 */

setconf()
{
	register struct genericconf *gc;
	register char *name;
	register int unit;
	dev_t	swapdev;
	extern	u_char	cons_scsi;	/* slic address of console scsi */
	extern struct sec_cib *cbdcib;	/* address of Console board device */
	extern unsigned boothowto;

	boothowto |= RB_SINGLE;

	/*
	 * Find Generic rootdev string.
	 *
	 * Skip past bootname, then till argument string.
	 * Error if string not found.
	 */

	for (name = CD_LOC->c_boot_name; *name != '\0'; name++)
		continue;
	while ((*name == '\0') && (name < &CD_LOC->c_boot_name[BNAMESIZE]))
		name++;
	if (name > &CD_LOC->c_boot_name[BNAMESIZE-MINARGLEN]) {
		CD_LOC->c_boot_name[BNAMESIZE-1] = '\0';
		goto error;
	}

	for (gc = genericconf; gc->gc_name; gc++) {
		if (gc->gc_name[0] == name[0] && 
		    gc->gc_name[1] == name[1]) {
			/*
			 * Determine unit number.
			 */
			if (name[2] >= '0' && name[2] <= '7') {
				unit = name[2] - '0';
			} else {
				printf("bad/missing unit number\n");
				break;
			}

			gc->gc_root = makedev(major(gc->gc_root), unit*8);
			rootdev = gc->gc_root;
			swdevt[0].sw_dev = swapdev =
				makedev(major(rootdev), minor(rootdev)+1);
			/*
			 * '*' means rootdev=swapdev
			 */
			if (name[3] == '*')
				rootdev = swapdev;
			else if (name[3] != '\0')
				printf("Extra characters ignored\n");

			printf("Generic; root on %s%d%c, swap on %s%d%c.\n",
				gc->gc_name, unit, 'a'+(minor(rootdev)&7),
				gc->gc_name, unit, 'a'+(minor(swapdev)&7));
			/* swap size set during autoconfigure */
			return;
		}
	}

error:
	printf("Generic boot device ");
	if (*name == NULL)
		printf("not specified.\n");
	else
		printf("\"%s\" is incorrect.\n", name);
	printf("Generic devices:");
	for (gc = genericconf; gc->gc_name; gc++)
		printf(" %s", gc->gc_name);
	printf(".\n");

	printf("Returning to Firmware\n");
	cbdcib->cib_inst = SINST_RETTODIAG;
	cbdcib->cib_status = SRD_BREAK;		/* Halt */
	mIntr(cons_scsi, 7, SDEV_SCSIBOARD);
	for(;;);				/*SCED takes control */
}
