/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*
 * $Header: file.h 2.6 89/07/10 $
 *
 * file.h
 *	File-descriptor structure.
 */

/* $Log:	file.h,v $
 */

#ifdef	KERNEL
/*
 * Descriptor table entry.
 * One for each kernel object.
 */
struct	file {
	int	f_flag;		/* see below */
	long	f_count;	/* reference count */
	short	f_type;		/* descriptor type */
	short	f_msgcount;	/* references from message queue */
				/* with uipc, access rights sending */
	lock_t	f_mutex;	/* lock access to this descriptor */
	short	f_next;		/* free-list thread */
	struct	fileops {
		int	(*fo_rw)();
		int	(*fo_ioctl)();
		int	(*fo_select)();
		int	(*fo_close)();
	} *f_ops;
 	caddr_t	f_data;		/* ptr to file specific struct (vnode/socket) */
	off_t	f_offset;
	off_t	f_uoffset;	/* offset for sysV dir translations */
	off_t	f_ucbboff;	/* another offset for sysV */
 	struct	ucred *f_cred;	/* credentials of user who opened file */
};

extern	struct	file *file, *fileNFILE;
extern	int	nfile;
extern	lock_t	file_list;
struct	file	*getf();
struct	file	*falloc();

/*
 * Lock/unlock, bump f_count macros.
 */

#define	FDLOCK(fp)	p_lock(&(fp)->f_mutex, SPLFD);
#define	FDUNLOCK(fp,s)	v_lock(&(fp)->f_mutex, (s));
#define	FDBUMP(fp)	{spl_t fd_spl; fd_spl = FDLOCK(fp); ++(fp)->f_count; FDUNLOCK(fp,fd_spl);}
#endif	KERNEL

/*
 * flags- also for fcntl call.
 */
#define	FOPEN		(-1)
#define	FREAD		000001		/* descriptor read/receive'able */
#define	FWRITE		000002		/* descriptor write/send'able */
#define	FNDELAY		000004		/* no delay */
#define	FAPPEND		000010		/* append on each write */
#define	FMARK		000020		/* mark during gc() */
#define	FDEFER		000040		/* defer for next gc pass */
#define	FASYNC		000100		/* signal pgrp when data ready */
#define	FSHLOCK		000200		/* shared lock present */
#define	FEXLOCK		000400		/* exclusive lock present */
#define	FSYNC		010000		/* synchronous write option */

/* bits to save after open */
#define	FMASK		(FREAD|FWRITE|FAPPEND|FASYNC|FSYNC)
#define	FCNTLCANT	(FREAD|FWRITE|FMARK|FDEFER|FSHLOCK|FEXLOCK)

/* open only modes */
#define	FCREAT		001000		/* create if nonexistent */
#define	FTRUNC		002000		/* truncate to zero length */
#define	FEXCL		004000		/* error if already created */

/* fcntl(2) requests--from <fcntl.h> */
#define	F_DUPFD	0	/* Duplicate fildes */
#define	F_GETFD	1	/* Get fildes flags */
#define	F_SETFD	2	/* Set fildes flags */
#define	F_GETFL	3	/* Get file flags */
#define	F_SETFL	4	/* Set file flags */
#define	F_GETOWN 5	/* Get owner */
#define F_SETOWN 6	/* Set owner */
#define F_GETLK	7	/* get lock */
#define F_SETLK	8	/* set lock */
#define F_SETLKW 9	/* set lock, with blocking */

/*
 * User definitions.
 */

/*
 * Open call.
 */
#define	O_RDONLY	000		/* open for reading */
#define	O_WRONLY	001		/* open for writing */
#define	O_RDWR		002		/* open for read & write */
#define	O_NDELAY	FNDELAY		/* non-blocking open */
#define	O_APPEND	FAPPEND		/* append on each write */
#define	O_CREAT		FCREAT		/* open with file create */
#define	O_TRUNC		FTRUNC		/* open with truncation */
#define	O_EXCL		FEXCL		/* error on create if file exists */

/*
 * Flock call.
 */
#define	LOCK_SH		1	/* shared lock */
#define	LOCK_EX		2	/* exclusive lock */
#define	LOCK_NB		4	/* don't block when locking */
#define	LOCK_UN		8	/* unlock */

/*
 * Access call.
 */
#define	F_OK		0	/* does file exist */
#define	X_OK		1	/* is it executable by caller */
#define	W_OK		2	/* writable by caller */
#define	R_OK		4	/* readable by caller */

/*
 * Lseek call.
 */
#define	L_SET		0	/* absolute offset */
#define	L_INCR		1	/* relative to current offset */
#define	L_XTND		2	/* relative to end of file */

#ifdef	KERNEL
#define	GETF(fp, fd) { \
	if ((unsigned)(fd) >= u.u_nofile || \
	    ((fp) = u.u_ofile[fd].of_file) == NULL) { \
		u.u_error = EBADF; \
		return; \
	} \
}

#define	GETFR(fp, fd) { \
	if ((unsigned)(fd) >= u.u_nofile || \
	    ((fp) = u.u_ofile[fd].of_file) == NULL) { \
		return (EBADF); \
	} \
}

#define	GET_OFILE(ofp, fd) { \
	if ((unsigned)(fd) >= u.u_nofile || \
	    ((ofp) = &u.u_ofile[fd])->of_file == NULL) { \
		u.u_error = EBADF; \
		return; \
	} \
}

#define	DTYPE_VNODE	1	/* file */
#define	DTYPE_SOCKET	2	/* communications endpoint */
#endif	KERNEL 
