/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: tcp_input.c 2.13 89/08/16 $";
#endif

/*
 * tcp_input.c
 * 	TCP input routine, follows pages 65-76 of the
 * 	protocol specification dated September, 1981 very closely.
 */

/* $Log:	tcp_input.c,v $
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/mutex.h"
#include "../machine/gate.h"
#include "../machine/intctl.h"
#include "../h/mbuf.h"
#include "../h/protosw.h"
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/errno.h"

#include "../net/if.h"
#include "../net/route.h"

#include "../netinet/in.h"
#include "../netinet/in_pcb.h"
#include "../netinet/in_systm.h"
#include "../netinet/ip.h"
#include "../netinet/ip_var.h"
#include "../netinet/tcp.h"
#include "../netinet/tcp_fsm.h"
#include "../netinet/tcp_seq.h"
#include "../netinet/tcp_timer.h"
#include "../netinet/tcp_var.h"
#include "../netinet/tcpip.h"
#include "../netinet/tcp_debug.h"

struct	tcpiphdr tcp_saveti;

struct	tcpcb *tcp_newtcpcb();

#define DODROP			/* for analysis of dropped packets */
#undef DODROP			/* for NO analysis of dropped packets */
#ifdef DODROP
short tcpdrops[64];		/* keep stats on drop codes */
#define DROPIT(x) tcpdrops[(x)]++;
#else
#define DROPIT(x) {;}
#endif DODROP

/*ARGSUSED*/
tcp_input(m0, ifp)
	struct mbuf *m0;
	struct ifnet *ifp;
{
	register struct tcpiphdr *ti;
	register struct mbuf *m;
	register struct tcpcb *tp = (struct tcpcb *)NULL;
	register int tiflags;
	struct inpcb *inp;
	struct mbuf *om = (struct mbuf *)NULL;
	int len, tlen, off;
	struct socket *so;
	int todrop, acked, needoutput = 0;
	short ostate;
	int debugit = 0;
	struct in_addr laddr;
	int dropsocket = 0;
	struct socket_peer * saved_sopp = (struct socket_peer *)NULL;
	struct socket_peer * saved_partial_sopp = (struct socket_peer *)NULL;
	int error;
	struct tcpiphdr *tinew;

	/*
	 * Get IP and TCP header together in first mbuf.
	 * Note: IP leaves IP header in first mbuf.
	 */
	tcpexstat.tcpex_rcvtotal++;

	m = m0;
	ti = mtod(m, struct tcpiphdr *);
	if (((struct ip *)ti)->ip_hl > (sizeof (struct ip) >> 2))
		ip_stripoptions((struct ip *)ti, (struct mbuf *)0);
	if (m->m_off > MMAXOFF || m->m_len < sizeof (struct tcpiphdr)) {
		if ((m = m_pullup(m, sizeof (struct tcpiphdr))) == 
				(struct mbuf *)NULL) {

			/*
			 * note: m_pullup can fail due to failure to allocate
			 * mbufs and so this tcpstat reflects a dropped
			 * packet => best effort fails.
			 *
			 * m_pullup does m_freem(m) if it fails.
			 */

			tcpstat.tcps_hdrops++;
			return;			/* packet dropped; no mbufs */
		}
		ti = mtod(m, struct tcpiphdr *);
	}

	/*
	 * Checksum extended TCP header and data.
	 */

	tlen = ((struct ip *)ti)->ip_len;
	len = sizeof (struct ip) + tlen;

	if (tcpcksumi) {

		/*
		 * The binary configurable value of tcpcksumi != 0
		 * means to compute input checksums.
		 * 4.2bsd uses the variable tcpcksum which is
		 * released == 0;
		 */

		ti->ti_next = ti->ti_prev = 0;
		ti->ti_x1 = 0;
		ti->ti_len = (u_short)tlen;
		ti->ti_len = htons((u_short)ti->ti_len);
		if (ti->ti_sum = in_cksum(m, len)) {
			tcpstat.tcps_badsum++;
			m_freem(m);
			return;		/* packet dropped; bad cksum */
		}
	}

	/*
	 * Check that TCP offset makes sense,
	 * pull out TCP options and adjust length.
	 */

	off = ti->ti_off << 2;
	if (off < sizeof (struct tcphdr) || off > tlen) {
		tcpstat.tcps_badoff++;
		m_freem(m);
		return;		/* packet dropped; bad offset */
	}

	tlen -= off;
	ti->ti_len = tlen;
	if (off > sizeof (struct tcphdr)) {
		if ((m = m_pullup(m, sizeof (struct ip) + off)) == 
				(struct mbuf *)NULL) {

			/*
			 * note: m_pullup can fail due to failure to allocate
			 * mbufs and so this tcpstat reflects a dropped
			 * packet => best effort fails.
			 *
			 * also note that m_pullup releases the mbuf chain if
			 * it fails to accomplish the m_pullup - so no need to
			 * m_freem. 
			 */

			tcpstat.tcps_hdrops++;
			DROPIT(0);
			return;		/* packet dropped; no mbufs */
		}

		ti = mtod(m, struct tcpiphdr *);

		/* N.B. M_DONTWAIT argument not used (also not used in 4.2) */

		om = m_get(M_DONTWAIT, MT_DATA);

		if (om == (struct mbuf *)NULL) {
			/*
			 * failed to m_get mbuf - forgetit
			 */
			DROPIT(1);
			m_freem(m);
			return;		/* packet dropped; no mbufs */
		}

		om->m_len = off - sizeof (struct tcphdr);

		{
		  caddr_t op = mtod(m, caddr_t) + sizeof (struct tcpiphdr);
		  bcopy(op, mtod(om, caddr_t), (unsigned)om->m_len);
		  m->m_len -= om->m_len;

		  /*
		   * account for removal of options  off now = tcphdr only
		   */

		  off = sizeof(struct tcphdr);

		  bcopy(op+om->m_len, op,
		   (unsigned)(m->m_len-sizeof (struct tcpiphdr)));
		}
	}

	tiflags = ti->ti_flags;

	/*
	 * Drop TCP and IP headers.
	 */

	off += sizeof (struct ip);
	m->m_off += off;
	m->m_len -= off;

	/*
	 * Convert TCP protocol specific fields to host format.
	 */

	ti->ti_seq = ntohl(ti->ti_seq);
	ti->ti_ack = ntohl(ti->ti_ack);
	ti->ti_win = ntohs(ti->ti_win);
	ti->ti_urp = ntohs(ti->ti_urp);

	/*
	 * Locate pcb for segment.
	 *
	 * in_pcblookup looks up an internet address in the inpcblist headed
	 * by head.  In this case head == &tcb which is an inpcb
	 * structure that serves as the head of the recognized internet 
	 * addresses of the tcp protocol engine.  in_pcblookup locks the list 
	 * while searching to guard against concurrent changes.
	 * When/if a pcb is found, it's sop_lock is acquired, it's
	 * sop_refcnt count is incremented and the pcb list is unlocked.
	 * NOTE the order of the locking is list head then list element.
	 * If a pcb is not found that is bound to the internet address,
	 * NULL is returned.  N.B. pcblookup spins on pcb lock.
	 */

	inp = in_pcblookup
		(&tcb, ti->ti_src, ti->ti_sport, ti->ti_dst, ti->ti_dport,
		   INPLOOKUP_WILDCARD, 0);

	/*
	 * If the state is CLOSED (i.e., TCB does not exist) then
	 * all data in the incoming segment is discarded.
	 */

	if (inp == (struct inpcb *)NULL) {

		/* pcb not found => state is CLOSED */

		DROPIT(2);

		if (om) (void) m_free(om);

		/*
		 * there is no tp
		 * tcp_respond_reset m_frees m via tcp_respond and ti.
		 * packet is dropped and RST sent since port is closed
		 */

		tcp_respond_reset((struct tcpcb *)NULL, ti, tiflags);
		return;
	}

	/*
	 * pcb is found and port is active - inp_sopp is locked
	 *
	 * save pointer to socket_peer for refcnt managment in case
	 * the pcb goes away because of this input packet.
	 */

	saved_sopp = inp->inp_sopp;
	tp = intotcpcb(inp);
	so = inp->inp_socket;
	if (tp == (struct tcpcb *)NULL || so == (struct socket *)NULL
	     || tp->t_state == TCPS_CLOSED) {

		/*
		 * tp == NULL => not tcp bound 
		 * 	or so == NULL => pcb not bound to socket
		 *	   or t_state == CLOSED
		 * send RST
		 *
		 * unlock the saved_sopp - cannot be last reference
		 * respond with tp == NULL and no further
		 * reference to saved_sopp.
		 */

		DROPIT(3);
		tcp_drop_locks(saved_sopp, saved_partial_sopp);
		tcp_respond_reset((struct tcpcb *)NULL, ti, tiflags);
		if (om) (void) m_free(om);
		return;
	}

	if (so->so_options & SO_DEBUG) {
		ostate = tp->t_state;
		tcp_saveti = *ti;	/* struct copy the tcp header */
		debugit++;
	}

	/*
	 * The socket_peer is locked.  This mutex's user access to the
	 * bound socket.  The socket that
	 * is locked can be a LISTENing socket in which case a new socket
	 * is created to queue to it.  It is the LISTENing socket that is
	 * locked with the socket_peer of the pcb via in_pcblookup.
	 *
	 * sonewconn is called to create this new connect socket. It
	 * returns a *locked* new socket attached and ready to bind.
	 * The order of the locking is queue entry then so_head.
	 * NOTE: the LISTENing socket is locked, but a window
	 * was opened during which the user could have deleted it.
	 * Make sure the LISTENing socket still exists before fooling
	 * with it again (viz. soqinsque).
	 * (FURTHER NOTE: if the LISTENing socket is deleted during
	 * sonewconn, there is no new connection.)
	 */

	if (so->so_options & SO_ACCEPTCONN) {
		/*
		 * so is a LISTEN socket which is accepting
		 * connections.  These connections are created by
		 * sonewconn and queued to the LISTENing socket.
		 *
		 * sonewconn creates a *new* socket structure, makes it look
		 * like the LISTENing socket except for ~SO_ACCEPTCONN.
		 * It is linked into the LISTENing socket's connect
		 * queues as a partial connection.
		 * tcp_usrreq.ATTACH is called to attach a new inpcb.
		 *
		 * if sonewconn() returns NULL - there is no new
		 * socket so drop this request.  This can happen if
		 * the queue is full or some other resource (e.g. mbufs) is not
		 * available.  It also happens if the LISTENing socket
		 * disappears during window.  saved_sopp is locked.
		 */

		so = sonewconn(so);
		if (so == (struct socket *)NULL) {

			/*
			 * sonewconn failed to get new socket
			 * manage sop, and release resources
			 */

			if (--(saved_sopp->sop_refcnt) == 0) {

				/* parent socket disappeared */

				DROPIT(4);
				(void) m_free(dtom(saved_sopp));
				tcp_respond_reset((struct tcpcb *)NULL,
						ti, tiflags);

				/* tcp_respond mfrees m so mark null */

				m = (struct mbuf *)NULL;

			} else { /* dropped for some other reason */

				DROPIT(5);
				v_lock(&saved_sopp->sop_lock, SPLNET);
			}

			if (om) (void) m_free(om);
			if (m) m_freem(m);

			/* packet dropped no new connection acquired */

			return;
		}

		/*
		 * both the new socket and the LISTENing socket are locked
		 * at this point and the LISTENing socket exists.
		 *
		 * inp is now the *new* inpcb bound to the *new* socket 
		 * which is  locked.  Also tp is a *new* tcpcb.
		 * The socket is not known to any user process
		 * until the accept returns but it is queued on the
		 * LISTENing socket referenced by so_head.
		 */

		dropsocket++;

		inp = (struct inpcb *)so->so_pcb;

		saved_partial_sopp = inp->inp_sopp;
		saved_partial_sopp->sop_refcnt++;

		inp->inp_laddr = ti->ti_dst;
		inp->inp_lport = ti->ti_dport;
		inp->inp_options = ip_srcroute();
		tp = intotcpcb(inp);

		/* state is LISTENing, but not SO_ACCEPTCONN */

		tp->t_state = TCPS_LISTEN;

		/* end SO_ACCEPTCONN */

	} else { /* not a new connect request, check for so_head */

		if (so->so_head) {
			
			/*
			 * this is a partial connection
			 *
			 * the pcb found in in_pcblookup (saved_sopp)
			 * is a partial connection whose refcnt includes
			 * this reference.  Save its sopp for later.
			 */

			DROPIT(39);
			SOLOCK(so->so_head);
			saved_partial_sopp = saved_sopp;
			saved_sopp = so->so_head->so_sopp;
			saved_sopp->sop_refcnt++;
			if (!(so->so_head->so_options & SO_ACCEPTCONN)) {

				/*
				 * head is disappearing soclose will
				 * soabort this so
			 	 */

				tcp_drop_locks(saved_sopp, saved_partial_sopp);
				DROPIT(6);
				tcp_respond_reset((struct tcpcb *)NULL,
							ti, tiflags);
				return;
			}
		}
	}

	/*
	 * Segment received on connection.
	 * Reset idle time and keep-alive timer.
	 *
	 * This goes for lookedup socket/pcb or new socket/pcb
	 * both of which are locked if appropriate.
	 *
	 * If it is a partial connection then the so_head is locked
	 * and it is still accepting connections (i.e. not in the
	 * process of being soclosed).
	 */

	if(!(tiflags & TH_RST))
		tp->t_idle = 0;
	tp->t_timer[TCPT_KEEP] = TCPTV_KEEP;


	/*
	 * Process options if not in LISTEN state,
	 * else do it below (after getting remote address).
	 */
	if (om && tp->t_state != TCPS_LISTEN) {
		tcp_dooptions(tp, om, ti);
		om = 0;
	}
	/*
	 * Calculate amount of space in receive window,
	 * and then do TCP input processing.
	 */

	{ int win;

	win = sbspace(&so->so_rcv);
	if (win < 0)
		win = 0;
	tp->rcv_wnd = MAX(win, (int)(tp->rcv_adv - tp->rcv_nxt));
	}

	switch (tp->t_state) {

	/*
	 * If the state is LISTEN then ignore segment if it contains an RST.
	 * If the segment contains an ACK then it is bad and send an RST.
	 * If it does not contain a SYN then it is not interesting; drop it.
	 * If resources cannot be acquired, drop packet.
	 * Otherwise initialize tp->rcv_nxt, and tp->irs, select an initial
	 * tp->iss, and send a segment:
	 *     <SEQ=ISS><ACK=RCV_NXT><CTL=SYN,ACK>
	 * Also initialize tp->snd_nxt to tp->iss+1 and tp->snd_una to tp->iss.
	 * Fill in remote peer address fields if not previously specified.
	 * Enter SYN_RECEIVED state, and process any other fields of this
	 * segment in this state.
	 *
	 * LISTEN state can be entered with *two* locks held -
	 * the lock of the "parent" or Listen socket and the new socket
	 * created to fullfill the connect.
	 *
	 * Both so and so_head need to be locked to avoid simultaneous
	 * modification (e.g. deletion), therefore, if not a new socket
	 * acquire the so_head lock.
	 */

	case TCPS_LISTEN: {
		struct mbuf *am = (struct mbuf *)NULL;
		register struct sockaddr_in *sin;

		/*
		 * Normally so is either a newly created
		 * socket, or one that was queued previously and
		 * not yet changed state.  lock the so_head and check it out.
		 * If this is a new socket, then head is already
		 * locked by sonewconn.  If the sonewconn failed, this
		 * is the parent listen socket that is undergoing
		 * deletion.  This is noted by the observation that
		 * the socket is in LISTEN state but not in any so_q.
		 */

		if (!(so->so_head)) {

			/*
			 * head was disappearing before we in_pcblooked it up.
			 * soclose finishes close
			 */

			tcp_drop_locks(saved_sopp, saved_partial_sopp);
			DROPIT(7);

			/* tcp_respond m_frees m */

			tcp_respond_reset((struct tcpcb *)NULL,
						ti, tiflags);
			return;
		}

		/*
		 * LISTENing socket head is accepting connects
		 * It is locked (saved_sopp) and the partially
		 * connected socket is locked (saved_partial_sopp)
		 */

		laddr = inp->inp_laddr;
		if (laddr.s_addr == INADDR_ANY)
			inp->inp_laddr = ti->ti_dst;

		/*
		 * check out received packet and available resources
		 */

		if ((tiflags & TH_RST) ||	/* if a RST */
		   (tiflags & TH_ACK) ||	/* or an ACK */
		   !(tiflags & TH_SYN) ||	/* or not a SYN */
		   !(am = m_get(M_DONTWAIT, MT_SONAME)) ||
					/* or can't get an mbuf for SONAME */
		   !(tp->t_template = tcp_template(tp)))
					/* or can't build a t_template */
		{
			/*
			 * something is wrong in LISTEN state,
			 * drop packet and back out.
			 * check for need to send RST
			 */

			DROPIT(8);
			if (!(tiflags & TH_RST) && (tiflags & TH_ACK)) {
				DROPIT(9);

				/* tp is always valid */

				tcp_respond(tp, ti, (tcp_seq)0,
						ti->ti_ack, TH_RST);

				/* tcp_respond eats m */

				m = (struct mbuf *)NULL;
			}

			/*
			 * so_head is still accepting connections.
			 * if it is a new socket take it out of the
			 * so_head so_q0 queue and abort it
			 */

			if (dropsocket) {

				/*
				 * newly created socket - backout
				 * by remqueing from so_head's so_q0.
				 */

				(void) soqremque(so, 0);

				/*
				 * unlock head to avoid deadlock with
				 * pcb list management.  Also
				 * dereference sop so soabort gets it.
				 */

				saved_sopp->sop_refcnt--;
				v_lock(&saved_sopp->sop_lock, SPLNET);
				saved_sopp = (struct socket_peer *)NULL;
				DROPIT(10);

				/* make tp NULL to avoid tcp_trace */

				tp = (struct tcpcb *)NULL;
				(void) soabort(so);
			}

			if (debugit && tp)
				tcp_trace(TA_DROP, ostate, tp, &tcp_saveti, 0);

			tcp_drop_locks(saved_sopp, saved_partial_sopp);

			/* release rest of resources */

			if (om) (void) m_free(om);
			if (am) (void) m_free(am);
			if (m) m_freem(m);
			return;
		}

		/*
		 * everything OK so far
		 *
		 * set up SONAME buffer for in_pcbconnect
		 */

		am->m_len = sizeof (struct sockaddr_in);
		sin = mtod(am, struct sockaddr_in *);
		sin->sin_family = AF_INET;
		sin->sin_addr = ti->ti_src;
		sin->sin_port = ti->ti_sport;

		/* 
		 * in_pcbconnect verifies that the
		 * connection is legitimate (e.g. this could be a
		 * retry at the connect attempt) and enters the tcb list. 
		 * in_pcbconnect expects the new socket/pcb to be locked
		 * but not the so_head - so unlock the so_head socket
		 * explicitly and check later for deletion.  During
		 * this WINDOW, everything can disappear but the
		 * socket_peers!  Note: that 
		 * the socket_peers have refcnts for these references since
		 * in_pcblookup does refcnt++ when it finds the LISTEN pcb
		 * or it is ++ above when it is locked and
		 * refcnt++ was done for new so after sonewconn.
		 */

		v_lock(&saved_sopp->sop_lock, SPLNET);
		error = in_pcbconnect(inp, am);
		(void) m_free(am);

		/*
		 * in_pcbconnect returns with inp_sop locked, but inp could
		 * have disappeared.  relock the so_head for state check
		 */

		(void) p_lock(&saved_sopp->sop_lock, SPLNET);

		if ((saved_sopp->sop_refcnt == 1) ||

			/*
			 * last reference => head gone => *both* are gone
			 */

		   (saved_partial_sopp->sop_refcnt == 1) ||

			/*
			 * inp disappeared without head going.
			 * Possible if timer does it.
			 */

		   !(so->so_head->so_options & SO_ACCEPTCONN)) {

			/*
			 * so_head is undergoing deletion, let it
			 * soclose this partial connection
			 */

			tcp_drop_locks(saved_sopp, saved_partial_sopp);
			DROPIT(11);
			return;
		}

		if (error) {

			/*
			 * in_pcbconnect failed, but so_head is still
			 * around.
			 */

			inp->inp_laddr = laddr;

			if (dropsocket) {

				/*
				 * abort this new partial connection
				 * by remqueing it from so_q0.
				 *
				 * unlock head to avoid deadlock with
				 * pcb list management.  Also
				 * dereference sop so soabort gets it.
				 */

				(void) soqremque(so, 0);
				saved_sopp->sop_refcnt--;
				v_lock(&saved_sopp->sop_lock, SPLNET);
				saved_sopp = (struct socket_peer *)NULL;
				DROPIT(12);

				/* make tp NULL to avoid tcp_trace */

				tp = (struct tcpcb *)NULL;
				(void) soabort(so);
			}

			if (debugit && tp)
				tcp_trace(TA_DROP, ostate, tp, &tcp_saveti, 0);

			tcp_drop_locks(saved_sopp, saved_partial_sopp);
			DROPIT(13);
			m_freem(m);
			return;
		}

		/*
		 * LISTENing socket has gotten expected packet and
		 * everything is connected ok.  so_head is locked.
		 * 
		 * recover things that tcp_template creates from tcb
		 */

		tinew = tp->t_template;
		tinew->ti_src = inp->inp_laddr;
		tinew->ti_dst = inp->inp_faddr;
		tinew->ti_sport = inp->inp_lport;
		tinew->ti_dport = inp->inp_fport;

		if (om) {
			tcp_dooptions(tp, om, ti);
			om = 0;
		}

		tp->iss = tcp_iss; tcp_iss += TCP_ISSINCR/2;
		tp->irs = ti->ti_seq;
		tcp_sendseqinit(tp);
		tcp_rcvseqinit(tp);
		tp->t_flags |= TF_ACKNOW;
		tp->t_state = TCPS_SYN_RECEIVED;
		tp->t_timer[TCPT_KEEP] = TCPTV_KEEP;

		dropsocket = 0;
		tcpexstat.tcpex_accepts++;

		goto trimthenstep6;

		/* end LISTEN case */

		}

	/*
	 * If the state is SYN_SENT:
	 *	if seg contains an ACK, but not for our SYN, drop the input.
	 *	if seg contains a RST, then drop the connection.
	 *	if seg does not contain SYN, then drop it.
	 * Otherwise this is an acceptable SYN segment
	 * initialize tp->rcv_nxt and tp->irs
	 * if seg contains ack then advance tp->snd_una
	 * if SYN has been acked change to ESTABLISHED else SYN_RCVD state
	 * arrange for segment to be acked (eventually)
	 * continue processing rest of data/controls, beginning with URG
	 */

	case TCPS_SYN_SENT:

		/*
		 * This state occurs on a socket asking for a connect
		 * Note: SEQ_LEQ here fixes bug when (int)iss < 0
		 */
		if ((tiflags & TH_ACK) &&
		    (SEQ_LEQ(ti->ti_ack, tp->iss) ||
		     SEQ_GT(ti->ti_ack, tp->snd_max))) {

			/* tcp_respond mfrees(m) */

			tcp_respond_reset(tp, ti, tiflags);
			if (debugit && tp)
				tcp_trace(TA_DROP, ostate, tp, &tcp_saveti, 0);

			/* saved_partial_sopp should be NULL */

			tcp_drop_locks(saved_sopp, saved_partial_sopp);
			DROPIT(14);
			return;
		}

		if (tiflags & TH_RST) {

			if (tiflags & TH_ACK)

				/*
				 * tcp_drop calls tcp_close, and in_pcbdetach
				 * which releases inpcb sop_lock still exists
				 * since refcnt > 0 still
				 *
				 * No tp to trace after tcp_drop.
				 */

				tp = tcp_drop(tp, ECONNREFUSED);

			if (debugit && tp)
				tcp_trace(TA_DROP, ostate, tp, &tcp_saveti, 0);

			tcp_drop_locks(saved_sopp, saved_partial_sopp);

			m_freem(m);
			DROPIT(15);
			return;
		}

		if ((tiflags & TH_SYN) == 0) {

			if (debugit && tp)
				tcp_trace(TA_DROP, ostate, tp, &tcp_saveti, 0);

			tcp_drop_locks(saved_sopp, saved_partial_sopp);

			DROPIT(16);
			m_freem(m);
			return;
		}

		tp->snd_una = ti->ti_ack;
		if (SEQ_LT(tp->snd_nxt, tp->snd_una))
			tp->snd_nxt = tp->snd_una;
		tp->t_timer[TCPT_REXMT] = 0;
		tp->irs = ti->ti_seq;
		tcp_rcvseqinit(tp);
		tp->t_flags |= TF_ACKNOW;

		if (SEQ_GT(tp->snd_una, tp->iss)) {

			/* connectee socket is now connected */

			tcpexstat.tcpex_connects++;
			DROPIT(17);
			soisconnected(so);
			tp->t_state = TCPS_ESTABLISHED;
			(void) tcp_reass(tp, (struct tcpiphdr *)NULL);
		} else {
			DROPIT(18);
			tp->t_state = TCPS_SYN_RECEIVED;
		}

		/* fall into ... */
trimthenstep6:

		/*
		 * both LISTEN and SYN_SENT cases execute the following
		 *
		 * Advance ti->ti_seq to correspond to first data byte.
		 * If data, trim to stay within window,
		 * dropping FIN if necessary.
		 */

		ti->ti_seq++;
		if (ti->ti_len > tp->rcv_wnd) {
			todrop = ti->ti_len - tp->rcv_wnd;
			m_adj(m, -todrop);
			ti->ti_len = tp->rcv_wnd;
			ti->ti_flags &= ~TH_FIN;
		}
		tp->snd_wl1 = ti->ti_seq - 1;
		goto step6;
	}	/* this is the end of the switch --------------------- */

	/*
	 * from the 4.2 source the above case statment defaults here and both
	 * cases of the switch use goto's to escape the switch.
	 */

#define MTXINU
#ifdef MTXINU
	/*
	 * the following check for dead socket *before* processing
	 * receive window information is to avoid a tcp hang in
	 * FIN_WAIT_2 <> LAST_ACK  vax4.2 does it *after* receive
	 * window processing.
	 *
	 * If data is received on a connection after the
	 * user processes are gone, then RST the other end.
	 */

	if ((so->so_state & SS_NOFDREF) && tp->t_state > TCPS_CLOSE_WAIT &&
	    ti->ti_len) {

		/*
		 * tcp_close calls in_pcbdetach which releases inpcb 
		 * sop_lock still exists since refcnt > 0 still
		 */
		tcpexstat.tcpex_drops++;

		tp = tcp_close(tp);	/* inpcb is GONE tp = NULL */
		tcp_drop_locks(saved_sopp, saved_partial_sopp);
		tcp_respond_reset((struct tcpcb *)NULL, ti, tiflags);
		DROPIT(19);
		return;
	}
#endif MTXINU
#undef MTXINU

	/*
	 * States other than LISTEN or SYN_SENT.
	 * First check that at least some bytes of segment are within 
	 * receive window. (sic 4.2)
	 */

	if (tp->rcv_wnd == 0) {

		/*
		 * rcv_wnd closed
		 *
		 * If window is closed can only take segments at
		 * window edge, and have to drop data and PUSH from
		 * incoming segments.
		 */

		DROPIT(20);
		if (tp->rcv_nxt != ti->ti_seq) {

			/* rcv_nxt is "edge" */

			DROPIT(21);

			/*
			 * tcp_ackit sends acknowledgement, tcp_traces
			 * and m_frees(m).
			 */

			tcp_ackit(ti, tlen, tiflags, tp, ostate);
			tcp_drop_locks(saved_sopp, saved_partial_sopp);

			return;
		}

		if (ti->ti_len > 0) {

			/*
			 * data present for closed window
			 * m_adj removes data
			 * ti->ti_len = 0 => packet is empty
			 * ti_flags &= removes PUSH and FIN flags
			 *
			 * now packet will seem lost to sender
			 */

			m_adj(m, ti->ti_len);
			ti->ti_len = 0;
			ti->ti_flags &= ~(TH_PUSH|TH_FIN);
		}	

	} else {

		/*
		 * rcv_wnd has room for data
		 *
		 * If segment begins before rcv_nxt, drop leading
		 * data (and SYN); if nothing left, just ack.
		 */

		DROPIT(22);
		todrop = tp->rcv_nxt - ti->ti_seq;
		if (todrop > 0) {

			/*
			 * seq starts *before* rcv_nxt
			 * indicates duplicated data
			 */

			if (tiflags & TH_SYN) {
				tiflags &= ~TH_SYN;
				ti->ti_flags &= ~TH_SYN;
				ti->ti_seq++;
				if (ti->ti_urp > 1) 
					ti->ti_urp--;
				else
					tiflags &= ~TH_URG;
				todrop--;
			}

			/* drop all data unless FIN */

			if (todrop > ti->ti_len ||  
			    todrop == ti->ti_len && (tiflags&TH_FIN) == 0)
			{

				/*
				 * tcp_ackit sends acknowledgement and
				 * m_frees(m).
				 */

				tcp_ackit(ti, tlen, tiflags, tp, ostate);
				tcp_drop_locks(saved_sopp, saved_partial_sopp);
				DROPIT(23);
				return;
			}

			/* m_adj by the amount to drop */

			m_adj(m, todrop);

			/* advance sequence number past to drop */

			ti->ti_seq += todrop;

			/* decrease len by to drop */

			ti->ti_len -= todrop;
			if (ti->ti_urp > todrop)
				ti->ti_urp -= todrop;
			else {
				tiflags &= ~TH_URG;
				ti->ti_flags &= ~TH_URG;
				ti->ti_urp = 0;
			}
		}

		/*
		 * If segment ends after window, drop trailing data
		 * (and PUSH and FIN); if nothing left, just ACK.
		 */

		todrop = (ti->ti_seq+ti->ti_len) - (tp->rcv_nxt+tp->rcv_wnd);

		if (todrop > 0) {

			/* cant take all of packet */

			if (todrop >= ti->ti_len) {

				/* cant take any */

				/*
				 * tcp_ackit sends acknowledgement and
				 * m_frees(m).
				 */
	
				tcp_ackit(ti, tlen, tiflags, tp, ostate);
				tcp_drop_locks(saved_sopp, saved_partial_sopp);
				DROPIT(24);
				return;
			}

			/* drop whatever */

			m_adj(m, -todrop);

			/* adjust the length */

			ti->ti_len -= todrop;

			/* turn off the flags */

			ti->ti_flags &= ~(TH_PUSH|TH_FIN);
		}

		/*
		 * note that if todrop <= 0 => rcv window is large enough to
		 * accept all of packet 
		 */

	/* end of the receive window processing ------------------ */

	}

/* if MTXINU define vax4.2 here for mtxinu backout XXX */

#undef vax4_2
#ifdef vax4_2	/*
		 * this segment of code is moved to execution *before*
		 * receive window processing to avoid *not* being
		 * executed at all.
		 * vax4.2 hangs in FIN_WAIT_2 <> LAST_ACK
		 *
		 * If data is received on a connection after the
		 * user processes are gone, then RST the other end.
		 */

	if ((so->so_state & SS_NOFDREF) && tp->t_state > TCPS_CLOSE_WAIT &&
	    ti->ti_len) {

		/*
		 * tcp_close calls in_pcbdetach which releases inpcb 
		 * sop_lock still exists since refcnt > 0 still
		 */
		tcpexstat.tcpex_drops++;
		tp = tcp_close(tp);	/* inpcb is GONE tp = NULL */
		DROPIT(25);

		/* tcp_respond_reset m_frees(m) */

		tcp_respond_reset((struct tcpcb *)NULL, ti, tiflags);
		tcp_drop_locks(saved_sopp, saved_partial_sopp);
		return;
	}
#endif vax4_2
#undef vax4_2		/* MTXINU consideration */

	/*
	 * If the RST bit is set examine the state:
	 *    SYN_RECEIVED STATE:
	 *	If passive open, return to LISTEN state.
	 *	If active open, inform user that connection was refused.
	 *    ESTABLISHED, FIN_WAIT_1, FIN_WAIT2, CLOSE_WAIT STATES:
	 *	Inform user that connection was reset, and close tcb.
	 *    CLOSING, LAST_ACK, TIME_WAIT STATES
	 *	Close the tcb.
	 */

	/*
	 * if socket has so_head, then *both* are locked
 	 * so_head is still accepting connections.
 	 */

	if (tiflags&TH_RST) {

		switch (tp->t_state) {

		case TCPS_SYN_RECEIVED:

			/*
			 * tcp_drop calls tcp_close, and in_pcbdetach which
			 * releases inpcb sop_lock still exists since
			 * refcnt > 0
		 	 * note tcp_drop -> tcp_close -> returns NULL
		 	 */

			tp = tcp_drop(tp, ECONNREFUSED);
			break;

		case TCPS_ESTABLISHED:
		case TCPS_FIN_WAIT_1:
		case TCPS_FIN_WAIT_2:
		case TCPS_CLOSE_WAIT:

			/*
			 * tcp_drop calls tcp_close, and in_pcbdetach which
			 * releases inpcb sop_lock still exists since
			 * refcnt > 0 still
		 	 */
	
			tp = tcp_drop(tp, ECONNRESET);
			break;

		case TCPS_CLOSING:
		case TCPS_LAST_ACK:
		case TCPS_TIME_WAIT:

			/*
			 * tcp_close calls in_pcbdetach which releases inpcb 
		 	 * sop_lock still exists since refcnt > 0 still
		 	 */

			tp = tcp_close(tp);
			break;
		}

		tcp_drop_locks(saved_sopp, saved_partial_sopp);
		m_freem(m);
		DROPIT(26);
		return;

	/* end of if() switch ----------------------------- */
	}

	/*
	 * If a SYN is in the window, then this is an
	 * error and we send an RST and drop the connection.
	 */

	if (tiflags & TH_SYN) {

		/*
		 * tcp_drop calls tcp_close, which calls in_pcbdetach which
		 * releases inpcb sop_lock still exists since refcnt > 0 still
		 */

		tp = tcp_drop(tp, ECONNRESET);
		tcp_respond_reset((struct tcpcb *)NULL, ti, tiflags);
		tcp_drop_locks(saved_sopp, saved_partial_sopp);
		DROPIT(27);
		return;
	}

	/*
	 * If the ACK bit is off we drop the segment and return.
	 */

	if ((tiflags & TH_ACK) == 0) {

		if (debugit && tp)
			tcp_trace(TA_DROP, ostate, tp, &tcp_saveti, 0);

		tcp_drop_locks(saved_sopp, saved_partial_sopp);
		m_freem(m);
		DROPIT(28);
		return;
	}
	
	/*
	 * Ack processing.	---------------------------------
	 */

	switch (tp->t_state) {

	/*
	 * In SYN_RECEIVED state if the ack ACKs our SYN then enter
	 * ESTABLISHED state and continue processing, otherwise
	 * send an RST.
	 */

	case TCPS_SYN_RECEIVED:

		if (SEQ_GT(tp->snd_una, ti->ti_ack) ||
		    SEQ_GT(ti->ti_ack, tp->snd_max)) {

			/* tcp_respond eats m via ti */

			tcp_respond_reset(tp, ti, tiflags);

			if (debugit && tp)
				tcp_trace(TA_DROP, ostate, tp, &tcp_saveti, 0);

			tcp_drop_locks(saved_sopp, saved_partial_sopp);
			DROPIT(29);
			return;
		}

		/* SYN ACKed */
		tcpexstat.tcpex_connects++;
		tp->snd_una++;
		if (SEQ_LT(tp->snd_nxt, tp->snd_una))
			tp->snd_nxt = tp->snd_una;
		tp->t_timer[TCPT_REXMT] = 0;

		/* soisconnected signals the acceptor */

		soisconnected(so);

		DROPIT(30);
		tp->t_state = TCPS_ESTABLISHED;
		(void) tcp_reass(tp, (struct tcpiphdr *)NULL);
		tp->snd_wl1 = ti->ti_seq - 1;

		/*
		 * now, theoretically, anything can happen on the new
		 * connection that can happen on an established connection.  
		 */

		/* fall into ... */

	/*
	 * In ESTABLISHED state: drop duplicate ACKs; ACK out of range
	 * ACKs.  If the ack is in the range
	 *	tp->snd_una < ti->ti_ack <= tp->snd_max
	 * then advance tp->snd_una to ti->ti_ack and drop
	 * data from the retransmission queue.  If this ACK reflects
	 * more up to date window information we update our window information.
	 */

	case TCPS_ESTABLISHED:
	case TCPS_FIN_WAIT_1:
	case TCPS_FIN_WAIT_2:
	case TCPS_CLOSE_WAIT:
	case TCPS_CLOSING:
	case TCPS_LAST_ACK:
	case TCPS_TIME_WAIT:

#define	ourfinisacked	(acked > 0)

		if (SEQ_LEQ(ti->ti_ack, tp->snd_una)) 
			break;			/* no newly acked data */

		if (SEQ_GT(ti->ti_ack, tp->snd_max)) {

			/*
			 * more acked than sent?
			 * tcp_ackit m_frees m
			 */

			tcp_ackit(ti, tlen, tiflags, tp, ostate);
			tcp_drop_locks(saved_sopp, saved_partial_sopp);
			DROPIT(31);
			return;
		}

	 	/* compute number of bytes being acked */

		acked = ti->ti_ack - tp->snd_una;

		/*
		 * If transmit timer is running and timed sequence
		 * number was acked, update smoothed round trip time.
		 *
		 * We keep mean round trip time "avg" and measured
		 * devicance "dev".
		 *
		 * 	  err = abs(rtt - avg)
		 * 	  dev = (1 - g2) * dev + g2 * err
		 * 	  avg = (1 - g1) * avg + g1 * rtt
		 *    
		 * For effciency, the values are store scaled, ie.
		 *     t_mrtt = avg * g1
		 *     t_drtt = dev * g2
		 *
		 * the values: g1,g2 are actually stored as g = 1/2^n
		 *     ie. if TCP_AVGSHIFT is  3 then  g1 = 1/8
		 *
		 * Initial values for average and mean deviance are set with
		 * first time estimate.
		 * Note: t_mrtt can not go to zero since t_rtt can not be zero.
		 */

		if (tp->t_rtt && SEQ_GT(ti->ti_ack, tp->t_rtseq)) {
			if (tp->t_mrtt == 0) {
				tp->t_mrtt = tp->t_rtt << TCP_AVGSHIFT;
				tp->t_drtt = tp->t_rtt << (TCP_MDEVSHIFT-1);
			} else {
				register int mrt;

				mrt = tp->t_rtt - (tp->t_mrtt >> TCP_AVGSHIFT);
				tp->t_mrtt += mrt;
				if(mrt < 0) mrt = -mrt;
				mrt -= (tp->t_drtt >> TCP_MDEVSHIFT);
				tp->t_drtt += mrt;
			}
			tp->t_rtt = 0;
		}


		/*
		 * If all outstanding data is acked, stop retransmit
		 * timer and remember to restart (more output or persist).
		 * If there is more data to be acked, restart retransmit
		 * timer.
		 */
		if (ti->ti_ack == tp->snd_max)  {  /* everything is acked */
			tp->t_timer[TCPT_REXMT] = 0;
			needoutput = 1;
		} else if (tp->t_timer[TCPT_PERSIST] == 0) {
			int rto;
			if (tp->t_mrtt == 0)
				rto = tcp_default_rexmt*PR_SLOWHZ;
			else
				rto = (tp->t_mrtt >> TCP_AVGSHIFT)
					+ (tp->t_drtt >> (TCP_MDEVSHIFT-1));
			TCPT_RANGESET(tp->t_timer[TCPT_REXMT], rto,
			    TCPTV_MIN, TCPTV_MAX);

			tp->t_rxtshift = 0;
		}
		/*
		 * When new data is acked, open the congestion window a bit.
		 */
		if (acked > 0)
			tp->snd_cwnd = MIN(11 * tp->snd_cwnd / 10, 65535);
		if (acked > so->so_snd.sb_cc) {	/* acked more than queued? */
			tp->snd_wnd -= so->so_snd.sb_cc;
			sbdrop(&so->so_snd, (int)so->so_snd.sb_cc);
		} else {	/* acked <= queued, dropem and adjust */
			sbdrop(&so->so_snd, acked);
			tp->snd_wnd -= acked;
			acked = 0;
		}


		/*
		 * If there is a process waiting or selecting,
		 * this is where user is awakened to transmit data.
		 */
		if (blocked_sema(&so->so_snd.sb_sbx->sbx_buf_wait) ||
		    so->so_snd.sb_sel)
			sowwakeup(so);

		/* update acked packets */

		tp->snd_una = ti->ti_ack;

		/* update snd up to those acked */

		if (SEQ_LT(tp->snd_nxt, tp->snd_una))
			tp->snd_nxt = tp->snd_una;

		switch (tp->t_state) {

		/*
		 * In FIN_WAIT_1 STATE in addition to the processing
		 * for the ESTABLISHED state if our FIN is now acknowledged
		 * then enter FIN_WAIT_2.
		 */

		case TCPS_FIN_WAIT_1:
			if (ourfinisacked) {
				/*
				 * If we can't receive any more
				 * data, then closing user can proceed.
				 * Starting the timer is contrary to the
				 * specification, but if we don't get a FIN
				 * we'll hang forever.
				 * when timer goes off we check to see if
				 * connection is idle before closing
				 */
				if (so->so_state & SS_CANTRCVMORE) {
					soisdisconnected(so);
					tp->t_timer[TCPT_2MSL] = TCPTV_MAXIDLE;
				}
				tp->t_state = TCPS_FIN_WAIT_2;
			}
			break;

	 	/*
		 * In CLOSING STATE in addition to the processing for
		 * the ESTABLISHED state if the ACK acknowledges our FIN
		 * then enter the TIME-WAIT state, otherwise ignore
		 * the segment.
		 */

		case TCPS_CLOSING:
			if (ourfinisacked) {
				tp->t_state = TCPS_TIME_WAIT;
				tcp_canceltimers(tp);
				tp->t_timer[TCPT_2MSL] = 2 * TCPTV_MSL;

				/*
				 * soisdisconnected assumes that the
				 * socket peer of so is *held*
				 */

				soisdisconnected(so);
			}
			break;

		/*
		 * If our FIN is now acknowledged, delete the TCB, enter
                 * the closed state and return.
		 * If any other AK, continue processing (for example,
		 * there may be data waiting to be sent ...).
		 */

		case TCPS_LAST_ACK:
			if (ourfinisacked) {

				/*
				 * tcp_close calls in_pcbdetach which releases
				 * inpcb sop_lock still exists since
				 * refcnt > 0 still
				 */

				tp = tcp_close(tp); /* inpcb is GONE */

				tcp_drop_locks(saved_sopp, saved_partial_sopp);
				m_freem(m);
				DROPIT(34);
				return;
			}
			break;


		/*
		 * In TIME_WAIT state the only thing that should arrive
		 * is a retransmission of the remote FIN.  Acknowledge
		 * it and restart the finack timer.
		 */

		case TCPS_TIME_WAIT:
			tp->t_timer[TCPT_2MSL] = 2 * TCPTV_MSL;
			tcp_ackit(ti, tlen, tiflags, tp, ostate);
			tcp_drop_locks(saved_sopp, saved_partial_sopp);
			DROPIT(35);
			return;
		}

		/* end of nested switch */

#undef ourfinisacked
	}

	/* end of ACK processing switch */

step6:
	/*
	 * Update window information.
	 * Don't look at window if no ACK: TAC's send garbage on first SYN.
	 */

	if ((tiflags & TH_ACK) &&
	    (SEQ_LT(tp->snd_wl1, ti->ti_seq) || tp->snd_wl1 == ti->ti_seq &&
	    (SEQ_LT(tp->snd_wl2, ti->ti_ack) ||
	     tp->snd_wl2 == ti->ti_ack && ti->ti_win > tp->snd_wnd))) {
		tp->snd_wnd = ti->ti_win;
		tp->snd_wl1 = ti->ti_seq;
		tp->snd_wl2 = ti->ti_ack;
		if (tp->snd_wnd > tp->max_sndwnd)
			tp->max_sndwnd = tp->snd_wnd;
		needoutput = 1;
	}

	/*
	 * Process segments with URG.
	 */

	if ((tiflags & TH_URG) && ti->ti_urp &&
	    TCPS_HAVERCVDFIN(tp->t_state) == 0) {

		/*
		 * If we receive accept
		 * random urgent pointers, we'll crash in
		 * soreceive.  It's hard to imagine someone
		 * actually wanting to send this much urgent data (sic 4.2).
		 */

		if (ti->ti_urp + so->so_rcv.sb_cc > 65535) {
			ti->ti_urp = 0;			/* XXX */
			tiflags &= ~TH_URG;		/* XXX */
			goto dodata;			/* XXX */
		}

		/*
		 * If this segment advances the known urgent pointer,
		 * then mark the data stream.  This should not happen
		 * in CLOSE_WAIT, CLOSING, LAST_ACK or TIME_WAIT STATES since
		 * a FIN has been received from the remote side. 
		 * In these states we ignore the URG.
		 */

		if (SEQ_GT(ti->ti_seq+ti->ti_urp, tp->rcv_up)) {
			tp->rcv_up = ti->ti_seq + ti->ti_urp;
			so->so_oobmark = so->so_rcv.sb_cc +
			    (tp->rcv_up - tp->rcv_nxt) - 1;
			if (so->so_oobmark == 0)
				so->so_state |= SS_RCVATMARK;
			sohasoutofband(so);
			tp->t_oobflags &= ~TCPOOB_HAVEDATA;
		}

		/*
		 * Remove out of band data so doesn't get presented to user.
		 * This can happen independent of advancing the URG pointer,
		 * but if two URG's are pending at once, some out-of-band
		 * data may creep in... ick (sic 4.2).
		 */

		if (ti->ti_urp <= ti->ti_len)
			tcp_pulloutofband(so, ti);
	} else
		/*
		 * If no out of band data is expected,
		 * pull receive urgent pointer along
		 * with the receive window.
		 */
		if (SEQ_GT(tp->rcv_nxt, tp->rcv_up))
			tp->rcv_up = tp->rcv_nxt;

dodata:
	/*
	 * Process the segment text, merging it into the TCP sequencing queue,
	 * and arranging for acknowledgment of receipt if necessary.
	 * This process logically involves adjusting tp->rcv_wnd as data
	 * is presented to the user (this happens in tcp_usrreq.c,
	 * case PRU_RCVD).  If a FIN has already been received on this
	 * connection then we just ignore the text.
	 */

	if ((ti->ti_len || (tiflags&TH_FIN)) &&
	    TCPS_HAVERCVDFIN(tp->t_state) == 0) {
		tiflags = tcp_reass(tp, ti);
		if (tcpnodelack == 0)	/* defined = 0 in conf_net */
			tp->t_flags |= TF_DELACK;
		else
			tp->t_flags |= TF_ACKNOW;
	} else {
		m_freem(m);
		tiflags &= ~TH_FIN;
	}

	/*
	 * If FIN is received ACK the FIN and let the user know
	 * that the connection is closing.
	 */

	if (tiflags & TH_FIN) {
		if (TCPS_HAVERCVDFIN(tp->t_state) == 0) {
			socantrcvmore(so);
			DROPIT(36);
			tp->t_flags |= TF_ACKNOW;
			tp->rcv_nxt++;
		}
		switch (tp->t_state) {

	 	/*
		 * In SYN_RECEIVED and ESTABLISHED STATES
		 * enter the CLOSE_WAIT state.
		 */

		case TCPS_SYN_RECEIVED:
			DROPIT(37);
		case TCPS_ESTABLISHED:
			tp->t_state = TCPS_CLOSE_WAIT;
			break;

	 	/*
		 * If still in FIN_WAIT_1 STATE FIN has not been acked so
		 * enter the CLOSING state.
		 */

		case TCPS_FIN_WAIT_1:
			tp->t_state = TCPS_CLOSING;
			break;

	 	/*
		 * In FIN_WAIT_2 state enter the TIME_WAIT state,
		 * starting the time-wait timer, turning off the other 
		 * standard timers.
		 */

		case TCPS_FIN_WAIT_2:
			tp->t_state = TCPS_TIME_WAIT;
			tcp_canceltimers(tp);
			tp->t_timer[TCPT_2MSL] = 2 * TCPTV_MSL;
			soisdisconnected(so);
			DROPIT(38);
			break;

		/*
		 * In TIME_WAIT state restart the 2 MSL time_wait timer.
		 */

		case TCPS_TIME_WAIT:
			tp->t_timer[TCPT_2MSL] = 2 * TCPTV_MSL;
			break;
		}
	}

	if (debugit)
		tcp_trace(TA_INPUT, ostate, tp, &tcp_saveti, 0);

	/*
	 * Return any desired output.
	 */
	if (needoutput || (tp->t_flags & TF_ACKNOW))
		(void) tcp_output(tp);

	/* finished with pcb - release socket_peer lock(s) */

	tcp_drop_locks(saved_sopp, saved_partial_sopp);
	return;
}

tcp_dooptions(tp, om, ti)
	struct tcpcb *tp;
	struct mbuf *om;
	struct tcpiphdr *ti;
{
	register u_char *cp;
	int opt, optlen, cnt;

	cp = mtod(om, u_char *);
	cnt = om->m_len;
	for (; cnt > 0; cnt -= optlen, cp += optlen) {
		opt = cp[0];
		if (opt == TCPOPT_EOL)
			break;
		if (opt == TCPOPT_NOP)
			optlen = 1;
		else {
			optlen = cp[1];
			if (optlen <= 0)
				break;
		}
		switch (opt) {

		default:
			break;

		case TCPOPT_MAXSEG:
			if (optlen != 4)
				continue;
			if (!(ti->ti_flags & TH_SYN))
				continue;
			tp->t_maxseg = *(u_short *)(cp + 2);
			tp->t_maxseg = ntohs((u_short)tp->t_maxseg);
			if( (int) tp->t_maxseg <= 0) {
				printf("tcp_input:t_maxseg <=0 from (0x%x)\n",
				       ti->ti_src);
				tp->t_maxseg = tcp_mss(tp);
			} else {
				tp->t_maxseg = MIN(tp->t_maxseg, tcp_mss(tp));
			}
			break;
		}
	}
	(void) m_free(om);
}

/*
 * Pull out of band byte out of a segment so
 * it doesn't appear in the user's data queue.
 * It is still reflected in the segment length for
 * sequencing purposes.
 */

tcp_pulloutofband(so, ti)
	struct socket *so;
	struct tcpiphdr *ti;
{
	register struct mbuf *m;
	int cnt = ti->ti_urp - 1;
	
	m = dtom(ti);
	while (cnt >= 0) {
		if (m->m_len > cnt) {
			char *cp = mtod(m, caddr_t) + cnt;
			struct tcpcb *tp = sototcpcb(so);

			tp->t_iobc = *cp;
			tp->t_oobflags |= TCPOOB_HAVEDATA;
			bcopy(cp+1, cp, (unsigned)(m->m_len - cnt - 1));
			m->m_len--;
			return;
		}
		cnt -= m->m_len;
		m = m->m_next;
		if (m == (struct mbuf *)NULL)
			break;
	}
	panic("tcp_pulloutofband");
}

/*
 * Insert segment ti into reassembly queue of tcp with
 * control block tp.  Return TH_FIN if reassembly now includes
 * a segment with FIN.
 */

tcp_reass(tp, ti)
	register struct tcpcb *tp;
	register struct tcpiphdr *ti;
{
	register struct tcpiphdr *q;
	struct socket *so = tp->t_inpcb->inp_socket;
	struct mbuf *m;
	int flags;

	/*
	 * Call with ti==0 after become established to
	 * force pre-ESTABLISHED data up to user socket.
	 */
	if (ti != (struct tcpiphdr *)NULL) {
		/*
		 * Find a segment which begins after this one does.
		 */

		for (q = tp->seg_next; q != (struct tcpiphdr *)tp;
		     q = (struct tcpiphdr *)q->ti_next)
			if (SEQ_GT(q->ti_seq, ti->ti_seq))
				break;

		/*
		 * If there is a preceding segment, it may provide some of
		 * our data already.  If so, drop the data from the incoming
		 * segment.  If it provides all of our data, drop us.
		 */

		if ((struct tcpiphdr *)q->ti_prev != (struct tcpiphdr *)tp) {
			register int i;
			q = (struct tcpiphdr *)q->ti_prev;

			/* conversion to int (in i) handles seq wraparound */

			i = q->ti_seq + q->ti_len - ti->ti_seq;
			if (i > 0) {
				if (i >= ti->ti_len) {
					m_freem(dtom(ti));
					return (0);
				}
				m_adj(dtom(ti), i);
				ti->ti_len -= i;
				ti->ti_seq += i;
			}
			q = (struct tcpiphdr *)(q->ti_next);
		}

		/*
		 * While we overlap succeeding segments trim them or,
		 * if they are completely covered, dequeue them.
		 */

		while (q != (struct tcpiphdr *)tp) {
			register int i = (ti->ti_seq + ti->ti_len) - q->ti_seq;
			if (i <= 0)
				break;
			if (i < q->ti_len) {
				q->ti_seq += i;
				q->ti_len -= i;
				m_adj(dtom(q), i);
				break;
			}
			q = (struct tcpiphdr *)q->ti_next;
			m = dtom(q->ti_prev);
			remque(q->ti_prev);
			m_freem(m);
		}
		tcpexstat.tcpex_rcvpack++;
		tcpexstat.tcpex_rcvbyte += ti->ti_len;

		/*
		 * Stick new segment in its place.
		 */
		insque(ti, q->ti_prev);
	}

	/*
	 * Present data to user, advancing rcv_nxt through
	 * completed sequence space.
	 */

	if (TCPS_HAVERCVDSYN(tp->t_state) == 0)
		return (0);
	ti = tp->seg_next;
	if (ti == (struct tcpiphdr *)tp || ti->ti_seq != tp->rcv_nxt)
		return (0);
	if (tp->t_state == TCPS_SYN_RECEIVED && ti->ti_len)
		return (0);
	do {
		tp->rcv_nxt += ti->ti_len;
		flags = ti->ti_flags & TH_FIN;
		remque(ti);
		m = dtom(ti);
		ti = (struct tcpiphdr *)ti->ti_next;
		if (so->so_state & SS_CANTRCVMORE)
			m_freem(m);
		else {
			sbappend(&so->so_rcv, m);
		}
	} while (ti != (struct tcpiphdr *)tp && ti->ti_seq == tp->rcv_nxt);

	sorwakeup(so);
	return (flags);
}

/* respond with RST */

tcp_respond_reset(tp, ti, tiflags)
	struct tcpcb * tp;
	struct tcpiphdr * ti;
	int	tiflags;
{

	/*
	 * tcp_respond uses mbufs at ti to respond.
	 * It does an m_freem() to free
	 * the rest of the mbuf chain headed by m.
	 *
	 * tp == NULL
	 */

	if (tiflags & TH_RST) {
		m_freem(dtom(ti));
		return;
	}
	if (tiflags & TH_ACK)

		/* if ACK then just RST */

		tcp_respond(tp, ti, (tcp_seq)0, ti->ti_ack, TH_RST);
	else {

		/* Not ACK so ACK and RST */

		if (tiflags & TH_SYN)
			ti->ti_len++;

		tcp_respond(tp, ti, ti->ti_seq+ti->ti_len,
				(tcp_seq)0, TH_RST|TH_ACK);
	}

	return;
}

/* ARGSUSED */
tcp_ackit(ti, tlen, tiflags, tp, ostate)
	struct tcpiphdr * ti;
	int tlen;
	int tiflags;
	struct tcpcb * tp;
	int ostate;
{


	/*
	 * Generate an ACK dropping incoming segment if it occupies
	 * sequence space, where the ACK reflects our state.
	 */

	if (tiflags&TH_RST) { /* do not ack an out of window RST */
		m_freem(dtom(ti));
		return;
	}
			
	tcp_respond(tp, ti, tp->rcv_nxt, tp->snd_nxt, TH_ACK);


	if (tp->t_inpcb->inp_socket->so_options & SO_DEBUG)
		tcp_trace(TA_RESPOND, ostate, tp, &tcp_saveti, 0);

	return;
}

tcp_drop_locks(saved_sopp, saved_partial_sopp)
	struct socket_peer * saved_sopp;
	struct socket_peer * saved_partial_sopp;
{
	if (saved_sopp) {
		if (--(saved_sopp->sop_refcnt) == 0)
			(void) m_free(dtom(saved_sopp));
		else {
			v_lock(&saved_sopp->sop_lock, SPLNET);
		}
	}
	if (saved_partial_sopp) {
		if (--(saved_partial_sopp->sop_refcnt) == 0)
			(void) m_free(dtom(saved_partial_sopp));
		else {
			v_lock(&saved_partial_sopp->sop_lock, SPLNET);
		}
	}
	return;
}

/*
 *  Determine a reasonable value for maxseg size.
 *  If the route is known, use one that can be handled
 *  on the given interface without forcing IP to fragment.
 *  If bigger than a page (CLBYTES), round down to nearest pagesize
 *  to utilize pagesize mbufs.
 *  If interface pointer is unavailable, or the destination isn't local,
 *  use a conservative size (512 or the default IP max size, but no more
 *  than the mtu of the interface through which we route),
 *  as we can't discover anything about intervening gateways or networks.
 *
 *  This is ugly, and doesn't belong at this level, but has to happen somehow.
 */
tcp_mss(tp)
	register struct tcpcb *tp;
{
	struct route *ro;
	struct ifnet *ifp;
	int mss;
	struct inpcb *inp;

	inp = tp->t_inpcb;
	ro = &inp->inp_route;
	if ((ro->ro_rt == (struct rtentry *)0) ||
	    (ifp = ro->ro_rt->rt_ifp) == (struct ifnet *)0) {
		/* No route yet, so try to acquire one */
		if (inp->inp_faddr.s_addr != INADDR_ANY) {
			ro->ro_dst.sa_family = AF_INET;
			((struct sockaddr_in *) &ro->ro_dst)->sin_addr =
				inp->inp_faddr;
			rtalloc(ro);
		}
		if ((ro->ro_rt == 0) || (ifp = ro->ro_rt->rt_ifp) == 0)
			return (TCP_MSS);
	}

	mss = ifp->if_mtu - sizeof(struct tcpiphdr);

	if (in_localaddr(inp->inp_faddr))	 /* on local net? */
		return (mss);
	else
		return (MIN(mss, TCP_MSS));
}
