/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: tcp_timer.c 2.5 89/08/14 $";
#endif

/*
 * tcp_timer.c
 *
 */

/* $Log:	tcp_timer.c,v $
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/mutex.h"
#include "../machine/gate.h"
#include "../machine/intctl.h"
#include "../h/mbuf.h"
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/protosw.h"
#include "../h/errno.h"

#include "../net/if.h"
#include "../net/route.h"

#include "../netinet/in.h"
#include "../netinet/in_pcb.h"
#include "../netinet/in_systm.h"
#include "../netinet/ip.h"
#include "../netinet/ip_var.h"
#include "../netinet/tcp.h"
#include "../netinet/tcp_fsm.h"
#include "../netinet/tcp_seq.h"
#include "../netinet/tcp_timer.h"
#include "../netinet/tcp_var.h"
#include "../netinet/tcpip.h"

struct tcpcb* timer_tcp_close();
struct tcpcb* timer_tcp_drop();

/*
 * Fast timeout routine for processing delayed acks
 * delayed acks allow the tcp engine to wait in order to "piggy
 * back" an acknowledgement onto a subsequent data packet by
 * having it wait a short while before acking.  When a data
 * packet comes along, we ack.
 */

tcp_fasttimo()
{
	register struct inpcb *inp;
	register struct tcpcb *tp;
	register spl_t splevel;

	/*
	 * Search through tcb's and update active timers.
	 * NOTE: list stays locked for entire scan!!
	 * this blocks users from insque/remque and
	 * tcp_input from looking for an entry.
	 */

	splevel = INP_LOCK(&tcb);	/* lock the tcb list */
	
	inp = tcb.inp_next;

	if (inp)

		/*
		 * inp->ip_next cannot change while the list
		 * lock is held!!
		 */

	   for (; inp != &tcb; inp = inp->inp_next) {

		/*
		 * use a conditional p_lock.  This is cheaper than spinning,
		 * and is OK since generally, ownership of the lock implies
		 * state change with respect to the timers.  (This might not
		 * be the case in extreme cases where there are many sibling
		 * processes sharing the socket and colliding on requests).
		 * NOTE BENE: inp_next cannot change while the list is
		 * locked.  This is done around insque/remque.
		 */

	   if(cp_lock(&inp->inp_sopp->sop_lock, SPLNET) != CPLOCKFAIL )
  	   {
		if(inp->inp_head) {	 /* not being detached */

			if ((tp = (struct tcpcb *)inp->inp_ppcb) &&
		    	(tp->t_flags & TF_DELACK)) {

				tp->t_flags &= ~TF_DELACK;
				tp->t_flags |= TF_ACKNOW;

				/*
				 * might not be too bad since no additional
				 * resources required and network output
				 * driver is simply behind a queue.
	 			 */

				(void) tcp_output(tp);
			}
		}
		INP_UNLOCK(inp, SPLNET);
	   }
	}	/* end for inp */

	INP_UNLOCK(&tcb, splevel);
}

/*
 * Tcp protocol timeout routine called every 500 ms.
 * Updates the timers in all active tcb's and
 * causes finite state machine actions if timers expire.
 */

tcp_slowtimo()
{
	register struct inpcb *ip, *ipnxt;
	register struct tcpcb *tp;
	register int i;

	register struct socket_peer *saved_sopp;
	spl_t splevel;

	/*
	 * Search through tcb's and update active timers.
	 * NOTE: list stays locked for entire scan!!
	 * this blocks users from insque/remque and
	 * tcp_input from looking for an entry.
	 */

	splevel = INP_LOCK(&tcb);	/* lock the tcb list */

	ip = tcb.inp_next;
	if (ip == (struct inpcb *) NULL) {	 /* shouldn't be possible */
		INP_UNLOCK(&tcb, splevel);
		return;
	}

	while (ip != &tcb) {

		/*
		 * use a conditional p_lock.  This is cheaper than spinning,
		 * and is OK since generally, ownership of the lock implies
		 * state change with respect to the timers.  (This might not
		 * be the case in extreme cases where there are many sibling
		 * processes sharing the socket and colliding on requests).
		 * NOTE BENE: inp_next cannot change without the list being
		 * locked.  This is done around insque/remque.
		 */

	   if(cp_lock(&ip->inp_sopp->sop_lock, SPLNET) != CPLOCKFAIL ) {

		/*
		 * note: ipnxt must be valid since the tcb list is LOCKED,
		 * therefore no remque/insque is permitted.
		 */

		ipnxt = ip->inp_next;
		saved_sopp = ip->inp_sopp;

		/*
		 * however, pcb can be in the process of being deleted
		 * *before* the inp_sopp->sop_lock is acquired above, in
		 * which case inp_head is set = NULL;
		 * note that ipnxt is still valid for next pcb since
		 * list LOCKED
		 */

		if(ip->inp_head == (struct inpcb *) NULL){

			/*
			 * nevermind, ip is on its way out
			 */

			goto tpgone;
		}

		/*
		 * this pcb could be in the list *before* a tcb is
		 * bound to it.
		 */

		tp = intotcpcb(ip);
		if (tp == (struct tcpcb *) NULL) { /* no tcb to process? */
			goto tpgone;
		}
		for (i = 0; i < TCPT_NTIMERS; i++) {
			if (tp->t_timer[i] && --tp->t_timer[i] == 0) {
				saved_sopp->sop_refcnt++; /* avoid deletion */
				(void) tcp_usrreq(tp->t_inpcb->inp_socket,
				    PRU_SLOWTIMO, (struct mbuf *) NULL,
				    (struct mbuf *)i, (struct mbuf *) NULL);

				/*
				 * if refcnt goes to zero => this slowtimo
				 * caused inp to be deleted, inp_detach has
				 * left the sop around because of the refcnt
				 * so release it.  Do not INP_UNLOCK cause
				 * inp is gone. socket_peer must be released.
				 * if pcb is deleted, then it is remqued and
				 * ipnext->inp_prev is updated.
				 * NOTE: inp_detach does not imply that this
				 * is the last reference to the socket_peer
				 * since socket might still have reference
				 * on time out.  (there may be some
				 * superfluous refcnt'ing here).
				 */

				saved_sopp->sop_refcnt--;
				if(saved_sopp->sop_refcnt == 0) {
					(void) m_free(dtom(saved_sopp));
					goto nextip;
				}

				/*
				 * the pcb can be deleted, but there is still
				 * a reference to the socket_peer (e.g. by
				 * the socket structure waiting for the
				 * disconnect to complete).  It must be
				 * unlocked without using the inp. This is
				 * done at tpgone.  Note that this breaks
				 * out of the for loop that goes through the
				 * timers since there are no more
				 * timers to go through if the pcb is gone.
				 */

				if(ipnxt->inp_prev != ip){
					goto tpgone;
				}
			}
		}

		/*
		 * timers done, tcb not deleted so increment stats
		 */

		tp->t_idle++;
		if (tp->t_rtt)
			tp->t_rtt++;
tpgone:
		v_lock(&saved_sopp->sop_lock, SPLNET);
nextip:
		ip = ipnxt;

	   }else{	 /* cp_lock == CPLOCKFAILS */

		/*
		 * ip->ip_next cannot change while list lock is held!!
		 */

		ip = ip->inp_next;
	   }
	}

	tcp_iss += TCP_ISSINCR/PR_SLOWHZ;		/* increment iss */
	if ((int)tcp_iss < 0)
		tcp_iss = 0;				/* compat 4.2 */
	INP_UNLOCK(&tcb, splevel);
}

/*
 * Cancel all timers for TCP tp.
 */

tcp_canceltimers(tp)
	struct tcpcb *tp;
{
	register int i;

	for (i = 0; i < TCPT_NTIMERS; i++)
		tp->t_timer[i] = 0;
}

/*
 * TCP timer processing.  Called from tcp_usrreq.PRU_SLOWTIMO
 */

struct tcpcb *
tcp_timers(tp, timer)
	register struct tcpcb *tp;
	int timer;
{
	register int rto;

	switch (timer) {
	/*
	 * 2 MSL timeout in shutdown went off.  If we're closed but
	 * still waiting for peer to close and connection has been idle
	 * too long, or if 2MSL time is up from TIME_WAIT, delete connection
	 * control block.  Otherwise, check again in a bit.
	 */
	case TCPT_2MSL:
		if (tp->t_state != TCPS_TIME_WAIT &&
		    tp->t_idle <= TCPTV_MAXIDLE)
			tp->t_timer[TCPT_2MSL] = TCPTV_KEEP;
		else
			tp = timer_tcp_close(tp);
		break;

	/*
	 * Retransmission timer went off.  Message has not
	 * been acked within retransmit interval.  Back off
	 * to a longer retransmit interval and retransmit all
	 * unacknowledged messages in the window.
	 */

	case TCPT_REXMT:
		tp->t_rxtshift++;
		if (tp->t_rxtshift > TCP_MAXRXTSHIFT) {
			tcpexstat.tcpex_timeoutdrop++;
			tp = timer_tcp_drop(tp, ETIMEDOUT);
			break;
		}
		tcpexstat.tcpex_rexmttimeo++;
		if (tp->t_mrtt == 0)
			rto = tcp_default_rexmt*PR_SLOWHZ;
		else
			rto = (tp->t_mrtt >> TCP_AVGSHIFT)
				+ (tp->t_drtt >> (TCP_MDEVSHIFT-1));

		rto *= tcp_backoff[tp->t_rxtshift - 1];
		TCPT_RANGESET(tp->t_timer[TCPT_REXMT], rto,
			    TCPTV_MIN, TCPTV_MAX);
#ifdef nottested
		/*
		 * If losing, let the lower level know
		 * and try for a better route.
		 */
		if (tp->t_rxtshift >= TCP_MAXRXTSHIFT / 4 ||
		    rto >= 10 * PR_SLOWHZ)
			in_losing(tp->t_inpcb);
#endif
		tp->snd_nxt = tp->snd_una;
		/*
		 * If timing a segment in this window,
		 * and we have already gotten some timing estimate,
		 * stop the timer.
		 */
		if (tp->t_rtt && tp->t_mrtt)
			tp->t_rtt = 0;
		(void) tcp_output(tp);
		break;

	/*
	 * Persistence timer into zero window.
	 * Force a byte to be output, if possible.
	 */

	case TCPT_PERSIST:
		tcpexstat.tcpex_persisttimeo++;
		tcp_setpersist(tp);
		tp->t_force = 1;
		(void) tcp_output(tp);
		tp->t_force = 0;
		break;

	/*
	 * Keep-alive timer went off; send something
	 * or drop connection if idle for too long.
	 *
	 * note: change from 4.2/4.3 if keepalive goes off before establishing
	 *       a connection, then ignore it, since the connection attempt
	 *       will abort via the max retries.
	 *       
	 */

	case TCPT_KEEP:
		tcpexstat.tcpex_keeptimeo++;

		if ( (tp->t_state == TCPS_ESTABLISHED
		      || tp->t_state == TCPS_CLOSE_WAIT)
		    && (tp->t_inpcb->inp_socket->so_options & SO_KEEPALIVE)) {
			/*
			 * dead connection?
			 */
		    	if (tp->t_idle >= TCPTV_MAXIDLE)
				goto dropit;
			/*
			 * Saying tp->rcv_nxt-1 lies about what
			 * we have received, and by the protocol spec
			 * requires the correspondent TCP to respond.
			 * Saying tp->snd_una-1 causes the transmitted
			 * byte to lie outside the receive window; this
			 * is important because we don't necessarily
			 * have a byte in the window to send (consider
			 * a one-way stream!)
			 */
			tcpexstat.tcpex_keepprobe++;
			/*
			 * note: tcp_respond can fail due to no mbufs -
			 * in which case, the ip_output does not occur,
			 * and the ti data base does not get updated.
			 * protocol will recover or eventually time out.
			 */

			tcp_respond(tp,
			    tp->t_template, tp->rcv_nxt-1, tp->snd_una-1, 0);
		}
		tp->t_timer[TCPT_KEEP] = TCPTV_KEEP;
		break;
	dropit:
		tcpexstat.tcpex_keepdrops++;
		tp = timer_tcp_drop(tp, ETIMEDOUT);
		break;
	}
	return (tp);
}

/*
 * the routines timer_tcp_close, timer_tcp_drop, and lin_pcbdetach
 * have been added to enable timer to do everything
 * with the tcb list locked including *removing* tcb's from the tcb list.
 */

/*
 * Close a TCP control block:
 *	discard all space held by the tcp
 *	discard internet protocol block
 *	wake up any sleepers
 *
 *	called with the inpcb socket_peer held
 *	after discarding inpcb, UNLOCKING the socket_peer is inappropriate
 */

#define DODROP			/* for analysis of dropped packets */
#undef DODROP			/* for analysis of dropped packets */
#ifdef DODROP			/* defined in ../netinet/tcp_input.c */
extern short tcpdrops[64];		/* keep stats on drop codes */
#define DROPIT(x) tcpdrops[(x)]++;
#else
#define DROPIT(x) {;}
#endif DODROP

struct tcpcb *
timer_tcp_close(tp)
	register struct tcpcb *tp;
{
	register struct tcpiphdr *t;
	struct inpcb *inp = tp->t_inpcb;
	struct socket *so = inp->inp_socket;
	register struct mbuf *m;
	struct socket * saved_head;

	DROPIT(60);

	t = tp->seg_next;
	while (t != (struct tcpiphdr *)tp) {
		t = (struct tcpiphdr *)t->ti_next;
		m = dtom(t->ti_prev);

		/*
		 * since this tcb no longer exists and the inpcb is locked,
		 * nothing should be looking at this queue so mutex is not
		 * an issue here.
 		 */

		remque(t->ti_prev);
		m_freem(m);
	}

	if (tp->t_template)
		(void) m_free(dtom(tp->t_template));
	if (tp->t_tcpopt)
		(void) m_free(dtom(tp->t_tcpopt));
	inp->inp_ppcb = (caddr_t) NULL;
	soisdisconnected(so);

	if(so->so_head) {

		DROPIT(63);
		saved_head = so->so_head;
		(void) SOLOCK(so->so_head);

		/*
		 * This is a partial connection - either remque it to avoid
		 * simultaneous deletion, or leave it in the accept queue and
		 * let soclose eventually in_pcbdetach.  
		 *
		 * if not accepting connections
		 * then it is either closing or full
		 *      in either case, we will get processed eventually.
		 */
		if(!(so->so_head->so_options & SO_ACCEPTCONN))
			panic("tcp timer remove from closed accept queue");

		if(soqremque(so, 0)) 		/* removed it from so_q0 */
			so->so_sopp->sop_refcnt--;
		/* else on so_q, let accept() (or soclose) remove it */
		DROPIT(62);
		SOUNLOCK(saved_head, SPLNET);
	}

	(void) m_free(dtom(tp));

	/*
	 * note locked version of in_pcbdetach assumes that the
	 * tcb list is locked.
	 */

	lin_pcbdetach(inp);
	return ((struct tcpcb *) NULL);
}

/*
 * locked version of in_pcbdetach which assumes that the pcb list
 * (viz. tcb) is locked before calling.  The pcb's socket_peer
 * is also locked (by tcp_slowtimo).
 */

lin_pcbdetach(inp)
	struct inpcb *inp;
{
	struct socket		*so = inp->inp_socket;
	struct socket_peer	*saved_sopp = inp->inp_sopp;

	/*
	 * socket_peer is locked so can safely sever the bond to the pcb.
	 * check for simultaneous detach by checking inp_head
	 * if simultaneous detach, then other detach will free inp
	 * so don't do anything.
	 */

	if(inp->inp_head != (struct inpcb *) NULL) {
		so->so_pcb = (caddr_t) NULL;
		sofree(so);
		if (inp->inp_route.ro_rt)
			rtfree(inp->inp_route.ro_rt);
		inp->inp_head = (struct inpcb *) NULL;
		remque(inp);
		(void) m_free(dtom(inp));
	}

	/*
	 * NOTE: this can be the last reference if called from
	 * tcp_abort->timer_tcp_close of partial connection.
	 */

	if(--(saved_sopp->sop_refcnt) <= 0)
		(void) m_free(dtom(saved_sopp));
}

/*
 * Drop a TCP connection, reporting
 * the specified error.  If connection is synchronized,
 * then send a RST to peer.
 */

struct tcpcb *
timer_tcp_drop(tp, errno)
	register struct tcpcb *tp;
	int errno;
{
	struct socket *so = tp->t_inpcb->inp_socket;

	if (TCPS_HAVERCVDSYN(tp->t_state)) {
		tp->t_state = TCPS_CLOSED;
		(void) tcp_output(tp);
	}
	so->so_error = errno;
	return (timer_tcp_close(tp));
}
