/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*
 * $Header: tcp_timer.h 2.4 89/08/07 $
 */

/* $Log:	tcp_timer.h,v $
 */

/*
 * Definitions of the TCP timers.  These timers are counted
 * down PR_SLOWHZ times a second.
 */
#define	TCPT_NTIMERS	4

#define	TCPT_REXMT	0		/* retransmit */
#define	TCPT_PERSIST	1		/* retransmit persistence */
#define	TCPT_KEEP	2		/* keep alive */
#define	TCPT_2MSL	3		/* 2*msl quiet time timer */

/*
 * The TCPT_REXMT timer is used to force retransmissions.
 * The TCP has the TCPT_REXMT timer set whenever segments
 * have been sent for which ACKs are expected but not yet
 * received.  If an ACK is received which advances tp->snd_una,
 * then the retransmit timer is cleared (if there are no more
 * outstanding segments) or reset to the base value (if there
 * are more ACKs expected).  Whenever the retransmit timer goes off,
 * we retransmit all unacknowledged segments, and do an exponential
 * backoff on the retransmit timer.
 *
 * The TCPT_PERSIST timer is used to keep window size information
 * flowing even if the window goes shut.  If all previous transmissions
 * have been acknowledged (so that there are no retransmissions in progress),
 * and the window is shut, then we start the TCPT_PERSIST timer, and at
 * intervals send a single byte into the peers window to force him to update
 * our window information.  We do this at most as often as TCPT_PERSMIN
 * time intervals, but no more frequently than the current estimate of
 * round-trip packet time.  The TCPT_PERSIST timer is cleared whenever
 * we receive a window update from the peer.
 *
 * The TCPT_KEEP timer is used to keep connections alive.  If an
 * connection is idle (no segments received) for TCPTV_KEEP amount of time,
 * but not yet established, then we drop the connection.  If the connection
 * is established, then we force the peer to send us a segment by sending:
 *	<SEQ=SND.UNA-1><ACK=RCV.NXT><CTL=ACK>
 * This segment is (deliberately) outside the window, and should elicit
 * an ack segment in response from the peer.  If, despite the TCPT_KEEP
 * initiated segments we cannot elicit a response from a peer in TCPT_MAXIDLE
 * amount of time, then we drop the connection.
 */

/*
 * Time constants.
 */
#define	TCPTV_MSL	(30*PR_SLOWHZ)	/* max seg lifetime hah!*/
#define	TCPTV_KEEP	( 45*PR_SLOWHZ)		/* keep alive - 45 secs */
#define	TCPTV_PERSMIN	(  5*PR_SLOWHZ)		/* retransmit persistence */

#define	TCPTV_MAXIDLE	( 12*TCPTV_KEEP)	/* maximum allowable idle
						   time before drop conn */

#define	TCPTV_MIN	(  1*PR_SLOWHZ)		/* minimum allowable value */

/* ASSUMPTION:
 *    (TCPTV_MAX << TCP_AVGSHIFT) < max unsigned short (2^16) 
 * because t_mrtt is a u_short and tcpcb must fit in an mbuf
 */
#define	TCPTV_MAX	(5*60*PR_SLOWHZ)	/* maximum allowable value */


#define	TCP_LINGERTIME	120			/* linger at most 2 minutes */

#define	TCP_MAXRXTSHIFT	12			/* maximum retransmits */

#ifdef	TCPTIMERS
char *tcptimers[] =
    { "REXMT", "PERSIST", "KEEP", "2MSL" };
#endif

/*
 * Retransmission smoothing constants.
 * Round trip time statistics are updated by each time a new value of
 * tp->t_rtt is available.   (see tcp_input.c)
 */
#define TCP_AVGSHIFT 3
#define TCP_MDEVSHIFT 2

/*
 * Force a time value to be in a certain range.
 */
#define	TCPT_RANGESET(tv, value, tvmin, tvmax) { \
	(tv) = (value); \
	if ((tv) < (tvmin)) \
		(tv) = (tvmin); \
	if ((tv) > (tvmax)) \
		(tv) = (tvmax); \
}


#ifdef KERNEL
/*
  Tunable configuration parameters set in conf_net.c
 */
extern int tcp_ttl;		/* default time to live */
extern int tcp_default_rexmt;	/* default retransmit time */
extern int tcp_backoff[];	/* tcp retransmit backoff table */
#endif
