/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*
 * $Header: tcp_var.h 2.3 89/08/14 $
 */

/* $Log:	tcp_var.h,v $
 */

/*
 * Kernel variables for tcp.
 */

/*
 * Tcp control block, one per tcp; fields:
 */
struct tcpcb {
	struct	tcpiphdr *seg_next;	/* sequencing queue */
	struct	tcpiphdr *seg_prev;
	short	t_state;		/* state of this connection */
	short	t_timer[TCPT_NTIMERS];	/* tcp timers */
	short	t_rxtshift;		/* log(2) of rexmt exp. backoff */
	struct	mbuf *t_tcpopt;		/* tcp options */
	u_short	t_maxseg;		/* maximum segment size */
	struct	mbuf *t_ipopt;		/* ip options XXX (temp) */
	char	t_force;		/* 1 if forcing out a byte */
	u_char	t_flags;
#define	TF_ACKNOW	0x01			/* ack peer immediately */
#define	TF_DELACK	0x02			/* ack, but try to delay it */
#define	TF_DONTKEEP	0x04			/* don't use keep-alives XXX */
#define	TF_NODELAY	0x04		/* don't delay packets to coalesce */
#define	TF_NOOPT	0x08			/* don't use tcp options */
#define	TF_SENTFIN	0x10		/* have sent FIN */
	struct	tcpiphdr *t_template;	/* skeletal packet for transmit */
	struct	inpcb *t_inpcb;		/* back pointer to internet pcb */
/*
 * The following fields are used as in the protocol specification.
 * See RFC783, Dec. 1981, page 21.
 */
/* send sequence variables */
	tcp_seq	snd_una;		/* send unacknowledged */
	tcp_seq	snd_nxt;		/* send next */
	tcp_seq	snd_up;			/* send urgent pointer */
	tcp_seq	snd_wl1;		/* window update seg seq number */
	tcp_seq	snd_wl2;		/* window update seg ack number */
	tcp_seq	iss;			/* initial send sequence number */
	u_short	snd_wnd;		/* send window */
/* receive sequence variables */
	u_short	rcv_wnd;		/* receive window */
	tcp_seq	rcv_nxt;		/* receive next */
	tcp_seq	rcv_up;			/* receive urgent pointer */
	tcp_seq	irs;			/* initial receive sequence number */
/*
 * Additional variables for this implementation.
 */
/* receive variables */
	tcp_seq	rcv_adv;		/* advertised window */
/* retransmit variables */
	tcp_seq	snd_max;		/* highest sequence number sent
					   used to recognize retransmits */
/* congestion control (for source quench) */
	u_short	snd_cwnd;		/* congestion-controlled window */
/* transmit timing stuff */
	short	t_idle;			/* inactivity time */
	short	t_rtt;			/* round trip time */
	u_short max_rcvd;		/* most peer has sent into window */
	tcp_seq	t_rtseq;		/* sequence number being timed */

/* new round trip time estimator */
	u_short	t_mrtt;			/* mean rtt of the connection */
	u_short t_drtt;			/* mean deviance of rtt */

	u_short	max_sndwnd;		/* largest window peer has offered */
/* out-of-band data */
	char	t_oobflags;		/* have some */
	char	t_iobc;			/* input character */
#define	TCPOOB_HAVEDATA	0x01
#define	TCPOOB_HADDATA	0x02
};

#define	intotcpcb(ip)	((struct tcpcb *)(ip)->inp_ppcb)
#define	sototcpcb(so)	(intotcpcb(sotoinpcb(so)))

struct	tcpstat {
	int	tcps_badsum;
	int	tcps_badoff;
	int	tcps_hdrops;
	int	tcps_badsegs;
	int	tcps_unack;
};

/*
 * extended tcp statistics
*/

struct tcpexstat {
	u_long	tcpex_connects;		/* connections established */
	u_long	tcpex_connattempt;	 /* connections initiated */
	u_long  tcpex_accepts;	   	/* connections accepted */
	u_long 	tcpex_drops;		/* conenctions dropped */
	u_long  tcpex_conndrops;	/* embryonic connections dropped */
	u_long  tcpex_closed;		/* conn. closed (with drops) */
	u_long	tcpex_timeoutdrop;	/* conn. dropped in rxmt timeout */
	u_long  tcpex_rexmttimeo;	/* rexmt timeouts */
	u_long	tcpex_persisttimeo;	/* persist timeouts */
	u_long  tcpex_keeptimeo;	/* keepalive timeouts */
	u_long	tcpex_keepprobe;	/* keepalive probes sent */
	u_long  tcpex_keepdrops;	/* conns dropped in keepalive */

	u_long  tcpex_sndtotal;		/* total packets sent */
	u_long	tcpex_sndpack;		/* data packets sent */
	u_long	tcpex_sndbyte;		/* data bytes sent */
	u_long  tcpex_sndprobe;		/* probes sent */
	u_long	tcpex_sndrexmitpack;	/* data packets retrans */
	u_long  tcpex_sndrexmitbyte;	/* data bytes retrans */
	
	u_long  tcpex_rcvtotal;		/* total packets received */
	u_long  tcpex_rcvpack;		/* packets recieved in sequence */
	u_long  tcpex_rcvbyte;		/* bytes recieved in sequence */
	
	};
#ifdef KERNEL
extern struct	inpcb tcb;		/* head of queue of active tcpcb's */
extern struct	tcpstat tcpstat;	/* tcp statistics */
extern struct   tcpexstat tcpexstat;	/* extended tcp statistics */
struct	tcpiphdr *tcp_template();
struct	tcpcb *tcp_close(), *tcp_drop();
struct	tcpcb *tcp_timers(), *tcp_disconnect(), *tcp_usrclosed();

extern	int	tcp_sendspace;		/* 4.2 sets to 2K */
extern	int	tcp_recvspace;
extern	int	tcpnodelack;
extern	int	tcpcksumi;		/* 0 => no tcpcksum on input */
extern	int	tcpcksumo;		/* 0 => no tcpcksum on input */
extern  tcp_seq tcp_iss;		/* initial sequent # in tcp_subr.c */
#endif
