/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: kern_descrip.c 2.10 87/04/17 $";
#endif

/*
 * kern_descrip.c
 *	Descriptor management.
 *
 * TODO:
 *	eliminate u.u_error side effects
 */

/* $Log:	kern_descrip.c,v $
 */

#include "../h/param.h"
#include "../h/systm.h"
#include "../h/mutex.h"
#include "../h/user.h"
#include "../h/kernel.h"
#include "../h/vfs.h"
#include "../h/vnode.h"
#include "../h/proc.h"
#include "../h/buf.h"
#include "../h/conf.h"
#include "../h/file.h"
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/stat.h"
#include "../h/flock.h"
#include "../h/vm.h"

#include "../h/ioctl.h"

#include "../machine/pte.h"
#include "../machine/gate.h"
#include "../machine/intctl.h"

/*
 * getdtablesize()
 *	Return max number of per-process file-descriptors.
 *
 * Also return high-water mark (highest in-use file-descriptor index).
 */

getdtablesize()
{
	u.u_r.r_val1 = u.u_nofile;
	u.u_r.r_val2 = u.u_lastfile;
}

/*
 * dup()
 *	System call.  Duplicate a given file-descriptor.
 */

dup()
{
	register struct a {
		int	i;
	} *uap = (struct a *) u.u_ap;
	register struct ofile *ofp;
	int	j;

	GET_OFILE(ofp, uap->i);
	j = ufalloc(0);
	if (j < 0)
		return;
	dupit(j, ofp->of_file, ofp->of_flags &~ UF_EXCLOSE);
}

/*
 * dup2()
 *	System call.  Duplicate a given file-descriptor to given file #.
 */

dup2()
{
	register struct a {
		int	i, j;
	} *uap = (struct a *) u.u_ap;
	register struct ofile *ofp;

	GET_OFILE(ofp, uap->i);

	if (uap->j < 0 || uap->j >= u.u_nofile) {
		u.u_error = EBADF;
		return;
	}

	u.u_r.r_val1 = uap->j;
	if (uap->i == uap->j)
		return;

	dupit(uap->j, ofp->of_file, ofp->of_flags &~ UF_EXCLOSE);
}

/*
 * dupit()
 *	Record the dup(), closing target if necessary.
 *
 * dup() always calls with fd referencing a null fp; dup2() may call
 * with in-use fd.
 */

dupit(fd, fp, flags)
	register struct	file *fp;
	int	fd;
	int	flags;
{
	register struct ofile *ofp = &u.u_ofile[fd];

	if (ofp->of_file) {
		struct	file	*oldfp;
		/*
		 * Must zap u_ofile[] before call closef(); see close().
		 * Unmap if was mapped.
		 */
		if (ofp->of_flags & UF_MAPPED)
			munmapfd(fd, 0);
		oldfp = ofp->of_file;
		ofp->of_file = NULL;
		closef(oldfp);
		if (u.u_error)
			return;
		ofp = &u.u_ofile[fd];		/* in case u_ofile moved */
	}
	ofp->of_file = fp;
	ofp->of_flags = flags;
	FDBUMP(fp);
	if (fd > u.u_lastfile)
		u.u_lastfile = fd;
}

/*
 * fcntl()
 *	The file control system call.
 */

fcntl()
{
	register struct a {
		int	fdes;
		int	cmd;
		int	arg;
	} *uap = (struct a *) u.u_ap;
	register struct ofile *ofp;
	register struct file *fp;
	register i;
	struct	flock	bf;
	int	value;
	spl_t	s_ipl;

	GET_OFILE(ofp, uap->fdes);
	fp = ofp->of_file;

	switch(uap->cmd) {
	case F_DUPFD:
		i = uap->arg;
		if (i < 0 || i >= u.u_nofile) {
			u.u_error = EINVAL;
			return;
		}
		if ((i = ufalloc(i)) < 0)
			return;
		dupit(i, fp, ofp->of_flags &~ UF_EXCLOSE);
		break;

	case F_GETFD:
		u.u_r.r_val1 = ofp->of_flags & 1;
		break;

	case F_SETFD:
		ofp->of_flags = (ofp->of_flags &~ 1) | (uap->arg & 1);
		break;

	case F_GETFL:
		u.u_r.r_val1 = fp->f_flag + FOPEN;
		break;

	case F_SETFL:
		/*
		 * The file lock is held until all operations completed. This
		 * means that the lower level code must not block on a 
		 * semaphore. In the case of FIONBIO and FIOASYNC this is
		 * true. This operation is kept atomic to avoid race with
		 * concurrent F_SETFL.
		 */
		s_ipl = FDLOCK(fp);
		fp->f_flag &= FCNTLCANT;
		fp->f_flag |= (uap->arg-FOPEN) &~ FCNTLCANT;
		value = fp->f_flag & FNDELAY;
		u.u_error = (*fp->f_ops->fo_ioctl)(fp, FIONBIO, (caddr_t)&value);
		if (u.u_error) {
			FDUNLOCK(fp, s_ipl);
			break;
		}
		value = fp->f_flag & FASYNC;
		u.u_error = (*fp->f_ops->fo_ioctl)(fp, FIOASYNC, (caddr_t)&value);
		if (u.u_error) {
			value = 0;
			(void)(*fp->f_ops->fo_ioctl)(fp, FIONBIO, (caddr_t)&value);
		}
		FDUNLOCK(fp, s_ipl);
		break;

	case F_GETOWN:
		u.u_error = fgetown(fp, &u.u_r.r_val1);
		break;

	case F_SETOWN:
		u.u_error = fsetown(fp, uap->arg);
		break;

	case F_GETLK:
		/* get record lock */
		if (copyin((caddr_t)uap->arg, (caddr_t)&bf, sizeof bf))
			u.u_error = EFAULT;
		else if ((i = getflck(fp, &bf)) != 0)
			u.u_error = i;
		else if (copyout((caddr_t)&bf, (caddr_t)uap->arg, sizeof bf))
			u.u_error = EFAULT;
		break;

	case F_SETLK:
		/* set record lock and return if blocked */
		if (copyin((caddr_t)uap->arg, (caddr_t)&bf, sizeof bf))
			u.u_error = EFAULT;
		else if ((i = setflck(fp, &bf, 0)) != 0)
			u.u_error = i;
		break;

	case F_SETLKW:
		/* set record lock and wait if blocked */
		if (copyin((caddr_t)uap->arg, (caddr_t)&bf, sizeof bf))
			u.u_error = EFAULT;
		else if ((i = setflck(fp, &bf, 1)) != 0)
			u.u_error = i;
		break;

	default:
		u.u_error = EINVAL;
	}
}

/*
 * fset()
 *	Called by ioctl to set/clear FIONBIO and FIOASYNC.
 *
 * see fcntl() FSETFL case.
 */

fset(fp, bit, value)
	register struct file *fp;
	int	bit;
	int	value;
{
	spl_t	s_ipl;
	int	retval;

	s_ipl = FDLOCK(fp);

	if (value)
		fp->f_flag |= bit;
	else
		fp->f_flag &= ~bit;
	retval = fioctl(fp, (int)(bit == FNDELAY ? FIONBIO : FIOASYNC),
			(caddr_t)&value);

	FDUNLOCK(fp, s_ipl);

	return(retval);
}

fgetown(fp, valuep)
	struct	file	*fp;
	int	*valuep;
{
	int	error;

	switch (fp->f_type) {

	case DTYPE_SOCKET:
		*valuep = ((struct socket *)fp->f_data)->so_pgrp;
		return(0);

	default:
		error = fioctl(fp, (int)TIOCGPGRP, (caddr_t)valuep);
		*valuep = -*valuep;
		return(error);
	}
}

fsetown(fp, value)
	register struct file *fp;
	int	value;
{
	if (fp->f_type == DTYPE_SOCKET) {
		((struct socket *)fp->f_data)->so_pgrp = value;
		return(0);
	}
	if (value > 0) {
		register struct proc *p;
		/*
		 * If process found, pfind locks its p_state.
		 */
		p = pfind(value);
		if (p == 0)
			return (EINVAL);
		value = p->p_pgrp;
		/*
		 * Free lock and back to base level.
		 */
		v_lock(&p->p_state, SPL0);
	} else
		value = -value;
	return(fioctl(fp, (int)TIOCSPGRP, (caddr_t)&value));
}

static
fioctl(fp, cmd, value)
	struct	file	*fp;
	int	cmd;
	caddr_t	value;
{
	return((*fp->f_ops->fo_ioctl)(fp, cmd, value));
}

/*
 * close()
 *	`close' system call.  Close a given file-descriptor.
 */

close()
{
	register struct a {
		int	i;
	} *uap = (struct a *) u.u_ap;
	register struct ofile *ofp;
	struct file *fp;

	GET_OFILE(ofp, uap->i);
	fp = ofp->of_file;

	/*
	 * Unmap it if currently mapped.
	 */

	if (ofp->of_flags & UF_MAPPED)
		munmapfd(uap->i, 0);

	/*
	 * Zap the ofile[] entry first, in case close is to driver that
	 * blocks, and signal wakes us out; closef() arranges file-table
	 * entry to be freed (if this is last ref) before calling driver
	 * close, and don't want to think entry is still active.
	 * This has implications on driver close procedures that allow
	 * signals to interrupt them.
	 */

	ofp->of_file = NULL;

	/*
	 * If just closed last open file descriptor, adjust high-water mark.
	 * Do this before calling closef() in case a driver close procedure
	 * allows process to swap (u_ofile might move).
	 */

	ASSERT_DEBUG(uap->i <= u.u_lastfile, "close: u_lastfile");

	if (uap->i == u.u_lastfile) do {
		--u.u_lastfile;
		--ofp;
	} while (u.u_lastfile >= 0 && ofp->of_file == NULL);

	/*
	 * Finally close the file-table entry.
	 */

	closef(fp);
	/* WHAT IF u.u_error ? */
}

/*
 * fstat()
 *	System-call.  Does a "stat" on an already open file.
 */

fstat()
{
	register struct file *fp;
	register struct a {
		int	fdes;
		struct	stat *sb;
	} *uap = (struct a *) u.u_ap;
	struct stat ub;

	fp = getf(uap->fdes);
	if (fp == 0)
		return;

	switch (fp->f_type) {

	case DTYPE_VNODE:
		/*
		 * Need locked vnode to insure consistent data.
		 */
		VN_LOCKNODE((struct vnode *)fp->f_data);
		u.u_error = vno_stat((struct vnode *)fp->f_data, &ub);
		VN_UNLOCKNODE((struct vnode *)fp->f_data);
		break;

	case DTYPE_SOCKET:
		if (u.u_tuniverse == U_ATT) {
 			att_pstat(&ub, fp);
			break;
		}
		u.u_error = soo_stat((struct socket *)fp->f_data, &ub);
		break;

	default:
		panic("fstat");
		/*NOTREACHED*/
	}
	if (u.u_error == 0)
		u.u_error = copyout((caddr_t)&ub, (caddr_t)uap->sb, sizeof(ub));
}

/*
 * ufalloc()
 *	Allocate a user file descriptor.
 */

ufalloc(i)
	register int	i;	/* smallest desired file-descriptor */
{
	register struct ofile *ofp;
	register struct ofile *lim = &u.u_ofile[u.u_nofile];

	for (ofp = &u.u_ofile[i]; ofp < lim; ofp++, i++) {
		if (ofp->of_file == NULL) {
			ofp->of_flags = 0;
			u.u_r.r_val1 = i;
			if (i > u.u_lastfile)
				u.u_lastfile = i;
			return(i);
		}
	}

	u.u_error = EMFILE;
	return(-1);
}

/*
 * finit()
 *	Init the file-table and file entries.
 *
 * Called in main() when the rest of the world is comming alive.
 */

lock_t	file_list;			/* to lock the list */
int	ffree_head;			/* head of free-list */

finit()
{
	register struct file *fp;
	register int	gateno;

	init_lock(&file_list, G_FILELIST);
	ffree_head = -1;			/* end of list indicator */
	gateno = 0;
	for (fp = file; fp < fileNFILE; fp++) {
		fp->f_next = ffree_head;
		ffree_head = fp - file;
		init_lock(&fp->f_mutex,
		       (gate_t)(G_FILMIN + (gateno++ % (G_FILMAX-G_FILMIN+1))));
	}
#ifdef	lint
	lint_ref_int(gateno);
#endif	lint

	/*
	 * Also initialize callers open-file table.
	 */

	ofile_ext_init();
}

/*
 * falloc()
 *	Allocate a user file descriptor and a file structure.
 *
 * Initialize the descriptor to point at the file structure.
 */

struct file *
falloc()
{
	register struct file *fp;
	register i;
	spl_t	s;

	i = ufalloc(0);
	if (i < 0)
		return (NULL);

	/*
	 * Lock file-list for allocation.
	 * If there is a free element, take it; else bitch.
	 */

	s = p_lock(&file_list, SPLFS);
	if (ffree_head < 0) {
		v_lock(&file_list, s);
		tablefull("file");
		u.u_error = ENFILE;
		return (NULL);
	}
	fp = &file[ffree_head];
	ffree_head = fp->f_next;

	u.u_ofile[i].of_file = fp;
	fp->f_data = 0;
	fp->f_count = 1;

#ifdef	ns32000
	ASSERT_DEBUG(fp->f_mutex.l_state == L_UNLOCKED, "falloc: locked");
#endif	ns32000
#ifdef	i386
	ASSERT_DEBUG(fp->f_mutex == L_UNLOCKED, "falloc: locked");
#endif	i386

	v_lock(&file_list, s);
	fp->f_offset = 0;			/* ok to do outside list lock */
	crhold(u.u_cred);
	fp->f_cred = u.u_cred;
	return (fp);
}

/*
 * ffree()
 *	Free a file-node.
 *
 * Called various places (closef(), unwinding in copen()) to give back
 * a previously allocated file node.
 */

ffree(fp)
	register struct file *fp;
{
	struct	ucred	*cr;
	spl_t	s;

	s = p_lock(&file_list, SPLFS);

	fp->f_next = ffree_head;
	ffree_head = fp - file;
	fp->f_type = 0;
	fp->f_count = 0;		/* for cleanliness */
					/* Not redundant if copen() unwind */
	cr = fp->f_cred;
	v_lock(&file_list, s);

	crfree(cr);
}

/*
 * getf()
 *	Convert a user supplied file descriptor into a pointer
 *	to a file structure.
 *
 * Only task is to check range of the descriptor.
 * Critical paths should use the GETF macro.
 */

struct file *
getf(f)
	int	f;
{
	struct	file	*fp;

	if ((unsigned)f >= u.u_nofile || (fp = u.u_ofile[f].of_file) == NULL) {
		u.u_error = EBADF;
		return (NULL);
	}
	return (fp);
}

/*
 * closef()
 *	Internal form of close.
 *
 * Decrement reference count on file structure, closes underling object
 * if count hits zero.
 */

closef(fp)
	register struct file *fp;
{
	spl_t	s;

	if (fp == NULL)
		return;
	cleanlocks(fp);
	s = FDLOCK(fp);
	if (fp->f_count > 1) {
		fp->f_count--;
		FDUNLOCK(fp, s);
		return;
	}

	/*
	 * Last ref going away.  We can unlock, since nobody's going
	 * to play with it any more.  Note that the "fo_close" procedure
	 * must ffree() the file-node.
	 *
	 * Must also decrement the vnode writer count if appropriate.
	 */

	fp->f_count = 0;	   /* for forceclose() and cleanliness */

	FDUNLOCK(fp, s);

	(*fp->f_ops->fo_close)(fp);
}

/*
 * flock()
 *	Apply an advisory lock on a file descriptor.
 */

flock()
{
	register struct a {
		int	fd;
		int	how;
	} *uap = (struct a *) u.u_ap;
	register struct file *fp;

	fp = getf(uap->fd);
	if (fp == NULL)
		return;

	if (fp->f_type != DTYPE_VNODE) {
		u.u_error = EOPNOTSUPP;
		return;
	}

	if (((struct vnode *)fp->f_data)->v_type == VFIFO) {
		u.u_error = EINVAL;
 		return;
	}

	if (!(uap->how & (LOCK_UN | LOCK_EX | LOCK_SH))) {
		u.u_error = EINVAL;
		return;
	}
	if (uap->how & LOCK_UN) {
		vno_unlock(fp, FSHLOCK|FEXLOCK);
		return;
	}

	/*
	 * No lock needed on fp here.  It wouldn't avoid any
	 * races that aren't already there.
	 */

	if ((fp->f_flag & FEXLOCK) && (uap->how & LOCK_EX) ||
	    (fp->f_flag & FSHLOCK) && (uap->how & LOCK_SH))
		return;
	u.u_error = vno_lock(fp, uap->how);
}

/*
 * The remainder of this module implements open-file table extension objects.
 */

extern	int		nofile_ext;		/* # allocated entries */
struct	ofile_ext	*ofile_ext;		/* boot time allocated list */

static	struct	ofile_ext *ofile_ext_free_list;	/* free list */
static	lock_t	ofile_ext_mutex;		/* mutex on free list */

struct	ofile_ext *ofile_ext_alloc();

/*
 * SZ_OFILE_EXT() is size in bytes of an ofile_ext structure for nfds fd's.
 * SZ_OFILE_EXT_SWAP() is the # of swap disk blocks to hold a given # bytes.
 * OFILE_EXT_SWBYTES() is the # bytes to swap in/out.
 * OFILE_EXT_SWPTE() is pointer to pte mapping the extension data.
 */

#define	SZ_OFILE_EXT(nfds)	((nfds) * sizeof(struct ofile))
#define	SZ_OFILE_EXT_SWAP(sz)	((long) ctod(clrnd(btoc(sz))))

#define	OFILE_EXT_SWBYTES(ofe) \
		((SZ_OFILE_EXT((ofe)->oe_nofile) + CLOFSET) & ~CLOFSET)
#define	OFILE_EXT_SWPTE(ofe) \
		(&Usrptmap[btokmx((struct pte *)((ofe)->oe_ofile))])

#define	LOCK_OFILE_EXT_LIST(s)	(s) = p_lock(&ofile_ext_mutex, SPLFS)
#define	UNLOCK_OFILE_EXT_LIST(s) v_lock(&ofile_ext_mutex, (s))

#define	LOCK_OFILE_EXT(ofe)	p_sema(&(ofe)->oe_mutex, PSWP)
#define	UNLOCK_OFILE_EXT(ofe)	v_sema(&(ofe)->oe_mutex)

/*
 * ofile_ext_init()
 *	Initialize list of open-file table extension objects and
 *	initialize open-file table in 1st process.
 *
 * Statically allocated list of objects allows representation of shared
 * ofile_ext's with swappable big parts.  If didn't need to swap, could
 * avoid need to keep seperate (swappable) data, and possibly allocate
 * from a heap.
 *
 * Called during system initialization.
 */

static
ofile_ext_init()
{
	register struct ofile_ext *ofe;
	register int i;

	/*
	 * Arrange that 1st process uses u-area local open-file table.
	 */

	u.u_lastfile = -1;
	u.u_nofile = sizeof(u.u_lofile) / sizeof(u.u_lofile[0]);
	u.u_ofile = u.u_lofile;
	bzero((caddr_t) u.u_lofile, sizeof(u.u_lofile));
	u.u_ofile_ext = NULL;

	/*
	 * Initialize list of open-file table extension objects.
	 */

	init_lock(&ofile_ext_mutex, G_FILELIST);

	for (ofe = ofile_ext, i = 0; i < nofile_ext; i++, ofe++) {
		ofe->oe_next = ofile_ext_free_list;
		ofile_ext_free_list = ofe;
		init_sema(&ofe->oe_mutex, 1, 0,
					G_FILMIN + i % (G_FILMAX-G_FILMIN+1));
	}
}

/*
 * setdtablesize()
 *	System call to alter size of the calling process's descriptor table.
 *
 * An "open file" in a process is represented by a struct ofile (see h/user.h).
 * Each process has by default NOFILE struct ofile's.  If the process
 * needs more, it calls setdtablesize() to allocate an ofile_ext structure.
 *
 * Once an "extension" is in use, must be careful no to use pointers to
 * ofile entries if the process can block and swap -- the ofile table will
 * move when the process swaps back in.
 *
 * The current implementation doesn't support sharing of open-file tables;
 * thus oe_refcnt==1, and oe_mutex, of_inuse aren't actually used (yet).
 */

setdtablesize()
{
	struct a {
		int	size;			/* # fd's desired */
	} *uap = (struct a *) u.u_ap;
	register int	nfds;
	register int	inuse;
	register struct	ofile_ext *ofe;
	extern	int	max_nofile;		/* system-defined MAX */

	if ((nfds = uap->size) < 0) {
		u.u_error = EINVAL;
		return;
	}

	/*
	 * Insure nfds is no larger than system-defined max,
	 * but at least NOFILE.
	 *
	 * Impose absolute max number fd's to avoid problems in select()
	 * (must be able to fit 6 fd_set's in an 8k buffer).
	 * For MAXBSIZE=8192 and NFDBITS=32, this works out to 10912
	 * (should be enough for even the most bizzare application ;-)
	 * This is a strange reason for a limit.
	 */

	nfds = MIN(nfds, ((MAXBSIZE * NBBY) / 6) & ~(NFDBITS-1));
	nfds = MIN(nfds, max_nofile);
	nfds = MAX(nfds, NOFILE);

	/*
	 * If not changing size of table, return current size.
	 * If shrinking, close higher numbered fd's and set up smaller table.
	 * Else (growing), allocate and setup larger table.
	 */

	if (nfds == u.u_nofile) {
		u.u_r.r_val1 = u.u_nofile;
		return;
	}

	if (nfds < u.u_nofile) {			/* shrinking */
		/*
		 * First insure can get an extension object if need one.
		 * Could re-use existing one but it's more awkward.
		 */
		if (nfds > NOFILE && (ofe = ofile_ext_alloc(nfds)) == NULL)
			return;
		/*
		 * Close all open file-descriptors above new table size.
		 */
		while (u.u_lastfile >= nfds) {
			struct file *fp = u.u_ofile[u.u_lastfile].of_file;
			u.u_ofile[u.u_lastfile].of_file = NULL;
			closef(fp);
			--u.u_lastfile;
		}
		while (u.u_lastfile>=0 && u.u_ofile[u.u_lastfile].of_file==NULL)
			--u.u_lastfile;
		/*
		 * Optimization: if shrink back to NOFILE, re-use local set.
		 */
		if (nfds == NOFILE) {
			ASSERT_DEBUG(u.u_ofile_ext != NULL,
					"setdtablesize: shrink NULL ext");
			bcopy((caddr_t) &u.u_ofile[0], (caddr_t) &u.u_lofile[0],
							sizeof(u.u_lofile));
			u.u_ofile = u.u_lofile;
			u.u_nofile = nfds;
			ofile_ext_release(u.u_ofile_ext);
			u.u_ofile_ext = NULL;
			u.u_r.r_val1 = nfds;
			return;
		}
	} else if ((ofe = ofile_ext_alloc(nfds)) == NULL) {	/* growing */
		return;
	}

	/*
	 * Growing, or shrinking to more than NOFILE.
	 * Have already allocated an open-file table extension object,
	 * now allocate memory for the variable size parts.
	 */

	(void) ofile_ext_alloc_mem(ofe, 1);

	/*
	 * Copy active part of ofile table into new one, zap rest of new ofile
	 * table.  If shrinking, u_lastfile guaranteed < nfds from above.
	 */

	inuse = u.u_lastfile + 1;

	bcopy(	(caddr_t) &u.u_ofile[0],
		(caddr_t) &ofe->oe_ofile[0],
		(unsigned) (inuse * sizeof(struct ofile)));

	bzero(	(caddr_t) &ofe->oe_ofile[inuse],
		(unsigned) ((nfds - inuse) * sizeof(struct ofile)));

	/*
	 * Delete old extension (if any) and install new one.
	 */

	if (u.u_ofile_ext)
		ofile_ext_release(u.u_ofile_ext);

	u.u_ofile_ext = ofe;
	u.u_ofile = ofe->oe_ofile;
	u.u_nofile = nfds;
	u.u_r.r_val1 = nfds;
}

/*
 * ofile_ext_release()
 *	Release a reference to an open-file table extension object.
 *
 * Doesn't do anything with ofile entries -- assumes caller released
 * relevant file-table references.
 *
 * If ref-count goes to zero, put the object on the free-list.
 */

ofile_ext_release(ofe)
	register struct ofile_ext *ofe;
{
	LOCK_OFILE_EXT(ofe);

	if (--ofe->oe_refcnt == 0) {
		/*
		 * No cache so can't race with other allocates.
		 */
		UNLOCK_OFILE_EXT(ofe);
		ofile_ext_free_mem(ofe);
		ofile_ext_free(ofe);
	} else {
		/*
		 * If all other references are swapped out, write out
		 * the current copy.
		 */
		if (--ofe->oe_ccount == 0) {
			swap(u.u_procp, ofe->oe_swaddr,
				OFILE_EXT_SWPTE(ofe), OFILE_EXT_SWBYTES(ofe),
				B_WRITE|B_UAREA, swapdev_vp);
			ofile_ext_free_mem(ofe);
		}
		UNLOCK_OFILE_EXT(ofe);
	}
}

/*
 * ofile_ext_clone()
 *	Duplicate an open-file table extension object (during fork()).
 *
 * Copies ofile entries but does not bump ref-counts on file-tble entries;
 * assumes caller handles that (ala newproc()).
 *
 * Returns new ofile_ext or NULL for failure.
 */

struct ofile_ext *
ofile_ext_clone(ofe, ofilep)
	register struct ofile_ext *ofe;
	struct ofile **ofilep;
{
	register struct ofile_ext *nfe;

	ASSERT_DEBUG(ofe != NULL, "ofile_dup: null ofile_ext");

	nfe = ofile_ext_alloc(ofe->oe_nofile);
	if (nfe == NULL)
		return(NULL);

	(void) ofile_ext_alloc_mem(nfe, 1);

	bcopy(	(caddr_t) ofe->oe_ofile,
		(caddr_t) nfe->oe_ofile,
		(unsigned) (ofe->oe_nofile * sizeof(struct ofile)));

	*ofilep = &nfe->oe_ofile[0];
	return(nfe);
}

/*
 * ofile_ext_swapout()
 *	Swap out a reference to an open-file table extension structure.
 *
 * If this is last in-memory reference, actually write the structure to
 * swap space and delete memory copy of the object.
 */

ofile_ext_swapout(ofe)
	register struct ofile_ext *ofe;
{
	LOCK_OFILE_EXT(ofe);

	if (--ofe->oe_ccount == 0) {
		swap(u.u_procp, ofe->oe_swaddr,
				OFILE_EXT_SWPTE(ofe), OFILE_EXT_SWBYTES(ofe),
				B_WRITE|B_UAREA, swapdev_vp);
		ofile_ext_free_mem(ofe);
	}

	UNLOCK_OFILE_EXT(ofe);
}

/*
 * ofile_ext_swapin()
 *	Attempt to swap-in an open-file table extension object.
 *
 * Returns TRUE if succeeds, else FALSE (insufficient memory).
 * If succeeds, also sets u_ofile field in process being swapped in.
 */

ofile_ext_swapin(ofe, ofilep)
	register struct	ofile_ext *ofe;
	struct	ofile	**ofilep;
{
	LOCK_OFILE_EXT(ofe);

	/*
	 * If in-core count == 0, try to allocate memory for this object.
	 * Otherwise, just bump in-core count.
	 */

	if (ofe->oe_ccount == 0) {
		if (!ofile_ext_alloc_mem(ofe, 0)) {
			UNLOCK_OFILE_EXT(ofe);
			return(0);			/* can't get memory */
		}
		/*
		 * Read it back from swap space.
		 * ofile_ext_alloc_mem() already initialized the pointers.
		 */
		swap(u.u_procp, ofe->oe_swaddr,
				OFILE_EXT_SWPTE(ofe), OFILE_EXT_SWBYTES(ofe),
				B_READ|B_UAREA, swapdev_vp);
		ofe->oe_ccount = 1;
	} else
		++ofe->oe_ccount;

	UNLOCK_OFILE_EXT(ofe);

	*ofilep = &ofe->oe_ofile[0];
	return(1);				/* success! */
}

/*
 * ofile_ext_alloc()
 *	Allocate an open-file table extension object.
 *
 * Does not guarantee there will be memory to implement the object.
 *
 * Note: could allocate the ofile_ext from the "heap"; these algorithms
 * only use the linked-list to keep track of a boot-time allocated set
 * of structures -- there is no other need to have an array of ofile_ext's.
 *
 * Returns ofile_ext pointer or NULL if fail (and sets u.u_error).
 */

static struct ofile_ext *
ofile_ext_alloc(nfds)
	int	nfds;
{
	register struct ofile_ext *ofe;
	register spl_t s;
	swblk_t	swaddr;
	long	nblk = SZ_OFILE_EXT_SWAP(SZ_OFILE_EXT(nfds));

	swaddr = swapalloc(nblk);
	if (swaddr == 0) {
		u.u_error = ENOMEM;
		return(NULL);
	}

	LOCK_OFILE_EXT_LIST(s);
	if ((ofe = ofile_ext_free_list) == NULL) {
		UNLOCK_OFILE_EXT_LIST(s);
		swapfree(nblk, swaddr);
		tablefull("open-file extension");
		u.u_error = ENFILE;
		return(NULL);
	}
	ofile_ext_free_list = ofe->oe_next;
	UNLOCK_OFILE_EXT_LIST(s);

	ofe->oe_nofile = nfds;
	ofe->oe_refcnt = 1;
	ofe->oe_ccount = 1;
	ofe->oe_swaddr = swaddr;
	return(ofe);
}

/*
 * ofile_ext_free()
 *	Put an ofile_ext back on to free-list & release swap space.
 *
 * See comments in ofile_ext_alloc() on possible use of heap.
 */

static
ofile_ext_free(ofe)
	register struct ofile_ext *ofe;
{
	register spl_t	s;

	swapfree(SZ_OFILE_EXT_SWAP(SZ_OFILE_EXT(ofe->oe_nofile)), ofe->oe_swaddr);

	LOCK_OFILE_EXT_LIST(s);
	ofe->oe_next = ofile_ext_free_list;
	ofile_ext_free_list = ofe;
	UNLOCK_OFILE_EXT_LIST(s);
}

/*
 * ofile_ext_alloc_mem()
 *	Try to allocate memory for an open-file table extension object.
 *
 * Only caller passing must_succeed == FALSE is ofile_ext_swapin().
 */

static
ofile_ext_alloc_mem(ofe, must_succeed)
	register struct ofile_ext *ofe;
	bool_t	must_succeed;
{
	ofe->oe_ofile = (struct ofile *) wmemall(SZ_OFILE_EXT(ofe->oe_nofile),
								must_succeed);
	return(ofe->oe_ofile != NULL);
}

/*
 * ofile_ext_free_mem()
 *	Free the memory associated with an open-file table extension object.
 */

static
ofile_ext_free_mem(ofe)
	register struct ofile_ext *ofe;
{
	wmemfree((caddr_t) ofe->oe_ofile, SZ_OFILE_EXT(ofe->oe_nofile));
	ofe->oe_ofile = NULL;
}
