/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: /ogc5/dynix.3.0.17/src/dynix.3.0.17/src/sys/common/sys/RCS/kern_time.c,v 1.2 90/12/28 05:12:52 bruce Exp $";
#endif

/*
 * kern_time.c
 *	Time of day and interval timer support.
 *
 * These routines provide the kernel entry points to get and set
 * the time-of-day and per-process interval timers.  Subroutines
 * here provide support for adding and subtracting timeval structures
 * and decrementing interval timers, optionally reloading the interval
 * timers when they expire.
 */

#include "../machine/reg.h"

#include "../h/param.h"
#include "../h/mutex.h"
#include "../h/user.h"
#include "../h/kernel.h"
#include "../h/proc.h"

#include "../machine/intctl.h"
#include "../machine/gate.h"

gettimeofday()
{
	register struct a {
		struct	timeval *tp;
		struct	timezone *tzp;
	} *uap = (struct a *)u.u_ap;
	struct timeval atv;
	extern unsigned sec0eaddr;
	GATESPL(s_ipl);

	/*
	 * Return number of users to programs that limit # of users.
	 * (value in R1 on return from gettimeofday()).
	 */
	u.u_r.r_val2 = sec0eaddr;

	/*
	 * Get time vector. Gate used to avoid possible 1 second error.
	 * Is this really necessary? User may be context switched and
	 * thus may be seconds later before user code sees the time.
	 * For consistency and to eliminate the slight probability that
	 * a process calls gettimeofday twice and the second call returning
	 * what appears to be an earlier time, the gate will remain.
	 */
	P_GATE(G_TIME, s_ipl); atv = time; V_GATE(G_TIME, s_ipl);

	u.u_error = copyout((caddr_t)&atv, (caddr_t)uap->tp, sizeof (atv));
	if (u.u_error)
		return;
	if (uap->tzp == 0)
		return;
	/* SHOULD HAVE PER-PROCESS TIMEZONE */
	u.u_error = copyout((caddr_t)&tz, (caddr_t)uap->tzp, sizeof (tz));
}

settimeofday()
{
	register struct a {
		struct	timeval *tv;
		struct	timezone *tzp;
	} *uap = (struct a *)u.u_ap;
	struct timeval atv;
	struct timezone atz;
	GATESPL(s);

	u.u_error = copyin((caddr_t)uap->tv, (caddr_t)&atv,
		sizeof (struct timeval));
	if (u.u_error)
		return;
	setthetime(&atv);
	if (uap->tzp && suser()) {
		u.u_error = copyin((caddr_t)uap->tzp, (caddr_t)&atz,
			sizeof (atz));
		if (u.u_error == 0) {
			P_GATE(G_TIME, s); tz = atz; V_GATE(G_TIME, s);
		}
	}
}

static
setthetime(tv)
	struct timeval *tv;
{
	GATESPL(s);

	if (!suser())
		return;
/* WHAT DO WE DO ABOUT PENDING REAL-TIME TIMEOUTS??? */
	boottime.tv_sec += tv->tv_sec - time.tv_sec;
	P_GATE(G_TIME, s); time = *tv; V_GATE(G_TIME, s);
	resettodr();
}

extern	int tickadj;			/* "standard" clock skew, us./tick */
int	tickdelta;			/* current clock skew, us. per tick */
long	timedelta;			/* unapplied time correction, us. */
long	bigadj = 1000000;		/* use 10x skew above bigadj us. */

adjtime()
{
	register struct a {
		struct timeval *delta;
		struct timeval *olddelta;
	} *uap = (struct a *)u.u_ap;
	struct timeval atv, oatv;
	register long ndelta;
	GATESPL(s);

	if (!suser()) 
		return;
	u.u_error = copyin((caddr_t)uap->delta, (caddr_t)&atv,
		sizeof (struct timeval));
	if (u.u_error)
		return;
	ndelta = atv.tv_sec * 1000000 + atv.tv_usec;
	if (timedelta == 0)
		if (ndelta > bigadj)
			tickdelta = 10 * tickadj;
		else
			tickdelta = tickadj;
	if (ndelta % tickdelta)
		ndelta = ndelta / tickadj * tickadj;

	P_GATE(G_TIME, s);
	if (uap->olddelta) {
		oatv.tv_sec = timedelta / 1000000;
		oatv.tv_usec = timedelta % 1000000;
	}
	timedelta = ndelta;
	V_GATE(G_TIME, s);

	if (uap->olddelta)
		(void) copyout((caddr_t)&oatv, (caddr_t)uap->olddelta,
			sizeof (struct timeval));
}

/*
 * Get value of an interval timer.  The process virtual and
 * profiling virtual time timers are kept in the u. area, since
 * they can be swapped out.  These are kept internally in the
 * way they are specified externally: in time until they expire.
 *
 * The real time interval timer is kept in the process table slot
 * for the process, and its value (it_value) is kept as an
 * absolute time rather than as a delta, so that it is easy to keep
 * periodic real-time signals from drifting.
 *
 * Virtual time timers are processed in the hardclock() routine of
 * kern_clock.c.  The real time timer is processed by a timeout
 * routine, called from the softclock() routine.  Since a callout
 * may be delayed in real time due to interrupt processing in the system,
 * it is possible for the real time timeout routine (realitexpire, given below),
 * to be delayed in real time past when it is supposed to occur.  It
 * does not suffice, therefore, to reload the real timer .it_value from the
 * real time timers .it_interval.  Rather, we compute the next time in
 * absolute time the timer should go off.
 *
 * The gate G_TIME synchronizes with the todclock.
 * The splhi synchronizes with the local hardclock in the 
 * virtual timer and profiling timer.
 */
getitimer()
{
	register struct a {
		u_int	which;
		struct	itimerval *itv;
	} *uap = (struct a *)u.u_ap;
	struct itimerval aitv;
	spl_t s;

	if (uap->which > 2) {
		u.u_error = EINVAL;
		return;
	}
	if (uap->which == ITIMER_REAL) {
		/*
		 * Convert from absolute to relative time in .it_value
		 * part of real time timer.  If time for real time timer
		 * has passed return 0, else return difference between
		 * current time and time for the timer to go off.
		 */
		P_GATE(G_TIME, s);
		aitv = u.u_procp->p_realtimer;
		if (timerisset(&aitv.it_value)) {
			if (timercmp(&aitv.it_value, &time, <))
				timerclear(&aitv.it_value);
			else
				timevalsub(&aitv.it_value, &time);
		}
		V_GATE(G_TIME, s);
	} else {
		s = splhi();
		aitv = u.u_timer[uap->which];
		splx(s);
	}
	u.u_error = copyout((caddr_t)&aitv, (caddr_t)uap->itv,
	    sizeof (struct itimerval));
}

setitimer()
{
	register struct a {
		u_int	which;
		struct	itimerval *itv, *oitv;
	} *uap = (struct a *)u.u_ap;
	register struct proc *p = u.u_procp;
	struct	itimerval aitv, aotv;
	bool_t	sigalrm = 0;			/* should post a SIGALRM? */
	int	s;
	extern	lock_t	time_lck;		/* see untimeout */

	if (uap->which > 2) {
		u.u_error = EINVAL;
		return;
	}
	u.u_error = copyin((caddr_t)uap->itv, (caddr_t)&aitv,
						sizeof (struct itimerval));
	if (u.u_error)
		return;

	if (itimerfix(&aitv.it_value) || itimerfix(&aitv.it_interval)) {
		u.u_error = EINVAL;
		return;
	}

	if (uap->which == ITIMER_REAL) {
#ifdef	ns32000
		s = drp_lock(&time_lck);
#endif	ns32000
#ifdef	i386
		s = p_lock(&time_lck, SPLHI);
		VOID_P_GATE(G_TIME);
#endif	i386
		/*
		 * Get current value and save away.
		 */
		aotv = p->p_realtimer;
		if (timerisset(&aotv.it_value)) {
			if (timercmp(&aotv.it_value, &time, <))
				timerclear(&aotv.it_value);
			else
				timevalsub(&aotv.it_value, &time);
		}

		/*
		 * Handle race where setitimer occurs between todclock()
		 * and softclock(). That is, if the timer expires and
		 * the untimeout occurs before the timeout event sends
		 * the SIGALRM.
		 */
		if (luntimeout(realitexpire, (caddr_t)p)) {
			if (aotv.it_value.tv_sec==0 && aotv.it_value.tv_usec==0)
				++sigalrm;
		}
		if (timerisset(&aitv.it_value)) {
			timevaladd(&aitv.it_value, &time);
			ltimeout(realitexpire,(caddr_t)p, hzto(&aitv.it_value));
		}
		p->p_realtimer = aitv;
		V_GATE(G_TIME, SPLHI);
		v_lock(&time_lck, s);
	} else {
		s = splhi();
		if (uap->oitv) {
			aotv = u.u_timer[uap->which];
		}
		u.u_timer[uap->which] = aitv;
		splx(s);
	}

	/*
	 * Now copy out previous value.
	 */
	if (uap->oitv) {
		u.u_error = copyout((caddr_t)&aotv, (caddr_t)uap->oitv,
						sizeof (struct itimerval));
	}

	/*
	 * If needed, post a SIGALRM.  Done here to avoid nesting gates/locks.
	 */
	if (sigalrm)
		psignal(p, SIGALRM);
}

/*
 * Real interval timer expired:
 * send process whose timer expired an alarm signal.
 * If time is not set up to reload, then just return.
 * Else compute next time timer should go off which is > current time.
 * This is where delay in processing this timeout causes multiple
 * SIGALRM calls to be compressed into one.
 */
realitexpire(p)
	register struct proc *p;
{
	GATESPL(s);

	psignal(p, SIGALRM);
	if (!timerisset(&p->p_realtimer.it_interval)) {
		timerclear(&p->p_realtimer.it_value);
		return;
	}
	for (;;) {
		P_GATE(G_TIME, s);
		timevaladd(&p->p_realtimer.it_value,
		    &p->p_realtimer.it_interval);
		if (timercmp(&p->p_realtimer.it_value, &time, >)) {
			ltimeout(realitexpire, (caddr_t)p,
			    hzto(&p->p_realtimer.it_value));
			V_GATE(G_TIME, s);
			return;
		}
		V_GATE(G_TIME, s);
	}
}

/*
 * Check that a proposed value to load into the .it_value or
 * .it_interval part of an interval timer is acceptable, and
 * fix it to have at least minimal value (i.e. if it is less
 * than the resolution of the clock, round it up.)
 */
itimerfix(tv)
	struct timeval *tv;
{

	if (tv->tv_sec < 0 || tv->tv_sec > 100000000 ||
	    tv->tv_usec < 0 || tv->tv_usec >= 1000000)
		return (EINVAL);
	if (tv->tv_sec == 0 && tv->tv_usec != 0 && tv->tv_usec < tick)
		tv->tv_usec = tick;
	return (0);
}

/*
 * Decrement an interval timer by a specified number
 * of microseconds, which must be less than a second,
 * i.e. < 1000000.  If the timer expires, then reload
 * it.  In this case, carry over (usec - old value) to
 * reducint the value reloaded into the timer so that
 * the timer does not drift.  This routine assumes
 * that it is called in a context where the timers
 * on which it is operating cannot change in value.
 */
itimerdecr(itp, usec)
	register struct itimerval *itp;
	int usec;
{

	if (itp->it_value.tv_usec < usec) {
		if (itp->it_value.tv_sec == 0) {
			/* expired, and already in next interval */
			usec -= itp->it_value.tv_usec;
			goto expire;
		}
		itp->it_value.tv_usec += 1000000;
		itp->it_value.tv_sec--;
	}
	itp->it_value.tv_usec -= usec;
	usec = 0;
	if (timerisset(&itp->it_value))
		return (1);
	/* expired, exactly at end of interval */
expire:
	if (timerisset(&itp->it_interval)) {
		itp->it_value = itp->it_interval;
		itp->it_value.tv_usec -= usec;
		if (itp->it_value.tv_usec < 0) {
			itp->it_value.tv_usec += 1000000;
			itp->it_value.tv_sec--;
		}
	} else
		itp->it_value.tv_usec = 0;		/* sec is already 0 */
	return (0);
}

/*
 * Add and subtract routines for timevals.
 * N.B.: subtract routine doesn't deal with
 * results which are before the beginning,
 * it just gets very confused in this case.
 * Caveat emptor.
 */
timevaladd(t1, t2)
	struct timeval *t1, *t2;
{

	t1->tv_sec += t2->tv_sec;
	t1->tv_usec += t2->tv_usec;
	timevalfix(t1);
}

timevalsub(t1, t2)
	struct timeval *t1, *t2;
{

	t1->tv_sec -= t2->tv_sec;
	t1->tv_usec -= t2->tv_usec;
	timevalfix(t1);
}

timevalfix(t1)
	struct timeval *t1;
{

	if (t1->tv_usec < 0) {
		t1->tv_sec--;
		t1->tv_usec += 1000000;
	}
	if (t1->tv_usec >= 1000000) {
		t1->tv_sec++;
		t1->tv_usec -= 1000000;
	}
}
