/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: uipc_usrreq.c 2.10 88/03/20 $";
#endif

/*
 * upic_usrreq.c
 *	Protocol engines for Unix domain protocol
 *	and Unix domain protocol control blocks
 */

/* $Log:	uipc_usrreq.c,v $
 */

#include "../h/param.h"
#include "../machine/intctl.h"
#include "../machine/gate.h"
#include "../h/systm.h"
#include "../h/mutex.h"	
#include "../h/user.h"
#include "../h/uio.h"
#include "../h/mbuf.h"
#include "../h/protosw.h"
#include "../h/socket.h"
#include "../h/socketvar.h"
#include "../h/unpcb.h"
#include "../h/un.h"
#include "../h/vnode.h"
#include "../h/file.h"

#include "../h/vmmeter.h"
#include "../machine/vmparam.h"
#include "../machine/pte.h"
#include "../machine/plocal.h"

/*
 * Unix communications domain.
 *
 * TODO:
 *	SEQPACKET, RDM
 *	rethink name space problems
 *	need a proper out-of-band
 */

struct	sockaddr sun_noname = { AF_UNIX };

sema_t uipc_sema;

uipc_init()
{
	init_sema(&uipc_sema,1,0,G_SOCK);
	return (0);
}

/*ARGSUSED*/
uipc_usrreq(so, req, m, nam, rights)	/* entered with so_lock *held* */
	struct socket *so;
	int req;
	struct mbuf *m, *nam, *rights;
{
	struct unpcb *unp = sotounpcb(so);
	register struct socket *so2;
	int error = 0;
	int sema_held = 0;

	if (req != PRU_SEND && rights && rights->m_len) {
		error = EOPNOTSUPP;
		goto release;
	}
	if (unp == (struct unpcb *)NULL && req != PRU_ATTACH) {
		error = EINVAL;
		goto release;
	}

	/*
	 * Check for cases that require the use of a global semaphore
	 * to provide mutual exclusion.  For example a uipc DGRAM connection
	 * graph is essentially unbounded and arbitrarily cross referenced.
	 * It is possible to have a many to many mapping of connections to
	 * each other and other horrible things.  Therefore, reduce this
	 * to the simplest case.  Mutex all attached and bound DGRAM operations
	 * with a semaphore such that only one such operation can be in
	 * affect at the same time.  This is also true for not yet connected
	 * AF_UNIX SOCK_STREAMs.  After they are connected (i.e. unp_conn
	 * has a value) then a socket_peer can be safely used for mutex.
	 * Sockets used for pipes never require use of the global sema.
	 * NOTE: pipes begin with the socket_peer and so are not affected
	 * by this global sema
	 */

	if (unp &&
		(
		unp->unp_vnode
		|| (so->so_state & SS_NOFDREF)
		|| ((so->so_state & SS_ISCONNECTED) == 0)
		|| (so->so_type == SOCK_DGRAM)
		))
		{

			/*
			 * attached (unp)
			 * bound (unp_vnode - e.g. listen)
			 * or not file referenced (SS_NOFDREF -
			 * 	e.g. sonewconn's socket on so_q's
			 * or not yet connected (connectee)
			 * or this is a SOCK_DGRAM
			 */

			/*
			 * if attached and bound DGRAM or unconnected
			 * (but attached) STREAM - need to use p_sema
			 * NOTE: non signalable (PZERO)
			 */

			p_sema_v_lock(&uipc_sema, PZERO,
				&so->so_sopp->sop_lock, SPLNET);

			sema_held++;

			/* SOLOCK for consistency with socket management */

			(void) SOLOCK(so);
	}

	switch (req) {

	case PRU_ATTACH:

		if (unp) {
			error = EISCONN;
			break;
		}
		error = unp_attach(so);
		break;

	case PRU_DETACH:	/* so_lock held */
		unp_detach(unp);
		break;

	case PRU_BIND:
		error = unp_bind(unp, nam);
		break;

	case PRU_LISTEN:
		if (unp->unp_vnode == 0)
			error = EINVAL;
		break;

	case PRU_CONNECT:
		error = unp_connect(so, nam);

		/*
		 * undo unp_connect lockee - the connectee.
		 * this socket cannot disappear due to global
		 * semaphore semantics.  Check for self reference.
		 */

		if (!error && so != unp->unp_conn->unp_socket)
			(void) SOUNLOCK(unp->unp_conn->unp_socket, SPLNET);

		break;

	case PRU_CONNECT2:	/* used by socketpair */
		error = unp_connect2(so, (struct mbuf *)0,
		    (struct socket *)nam);
		break;

	case PRU_DISCONNECT:
		unp_disconnect(unp);
		break;

	case PRU_ACCEPT:
		nam->m_len = unp->unp_remaddr->m_len;
		bcopy(mtod(unp->unp_remaddr, caddr_t),
		    mtod(nam, caddr_t), (unsigned)nam->m_len);
		break;

	case PRU_SHUTDOWN:
		socantsendmore(so);
		unp_usrclosed(unp);	/* N.B. NO-OP routine */
		break;

	case PRU_RCVD:
		switch (so->so_type) {

		case SOCK_DGRAM:
			panic("uipc 1");
			/*NOTREACHED*/

		case SOCK_STREAM:

			/*
			 * socket_peer lock is held so both sockets are
			 * locked (e.g. pipe)
			 */

#define	rcv (&so->so_rcv)
#define snd (&so2->so_snd)

			if (unp->unp_conn == 0)
				break;
			so2 = unp->unp_conn->unp_socket;

			/*
			 * Transfer resources back to send port
			 * and wakeup any waiting to write.
			 */

			snd->sb_mbmax += rcv->sb_mbmax - rcv->sb_mbcnt;
			rcv->sb_mbmax = rcv->sb_mbcnt;
			snd->sb_hiwat += rcv->sb_hiwat - rcv->sb_cc;
			rcv->sb_hiwat = rcv->sb_cc;

			/*
			 * N.B. sbwakeup can take affect immediately
			 * on multi-processors
			 */

			sbwakeup(snd, 0);
#undef snd
#undef rcv
			break;

		default:
			panic("uipc 2");
		}
		break;

	case PRU_SEND:

		/*
		 * socket buffer is locked when called from sosend routine
		 * for send, sendto, and sendmsg - also by soo_rw when
		 * WRITE done.
		 */

		switch (so->so_type) {

		case SOCK_DGRAM:

			if (rights) {
				error = EOPNOTSUPP;
				break;
			}
			if (nam) {
				if (unp->unp_conn) {
					error = EISCONN;
					break;
				}
				error = unp_connect(so, nam);
				if (error)
					break;
			} else {
				if (unp->unp_conn == 0
				    || unp->unp_conn->unp_socket == 0) {
					error = ENOTCONN;
					break;
				}
				(void) SOLOCK(unp->unp_conn->unp_socket);
			}

			so2 = unp->unp_conn->unp_socket;

			if (sbspace(&so2->so_rcv) > 0) {

				/*
				 * There's no record of source socket's
				 * name, so send null name for the moment.
				 */

				/*
				 * vax4.2 has an error here in that the
				 * sbappendaddr return value is ignored.
				 * sbappendaddr returns (1) if the append
				 * is successful, (0) if not.  If (0) then
				 * the mbufs must be released (best effort
				 * to deliver fails).  This is performed
				 * before return if m != NULL.  Therefore,
				 * if successful, set m == NULL
				 * and if not, do not set m == NULL.
				 */

				if (sbappendaddr(&so2->so_rcv,
				    &sun_noname, m, rights,
				    so2->so_proto->pr_flags & PR_RIGHTS)) {
					m = (struct mbuf *)NULL;
				}

				sbwakeup(&so2->so_rcv, 0);
			}

			if (nam)
				unp_disconnect(unp);

			/*
			 * unlock the connectee, note this
			 * socket cannot disappear because of global
			 * semaphore semantics.
			 * Check for self reference.
			 */

			if (so2 != so)
				(void) SOUNLOCK(so2, SPLNET);
			break;

		case SOCK_STREAM:
#define	rcv (&so2->so_rcv)
#define	snd (&so->so_snd)

			if (rights && rights->m_len) {
				error = EOPNOTSUPP;
				break;
			}

			ASSERT(unp->unp_conn != 0, "uipc 3");

			so2 = unp->unp_conn->unp_socket;

			/*
			 * Send to paired receive port, and then
			 * give it enough resources to hold what it already has.
			 * Wake up readers.
			 */

			/*
			 * N.B. for TMP system - requires that destination
			 * socket be locked - not so in the case of mono-p
			 *
			 * socket_peer lock technique means that peer
			 * (i.e. connected) socket is locked so it is OK
			 * to fiddle with peer's socket struct.
			 */

			sbappend(rcv, m);

			m = (struct mbuf *)NULL;

			snd->sb_mbmax -= rcv->sb_mbcnt - rcv->sb_mbmax;
			rcv->sb_mbmax = rcv->sb_mbcnt;
			snd->sb_hiwat -= rcv->sb_cc - rcv->sb_hiwat;
			rcv->sb_hiwat = rcv->sb_cc;

			/*
			 * note that with multiple processors,
			 * sbwakeup can happen immediately.
			 */

			sbwakeup(rcv, 0);

#undef snd
#undef rcv
			break;

		default:
			panic("uipc 4");
		}
		break;

	case PRU_ABORT:

		/* so_lock *held* */

		unp_drop(unp, ECONNABORTED);
		break;

	/* SOME AS YET UNIMPLEMENTED HOOKS (sic 4.2) */

	case PRU_CONTROL:
	case PRU_SENSE:
	case PRU_RCVOOB:

		error = EOPNOTSUPP;

		/*
		 * m = NULL prevents these unsupported ops from getting
		 * non-mbuf argument m_freed!  soo_ioctl() and soo_stat()
		 * use non-mbufs - sorecv() releases mbuf given to
		 * OOB pr_usrreq()
		 */

		m = (struct mbuf *)NULL;

		break;

	/* END UNIMPLEMENTED HOOKS */

	case PRU_SENDOOB:

		/*
		 * from 4.3: note sosend() assumes pr_usrreq() eats mbuf
		 */

		error = EOPNOTSUPP;

	case PRU_SOCKADDR:
	case PRU_PEERADDR:
	case PRU_SLOWTIMO:
		break;

	default:
		panic("piusrreq");
	}
release:
	if (m) {
		m_freem(m);
	}
	if (sema_held) {
		(void) v_sema(&uipc_sema);
	}

	return (error);
}

unp_attach(so)
	struct socket *so;
{
	register struct mbuf *m;
	register struct unpcb *unp;
	register struct socket_peer * sop;
	int error;
	
	error = soreserve(so, unp_sendspace, unp_recvspace);

	if (error)
		return (error);

	/*
	 * N.B. M_DONTWAIT argument is not used (also not used in 4.2)
	 *
	 * however, this routine is called when a socket is first being
	 * created, therefore if we run out of buffers, return an error
	 * and socreate returns ENOBUFS to the user.
	 */

	m = m_getclr(M_DONTWAIT, MT_PCB);

	if (m == (struct mbuf *)NULL)
		return (ENOBUFS);

	unp = mtod(m, struct unpcb *);

	m = m_getclr(M_DONTWAIT, MT_SOPEER);
	if (m == (struct mbuf *)NULL) {
		(void) m_free(dtom(unp));
		return (ENOBUFS);
	}

	sop = mtod(m, struct socket_peer *);

	init_lock(&sop->sop_lock, G_SOCK);
	
	so->so_sopp = sop;
	
	sop->sop_refcnt = 2;

	/* point socket to pcb and pcb to socket */

	so->so_pcb = (caddr_t)unp;
	unp->unp_socket = so;

	return (0);
}

unp_detach(unp)
	register struct unpcb *unp;
{
	/* so_lock is *held* */
	if (unp->unp_vnode) {
		unp->unp_vnode->v_socket = (struct socket *)NULL;
		VN_RELE(unp->unp_vnode);
		unp->unp_vnode = (struct vnode *)NULL;
	}

	if (unp->unp_conn) {

		/*
		 * subtle, if this is still a connected socket pair
		 * (e.g. pipe), then if one is disconnected and the other
		 * is not, make the second one disconnected here in order
		 * to close out waiters etc.
		 */

		unp_disconnect(unp);	/* mark disconnected */

	}

	while (unp->unp_refs) {
		unp_drop(unp->unp_refs, ECONNRESET);
	}

	/*N.B. so_lock *held* */

	soisdisconnected(unp->unp_socket);

	unp->unp_socket->so_pcb = (caddr_t)NULL;

	if (--unp->unp_socket->so_sopp->sop_refcnt == 0) {
		(void) m_free(dtom(unp->unp_socket->so_sopp));
	}

	m_freem(unp->unp_remaddr);
	(void) m_free(dtom(unp));

	return;
}

unp_bind(unp, nam)
	struct unpcb *unp;
	struct mbuf *nam;
{
	struct sockaddr_un *soun = mtod(nam, struct sockaddr_un *);
	struct vnode *vp;
	struct vattr vattr;
	int error;

	/* not checked in 4.2bsd */

 	if (unp->unp_vnode)

		/* already bound */

 		return (EINVAL);
 
	if (nam->m_len == MLEN)
		return (EINVAL);
	*(mtod(nam, caddr_t) + nam->m_len) = 0;

	/* SHOULD BE ABLE TO ADOPT EXISTING AND wakeup() ALA FIFO's */

	vattr_null(&vattr);
	vattr.va_type = VSOCK;
	vattr.va_mode = 0777;
	error = vn_create(soun->sun_path, UIOSEG_KERNEL, &vattr, EXCL, 0, &vp);
	if (error) {
		if (error == EEXIST)
			return (EADDRINUSE);
		return (error);
	}
	vp->v_socket = unp->unp_socket;		/* socket is bound */
	unp->unp_vnode = vp;

	/* We return with vnode unlocked but referenced */

	VN_UNLOCKNODE(vp);

	return (0);
}

unp_connect(so, nam)
	struct socket *so;
	struct mbuf *nam;
{
	register struct sockaddr_un *soun = mtod(nam, struct sockaddr_un *);
	struct vnode *vp;
	int error;
	register struct socket *so2;

	register struct socket *asock;	/* accept socket */
	struct socket_peer * asock_sopp;

	if (nam->m_len + (nam->m_off - MMINOFF) == MLEN)
		return (EMSGSIZE);

	*(mtod(nam, caddr_t) + nam->m_len) = 0;

	error =
	    lookupname(soun->sun_path, UIOSEG_KERNEL, FOLLOW_LINK,
		(struct vnode **)0, &vp);

	if (error)
		return (error);

	if (vp->v_type != VSOCK) {
		VN_PUT(vp);
		return (ENOTSOCK);
	}

	so2 = vp->v_socket;

	if (so2 == (struct socket *)NULL) {
		VN_PUT(vp);
		return (ECONNREFUSED);
	}

	if (so != so2)
		(void) SOLOCK(so2);

	if (so->so_type != so2->so_type) {
		if (so != so2)
			SOUNLOCK(so2, SPLNET);
		VN_PUT(vp);
		return (EPROTOTYPE);
	}

	switch(so->so_type) {

	case SOCK_DGRAM:
		if ((error = unp_connect2(so, nam, so2))) {
			if (so != so2)
				SOUNLOCK(so2, SPLNET);
		}

		/*
		 * note: if (!error) so2 is still locked
		 */

		VN_PUT(vp);
		return (error);

	case SOCK_STREAM:

		if ((so2->so_options&SO_ACCEPTCONN) == 0) {
			if (so2 != so)
				(void) SOUNLOCK(so2, SPLNET);
			VN_PUT(vp);
			return (ECONNREFUSED);
		}

		asock = so2;
		asock_sopp = so2->so_sopp;
		asock_sopp->sop_refcnt++;

		/*
		 * note, check for ACCEPTCONN must be done after
		 * sonewconn since it is checking for concurrent
		 * deletion of so_head.
		 */

		if ((so2 = sonewconn(so2)) == (struct socket *)NULL
		   || !(asock->so_options & SO_ACCEPTCONN)) {

			/*
			 * there is no new so, or there is and the
			 * head is disappearing. check for
			 * self reference and allow so_head to finish
			 */

			if (so2)
				(void) SOUNLOCK(so2, SPLNET);

			/*
			 * refcnt-- before check since self-reference
			 * also refcnt++  - can this be more efficient?
			 */

			asock_sopp->sop_refcnt--;
			if (so != asock)
				if (asock_sopp->sop_refcnt)
					(void) v_lock(&asock_sopp->sop_lock,
							SPLNET);
				else
					(void) m_free(dtom(asock_sopp));
			VN_PUT(vp);
			return (ECONNREFUSED);
		}

		if ((error = unp_connect2(so, nam, so2)))
			(void) soabort(so2);
		else {

			/*
			 * so and so2 connected ok
			 */

			so2->so_state &= ~(SS_ISCONNECTING|SS_ISDISCONNECTING);
			so2->so_state |= SS_ISCONNECTED;

			/*
			 * wake up accept socket waiter 
			 *
			 * note that if this is a connect request to an
			 * accept socket, so2 is new socket on accept side,
			 * not the accept socket. There are not always
			 * connect waiters.
			 */

			(void) vall_sema(asock->so_conn_semap);
		}

		/*
		 * can now unlock the so_head (asock)
		 * successful new socket is unlocked by caller (_CONNECT)
		 */

		asock_sopp->sop_refcnt--;

		if (asock != so)
			(void) SOUNLOCK(asock, SPLNET);

		VN_PUT(vp);

		return (error);

	default:
		if (so != so2)
			SOUNLOCK(so2, SPLNET);
		VN_PUT(vp);
		return (EPROTOTYPE);	/* should not happen */
	}

}

/*
 * unp_connect2 - connect two unix domain sockets - both so and so2 are
 * locked upon entry. If SOCK_STREAM, the acceptee socket is also
 * locked (for soisconnected).  Also, if a SOCK_STREAM connect
 * request, so2 is a new socket that is locked, but has no reference to
 * it yet (will get reference when return to accept call); therefore it
 * is OK to fiddle with the so_sopp and point it to so's socket_peer.
 */

unp_connect2(so, sonam, so2)
	register struct socket *so;
	struct mbuf *sonam;
	register struct socket *so2;
{
	register struct unpcb *unp = sotounpcb(so);
	register struct unpcb *unp2;

	if (so2->so_type != so->so_type)
		return (EPROTOTYPE);

	unp2 = sotounpcb(so2);
	unp->unp_conn = unp2;

	switch (so->so_type) {

	case SOCK_DGRAM:

		/*
		 * More thought needed with respect to multiple connections
		 * to a single SOCK_DGRAM socket?
		 * Currently mutex is via global uipc_sema.
		 */

		unp->unp_nextref = unp2->unp_refs;
		unp2->unp_refs = unp;

		/* FIX moved from unp_connect */

		so2->so_state &= ~(SS_ISCONNECTING|SS_ISDISCONNECTING);
		so2->so_state |= SS_ISCONNECTED;

		so->so_state &= ~(SS_ISCONNECTING|SS_ISDISCONNECTING);
		so->so_state |= SS_ISCONNECTED;

		(void) vall_sema(so2->so_conn_semap);

		break;

	case SOCK_STREAM:

		unp2->unp_conn = unp;

		if (sonam) {

			/*
			 * XXX - in 4.2 not checked -
			 * mcopy can return NULL => ENOBUFS!
			 */

			unp2->unp_remaddr = m_copy(sonam, 0, (int)M_COPYALL);

			if (unp2->unp_remaddr == (struct mbuf*)NULL) {

				/*
				 * unp_connect2 fails because m_copy failed
				 * unravel attempted connect2 and return error
				 */

				unp2->unp_conn = (struct unpcb*)NULL;
				unp->unp_conn = (struct unpcb*)NULL;
				return (ENOBUFS);
			}
		}

		/*
		 * point so2 to so's socket_peer.  This is OK since so2 is a
		 * new socket.  (for the case of socketpair *both* so and so2
		 * are new sockets; however, for an accept/connect sequence,
		 * the connect socket is so - and it does have a reference.
		 * point so2->so_sopp to so's socket_peer before signalling
		 * soisconnected.  Note: this leaves the unreferenced lock
		 * in the released sopp locked, but that is OK.
		 */

		(void) m_free(dtom(so2->so_sopp));

		so2->so_sopp = so->so_sopp;

		/* need ref for each detach and soclose */
	
		so->so_sopp->sop_refcnt = 4;

		soisconnected(so2);

		soisconnected(so);

		break;

	default:
		panic("unp_connect2");
	}
	return (0);
}

unp_disconnect(unp)		/* n.b. so_lock is held */
	struct unpcb *unp;
{
	register struct unpcb *unp2 = unp->unp_conn;

	if (unp2 == (struct unpcb *)NULL)
		return;

	unp->unp_conn = (struct unpcb *)NULL;

	switch (unp->unp_socket->so_type) {

	case SOCK_DGRAM:
		if (unp2->unp_refs == unp)
			unp2->unp_refs = unp->unp_nextref;
		else {
			unp2 = unp2->unp_refs;
			for (;;) {
				ASSERT(unp2 != (struct unpcb *)NULL,
						"unp_disconnect");
				if (unp2->unp_nextref == unp)
					break;
				unp2 = unp2->unp_nextref;
			}
			unp2->unp_nextref = unp->unp_nextref;
		}
		unp->unp_nextref = (struct unpcb *)NULL;
		unp->unp_socket->so_state &= ~SS_ISCONNECTED;
		break;

	case SOCK_STREAM:
		soisdisconnected(unp->unp_socket);
		unp2->unp_conn = (struct unpcb *)NULL;
		soisdisconnected(unp2->unp_socket);
		break;
	}
	return;
}

/*ARGSUSED*/
unp_usrclosed(unp)
	struct unpcb *unp;
{
	return;
}

unp_drop(unp, errno)
	struct unpcb *unp;
	int errno;
{
	struct socket *so = unp->unp_socket;
	struct socket * head = so->so_head;

	so->so_error = errno;
	unp_disconnect(unp);

	if (head) {

		/*
		 * this socket is still in a socket queue
		 * but it is not accepted - soclose() is in process
		 * of soaborting it.  accept() will not accept during
		 * so_head's window since it is marked ~SO_ACCEPTCONN
		 * Formerly connected peer is mutexed via socket_peer
		 * and now unp_disconnected.
		 *
		 * release *all* resources.
		 */

		(void) SOLOCK(head);
		so->so_pcb = (caddr_t) 0;
		m_freem(unp->unp_remaddr);
		(void) m_free(dtom(unp));
		(void) m_free(dtom(so->so_sopp));
		sofree(so);
		SOUNLOCK(head, SPLNET);
	}
	return;
}
