/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990, 1991
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: /usr/src/dynix.3.0.17/src/sys/i386/conf/RCS/conf_zd.c,v 2.2 91/03/22 00:56:34 root Exp $";
#endif

/*
 * configuration of disks on the ZDC
 *
 * WARNING: If the tables are partially filled in unpredictable results
 *	    *WILL* occur.
 */

#include "../h/param.h"
#include "../h/mutex.h"
#include "../h/buf.h"
#include "../h/time.h"
#include "../zdc/zdc.h"

/*
 *	Part A.	Partition tables.
 *
 * On all drives, the first cylinder is reserved for disk description data
 * and the last two cylinders are reserved for diagnostics. No partition
 * which will contain a filesystem should include any of these cylinders.
 *
 * N.B.: The stand-alone driver knows these offsets.
 *
 * The zdparts table is indexed by drive type.
 * Note that this list is order dependent. New entries *must*
 * correspond with the drive type.
 *
 * NOTE: The newfs utility ASSUMES that the 'c' partition starts at the
 *	 beginning of the disk when writing the bootstrap program.
 *	 The bootstrap program is written when a root filesystem is created.
 *	 The newfs utility ASSUMES that the 'a' partition is the root
 *	 filesystem. However, by writing the bootstrap to partition 'c' the
 *	 'a' partition may be moved to the middle of the disk to reduce
 *	 seek latency.
 *	 If the 'c' partition is changed so that it does not include the
 *	 start of the disk, then be sure to use the "-n" option to newfs
 *	 and use /stand/installboot to write the bootstrap program (at least
 *	 1 partition must start at the beginning of the disk).
 */

struct zdsize m2333k[NUMPARTS] = {	/* Fujitsu M2333K (swallow) */
	 25*66*10,	335,	/* A=cyl 335 thru 359 */
	102*66*10,	360,	/* B=cyl 360 thru 461 */
	820*66*10,	  1,	/* C=cyl   1 thru 820 */
	257*66*10,	  1,	/* D=cyl   1 thru 257 */
	102*66*10,	258,	/* E=cyl 258 thru 359 */
	359*66*10,	  1,	/* F=cyl   1 thru 359 */
	359*66*10,	462,	/* G=cyl 462 thru 820 */
	334*66*10,	  1,	/* H=cyl   1 thru 334 */
};

struct zdsize m2351a[NUMPARTS] = {	/* Fujitsu M2351A (Eagle) */
	 18*46*20,	366,	/* A=cyl 366 thru 383 */
	 73*46*20,	384,	/* B=cyl 384 thru 456 */
	839*46*20,	  1,	/* C=cyl   1 thru 839 */
	419*46*20,	  1,	/* D=cyl   1 thru 419 */
	419*46*20,	420,	/* E=cyl 420 thru 838 */
	383*46*20,	  1,	/* F=cyl   1 thru 383 */
	383*46*20,	457,	/* G=cyl 457 thru 839 */
	365*46*20,	  1,	/* H=cyl   1 thru 365 */
};

struct zdsize m2344k[NUMPARTS] = {	/* Fujitsu M2344K (swallow 4) */
	 10*66*27,	282,	/* A=cyl 282 thru 291 */
	 38*66*27,	292,	/* B=cyl 292 thru 329 */
	621*66*27,	  1,	/* C=cyl   1 thru 621 */
	195*66*27,	  1,	/* D=cyl   1 thru 195 */
	 96*66*27,	196,	/* E=cyl 196 thru 291 */
	291*66*27,	  1,	/* F=cyl   1 thru 291 */
	292*66*27,	330,	/* G=cyl 330 thru 621 */
	281*66*27,	  1,	/* H=cyl   1 thru 281 */
};

struct zdsize m2382k[NUMPARTS] = {	/* Fujitsu M2382K (swallow 5) */
	 8*81*27,	348,	/* A=cyl 348 thru 355 */
	 32*81*27,	356,	/* B=cyl 356 thru 387 */
	742*81*27,	  1,	/* C=cyl   1 thru 742 */
	371*81*27,	  1,	/* D=cyl   1 thru 371 */
	371*81*27,	372,	/* E=cyl 372 thru 742 */
	355*81*27,	  1,	/* F=cyl   1 thru 355 */
	355*81*27,	388,	/* G=cyl 388 thru 742 */
	347*81*27,	  1,	/* H=cyl   1 thru 347 */
};

struct zdsize cdc9715_340[NUMPARTS] = {	/* CDC 9715-340 (FSD) */
	 21*34*24,	292,	/* A=cyl 292 thru 312 */
	 83*34*24,	313,	/* B=cyl 313 thru 395 */
	708*34*24,	  1,	/* C=cyl   1 thru 708 */
	354*34*24,	  1,	/* D=cyl   1 thru 354 */
	354*34*24,	355,	/* E=cyl 355 thru 708 */
	312*34*24,	  1,	/* F=cyl   1 thru 312 */
	313*34*24,	396,	/* G=cyl 396 thru 708 */
	291*34*24,	  1,	/* H=cyl   1 thru 291 */
};

struct zdsize cdc9771_800[NUMPARTS] = {	/* CDC 9771-800 (XMD) */
	 12*85*16,	474,	/* A=cyl 474 thru 485 */
	 50*85*16,	486,	/* B=cyl 486 thru 535 */
	1021*85*16,	  1,	/* C=cyl   1 thru 1021 */
	511*85*16,	  1,	/* D=cyl   1 thru 511 */
	510*85*16,	512,	/* E=cyl 512 thru 1021 */
	485*85*16,	  1,	/* F=cyl   1 thru 485 */
	486*85*16,	536,	/* G=cyl 536 thru 1021 */
	473*85*16,	  1,	/* H=cyl   1 thru 473 */
};

struct zdsize cdc9720_850[NUMPARTS] = {	/* CDC 9720-850 (EMD) */
	 17*68*15,	640,	/* A=cyl 640 thru 656 */
	 66*68*15,	657,	/* B=cyl 657 thru 722 */
	1378*68*15,	  1,	/* C=cyl   1 thru 1378 */
	689*68*15,	  1,	/* D=cyl   1 thru 689 */
	689*68*15,	690,	/* E=cyl 690 thru 1378 */
	656*68*15,	  1,	/* F=cyl   1 thru 656 */
	656*68*15,	723,	/* G=cyl 723 thru 1378 */
	639*68*15,	  1,	/* H=cyl   1 thru 639 */
};

struct zdsize cdc9720_1230[NUMPARTS] = { /* CDC 9720-1230 (EMD) */
	  16*83*15,	769,	/* A=cyl 769 thru 784 */
 	  55*83*15,	785,	/* B=cyl 785 thru 839 */
        1632*83*15,	  1,	/* C=cyl   1 thru 1632 */
 	 816*83*15,	  1,	/* D=cyl   1 thru 816 */
 	 816*83*15,	817,	/* E=cyl 817 thru 1632 */
 	 784*83*15,	  1,	/* F=cyl   1 thru 784 */
 	 793*83*15,	840,	/* G=cyl 840 thru 1632 */
 	 768*83*15,	  1,	/* H=cyl   1 thru 768 */
};

struct zdsize sabre5_1230[NUMPARTS] = { /* SABRE5-1230 (EMD) new format */
	  16*83*15,	769,	/* A=cyl 769 thru 784 */
 	  55*83*15,	785,	/* B=cyl 785 thru 839 */
        1632*83*15,	  1,	/* C=cyl   1 thru 1632 */
 	 816*83*15,	  1,	/* D=cyl   1 thru 816 */
 	 816*83*15,	817,	/* E=cyl 817 thru 1632 */
 	 784*83*15,	  1,	/* F=cyl   1 thru 784 */
 	 793*83*15,	840,	/* G=cyl 840 thru 1632 */
 	 768*83*15,	  1,	/* H=cyl   1 thru 768 */
};

/*
 * A null zdsize array for Reserved drive types
 * Done this way since driver already checks for a partition
 * length of zero.  Otherwise additonal code to check
 * for NULL pointer in zdparts[] would be required.
 */
struct zdsize zd_nulltype[NUMPARTS] = {	/* Used for reserved types */
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
	0,	0,
};

struct zdsize *zdparts[] = {
	m2333k,			/* 0 - Fujitsu M2333K (swallow) */
	m2351a,			/* 1 - Fujitsu M2351A (Eagle) */
	m2344k,			/* 2 - Fujitsu M2344K (swallow 4) */
	m2382k,			/* 3 - Fujitsu M2382K (swallow 5) */
	cdc9720_850,		/* 4 - CDC 9720-850 (EMD) */
	zd_nulltype,		/* 5 - Reserved */
	zd_nulltype,		/* 6 - Reserved */
	zd_nulltype,		/* 7 - Reserved */
	zd_nulltype,		/* 8 - Reserved */
	zd_nulltype,		/* 9 - Reserved */
	zd_nulltype,		/* 10 - Reserved */
	zd_nulltype,		/* 11 - Reserved */
	zd_nulltype,		/* 12 - Reserved */
	zd_nulltype,		/* 13 - Reserved */
	zd_nulltype,		/* 14 - Reserved */
	zd_nulltype,		/* 15 - Reserved */
	cdc9715_340,		/* 16 - CDC 9715-340 MB (FSD) */
	cdc9771_800,		/* 17 - CDC 9771-800 MB (XMD) */
	zd_nulltype,		/* 18 - Reserved */
	zd_nulltype,		/* 19 - Reserved */
	cdc9720_1230,           /* 20 - CDC 9720-1230 MB (EMD) */
	sabre5_1230,            /* 21 - SABRE5-1230 MB (EMD) new format */
};
int	zdntypes	= sizeof (zdparts) / sizeof(struct zdsize *);

/*
 *	Part B. Global Information.
 */

/*
 * If the number of zdc_iovpercb is 0 then the driver will automatically
 * allocate enough iovecs to handle max_RAW_IO (see physio()).
 * If the number of zdc_iovpercb is non-zero, then the value specified
 * is allocated if it is less than max_RAW_IO. If the specified value is
 * greater than max_RAW_IO the number is reduced to handle max_RAW_IO.
 */
int	zdc_iovpercb	= 0;			/* no of iovecs per cb */

int	zdc_err_bin	= 7;			/* Error interrupt bin */
int	zdc_cb_bin	= 5;			/* CB interrupt bin */
gate_t	zdcgate		= 62;			/* gate for zdc locks/semas */
/*
 * polled command timeout, should be approx. 30 secs.
 */
unsigned int	zdccmdtime	= 30000000;
/*
 * controller ready timeout, should be >= 60 secs.
 */
unsigned int	zdcinitime	= 2400000;
short	zdcretry	= 10;			/* retry count */
int	zdc_AB_throttle	= 2;			/* Channel A&B DMA throttle */
int	zdc_C_throttle	= 2;			/* Channel C DMA throttle */
u_char	zdctrl = ZDC_DUMPONPANIC;		/* ZDC_INIT control bits */
