/* $Header: asm.h 1.8 89/02/27 $ */

/*
 * asm.h
 *	Standard ASM use definitions.  i386 version.
 *
 * assym.h contains structure offsets and the like, generated by genassym.c
 */

/* $Log:	asm.h,v $
 */

/*
 * ENTRY() declares a text .globl and aligns the start of the procedure.
 */

#define	ENTRY(x)	.text; .globl _/**/x; .align 2; _/**/x:

/*
 * BP relative stack arguments for asm procedures that obey C calling
 * convention, establishing a stack frame.
 */

#define	BPARGOFF 8
#define	BPARG0	BPARGOFF+0(%ebp)
#define	BPARG1	BPARGOFF+4(%ebp)
#define	BPARG2	BPARGOFF+8(%ebp)
#define	BPARG3	BPARGOFF+12(%ebp)

/*
 * SP relative stack arguments for asm procedures that don't "enter".
 */

#define	SPARGOFF 4
#define	SPARG0	SPARGOFF+0(%esp)
#define	SPARG1	SPARGOFF+4(%esp)
#define	SPARG2	SPARGOFF+8(%esp)
#define	SPARG3	SPARGOFF+12(%esp)

/*
 * Call, return "instructions", in case these need to change.
 */

#define	CALL	call
#define	RETURN	ret

#if	defined(KXX) || defined(SLIC_GATES)

/*
 * In-line ASM macros for portable gate and lock implementations.
 * Macros don't block interrupts; caller must do this.
 *
 * P_GATE_ASM	grab a gate, used in ASM code.
 * CP_GATE_ASM	conditionally grab a gate, used in ASM code.
 * V_GATE_ASM	release a gate in ASM code.
 *
 * These macros touch only EAX.
 */

/*
 * P_SLIC_GATE and V_SLIC_GATE lock/unlock slic gate whose number is in %al.
 * Since these may be called from (eg) p_gate() and cp_gate(), the must be
 * sure to save/restore interrupt enable flag.  This is ok in the prototype
 * (K20), and a non-issue on the real HW.
 */

#define	P_SLIC_GATE(gateno) \
	pushfl; cli; \
	movb	$GATE_GROUP, SL_DEST+VA_SLIC; \
	movb	gateno, SL_SMESSAGE+VA_SLIC; \
9:	movb	$SL_REQG, SL_CMD_STAT+VA_SLIC; \
8:	testb	$SL_BUSY, SL_CMD_STAT+VA_SLIC; \
	jne	8b; \
	testb	$SL_OK, SL_CMD_STAT+VA_SLIC; \
	je	9b

#define	V_SLIC_GATE(gateno) \
	movb	$GATE_GROUP, SL_DEST+VA_SLIC; \
	movb	gateno, SL_SMESSAGE+VA_SLIC; \
	movb	$SL_RELG, SL_CMD_STAT+VA_SLIC; \
9:	testb	$SL_BUSY, SL_CMD_STAT+VA_SLIC; \
	jne	9b; \
	popfl

#define	P_GATE_ASM(gaddr) \
	leal	gaddr, %eax; \
	shrb	$2, %al; \
	andb	$0x3f, %al; \
5:	cmpb	$G_UNLOCKED, gaddr; \
	jne	5b; \
	P_SLIC_GATE(%al); \
	movb	gaddr, %ah; \
	cmpb	$G_UNLOCKED, %ah; \
	jne	6f; \
	movb	$G_LOCKED, gaddr; \
6:	V_SLIC_GATE(%al); \
	cmpb	$G_UNLOCKED, %ah; \
	jne	5b

#define	CP_GATE_ASM(gaddr,fail) \
	leal	gaddr, %eax; \
	shrb	$2, %al; \
	andb	$0x3f, %al; \
	P_SLIC_GATE(%al); \
	movb	gaddr, %ah; \
	cmpb	$G_UNLOCKED, %ah; \
	jne	6f; \
	movb	$G_LOCKED, gaddr; \
6:	V_SLIC_GATE(%al); \
	cmpb	$G_UNLOCKED, %ah; \
	jne	fail

#define	V_GATE_ASM(gaddr)	movb	$G_UNLOCKED, gaddr

#else	Real HW

/*
 * Gate and lock accesses on real HW are in-line expanded.
 * See machine/mutex.h
 */

#define	V_GATE_ASM(gaddr) \
	movb	$G_UNLOCKED, %al; \
	xchgb	%al, gaddr

#endif	KXX

/*
 * V_LOCK_ASM() assumes gate and lock data-structure and values are idential.
 */

#define	V_LOCK_ASM(lock)	V_GATE_ASM(lock)

/*
 * SPL_ASM(new,old)	raise SPL to "new", put old value in "old" (mod's %ah).
 * SPLX_ASM(old)	lower SPL back to "old".
 *
 * See machine/intctl.h for detail on spl synch with SLIC.
 * need 8 clocks at 16Mhz 10 at 20 Mhz  12 at 24 Mhz etc
 * modelC requires an extra 50 ns
 * all machines may run at 10% margins.
 * Note: movb X,r	will contribute 2 cycles
 *	 nop	        is 3 cycles
 *	 movl	r,r	is 2 cycles
 */

/***************SLICSYNC 2 ***************************************/
#if MHz == 16
#define	SPL_ASM(new,old) \
	movb	VA_SLIC+SL_LMASK, old; \
	movb	new, VA_SLIC+SL_LMASK; \
 	movb	VA_SLIC+SL_LMASK, %ah; \
	nop; nop;
#else
#if MHz == 20
#define	SPL_ASM(new,old) \
	movb	VA_SLIC+SL_LMASK, old; \
	movb	new, VA_SLIC+SL_LMASK; \
 	movb	VA_SLIC+SL_LMASK, %ah; \
	movl	%eax,%eax; \
	movl	%eax,%eax; \
	nop; nop; 
#else
	ERROR not 16Mhz nor 20Hz
#endif
#endif

#define	SPLX_ASM(old)	movb	old, VA_SLIC+SL_LMASK
