/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: dbg_machdep.c 2.6 86/09/13 $";
#endif

/*
 * dbg_machdep.c
 *	Machine dependent debugger support routines.
 *
 * Conditionals:
 *	-DWATCHPT:	turn on watchpoint code.
 *
 * i386 version.
 */

/* $Log:	dbg_machdep.c,v $
 */

#include "../h/param.h"
#include "../h/mutex.h"
#include "../h/systm.h"
#include "../h/user.h"
#include "../h/proc.h"
#include "../h/vm.h"

#include "../machine/reg.h"
#include "../machine/pte.h"
#include "../machine/psl.h"
#include "../machine/mmu.h"
#include "../machine/intctl.h"
#include "../machine/plocal.h"

/*
 * Order of registers must match struct pt_regset.
 */

int	ipcreg[] = {EAX,EBX,ECX,EDX,ESI,EDI,EBP,ESP,EIP};
int	nipcreg	= (sizeof(ipcreg)/sizeof(ipcreg[0]));

/*
 * dbg_watchpt_set()
 *	Set a watchpoint.
 */

/*ARGSUSED*/
dbg_watchpt_set(wpt)
	struct	pt_watchpt *wpt;
{
#ifdef	WATCHPT
	/*
	 * Must insure address is within legit user space
	 * (real problems could happen if certain kernel addresses
	 * were given).  Need to pick a free watchpoint (if any)
	 * and set it up.  No need to program up debug registers,
	 * this process necessarily context switches before
	 * runs user code again.
	 *
	 * May need to not turn on until return from mpt_stop??
	 * Think is ok; worst case is user asks to write a variable
	 * that a watchpoint is set in -- gets another trap, but
	 * that's what he asked for.
	 */
#else	!WATCHPT
	return(EINVAL);				/* for now */
#endif	WATCHPT
}

/*
 * dbg_watchpt_clear()
 *	Clear a watchpoint.
 */

/*ARGSUSED*/
dbg_watchpt_clear(wpt)
	struct	pt_watchpt *wpt;
{
#ifdef	WATCHPT
	/*
	 * See if currently have this address watchpoint'd.
	 * If so, disable it by clearing appropriate wp_control
	 * Gi bit.  If last just went off, set wp_control == DCR_OFF.
	 *
	 * Again, no need to turn off since process will context
	 * switch before really continuing.
	 */
#else	!WATCHPT
	return(EINVAL);				/* for now */
#endif	WATCHPT
}

/*
 * watchpoint()
 *	A watchpoint() debug trap occurred (in user or kernel mode).
 *
 * Turn off watchpoints() and enter mpt_stop().
 *
 * Caller already checked for and handled single-step trap.
 */

/*ARGSUSED*/
watchpoint(dsr, kernel)
	int	dsr;			/* Debug Status Register (DR6) */
	bool_t	kernel;			/* trap came from kernel mode? */
{
#ifdef	WATCHPT
	register struct proc *p = u.u_procp;

	ASSERT(dsr & DSR_WATCHPT, "watchpoint: dsr");
	ASSERT(l.watchptON, "watchpoint: !watchptON");

	/*
	 * Disable watchpoints in HW and arrange SW doesn't turn on again.
	 */

	WRITE_DCR(DCR_OFF);			/* turn OFF HW watchpoints */
	u.u_watchpt.wp_control = DCR_OFF;	/* turn OFF SW shadow */
	l.watchptON = 0;			/* really OFF */

	/*
	 * If came from kernel mode, do XXX (not yet defined).
	 */

	if (kernel) {
		panic("watchpoint: kernel mode not yet supported");
		/*NOTREACHED*/
	}

	/*
	 * Lock proc-state and stop for debugger.
	 * Could race with debugger exit, thus only stop if still supposed to.
	 */

	(void) p_lock(&p->p_state, SPLHI);

	if (p->p_flag & SMPTRC)
		mpt_stop(PTS_WATCHPT_AFTER);

	v_lock(&p->p_state, SPL0);
#else	!WATCHPT
	panic("watchpoint");
#endif	WATCHPT
}
