/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)coredump.c	1.7 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: coredump.c 2.11 87/03/31 $";
#endif

/*
 * Deal with the core dump anachronism.
 */

#include "defs.h"
#include "coredump.h"
#include "machine.h"
#include "object.h"
#include "main.h"
#include <sys/param.h>
#include <sys/dir.h>
#include <machine/psl.h>
#include <machine/pte.h>
#include <sys/user.h>
#include "multi-p.h"
#include <sys/vm.h>
#include <machine/reg.h>
#include <machine/plocal.h>

#ifndef public
#include "machine.h"
#endif

#define MAXSTKADDR USRSTACK	/* highest stack address */

/*
 * file descriptors for reading from core/object.  These are reopened
 * if necessary.
 */
private String lastobj;
private String lastcore;
private File objfile = NULL, corefile = NULL;

/*
 * return descriptor for file "newname".  if the "newname" file is the
 * same as the "oldname" file, re-use the descriptor "file".  if not, open 
 * "newname", and store the new name in "oldname".  implements a one-element 
 * cache for descriptors.
 */
private File coreopen(oldname, newname, file)
	String *oldname;
	String newname;
	File file;
{
	if (newname == nil)
		return(NULL);
	if (*oldname != newname || file == NULL) {
		if (file != NULL)
			fclose(file);
		if ((file = fopen(newname, "r")) == NULL)
			*oldname = nil;
		else
			*oldname = newname;
	}
	return(file);
}

/*
 * get register values out of the core dump's u-area.  'ar0' is the
 * address of registers on the kernel stack, biased to point into
 * u.  all other registers are relative to it.
 */
private getregs(u)
    struct user *u;
{
    Word *ar0;
#ifdef i386
    extern double t_to_d();
#endif

    ar0 = (Word *) ((Word)u->u_ar0 - VA_UAREA + (Word)u);
#ifdef ns32000
    setreg(0, ar0[R0]);
    setreg(1, ar0[R1]);
    setreg(2, ar0[R2]);
    setreg(3, ar0[R3]);
    setreg(4, ar0[R4]);
    setreg(5, ar0[R5]);
    setreg(6, ar0[R6]);
    setreg(7, ar0[R7]);
    setreg(FRP, ar0[FP]);
    setreg(STKP, ar0[SP]);
    setreg(PROGCTR, ar0[PC]);
    setreg(MODREG, ar0[MODPSR]);
    setreg(SB, 0);
    setreg(F0, u->u_fpusave.fpu_f0);
    setreg(F1, u->u_fpusave.fpu_f1);
    setreg(F2, u->u_fpusave.fpu_f2);
    setreg(F3, u->u_fpusave.fpu_f3);
    setreg(F4, u->u_fpusave.fpu_f4);
    setreg(F5, u->u_fpusave.fpu_f5);
    setreg(F6, u->u_fpusave.fpu_f6);
    setreg(F7, u->u_fpusave.fpu_f7);
    setreg(FSR, u->u_fpusave.fpu_fsr);
#endif
#ifdef i386
    setreg(REG_EAX, ar0[EAX]);
    setreg(REG_EBX, ar0[EBX]);
    setreg(REG_ECX, ar0[ECX]);
    setreg(REG_EDX, ar0[EDX]);
    setreg(REG_ESI, ar0[ESI]);
    setreg(REG_EDI, ar0[EDI]);
    setreg(REG_EBP, ar0[EBP]);
    setreg(REG_ESP, ar0[ESP]);
    setreg(REG_EIP, ar0[EIP]);
    setreg(REG_EFLAGS, ar0[FLAGS]);
    setreg(REG_FPA1, u->u_fpasave.fpa_regs[0]);
    setreg(REG_FPA2, u->u_fpasave.fpa_regs[1]);
    setreg(REG_FPA3, u->u_fpasave.fpa_regs[2]);
    setreg(REG_FPA4, u->u_fpasave.fpa_regs[3]);
    setreg(REG_FPA5, u->u_fpasave.fpa_regs[4]);
    setreg(REG_FPA6, u->u_fpasave.fpa_regs[5]);
    setreg(REG_FPA7, u->u_fpasave.fpa_regs[6]);
    setreg(REG_FPA8, u->u_fpasave.fpa_regs[7]);
    setreg(REG_FPA9, u->u_fpasave.fpa_regs[8]);
    setreg(REG_FPA10, u->u_fpasave.fpa_regs[9]);
    setreg(REG_FPA11, u->u_fpasave.fpa_regs[10]);
    setreg(REG_FPA12, u->u_fpasave.fpa_regs[11]);
    setreg(REG_FPA13, u->u_fpasave.fpa_regs[12]);
    setreg(REG_FPA14, u->u_fpasave.fpa_regs[13]);
    setreg(REG_FPA15, u->u_fpasave.fpa_regs[14]);
    setreg(REG_FPA16, u->u_fpasave.fpa_regs[15]);
    setreg(REG_FPA17, u->u_fpasave.fpa_regs[16]);
    setreg(REG_FPA18, u->u_fpasave.fpa_regs[17]);
    setreg(REG_FPA19, u->u_fpasave.fpa_regs[18]);
    setreg(REG_FPA20, u->u_fpasave.fpa_regs[19]);
    setreg(REG_FPA21, u->u_fpasave.fpa_regs[20]);
    setreg(REG_FPA22, u->u_fpasave.fpa_regs[21]);
    setreg(REG_FPA23, u->u_fpasave.fpa_regs[22]);
    setreg(REG_FPA24, u->u_fpasave.fpa_regs[23]);
    setreg(REG_FPA25, u->u_fpasave.fpa_regs[24]);
    setreg(REG_FPA26, u->u_fpasave.fpa_regs[25]);
    setreg(REG_FPA27, u->u_fpasave.fpa_regs[26]);
    setreg(REG_FPA28, u->u_fpasave.fpa_regs[27]);
    setreg(REG_FPA29, u->u_fpasave.fpa_regs[28]);
    setreg(REG_FPA30, u->u_fpasave.fpa_regs[29]);
    setreg(REG_FPA31, u->u_fpasave.fpa_regs[30]);
    bcopy((char *)(u->u_fpusave.fpu_stack[0]), (char *)(wp->p_regs.pr_fpu.fpu_stack[0]), 10/* 80 bits */);
    bcopy((char *)(u->u_fpusave.fpu_stack[1]), (char *)(wp->p_regs.pr_fpu.fpu_stack[1]), 10/* 80 bits */);
    bcopy((char *)(u->u_fpusave.fpu_stack[2]), (char *)(wp->p_regs.pr_fpu.fpu_stack[2]), 10/* 80 bits */);
    bcopy((char *)(u->u_fpusave.fpu_stack[3]), (char *)(wp->p_regs.pr_fpu.fpu_stack[3]), 10/* 80 bits */);
    bcopy((char *)(u->u_fpusave.fpu_stack[4]), (char *)(wp->p_regs.pr_fpu.fpu_stack[4]), 10/* 80 bits */);
    bcopy((char *)(u->u_fpusave.fpu_stack[5]), (char *)(wp->p_regs.pr_fpu.fpu_stack[5]), 10/* 80 bits */);
    bcopy((char *)(u->u_fpusave.fpu_stack[6]), (char *)(wp->p_regs.pr_fpu.fpu_stack[6]), 10/* 80 bits */);
    bcopy((char *)(u->u_fpusave.fpu_stack[7]), (char *)(wp->p_regs.pr_fpu.fpu_stack[7]), 10/* 80 bits */);
#endif
}

/*
 * Read the user area information from the core dump.
 */

public boolean coredump_readin(filename)
String filename;
{
    register struct user *up;
    union {
        struct user u;
        char dummy[ctob(UPAGES)];
    } ustruct;
    char *objname;
    File objfile = NULL;
    Object *o;

    nosyserrs();

    /*
     * open the core file.
     */
    corefile = coreopen(&lastcore, filename, corefile);
    if (corefile == NULL) {
	beginerrmsg();
	perror(filename);
	syserrs();
	return(false);
    }

    /*
     * read the u-area out of core file.
     */
    up = &(ustruct.u);
    fseek(corefile, 0, 0);
    if (fread((char *)up, ctob(UPAGES), 1, corefile) != 1) {
	beginerrmsg();
	fprintf(stderr, "%s: not a core file\n", filename);
	syserrs();
	return(false);
    }

    /*
     * open the associated object file for text reads and segment sizes.
     */
    objname = up->u_comm;
    o = create_object(objname);
    objfile = coreopen(&lastobj, o->o_pathname, objfile);
    if (objfile == NULL) {
	beginerrmsg();
	fprintf(stderr, "%s: no corresponding object file `%s'\n", filename, objname);
	syserrs();
	return(false);
    }

    syserrs();

    /*
     * fill in process structure from core info.
     */
    curobj = wp->p_obj = o;
    fseek(objfile, 0, 0);
    get(objfile, o->o_hdr);
    getregs(&ustruct);
    wp->p_stopinfo = BY_EVENT;
    wp->p_reason = up->u_arg[0];
    wp->p_corename = strdup(filename);
    o->o_datamap.seekaddr = ctob(UPAGES);
    o->o_stkmap.begin = MAXSTKADDR - ctob(up->u_ssize);
    o->o_stkmap.end = MAXSTKADDR;
    /* subtracting u_tsize from u_dsize for merged text+mmap in V3.0 */
    o->o_stkmap.seekaddr = o->o_datamap.seekaddr + ctob(up->u_dsize - up->u_tsize);
    if (o->o_hdr.a_magic == XMAGIC || o->o_hdr.a_magic == ZMAGIC) {
	o->o_datamap.begin = (Address) ptob(btop(ctob(up->u_tsize) - 1) + 1);
	o->o_datamap.end = o->o_datamap.begin + ctob(up->u_dsize - up->u_tsize);
    } else {
	beginerrmsg();
	fprintf(stderr, "%s: bad magic number\n", o->o_pathname);
	return(false);
    }
    return(true);
}

public coredump_readtext(buff, addr, nbytes)
char *buff;
Address addr;
int nbytes;
{
    if (curobj == nil)
	error("no object file");
    if (addr < N_ADDRADJ(curobj->o_hdr)) {
	if (curobj->o_hdr.a_magic == ZMAGIC) {
	    bzero(buff, (unsigned)nbytes);
	    return;
	} else /* curobj->o_hdr.a_magic == XMAGIC */
	    error("text address 0x%x invalid", addr);
	    /*NOTREACHED*/
    }
    objfile = coreopen(&lastobj, curobj->o_pathname, objfile);
    if (objfile == NULL) 
	error("can't open %s\n", curobj->o_pathname);
    fseek(objfile, (long) (N_TXTOFF(curobj->o_hdr) + addr - N_ADDRADJ(curobj->o_hdr)), 0);
    fread(buff, nbytes, sizeof(Byte), objfile);
}

public coredump_readdata(buff, addr, nbytes)
char *buff;
Address addr;
int nbytes;
{
    Address a;

    if (curobj == nil)
	error("no object file");
    a = addr;
    if (a < curobj->o_datamap.begin)
	    coredump_readtext(buff, a, nbytes);
    else if (a > curobj->o_stkmap.end)
	error("data address 0x%x too high (ub = 0x%x)", a, curobj->o_stkmap.end);
    else {
	if (wp->p_corename == nil)
	    error("no core file");
	corefile = coreopen(&lastcore, wp->p_corename, corefile);
	if (corefile == NULL) 
	    error("can't open %s\n", wp->p_corename);
	readfromfile(corefile, a, buff, nbytes);
    }
}

public coredump_readuser(buff, addr, nbytes)
char *buff;
Address addr;
int nbytes;
{
    if (wp->p_corename == nil)
	error("no core file");
    corefile = coreopen(&lastcore, wp->p_corename, corefile);
    if (corefile == NULL) 
	error("can't open %s\n", wp->p_corename);
    fseek(corefile, addr, 0);
    fread(buff, nbytes, 1, corefile);
}

private readfromfile (f, a, buff, nbytes)
File f;
Address a;
char *buff;
integer nbytes;
{
    integer fileaddr;

    if (a < curobj->o_stkmap.begin) {
	fileaddr = curobj->o_datamap.seekaddr + a - curobj->o_datamap.begin;
    } else {
	fileaddr = curobj->o_stkmap.seekaddr + a - curobj->o_stkmap.begin;
    }
    fseek(f, (long)fileaddr, 0);
    fread(buff, nbytes, sizeof(Byte), f);
}

public int errcode()
{
	int code;
	uread(&code, &((struct user *) 0)->u_code, sizeof(code));
	return(code);
}

private char uname[MAXNAMLEN + 1];
public String processname()
{
	uread(uname, ((struct user *) 0)->u_comm, sizeof(uname));
	return(uname);
}
