/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef lint
static char rcsid[] = "$Header: dbgparse.c 2.22 87/06/26 $";
#endif

/*
 * dbgparse.c -- parse a `.dbg' file
 */

#include "defs.h"
#include "dbgsym.h"
#include "dbgparse.h"
#include "machine.h"
#include "mappings.h"
#include "names.h"

#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef public
/*
 * structure used to pass fortran array bound information to dbgsym.c.
 */
#define USUALSIZE 1
#define ALLOWDEBUG			/* allow debugging output */
struct fortdim {
	int	fdim_lbound;
	int	fdim_hbound;
	int	fdim_size;
	char	fdim_flags;
};
struct fortdim usualtab[USUALSIZE];
struct fortdim *fdimtab;		/* to pass fortran definitions */

struct reg_save_mask {			/* used while readig mask info */
	unsigned short r_mask;	/* general registers & 387 fp regs */
	unsigned long  w_mask;	/* wytek registers */
};
#endif public

#define	NOSAVE	0
#define SAVE	1
#define SKIP	2

struct dbginfo {
	Name		name;	/* name of module with symbolic info */
	int		offset;	/* offset of info in dbg file */
	struct dbginfo	*next; 	/* pointer to next info record  */
};

private	File	dbg;		/* dbg file descriptor */
private int	dbg_debug = 0;	/* trace reading dbg file? */
private int	dbg_dump = 0;	/* trace bytes of dbg file, too */
private struct dbginfo *infolist;/* which modules have symbolic info */

private char	**fname_tab;	/* names of source files */
private char	*dbgfilename;

#define skip(n)	fseek(dbg, (n), 1);
#define dprintf (!dbg_debug) ? 0 : printf

#define CVRT_VERSION	0xF0		/* debug file version we expect */

private char *getname();
private unsigned int getdbg();
#ifdef ns32000
private unsigned short convert_svs_32k_regs();
#endif
#ifdef i386
private unsigned short convert_svs_386_regs();
#endif

/*
 * determine if there is a dbg file with the given name.  if so, open it
 * (fd in 'dbg') and scan through to find which modules have symbolic
 * info contained inside.
 */
public dbg_open(dbg_file_name)
	char	*dbg_file_name;
{
	struct stat s;
	long loc;
	unsigned int file_size;
	char *name;
	char *suffix;
	struct dbginfo *dp;
	unsigned char fmt;

	/*
	 * is the dbg file there?
	 */
	dbg = fopen(dbg_file_name, "r");
	if (dbg == NULL)
		return;

	strip_ = true;		/* must remove trailing underscores */
				/* as symbols are read from a.out   */
	/*
	 * a ".dbg" file can consist of multiple concatenated ".dbg" files.
	 * go through and find out what the original files were.
	 */

	(void) stat(dbg_file_name, &s);
	while ((loc = ftell(dbg)) < s.st_size) {

		fmt = getdbg(1);		/* file format */
		if (fmt != CVRT_VERSION) {
		    dbg_close();
		    printf("\nBad .dbg file found.\n");
		    printf("Saw version 0x%x, exepected 0x%x\n",
		      fmt, CVRT_VERSION);
		    if (fmt == 0x0)
		    	printf("Version 0 indicates old compiler used.\n");
		    if (fmt == 0x1)
		       printf("Version 1 indicates conversion program not run.\n");
		    return;
		}
		file_size = getdbg(4);		/* file size */
#ifdef ALLOWDEBUG
		dprintf("file size: %d\n", file_size);
#endif ALLOWDEBUG

		/*
		 * get the name of this dbg file, and salt it away in the
		 * info list.  later on, when we encounter a ".o" with the
		 * same name as the dbg file, the symbolic info at 'loc'
		 * will be read.
		 */
		name = getname(NOSAVE);
#ifdef ALLOWDEBUG
		dprintf("name: '%s'\n", name);
#endif ALLOWDEBUG
		suffix = rindex(name, '.');
		if (suffix && streq(suffix, ".dbg"))
			*suffix = '\0';
		dp = (struct dbginfo *)malloc(sizeof(struct dbginfo));
		dp->name = identname(name, false);
		dp->offset = loc + 5;
		dp->next = infolist;
		infolist = dp;
#ifdef ALLOWDEBUG
		dprintf("saving loc: %d\n", loc);
#endif ALLOWDEBUG

		/*
		 * skip to end of this dbg file.
		 */
		fseek(dbg, (long) (loc + file_size), 0);
	}
}

/*
 * close 'dbg' file descriptor, and free space 
 */
public dbg_close()
{
	struct dbginfo *dp;
	
	if (dbg != NULL)
		fclose(dbg);
	dbg = NULL;

	for (dp = infolist; dp; dp = dp->next)
		free(dp);
	infolist = 0;
}

/*
 * read in symbolic info for module 'modname', if there is any.  this is where
 * fortran and pascal symbolic info comes from.
 */
public dbg_read(modname, mod_offset)
	Name modname;
	int mod_offset;
{
	struct dbginfo *dp;

	for (dp = infolist; dp != 0; dp = dp->next)
		if (dp->name == modname)
			break;

	if (dp) {
#ifdef ALLOWDEBUG
		dprintf("found '%s' at %d\n", ident(modname), dp->offset);
#endif ALLOWDEBUG
		fseek(dbg, dp->offset, 0);
		sourcename();
		vars_and_types(mod_offset);
		statement_offsets(mod_offset);
		dbg_modend();
		free((char *)fname_tab);
	}
}


/*
 * source names.
 */

private sourcename()
{
	register Byte *bp, len;
	int nfiles, finfo_size;
	unsigned int strsize;
	char **fp;
	Byte *strtab;

	/*
	 * Read in file table.  An extra byte is allocated to leave
	 * room for a trailing '\0' on the last name when the names
	 * are converted to C strings.
	 */

	dbgfilename = getname(SAVE);
	nfiles = getdbg(4);
	finfo_size = getdbg(4);
#ifdef ALLOWDEBUG
	dprintf("SOURCENAME\n");
	dprintf("dbgfilename: '%s'\n", dbgfilename);
	dprintf("nfiles: %d\n", nfiles);
	dprintf("file info size: %d\n", finfo_size);
#endif ALLOWDEBUG

	strsize = finfo_size + 1;

	if (strsize != 1) {
		strtab = (Byte *)malloc(strsize);
		fread((char *)strtab, strsize - 1, 1, dbg);
		strtab[strsize - 1] = 0;

		/*
		 * count number of files in the file name table and allocate
		 * name table for them.  this is an array of pointers to 
		 * strings which will be indexed by the 'file name index'
		 * in statement_offsets().
		 */
		nfiles = 0;
		for (bp = strtab; bp <= &strtab[strsize]; bp += (*bp + 1))
			nfiles++;
		fname_tab = (char **)malloc(nfiles * sizeof(char *));

		/*
		 * fill in name table and convert it.  it's made up of strings
		 * where the leading byte is the length.  translate these into
		 * strings with a trailing null.
		 */
		bp = (Byte *)strtab;
		fp = fname_tab;
		len = *bp;
		while (len != 0) {
		     *bp = 0;
		     bp++;
		     *fp++ = (char *)bp;
		     bp += len;
		     len = *bp;
		}
	}
}

/*
 * variable and type definitions.
 */
private vars_and_types(mod_offset)
	int mod_offset;
{
	Byte lang, level;
	short namelid, outlid;
	char *name, *outer, *user;
	boolean first;
	int offset1, offset2;

#ifdef ALLOWDEBUG
	dprintf("VARS_AND_TYPES\n");
#endif ALLOWDEBUG
	first = true;
	while(getdbg(1) != 0xff) {
		skip(-1);
		startline("proc rec");
#ifdef ALLOWDEBUG
		dprintf("file loc: %d\n", ftell(dbg));
#endif ALLOWDEBUG
		lang = getdbg(1);	/* language: 0==>pascal, 1==>fortran */
		level =	getdbg(1);	/* static level */
#ifdef ALLOWDEBUG
		dprintf("lang: %d\n", lang);
		dprintf("level: %d file loc: %d\n", level, ftell(dbg));
#endif ALLOWDEBUG
		namelid = getdbg(2);	/* link id for name */
		name = getname(NOSAVE);	/* name */
		offset1 = getdbg(4);	/* entry point offset */
#ifdef ALLOWDEBUG
		dprintf("link id: 0x%x\n", namelid);
		dprintf("name: '%s', offset: 0x%x\n", name, offset1);
		dprintf("file loc = %d\n", ftell(dbg));
#endif ALLOWDEBUG
		outlid = getdbg(2);	/* link id for outer name */
		outer =	getname(NOSAVE);  /* outer name */
		offset2 = getdbg(4);	/* entry point offset */
		user = getname(NOSAVE);	/* user name */
#ifdef ALLOWDEBUG
		dprintf("outer: 0x%x  '%s', offset: 0x%x\n",
			outlid, outer, offset2);
#endif ALLOWDEBUG
		if (first) {
			dbg_modbegin((int)lang);
			first = false;
		}
		dbg_func(lang, level, name, namelid, offset1,
			 outer, outlid, offset2, user, mod_offset);
#ifdef ALLOWDEBUG
		dprintf("types, file loc: %d\n", ftell(dbg));
#endif ALLOWDEBUG
		types();
#ifdef ALLOWDEBUG
		dprintf("variables, file loc: %d\n", ftell(dbg));
#endif ALLOWDEBUG
		variables();
	}
}


/*
 * deal with type definitions.
 */
private types()
{
	int packed;
	short typeno;
	Byte kind;
	Byte c;
	short s, s1, s2;
	long l, l1, l2;
	int i;
	struct fortdim *dp;

#ifdef ALLOWDEBUG
	dprintf("TYPES\n");
	startline("typeinfo rec (1)");
#endif ALLOWDEBUG
	while( (typeno=getdbg(2)) != 0 ) {
#ifdef ALLOWDEBUG
		dprintf("type #%d ", typeno);
#endif ALLOWDEBUG
		kind = getdbg(1);
		packed = kind&0x10; /* set if packed type */
		kind &= 0xf;
		switch(kind) {
			case 0:	/* scalar */
				s = getdbg(2);	/* scalar max value */
#ifdef ALLOWDEBUG
				dprintf("scalar maxval %d\n", s);
#endif ALLOWDEBUG
				dbg_type(kind, typeno, s);
				break;
			case 1: /* subrange */
				s = getdbg(2);	/* subrange of */
				l1 = getdbg(4);	/* min val */
				l2 = getdbg(4);	/* max val */
#ifdef ALLOWDEBUG
				dprintf("subrange of %d %d..%d\n", s, l1, l2);
#endif ALLOWDEBUG
				dbg_type(kind, typeno, s, l1, l2);
				break;
			case 2:	/* pointer */
				s = getdbg(2);	/* pointer to */
#ifdef ALLOWDEBUG
				dprintf("pointer to %d\n", s);
#endif ALLOWDEBUG
				dbg_type(kind, typeno, s);
				break;
			case 3:	/* set */
				s = getdbg(2);	/* set of */
#ifdef ALLOWDEBUG
				dprintf("set of %d\n", s);
#endif ALLOWDEBUG
				dbg_type(kind, typeno, s);
				break;
			case 4: /* array */
#ifdef ALLOWDEBUG
				if (packed)
					dprintf("packed ");
#endif ALLOWDEBUG
				s1 = getdbg(2);	/* indexed by */
				s2 = getdbg(2);	/* array of */
#ifdef ALLOWDEBUG
				dprintf("array of %d ", s2);
				dprintf("indexed by %d ", s1);
#endif ALLOWDEBUG
				if (packed) {
					c = getdbg(1);	/* packinfo */
#ifdef ALLOWDEBUG
					dprintf("packinfo=0x%x", c);
#endif ALLOWDEBUG
				} else 
					c = 0;
#ifdef ALLOWDEBUG
				dprintf("\n");
#endif ALLOWDEBUG
				dbg_type(kind, typeno, c, s2, s1);
				break;
			case 5: /* string */
				c = getdbg(1);		/* str length */
#ifdef ALLOWDEBUG
				dprintf("string length %d\n", c);
#endif ALLOWDEBUG
				dbg_type(kind, typeno, c);
				break;
			case 6:	/* file */
				s = getdbg(2);		/* file of */
#ifdef ALLOWDEBUG
				dprintf("file of %d\n", s);
#endif ALLOWDEBUG
				dbg_type(kind, typeno, s);
				break;
			case 7: /* record */
				l = getdbg(4);		/* record size */
#ifdef ALLOWDEBUG
				if (packed)
					dprintf("packed ");
				dprintf("record size %d\n", l);
#endif ALLOWDEBUG
				dbg_type(kind, typeno, packed, l);
#ifdef ALLOWDEBUG
				startline("rec fld rec (1)");
#endif ALLOWDEBUG
				while (getdbg(1) != 0) {
					skip(-1);
					dofield(packed);
					startline("rec fld rec (2)");
				}
				break;
			case 9:	/* character */
				s = getdbg(2);		/* char length */
#ifdef ALLOWDEBUG
				dprintf("character length %d\n", (int)s);
#endif ALLOWDEBUG
				dbg_type(kind, typeno, s);
				break;

			case 10: /* fortran array */
				c = getdbg(1);		/* dims */
				s = getdbg(2);		/* array of type */
#ifdef ALLOWDEBUG
				dprintf("fortran array of %d dim %d (", s, c);
#endif ALLOWDEBUG
				if (c > USUALSIZE)
					fdimtab = (struct fortdim *)malloc(c * sizeof(struct fortdim));
				else 
					fdimtab = usualtab;
				dp = fdimtab;
				for (i = 0; i < c; i++) {
					dp->fdim_flags = getdbg(1);
					dp->fdim_lbound = getdbg(4); 
					dp->fdim_hbound = getdbg(4);
					dp->fdim_size = getdbg(4);
#ifdef ALLOWDEBUG
					if (dp->fdim_flags & 0x1) 
						dprintf("*");
					dprintf("%d..", dp->fdim_lbound);
					if (dp->fdim_flags & 0x2) 
						dprintf("*");
					dprintf("%d-", dp->fdim_hbound);
					if (dp->fdim_flags)
						dprintf("*");
					dprintf("%d", dp->fdim_size);
					if (i != (c-1))
						dprintf(",");
#endif ALLOWDEBUG
					dp++;
				}
#ifdef ALLOWDEBUG
				dprintf(")\n");
#endif ALLOWDEBUG
				dbg_type(kind, typeno, s, c);
				if (c > USUALSIZE)
					free(fdimtab);
				break;
			default:
				error("types: unknown kind %d", kind);
				break;
		}
#ifdef ALLOWDEBUG
		startline("typeinfo rec (2)");
#endif ALLOWDEBUG
	}
}


private dofield(packed)
	int packed;
{
	unsigned int offset;
	short fieldtype;
	unsigned char leftbit, numbits;
	char *name;

	name = getname(SAVE);
	fieldtype = getdbg(2);
	offset = getdbg(4) * 8;
#ifdef ALLOWDEBUG
	dprintf("field %s offset %d", name, offset);
#endif ALLOWDEBUG
	numbits = 0;
	if (packed) {
		numbits = getdbg(1);
		leftbit = getdbg(1);
		offset += leftbit;
#ifdef ALLOWDEBUG
		dprintf(" packed, leftbit = %d numbits = %d", leftbit, numbits);
#endif ALLOWDEBUG
	}
#ifdef ALLOWDEBUG
	dprintf("\n");
#endif ALLOWDEBUG
	dbg_field(name, fieldtype, offset, numbits);
}


/*
 * deal with variables.  note that getlocation() does a getname(NOSAVE)
 * too, but this is safe since getname() maintains two buffers and
 * returns alternate ones.
 */
private variables()
{
	short vtype;
	char *name;
	struct dbgmode m;

#ifdef ALLOWDEBUG
	dprintf("VARIABLES\n");
	startline("var rec (1)");
#endif ALLOWDEBUG
	while(getdbg(1) != 0) {
		skip(-1);
		name = getname(NOSAVE);
		vtype = getdbg(2);
#ifdef ALLOWDEBUG
		dprintf("variable '%s' type=%d", name, vtype);
		if (dbg_debug)
			fflush(stdout);
#endif ALLOWDEBUG
		getlocation(&m);
		dbg_var(name, vtype, &m);
#ifdef ALLOWDEBUG
		dprintf("\n");
#endif ALLOWDEBUG
		startline("var rec (2)");
		if (dbg_debug)
			fflush(stdout);
	}
}


/*
 * find address and class for variable
 */
private getlocation(m)
	struct dbgmode *m;
{
	unsigned char format;
	unsigned char w_reg;
	short s;
	short regaddr;

	format = getdbg(1);
	m->m_rmode = format&0xf;	/* storage mode */
	m->m_regnum = (format>>4)&0xf;	/* which register number */
	m->m_linkname = NULL;		/* for debugging */
	m->m_linkid = 0;

#ifdef ALLOWDEBUG
	dprintf("mode: %02d  %02d  (0x%x)\n", ((format >> 4) &0xf),
	    (format & 0xf), format);
#endif ALLOWDEBUG

	/*
	 * Register conversions are done in this routine.
	 * The conversion is dependent upon the storage mode.
	 *
	 * See .dbg info for "Variable Descriptors".
	 */

	m->m_indirect = (m->m_rmode & 1);	/* odd modes are indirect */
	switch(m->m_rmode) {
		case 0: /* fp or sb with 2 byte offset*/
		case 1: /* fp or sb pointer with 2 byte offset*/
			m->m_regnum = reg_conv01(m->m_regnum);
			s = getdbg(2);		/* XXX compiler bug */
			m->m_offset = s;	/* XXX compiler bug */
			break;
		case 6: /* fp or sb with 4 byte offset*/
		case 7: /* fp or sb pointer with 4 byte offset*/
			m->m_regnum = reg_conv01(m->m_regnum);
			m->m_offset = getdbg(4);
			break;
		case 2: /* register */
		case 3: /* register pointer */
			m->m_regnum = reg_conv23(m->m_regnum);
			break;
		case 4: /* external data area */
			m->m_rmode = -1;
			m->m_linkid = getdbg(2);
			m->m_linkname = getname(NOSAVE);
			m->m_offset = getdbg(4);
			break;
#ifdef i386
		case 8:
			w_reg = getdbg(1);
			if (w_reg < 2 or w_reg > 31)
			    error("dbg getlocation: bad weitek register %d\n", w_reg);
			m->m_regnum = w_reg + FPA_BASE - 1;
			break;
#endif i386
		default:
			error("dbg getlocation: unknown mode %d", m->m_rmode);
	}

	/*
	 * Next two bytes are added by dbgcvrt program.  They are for
	 * future use when we want to debug optimized code with
	 * register spill and load.
	 */

	regaddr = getdbg(2);
	if ((regaddr & 0xffff) != 0xffff) {
		error("dbg getlocation: cannot deal with reg addr 0x%x",
		    regaddr);
	}
}

/*
 * Make a conversion of the register number from the mapping used by SVS to
 * the registers numbers used by the debugger.  These are generally set
 * up for direct mapping with the saved reg struct.
 *
 * Two versions of this routine are used, the "01" routine is for storage
 * modes 0 & 1 (same as 6 & 7), the "23" routine is for storage modes 2 & 3.
 */

private reg_conv01(n)
{
#ifdef ns32000
	switch (n) {
	case 0: case 1: case 2: case 3:		/* r0 ... r4 */
	case 4: case 5: case 6: case 7:		/* r5 ... r7 */
		return(n);
	case 13:
		return(SB);
	case 14:
		return(FRP);
	default:
		printf("\nreg_conv: bogus register used: %d\n", n);
		error("cannot use reg for storage mode 0, 1");
	}
#endif ns32000
#ifdef i386
	switch (n) {
	case 0:
		return(REG_EAX);	/* r0 */
	case 1:
		return(REG_ECX);	/* r1 */
	case 2:
		return(REG_EDX);	/* r2 */
	case 3:
		return(REG_EBX);	/* r5 */
	case 4:
		return(REG_ESP);	/* sp */
	case 5:
		return(REG_EBP);	/* fp */
	case 6:
		return(REG_ESI);	/* r6 */
	case 7:
		return(REG_EDI);	/* r7 */
	case 8:
		return(REG_FP0);	/* fp0 */
	case 9:
		return(REG_FP1);	/* fp1 */
	case 10:
		return(REG_FP2);	/* fp2 */
	case 11:
		return(REG_FP2);	/* fp3 */
	case 12:
		return(REG_FP4);	/* fp4 */
	case 13:
		return(REG_FP5);	/* fp5 */
	case 14:
		return(REG_FP6);	/* fp6 */
	case 15:
		return(REG_FP7);	/* fp7 */
	default:
		printf("\nreg_conv: bogus register used: %d\n", n);
		panic("cannot use reg for storage mode 0, 1");
	}
#endif i386
}

private reg_conv23(n)
{
#ifdef ns32000
	switch (n) {
	case 0: case 1: case 2: case 3:		/* r0 ... r3 */
	case 4: case 5: case 6: case 7:		/* r4 ... r7 */
	case 8: case 9: case 10: case 11:	/* f0 ... f4 */
	case 12: case 13: case 14: case 15:	/* f5 ... f7 */
		return(n);
	default:
		printf("\nreg_conv: bogus register used: %d\n", n);
		panic("cannot use reg for storage mode 0, 1");
	}
	return(n);	/* already mapped for 32K */
#endif ns32000
#ifdef i386
	switch (n) {
	case 0:
		return(REG_EAX);	/* r0 */
	case 1:
		return(REG_ECX);	/* r1 */
	case 2:
		return(REG_EDX);	/* r2 */
	case 3:
		return(REG_EBX);	/* r5 */
	case 4:
		return(REG_ESP);	/* sp */
	case 5:
		return(REG_EBP);	/* fp */
	case 6:
		return(REG_ESI);	/* r6 */
	case 7:
		return(REG_EDI);	/* r7 */
	case 8:
		return(REG_FP0);	/* fp0 */
	case 9:
		return(REG_FP1);	/* fp1 */
	case 10:
		return(REG_FP2);	/* fp2 */
	case 11:
		return(REG_FP2);	/* fp3 */
	case 12:
		return(REG_FP4);	/* fp4 */
	case 13:
		return(REG_FP5);	/* fp5 */
	case 14:
		return(REG_FP6);	/* fp6 */
	case 15:
		return(REG_FP7);	/* fp7 */
	default:
		printf("\nreg_conv: bogus register used: %d\n", n);
		panic("cannot use reg for storage mode 2, 3");
	}
#endif i386
}


/*
 * statement offset info.
 */
private statement_offsets(mod_offset)
	int mod_offset;
{
	short numentries, nstatements;
	Byte reginfo;
	int entry;
	int fp_disp;
	int i;
	short linkid, firstlink;
	int addr, lineno;
	char *ename;
	boolean first_time = true;
	short fileindex, lastfindex;
	Linetab dummy;
	int firstl;
	register Linetab *lastlp;
	int nlines;
	char *params, *getparamlist();
	int entry_offset;
	int f_addr;
	struct reg_save_mask regmask;
	unsigned char cpu_regs, fp_regs;

#ifdef lint
	/* lastfindex won't really be used before set because of first_time */
	lastfindex = 0;
#endif

#ifdef ALLOWDEBUG
	startline("#line info recs (1)");
	dprintf("file loc: %d\n", ftell(dbg));
#endif ALLOWDEBUG
	while ((numentries=getdbg(2)) != 0) {
		reginfo = getdbg(1);
		regmask.r_mask = 0;
		regmask.w_mask = 0;
		cpu_regs = 0;
		fp_regs = 0;

		switch (reginfo) {
		case 0:			/* no register save info */
			break;
#ifdef ns32000
		case 2:			/* 32K reg info */
			cpu_regs = getdbg(1);
			fp_regs = getdbg(1);
			regmask.r_mask = convert_svs_32k_regs(cpu_regs, fp_regs);
			fp_disp = getdbg(4);
			break;
#endif ns32000
#ifdef i386
		case 4:			/* 386 reg info */
			cpu_regs = getdbg(1);
			regmask.r_mask = convert_svs_386_regs(cpu_regs, fp_regs);
			regmask.w_mask = getdbg(4);
			fp_disp = getdbg(4);
			break;
		case 5:			/* 386 reg info */
			cpu_regs = getdbg(1);
			fp_regs = getdbg(1);
			regmask.r_mask = convert_svs_386_regs(cpu_regs, fp_regs);
			fp_disp = getdbg(4);
			break;
#endif i386
		default:
			error("dbg statement_offsets: reginfo is %d\n", reginfo);
		}

		for (i = 0; i < numentries; i++) {
#ifdef ALLOWDEBUG
			startline("entrypt rec ");
			dprintf("file loc: %d\n", ftell(dbg));
#endif ALLOWDEBUG
			linkid = getdbg(2);
			if (i == 0)
				firstlink = linkid;
			ename = getname(NOSAVE);
			entry = getdbg(4);
			entry_offset = getdbg(4);
#ifdef ALLOWDEBUG
			dprintf("link id %d '%s' entry at 0x%x, offset: 0x%x\n",
			    linkid, ename, entry, entry_offset);
#endif ALLOWDEBUG
			params = getparamlist();
			dbg_entry(numentries, i == 0 ? 0 : numentries - i,
				ename, linkid, entry, regmask, fp_disp,
				mod_offset, entry_offset);
			dbg_param(params);
		}
		dummy.line = 0x7fffffff;
		dummy.addr = 0xffffffff; 	/* addr is unsigned */
		lastlp = &dummy;
		firstl = linep - linetab;
#ifdef ALLOWDEBUG
		startline("#statements");
#endif ALLOWDEBUG
		nlines = nstatements = getdbg(2);
		addr = 0;
		for (i = 0; i < nstatements; i++) {
			addr += getdbg(2);
			lineno = getdbg(4);
			fileindex = getdbg(2);
			if (lineno == 0) {
				nlines--;
				continue;
			}
			if (fileindex == -1) {
				panic("fileindex == -1 in '%s'", dbgfilename);
			}
			if (first_time || fileindex != lastfindex) {
				f_addr = addr + mod_offset + entry_offset;
				dbg_newfile(f_addr, fname_tab[fileindex]);
				first_time = false;
				lastfindex = fileindex;
			}
			if (linep == endlinetab) {
				realloc_linetab();
				lastlp = linep - 1;
			}
			if (lastlp->line == lineno) {
				lastlp->addr = min(addr, lastlp->addr);
				nlines--;
			} else if (lastlp->addr == addr) {
				lastlp->line = lineno;
				nlines--;
			} else {
				linep->addr = addr;
				linep->line = lineno;
				lastlp = linep;
				linep++;
			}
		}
		dbg_setlines(numentries, nlines, firstl);
#ifdef ALLOWDEBUG
		startline("#line info recs (2)");
#endif ALLOWDEBUG
	}
}

/*
 * read in a list of names, convert to C strings, and return a pointer
 * to the list.  the list is terminated by a name with length zero.
 * the list will hang out until this routine is called again.
 */
private char *getparamlist()
{
	Byte *bp, len;
	long bytesofparams;
	static char *parambuf = 0;
	static long pallocsize = 0;

	bytesofparams = getdbg(4);
#ifdef ALLOWDEBUG
	dprintf("param list size %d at loc %d\n", bytesofparams,
	    (ftell(dbg) - 4));
#endif ALLOWDEBUG
	if (pallocsize < bytesofparams + 2) {
		if (parambuf)
			free(parambuf);
		parambuf = (char *)malloc(bytesofparams + 2);
		pallocsize = bytesofparams + 2;
	}
	parambuf[bytesofparams] = '\0';
	parambuf[bytesofparams + 1] = '\0';
	if (bytesofparams == 0)
		return(parambuf);
	fread(parambuf, bytesofparams, 1, dbg);
	bp = (Byte *)parambuf;
	len = *bp;
	while (len != 0) {
		*bp = 0;
		bp++;
		bp += len;
		len = *bp;
	}
	return(&parambuf[1]);
}


/*
 * get a name out of the dbg file, allocating space for it if 
 * requested, otherwise reading it into a static buffer.  three
 * static buffers are maintained, taking sequenced requests.
 * this means that three (only) getname(NOSAVE) calls can be made without
 * destroying the static data.
 */
private char	namebuf1[256+1];	/* place to read names into */
private char	namebuf2[256+1];	/* place to read names into */
private char	namebuf3[256+1];	/* place to read names into */
private int	bufswitch;		/* select which buffer to use */
private char	noname[] = "!!noname!!";

private char *getname(alloc)
	int alloc;
{
	register Byte namelen;
	register char *s;

	namelen = getdbg(1);
	if (alloc == SKIP) {
		skip(namelen);
		return(noname);
	}
	if (alloc == SAVE)
		s = (char *)stralloc((int)(namelen + 1));
	else {
		switch (bufswitch) {
		case 0:
			s = namebuf1;
			bufswitch = 1;
			break;
		case 1:
			s = namebuf2;
			bufswitch = 2;
			break;
		case 2:
			s = namebuf3;
			bufswitch = 0;
			break;
		default:
			panic("bufswitch botched in getname");
		}
	}
	if (namelen != 0)
		fread(s, namelen, 1, dbg);
	s[namelen] = '\0';
	return(s);
}


/*
 * return an size-byte integer from the dbg file
 */
private unsigned int getdbg(size)
	int size;
{
	unsigned char  v_1;
	unsigned short v_2;
	unsigned long  v_4;
	int nread;
	unsigned int i;

	switch (size) {
	case 1:
		nread = fread((char *)&v_1, size, 1, dbg);
		i = (unsigned int) v_1;
		break;
	case 2:
		nread = fread((char *)&v_2, size, 1, dbg);
		i = (unsigned int) v_2;
		break;
	case 4:
		nread = fread((char *)&v_4, size, 1, dbg);
		i = (unsigned int) v_4;
		break;
	default:
		error("dbg get: bad size");
		break;
	}
	if (nread != 1)
		error("dbg get: bad fread");
	return(i);

#ifdef notdef
	union {
		unsigned char	c[4];
		long		i;
	} buf;
	int nread;
	unsigned char t;

	buf.i = 0;
	nread = fread(&buf, size, 1, dbg);
	if (nread != 1)
		error("dbg get: bad fread");
	switch(size) {
	case 1:
		break;
	case 2:
		t = buf.c[0];
		buf.c[0] = buf.c[1];
		buf.c[1] = t;
		break;
	case 3:
		t = buf.c[0];
		buf.c[0] = buf.c[2];
		buf.c[2] = t;
		break;
	case 4:
		t = buf.c[0];
		buf.c[0] = buf.c[3];
		buf.c[3] = t;
		t = buf.c[1];
		buf.c[1] = buf.c[2];
		buf.c[2] = t;
		break;
	default:
		error("dbg get: bad size");
		break;
	}
	return(buf.i);
#endif notdef
}

/*
 * startline is called before starting to read a "new record".  records
 * are defined by what i decide to print out on one line in the output
 * of this program.  startline synchronizes a higher level object, 
 * named by the argument, with the current file pointer location.  thus
 * it can be used to show what's about to be interpreted.
 */

#define bytes_to_dump	14

private startline(s)
	char *s;
{
	unsigned int loc;
	register int i;
	unsigned char dumpbuf[bytes_to_dump];

	if (dbg_debug && dbg_dump) {
		loc = ftell(dbg);
		fread(dumpbuf, bytes_to_dump, 1, dbg);
		fseek(dbg, loc, 0);
		printf(" %-12.12s ", s);
		printf("%6d: ", loc);
		for (i = 0; i < bytes_to_dump; i++)
			printf("%02x ", dumpbuf[i]);
		printf("\"");
		for (i = 0; i < bytes_to_dump; i++) {
			register char c = dumpbuf[i];
			if (isprint(c) && c != 0xff && c != 0)
				putchar(c);
			else
				putchar('.');
		}
		printf("\"");
		printf("\n");
	}
}

/*
 * Map from the SVS compiler generated register mask used for general
 * registers and floating point registers to the bits used by the C
 * compiler.  The register numbers used by cpp will match the definitions
 * in "machine.c".  Note that the SVS compiler lists the saved registers
 * for procedures in the debug info.  The C compiler does not do this
 * so the information is picked up when we build a frame.  For the 32K we
 * just look at the mask on an enter instruction.  For the 386 we have
 * to disassemble the code after an entry and build a mask.
 */

#define bis(b, n)	( ((b) & ( 1 << (n))) != 0)
#define set(b, n)	(b) |= ( 1 << (n))

#ifdef ns32000
private unsigned short convert_svs_32k_regs(gen, fp)
	unsigned char gen, fp;
{
	unsigned short ret;

	ret = fp << 8;
	ret |= gen;

	return(ret);
}
#endif ns32000

#ifdef i386
#define SVS_REG_EAX	0
#define SVS_REG_ECX	1
#define SVS_REG_EDX	2
#define SVS_REG_EBX	3	/* ok to save */
#define SVS_REG_ESP	4
#define SVS_REG_EBP	5
#define SVS_REG_ESI	6	/* ok to save */
#define SVS_REG_EDI	7	/* ok to save */

#define SVS_REG_FP0	0
#define SVS_REG_FP1	1
#define SVS_REG_FP2	2
#define SVS_REG_FP3	3
#define SVS_REG_FP4	4
#define SVS_REG_FP5	5
#define SVS_REG_FP6	6
#define SVS_REG_FP7	7

private unsigned short convert_svs_386_regs(gen, fp)
	unsigned char gen, fp;
{
	unsigned short ret = 0;

	if (bis(gen, SVS_REG_EAX)) set(ret, REG_EAX);
	if (bis(gen, SVS_REG_EBX)) set(ret, REG_EBX);
	if (bis(gen, SVS_REG_ECX)) set(ret, REG_ECX);
	if (bis(gen, SVS_REG_ESI)) set(ret, REG_ESI);
	if (bis(gen, SVS_REG_EDI)) set(ret, REG_EDI);

	if (bis(gen, SVS_REG_EAX))
		printf("\nWarning, saw reg EAX saved in procedure\n");
	if (bis(gen, SVS_REG_ECX))
		printf("\nWarning, saw reg ECX saved in procedure\n");
	if (bis(gen, SVS_REG_EDX))
		printf("\nWarning, saw reg EDX saved in procedure\n");
	if (bis(gen, SVS_REG_ESP))
		printf("\nWarning, saw reg ESP saved in procedure\n");
	if (bis(gen, SVS_REG_EBP))
		printf("\nWarning, saw reg EBP saved in procedure\n");

	if (bis(fp, SVS_REG_FP0)) set(ret, REG_FP0);
	if (bis(fp, SVS_REG_FP1)) set(ret, REG_FP1);
	if (bis(fp, SVS_REG_FP2)) set(ret, REG_FP2);
	if (bis(fp, SVS_REG_FP3)) set(ret, REG_FP3);
	if (bis(fp, SVS_REG_FP4)) set(ret, REG_FP4);
	if (bis(fp, SVS_REG_FP5)) set(ret, REG_FP5);
	if (bis(fp, SVS_REG_FP6)) set(ret, REG_FP6);
	if (bis(fp, SVS_REG_FP7)) set(ret, REG_FP6);

	return(ret);
}
#endif i386
