/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef lint
static char rcsid[] = "$Header: dbgsym.c 2.25 87/10/27 $";
#endif

/*
 * dbgsym.c -- create symbol table entries from dbg file
 */

#include <ctype.h>
#include "defs.h"
#include "dbgsym.h"
#include "dbgparse.h"
#include "object.h"
#include "symbols.h"
#include "languages.h"
#include "stabstring.h"
#include "mappings.h"
#include "machine.h"
#include "main.h"
#include "names.h"

#ifndef public

/* #define DEBUGLINETAB	1	/* debugging of line table */
#define NEWDBG	1	/* new addressing information in .dbg files */

#define	NBUILTINS	16

/* XXX can we throw away the linkname and linkid for externs? */
struct dbgmode {
	int	m_rmode;	/* storage mode */
	int	m_regnum;	/* register number */
	int	m_offset;	/* offset from reg or base addr */
	int	m_indirect;	/* mode is indirect off reg */
	char 	*m_linkname;	/* name of base if ext addr */
	short	m_linkid;	/* linkid of base */
};

#endif public

#define whichlang(l)	((l) ? lang_fortran : lang_pascal)


private	Symbol	currec;		/* pointer to current record being made */
private Symbol	curentry;	/* pointer to current entrypoint */
private char	*progdata_name;
private short	progdata_linkid;
private char	*firstprocname;
private short	firstproclink;

char	*aoutname();
Symbol	getproc();


/*
 * since types can be referenced in the dbg file before they're defined,
 * we may have to defer assigning the type until the type is finally seen.
 * also, the size of an object may not be known until the type is seen.
 * these fixups are done when the end of the dbg file is reached.  see
 * dofixups() below.
 */
typedef enum {
	FIX_NOTHING = 0,	/* used to mark end of lists */
	FIX_TYPE,		/* fill in type info */
	FIX_TYPECHAIN,		/* fill in type info in 'chain' field */
	FIX_FIELDSIZE		/* fill in field size */
} fixup_op;

typedef struct fixup_rec {
	struct fixup_rec *f_next; /* next fixup record */
	Symbol		f_sym;	/* symbol to fix up */
	fixup_op	f_op;	/* what to do to it */
	short		f_type; /* type number for type fixups */
} fixup_rec;
private fixup_rec *fixup_list;	/* list of symbols to fix up */


/*
 * record used to fix up line numbers.  the dbg file lists lines as
 * offsets from the beginning of procedures.  since the address of
 * a procedure isn't known when the dbg file is read in, we have
 * to add in the address of the procedure when we finally know it.
 * see fixlines() below.
 */
struct fixline {
	Symbol		proc;
	int		line;
	int		nlines;
	struct fixline	*next;
	int		prologue_size;
};
private struct fixline *fixlinelist = 0;


/*
 * allocate fixline structures in blocks for space efficiency.
 */

#define FIXLINECHUNK POOLSIZE(4096, sizeof(struct fixline))

struct fixlinepool {
	struct fixline item[FIXLINECHUNK];
	struct fixlinepool *prevpool;
};
typedef struct fixlinepool fixlinepool;
private fixlinepool *fl_pool, *newpool;
private int fl_left = 0;
private int total_fixlines;

private struct fixline *alloc_fixline()
{
	fl_left--;
	if (fl_left < 0) {
		newpool = (fixlinepool *)malloc(sizeof(fixlinepool));
		newpool->prevpool = fl_pool;
		fl_pool = newpool;
		fl_left = FIXLINECHUNK - 1;
		total_fixlines++;
	}
	return(&fl_pool->item[fl_left]);
}

private free_fixlines()
{
	register fixlinepool *fp = fl_pool;
	while(fp) {
		free(fp);
		fp = fp->prevpool;
	}
	fixlinelist = 0;
	fl_left = 0;
}


/*
 * A new file has been encountered.  Add it to file file table.
 * It used to be that absolute addresses weren't know at this point so
 * a symbol would be created and later a fix made to the address of the
 * table.  New info in .dbg file provides correct info now so things are
 * much simpler.  What we are putting in the table is the address of the
 * first line of code in the file.  This is absolute address of file,
 * plus distance to first routine in file, plus distance to first
 * "real code" in that routine.
 * See the call in dbgparse.c : statement_offsets().
 */

private struct fixline *realentry;
private char realentry_name[256 + 1];
private Symbol lastentry, firstentry;

dbg_newfile(offset, filename)
	integer offset;
	char *filename;
{
	enterfile(filename, offset);
}


/*
 * a 'module' for pascal and fortran corresponds to a compilation
 * unit.  there is one dbg file for each compilation unit, so
 * this gets called when we start reading in a new dbg file.
 * don't have to have a block stack as in object.c since the level
 * of each procedure or function is given in the dbg file.
 */
public dbg_modbegin(language)
	int	language;
{
	curlang = whichlang(language);
	curmodule->language = curlang;
	if (program->language == nil)
		program->language = curlang;
	initTypeTable();
}


public dbg_modend()
{
	dofixups();
	linksym(curmodule, firstprocname, firstproclink);

	/*
	 * Setting "curblock" to "program" prevents unstacking the
	 * block (call to exitblock() will not be made).  This
	 * causes the debugger to abort once we exceed the MAXBLKDEPTH
	 * limit.  See array "blkstack" in file "object.c".
	 *
	 * Next line removed 14-Nov-86.
	 *
	 * curblock = program;
	 */
}


/*
 * deal with a variable that's in a common.
 */

private char internal_name[256 + 1];

private docommon(var, extname, linkid, offset)
	Symbol var;			/* the variable */
	char *extname;			/* external name of its common */
	short linkid;			/* linkid of common */
	int offset;			/* offset of var in common */
{
	register Symbol c;
	boolean different;		/* extname same as internal_name? */
	register Name n;
	boolean savedvar = false;	/* common for save'd variable? */
	String p;
	boolean do_lc;			/* we need to lower case internal name */

	/*
	 * Note that previous releases placed common areas inside
	 * slashes /comname/ , but now we use underscores, _comname_ .
	 *
	 * The flags "-e" and "+e" save the slashes for backwards
	 * compatibility.  (As well as changing arg passing order.)
	 */

	/*
	 * the external name for blank common is "//", but it's "/ /" 
	 * in the dbg file.   sigh.
	 */
	if (strcmp(extname, "/ /") == 0) {
		strcpy(extname, "//");
		strcpy(internal_name, "blank");
		different = true;
	}
	/*
	 * if not compiled with "-e" or "+e" then common has the underscores.
	 */
	 else if (strcmp(extname, "_ _") == 0) {
		strcpy(extname, "//");
		strcpy(internal_name, "blank");
		different = true;
	 }
	/*
	 * for normal commons, make the internal name by removing the slashes
	 * from the external name.  Test here is for common compiled with
	 * the "-e" or "+e" flag.
	 */
	else if (extname[0] == '/') {
		common_name(internal_name, extname, &different);
	}
	/*
	 * Without the "-e" or "+e" we will have the underscore, but it
	 * will be removed as symbols are read from the a.out file.
	 */
	 else if (extname[0] == '_') {
		common_name(internal_name, extname, &different);
		extname = internal_name;
		different = false;
	 }
	/*
	 * save'd variables are implemented by putting them in a 
	 * common named "=proc" where 'proc' is the name of the 
	 * procedure that contains the variable.
	 *
	 * We could determine if common or saved variable by
	 * checking for the "=", but presently assume it is saved
	 * variable.
	 */
	else {
		if (extname[0] != '=')
		    printf("\nWarning, docommon: no leading '=' in '%s'\n",
		      extname);
		strcpy(internal_name, extname);
		different = false;
		savedvar = true;
	}

	/*
	 * New compiler does the lower case conversion, so check to see
	 * if this step is necessary.  If we can avoid it we may be able
	 * to skip the call to linksym.
	 */

	p = internal_name;
	do_lc = false;

	while (*p) {
		if (isupper(*p))
			do_lc = true;
		p++;
	}
	if (do_lc) {
		lowercase(internal_name);
		if (not uppercase_flag)
			different = true;
	}

	/*
	 * Still have to compare what we ended up with because of the
	 * possibility of underscores or slashes.
	 */

	if (not streq(internal_name, extname))
		different = true;

	/*
	 * find the common, make a new symbol for it if necessary.
	 * there is a single COMMON symbol for each common name in the program.
	 */
	n = identname(internal_name, false);
	find(c, n) where
		c->class == COMMON
	endfind(c);
	if (c == nil) {
		c = insert(n); 
		if (different || linkid != -1)
			linksym(c, extname, linkid);
		c->class = COMMON;
		c->language = curlang;
		c->block = program;
		c->level = program->level + 1;
		c->symvalue.common.chain = nil;
		c->symvalue.common.fixed_addr = false;
	}

	/*
	 * insert the variable into the common.  the common-relative offsets
	 * are fixed up in check_global().  fixcommonvars() in object.c
	 * assumes the head of the list is in common.offset of 'c'.
	 * if this is a save'd variable instead of a real common, arrange
	 * that the variable is not "contained" in the common (since the
	 * common is an artifact of the implemetation).
	 */
	var->level = 1;
	if (not savedvar)
		var->chain = c;
	else
		var->chain = nil;
	var->symvalue.common.offset = offset;
	var->symvalue.common.chain = c->symvalue.common.chain;
	var->symvalue.common.fixed_addr = false;
	c->symvalue.common.chain = var;
	c->symvalue.common.offset = (integer) c->symvalue.common.chain;
}


/*
 * Make a common name usable by deleting all the slashes or underscores.
 * Copies source into target.
 */
private common_name(target, source, different)
	char *target;
	char *source;
	boolean *different;
{
	register char *head, *tail;
	*different = false;

	head = source;
	tail = target;
	while(*head != '\0') {
		if (*head == '/' or *head == '_')
			*different = true;
		else
			*tail++ = *head;
		head++;
	}
	*tail = '\0';
}

/*
 * make a symbol table entry (etc) for a procedure.  assumes 'name' and
 * 'outer' aren't permanently allocated.
 */

public dbg_func(lang, level, name, namelink, offset1,
		outer, outlink, offset2, user, mod_offset)
	Byte lang, level;
	char	*name;		/* name of the procedure to create sym for */
	short	namelink;	/* linkid for new procedure name */
	int	offset1;	/* offset for this entry point (name) */
	char	*outer;		/* name of procedure that contains 'name' */
	short	outlink;	/* linkid for container */
	int	offset2;	/* offset for this entry point (outer) */
	char	*user;		/* user name */
	int	mod_offset;	/* offset for the load addr of this module */
{
	register Symbol s;	/* symbol for the procedure or function */
	register Symbol t;	/* symbol for the outer name */
	Name	internal;	/* internal name */
	Name	enclosing;	/* outer block name */

	/*
	 * get a symbol and fill it in
	 */
#ifdef NEWDBG
	lowercase(name);
	internal = identname(name, false);
	find(s, internal) where
		s->module == curmodule and s->symvalue.funcv.linkid == namelink
	endfind(s);
	if (s == nil) {
	    s = insert(internal);
	    s->block = curmodule;
	    s->symvalue.funcv.linkid = namelink;
	} else if (not testing) {
	    if (s->symvalue.funcv.beginaddr != (mod_offset + offset1)) {
		printf("Warning: dbg_func: existing symbol '%s'\n",
		    symname(s));
		printf(" addr 0x%x found, expected 0x%x\n",
		    s->symvalue.funcv.beginaddr, mod_offset + offset1);
	    }
	}
	s->symvalue.funcv.beginaddr = mod_offset + offset1;
#else
	s = getproc(name, namelink);
#endif NEWDBG
	s->module = curmodule;
	s->class = PROC;
	s->level = level;
	s->language = whichlang(lang);
	s->chain = nil;

	/*
	 * set current stuff
	 */
	curblock = s;
	if (curlang != s->language)
		curlang = s->language;

	/*
	 * set s->block to outer name, or if there isn't one, remember
	 * the base for program locals in "NAME/D".
	 */
	if (strlen(outer) != 0) {
#ifdef NEWDBG
		lowercase(outer);
		enclosing = identname(outer, false);
		find(t, enclosing) where
			t->module == curmodule and 
			t->symvalue.funcv.linkid == outlink
		endfind(t);
		if (t == nil) {
		    t = insert(enclosing);
		    t->block = curmodule;	/* may be changed */
		    t->symvalue.funcv.linkid = outlink;
		    t->symvalue.funcv.beginaddr = mod_offset + offset2;
		} else
		    if (t->symvalue.funcv.beginaddr != (mod_offset + offset2)) {
			if (not testing) {
			  printf("Warning: dbg_func: existing symbol '%s'\n",
			    symname(t));
			  printf(" addr 0x%x found, expected 0x%x\n",
		    	    t->symvalue.funcv.beginaddr, mod_offset + offset2);
			}
		        t->symvalue.funcv.beginaddr = mod_offset + offset2;
		    }
#else
		t = getproc(outer, outlink);
#endif NEWDBG
		t->module = curmodule;
		t->class = PROC;
		t->level = level - 1;
		t->language = s->language;
		s->block = t;
	} else {
		s->block = curmodule;
		progdata_name = (char *)malloc(strlen(name) + 3);
		strcpy(progdata_name, name);
		strcat(progdata_name, "/D");
		progdata_linkid = namelink;
	}
}

/* 
 * find or allocate a symbol for a procedure.   have to search
 * by external name since they're the one's that are unique.
 * links the external name.
 *
 * this is the only correct way to get a symbol for a procedure!
 */
private Symbol getproc(name, link)
	char	*name;
	short	link;
{
	register Symbol s;	/* the symbol to return */
	Name internal;		/* the name inside dbx */
	Name external;		/* the name used to link w/ a.out symbol */

	external = identname(aoutname(name, link), false);
	lowercase(name);
	internal = identname(name, false);
	find(s, external) where
		s->class == EXTREF && s->block == curmodule
		and s->symvalue.extref->class != BADUSE		/* file name */
	endfind(s);
	if (s == nil) {
		s = insert(internal); 
		s->module = curmodule;
		linksym(s, ident(external), link);
	} else {
		s = s->symvalue.extref;
	}
	return(s);
}


/*
 * connect an EXTREF to the given symbol.  if 'linkid'
 * is zero or greater, the EXTREF's name will be '$linkid'.  otherwise
 * it's 'name'.  these get fixed
 * up later, when reading the linker symbols.   see fix_pascal_internal()
 * in object.c.  this procedure is the analog of 'getExtRef' in stabstring.c.
 */
private linksym(s, name, linkid)
	Symbol s;
	char *name;
	register short linkid;
{
	char *funnyname;
	Symbol t;

	/*
	 * create the funny name
	 */
	funnyname = aoutname(name, linkid);

#ifdef NOTDEF
	if (not testing)
		printf("Linking extref `%s' to symbol `%s'.\n", funnyname, symname(s));
#endif
	/*
	 * stick it into the symbol table, connecting to the symbol.
	 */
	t = insert(identname(funnyname, false));
	t->language = s->language;
	t->class = EXTREF;
	t->module = curmodule;
	t->block = curmodule;
	t->level = program->level;
	t->symvalue.extref = s;
}

/*
 * returns the name in the a.out symbol table that the given (name, limkid)
 * will be referenced by.
 */
private char *aoutname(name, linkid)
	char *name;
	register short linkid;
{
	static char buf[10];
	register char *cp;
	register int i;

	if (linkid >= 0) {
		cp = buf;
		*cp++ = '$';
		for (i = 0; i < 7; i++) {
			*cp++ = (linkid % 10) + '0';
			linkid /= 10;
		}
		*cp = '\0';
		return(buf);
	} else {
		if (*name == '_')
			return(name + 1);
		else
			return(name);
	}
}


/*
 * Construct a symbol for a variable.
 * Location can be register, fp relative, common:
 *
 * register -- Set level negative, reg num in symvalue.offset
 * fp relative -- put offfset in symvalue.offset
 * common -- symvalue.offset is offset in common now, fixed up later
 *           to be actual mem address
 */

public dbg_var(name, type, m)
	char *name;
	int type;
	struct dbgmode *m;
{
	register Symbol s;
	register Name n;


	lowercase(name);
	n = identname(name, false);
	s = insert(n);
	s->level = curblock->level + 1;
	s->language = curlang;
	s->block = curblock;
	typelink(s, type);
	if (m->m_indirect)
		s->class = REF;
	else
		s->class = VAR;

	if ( (m->m_rmode == 0 or m->m_rmode == 1 or 
	      m->m_rmode == 6 or m->m_rmode == 7 )   and
	    ( m->m_regnum == FRP)) 			/* local or param */
		s->symvalue.offset = m->m_offset;
	else if (m->m_rmode == -1) 			/* extern */
		docommon(s, m->m_linkname, m->m_linkid, m->m_offset);
#ifdef	ns32000
	else if ( (m->m_rmode == 0 or m->m_rmode == 1 or 
	           m->m_rmode == 6 or m->m_rmode == 7 )   and
		 ( m->m_regnum == SB)) 			/* static */
		docommon(s, progdata_name, progdata_linkid, m->m_offset);
#endif
	else {						/* register */
		s->level = - s->level;
		s->symvalue.offset = m->m_regnum;
	} 
}

/*VARARGS2*/
public dbg_type(kind, typeno, arg1, arg2, arg3)
	Byte kind;
	short typeno;
	int arg1, arg2, arg3;
{
	register Symbol t;
	register struct fortdim *fdp;
	char name[256];

	if (kind == 7) {		/* record */
		/* 
		 * have to create a name for records.  ugh!  this is to
		 * avoid recursion when printing the type of records
		 * with pointers to themselves.
		 */
		sprintf(name, "$record%d", typeno);
		t = insert(identname(name, false));
	} else
		t = symbol_alloc();

	/*
	 * add the type to the type table
	 */
	typeno += NBUILTINS;
	t->class = TYPE;
	t->language = curlang;
	t->block = curblock;
	t->type = symbol_alloc();
	typetable[typeno] = t;

	/*
	 * fill in info about the type
	 */
	t = t->type;
	t->language = curlang;
	t->level = curblock->level + 1;
	t->block = curblock;
	switch(kind) {
	case 0:	/* scalar -- max value */
		t->class = SCAL;
		t->symvalue.iconval = arg1;
		t->chain = nil;			/* XXX should have names */
		break;

	case 1: /* subrange -- subrange of, min val, max val */
		t->class = RANGE;
		typelink(t, arg1);
		t->symvalue.rangev.lowertype = R_CONST;
		t->symvalue.rangev.uppertype = R_CONST;
		t->symvalue.rangev.lower = arg2;
		t->symvalue.rangev.upper = arg3;
		break;

	case 2:	/* pointer -- ptr to */
		t->class = PTR;
		typelink(t, arg1);
		break;

	case 3:	/* set --  set of */
		t->class = SET;
		typelink(t, arg1);
		break;

	case 4: /* array  -- packinfo, array of, indexed by */
		t->class = ARRAY; /* XXX ignoring array packinfo for now */
		typelink(t, arg2);
		t->chain = typetable[arg3 + NBUILTINS];
		if (t->chain == nil)
			fixlater(FIX_TYPECHAIN, t, arg3 + NBUILTINS);
		break;

	case 5: /* string -- str length */
		if (t->language != lang_pascal)
			warning("dbg type: non-pascal string type");
		t->class = STR;
		t->symvalue.offset = arg1 + 1;
		break;

	case 6:	/* file -- file of */
		t->class = FILET;
		typelink(t, arg1);
		break;

	case 7: /* record -- packed, size */
		/* XXX sometime have to deal with packed records */
		t->class = RECORD;
		t->symvalue.offset = arg2;	/* size of entire record */
		currec = t;
		break;

	case 9:	/* character -- length */
		if (t->language != lang_fortran)
			warning("dbg type: non-fortran character type");
		t->class = STR;
		t->symvalue.offset = arg1;
		break;

	case 10: /* fortran array -- array of, #dimensions*/
		/* dimension info is in fdimtab */
		typelink(t, arg1);
		fdp = &fdimtab[arg2 - 1];
		while(arg2--) {				/* for each dimension */
			t->class = ARRAY;
			t->chain = symbol_alloc();
			t->chain->class = RANGE;
			t->chain->language = lang_fortran;
			t->chain->block = curblock;
			if (fdp->fdim_flags & 1) {
			    t->chain->symvalue.rangev.lowertype = R_TEMP;
			    t->chain->symvalue.rangev.lower = fdp->fdim_lbound;
			} else {
			    t->chain->symvalue.rangev.lowertype = R_CONST;
			    t->chain->symvalue.rangev.lower = fdp->fdim_lbound;
			}
			if (fdp->fdim_flags & 2) {
			    if (fdp->fdim_size < 0) {
				    t->chain->symvalue.rangev.uppertype = R_TEMP;
				    t->chain->symvalue.rangev.upper = fdp->fdim_size;
			    } else {
				    t->chain->symvalue.rangev.uppertype = R_ASSUMED;
			    }
			} else {
			    t->chain->symvalue.rangev.uppertype = R_CONST;
			    t->chain->symvalue.rangev.upper = fdp->fdim_hbound;
			}
			if (arg2) {
			    t->type = symbol_alloc();
			    t = t->type;
			    t->language = lang_fortran;
			    t->block = curblock;
			}
			fdp--;
		}
		t->class = ARRAY;
		typelink(t, arg1);
		break;

	default:
		warning("dbg type: unknown type %d", kind);
	}
}
	

/*
 * allocate a type for "character*n", and return it.
 */
private Symbol chartype(len)
	int len;
{
	register Symbol t, s;

	t = symbol_alloc();
	t->class = TYPE;
	t->language = curlang;
	t->block = curblock;
	t->type = symbol_alloc();
	s = t->type;
	s->language = curlang;
	s->level = curblock->level + 1;
	s->class = STR;
	s->block = curblock;
	s->symvalue.offset = len;
	return(t);
}


public dbg_field(name, type, offset, size)
	char		*name;
	short		type;
	unsigned int	offset;		/* from start of rec in bits */
	unsigned char	size;		/* of field in bits */
{
	register Symbol s, f;

	lowercase(name);
	s = newSymbol(identname(name, true), curblock->level + 1,
		FIELD, (Symbol) nil, (Symbol) nil);
	s->language = curlang;
	typelink(s, type);
	s->symvalue.field.offset = offset;
	if (size == 0)
		fixlater(FIX_FIELDSIZE, s, 0);
	else
		s->symvalue.field.length = size;
	f = currec;
	while (f->chain != nil && f->chain->symvalue.field.offset < offset)
		f = f->chain;
	s->chain = f->chain;
	f->chain = s;
}

public dbg_setlines(nentries, nstmts, line)
	short nentries;
	short nstmts;
	int line;		/* index to start of lines for this entry */
{
	register Symbol s;

	s = realentry->proc;
	s->symvalue.funcv.src = true;
	realentry->line = line;
	realentry->nlines = nstmts;
	if (nentries > 1) {
		s->symvalue.funcv.isentry = true;
		lastentry->symvalue.entryv.nextentry = firstentry;
	}
}


public dbg_entry(nents, index, name, link, entry_offset,
		 regsavemask, nlocals, mod_offset, entry_loc)
	short	nents;		/* number of entries */
	short	index;
	char	*name;
	short	link;
	int	entry_offset;	/* distance from address of proc to start
				   of code */
	struct	reg_save_mask regsavemask;
	int	nlocals;
	int	mod_offset;	/* link addr of module */
	int	entry_loc;	/* offset from link addr of entry point */
{
	register Symbol s, e;
	register struct fixline *f;
	Name internal;

	/*
	 * catch the real name of the entry before it gets converted
	 * to lower case
	 */
	if (index == 0)
		strcpy(realentry_name, name);
	
	/*
	 * get the symbol for the entry
	 */
#ifdef	NEWDBG
	lowercase(name);
	internal = identname(name, false);
	find(s, internal) where
		s->module == curmodule and s->symvalue.funcv.linkid == link
	endfind(s);
	if (s == nil) {
		s = insert(internal);
		s->block = curblock;
		s->module = curmodule;
		s->class = PROC;		/* what about FUNC ? */
		s->language = curlang;
		s->symvalue.funcv.linkid = link;
		s->symvalue.funcv.beginaddr = mod_offset + entry_loc;
	} else
	    if (s->symvalue.funcv.beginaddr != (mod_offset + entry_loc)) {
		if (not testing) {
		    printf("Warning: dbg_entry: existing symbol '%s'\n",
			symname(s));
		    printf(" addr 0x%x found, expected 0x%x\n",
			s->symvalue.funcv.beginaddr, mod_offset + entry_loc);
		}
		s->symvalue.funcv.beginaddr = mod_offset + entry_loc;
	    }
#else
	s = getproc(name, link);
#endif NEWDBG
	if (s->module == nil)
		s->module = curmodule;
	if (s->block == nil)
		s->block = curblock;
	s->symvalue.funcv.r_mask = regsavemask.r_mask;
	s->symvalue.funcv.w_mask = regsavemask.w_mask;
	s->symvalue.funcv.have_rmasks = true;
	s->symvalue.funcv.nlocals = -nlocals;
	s->symvalue.funcv.entryoff = entry_offset;

	/*
	 * set up to get absolute address for proc and lines later
	 */
	f = alloc_fixline();
	f->proc = s;
	f->prologue_size = entry_offset;
	f->next = fixlinelist;
	fixlinelist = f;

	/*
	 * don't generate line entries for funcs with no source lines
	 */
	if (entry_offset == -1) {
		s->symvalue.funcv.entryoff = 0;
		s->symvalue.funcv.src = false;
		f->prologue_size = 0;
		f->nlines = 0;
	}

	if (index == 0) {
		curblock = s;
		realentry = f;
	}
	curentry = s;

	if (nents > 1) {

		/*
		 * multiple entry points are represented with the addition
		 * of a ENTRY symbol on the 'chain' between the PROC symbol
		 * and the parameter list.  make the symbol.
		 */
		e = symbol_alloc();
		e->name = s->name;
		e->class = ENTRY;
		e->block = s->block;
		e->module = s->module;
		e->language = s->language;
		e->symvalue.entryv.entryid = index;
		e->symvalue.entryv.eid_offset = 0; 	/* not known yet */

		/*
		 * the first entry is special, because variables
		 * actually point to it.  for the others,
		 * the symbol returned by getproc() won't have been seen yet,
		 * so fill it in.
		 */
		if (index == 0) {
			lastentry = e;
			firstentry = e;
		} else {
			s->block = realentry->proc->block;
			s->level = realentry->proc->level;
			s->language = s->block->language;
			s->symvalue.funcv.src = true;
			s->class = PROC;
		}
		s->symvalue.funcv.isentry = true;
		s->chain = e;

		/* 
		 * the entrypoints are linked together in a circular list.
		 */
		lastentry->symvalue.entryv.nextentry = e;
		lastentry = e;
	}
}


/*
 * it would be nice to be able to give the offset of procedure params,
 * but since they don't come in through dbg_var, and the sizes aren't
 * all known until the end of the dbg file, it's kind of hard to figure
 * it out.  i don't want to do another fixup scheme for these just now.
 */
private Symbol kludge_up_a_proc_param(n)
	Name n;
{
	register Symbol s, t;

	t = newSymbol((Name) nil, nil, FPROC, (Symbol) nil, (Symbol) nil);
	t = newSymbol((Name) nil, nil, TYPE, t, (Symbol) nil);
	s = insert(n);
	s->level = curblock->level + 1;
	s->class = VAR;
	s->type = t;
	s->block = curblock;
	s->language = curlang;
	return(s);
}


/*
 * set up parameters for the current entry point.
 * 'curblock' is the symbol for the 'real' entry, i.e. the procedure
 * that was entered via dbg_func().  'curentry' is the symbol for the 
 * entry point we're currently dealing with -- it was created by
 * dbg_entry as one of multiple entry points.  'curparam' is just
 * a pointer that says where to chain the parameter.  variables in
 * 'curentry' really belong to 'curblock'.
 */
public dbg_param(params)
	char *params;
{
	register int len;
	register Symbol s, copysym;
	register Name n;
	boolean copyargs = false;

	if (curentry->symvalue.funcv.isentry) {
		curparam = curentry->chain;
		if (curentry != curblock)
			copyargs = true;
	} else
		curparam = curentry;

	while((len=strlen(params)) != 0) {
		lowercase(params);
		n = identname(params, false);
		find(s, n) where
			s->block == curblock and
			(s->class == REF or s->class == VAR)
		endfind(s);
		if (s == nil)
			s = kludge_up_a_proc_param(n);
		if (copyargs) {
			/*
			 * XXX this should be done by a procedure in symbols.c
			 */
			copysym = insert(n);
			copysym->language = s->language;
			copysym->class = s->class;
			copysym->level = s->level;
			copysym->type = s->type;
			copysym->chain = nil;
			copysym->symvalue = s->symvalue;
			copysym->block = curentry;
			s = copysym;
		}
		if (curparam->chain != nil)
			warning("dbg param: non-nil chain in curentry");
		curparam->chain = s;
		curparam = s;
		params += len + 1;
	}
}


public fixlines()
{
	register struct fixline *f;
	register Address procaddr;
	register Linetab *lp;
	register Linetab *firstline;
#ifdef DEBUGLINETAB
	int i;
#endif DEBUGLINETAB

	f = fixlinelist;
	while (f != nil) {
		procaddr = codeloc(f->proc);
		codeloc(f->proc) += f->prologue_size;	/* XXX is this ok? */
#ifdef DEBUGLINETAB
		printf("fixing '%s', %d lines, addr is 0x%x\n",
		    symname(f->proc), f->nlines, procaddr);
		i = f->line;
#endif DEBUGLINETAB
		/*
		 * entry points don't have any lines and don't go into
		 * the function table.
		 */
		if (f->proc->symvalue.funcv.isentry == false
		    or f->proc->chain->symvalue.entryv.entryid == 0) {

			/*
			 * add to function table
			 */
			newfunc(f->proc, procaddr);

			/*
			 * make line addresses absolute
			 */
			firstline = linetab + f->line;
			for (lp = firstline; lp < firstline + f->nlines; lp++) {
#ifdef DEBUGLINETAB
				printf("linetab[%d] = line %d, addr 0x%x : 0x%x\n",
				   i, lp->line, lp->addr, lp->addr + procaddr);
				i++;
#endif DEBUGLINETAB
				lp->addr += procaddr;
			}
		}
		f = f->next;
	}
	free_fixlines();
}


/*
 * It's possible that the type isn't known yet.  If not, add it to 
 * the list to be fixed later.  This procedure is the *only* way to 
 * connect a type to a symbol, since it knows oddities of type
 * numbering.
 */
private typelink(s, type)
	Symbol s;
	short type;
{
	register Symbol ty;
	Symbol chartype();

	if (type <= -100) {
		s->type = chartype(- (type + 100) );
		return;
	}

	type += NBUILTINS;
	ty = typetable[type];
	if (ty == nil) {	/* haven't seen it yet */
		fixlater(FIX_TYPE, s, type);
	} else {
		if (type < NBUILTINS && ty->type->class == RANGE &&
		    ty->type->symvalue.rangev.lower == 0 &&
		    ty->type->symvalue.rangev.upper == 0)
			warning("dbg var: type `%s' for `%s' unsupported in %s",
				ident(ty->name), s->name, language_name(curlang));
	}
	s->type = ty;
}


/*
 * put symbol on a list to be completed later.
 */
fixlater(what, s, type)
	fixup_op what;
	Symbol s;
	short type;
{
	register fixup_rec *f;

	f = (struct fixup_rec *)malloc(sizeof(struct fixup_rec)); /*XXXfaster?*/
	f->f_op = what;
	f->f_sym = s;
	f->f_type = type;
	f->f_next = fixup_list;
	fixup_list = f;
}


/*
 * Fix up symbols that weren't completed earlier.
 */
dofixups()
{
	register fixup_rec *f;

	/*
	 * fix types.  this must be done before fixing sizes so
	 * that all the sizes can be known!
	 */
	f = fixup_list;
	while(f != nil) {
		switch(f->f_op) {
		case FIX_TYPE:
			assert(typetable[f->f_type] != nil);
			f->f_sym->type = typetable[f->f_type];
			break;
		case FIX_TYPECHAIN:
			assert(typetable[f->f_type] != nil);
			f->f_sym->chain = typetable[f->f_type];
			break;
		}
		f = f->f_next;
	}

	/*
	 * fix up field sizes.
	 */
	f = fixup_list;
	while(f != nil) {
		if (f->f_op == FIX_FIELDSIZE)
			f->f_sym->symvalue.field.length = 
				size(f->f_sym->type) * BITSPERBYTE;
		free(f);
		f = f->f_next;
	}

	fixup_list = nil;
}

lowercase(s)
	register char *s;
{
	register char c;

	if (uppercase_flag)
		return;
	while ((c = *s) != '\0') {
		if (isupper(c))
			*s = tolower(c);
		s++;
	}
}


