/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)debug.c	1.5 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: debug.c 2.10 87/11/09 $";
#endif

/*
 *  Debug routines
 */

#include "defs.h"
#include "tree.h"
#include "operators.h"
#include "eval.h"
#include "events.h"
#include "symbols.h"
#include "scanner.h"
#include "source.h"
#include "object.h"
#include "main.h"
#include "mappings.h"
#include "machine.h"
#include "debug.h"
#include <signal.h>

public boolean tracetree;	/* trace building of parse trees */
public boolean traceeval;	/* trace tree evaluation */
public boolean traceassign;	/* trace assign command */
extern boolean debug_ptrace;	/* trace ptrace calls */
extern boolean debug_process;	/* trace process state changes */
extern boolean debug_procstack;	/* trace process stack management */
extern boolean debug_globals;	/* trace globals fussing */
extern boolean debug_queue;	/* trace command queueing */
extern boolean debug_stopall;	/* trace stopall() */
extern boolean debug_regsave;	/* debug register save code */
extern boolean debug_frame;	/* debug stack frame manipulation */
extern boolean debug_stab;	/* debug stab info */
extern boolean debug_which;	/* debug symbol resolution (which) code */
extern boolean notty;		/* don't play with tty */

enum debugfunc {
	NOFUNC, SCAN, BPTS, EXEC, BUILD, EVAL, PROCESS, SYMS,
	PTRACE, REGSAVE, FRAME, FUNCTAB, SYMTAB, LINETAB, QUEUE, STOPALL, 
	NOTTY, TTY, BLOCKS, ASSIGN, GLOBALS, PROCSTACK, STAB, WHICH
};

struct debugentry {
	String name;
	enum debugfunc func;
	String help;
} debugtable[] = {
	"scan",		SCAN,		"trace scanner return values",
	"bpts",		BPTS,		"trace breakpoints and dump bpt list",
	"exec",		EXEC,		"trace execution",
	"build",	BUILD,		"trace tree building",
	"eval",		EVAL,		"trace tree evaluation",
	"process",	PROCESS,	"trace process state changes",
	"procstack",	PROCSTACK,	"trace process stack management",
	"globals",	GLOBALS,	"trace globals save/restore",
	"queue",	QUEUE,		"trace command queue",
	"stopall",	STOPALL,	"trace stop all",
	"syms",		SYMS,		"trace reading symbols",
	"blocks",	BLOCKS,		"trace blocks while reading symbols",
	"assign",	ASSIGN,		"trace assign command",
	"notty",	NOTTY,		"don't play with tty",
	"tty",		TTY,		"dump info about ttys and pgrps",
	"ptrace",	PTRACE,		"show ptrace calls",
	"regsave",	REGSAVE,	"debug register save operation",
	"frame",	FRAME,		"debug stack frame manipulation",
	"functab",	FUNCTAB,	"dump function table",
	"symtab",	SYMTAB,		"dump symbol table",
	"linetab",	LINETAB,	"dump line and file tables",
	"stab",		STAB,		"trace reading of stab info",
	"which",	WHICH,		"debug symbol resolution (which) code",
	0,		NOFUNC,		0
};

/*
 * Dynamically turn on/off a debug flag, or display some information.
 */

public debug (p)
Node p;
{
    String flag;
    register struct debugentry *dp;
    boolean value;

    flag = p->value.scon;

    if (flag == nil) {

	printf("Debug flags (use `-flag' to clear):\n");
	for (dp = debugtable; dp->func != NOFUNC; dp++)
		printf("%s\t%s\n", dp->name, dp->help);

    } else {

	if (*flag == '-') {
		flag++;
		value = false;
	} else
		value = true;

	for (dp = debugtable; dp->func != NOFUNC; dp++)
		if (strcmp(dp->name, flag) == 0)
			break;

	switch(dp->func) {

	default:
	    error("no such debug flag");
	    break;
	
	case SCAN:
#           ifdef LEXDEBUG
		lexdebug = value;
#           else
		error("can't debug scanner (not compiled with LEXDEBUG)");
#           endif
	    break;
	
	case STAB:
#	    ifdef STABDEBUG
		debug_stab = value;
#	    else
		error("can't debug stab info (not compiled with SCANDEBUG)");
#	    endif
	    break;

	case SYMS:
	    tracesyms = value;
	    break;

	case BLOCKS:
	    traceblocks = value;
	    break;
	
	case ASSIGN:
	    traceassign = value;
	    break;

	case BPTS:
	    tracebpts = value;
	    if (tracebpts)
		printbps();
	    break;

	case EXEC:
	    traceexec = value;
	    break;

	case BUILD:
	    tracetree = value;
	    break;

	case EVAL:
	    traceeval = value;
	    break;

	case FUNCTAB:
	    dumpfunctab();
	    break;

	case SYMTAB:
	    symbol_dump((Symbol) nil);
	    break;

	case PTRACE:
	    debug_ptrace = value;
	    break;

	case PROCESS:
	    debug_process = value;
	    break;

	case PROCSTACK:
	    debug_procstack = value;
	    break;

	case GLOBALS:
	    debug_globals = value;
	    break;

	case LINETAB:
	    dumplinetab();
	    break;

	case NOTTY:
	    notty = value;
	    break;

	case TTY:
	    dumpttys();
	    break;

	case QUEUE:
	    debug_queue = value;
	    if (debug_queue)
		printq();
	    break;

	case STOPALL:
	    debug_stopall = value;
	    break;

	case REGSAVE:
	    debug_regsave = value;
	    break;
	
	case FRAME:
	    debug_frame = value;
	    break;

	case WHICH:
	    debug_which = value;
	    break;
	}
    }
}

private String leafname[] = {
    "nop", "name", "sym", "lcon", "ccon", "fcon", "scon", "rval", "index"
};

public String opname (op)
Operator op;
{
    String s;
    static char buf[100];

    switch (op) {
	case O_ITOF:
	    s = "itof";
	    break;

	case O_QLINE:
	    s = "qline";
	    break;

	default:
	    if (ord(op) <= ord(O_INDEX)) {
		s = leafname[ord(op)];
	    } else {
		s = opinfo[ord(op)].opstring;
		if (s == nil) {
		    sprintf(buf, "[op %d]", op);
		    s = buf;
		}
	    }
	    break;
    }
    return s;
}

/*
 * print all symbols with same name as 's'.
 */
public pchain(s)
Symbol s;
{
    register Name n;
    register Symbol t;

    checkref(s);
    n = s->name;
    t = lookup(n);
    psym(t);
    t = t->next_sym;
    while (t != nil) {
	if (t->name == n) {
	    psym(t);
	}
	t = t->next_sym;
    }
}

/* 
 * flush i/o buffers -- for use by the debugger
 */
xxflush()
{
	printf("stdout == 0x%x; stderr == 0x%x\n", stdout, stderr);
	fflush(stdout);
	fflush(stderr);
}

xpsym(i)
	integer i;
{
	psym( /*NOSTRICT*/ i);
}
