/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/*	defs.h	1.4	85/03/01	*/

/*
 * Public definitions, common to all.
 */

#include <stdio.h>

#define new(type)           ((type) malloc(sizeof(struct type)))
#define newarr(type, n)     ((type *) malloc((unsigned) (n) * sizeof(type)))
#define dispose(ptr)        { free((char *) ptr); ptr = 0; }

/*
 * macros to determine an efficient value for 'NITEMS' for "pools" like:
 *    struct {
 *	struct xxx *next;
 *	struct yyy[NITEMS];
 *    } xxx;
 * efficiency means pick the largest value of NITEMS that will fit in
 * a malloc() "bucket" (2^x), taking into account the space malloc uses.
 * for the above struct, do "#define NITEMS POOLSIZE(2048,sizeof(struct yyy))".
 */
#define MALLOC_OVERHEAD	16
#define NITEMS(bucket,itemsize) \
	(((bucket) - sizeof(char *) - MALLOC_OVERHEAD) / (itemsize))
#define POOLSIZE(bucket,itemsize) \
	((((NITEMS((bucket),(itemsize))*(itemsize))+3)&~3)/(itemsize))

#define public
#define private static

#define ord(enumcon) ((unsigned int) enumcon)
#define nil 0
#define and &&
#define or ||
#define not !
#define div /
#define mod %
#define max(a, b)    ((a) > (b) ? (a) : (b))
#define min(a, b)    ((a) < (b) ? (a) : (b))

#define assert(b) { \
    if (not(b)) { \
	panic("assertion failed at line %d in file %s", __LINE__, __FILE__); \
    } \
}

#define badcaseval(v) { \
    panic("unexpected value %d at line %d in file %s", v, __LINE__, __FILE__); \
}

#define checkref(p) { \
    if (p == nil) { \
	panic("reference through nil pointer at line %d in file %s", \
	    __LINE__, __FILE__); \
    } \
}

typedef int Integer;
typedef int integer;
typedef char Char;
typedef double Real;
typedef double real;
typedef enum { false = 0, true = 1 } Boolean;
#define FALSE 0
#define TRUE 1
typedef Boolean boolean;
typedef char *String;
typedef char *pointer;

#define strdup(s)       strcpy(malloc((unsigned) strlen(s) + 1), s)
#define streq(s1, s2)   (strcmp(s1, s2) == 0)

typedef FILE *File;
typedef int Fileid;
typedef String Filename;

#define get(f, var) fread((char *) &(var), sizeof(var), 1, f)
#define put(f, var) fwrite((char *) &(var), sizeof(var), 1, f)

#undef FILE

extern long atol();
extern double atof();
extern char *malloc();
extern String strcpy(), index(), rindex();
extern int strlen();

extern String cmdname;
extern String errfilename;
extern short errlineno;
extern int debug_flag[];

#define DEFSIG	-1

#ifdef DBX
#include "tty.h"
#endif
