/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)eval.c	1.15 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: eval.c 2.24 88/03/15 $";
#endif

/*
 * Tree evaluation.
 */

#include "defs.h"
#include "main.h"
#include "tree.h"
#include "operators.h"
#include "debug.h"
#include "eval.h"
#include "events.h"
#include "symbols.h"
#include "scanner.h"
#include "source.h"
#include "object.h"
#include "mappings.h"
#include "multi-p.h"
#include "runtime.h"
#include "machine.h"
#include <signal.h>

#ifndef public

#include "machine.h"

#define STACKSIZE 20000

typedef Char Stack;

#define push(type, value) { \
    ((type *) (sp += sizeof(type)))[-1] = (value); \
}

#define pop(type) ( \
    (*((type *) (sp -= sizeof(type)))) \
)

#define popn(n, dest) { \
    sp -= n; \
    bcopy(sp, (char *)dest, (unsigned int)n); \
}

#define alignstack() { \
    sp = (Stack *) (( ((int) sp) + sizeof(int) - 1)&~(sizeof(int) - 1)); \
}

#endif

public Stack stack[STACKSIZE];
public Stack *sp = &stack[0];
public Boolean useInstLoc = false;

#define chksp() \
{ \
    if (sp < &stack[0]) { \
	sp = &stack[0]; \
	error("internal error: stack underflow"); \
    } \
}

#define poparg(n, r, fr) { \
    eval(p->value.arg[n]); \
    if (isreal(p->op)) { \
	if (size(p->value.arg[n]->nodetype) == sizeof(float)) { \
	    fr = pop(float); \
	} else { \
	    fr = pop(double); \
	} \
    } else if (isint(p->op)) { \
	r = popsmall(p->value.arg[n]->nodetype); \
    } \
}

#define Boolrep char	/* underlying representation type for booleans */

/*
 * Command-level evaluation.
 */

public topeval (p)
Node p;
{
    if (traceeval) {
	fprintf(stderr, "topeval(");
	prtree(stderr, p);
	fprintf(stderr, ")\n");
	fflush(stderr);
    }
    eval(p);
}

/*
 * Evaluate a parse tree leaving the value on the top of the stack.
 */

public eval(p)
register Node p;
{
    long r0, r1;
    double fr0, fr1;
    Address addr;
    long n;
    int len;
    Symbol s;
    Node n1, n2;
    String str;
    long bytes_on_stack;

    bytes_on_stack = sp - &stack[0];

    checkref(p);
    if (traceeval) {
	fprintf(stderr, "begin eval node 0x%x, op= '%s', stack= %d bytes\n",
	    p, opname(p->op), bytes_on_stack);
    }

    if (bytes_on_stack < 0) {
	sp = &stack[0];
	error("internal error: stack underflow");
    }
    if (bytes_on_stack > STACKSIZE) {
	sp = &stack[0];
	error("internal error: stack overflow");
    }


    switch (degree(p->op)) {
	case BINARY:
	    poparg(1, r1, fr1);
	    poparg(0, r0, fr0);
	    break;

	case UNARY:
	    poparg(0, r0, fr0);
	    break;

	default:
	    /* do nothing */;
    }
    switch (p->op) {
	case O_SYM:
	    s = p->value.sym;
	    if (s == retaddrsym) {
		push(long, return_addr(nil));
	    } else if (isregname(s)) {
		if (isfpreg(s)) {
		    push(double, get_fpu_reg(regnum(s)));
		} else if (isfreg8(s)) {
		    pushregdouble(s);
		} else if (isreg(s)) {
		    pushsmall(s, (long)address(s, (Frame)nil));
		} else {
		    error("internal error: eval(): reg error '%s'", identname(s));
		}
	    } else if (isvariable(s)) {
		if (s != program and not isactive(container(s))) {
		    error("\"%s\" is not active", symname(s));
		}
		if (isvarparam(s) and not isopenarray(s)) {
		    if (isreg(s)) {
		        push(Address, address(s, (Frame)nil));
		    } else {
		        rpush(address(s, (Frame)nil), sizeof(Address));
		    }
		} else if (isfpreg(s)) {
		    push(double, get_fpu_reg(regnum(s)));
		} else if (isfreg8(s)) {
		    pushregdouble(s);
		} else if (isreg(s)) {
		    pushsmall(s, (long)address(s, (Frame)nil));
		} else {
		    push(Address, address(s, (Frame)nil));
		}
	    } else if (isblock(s)) {
		push(Symbol, s);
	    } else if (isconst(s)) {
		eval(constval(s));
	    } else {
		error("can't evaluate a %s", classname(s));
	    }
	    break;

	case O_LCON:
	case O_CCON:
	    r0 = p->value.lcon;
	    pushsmall(p->nodetype, r0);
	    break;

	case O_FCON:
	    push(double, p->value.fcon);
	    break;

	case O_SCON:
	    len = size(p->nodetype);
	    mov(p->value.scon, sp, len);
	    sp += len;
	    break;

	case O_INDEX:
	    s = p->value.arg[0]->nodetype;
	    p->value.arg[0]->nodetype = t_addr;
	    eval(p->value.arg[0]);
	    p->value.arg[0]->nodetype = s;
	    n = pop(Address);
	    eval(p->value.arg[1]);
	    evalindex(s, n, popsmall(p->value.arg[1]->nodetype));
	    break;

	case O_DOT:
	    s = p->value.arg[1]->value.sym;
	    eval(p->value.arg[0]);
	    n = pop(long);
	    push(long, n + (s->symvalue.field.offset div 8));
	    break;

	/*
	 * Get the value of the expression addressed by the top of the stack.
	 * Push the result back on the stack.
	 */

	case O_INDIR:
	case O_RVAL:

	    /*
	     * Don't RVAL register param symbol whose frame has been 
	     * entered, since value on stack is register contents.
	     * Also don't RVAL a module.
	     */
	    if (p->value.arg[0]->op == O_SYM) {
		s = p->value.arg[0]->value.sym;
		if (isreg(s) and isparam(s) and isentered(symframe(s, nil)))
		    break;
		if (s->class == MODULE)
		    break;
	    }

	    addr = pop(long);
	    if (addr == 0) {
		error("reference through nil pointer");
	    }
	    len = size(p->nodetype);
	    rpush(addr, len);
	    break;

	/*
	 * Move the stack pointer so that the top of the stack has
	 * something corresponding to the size of the current node type.
	 * If this new type is bigger than the subtree (len > 0),
	 * then the stack is padded with nulls.  If it's smaller,
	 * the stack is just dropped by the appropriate amount.
	 */
	case O_TYPERENAME:
	    len = size(p->nodetype) - size(p->value.arg[0]->nodetype);
	    if (len > 0) {
		for (n = 0; n < len; n++) {
		    *sp++ = '\0';
		}
	    } else if (len < 0) {
		sp += len;
	    }
	    break;

	case O_COMMA:
	    eval(p->value.arg[0]);
	    if (p->value.arg[1] != nil) {
		eval(p->value.arg[1]);
	    }
	    break;

	case O_ITOF:
	    push(double, (double) r0);
	    break;

	case O_ADD:
	    push(long, r0+r1);
	    break;

	case O_ADDF:
	    push(double, fr0+fr1);
	    break;

	case O_SUB:
	    push(long, r0-r1);
	    break;

	case O_SUBF:
	    push(double, fr0-fr1);
	    break;

	case O_NEG:
	    push(long, -r0);
	    break;

	case O_NEGF:
	    push(double, -fr0);
	    break;

	case O_MUL:
	    push(long, r0*r1);
	    break;

	case O_MULF:
	    push(double, fr0*fr1);
	    break;

	case O_DIVF:
	    if (fr1 == 0) {
		error("error: division by 0");
	    }
	    push(double, fr0 / fr1);
	    break;

	case O_DIV:
	    if (r1 == 0) {
		error("error: div by 0");
	    }
	    push(long, r0 div r1);
	    break;

	case O_MOD:
	    if (r1 == 0) {
		error("error: mod by 0");
	    }
	    push(long, r0 mod r1);
	    break;

	case O_LT:
	    push(Boolrep, r0 < r1);
	    break;

	case O_LTF:
	    push(Boolrep, fr0 < fr1);
	    break;

	case O_LE:
	    push(Boolrep, r0 <= r1);
	    break;

	case O_LEF:
	    push(Boolrep, fr0 <= fr1);
	    break;

	case O_GT:
	    push(Boolrep, r0 > r1);
	    break;

	case O_GTF:
	    push(Boolrep, fr0 > fr1);
	    break;

	case O_EQ:
	    push(Boolrep, r0 == r1);
	    break;

	case O_EQF:
	    push(Boolrep, fr0 == fr1);
	    break;

	case O_NE:
	    push(Boolrep, r0 != r1);
	    break;

	case O_NEF:
	    push(Boolrep, fr0 != fr1);
	    break;

	case O_AND:
	    push(Boolrep, r0 and r1);
	    break;

	case O_OR:
	    push(Boolrep, r0 or r1);
	    break;

	case O_ASSIGN:
	    assign(p->value.arg[0], p->value.arg[1]);
	    break;

	case O_BACKGROUND:
	    {
		boolean save;

		save = synchronous;
		synchronous = false;
		eval(p->value.arg[0]);
		synchronous = save;
	    }
	    break;

	case O_CHFILE:
	    changefile(p->value.list);
	    break;

	case O_CONT:
	    tempbp(p->value.arg[1], BY_CMD);
	    cont(p->value.arg[0]->value.lcon, true);
	    break;

	case O_CONTALL:
	    contall(p->value.arg[0]);
	    break;

	case O_KILL:
	    kill_process();
	    makeprocs();
	    wfixdefprocs(wp->p_id);
	    break;

	case O_KILLALL:
	    kill_all();
	    makeprocs();
	    wfixdefprocs(wp->p_id);
	    break;

	case O_LIST:
	    list(p, false);
	    break;

	case O_LISTI:
	    list(p, true);
	    break;

	case O_FUNC:
	    func(p->value.arg[0]);
	    break;

	case O_EXAMINE:
	    eval(p->value.examine.beginaddr);
	    r0 = pop(long);
	    if (p->value.examine.endaddr == nil) {
		n = p->value.examine.count;
		if (n == 0) {
		    printvalue(r0, p->value.examine.mode);
		} else if (streq(p->value.examine.mode, "i")) {
		    printninst(n, (Address) r0);
		} else {
		    printndata(n, (Address) r0, p->value.examine.mode);
		}
	    } else {
		eval(p->value.examine.endaddr);
		r1 = pop(long);
		if (streq(p->value.examine.mode, "i")) {
		    printinst((Address)r0, (Address)r1);
		} else {
		    printdata((Address)r0, (Address)r1, p->value.examine.mode);
		}
	    }
	    break;

	case O_PRINT:
	   /*
	    * NOTE:
	    * Special case check for floating point registger print.
	    * A different display format is used.  We check here to see
	    * if an arg of type symbol was presented and if check to see
	    * if is is one of the floating point registers.
	    * If "print_fpu_reg()" does the display it will return true
	    * and we know the print command has been completed.
	    */

	    /*
	     * loop through args, evaluating and printing.  have to push
	     * proc for printval() because printString() in c.c reads from
	     * the process to get the value of strings.
	     */
	    for (n1 = p->value.arg[0]; n1 != nil; n1 = n1->value.arg[1]) {
		if (n1->op == O_COMMA and
		    n1->value.arg[0]->op == O_SYM and
		    print_fpu_reg(symname(n1->value.arg[0]->value.sym)) ) {
			/* work done by print_fpu_reg */
	        } else {
		    eval(n1->value.arg[0]);
		    if (n1->value.arg[0]->op == O_PUSHPROC) /* kludge */
		        pushpid(n1->value.arg[0]->value.arg[0]->value.lcon);
		    printval(n1->value.arg[0]->nodetype);
		    if (n1->value.arg[0]->op == O_PUSHPROC) /* kludge */
		        popproc();
		}
		putchar(' ');
	    }
	    putchar('\n');
	    break;
	
	case O_PRINTFPUSTACK:
	    printfpustack();
	    break;
	
	case O_PRINTNEWS:
	    printnews();
	    break;

	case O_PS:
	    if (p->value.arg[0])
		ps(p->value.arg[0]->value.lcon);
	    else
	        ps(0);
	    break;

	case O_PSYM:
	    if (p->value.arg[0]->op == O_SYM) {
		symbol_dump(p->value.arg[0]->value.sym);
	    } else {
		psym(p->value.arg[0]->value.lcon);
	    }
	    break;

	case O_PUSHPROC:
	    pushpid(p->value.arg[0]->value.lcon);
	    eval(p->value.arg[1]);
	    popproc();
	    break;

	case O_QLINE:
	    eval(p->value.arg[1]);
	    break;

	case O_SETPROC:
	    setproc(p->value.arg[0]->value.lcon);
	    break;

	case O_SIGNAL:
	    sendsig(p->value.arg[0]->value.lcon,	/* procid */
	            p->value.arg[1]->value.lcon);	/* signal */
	    break;

	case O_STEP:
	    wp->p_stepi = (Boolean) (not p->value.step.source);
	    if (p->value.step.skipcalls) {
		next(0);
	    } else {
		step(0);
	    }
	    break;

	case O_STOPALL:
	    stopall();
	    break;

	case O_RELEASE:
	    release(p->value.arg[0]->value.lcon);	/* procid */
	    makeprocs();
	    wfixdefprocs(wp->p_id);
	    break;

	case O_WAIT:
	    waitfor(wp);
	    break;

	case O_WHATIS:
	    if (p->value.arg[0]->op == O_SYM) {
		printdecl(p->value.arg[0]->value.sym);
	    } else {
		printdecl(p->value.arg[0]->nodetype);
	    }
	    break;

	case O_WHERE:
	    wherecmd();
	    break;

	case O_WHEREIS:
	    if (p->value.arg[0]->op == O_SYM) {
		printwhereis(stdout, p->value.arg[0]->value.sym);
	    } else {
		printwhereis(stdout, p->value.arg[0]->nodetype);
	    }
	    break;

	case O_WHICH:
	    if (p->value.arg[0]->op == O_SYM) {
		printwhich(stdout, p->value.arg[0]->value.sym);
	    } else {
		printwhich(stdout, p->value.arg[0]->nodetype);
	    }
	    putchar('\n');
	    break;

	case O_WINDOW:
	    wopen(p->value.arg[1]->value.scon, p->value.arg[0]->value.lcon);
	    break;

	case O_ALIAS:
	    n1 = p->value.arg[0];
	    n2 = p->value.arg[1];
	    if (n2 == nil) {
		if (n1 == nil) {
		    alias(nil, nil, nil);
		} else {
		    alias(n1->value.name, nil, nil);
		}
	    } else if (n2->op == O_NAME) {
		str = ident(n2->value.name);
		alias(n1->value.name, nil, strdup(str));
	    } else {
		if (n1->op == O_COMMA) {
		    alias(
			n1->value.arg[0]->value.name,
			(List) n1->value.arg[1],
			n2->value.scon
		    );
		} else {
		    alias(n1->value.name, nil, n2->value.scon);
		}
	    }
	    break;

	case O_UNALIAS:
	    unalias(p->value.arg[0]->value.name);
	    break;

	case O_CALLPROC:
	    callproc(p, false);
	    break;

	case O_CALL:
	    callproc(p, true);
	    break;

	case O_CATCH:
	    if (p->value.lcon == 0) {
		printsigscaught();
	    } else {
		psigtrace(p->value.lcon, true);
	    }
	    break;

	case O_EDIT:
	    edit(p->value.scon);
	    break;

        case O_DEBUG:
            debug(p);
	    break;

	case O_DOWN:
	    checkref(p->value.arg[0]);
	    assert(p->value.arg[0]->op == O_LCON);
	    down(p->value.arg[0]->value.lcon);
	    break;

	case O_DUMP:
	    if (p->value.arg[0] == nil) {
		dumpall();
	    } else {
		s = p->value.arg[0]->value.sym;
		if (s == curfunc) {
		    dump(nil);
		} else {
		    dump(s);
		}
	    }
	    break;

	case O_HELP:
	    help();
	    break;

	case O_IGNORE:
	    if (p->value.lcon == 0) {
		printsigsignored();
	    } else {
		psigtrace(p->value.lcon, false);
	    }
	    break;

	case O_RETURN:
	    if (p->value.arg[0] == nil) {
		rtnfunc(nil);
	    } else {
		assert(p->value.arg[0]->op == O_SYM);
		rtnfunc(p->value.arg[0]->value.sym);
	    }
	    break;

	case O_RUN:
	    run();
	    break;

	case O_SET:
	    set(p->value.arg[0], p->value.arg[1]);
	    break;

	case O_SEARCH:
	    search(p->value.arg[0]->value.lcon, p->value.arg[1]->value.scon);
	    break;

	case O_SOURCE:
	    setinput(p->value.scon);
	    break;

	case O_STATUS:
	    status();
	    break;

	case O_TRACE:
	case O_TRACEI:
	    trace(p);
	    break;

	case O_STOP:
	case O_STOPI:
	    stop(p);
	    break;

	case O_UNSET:
	    undefvar(p->value.arg[0]->value.name);
	    break;

	case O_UP:
	    checkref(p->value.arg[0]);
	    assert(p->value.arg[0]->op == O_LCON);
	    up(p->value.arg[0]->value.lcon);
	    break;

	case O_ADDEVENT:
	    addevent(ALLPROCS, p->value.event.cond, p->value.event.actions);
	    break;

	case O_DELETE:
	    n1 = p->value.arg[0];
	    while (n1->op == O_COMMA) {
		n2 = n1->value.arg[0];
		assert(n2->op == O_LCON);
		if (not delevent((unsigned int) n2->value.lcon, true)) {
		    error("unknown event %ld", n2->value.lcon);
		}
		n1 = n1->value.arg[1];
	    }
	    assert(n1->op == O_LCON);
	    if (not delevent((unsigned int) n1->value.lcon, true)) {
		error("unknown event %ld", n1->value.lcon);
	    }
	    break;

	case O_IF:
	    if (cond(p->value.event.cond)) {
		evalcmdlist(p->value.event.actions);
	    }
	    break;

	case O_ONCE:
	    event_once(THISPROC, p->value.event.cond, p->value.event.actions);
	    break;

	case O_PRINTCALL:
	    printprocid();
	    printcall(p->value.sym, whatblock(return_addr(nil)));
	    break;

	case O_PRINTIFCHANGED:
	    printifchanged(p->value.arg[0]);
	    break;

	case O_PRINTRTN:
	    printprocid();
	    printrtn(p->value.sym);
	    break;

	case O_PRINTSRCPOS:
	    printprocid();
	    getsrcpos();
	    if (p->value.arg[0] == nil) {
		printsrcpos();
		putchar('\n');
		printlines(curline, curline, false);
	    } else if (p->value.arg[0]->op == O_QLINE) {
		if (p->value.arg[0]->value.arg[1]->value.lcon == 0) {
		    printf("tracei: ");
		    printinst(reg(PROGCTR), reg(PROGCTR));
		} else {
		    if (canReadSource()) {
			printf("trace:  ");
			printlines(curline, curline, false);
		    }
		}
	    } else {
		printsrcpos();
		printf(": ");
		eval(p->value.arg[0]);
		prtree(stdout, p->value.arg[0]);
		printf(" = ");
		printval(p->value.arg[0]->nodetype);
		putchar('\n');
	    }
	    break;

	case O_PROCFLUSHED:
	    procdone(p->value.sym);
	    break;

	case O_PROCRTN:
	    procreturn(p->value.arg[0]->value.sym,   /* proc */
	               p->value.arg[1]->value.lcon,  /* stack pointer */
	               p->value.arg[2]->value.lcon); /* event id */
	    break;

	case O_STOPIFCHANGED:
	    stopifchanged(p->value.arg[0], p->value.arg[1]->value.lcon);
	    break;

	case O_DBXSTOP:
	    setstopped(p->value.lcon);
	    break;

	case O_STOPX:
	    if ((int)p->value.lcon == ALLPROCS)
		stopall();
	    setstopped(BY_BPT);
	    break;

	case O_TRACEON:
	    traceon(p->value.trace.inst, p->value.trace.event,
		p->value.trace.actions, p->value.trace.inherit);
	    break;

	case O_TRACEOFF:
	    traceoff(p->value.lcon);
	    break;

	case O_RGERR:
	    range_err(0);
	    break;

	case O_SRERR:
	    range_err(1);
	    break;

	default:
	    panic("eval: bad op %d", p->op);
    }
    if (traceeval) { 
	fprintf(stderr, "end eval %s\n", opname(p->op));
    }
}

/*
 * Evaluate a list of commands.
 */

public evalcmdlist(cl)
Cmdlist cl;
{
    Command c;

    foreach (Command, c, cl)
	evalcmd(c);
    endfor
}

/*
 * Push "len" bytes onto the expression stack from address "addr"
 * in the process.  If there isn't room on the stack, print an error message.
 */

public rpush(addr, len)
Address addr;
int len;
{
    if (not canpush(len)) {
	error("expression too large to evaluate");
    } else {
	chksp();
	dread(sp, addr, len);
	sp += len;
    }
}

/*
 * Check if the stack has n bytes available.
 */

public Boolean canpush(n)
Integer n;
{
    return (Boolean) (sp + n < &stack[STACKSIZE]);
}

/*
 * Push a small scalar of the given type onto the stack.
 */

public pushsmall(t, v)
Symbol t;
long v;
{
    register Integer s;

    s = size(t);
    switch (s) {
	case sizeof(char):
	    push(char, v);
	    break;

	case sizeof(short):
	    push(short, v);
	    break;

	case sizeof(long):
	    push(long, v);
	    break;

	default:
	    panic("bad size %d in popsmall", s);
    }
}

/*
 * Pop an item of the given type which is assumed to be no larger
 * than a long and return it expanded into a long.
 */

public long popsmall(t)
Symbol t;
{
    register integer n;
    long r;

    n = size(t);
    if (n == sizeof(char)) {
	if (t->class == RANGE and t->symvalue.rangev.lower >= 0) {
	    r = (long) pop(unsigned char);
	} else {
	    r = (long) pop(char);
	}
    } else if (n == sizeof(short)) {
	if (t->class == RANGE and t->symvalue.rangev.lower >= 0) {
	    r = (long) pop(unsigned short);
	} else {
	    r = (long) pop(short);
	}
    } else if (n == sizeof(long)) {
	r = pop(long);
    } else {
	error("[internal error: size %d in popsmall]", n);
    }
    return r;
}

/*
 * Evaluate a conditional expression.
 */

public Boolean cond(p)
Node p;
{
    register Boolean b;

    if (p == nil) {
	b = true;
    } else {
	eval(p);
	b = (Boolean) pop(Boolrep);
    }
    return b;
}

/*
 * Return the address corresponding to a given tree.
 */

public Address lval(p)
Node p;
{
    if (p->op == O_RVAL) {
	eval(p->value.arg[0]);
    } else {
	eval(p);
    }
    return (Address) (pop(long));
}

/*
 * set a temporary breakpoint at the source line given in 'place' 
 * in the current process.
 */
private tempbp(place, why)
Node place;
Stopinfo why;
{
    Node event, action;

    if (place != nil) {
	action = build(O_DBXSTOP, why);
	event = build(O_EQ, build(O_SYM, linesym), place);
	inherit(event_temp(wp->p_id, event, buildcmdlist(action)));
    }
}

/*
 * Process a trace command, translating into the appropriate events
 * and associated actions.
 */

public trace(p)
Node p;
{
    Node exp, place, cond, procs;
    Node left;
    int who;

    exp = p->value.arg[0];
    if (exp != nil and exp->nodetype != nil and exp->nodetype->class == COMMON)
	error("Cannot trace the name of a common block.");
    place = p->value.arg[1];
    cond = p->value.arg[2];
    procs = p->value.arg[3];
    pushproto(procs);
    if (procs->op == O_LCON) {
	who = procs->value.lcon;
    } else {
	who = 0;
    }
    if (who <= 0) {
	who = ALLPROCS;
    }
    if (exp == nil) {
	traceall(p->op, place, cond, who);
    } else if (exp->op == O_QLINE or exp->op == O_LCON) {
	traceinst(p->op, exp, cond, who);
    } else if (place != nil and (place->op == O_QLINE or place->op == O_LCON)) {
	traceat(p->op, exp, place, cond, who);
    } else {
	left = exp;
	if (left->op == O_RVAL or left->op == O_CALL) {
	    left = left->value.arg[0];
	}
	if (left->op == O_SYM and isblock(left->value.sym)) {
	    traceproc(left->value.sym, cond, who);
	} else {
	    tracedata(p->op, exp, place, cond, who);
	}
    }
    popproc();
}

/*
 * Set a breakpoint that will turn on tracing.
 */

private traceall(op, place, cond, who)
Operator op;
Node place;
Node cond;
int who;
{
    Address addr;
    Symbol s;
    Node event;
    Command action;

    /*
     * figure out when to turn on tracing.  codeloc() for a fortran
     * or pascal symbol is the address of the first line of source rather
     * than the address of the prologue.  if we turn tracing on then,
     * the first line of the subroutine won't get traced.  so we turn
     * on tracing at entryloc() for these with an instruction breakpoint.
     * this is pretty gross... eventually the compiler should put out
     * line number info for the prologue.
     */
    if (place == nil) {
	event = build(O_EQ, build(O_SYM, procsym), build(O_SYM, program));
    } else {
	s = place->value.sym;
	if (s->language == lang_pascal || s->language == lang_fortran) {
	    addr = entryloc(place->value.sym);
	    event = build(O_EQ, build(O_SYM, pcsym), build(O_LCON, addr));
	} else
	    event = build(O_EQ, build(O_SYM, procsym), build(O_SYM, s));
    }

    action = build(O_PRINTSRCPOS,
	build(O_QLINE, nil, build(O_LCON, (op == O_TRACE) ? 1 : 0)));
    if (cond != nil) {
	action = build(O_IF, cond, buildcmdlist(action));
    }
    action = build(O_TRACEON, (op == O_TRACEI), buildcmdlist(action));
    action->value.trace.inherit = (who == ALLPROCS ? true : false);
    action->value.trace.event = addevent(who, event, buildcmdlist(action));
    if (notsourcing()) {
	printevent(action->value.trace.event);
    }
}

/*
 * Set up the appropriate breakpoint for tracing an instruction.
 */

private traceinst(op, exp, cond, who)
Operator op;
Node exp;
Node cond;
int who;
{
    Node event;
    Command action;
    Event e;

    if (op == O_TRACEI and exp->op == O_LCON) {
	event = build(O_EQ, build(O_SYM, pcsym), exp);
    } else {
	if (exp->op == O_LCON) {
	    checksource();
	    exp = build(O_QLINE, build(O_SCON, strdup(cursource)), exp);
	}
	event = build(O_EQ, build(O_SYM, linesym), exp);
    }
    action = build(O_PRINTSRCPOS, exp);
    if (cond) {
	action = build(O_IF, cond, buildcmdlist(action));
    }
    e = addevent(who, event, buildcmdlist(action));
    if (notsourcing()) {
	printevent(e);
    }
}

/*
 * Set a breakpoint to print an expression at a given line or address.
 */

private traceat(op, exp, place, cond, who)
Operator op;
Node exp;
Node place;
Node cond;
int who;
{
    Node event;
    Command action;
    Event e;

    if (op == O_TRACEI) {
	event = build(O_EQ, build(O_SYM, pcsym), place);
    } else {
	if (place->op == O_LCON) {
	    checksource();
	    place = build(O_QLINE, build(O_SCON, strdup(cursource)), place);
	}
	event = build(O_EQ, build(O_SYM, linesym), place);
    }
    action = build(O_PRINTSRCPOS, exp);
    if (cond != nil) {
	action = build(O_IF, cond, buildcmdlist(action));
    }
    e = addevent(who, event, buildcmdlist(action));
    if (notsourcing()) {
	printevent(e);
    }
}

/*
 * Construct event for tracing a procedure.
 *
 * What we want here is
 *
 * 	when $proc = p do
 *	    if <condition> then
 *	        printcall;
 *	        once $pc = $retaddr do
 *	            printrtn;
 *	        end;
 *	    end if;
 *	end;
 *
 * Note that "once" is like "when" except that the event
 * deletes itself as part of its associated action.
 */

private traceproc(p, cond, who)
Symbol p;
Node cond;
{
    Node event;
    Command action;
    Cmdlist actionlist;
    Event e;

    action = build(O_PRINTCALL, p);
    actionlist = list_alloc();
    cmdlist_append(action, actionlist);
    event = build(O_EQ, build(O_SYM, pcsym), build(O_SYM, retaddrsym));
    action = build(O_ONCE, event, buildcmdlist(build(O_PRINTRTN, p)));
    cmdlist_append(action, actionlist);
    if (cond != nil) {
	actionlist = buildcmdlist(build(O_IF, cond, actionlist));
    }
    event = build(O_EQ, build(O_SYM, procsym), build(O_SYM, p));
    e = addevent(who, event, actionlist);
    if (notsourcing()) {
	printevent(e);
    }
}

/*
 * Set up breakpoint for tracing data.
 */

private tracedata(op, exp, place, cond, who)
Operator op;
Node exp;
Node place;
Node cond;
int who;
{
    Symbol p;
    Node event;
    Command action;

    p = (place == nil) ? tcontainer(exp) : place->value.sym;
    if (p == nil or p->class == MODULE or p->class == COMMON) {
	p = program;
    }
    action = build(O_PRINTIFCHANGED, exp);
    if (cond != nil) {
	action = build(O_IF, cond, buildcmdlist(action));
    }
    action = build(O_TRACEON, (op == O_TRACEI), buildcmdlist(action));
    event = build(O_EQ, build(O_SYM, procsym), build(O_SYM, p));
    action->value.trace.inherit = (who == ALLPROCS ? true : false);
    action->value.trace.event = addevent(who, event, buildcmdlist(action));
    if (notsourcing()) {
	printevent(action->value.trace.event);
    }
}

/*
 * Setting and unsetting of stops.
 */

public stop(p)
Node p;
{
    Node exp, place, cond, t, procs;
    Symbol s;
    Command action;
    Event e;
    int who;

    exp = p->value.arg[0];
    place = p->value.arg[1];
    cond = p->value.arg[2];
    procs = p->value.arg[3];
    pushproto(procs);
    if (procs->op == O_LCON) {
	who = procs->value.lcon;
    } else {
	who = 0;
    }
    if (exp != nil) {
	stopvar(p->op, exp, place, cond, who);
    } else {
	action = build(O_STOPX, who);
	if (who <= 0)
		who = ALLPROCS;
	if (cond != nil) {
	    action = build(O_IF, cond, buildcmdlist(action));
	}
	if (place == nil or place->op == O_SYM) {
	    if (place == nil) {
		s = program;
	    } else {
		s = place->value.sym;
	    }
	    t = build(O_EQ, build(O_SYM, procsym), build(O_SYM, s));
	    if (cond != nil) {
		action = build(O_TRACEON, (p->op == O_STOPI),
		    buildcmdlist(action));
		action->value.trace.inherit = (who == ALLPROCS ? true : false);
		e = addevent(who, t, buildcmdlist(action));
		action->value.trace.event = e;
	    } else {
		e = addevent(who, t, buildcmdlist(action));
	    }
	    if (notsourcing()) {
		printevent(e);
	    }
	} else {
	    stopinst(p->op, place, action, who);
	}
    }
    popproc();
}

private stopinst(op, place, action, who)
Operator op;
Node place;
Command action;
int who;
{
    Node event;
    Event e;

    if (op == O_STOP) {
	event = build(O_EQ, build(O_SYM, linesym), place);
    } else {
	event = build(O_EQ, build(O_SYM, pcsym), place);
    }
    e = addevent(who, event, buildcmdlist(action));
    if (notsourcing()) {
	printevent(e);
    }
}

/*
 * Implement stopping on assignment to a variable by adding it to
 * the variable list.
 */

private stopvar(op, exp, place, cond, who)
Operator op;
Node exp;
Node place;
Node cond;
int who;
{
    Symbol p;
    Node event;
    Command action;

    if (place == nil) {
	if (exp->op == O_LCON) {
	    p = program;
	} else {
	    p = tcontainer(exp);
	    if (p == nil or p->class == MODULE) {
		p = program;
	    }
	}
    } else {
	p = place->value.sym;
    }
    action = build(O_STOPIFCHANGED, exp, build(O_LCON, who == ALLPROCS));
    if (who <= 0)
	who = ALLPROCS;
    if (cond != nil) {
	action = build(O_IF, cond, buildcmdlist(action));
    }
    action = build(O_TRACEON, (op == O_STOPI), buildcmdlist(action));
    event = build(O_EQ, build(O_SYM, procsym), build(O_SYM, p));
    action->value.trace.inherit = (who == ALLPROCS ? true : false);
    action->value.trace.event = addevent(who, event, buildcmdlist(action));
    if (notsourcing()) {
	printevent(action->value.trace.event);
    }
}

/*
 * Is a 4 byte expression a floating point number?
 * ie. "float", "real*4", or "real".
 */

private Boolean is4bytefp(exp)
Node exp;
{
	register Symbol t;

	if (exp == nil or exp->nodetype == nil)
		return(false);

	t = exp->nodetype;
	if (t->class == TYPE) {
		if ( istypename(t, "float" ) or
		     istypename(t, "real*4") or
		     istypename(t, "real"  )        )
			return(true);
	}

	if (exp->nodetype->type == nil or exp->nodetype->type->class != TYPE)
		return(false);

	t = exp->nodetype->type;	
	if ( istypename(t, "float" ) or
	     istypename(t, "real*4") or
	     istypename(t, "real"  )        )
		return(true);

	return(false);
}

/*
 * Assign the value of an expression to a variable (or term).
 */


# ifndef OLDASSIGN	/* New assignment routine, old one follows. */

/*
 * Assign stmt is of the form:  " assign A = B ", where "A" could be a symbol
 * or address.  The value for the assignment, "B" can be any expression,
 * either a symbol or an address.  This routine initially decides if "A"
 * is a symbol stored in a register.  The outcome of this decision frorces
 * the code through one of two paths.
 *
 * In either case we just "eval(exp)" to get the value of the expression on
 * the stack.  Then we have to transfer the value from the stack to "A".
 * Some size (type) conversion may be necessary.
 *
 * If the variable being assigned to is a register and not in the current
 * procedure the variable may be saved on the stack.  Some difficulty in
 * dealing with this becuause we don't want to change the register, instead
 * changing what has been saved on the stack.  Problem compounded by floating
 * point numbers.  Probably could be done but too much work for now.  To
 * implement take a look at routine "address()" and see how it deals with
 * registers.  Would probably want to either figure out address of where
 * variable was saved and jump down to code that deals with non-registers,
 * or maybe something like "savereg()" that writes into frame instead of
 * returning value from the frame.
 */

public assign(var, exp)
Node var;
Node exp;
{
    integer varsize;		/* # bytes in variable */
    integer expsize;		/* # bytes in expression */
				/* variables for conversions */
    int nreg;			/* register # of variable */
    boolean exp_4bytefp;	/* is expression a 4 byte fp value */
    Address addr;
    char cvalue;
    long lvalue;
    float fvalue;
    double dvalue;
    int pad;
    union {
	long  v_l;
	float v_f;
	Word  v_w;
    } v;

    Frame frp, nframe;		/* frame for register variable */
    Address reg_fp;		/* frame poiner of frame holding register */
    Address cur_fp;		/* current frame pointer */
    Symbol s;			/* symbol of this variable */

    if (var->op == O_SYM and regnum(var->value.sym) != -1) {
    /* Assign to a symbol which is held in a register. */
	if (isfpreg(var->value.sym))
	    varsize = sizeof(double);
	else
	    varsize = size(var->nodetype);
	if (exp->op == O_SYM and isfpreg(exp->nodetype))
	    expsize = sizeof(double);
	else
	    expsize = size(exp->nodetype);
	nreg = regnum(var->value.sym);
	/*
	 * Get the frame for this reg.  If nil assume was specified as internal
	 * register name and is ok.  Otherwise see if the frame pointer for the
	 * symbol matches current frame pointer.  If so all is well and we can
	 * do the assignment.  Otherwise, register variable may have been saved
	 * on the stack.
	 */
	nframe = nil;
	s = var->value.sym;
	frp = symframe(s, nframe);
	if (frp != nil) {
	    reg_fp = savereg(FRP, frp);
	    cur_fp = reg(FRP);
	} else
	    reg_fp = cur_fp = 0;
	if (traceassign) {
	    printf("assign: Assigning to a symbol held in a register.\n");
	    printf("varsize= %d, expsize= %d, regnum= %d\n",
		varsize, expsize, nreg);
	    printf("var:  ");
	    psym(s);
	    if (frp == nil) {
		printf("got nil frame - internal reg?\n");
	    } else {
		printf("fp of frame active for symbol is 0x%x\n", reg_fp);
		printf("Current fp is 0x%x\n", cur_fp);
	    }
	    fflush(stdout);
	}
	if (reg_fp != cur_fp)
	    error("cannot alter register unless in current procedure");
	/*
	 * Evaluate the expression which has the value for the assignment.
	 * This leaves it on the stack.  When we have a 4 bytes for the
	 * expression we need to know if this is a 4 byte floating point
	 * number (such as a Fortran real variable).
	 */
	exp_4bytefp = is4bytefp(exp);
	eval(exp);
	/* Assign to var, depending on size of var. */
	if (varsize == sizeof(double)) {
	    /* Variable is a double. */
	    if (expsize == sizeof(double) ) {
		/* Expression is a double. */
		dvalue = (double) pop(double);
	    } else if (expsize == sizeof(float) and exp_4bytefp) {
		/* Expression is a 4 byte floating point number. */
		popn(expsize, &fvalue);
		dvalue = (double) fvalue;
	    } else {
		/* Expression is 4 byte, or smaller, value. */
		dvalue = (double) popsmall(exp->nodetype);
	    }
	    if (traceassign) {
		printf("assigning %f to variable\n", dvalue);
		fflush(stdout);
	    }
	    setregfp(nreg, dvalue);
	} else {
	    /* Variable is not a double. */
	    if (expsize == sizeof(double)) {
		/* Expression is a double. */
		dvalue = (double) pop(double);
		if (varsize == sizeof(float) and is4bytefp(var)) {
		    /* Variable is a 4 byte floating point number. */
		    if (dvalue > 3.4E38 or dvalue < -3.4E38 )
			error("value out of range for assign");
		    else
			v.v_f = (float) dvalue;
		} else {
		    /* Variable is 4 byte or less value. */
		    if (dvalue > 2147483647.0 or dvalue < -2147483647.0 )
			error("value out of range for assign");
		    else
			v.v_l = (long) dvalue;
		}
	    } else if (expsize == sizeof(float) and exp_4bytefp) {
		/* Expression is a 4 byte floating point number. */
		popn(expsize, &fvalue);
		if (varsize == sizeof(float) and is4bytefp(var)) {
		    /* Variable is a 4 byte floating point number. */
		    v.v_f = fvalue;
		} else {
		    /* Variable is 4 byte or less value. */
		    if (fvalue > 2147483647.0 or fvalue < -2147483647.0 )
			error("value out of range for assign");
		    else
			v.v_l = (long) fvalue;
		}
	    } else {
		/* Variable and expression are both 4 byte or less, not fp. */
		v.v_l = popsmall(exp->nodetype);
	    }
	    setreg(nreg, v.v_w);
	}
    } else {	/* not a register */
	addr = lval(var);
	varsize = size(var->nodetype);
	if (exp->op == O_SYM and isfpreg(exp->nodetype))
	    expsize = sizeof(double);
	else
	    expsize = size(exp->nodetype);
	if (traceassign) {
	    printf("assign: varsize= %d, expsize= %d, addr= 0x%x\n",
		varsize, expsize, addr);
	    fflush(stdout);
	}
	eval(exp);
	if (rtype(var->nodetype)->class == STR) {
	    if (traceassign) {
		printf("assign: rtype(var->nodetype)->class = STR\n");
		fflush(stdout);
	    }
	    sp -= expsize;
	    if (exp->op == O_SCON)
		expsize--;			/*toss trailing '\0'*/
	    if (var->nodetype->language == lang_pascal) {
		varsize--;			/*varsize includes size byte*/
		cvalue = min(varsize, expsize);
		dwrite(&cvalue, addr, 1);
		addr++;				/*put data after size byte*/
	    }
	    if (var->nodetype->language == lang_fortran && expsize < varsize) {
		pad = varsize - expsize;
		cvalue = ' ';
		while (pad) {
		    dwrite(&cvalue, addr + varsize - pad, 1);
		    pad--;
		}
	    }
	    dwrite(sp, addr, min(expsize, varsize));
	    return;
	}
	if (varsize == sizeof(float) and expsize == sizeof(double)) {
	    fvalue = (float) pop(double);
	    if (traceassign) {
		printf("assign: %d && %d, fvalue = pop(double) = %f\n",
		    sizeof(float), sizeof(double), (double) fvalue);
		fflush(stdout);
	    }
	    if (is4bytefp(var)) {  /* now assign float to float */
		if (traceassign) {
		    printf("assign: float to float\n");
		    fflush(stdout);
		}
		dwrite(&fvalue, addr, sizeof(fvalue));
	    } else {		  /* assign float to int */
		if (traceassign) {
		    printf("assign: float to int\n");
		    fflush(stdout);
		}
		if (fvalue > 2147483647.0 or fvalue < -2147483647.0)
		    error("value out of range for assign");
		lvalue = (int) fvalue;
		dwrite(&lvalue, addr, sizeof(lvalue));
	    }
	} else if (varsize == sizeof(double) and expsize == sizeof(double)) {
	    /* DEM 9-Jul-86 */
	    dvalue = (double) pop(double);
	    if (traceassign) {
		printf("assign: %d && %d, dvalue = pop(double) = %f\n",
		    sizeof(double), sizeof(double), dvalue);
		fflush(stdout);
	    }
	    dwrite(&dvalue, addr, sizeof(dvalue));
	} else if (varsize == sizeof(double)) {
	    /* DEM 9-Jul-86 */
	    if (expsize == sizeof(float) and is4bytefp(exp)) {
		popn(expsize, &fvalue);
		dvalue = (double) fvalue;
		if (traceassign) {
		    printf("assign: %d && %d, dvalue = popn = %f, using setregfp\n",
			sizeof(double), expsize, dvalue);
		    fflush(stdout);
		    }
	    } else {
		dvalue = (double) popsmall(exp->nodetype);
		if (traceassign) {
		    printf("assign: %d && %d, dvalue = popsmall = %f, using setregfp\n",
			sizeof(double), expsize, dvalue);
		    fflush(stdout);
		    }
	    }
	    dwrite(&dvalue, addr, sizeof(dvalue));
	} else {
	    if (varsize <= sizeof(long)) {
		if (traceassign) {
		    printf("assign: varsize <= sizeof(long) :: %d <= %d\n",
			varsize, sizeof(long));
		    printf("assign: using popn(%d, &lvalue)\n", expsize);
		    fflush(stdout);
		}
		popn(expsize, &lvalue);
		if (expsize == sizeof(char))
		    lvalue = (char) lvalue;
		else if (expsize == sizeof(short))
		    lvalue = (short) lvalue;
		if (traceassign) {
		    if (expsize == sizeof(char))
			printf("assign: expsize = sizeof(char) :: %d == %d\n",
			    expsize, sizeof(char));
		    else if (expsize == sizeof(short))
			printf("assign: expsize = sizeof(short) :: %d == %d\n",
			    expsize, sizeof(short));
		    else
			printf("assign: expsize= %d, no cast of lvalue (%d)\n",
			    sizeof(lvalue));
		    fflush(stdout);
		}
		dwrite(&lvalue, addr, varsize);
	    } else {
		if (expsize <= varsize) {
		    sp -= expsize;
		    dwrite(sp, addr, expsize);
		} else {
		    sp -= expsize;
		    dwrite(sp, addr, varsize);
		}
		if (traceassign) {
		    if (expsize <= varsize) {
			printf("assign: expsize <= varsize :: %d <= %d\n",
			    expsize, varsize);
			printf("assign: dwrite(, , expsize)\n");
		    } else {
			printf("assign: expsize > varsize :: %d > %d\n",
			    expsize, varsize);
			printf("assign: dwrite(, , varsize)\n");
		    }
		    fflush(stdout);
		}
	    }
	}
    }
}

# else 		/* Old assignment routine, prior to March 1987 rewrite. */

public assign(var, exp)
Node var;
Node exp;
{
    Address addr;
    integer varsize, expsize;
    char cvalue;
    long lvalue;
    float fvalue;
    double dvalue;
    int pad;

    if (var->op == O_SYM and regnum(var->value.sym) != -1) {
	varsize = size(var->nodetype);
	expsize = size(exp->nodetype);
	if (traceassign) {
	    printf("assign: varsize= %d, expsize= %d, regnum= %d\n",
		varsize, expsize, regnum(var->value.sym));
	    fflush(stdout);
	}
	eval(exp);
	if (varsize == sizeof(double) and expsize == sizeof(double)) {
	    /* DEM 8-Jul-86 */
	    dvalue = (double) pop(double);
	    if (traceassign) {
		printf("assign: %d && %d, dvalue = pop(double) = %f, using setregfp\n",
		    sizeof(double), sizeof(double), dvalue);
		fflush(stdout);
	    }
	    setregfp(regnum(var->value.sym), dvalue);
	} else if (varsize == sizeof(double)) {
	    /* DEM 8-Jul-86 */
	    if (expsize == sizeof(float) and is4bytefp(exp)) {
		popn(expsize, &fvalue);
		dvalue = (double) fvalue;
		if (traceassign) {
		    printf("assign: %d && %d, dvalue = popn = %f, using setregfp\n",
			sizeof(double), expsize, dvalue);
		    fflush(stdout);
		    }
	    } else {
		dvalue = (double) popsmall(exp->nodetype);
		if (traceassign) {
		    printf("assign: %d && %d, dvalue = popsmall = %f, using setregfp\n",
			sizeof(double), expsize, dvalue);
		    fflush(stdout);
		    }
	    }
	    setregfp(regnum(var->value.sym), dvalue);
	} else
	    if (traceassign) {
		printf("assign: %d && %d, using setreg(%d, popsmall())\n",
		    varsize, expsize, regnum(var->value.sym));
		fflush(stdout);
	    }
	    setreg(regnum(var->value.sym), popsmall(exp->nodetype));
    } else {
	addr = lval(var);
	varsize = size(var->nodetype);
	expsize = size(exp->nodetype);
	if (traceassign) {
	    printf("assign: varsize= %d, expsize= %d, addr= 0x%x\n",
		varsize, expsize, addr);
	    fflush(stdout);
	}
	eval(exp);
	if (rtype(var->nodetype)->class == STR) {
	    if (traceassign) {
		printf("assign: rtype(var->nodetype)->class = STR\n");
		fflush(stdout);
	    }
	    sp -= expsize;
	    if (exp->op == O_SCON)
		expsize--;			/*toss trailing '\0'*/
	    if (var->nodetype->language == lang_pascal) {
		varsize--;			/*varsize includes size byte*/
		cvalue = min(varsize, expsize);
		dwrite(&cvalue, addr, 1);
		addr++;				/*put data after size byte*/
	    }
	    if (var->nodetype->language == lang_fortran && expsize < varsize) {
		pad = varsize - expsize;
		cvalue = ' ';
		while (pad) {
		    dwrite(&cvalue, addr + varsize - pad, 1);
		    pad--;
		}
	    }
	    dwrite(sp, addr, min(expsize, varsize));
	    return;
	}
	if (varsize == sizeof(float) and expsize == sizeof(double)) {
	    fvalue = (float) pop(double);
	    if (traceassign) {
		printf("assign: %d && %d, fvalue = pop(double) = %f\n",
		    sizeof(float), sizeof(double), (double) fvalue);
		fflush(stdout);
	    }
	    dwrite(&fvalue, addr, sizeof(fvalue));
	} else if (varsize == sizeof(double) and expsize == sizeof(double)) {
	    /* DEM 9-Jul-86 */
	    dvalue = (double) pop(double);
	    if (traceassign) {
		printf("assign: %d && %d, dvalue = pop(double) = %f\n",
		    sizeof(double), sizeof(double), dvalue);
		fflush(stdout);
	    }
	    dwrite(&dvalue, addr, sizeof(dvalue));
	} else if (varsize == sizeof(double)) {
	    /* DEM 9-Jul-86 */
	    if (expsize == sizeof(float) and is4bytefp(exp)) {
		popn(expsize, &fvalue);
		dvalue = (double) fvalue;
		if (traceassign) {
		    printf("assign: %d && %d, dvalue = popn = %f, using setregfp\n",
			sizeof(double), expsize, dvalue);
		    fflush(stdout);
		    }
	    } else {
		dvalue = (double) popsmall(exp->nodetype);
		if (traceassign) {
		    printf("assign: %d && %d, dvalue = popsmall = %f, using setregfp\n",
			sizeof(double), expsize, dvalue);
		    fflush(stdout);
		    }
	    }
	    dwrite(&dvalue, addr, sizeof(dvalue));
	} else {
	    if (varsize <= sizeof(long)) {
		if (traceassign) {
		    printf("assign: varsize <= sizeof(long) :: %d <= %d\n",
			varsize, sizeof(long));
		    printf("assign: using popn(%d, &lvalue)\n", expsize);
		    fflush(stdout);
		}
		popn(expsize, &lvalue);
		if (expsize == sizeof(char))
		    lvalue = (char) lvalue;
		else if (expsize == sizeof(short))
		    lvalue = (short) lvalue;
		if (traceassign) {
		    if (expsize == sizeof(char))
			printf("assign: expsize = sizeof(char) :: %d == %d\n",
			    expsize, sizeof(char));
		    else if (expsize == sizeof(short))
			printf("assign: expsize = sizeof(short) :: %d == %d\n",
			    expsize, sizeof(short));
		    else
			printf("assign: expsize= %d, no cast of lvalue (%d)\n",
			    sizeof(lvalue));
		    fflush(stdout);
		}
		dwrite(&lvalue, addr, varsize);
	    } else {
		if (expsize <= varsize) {
		    sp -= expsize;
		    dwrite(sp, addr, expsize);
		} else {
		    sp -= expsize;
		    dwrite(sp, addr, varsize);
		}
		if (traceassign) {
		    if (expsize <= varsize) {
			printf("assign: expsize <= varsize :: %d <= %d\n",
			    expsize, varsize);
			printf("assign: dwrite(, , expsize)\n");
		    } else {
			printf("assign: expsize > varsize :: %d > %d\n",
			    expsize, varsize);
			printf("assign: dwrite(, , varsize)\n");
		    }
		    fflush(stdout);
		}
	    }
	}
    }
}

#  endif OLDASSIGN


/*
 * Set a debugger variable.
 *
 * The set command had several problems.  It used to unrval() the expression
 * here and in commands.y would introduce the O_RVAL node.  This broke all
 * other uses of the set command so seems better to just leave it alone here.
 * Note: could now collapse last two else's into one, but want to preserve
 * for the diff file.
 */

private set (var, exp)
Node var, exp;
{
    if (var == nil) {
	defvar(nil, nil);
    } else if (exp == nil) {
	defvar(var->value.name, nil);
    } else {
	defvar(var->value.name, exp);		/* used to unrval(exp) */
    }
}

/*
 * Execute a list command.
 */

private list (p, listi)
Node p;
boolean listi;
{
    Symbol f;
    Address addr;
    Lineno line, l1, l2;

    if (p->value.arg[0]->op == O_SYM) {
	f = p->value.arg[0]->value.sym;
	addr = firstline(f);
	if (addr == NOADDR) {
	    error("no source lines for \"%s\"", symname(f));
	}
	setsource(srcfilename(addr));
	line = srcline(addr);
	getsrcwindow(line, &l1, &l2);
    } else {
	eval(p->value.arg[0]);
	l1 = (Lineno) (pop(long));
	eval(p->value.arg[1]);
	l2 = (Lineno) (pop(long));
    }
    printlines(l1, l2, listi);
}

/*
 * Execute a func command.
 */

private func (p)
Node p;
{
    Symbol s, f;
    Address addr;

    if (p == nil) {
	printname(stdout, curfunc);
	putchar('\n');
    } else {
	s = p->value.sym;
	if (isroutine(s)) {
	    setcurfunc(s);
	} else {
	    find(f, s->name) where isroutine(f) endfind(f);
	    if (f == nil) {
		error("%s is not a procedure or function", symname(s));
	    }
	    setcurfunc(f);
	}
	addr = codeloc(curfunc);
	if (addr != NOADDR) {
	    setsource(srcfilename(addr));
	    cursrcline = srcline(addr);
	}
    }
}

/*
 * Give the user some help.
 */

public help()
{
    puts("Refer to \"DYNIX Pdbx Debugger User's Manual\" for complete details");
    puts("run                    - begin execution of the program");
    puts("ps                     - print status of processes/core dumps");
    puts("print <exp>            - print the value of the expression");
    puts("where                  - print currently active procedures");
    puts("%<number>              - switch to context of process `number'");
    puts("stop at <line>         - suspend execution at the line");
    puts("stop in <proc>         - suspend execution when <proc> is called");
    puts("cont                   - continue execution");
    puts("step                   - single step one line");
    puts("next                   - step to next line (skip over calls)");
    puts("trace <line#>          - trace execution of the line");
    puts("trace <proc>           - trace calls to the procedure");
    puts("trace <var>            - trace changes to the variable");
    puts("trace <exp> at <line#> - print <exp> when <line> is reached");
    puts("status                 - print trace/stop's in effect");
    puts("delete <number>        - remove trace or stop of given number");
    puts("call <proc>            - call a procedure in program");
    puts("whatis <name>          - print the declaration of the name");
    puts("list <line>, <line>    - list source lines");
    puts("quit                   - exit dbx");
}

/*
 * Divert output to the given file name.
 * Cannot redirect to an existing file.
 */

private int so_fd;
private Boolean notstdout;

public setout(filename)
String filename;
{
    File f;

    f = fopen(filename, "r");
    if (f != nil) {
	fclose(f);
	error("%s: file already exists", filename);
    } else {
	so_fd = dup(1);
	close(1);
	if (creat(filename, 0666) == nil) {
	    unsetout();
	    error("can't create %s", filename);
	}
	notstdout = true;
    }
}

/*
 * Revert output to standard output.
 */

public unsetout()
{
    fflush(stdout);
    close(1);
    if (dup(so_fd) != 1) {
	panic("standard out dup failed");
    }
    close(so_fd);
    notstdout = false;
}

/*
 * Determine is standard output is currently being redirected
 * to a file (as far as we know).
 */

public Boolean isredirected()
{
    return notstdout;
}

/*
 * Initialize stack.  Used to clean up stack after error.
 */

public reset_stack()
{
	sp = &stack[0];
}
