/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */

#ifndef lint
static	char sccsid[] = "@(#)events.c	1.7 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: events.c 2.7 87/09/15 $";
#endif

/*
 * Event/breakpoint managment.
 */

#include "defs.h"
#include "events.h"
#include "main.h"
#include "symbols.h"
#include "tree.h"
#include "eval.h"
#include "source.h"
#include "mappings.h"
#include "runtime.h"
#include "multi-p.h"
#include "machine.h"
#include "lists.h"

#ifndef public
typedef struct Event *Event;
typedef struct Breakpoint *Breakpoint;

#include "symbols.h"

Symbol linesym;
Symbol procsym;
Symbol pcsym;
Symbol retaddrsym;
int    cureventid;	/* set to event id before calling evalcmdlist()
			   for event.  used by setstopped() to identify 
			   which event stopped the process */
/*boolean objselected;*/ /* if true, don't allow translate() to pick the
			   object on the basis of the filename. */

#define ALLPROCS	-1
#define THISPROC	(wp->p_id)

/*
 * create a permanent, user-visible event (e.g. "stop at 10")
 */
#define addevent(p, cond, cmdlist) \
	event_alloc(p, false, false, false, cond, cmdlist)

/*
 * create an "invisible" event that deletes itself when hit (e.g. O_ONCE).
 */
#define event_once(p, cond, cmdlist) \
	event_alloc(p, true, true, false, cond, cmdlist)

/*
 * invisible event that self destructs even if it isn't hit
 * ("cont to", "return").
 */
#define event_temp(p, cond, cmdlist) \
	event_alloc(p, true, true, true, cond, cmdlist)

/*
 * create a permanent, invisible event.
 */
#define dbxevent(p, cond, cmdlist) \
	event_alloc(p, false, true, false, cond, cmdlist)

#endif

struct Event {
    int id;
    boolean once;		/* delete when hit */
    boolean temp;		/* delete when process stops for any reason */
    Node condition;
    Cmdlist actions;
    int refcount;		/* # of breakpoints that reference this event */
    boolean deleted;		/* been deleted?  keeps out of status output */
    Object *obj;		/* object file this event is for */
};

#define for_this_process(b)	(not (wp->p_flags&PF_NO_BPTS)		\
				 and ((b)->procid == ALLPROCS		\
				      or (b)->procid == wp->p_id))
#define for_dbx(e)		((e)->id < 0)

struct Breakpoint {
    Event event;
    Address bpaddr;
    Lineno bpline;
    Cmdlist actions;
    boolean once;
    boolean temp;
    boolean deleted;
    int procid;
    int refcount;		/* # of processes this bpt is in */
    boolean inherit;		/* inherit this breakpoint over fork? */
};

typedef List Eventlist;
typedef List Bplist;

#define eventlist_append(event, el) list_append(list_item(event), (ListItem) nil, el)
#define bplist_append(bp, bl) list_append(list_item(bp), (ListItem) nil, bl)

private Eventlist eventlist;		/* list of active events */
#define bplist (curobj->o_bplist)	/* list of active breakpoints */
private Event curevent;			/* most recently created event */
private integer eventid;		/* id number of current event */
private integer dbxeventid;		/* same for dbx internal events */
private integer trid;			/* id number of current trace */
private Cmdlist printnewsact;		/* action to call printnews */

typedef struct Trcmd {
    Integer trid;
    Event event;
    Cmdlist cmdlist;
    Integer procid;
    boolean inherit;
} *Trcmd;

#define eachline (curobj->o_eachline) /* commands to execute after each line */
#define eachinst (curobj->o_eachinst) /* commands to execute after each instruction */

struct events_globals {
	boolean events_bps_in;
};
public struct events_globals *events_globals_ptr;
public int events_globals_size = sizeof(struct events_globals);
#define	bps_in	(events_globals_ptr->events_bps_in)

private Breakpoint bp_alloc();

/*
 * Initialize breakpoint information.
 */

private Symbol builtinsym(str, class, type)
String str;
Symclass class;
Symbol type;
{
    Symbol s;

    s = insert(identname(str, true));
    s->language = findlanguage(".s");
    s->class = class;
    s->type = type;
    return s;
}

public bpinit()
{
    linesym = builtinsym("$line", VAR, t_int);
    procsym = builtinsym("$proc", PROC, (Symbol) nil);
    pcsym = lookup(identname("$pc", true));
    if (pcsym == nil) {
	panic("can't find $pc");
    }
    retaddrsym = builtinsym("$retaddr", VAR, t_int);
    eventlist = list_alloc();
    bplist = list_alloc();
    eachline = list_alloc();
    eachinst = list_alloc();
    printnewsact = buildcmdlist(build(O_PRINTNEWS));  /* XXX called too often */
    catch_range_error(0);
    catch_range_error(1);
}

public integer event_id(e)
Event e;
{
    if (e)
	return(e->id);
    else
	return(0);
}

public printwhofor(f, e)
File f;
Event e;
{
    register Breakpoint b;
    foreach(Breakpoint, b, bplist)
	if (b->event == e)
	    goto found;
    endfor
    panic("no breakpoint for event %d curobj %%%s", e->id, curobj->o_name);

found:
    if (b->procid != ALLPROCS)
	fprintf(f, "%%%d ", b->procid);
    else if (multi_objs)
	fprintf(f, "%%%s ", curobj->o_name);
}

/*
 * Trap an event and do the associated commands when it occurs.
 */

public Event event_alloc(procid, isonce, is_for_dbx, istmp, econd, cmdlist)
int procid;
boolean isonce;
boolean is_for_dbx;
boolean istmp;
Node econd;
Cmdlist cmdlist;
{
    register Event e;

    e = new(Event);
    e->id = is_for_dbx ? --dbxeventid : ++eventid;
    e->once = isonce;
    e->temp = istmp;
    e->condition = econd;
    e->actions = cmdlist;
    e->deleted = false;
    e->obj = curobj;
    eventlist_append(e, eventlist);
    curevent = e;
    translate(procid, e);
    return e;
}

/*
 * decide whether to turn off wp->p_tracei and/or wp->p_tracing
 */
private rmtrace()
{
    int ntraces;
    register Trcmd t;

    ntraces = 0;
    foreach (Trcmd, t, eachinst)
	if (for_this_process(t))
	    ntraces++;
    endfor
    if (ntraces == 0) {
	wp->p_tracei = false;
	ntraces = 0;
	foreach (Trcmd, t, eachline)
	    if (for_this_process(t))
	        ntraces++;
	endfor
	if (ntraces == 0)
	    wp->p_tracing = false;
    }
}

/*
 * Delete the event with the given id.
 * Returns whether it's successful or not.
 */

public boolean delevent (id, allprocs)
unsigned int id;
boolean allprocs;
{
    Event e;
    Breakpoint bp;
    Trcmd t;
    boolean found;

    found = false;
    foreach (Event, e, eventlist)
	if (e->id == id) {
	    found = true;
	    foreach (Breakpoint, bp, bplist)
		if (bp->event == e and (allprocs or for_this_process(bp))) {
		    if (tracebpts) {
			fprintf(stderr, "%%%d deleting breakpoint at 0x%x\n", 
				wp->p_id, bp->bpaddr);
		    }
		    delete_bp(bp, list_curitem(bplist));
		}
	    endfor
	    e->deleted = true;
	    if (e->refcount == 0)
		list_delete(list_curitem(eventlist), eventlist);
	    break;
	}
    endfor
    foreach (Trcmd, t, eachline)
	if (t->event->id == id) {
	    found = true;
	    printrmtr(t);
	    list_delete(list_curitem(eachline), eachline);
	}
    endfor
    foreach (Trcmd, t, eachinst)
	if (t->event->id == id) {
	    found = true;
	    printrmtr(t);
	    list_delete(list_curitem(eachinst), eachinst);
	}
    endfor
    doallp(rmtrace);
    return found;
}

/*
 * evaluate a condition for a process if it's appropriate.  used by
 * translate (via evalall()) to turn on bpts for conditions that are
 * true in a stopped process.
 */
public evalevent(proc, actions)
    Symbol proc;
    Cmdlist actions;
{
    if (isprocess() and isactive(proc) and reg(PROGCTR) != codeloc(program))
	evalcmdlist(actions);
}

/*
 * Translate an event into the appropriate breakpoints and actions.
 * While we're at it, turn on the breakpoints if the condition is true.
 */

private translate(procid, e)
int procid;
Event e;
{
    Symbol s;
    Node place;
    Lineno line;
    Address addr;
    String filename;
    boolean pushed = false;

    checkref(e->condition);
    switch (e->condition->op) {
	case O_EQ:
	    if (e->condition->value.arg[0]->op == O_SYM) {
		s = e->condition->value.arg[0]->value.sym;
		place = e->condition->value.arg[1];
		if (s == linesym) {
		    filename = nil;
		    if (place->op == O_QLINE) {
			filename = place->value.arg[0]->value.scon;
			line = place->value.arg[1]->value.lcon;
			pushed = protofile(filename);
			addr = objaddr(line, filename);
		    } else {
			eval(place);
			line = pop(long);
			if (not canReadSource()) {
			    if (not delevent(e->id, true)) {
			        printf("!!dbx.translate: can't undo event %d\n",
				e->id);
			    }
			    checksource();	/* for the error message */
			}
			addr = objaddr(line, cursource);
			e->condition->value.arg[1] = 
			    build(O_QLINE, build(O_SCON, strdup(cursource)), place);
		    }
		    if (addr == NOADDR) {
			if (not delevent(e->id, true)) {
			    printf("!! dbx.translate: can't undo event %d?\n",
				e->id);
			}
			beginerrmsg();
			fprintf(stderr, "no executable code at line ");
			prtree(stderr, place);
			enderrmsg();
		    }
		    (void) bp_alloc(procid, e, addr, line, e->actions);
		    if (pushed) {
			e->obj = curobj;
			popproc();
		    }
		} else if (s == procsym) {
		    eval(place);
		    s = pop(Symbol);
		    (void) bp_alloc(procid, e, codeloc(s), 0, e->actions);
		    evalall(procid, s, e->actions);
		} else if (s == pcsym) {
		    eval(place);
		    (void) bp_alloc(procid, e, pop(Address), 0, e->actions);
		} else {
		    condbp(procid, e);
		}
	    } else {
		condbp(procid, e);
	    }
	    break;

	/*
	 * These should be handled specially.
	 * But for now I'm ignoring the problem.
	 */
	case O_AND:
	case O_OR:
	default:
	    condbp(procid, e);
	    break;
    }
}


/*
 * Create a breakpoint for a condition that cannot be pinpointed
 * to happening at a particular address, but one for which we
 * must single step and check the condition after each statement.
 */

private condbp(procid, e)
int procid;
Event e;
{
    Symbol p;
    Cmdlist actions;

    p = tcontainer(e->condition);
    if (p == nil or p->class == MODULE) {
	p = program;
    }
    actions = buildcmdlist(build(O_IF, e->condition, e->actions));
    actions = buildcmdlist(build(O_TRACEON, false, actions));
    (void) bp_alloc(procid, e, codeloc(p), 0, actions);
}

/*
 * Determine the deepest nested subprogram that still contains
 * all elements in the given expression.
 */

public Symbol tcontainer(exp)
Node exp;
{
    Integer i;
    Symbol s, t, u, v;

    checkref(exp);
    s = nil;
    if (exp->op == O_SYM) {
	s = container(exp->value.sym);
    } else if (not isleaf(exp->op)) {
	for (i = 0; i < nargs(exp->op); i++) {
	    t = tcontainer(exp->value.arg[i]);
	    if (t != nil) {
		if (s == nil) {
		    s = t;
		} else {
		    u = s;
		    v = t;
		    while (u != v and u != nil) {
			u = container(u);
			v = container(v);
		    }
		    if (u == nil) {
			panic("bad ancestry for \"%s\"", symname(s));
		    } else {
			s = u;
		    }
		}
	    }
	}
    }
    return s;
}

/*
 * Determine if the given function can be executed at full speed.
 * This can only be done if there are no breakpoints within the function.
 */

public boolean canskip(f)
Symbol f;
{
    Breakpoint p;
    boolean ok;

    ok = true;
    foreach (Breakpoint, p, bplist)
	if (not p->deleted and for_this_process(p) and whatblock(p->bpaddr) == f) {
	    ok = false;
	    break;
	}
    endfor
    return ok;
}

/*
 * Print out what's currently being traced by looking at
 * the currently active events.
 *
 * Some convolution here to translate internal representation
 * of events back into something more palatable.
 */

public status()
{
    Event e;

    foreach (Event, e, eventlist)
	if (not for_dbx(e) and not e->deleted) {
	    printevent(e);
	}
    endfor
}

public printevent(e)
Event e;
{
    Command cmd;

    checkref(e->obj);
    pushproc(e->obj->o_proto);
    if (not isredirected()) {
	printeventid(e->id);
    }
    cmd = list_element(Command, list_head(e->actions));
    if (cmd->op == O_PRINTCALL) {
	printf("trace ");
	printname(stdout, cmd->value.sym);
    } else {
	if (list_size(e->actions) > 1) {
	    printf("{ ");
	}
	foreach (Command, cmd, e->actions)
	    printcmd(stdout, cmd);
	    if (not list_islast()) {
		printf("; ");
	    }
	endfor
	if (list_size(e->actions) > 1) {
	    printf(" }");
	}
	printcond(e->condition);
    }
    printf("\n");
    popproc();
}

public printeventid (id)
integer id;
{
    printf("[%d] ", id);
}

/*
 * Print out a condition.
 */

private printcond(cond)
Node cond;
{
    Symbol s;
    Node place;

    if (cond->op == O_EQ and cond->value.arg[0]->op == O_SYM) {
	s = cond->value.arg[0]->value.sym;
	place = cond->value.arg[1];
	if (s == procsym) {
	    if (place->value.sym != program) {
		printf(" in ");
		printname(stdout, place->value.sym);
	    }
	} else if (s == linesym) {
	    printf(" at ");
	    prtree(stdout, place);
	} else if (s == pcsym or s == retaddrsym) {
	    printf("i at ");
	    prtree(stdout, place);
	} else {
	    printf(" when ");
	    prtree(stdout, cond);
	}
    } else {
	printf(" when ");
	prtree(stdout, cond);
    }
}

/*
 * Add a breakpoint to the list and return it.
 */

private Breakpoint bp_alloc(procid, e, addr, line, actions)
int procid;
Event e;
Address addr;
Lineno line;
Cmdlist actions;
{
    register Breakpoint p;

    e->refcount++;
    p = new(Breakpoint);
    p->event = e;
    p->bpaddr = addr;
    p->bpline = line;
    p->actions = actions;
    p->once = e->once;
    p->temp = e->temp;
    p->deleted = false;
    p->refcount = 0;
    p->procid = procid;
    p->inherit = false;
    if (tracebpts) {
	if (e == nil) {
	    fprintf(stderr, "%%%d new bp at 0x%x for event ??\n", wp->p_id, addr, e->id);
	} else {
	    fprintf(stderr, "%%%d new bp at 0x%x for event %d\n", wp->p_id, addr, e->id);
	}
    }
    bplist_append(p, bplist);
    return p;
}

/*
 * make event 'e' become inherited over fork by causing the breakpoints
 * that support it to be inherited.  only meaningful for single-process
 * breakpoints.
 */
public Event inherit(e)
register Event e;
{
    register Breakpoint bp;

    foreach (Breakpoint, bp, bplist)
	if (bp->event == e)
	    bp->inherit = true;
    endfor
    return(e);
}

/*
 * Free all storage in the event and breakpoint tables.
 */

public bpfree()
{
    register Event e;

    fixbps();
    foreach (Event, e, eventlist)
	if (not delevent(e->id, true)) {
	    printf("!! dbx.bpfree: can't delete event %d\n", e->id);
	}
	list_delete(list_curitem(eventlist), eventlist);
    endfor
}

/*
 * Determine if the program stopped at a known breakpoint or if we ought to
 * print something out because we're tracing.  if so queue the 
 * associated commands.  return whether a breakpoint was found, not
 * counting the trace stuff.
 */

public boolean bpact()
{
    register Breakpoint p;
    boolean found = false;

    /*
     * make sure we're not in the process of killing this process
     */
    if (killing(wp->p_id))
	return(false);

    /*
     * check for breakpoints 
     */
    foreach (Breakpoint, p, bplist)
	if (p->deleted) {
	    found = true;
	    delete_bp(p, list_curitem(bplist));
	} else if (for_this_process(p) and p->bpaddr == reg(PROGCTR)) {
	    found = true;
	    if (tracebpts) {
	        fprintf(stderr, "%%%d breakpoint for event %d found at location 0x%x\n",
	            wp->p_id, p->event->id, reg(PROGCTR));
	    }
	    assert(wp->p_obj == p->event->obj);
	    qcommand(p->actions, wp->p_id, p->event->id);
	    if (p->once || p->temp) {
	        delete_bp(p, list_curitem(bplist));
	    }
	}
    endfor

    /*
     * if we're tracing, queue O_PRINTNEWS.  this happens after the
     * breakpoints because they turn tracing on and off. 
     */
    if (tracing())
	qcommand(printnewsact, wp->p_id, 0);

     return(found);
}

/*
 * are we at a breakpoint?
 */
public boolean at_bpt()
{
    register Breakpoint p;

    foreach (Breakpoint, p, bplist)
	if (for_this_process(p) and p->bpaddr == reg(PROGCTR))
	    return(true);
    endfor
    return(false);
}

/*
 * Begin single stepping and executing the given commands after each step.
 * If the first argument is true step by instructions, otherwise
 * step by source lines.
 *
 * We automatically set a breakpoint at the end of the current procedure
 * to turn off the given tracing.
 */

public traceon(inst, event, cmdlist, inherit)
boolean inst;
Event event;
Cmdlist cmdlist;
boolean inherit;
{
    register Trcmd trcmd;
    Breakpoint bp;
    Cmdlist actions;
    Address ret;
    Event e;

    if (event == nil) {
	e = curevent;
    } else {
	e = event;
    }
    trcmd = new(Trcmd);
    ++trid;
    trcmd->procid = wp->p_id; 
    trcmd->trid = trid;
    trcmd->event = e;
    trcmd->cmdlist = cmdlist;
    trcmd->inherit = inherit;
    wp->p_tracing = true;
    wp->p_oldpc = 0;
    if (inst) {
	wp->p_tracei = true;
	list_append(list_item(trcmd), (ListItem) nil, eachinst);
    } else {
	list_append(list_item(trcmd), (ListItem) nil, eachline);
    }
    ret = return_addr(nil);
    if (ret != 0) {
	actions = buildcmdlist(build(O_TRACEOFF, trcmd->trid));
	bp = bp_alloc(THISPROC, e, (Address) ret, 0, actions);
	bp->once = true;
	bp->inherit = true;
    }
    if (tracebpts) {
	fprintf(stderr, "%%%d adding trace %d for event %d\n",
		wp->p_id, trcmd->trid, e->id);
    }
    printnews();
}

/*
 * Turn off some kind of tracing.
 * Strictly an internal command, this cannot be invoked by the user.
 */

public traceoff(id)
Integer id;
{
    register Trcmd t;
    register boolean found;

    found = false;
    foreach (Trcmd, t, eachline)
	if (t->trid == id and for_this_process(t)) {
	    printrmtr(t);
	    list_delete(list_curitem(eachline), eachline);
	    found = true;
	    break;
	}
    endfor
    if (not found) {
	foreach (Trcmd, t, eachinst)
	    if (t->trid == id and for_this_process(t)) {
		printrmtr(t);
		list_delete(list_curitem(eachinst), eachinst);
		found = true;
		break;
	    }
	endfor
	if (not found) {
	    beginerrmsg();
	    fprintf(stderr, "[internal error: trace id %d not found]\n", id);
	}
    }
    rmtrace();
}

public copy_trcmds(parent)
    int parent;
{
    register Trcmd t, newt;

    foreach (Trcmd, t, eachline)
	if (t->procid == parent && t->inherit) {
	    newt = new(Trcmd);		/* XXX int side malloc */
	    *newt = *t;
	    newt->procid = wp->p_id;
	    list_append(list_item(newt), (ListItem) nil, eachline);
	}
    endfor

    foreach (Trcmd, t, eachinst)
	if (t->procid == parent && t->inherit) {
	    newt = new(Trcmd);		/* XXX int side malloc */
	    *newt = *t;
	    newt->procid = wp->p_id;
	    list_append(list_item(newt), (ListItem) nil, eachinst);
	}
    endfor

    rmtrace();
}

/*
 * If breakpoints are being traced, note that a Trcmd is being deleted.
 */

private printrmtr(t)
Trcmd t;
{
    if (tracebpts) {
	fprintf(stderr, "%%%d removing trace %d", wp->p_id, t->trid);
	if (t->event != nil) {
	    fprintf(stderr, " for event %d", t->event->id);
	}
	fprintf(stderr, "\n");
    }
}

/*
 * Print out news during single step tracing.  don't do it if the process
 * is about to stop.
 */

public printnews()
{
    register Trcmd t;
    Address addr = reg(PROGCTR);

    if (tracing() && wp->p_newstate != STOPPED) {

	if (iscall(wp->p_oldpc) || entryloc(whatblock(addr)) == addr)
	    callnews(true);

	if (linelookup(addr)) {
	    foreach (Trcmd, t, eachline)
		if (for_this_process(t)) {
		    cureventid = t->event->id;
		    evalcmdlist(t->cmdlist);
		}
	    endfor
	} 

	if (wp->p_tracei) {
	    foreach (Trcmd, t, eachinst)
		if (for_this_process(t)) {
		    cureventid = t->event->id;
		    evalcmdlist(t->cmdlist);
		}
	    endfor
	}

	if (isreturn(reg(PROGCTR)))
	    callnews(false);

	cureventid = 0;
    }
}

/*
 * A procedure call/return has occurred while single-stepping,
 * note it if we're tracing lines.
 */

private boolean chklist();

public callnews(iscall)
boolean iscall;
{
    if (wp->p_tracei or not nosource(curfunc)) {
	if (not chklist(eachline, iscall)) {
	    chklist(eachinst, iscall);
	}
    }
}

private boolean chklist(list, iscall)
List list;
boolean iscall;
{
    register Trcmd t;
    register Command cmd;

    foreach (Trcmd, t, list)
	if (for_this_process(t)) {
	    foreach (Command, cmd, t->cmdlist)
	        if (cmd->op == O_PRINTSRCPOS and
	          (cmd->value.arg[0] == nil or cmd->value.arg[0]->op == O_QLINE)) {
		    if (iscall) {
		        printentry(curfunc);
		    } else {
		        printexit(curfunc);
		    }
		    return true;
	        }
	    endfor
	}
    endfor
    return false;
}

/*
 * When tracing variables we keep a copy of their most recent value
 * and compare it to the current one each time a breakpoint occurs.
 * MAXTRSIZE is the maximum size variable we allow.
 */

#define MAXTRSIZE 512

/*
 * List of variables being watched.
 */

typedef struct Trinfo *Trinfo;

struct Trinfo {
    Node variable;
    Address traddr;
    Symbol trblock;
    char *trvalue;
    int procid;
    Lineno prevline;
    Filename prevfile;
    Address prevaddr;
};

private List trinfolist;

/*
 * Find the trace information record associated with the given record.
 * If there isn't one then create it and add it to the list.
 */

private Trinfo findtrinfo(p)
Node p;
{
    register Trinfo tp;
    boolean isnew;

    isnew = true;
    if (trinfolist == nil) {
	trinfolist = list_alloc();
    } else {
	foreach (Trinfo, tp, trinfolist)
	    if (tp->variable == p && tp->procid == wp->p_id) {
		isnew = false;
		break;
	    }
	endfor
    }
    if (isnew) {
	if (tracebpts) {
	    fprintf(stderr, "%%%d adding trinfo for \"", wp->p_id);
	    prtree(stderr, p);
	    fprintf(stderr, "\"\n");
	}
	tp = new(Trinfo);
	tp->variable = p;
	tp->traddr = lval(p);
	tp->trvalue = nil;
	tp->procid = wp->p_id;
	list_append(list_item(tp), nil, trinfolist);
    }
    return tp;
}

/*
 * Print out the value of a variable if it has changed since the
 * last time we checked.
 */

public printifchanged(p)
Node p;
{
    register Trinfo tp;
    register int n;
    char buff[MAXTRSIZE];
    Filename curfile;
    Address curpc;

    tp = findtrinfo(p);
    n = size(p->nodetype);
    if (tp->trvalue == nil and not ptrace_read_ok(tp->traddr)) {
	printf("Error: cannot read location 0x%x for event [%d], ",
	    tp->traddr, cureventid);
	printf("deleting event.\n");
	delevent(cureventid, true);
	return;
    }
    dread(buff, tp->traddr, n);
    curpc = reg(PROGCTR);
    curfile = srcfilename(curpc);
    if (tp->trvalue == nil) {
	printprocid();
	tp->trvalue = newarr(char, n);
	mov(buff, tp->trvalue, n);
	mov(buff, sp, n);
	sp += n;
	if (curline == nil || curfile == nil)
	    printf("initially (at pc = 0x%x):\t", curpc);
	else
	    printf("initially (at line %d in \"%s\"):\t", curline, curfile);
	prtree(stdout, p);
	printf(" = ");
	printval(p->nodetype);
	putchar('\n');
    } else if (cmp(tp->trvalue, buff, n) != 0) {
	printprocid();
	mov(buff, tp->trvalue, n);
	mov(buff, sp, n);
	sp += n;
	if (tp->prevline == nil || tp->prevfile == nil)
	    printf("after (pc = 0x%x):\t", tp->prevaddr);
	else
	    printf("after line %d in \"%s\":\t", tp->prevline, tp->prevfile);
	prtree(stdout, p);
	printf(" = ");
	printval(p->nodetype);
	putchar('\n');
    }
    tp->prevline = curline;
    tp->prevfile = curfile;
    tp->prevaddr = curpc;
}

/*
 * Stop if the value of the given expression has changed.
 */

public stopifchanged(p, all)
Node p;
boolean all;
{
    register Trinfo tp;
    register int n;
    char buff[MAXTRSIZE];
    Address curpc;

    curpc = reg(PROGCTR);
    tp = findtrinfo(p);
    n = size(p->nodetype);
    if (tp->trvalue == nil and not ptrace_read_ok(tp->traddr)) {
	printf("Error: cannot read location 0x%x for event [%d], ",
	    tp->traddr, cureventid);
	printf("deleting event.\n");
	delevent(cureventid, true);
	return;
    }
    dread(buff, tp->traddr, n);
    if (tp->trvalue == nil) {
	tp->trvalue = newarr(char, n);
	mov(buff, tp->trvalue, n);
    } else if (cmp(tp->trvalue, buff, n) != 0) {
	if (all)
		stopall();
	mov(buff, tp->trvalue, n);
	mov(buff, sp, n);
	sp += n;
	printprocid();
	if (tp->prevline == nil)
	    printf("after (pc = 0x%x):\t", tp->prevaddr);
	else
	    printf("after line %d:\t", tp->prevline);
	prtree(stdout, p);
	printf(" = ");
	printval(p->nodetype);
	putchar('\n');
	setstopped(BY_BPT);
    }
    tp->prevline = curline;
    tp->prevaddr = curpc;
}

/*
 * Free the tracing table.
 */

public trfree()
{
    register Trinfo tp;

    foreach (Trinfo, tp, trinfolist)
	dispose(tp->trvalue);
	dispose(tp);
	list_delete(list_curitem(trinfolist), trinfolist);
    endfor
}

/*
 * clean up old breakpoints before running.
 */

public fixbps()
{
    register Trcmd t;
    register Event e;

    wp->p_tracing = false;
    trfree();
    foreach (Event, e, eventlist)
	if (e->once || e->temp) {
	    if (not delevent(e->id, true)) {
		printf("!! dbx.fixbps: can't find event %d\n", e->id);
	    }
	}
    endfor
    foreach (Trcmd, t, eachline)
	printrmtr(t);
	list_delete(list_curitem(eachline), eachline);
    endfor
    foreach (Trcmd, t, eachinst)
	printrmtr(t);
	list_delete(list_curitem(eachinst), eachinst);
    endfor

    clearbps();
}

/*
 * Set all breakpoints in object code.
 */

public setallbps()
{
    register Breakpoint p;

    if (not bps_in) {
        if (tracebpts)
	    fprintf(stderr, "%%%d (%s) setallbps\n", wp->p_id, curobj->o_name);
        foreach (Breakpoint, p, bplist)
	    if (for_this_process(p))
                insert_bp(p);
        endfor
	bps_in = true;
    }
}

/*
* Undo damage done by "setallbps".
*/

public unsetallbps()
{
    register Breakpoint p;

    if (bps_in) {
        if (tracebpts)
	    fprintf(stderr, "%%%d (%s) unsetallbps\n", wp->p_id, curobj->o_name);
        foreach (Breakpoint, p, bplist)
            if (for_this_process(p))
	        remove_bp(p, list_curitem(bplist));
        endfor
	bps_in = false;
    }
}

/*
 * delete a breakpoint
 */
private delete_bp(bp, item)
    Breakpoint bp;
    ListItem item;
{
    assert(bp == list_element(Breakpoint, item));
    bp->deleted = true;
    if (bp->refcount == 0) {
        if (tracebpts)
		fprintf(stderr, "%%%d deleting bpt 0x%x at 0x%x event 0x%x\n",
			wp->p_id, bp, bp->bpaddr, bp->event);
	assert(bp->event->refcount > 0);
	bp->event->refcount--;
	list_delete(item, bplist);
	/* XXX free it? */
    }
}

/*
 * delete all the breakpoints for process 'id' or just the temporary ones.
 */
public delbpts(id, everybpt)
    int id;
    boolean everybpt;
{
    register Breakpoint bp;

    foreach (Breakpoint, bp, bplist)
	if (bp->procid == id and (everybpt or bp->temp)) {
	    delete_bp(bp, list_curitem(bplist));
	}
    endfor
}

/*
 * take a breakpoint out of a process
 */

private remove_bp(bp, item)
    Breakpoint bp;
    ListItem item;
{
    assert(bp == list_element(Breakpoint, item));
    if (bp->refcount) {
	if (wp->p_state != EXITED)	/* can't do anything w/ EXITED procs */
		unsetbp(bp->bpaddr);
	bp->refcount--;
	if (tracebpts)
		fprintf(stderr, "%%%d removing bpt 0x%x at 0x%x event 0x%x\n",
			wp->p_id, bp, bp->bpaddr, bp->event);
    }
    if (bp->deleted)
	delete_bp(bp, item);
}

private insert_bp(bp)
    Breakpoint bp;
{
    if (not bp->deleted) {
	if (tracebpts)
		fprintf(stderr, "%%%d inserting bpt 0x%x at 0x%x event 0x%x\n",
			wp->p_id, bp, bp->bpaddr, bp->event);
	if (ptrace_write_ok(bp->bpaddr)) {
	    setbp(bp->bpaddr);
	    bp->refcount++;
	} else {
	    printf("Error: cannot set breakpoint at 0x%x", bp->bpaddr);
	    if (bp->event) {
		printf(" for event [%d], deleting event.\n", bp->event->id);
		delevent(bp->event->id, true);
	    } else {
		printf("\n");
	    }
	}
    }
}

/*
 * copy the appropriate breakpoints from the process with id "parent"
 * into the current process.  this is called when the given process 
 * forks.  if the bp is 'in' in the parent, it will be 'in' in the child.
 */
copy_bps(parent)
    int parent;
{
    register Breakpoint bp;
    register Breakpoint newbp;

    assert(parent != wp->p_id);			/* not forking onto self! */
    foreach (Breakpoint, bp, bplist)
	if (bp->procid == parent) {	/* is a per-process bpt for parent */
		if (bp->inherit) {	/* we want to duplicate it in child */
			/*
			 * clone bpt in the new process.
			 */
			assert(bp->refcount <= 1);/* because it's per-process */
			newbp = new(Breakpoint);  /* XXX malloc on int side */
			*newbp = *bp;
			newbp->procid = wp->p_id;
			newbp->event->refcount++;
			bplist_append(newbp, bplist);
			if (tracebpts)
			    fprintf(stderr,
				"%%%d copy bp at 0x%x for event %d from %%%d\n",
				wp->p_id, bp->bpaddr, bp->event->id, parent);
		} else {		/*  don't want bpt in new process */
			/*
			 * breakpoints are copied along with the fork, so
			 * if we're not inheriting, have to delete them.
			 */
			unsetbp(bp->bpaddr);
			if (tracebpts)
			    fprintf(stderr,
			      "%%%d forget bp at 0x%x for event %d from %%%d\n",
			      wp->p_id, bp->bpaddr, bp->event->id, parent);
		}
	} else if (bp->procid == ALLPROCS) {
		/*
		 * if this is an 'all-process' breakpoint, just bump its
		 * refcount since another process is using it.
		 */
		bp->refcount++;
	}
    endfor
}

public printbps()
{
    register Breakpoint bp;
    register Event e;

    foreach (Breakpoint, bp, bplist)
	printf("bp 0x%x at 0x%x for event 0x%x refcount %d ",
		bp, bp->bpaddr, bp->event, bp->refcount);
	if (bp->procid == ALLPROCS)
		printf("ALL ");
	else 
		printf("%%%d ", bp->procid);
	if (bp->once)
		printf("once ");
	if (bp->temp)
		printf("temp ");
	if (bp->deleted)
		printf("deleted ");
	printf("\n");
    endfor

    foreach(Event, e, eventlist)
	printf("event 0x%x id %d refcount %d ", e, e->id, e->refcount);
	if (e->once)
		printf("once ");
	if (e->temp)
		printf("temp ");
	printf("\n");
    endfor

    fflush(stdout);

    printbpsavelist();
}
