/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)fortran.c	1.6 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: fortran.c 2.8 87/10/09 $";
#endif

/*
 * FORTRAN dependent symbol routines.
 */

#include "defs.h"
#include "symbols.h"
#include "printsym.h"
#include "languages.h"
#include "fortran.h"
#include "tree.h"
#include "eval.h"
#include "operators.h"
#include "mappings.h"
#include "runtime.h"
#include "machine.h"
#include "dbgsym.h"
#include "keywords.h"

#define isfloat(range) ( \
    range->symvalue.rangev.upper == 0 and range->symvalue.rangev.lower > 0 \
)

#define isrange(t, name) (t->class == RANGE and istypename(t->type, name))

#define MAXDIM  20

/*
 * Initialize FORTRAN language information.
 */

public Language fortran_init()
{
    Language fort;

    fort = language_define("fortran", ".f");
    language_setop(fort, L_PRINTDECL, fortran_printdecl);
    language_setop(fort, L_PRINTVAL, fortran_printval);
    language_setop(fort, L_TYPEMATCH, fortran_typematch);
    language_setop(fort, L_BUILDAREF, fortran_buildaref);
    language_setop(fort, L_EVALAREF, fortran_evalaref);
    language_setop(fort, L_MODINIT, fortran_modinit);
    language_setop(fort, L_HASMODULES, fortran_hasmodules);
    language_setop(fort, L_PASSADDR, fortran_passaddr);
    return(fort);
}

/*
 * Test if two types are compatible.
 *
 * Integers and reals are not compatible since they cannot always be mixed.
 */

public Boolean fortran_typematch(type1, type2)
Symbol type1, type2;
{
    Boolean b;
    Symbol t1, t2;

    t1 = rtype(type1);
    t2 = rtype(type2);
    if (t1 == nil or (t1->class != STR and t1->type == nil) or
        t2 == nil or (t2->class != STR and t2->type == nil)) {
	b = false;
    } else if (varIsSet("$nostrict")) {
	b = true;
    } else
	b = (Boolean) (typeclass(t1) == typeclass(t2));
    return b;
}

private typeclass(t)
    register Symbol t;
{
    if (t->class == STR || t == t_char->type || istypename(t->type, "char") 
	  || istypename(t->type, "$char"))
	return(1);
    else if (t == t_int->type || strncmp(symname(t->type), "integer", 7) == 0)
	return(2);
    else if (t == t_boolean->type || strncmp(symname(t->type), "logical", 7) == 0)
	return(3);
    else if (t == t_real->type || strncmp(symname(t->type), "real", 4) == 0)
	return(4);
    else if (strncmp(symname(t->type), "complex", 7) == 0)
	return(5);
    error("unknown type class for %s", symname(t));
    /*NOTREACHED*/
}

private String typename(sym)
    Symbol sym;
{
    register Symbol s;
    static char buf[20];
    int size;

    s = sym;
    do {
	s = s->type;
    } while(s->class == TYPE || s->class == ARRAY);

    if (s->class == STR) {
	size = s->symvalue.offset;
	if (size == 0)
	    size = getcharsize(s);
	if (size > 0)
	    sprintf(buf, "character*%d", size);
	else 
	    sprintf(buf, "character*(*)");
        return(buf);
    } else
	return(symname(s));
}

private Symbol mksubs(pbuf,st)
Symbol st;
char  **pbuf;
{   
   int lb, ub;
   Symbol r;

   if (st->class != ARRAY)
	return;
   else {
          mksubs(pbuf, st->type);
          assert( (r = st->chain)->class == RANGE);
	  **pbuf = '\0';
          if(r->symvalue.rangev.lowertype == R_ARG or
              r->symvalue.rangev.lowertype == R_TEMP) {
	      if( ! getbound(st,r->symvalue.rangev.lower, 
                    r->symvalue.rangev.lowertype, &lb) )
		sprintf(*pbuf,"?:");
	      else if (lb != 1)
		sprintf(*pbuf, "%d:", lb);
	  } else {
		lb = r->symvalue.rangev.lower;
		if (lb != 1)
		    sprintf(*pbuf,"%d:",lb);
		}
    	  *pbuf += strlen(*pbuf);

	  if (r->symvalue.rangev.uppertype == R_ASSUMED) {
	      sprintf(*pbuf, "*, ");
          } else if(r->symvalue.rangev.uppertype == R_ARG or
             r->symvalue.rangev.uppertype == R_TEMP) {
	      if( ! getbound(st,r->symvalue.rangev.upper, 
                    r->symvalue.rangev.uppertype, &ub) )
		sprintf(*pbuf,"?, ");
	      else 
#ifdef i386
		sprintf(*pbuf, "%d, ", ub + lb -1);
#else
		sprintf(*pbuf, "%d, ", ub + lb);
#endif
	  }
          else {
		ub = r->symvalue.rangev.upper;
		sprintf(*pbuf,"%d, ",ub);
		}
    	  *pbuf += strlen(*pbuf);

       }
}

/*
 * Print out the declaration of a FORTRAN variable.
 */

public fortran_printdecl(s)
    Symbol s;
{
    Symbol t;
    char bounds[130], *p;

    switch (s->class) {

	case CONST:
	    
	    printf("parameter %s = ", symname(s));
            printval(s);
	    break;

        case REF:
            printf(" (dummy argument) ");

	case VAR:
	    t = s->type->type;
	    switch (t->class) {

	    case ARRAY:
		p = bounds;
                mksubs(&p, t);
                *(p - 2) = '\0';   /* get rid of trailing ", " */
		printf("%s %s(%s) ", typename(s), symname(s), bounds);
		break;

	    default:
		printf("%s %s", typename(s), symname(s));

	    }
	    break;

	case FUNC:
	    if (not istypename(s->type, "void")) {
                printf(" %s function ", typename(s) );
	    }
	    else printf(" subroutine");
	    printf(" %s ", symname(s));
	    fortran_listparams(s);
	    break;

	case PROC:
	    if (s->symvalue.funcv.isentry)
		printf("entry");
	    else
	        printf("subroutine");
	    printf(" %s", symname(s));
	    fortran_listparams(s);
	    break;

	case MODULE:
	    printf("source file \"%s.c\"", symname(s));
	    break;

	case PROG:
	    printf("executable file \"%s\"", symname(s));
	    break;

	case COMMON:
	    printf("common /%s/ ", symname(s));
	    showcommonvars(s);
	    break;

	default:
	    error("class %s in fortran_printdecl", classname(s));
    }
    putchar('\n');
}

private common_offset_compare(c1, c2)
	Symbol *c1, *c2;
{
	return((*c1)->symvalue.common.offset - (*c2)->symvalue.common.offset);
}

/*
 * print the names of the variables in common c.  the variables
 * are chained in more or less random order, so sort them in order
 * of increasing offset.  the sort is done here rather than when
 * the symbols are read in to speed up the startup time.
 */
private showcommonvars(com)
	Symbol com;
{
	register Symbol c;
	Symbol		*list;
	register Symbol	*l;
	int 		nvars = 0;
	register Symbol	*head, *tail;

	/*
	 * count the variables
	 */
	c = com->symvalue.common.chain;
	while (c) {
		nvars++;
		c = c->symvalue.common.chain;
	}

	/*
	 * allocate space for an array, fill it in
	 */
	list = (Symbol *)malloc(nvars * sizeof(Symbol));
	c = com->symvalue.common.chain;
	l = list;
	while (c) {
		*l++ = c;
		c = c->symvalue.common.chain;
	}

	/*
	 * sort the array and remove duplicates.
	 */
	qsort(list, nvars, sizeof(Symbol), common_offset_compare);
	head = list;
	tail = list;
	while (head < list + nvars - 1) {
		if ((*head)->name == (*(head+1))->name && (*head)->symvalue.common.offset == (*(head+1))->symvalue.common.offset)
			head++;
		else
			*tail++ = *head++;
	}
	*tail++ = *head;
	nvars = tail - list;

	/*
	 * print out the symbols, and free the array
	 */
	for(l = list; l < list + nvars; l++) {
		printf("%s", symname((*l)));
		if (l < list + (nvars - 1))
			printf(", ");
	}
	free(list);
}


/*
 * List the parameters of a procedure or function.
 * No attempt is made to combine like types.
 */

public fortran_listparams(s)
Symbol s;
{
    register Symbol t;

    putchar('(');
    if (s->symvalue.funcv.isentry)
	s = s->chain;
    for (t = s->chain; t != nil; t = t->chain) {
	printf("%s", symname(t));
	if (t->chain != nil) {
	    printf(", ");
	}
    }
    putchar(')');
    if (s->chain != nil) {
	printf("\n");
	for (t = s->chain; t != nil; t = t->chain) {
	    printdecl(t);
	}
    } else {
	putchar('\n');
    }
}

/*
 * Print out the value on the top of the expression stack
 * in the format for the type of the given symbol.
 */

public fortran_printval(s)
Symbol s;
{
    register Symbol t;
    register int len;
    double d1, d2;

    switch (s->class) {
	case REF:
	case CONST:
	case VAR:
	case FVAR:
	case TAG:
	case TYPE:
	    fortran_printval(s->type);
	    break;

	case STR:
	    len = s->symvalue.offset;
	    if (len == 0)
		len = getcharsize(s);
	    sp -= len;
	    printf("'%.*s'", len, sp);
	    break;

	case ARRAY:
	    t = rtype(s->type);
	    if (t->class == RANGE and istypename(t->type, "char")) {
		len = size(s);
		sp -= len;
		printf("\"%.*s\"", len, sp);
	    } else {
		fortran_printarray(s);
	    }
	    break;

	case RANGE:
	     if (isfloat(s)) {
		switch (s->symvalue.rangev.lower) {
		    case sizeof(float):
			prtreal(pop(float));
			break;

		    case sizeof(double):
			if (istypename(s->type,"complex")) {
			    d2 = pop(float);
			    d1 = pop(float);
			    printf("(");
			    prtreal(d1);
			    printf(",");
			    prtreal(d2);
			    printf(")");
			} else {
			    prtreal(pop(double));
			}
			break;

		    default:
			panic("bad size \"%d\" for real",
                                  t->symvalue.rangev.lower);
			break;
		}
	    } else {
		printint(popsmall(s), s);
	    }
	    break;

	default:
	    if (ord(s->class) > ord(TYPEREF)) {
		panic("printval: bad class %d", ord(s->class));
	    }
	    error("don't know how to print a %s", fortran_classname(s));
	    /* NOTREACHED */
    }
}

/*
 * Print out an int 
 */

private printint(i, t)
Integer i;
register Symbol t;
{
    if (varIsSet("$hexints"))
	printf("0x%x", i);
    else if ((i == 0 || i == 1) && strncmp(symname(t), "logical", 7) == 0)
	printf(i == 0 ? "false" : "true");
    else
	printf("%d", i);
}


#ifdef notdef
/*
 * Print out a null-terminated string (pointer to char)
 * starting at the given address.
 */

private printstring(addr)
Address addr;
{
    register Address a;
    register Integer i;
    register Boolean endofstring;
    union {
	char ch[sizeof(Word)];
	int word;
    } u;

    putchar('"');
    a = addr;
    endofstring = false;
    while (not endofstring) {
	dread(&u, a, sizeof(u));
	i = 0;
	do {
	    if (u.ch[i] == '\0') {
		endofstring = true;
	    } else {
		printchar(u.ch[i]);
	    }
	    ++i;
	} while (i < sizeof(Word) and not endofstring);
	a += sizeof(Word);
    }
    putchar('"');
}
#endif notdef

/*
 * Return the FORTRAN name for the particular class of a symbol.
 */

public String fortran_classname(s)
Symbol s;
{
    String str;

    switch (s->class) {
	case REF:
	    str = "dummy argument";
	    break;

	case CONST:
	    str = "parameter";
	    break;

	default:
	    str = classname(s);
    }
    return str;
}

/* reverses the indices from the expr_list; should be folded into buildaref
 * and done as one recursive routine
 */
Node private rev_index(here,n)
register Node here,n;
{
 
  register Node i;

  if( here == nil  or  here == n) i=nil;
  else if( here->value.arg[1] == n) i = here;
  else i=rev_index(here->value.arg[1],n);
  return i;
}

public Node fortran_buildaref(a, slist)
Node a, slist;
{
    register Symbol as;      /* array of array of .. cursor */
    register Node en;        /* Expr list cursor */
    Symbol etype;            /* Type of subscript expr */
    Node esub, tree;         /* Subscript expression ptr and tree to be built*/

    tree=a;

    as = rtype(tree->nodetype);     /* node->sym.type->array*/
    if ( not (
               (tree->nodetype->class == VAR or tree->nodetype->class == REF)
                and as->class == ARRAY
             ) ) {
	beginerrmsg();
	prtree(stderr, a);
	fprintf(stderr, " is not an array");
	/*fprintf(stderr, " a-> %x as %x ", tree->nodetype, as ); OUT*/
	enderrmsg();
    } else {
	for (en = rev_index(slist,nil); en != nil and as->class == ARRAY;
                     en = rev_index(slist,en), as = as->type) {
	    esub = en->value.arg[0];
	    etype = rtype(esub->nodetype);
            assert(as->chain->class == RANGE);
	    if ( not compatible( t_int, etype) ) {
		beginerrmsg();
		fprintf(stderr, "subscript ");
		prtree(stderr, esub);
		fprintf(stderr, " is type %s ",symname(etype->type) );
		enderrmsg();
	    }
	    tree = build(O_INDEX, tree, esub);
	    tree->nodetype = as->type;
	}
	if (en != nil or
             (as->class == ARRAY && (not istypename(as->type,"char"))) ) {
	    beginerrmsg();
	    if (en != nil) {
		fprintf(stderr, "too many subscripts for ");
	    } else {
		fprintf(stderr, "not enough subscripts for ");
	    }
	    prtree(stderr, tree);
	    enderrmsg();
	}
    }
    return tree;
}

/*
 * Evaluate a subscript index.
 */

public fortran_evalaref(s, base, i)
Symbol s;
Address base;
long i;
{
    Symbol r, t;
    long lb, ub;

    t = rtype(s);
    if (t->class == STR) {
	push(long, base + i);
	return;
    }
    r = t->chain;
    if (
        r->symvalue.rangev.lowertype == R_ARG or
        r->symvalue.rangev.lowertype == R_TEMP
    ) {
        if (not getbound(
    	    s, r->symvalue.rangev.lower, r->symvalue.rangev.lowertype, &lb
        )) {
	    error("dynamic bounds not currently available");
        }
    } else {
        lb = r->symvalue.rangev.lower;
    }
    if (i < lb) {
        error("subscript out of range");
    }
    if (r->symvalue.rangev.uppertype != R_ASSUMED) {
	if (
	    r->symvalue.rangev.uppertype == R_ARG or
            r->symvalue.rangev.uppertype == R_TEMP
	) {
	    if (not getbound(
		s, r->symvalue.rangev.upper, r->symvalue.rangev.uppertype, &ub
	    )) {
		error("dynamic bounds not currently available");
	    } else {
#ifdef i386
		ub -= 1;
#endif
		ub += lb;
	    }
	} else {
	    ub = r->symvalue.rangev.upper;
	}
	if (i > ub) {
	    error("subscript out of range");
	}
    }
    push(long, base + (i - lb) * size(t->type));
}

private fortran_printarray(a)
Symbol a;
{
struct Bounds { int lb, val, ub; } dim[MAXDIM];

Symbol sc,st,eltype;
char buf[50];
char *subscr;
int i,ndim,elsize;
Boolean done;

st = a;

sp -= size(a);
ndim=0;

for(;;){
          sc = st->chain;

	  if (sc->symvalue.rangev.uppertype == R_ASSUMED)
	      error("can't print assumed size array");

          if(sc->symvalue.rangev.lowertype == R_ARG or
             sc->symvalue.rangev.lowertype == R_TEMP) {
	      if( ! getbound(a,sc->symvalue.rangev.lower, 
                    sc->symvalue.rangev.lowertype, &dim[ndim].lb) )
		error(" dynamic bounds not currently available");
	  }
	  else dim[ndim].lb = sc->symvalue.rangev.lower;

          if(sc->symvalue.rangev.uppertype == R_ARG or
             sc->symvalue.rangev.uppertype == R_TEMP) {
	      if( ! getbound(a,sc->symvalue.rangev.upper, 
                    sc->symvalue.rangev.uppertype, &dim[ndim].ub) )
	          error(" dynamic bounds not currently available");
	      else
#ifdef i386
			dim[ndim].ub -= 1;
#endif
		  dim[ndim].ub += dim[ndim].lb;
	  }
	  else dim[ndim].ub = sc->symvalue.rangev.upper;

          ndim ++;
          if (st->type->class == ARRAY) st=st->type;
	  else break;
     }

if(istypename(st->type,"char")) {
		eltype = st;
		ndim--;
	}
else eltype=st->type;
elsize=size(eltype);
sp += elsize;
 /*printf("ndim %d elsize %lx in fortran_printarray\n",ndim,elsize);OUT*/

ndim--;
for (i=0;i<=ndim;i++){
	  dim[i].val=dim[i].lb;
	  /*OUT printf(" %d %d %d \n",i,dim[i].lb,dim[i].ub);
	    fflush(stdout); OUT*/
}


for(;;) {
	buf[0]=',';
	subscr = buf+1;

	for (i=ndim-1;i>=0;i--)  {

		sprintf(subscr,"%d,",dim[i].val);
        	subscr += strlen(subscr);
	}
        *--subscr = '\0';

	for(i=dim[ndim].lb;i<=dim[ndim].ub;i++) {
	      	printf("[%d%s]\t",i,buf);
		printval(eltype);
	      	printf("\n");
		sp += 2*elsize;
	}
        dim[ndim].val=dim[ndim].ub;

        i=ndim-1;
        if (i<0) break;

        done=false;
        do {
		dim[i].val++;
		if(dim[i].val > dim[i].ub) { 
			dim[i].val = dim[i].lb;
			if(--i<0) done=true;
		}
		else done=true;
         }
	 while (not done);
         if (i<0) break;
     }
}

/*
 * Until we can print assumed size arrays we need to detect them
 * in printv() and not try and print -- otherwise none of the
 * other args will be printed.
 */

public boolean trap_assumed_size_array(s)
Symbol s;
{
	Symbol t;
	boolean done;

	if (s->language != lang_fortran)
		return(false);

	done = false;
	while ( (s != nil) and not done) {
		switch (s->class) {
		case REF:
		case CONST:
		case VAR:
		case FVAR:
		case TAG:
		case TYPE:
			s = s->type;
			break;
		default:
			done = true;
		}
	}

	if (s == nil or s->class != ARRAY or s->chain == nil)
		return(false);
	
	t = rtype(s);
	if (t->class == RANGE and istypename(t->type, "char"))
		return(false);
	
	for (;;) {
		t = s->chain;
		if (t->symvalue.rangev.uppertype == R_ASSUMED)
			return(true);
		if (s->type->class == ARRAY)
			s = s->type;
		else
			break;
		if (s->chain == nil)
			break;
	}
	return(false);
}

private Symbol inittype[NBUILTINS + 1];

private addType (n, s, lower, upper)
integer n;
String s;
long lower, upper;
{
    register Symbol t;

    if (n > NBUILTINS) {
	panic("initial FORTRAN type number too large for '%s'", s);
    }
    t = insert(identname(s, true));
    t->language = lang_fortran;
    t->class = TYPE;
    t->type = newSymbol(t->name, 0, RANGE, t, nil);
    t->type->symvalue.rangev.lower = lower;
    t->type->symvalue.rangev.upper = upper;
    t->type->language = lang_pascal;
    inittype[n] = t;
}

/*
 * Initialize typetable at beginning of a module.
 */

private boolean initialized = false;

public fortran_modinit (typetable)
Symbol typetable[];
{
    register integer i;

    if (not initialized) {
	initTypes();
	initialized = true;
    }
    for (i = 1; i <= NBUILTINS; i++) {
	typetable[i] = inittype[i];
    }
}


private initTypes ()
{
    addType(15, "integer*1", -128, 127);
    addType(14, "integer*2", -32768, 32767);
    addType(13, "integer*4", 0x80000000L, 0x7fffffffL);
    addType(12, "unsigned integer*1", 0, 0);
    addType(11, "unsigned integer*2", 0, 0);
    addType(10, "unsigned integer*4", 0, 0);
    addType(9, "char", -128, 127);	/* one byte char */
    addType(8, "char", -32768, 32767);	/* two byte char */
    addType(7, "real*4", 4L, 0L);
    addType(6, "real*8", 8L, 0L);
    addType(5, "logical*1", -128, 127);
    addType(4, "logical*2", -32768, 32767);
    addType(3, "logical*4", 0x80000000, 0x7fffffff);
    addType(2, "file", 0, 0);
    addType(1, "complex", 8L, 0L);
    initialized = true;
}

public boolean fortran_hasmodules ()
{
    return false;
}

/*ARGSUSED*/
public boolean fortran_passaddr (param, exprtype)
Symbol param, exprtype;
{
    return false;
}
