/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)library.c	1.7 (Berkeley) 3/28/85";
static char rcsid[] = "$Header: library.c 2.8 88/03/10 $";
#endif

/*
 * General purpose routines.
 */

#include <errno.h>
#include <signal.h>

#ifdef DBX

#include "defs.h"
#include "library.h"

#else 

#include <stdio.h>
#define public
#define private static
#define and &&
#define or ||
#define not !
#define ord(enumcon)	((int) enumcon)

typedef int integer;
typedef char *String;
typedef enum { false, true } boolean;
typedef String Filename;
typedef FILE *File;
typedef int Fileid;
#undef FILE

#endif

#define NIL(type)	((type) 0)
#define setrep(s)	(1 << ((s) - 1))

String cmdname;			/* name of command for error messages */
Filename errfilename;		/* current file associated with error */
short errlineno;		/* line number associated with error */

/*
 * Definitions for doing memory allocation.
 */

extern char *malloc();

#ifndef DBX

#define newarr(type, n)     ((type *) malloc((unsigned) (n) * sizeof(type)))
#define dispose(p)	{ free((char *) p); p = 0; }

/*
 * Macros for doing freads + fwrites.
 */

#define get(fp, var)	fread((char *) &(var), sizeof(var), 1, fp)
#define put(fp, var)	fwrite((char *) &(var), sizeof(var), 1, fp)

/*
 * String definitions.
 */

extern String strcpy(), index(), rindex();
extern int strlen();

#define strdup(s)		strcpy(malloc((unsigned) strlen(s) + 1), s)
#define streq(s1, s2)	(strcmp(s1, s2) == 0)

/*
 * The rest of this (up to "endif not dbx") is repeated because we need
 * it for both if DBX is not defined, and also inclded in the "library.h"
 * file that is built.
 */

typedef int INTFUNC();
INTFUNC *onsyserr();

typedef struct {
    INTFUNC *func;
} ERRINFO;

#define ERR_IGNORE ((INTFUNC *) 0)
#define ERR_CATCH  ((INTFUNC *) 1)
#endif not DBX

/*
 * Now repeat the definitions for extraction into "library.h".
 */

#ifdef DBX
#ifndef public
typedef int INTFUNC();
INTFUNC *onsyserr();

typedef struct {
    INTFUNC *func;
} ERRINFO;

#define ERR_IGNORE ((INTFUNC *) 0)
#define ERR_CATCH  ((INTFUNC *) 1)

#endif
#endif DBX

/*
 * Call a program.
 *
 * Four entries:
 *
 *	call, callv - call a program and wait for it, returning status
 *	back, backv - call a program and don't wait, returning process id
 *
 * The command's standard input and output are passed as FILE's.
 */


#define MAXNARGS 1000    /* unchecked upper limit on max num of arguments */
#define BADEXEC 127	/* exec fails */

#define ischild(pid)    ((pid) == 0)

/* VARARGS3 */
public int call(name, in, out, args)
String name;
String in;
String out;
String args;
{
    String *ap, *argp;
    String argv[MAXNARGS];

    argp = &argv[0];
    *argp++ = name;
    ap = &args;
    while (*ap != NIL(String)) {
	*argp++ = *ap++;
    }
    *argp = NIL(String);
    return callv(name, in, out, argv);
}

/* VARARGS3 */
public int back(name, in, out, args)
String name;
String in;
String out;
String args;
{
    String *ap, *argp;
    String argv[MAXNARGS];

    argp = &argv[0];
    *argp++ = name;
    ap = &args;
    while (*ap != NIL(String)) {
	*argp++ = *ap++;
    }
    *argp = NIL(String);
    return backv(name, in, out, argv);
}

#ifdef DBX
#ifndef public
#define callv(n, i, o, a)	runalone(n, a, i, o, false)  /* call program */
#define backv(n, i, o, a)	runalone(n, a, i, o, true) /* call in bg */
#endif
#endif DBX

#ifndef DBX
private int callv(name, in, out, argv)
String name;
String in;
String out;
String *argv;
{
    int pid, status;

    pid = backv(name, in, out, argv);
    pwait(pid, &status);
    return status;
}
#endif not DBX

#ifndef DBX
private int backv(name, in, out, argv)
String name;
String in;
String out;
String *argv;
{
    int pid;

    fflush(stdout);
    if (ischild(pid = fork())) {
	if (in)
		infrom(in);
	if (out)
		outto(out);
	(void) onsyserr(EACCES, ERR_IGNORE);
	execvp(name, argv);
	_exit(BADEXEC);
    }
    return pid;
}
#endif not DBX

/*
 * Swap file numbers so as to redirect standard input and output.
 */

private fswap(oldfd, newfd)
int oldfd;
int newfd;
{
    if (oldfd != newfd) {
	close(oldfd);
	dup(newfd);
	close(newfd);
    }
}

/*
 * Redirect input.
 * Assuming this is called from a child, we should be careful to avoid
 * (possibly) shared standard I/O buffers.
 */

public infrom (filename)
String filename;
{
    Fileid in;

    in = open(filename, 0);
    if (in == -1) {
	write(2, "can't read ", 11);
	write(2, filename, strlen(filename));
	write(2, "\n", 1);
	_exit(1);
    }
    fswap(0, in);
}

/*
 * Redirect standard output.
 * Same assumptions as for "infrom" above.
 */

public outto (filename)
String filename;
{
    Fileid out;

    out = creat(filename, 0666);
    if (out == -1) {
	write(2, "can't write ", 12);
	write(2, filename, strlen(filename));
	write(2, "\n", 1);
	_exit(1);
    }
    fswap(1, out);
}

/*
 * Invoke a shell on a command line.
 */

#define DEF_SHELL	"/bin/csh"

public shell(s)
String s;
{
    extern String getenv();
    String sh;

    if ((sh = getenv("SHELL")) == NIL(String)) {
	sh = DEF_SHELL;
    }
    if (s != NIL(String) and *s != '\0') {
	call(sh, 0, 0, "-c", s, 0);
    } else {
	call(sh, 0, 0, 0);
    }
}

/*
 * wait for process 'pid', ignoring all others.
 */
public pwait(pid, statusp)
int pid, *statusp;
{
	int pnum, status;

	do {
	    pnum = wait(&status);
	} while (pnum != pid and pnum >= 0);
	*statusp = status;
}



/*
 * System call error handler.
 *
 * The syserr routine is called when a system call is about to
 * set the c-bit to report an error.  Certain errors are caught
 * and cause the process to print a message and immediately exit.
 */

extern int sys_nerr;
extern char *sys_errlist[];
private boolean ignore_errs = false;
 
/*
 * Before calling syserr, the integer errno is set to contain the
 * number of the error.  The routine "_mycerror" is a dummy which
 * is used to force the loader to get my version of cerror rather
 * than the usual one.
 */

extern int errno;
extern _mycerror();

/*
 * Initialize error information, setting defaults for handling errors.
 */

private ERRINFO *errinfo;

private initErrInfo ()
{
    integer i;

    errinfo = newarr(ERRINFO, sys_nerr);
    for (i = 0; i < sys_nerr; i++) {
	errinfo[i].func = ERR_CATCH;
    }
    errinfo[0].func = ERR_IGNORE;
    errinfo[EPERM].func = ERR_IGNORE;
    errinfo[ENOENT].func = ERR_IGNORE;
    errinfo[ESRCH].func = ERR_IGNORE;
    errinfo[EBADF].func = ERR_IGNORE;
    errinfo[ENOTTY].func = ERR_IGNORE;
    errinfo[EOPNOTSUPP].func = ERR_IGNORE;
}

public syserr()
{
    ERRINFO *e;

    if (ignore_errs)
	return;
    if (errno < 0 or errno > sys_nerr) {
	fatal("errno %d", errno);
    } else {
	if (errinfo == NIL(ERRINFO *)) {
	    initErrInfo();
	}
	e = &(errinfo[errno]);
	if (e->func == ERR_CATCH) {
	    fatal(sys_errlist[errno]);
	} else if (e->func != ERR_IGNORE) {
	    (*e->func)();
	}
    }
}

/*
 * Catcherrs' purpose is to initialize the errinfo table, get this module
 * loaded, and make sure my cerror is loaded (only applicable when this is
 * in a library).
 */

public catcherrs()
{
    _mycerror();
    initErrInfo();
}

/*
 * Turn off the error catching mechanism completely by having all errors
 * ignored.  This is most useful between a fork and an exec.
 */

public nocatcherrs()
{
    integer i;

    for (i = 0; i < sys_nerr; i++) {
	errinfo[i].func = ERR_IGNORE;
    }
}

/*
 * Change the action on receipt of an error.
 */

public INTFUNC *onsyserr(n, f)
int n;
INTFUNC *f;
{
    INTFUNC *old_handler;

    if (errinfo == NIL(ERRINFO *)) {
	initErrInfo();
    }
    old_handler = errinfo[n].func;
    errinfo[n].func = f;
    return(old_handler);
}

public nosyserrs()
{
    ignore_errs = true;
}

public syserrs()
{
    ignore_errs = false;
}

/*
 * Print the message associated with the given signal.
 * Like a "perror" for signals.
 */

public int sys_nsig = NSIG;

public psignal(n, s)
integer n;
String s;
{
    String msg;
    integer len;
    extern String sys_siglist[];

    if (n >= 0 and n < sys_nsig) {
	msg = sys_siglist[n];
    } else {
	msg = "Unknown signal";
    }
    len = strlen(s);
    if (len > 0) {
	write(2, s, len);
	write(2, ": ", 2);
    }
    write(2, msg, strlen(msg));
    write(2, "\n", 1);
}

/*
 * Standard error handling routines.
 */

private short nerrs;
private short nwarnings;
public boolean readingsyms;

/*
 * Main driver of error message reporting.
 */

/* VARARGS2 */
private errmsg(errname, shouldquit, s, a, b, c, d, e, f, g, h, i, j, k, l, m)
String errname;
boolean shouldquit;
String s;
{
    fflush(stdout);
#ifdef DBX
    if (readingsyms)
	putc('\n', stderr);
#endif
    if (shouldquit and cmdname != NIL(String)) {
	fprintf(stderr, "%s: ", cmdname);
    }
    if (errfilename != NIL(Filename)) {
	fprintf(stderr, "%s: ", errfilename);
    }
    if (errlineno > 0) {
	fprintf(stderr, "%d: ", errlineno);
    }
    if (errname != NIL(String)) {
	fprintf(stderr, "%s: ", errname);
    }
    fprintf(stderr, s, a, b, c, d, e, f, g, h, i, j, k, l, m);
#ifdef DBX
    if (shouldquit) {
	putc('\n', stderr);
	quit(1);
    }
#else not DBX
    putc('\n', stderr);
    if (shouldquit) {
	quit(1);
    }
#endif not DBX
}

/*
 * For when printf isn't sufficient for printing the error message ...
 */

public beginerrmsg()
{
#ifdef DBX
    if (readingsyms)
	putc('\n', stderr);
#endif DBX
    fflush(stdout);
    if (errfilename != NIL(String)) {
	fprintf(stderr, "%s: ", errfilename);
    }
    if (errlineno > 0) {
	fprintf(stderr, "%d: ", errlineno);
    }
}

public enderrmsg()
{
    putc('\n', stderr);
    erecover();
}

/*
 * The messages are listed in increasing order of seriousness.
 *
 * First are warnings.
 */

/* VARARGS1 */
public warning(s, a, b, c, d, e, f, g, h, i, j, k, l, m)
String s;
{
    nwarnings++;
    errmsg("warning", false, s, a, b, c, d, e, f, g, h, i, j, k, l, m);
#ifdef DBX
    if (readingsyms)
	fprintf(stderr, " ...");
    else
	putc('\n', stderr);
#endif
}

/*
 * Errors are a little worse, they mean something is wrong,
 * but not so bad that processing can't continue.
 *
 * The routine "erecover" is called to recover from the error,
 * a default routine is provided that does nothing.
 */

/* VARARGS1 */
public error(s, a, b, c, d, e, f, g, h, i, j, k, l, m)
String s;
{
    extern erecover();

    nerrs++;
    errmsg(NIL(String), false, s, a, b, c, d, e, f, g, h, i, j, k, l, m);
#ifdef DBX
    putc('\n', stderr);
#endif
    erecover();
}

/*
 * Non-recoverable user error.
 */

/* VARARGS1 */
public fatal(s, a, b, c, d, e, f, g, h, i, j, k, l, m)
String s;
{
    errmsg("fatal error", true, s, a, b, c, d, e, f, g, h, i, j, k, l, m);
}

/*
 * Panics indicate an internal program error.
 */

/* VARARGS1 */
public panic(s, a, b, c, d, e, f, g, h, i, j, k, l, m)
String s;
{
    errmsg("internal error", true, s, a, b, c, d, e, f, g, h, i, j, k, l, m);
}

short numerrors()
{
    short r;

    r = nerrs;
    nerrs = 0;
    return r;
}

short numwarnings()
{
    short r;

    r = nwarnings;
    nwarnings = 0;
    return r;
}

/*
 * Error Handling:
 */

#include <setjmp.h>
#define ERRORSTACKSIZE	20
#ifndef public
#define seterror()	_setjmp(push_jmpbuf())
#endif
private jmp_buf jarray[ERRORSTACKSIZE];
private int	jindex = 0;

public int *push_jmpbuf()
{
	if (jindex == ERRORSTACKSIZE) {
		fprintf(stderr, "fatal error: error handler overflow");
		exit();
	}
	return(jarray[jindex++]);
}

public unseterror()
{
	if (jindex == 0) {
		fprintf(stderr, "fatal error: error handler underflow");
		exit(1);
	} else
		jindex--;
}

public erecover()
{
#ifdef DBX
	if (intside())
		panic("interrupt side error");
#endif DBX
	if (jindex == 0)
		quit(0);
#ifdef DBX
	reset_stack();	/* to prevent stack underflow problem */
#endif DBX
	jindex--;
	_longjmp(jarray[jindex], 1);
}


/*
 * Compare n-byte areas pointed to by s1 and s2
 * if n is 0 then compare up until one has a null byte.
 */

public int cmp(s1, s2, n)
register char *s1, *s2;
register int n;
{
    if (s1 == NIL(char *) || s2 == NIL(char *)) {
	panic("cmp: nil pointer");
    }
    if (n == 0) {
	while (*s1 == *s2++) {
	    if (*s1++ == '\0') {
		return(0);
	    }
	}
	return(*s1 - *(s2-1));
    } else {
	for (; n != 0; n--) {
	    if (*s1++ != *s2++) {
		return(*(s1-1) - *(s2-1));
	    }
	}
	return(0);
    }
}

/*
 * Move n bytes from src to dest.
 * If n is 0 move until a null is found.
 */

public mov(src, dest, n)
register char *src, *dest;
register int n;
{
    if (src == NIL(char *))
	panic("mov: nil source");
    if (dest == NIL(char *))
	panic("mov: nil destination");
    if (n != 0) {
	for (; n != 0; n--) {
	    *dest++ = *src++;
	}
    } else {
	while ((*dest++ = *src++) != '\0');
    }
}

/*
 * Signal name manipulation.
 */

private String signames[] = {
	0, "HUP", "INT", "QUIT", "ILL", "TRAP", "IOT", "EMT",
	"FPE", "KILL", "BUS", "SEGV", "SYS", "PIPE", "ALRM", "TERM",
	"URG", "STOP", "TSTP", "CONT", "CHLD", "TTIN", "TTOU", "IO",
	"XCPU", "XFSZ", "VTALRM", "PROF", "WINCH", 0, "USR1", "USR2",
	0, "FORK", "EXEC", "EXIT"
};

/*
 * Get the signal number associated with a given name.
 * The name is first translated to upper case if necessary.
 */

public integer siglookup (s)
String s;
{
    register char *p, *q;
    char buf[100];
    integer i;

    p = s;
    q = buf;
    while (*p != '\0') {
	if (*p >= 'a' and *p <= 'z') {
	    *q = (*p - 'a') + 'A';
	} else {
	    *q = *p;
	}
	++p;
	++q;
    }
    *q = '\0';
    p = buf;
    if (buf[0] == 'S' and buf[1] == 'I' and buf[2] == 'G') {
	p += 3;
    }
    i = 1;
    for (;;) {
	if (i > NSIG + 3) {
	    error("signal \"%s\" unknown", s);
	    i = 0;
	    break;
	}
	if (signames[i] != NIL(String) and streq(signames[i], p)) {
	    break;
	}
	++i;
    }
    return i;
}

public printsigs (set)
integer set;
{
    integer s;
    char separator[2];

    separator[0] = '\0';
    for (s = 1; s < NSIG; s++) {
	if (set & setrep(s)) {
	    if (signames[s] != NIL(String)) {
		printf("%s%s", separator, signames[s]);
		separator[0] = ' ';
		separator[1] = '\0';
	    }
	}
    }
    if (separator[0] == ' ') {
	putchar('\n');
    }
}
