/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)main.c	1.11 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: main.c 2.18 87/10/15 $";
#endif

/*
 * Debugger main routine.
 */

#include "defs.h"
#include <signal.h>
#include <sys/param.h>
#include <machine/fpu.h>
#include <errno.h>
#include <sys/file.h>
#include "main.h"
#include "eval.h"
#include "debug.h"
#include "symbols.h"
#include "scanner.h"
#include "keywords.h"
#include "runtime.h"
#include "source.h"
#include "object.h"
#include "mappings.h"
#include "coredump.h"
#include "library.h"
#include "tty.h"

public boolean runfirst;		/* run program immediately */
public boolean interactive;		/* standard input IS a terminal */
public boolean lexdebug;		/* trace scanner return values */
public boolean tracebpts;		/* trace create/delete breakpoints */
public boolean traceexec;		/* trace execution */
public boolean tracesyms;		/* print symbols are they are read */
public boolean traceblocks;		/* trace blocks while reading symbols */
public boolean vaddrs;			/* map addresses through page tables */
public boolean testing;			/* be quiet about version stuff */
public boolean uppercase_flag;		/* preserve upper case in names from dbg file */
public boolean is_pdbx; 		/* are we parallel debugger */
public boolean is_dbx;			/* are we "old style" dbx debugger */
public boolean async_flag = false;	/* value of -a flag */
public boolean synchronous = true;	/* wait for processes */
public boolean Oflag = false;		/* was there a "-O" flag specified? */

private boolean foundcore;		/* found core file name on cmd line? */
private Boolean foundobj;		/* found object name on cmd line? */
private String defobjname = "a.out";	/* object name to default to */
private Boolean initdone = false;	/* true if initialization done */
private char outbuf[BUFSIZ];		/* standard output buffer */
private char namebuf[512];		/* possible name of object file */
private int firstarg;			/* first program argument (for -r) */

private catchintr();
private cmdline_erecover();

/*
 * Main program.
 */

main(argc, argv)
int argc;
String argv[];
{
    register integer i;
    extern String date;
    extern integer versionNumber;

    winit();
    if (seterror())
	quit(1);
    cmdname = argv[0];
    if (argc > 1 && strcmp(argv[1], "-q") == 0) {
	testing = true;
	argv++; argc--;
    }
    catcherrs();
    (void) onsyserr(EINTR, ERR_IGNORE);
    setbuf(stdout, outbuf);
    if (not testing)
	printf("dbx version 5.%d of %s.\nType 'help' for help.\n",
	    versionNumber, date);
    fflush(stdout);
    set_prog_name(cmdname);
    process_init();
    language_init();
    machine_init();
    scanargs(argc, argv);
    unseterror();
    while (seterror()) {
	cmdline_erecover();
        runfirst = false;
    }
    if (runfirst) {
	arginit();
	for (i = firstarg; i < argc; i++) {
	    newarg(argv[i]);
	}
	run();
	/* NOTREACHED */
    } else {
	init();
    }
    unseterror();
    while (seterror()) {
	cmdline_erecover();
    }
    signal(SIGINT, catchintr);
    signal(SIGFPE, catchfpe);
    yyparse();
    putchar('\n');
    quit(0);
}

/*
 * Initialize the world, including setting initial input file
 * if the file exists.
 */

public init()
{
    File f;
    String home;
    char buf[100];
    extern String getenv();

    enterkeywords();
    scanner_init();
    setuptty();
    if (not runfirst)
	makeprocs();
    curobj = find_object(defobjname);
    f = fopen(initfile, "r");
    if (f != nil) {
	fclose(f);
	setinput(initfile);
    } else {
	home = getenv("HOME");
	if (home != nil) {
	    sprintf(buf, "%s/%s", home, initfile);
	    f = fopen(buf, "r");
	    if (f != nil) {
		fclose(f);
		setinput(strdup(buf));
	    }
	}
    }
    initdone = true;
}

/*
 * After a non-fatal error we skip the rest of the current input line, and
 * jump back to command parsing.
 */

private cmdline_erecover()
{
    if (initdone) {
	process_reset();
	(void) sigsetmask(defsigmask);
	gobble();
    }
}

/*
 * This routine is called when an interrupt occurs.
 */

private catchintr()
{
    if (isredirected()) {
	fflush(stdout);
	unsetout();
    }
    putchar('\n');
    erecover();
}

/*
 * Scan the argument list.
 */

private scanargs(argc, argv)
int argc;
String argv[];
{
    register int i, j;

    runfirst = false;
    interactive = false;
    lexdebug = false;
    tracebpts = false;
    traceexec = false;
    tracesyms = false;
    traceblocks = false;
    vaddrs = false;
    foundobj = false;
    foundcore = false;
    has_mplib = false;
    genericsrcpath = list_alloc();
    i = 1;
    while (i < argc and not runfirst) {
	if (argv[i][0] == '-') {
	    if (streq(argv[i], "-I")) {
		++i;
		if (i >= argc) {
		    fatal("missing directory for -I");
		}
		list_append(list_item(argv[i]), nil, genericsrcpath);
	    } else if (streq(argv[i], "-O") or streq(argv[i], "-Q")) {
		++i;
		if (i >= argc) {
		    fatal("missing object file name for %s", argv[i-1]);
		}
		multi_objs = true;
		Oflag = true;
		readobj(argv[i], streq(argv[i-1], "-Q"));
		foundobj = true;
	        defobjname = argv[i];
	    } else if (streq(argv[i], "-C")) {
		++i;
		if (i >= argc) {
		    fatal("missing core file name for -C");
		}
		foundcore = true;
		(void) readcore(argv[i]);
	    } else if (streq(argv[i], "-c")) {
		++i;
		if (i >= argc) {
		    fatal("missing command file name for -c");
		}
		initfile = argv[i];
	    } else if (streq(argv[i], "-d")) {
		++i;
		if (i >= argc) {
		    fatal("missing debug option for -d");
		}
		debug(build(O_DEBUG, argv[i]));
	    } else if (testing and streq(argv[i], "-is_dbx")) {
		is_dbx = true;
		is_pdbx = false;
	    } else if (testing and streq(argv[i], "-is_pdbx")) {
		is_dbx = false;
		is_pdbx = true;
	    } else {
		for (j = 1; argv[i][j] != '\0'; j++) {
		    setoption(argv[i][j]);
		}
	    }
	} else if (not foundobj) {
	    readobj(argv[i], false);
	    defobjname = argv[i];
	    foundobj = true;
	} else if (not foundcore) {
	    foundcore = true;
	    (void) readcore(argv[i]);
	} else {
	    fatal("extraneous argument %s", argv[i]);
	}
	++i;
    }
    firstarg = i;
    if (not foundobj) {
	if (isatty(0)) {
	    printf("enter object file name (default is `%s'): ", defobjname);
	    fflush(stdout);
	    gets(namebuf);
	    if (namebuf[0] != '\0') {
	        defobjname = namebuf;
	    }
	}
	readobj(defobjname, false);
    } 
    if (not foundcore and access("core", F_OK) == 0) {
	    foundcore = true;
	    (void) readcore("core");
    }
}

/*
 * Take appropriate action for recognized command argument.
 */

private setoption(c)
char c;
{
    switch (c) {
	case 'a':
	    async_flag = true;
	    synchronous = false;
	    break;

	case 'i':
	    interactive = true;
	    break;

	case 'r':   /* run program before accepting commands */
	    warning("-r flag is not implemented");
	    break;

	case 'u':
	    uppercase_flag = true;
	    break;

	default:
	    fatal("unknown option '%c'", c);
    }
}

/*
 * Exit gracefully.
 */

public quit(r)
Integer r;
{
    (void) onsyserr(EINVAL, ERR_IGNORE);
    killkids();
    untty();
    wquit();
    exit(r);
}

/*
 * The debugger acts a bit diffenently depending on the name used
 * to invoke it.  Expected names are "pdbx" and "dbx".  Flags set
 * are "is_pdbx" and "is_dbx".  The difference in operation involves
 * signals that are caught/ignored and signal catching for descendents
 * of the process being debugged.
 *
 * We have to check the last path of the name used to invoke the debugger
 * since it could be invoked as "/usr/ucb/dbx".
 */

private set_prog_name(cmdname)
	String cmdname;
{
	String p;

	is_pdbx = false;
	is_dbx = false;

	p = rindex(cmdname, '/');
	if (p == nil)
		p = cmdname;
	else {
		p++;
		if (! *p)
			return;
	}

	if (streq(p, "pdbx"))
		is_pdbx = true;
	else if (streq(p, "dbx"))
		is_dbx = true;
}
