/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)mappings.c	1.6 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: mappings.c 2.8 87/01/14 $";
#endif

/*
 * Source-to-object and vice versa mappings.
 */

#include "defs.h"
#include "symbols.h"
#include "object.h"
#include "mappings.h"
#include "source.h"
#include "machine.h"

#ifndef public
#include "machine.h"
#include "source.h"

/*
 * structures to support per-object line and file tables.  filetab and
 * linetab are maintained in the Object structure.
 * filep, linep, endfiletab, and endlinetab are only used while reading
 * the object file, so they don't have to be saved in the Object structure.
 */

typedef struct {
    Address addr;
    String filename;
    Lineno lineindex;		/* index to first linetab entry */
} Filetab;

typedef struct {
    Lineno line;
    Address addr;
} Linetab;

typedef struct aof AddrOfFunc;

#define filetab		(curobj->o_filetab)
#define linetab		(curobj->o_linetab)
Filetab *filep;		/* where to put next file */
Filetab *endfiletab;	/* checked for overflowing file table */
Linetab *linep;		/* where to put next line */
Linetab *endlinetab;	/* checked for overflowing line table */

#define NOADDR ((Address) -1)	/* no address for line or procedure */

#endif

public realloc_linetab()
{
	int index;

	index = linep - linetab;
	nlhdr.nlines *= 4;
	linetab = (Linetab *)realloc(linetab, nlhdr.nlines * sizeof(Linetab));
	linep = &linetab[index];
	endlinetab = &linetab[nlhdr.nlines];
}

public realloc_filetab()
{
	int index;

	index = filep - filetab;
	nlhdr.nfiles *= 4;
	filetab = (Filetab *)realloc(filetab, nlhdr.nfiles * sizeof(Filetab));
	filep = &filetab[index];
	endfiletab = &filetab[nlhdr.nfiles];
}

/*
 * Get the source file name associated with a given address.
 */

public String srcfilename(addr)
Address addr;
{
    register Address i, j, k;
    Address a;
    Filetab *ftp;
    String s;

    s = nil;
    if (nlhdr.nfiles != 0 and addr >= filetab[0].addr) {
	i = 0;
	j = nlhdr.nfiles - 1;
	while (i < j) {
	    k = (i + j) / 2;
	    ftp = &filetab[k];
	    a = ftp->addr;
	    if (a == addr) {
		s = ftp->filename;
		break;
	    } else if (addr > a) {
		i = k + 1;
	    } else {
		j = k - 1;
	    }
	}
	if (s == nil) {
	    if (addr >= filetab[i].addr) {
		s = filetab[i].filename;
	    } else {
		s = filetab[i-1].filename;
	    }
	}
    }
    return s;
}

/*
 * Find the line associated with the given address.
 * If the second parameter is true, then the address must match
 * a source line exactly.  Otherwise the nearest source line
 * below the given address is returned.
 *
 * Return the index of the line table entry or -1 if none suitable.
 */

private integer findline (addr, exact)
Address addr;
Boolean exact;
{
    register Address i, j, k;
    register Lineno r;
    register Address a;

    if (nlhdr.nlines == 0 or addr < linetab[0].addr) {
	r = -1;
    } else {
	i = 0;
	j = nlhdr.nlines - 1;
	if (addr == linetab[i].addr) {
	    r = i;
	} else if (addr == linetab[j].addr) {
	    r = j;
	} else if (addr > linetab[j].addr) {
	    r = exact ? -1 : j;
	} else {
	    do {
		k = (i + j) div 2;
		a = linetab[k].addr;
	    if (a == addr) break;
		if (addr > a) {
		    i = k + 1;
		} else {
		    j = k - 1;
		}
	    } while (i <= j);
	    if (a == addr) {
		r = k;
	    } else if (exact) {
		r = -1;
	    } else if (addr > linetab[i].addr) {
		r = i;
	    } else {
		r = i - 1;
	    }
	}
    }
    return r;
}

/*
 * Lookup the source line number nearest (from below) to an address.
 *
 * It is possible (unfortunately) that the compiler will generate
 * code before line number for a procedure.  Therefore we check
 * to see that the found line is in the same procedure as the given address.
 * If it isn't, then we walk forward until the first suitable line is found.
 */

public Lineno srcline(addr)
Address addr;
{
    integer i;
    Lineno r;
    Symbol f1, f2;

    i = findline(addr, false);
    if (i == -1) {
	r = 0;
    } else {
	r = linetab[i].line;
	if (linetab[i].addr != addr) {
	    f1 = whatblock(addr);
	    if (nosource(f1)) {
		r = 0;
	    } else {
		f2 = whatblock(linetab[i].addr + 1);
		if (f1 != f2) {
		    do {
			++i;
		    } while (linetab[i].addr < addr and i < nlhdr.nlines);
		    r = linetab[i].line;
		}
	    }
	}
    }
    return r;
}

/*
 * Look for a line exactly corresponding to the given address.
 */

public Lineno linelookup(addr)
Address addr;
{
    integer i;
    Lineno r;

    i = findline(addr, true);
    if (i == -1) {
	r = 0;
    } else {
	r = linetab[i].line;
    }
    return r;
}

/*
 * return true if file 'filename' is listed in the file table for 'obj'.
 */

public boolean havefile(filename, obj)
    register String filename;
    Object *obj;
{
    register Filetab *ftp;
    Filetab *tab;
    register Filetab *end;

    if (obj == nil or obj->o_nlhdr.nlines == 0) {
	return(false);
    }

    tab = obj->o_filetab;
    end = &tab[obj->o_nlhdr.nfiles];
    for (ftp = tab; ftp < end; ftp++) {
	if (filename == nil or ftp->filename == nil) {
	    printf("\nWarning: havefile() called with nil filename.\n");
	    return(false);
	}
	if (streq(ftp->filename, filename)) {
	    return(true);
	}
    }

    return(false);
}

/*
 * Lookup the object address of a given line from the named file.
 *
 * Potentially all files in the file table need to be checked
 * until the line is found since a particular file name may appear
 * more than once in the file table (caused by includes).
 */

public Address objaddr(line, name)
Lineno line;
String name;
{
    register Filetab *ftp;
    register Lineno i, j;
    Boolean foundfile;

    if (nlhdr.nlines == 0) {
	return NOADDR;
    }
    if (name == nil) {
	if (not canReadSource())
	    return NOADDR;
	name = cursource;
    }
    foundfile = false;
    for (ftp = &filetab[0]; ftp < &filetab[nlhdr.nfiles]; ftp++) {
	if (streq(ftp->filename, name)) {
	    foundfile = true;
	    i = ftp->lineindex;
	    if (ftp == &filetab[nlhdr.nfiles-1]) {
		j = nlhdr.nlines;
	    } else {
		j = (ftp + 1)->lineindex;
	    }
	    while (i < j) {
		if (linetab[i].line == line) {
		    return linetab[i].addr;
		}
		i++;
	    }
	}
    }
    if (not foundfile) {
	beginerrmsg();
	fprintf(stderr, "source file \"%s\" not compiled with -g", name);
	if (multi_objs)
	    fprintf(stderr, " in object file %s", curobj->o_name);
	enderrmsg();
    }
    return NOADDR;
}

/*
 * Table for going from object addresses to the functions in which they belong.
 */

#define NFUNCS 500	/* initial size of function table */

struct aof {
    Symbol func;
    Address addr;
};

#define functab		(curobj->o_functab)
#define nfuncs		(curobj->o_nfuncs)
private int functablesize = 0;

/*
 * Insert a new function into the table.
 *
 * Fortran programs with BLOCK DATA will generate a symbol " -----",
 * which will have an address that is the same as that for the main
 * program.  Enering this in the table can cause problems as it may
 * overwrite the main program name when the table is ordered later.
 * This is the reason that the symbol is ignored.
 */

public newfunc(f, addr)
Symbol f;
Address addr;
{
    register AddrOfFunc *af;
    AddrOfFunc *newfunctab;

    if (f->name == identname(" -----", false))	/* block data */
	return;

    if (nfuncs >= functablesize) {
	if (functablesize == 0) {
	    functab = newarr(AddrOfFunc, NFUNCS);
	    functablesize = NFUNCS;
	} else {
	    functablesize *= 2;
	    newfunctab = newarr(AddrOfFunc, functablesize);
	    bcopy(functab, newfunctab, nfuncs * sizeof(AddrOfFunc));
	    dispose(functab);
	    functab = newfunctab;
	}
    }
    af = &functab[nfuncs];
    af->func = f;
    af->addr = addr;
    ++nfuncs;
}

mappings_init()
{
	nfuncs = 0;
	functablesize = 0;
}

/*
 * Return the function that begins at the given address.
 */

public Symbol whatblock(addr)
Address addr;
{
    register int i, j, k;
    Address a;

    if (functab == nil)
	return(program);
    i = 0;
    j = nfuncs - 1;
    if (addr < functab[i].addr) {
	return program;
    } else if (addr == functab[i].addr) {
	return functab[i].func;
    } else if (addr >= functab[j].addr) {
	return functab[j].func;
    }
    while (i <= j) {
	k = (i + j) / 2;
	a = functab[k].addr;
	if (a == addr) {
	    return functab[k].func;
	} else if (addr > a) {
	    i = k+1;
	} else {
	    j = k-1;
	}
    }
    if (addr > functab[i].addr) {
	return functab[i].func;
    } else {
	return functab[i-1].func;
    }
    /* NOTREACHED */
}

/*
 * Order the functab.
 */

private int cmpfunc(f1, f2)
AddrOfFunc *f1, *f2;
{
    register Address a1, a2;

    a1 = (*f1).addr;
    a2 = (*f2).addr;
    return ( (a1 < a2) ? -1 : ( (a1 == a2) ? 0 : 1 ) );
}

public ordfunctab()
{
    qsort(functab, nfuncs, sizeof(AddrOfFunc), cmpfunc);
}

/*
 * remove functions with duplicate addresses, preferring those with source.
 * for "inline" functions (C blocks), we want to prefer the last instance.
 */
remdupfuncs()
{
    register AddrOfFunc *head, *tail, *end, *best;

    tail = functab;
    head = functab;
    best = functab;
    end = &functab[nfuncs];
    do {
	if (not nosource(head->func) and not isinline(head->func))
	    best = head;
	if (head->addr != (head+1)->addr) {
	    *tail++ = *best;
	    best = ++head;
	} else {
	    nfuncs--;
	    head++;
	}
    } while (head < end);
}

/*
 * Clear out the functab, used when re-reading the object information.
 */

public clrfunctab()
{
    nfuncs = 0;
}

public dumpfunctab()
{
    int i;

    for (i = 0; i < nfuncs; i++) { 
	printf("%#x: ", functab[i].addr);
	psym(functab[i].func);
    }
}


public dumplinetab()
{
	register Lineno l;
	register int f;

	f = 0;
	for (l = 0; l < nlhdr.nlines; l++) {
		printf("%8d %#10x ", linetab[l].line, linetab[l].addr);
		while(filetab[f].lineindex < l && f < nlhdr.nfiles)
			f++;
		if (filetab[f].lineindex == l) {
			printf("\"%s\" ", 
			  (filetab[f].filename == NULL) ? 
			  	"(null)" : filetab[f].filename );
		}
		printf("\n");
	}
	fflush(stdout);
}

/*
 * return the address of the line after address 'a'.
 */
public Address nextline(a)
    Address a;
{
    int i;

    i = findline(a, true);
    if (i == -1)
	return(NOADDR);
    i++;
    if (i > nlhdr.nlines)
	return(a + 10);
    else
	return(linetab[i].addr);
}
