/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef lint
static char rcsid[] = "$Header: multi-p.c 1.28 88/03/17 $";
#endif

/*
 * multi-p.c -- multiprocess control
 */

#include <ctype.h>
#include <errno.h>
#include <signal.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/proc.h>
#include "defs.h"
#include "main.h"
#include "machine.h"
#include "events.h"
#include "object.h"
#include "multi-p.h"
#include "runtime.h"
#include "tree.h"
#include "tty.h"
#include "library.h"

#define	MAXARGS	20
#define mask(b)	(1 << ((b) - 1))


/* debugging flags */
public boolean	debug_ptrace = false;	/* trace calls to ptrace */
public boolean	debug_process = false;	/* trace process state changes */
public boolean	debug_procstack = false;/* trace process stack management */
public boolean	debug_stopall = false;	/* trace "stop all" command */
public boolean	debug_globals = false;	/* trace fussing with globals */

/* random */
extern	int	errno;
boolean		stopallmode = false;

extern char *sprintf();
extern char *strcpy();
extern char *strcat();

#ifndef public

# ifndef PTS_FORK
# ifndef MAXPATHLEN
#	include <sys/param.h>
# endif
#	include <machine/fpu.h>
#	include <machine/ptrace.h>
# endif

typedef struct process Process;
#include "object.h"

#define PROCIDFMT	"%%%-2d "	/* how to print a process id */

enum {
	USERSIDE,			/* handling user input, etc */
	INTSIDE				/* handling a child event interrupt */
} md_state;

/*
 * structure to describe a process
 */

typedef enum {
	UNKNOWN,
	IMAGE,		/* read from program a.out */
	COREDUMP,	/* read from coredump */
	STOPPED,	/* md has decided process should be stopped and it is */
	RUNNING,	/* process is really running */
	PENDING,	/* process stopped while md figures out what to do */
	EXITED,		/* process is gone */
	DELETED,	/* process removed from process list */
	PROTO		/* program prototype -- used for searches only */
} Pstate;

typedef enum {
	INIT,		/* initialize state machine */
	STEP,		/* do a step */
	STEPOVER,	/* do step over procedure call */
	CONT		/* continue execution */
} Cmdstate;

/*
 * why a process stopped, mostly for the benefit of printstatus().
 */
typedef enum {
	BY_EVENT,	/* look in p_reason for why */
	BY_BPT,		/* stopped by a breakpoint */
	BY_STEP,	/* stopped by single step */
	BY_STOP,	/* by "stop" command */
	BY_ERROR,	/* error in doqueue */
	BY_CALLRTN,	/* returned from 'call' command */
	BY_RTERR,	/* hit a runtime error */
	BY_CMD,		/* look in p_cmd for why */
	BY_STOPALL,	/* by "stop all" command */
	BY_RETURN	/* by "return" command */
} Stopinfo;

#define PF_CATCH_FORK	1	/* stop for fork */
#define PF_CATCH_EXEC 	2	/* stop for exec */
#define PF_CATCH_EXIT 	4	/* stop for exec */
#define	PF_INHERIT_BPTS	8	/* inherit breakpoints from parent */
#define PF_SETFLAGS	16	/* keep p_flags == default_flags */
#define PF_DELETE_EXIT	32	/* delete process when it exits */
#define PF_NO_BPTS	64	/* don't insert bpts into process */

#define IGNORESIGS	-1	/* if in p_signal, all sigs ignored */

/*
 * A cache of the instruction segment is kept to reduce the number
 * of system calls.  Might be better just to read the entire
 * code space into memory.
 */

#define CSIZE	1003       /* size of instruction cache */
# ifdef i386
#define PSSIZE	2048	   /* size of per-process stack */
# else
#define PSSIZE	1024	   /* size of per-process stack */
# endif

typedef struct {
    Word addr;
    Word val;
} CacheWord;

struct process {

	/* current state */

	int			p_id;		/* small integer to name proc */
	Pstate			p_state;	/* what process is up to now */
	struct pt_stop		p_s;		/* why process is stopped */
	Stopinfo		p_stopinfo;	/* details about stopped state*/
	int			p_stopeventid;	/* of event that caused stop */
	int			p_signal;	/* signal that stopped process*/
	struct pt_regset	p_regs;		/* registers after stop */
	int 			p_parent;	/* id of parent process */
	Object			*p_obj;		/* info about a.out for proc */
	struct Job		*p_job;		/* tty/pgrp info */

	/* process attributes */

	int			p_sigmask;	/* signals to be ignored */
	unsigned int		p_flags;	/* what to do for events */
	boolean			p_tracing;	/* cont means repeatedly step*/
	boolean			p_tracei;	/* tracei rather than trace */
	boolean			p_stepi;	/* stepi/nexti, not step */ 
	boolean			p_quiet;	/* don't say when stopped */

	/* how to change state */

	Pstate			(*p_cmd)();	/* command causing proc to run*/
	Cmdstate		p_cmdstate;	/* state of command */
	int			p_contmode;	/* how to continue (step/cont)*/
	Address			p_oldpc;	/* pc before last continue */
	Address			p_tempbp;	/* temporary breakpoint */
	Pstate			p_newstate;	/* what to do next */
	boolean			p_pending;	/* has commands to eval()? */
	boolean			p_trap_pending;	/* trap pending from stopall?*/
	boolean			p_stepover;	/* false==>step true==>next*/

	/* administration */

	String			p_corename;	/* core file name if COREDUMP */
	pointer			p_savearea;	/* global save area */
	CacheWord		p_tcache[CSIZE];/* text segment cache */
	struct process		*p_next;	/* for process list */
	boolean			p_regschanged;	/* need to write regs back? */
	char			p_stack[PSSIZE];/* per-process stack */
	int			p_sp;		/* per-process stack ptr */


};

#define p_pid			p_s.ps_pid
#define	p_reason		p_s.ps_reason

#endif public


/*
 * data to maintain processes
 */

public	Process *wp;			/* current working process */
private Process	*processlist = 0;	/* list of processes */
private Process	*curprocess;		/* set by 'process' command */
private int	nextprocid = 1;		/* next procid to use */
private Boolean	retry;			/* got nowhere; continue without
					   consulting state machine again */
private Pstate	pstep();		/* state machine for step (and next) */
private Pstate	pcont();		/* state machine for cont */
private Pstate	pnothing();		/* dummy state machine */
private Process *process_alloc();
private unsigned int default_flags;
private int	default_sigmask;	/* for processes dbx controls */

#define MAXNCMDARGS 100		        /* maximum number of args to RUN */
private int argc;
private String argv[MAXNCMDARGS];
private String infile, outfile;

/*
 * routines to manipulate processes
 */

/*
 * return process structure for process with pid 'pid'.
 * linear search for now, probably want a hash table.
 */
private Process *find_process(pid)
	register int		pid;
{
	register Process	*p;

	if (pid == 0)
		return(wp);
	p = processlist;
	while(p && p->p_pid != pid)
		p = p->p_next;
	return(p);
}

/*
 * return process structure for process with id 'procid'.
 * linear search for now, probably want a hash table.
 * only public so that other modules can check whether a 
 * given procid is valid.
 */
public Process *find_procid(procid)
	register int		procid;
{
	register Process	*p;

	if (procid == 0)
		return(wp);
	p = processlist;
	while(p && p->p_id != procid)
		p = p->p_next;
	return(p);
}

/*
 * return process with id 'procid'.  if no such, error.
 */
private Process *findpid(procid)
	int procid;
{
	Process *p;

	p = find_procid(procid);
	if (p == 0)
		error("No process with id %d", procid);
	return(p);
}

/*
 * push process with id 'procid'.  if no such, error.
 */
public pushpid(procid)
	int	procid;
{
	pushproc(findpid(procid));
}

public int firstprocid()
{
	assert(processlist != 0);
	return(processlist->p_id);
}


/*
 * User side routines that call create_process *must* block SIGCHLD!
 * Since the interrupt side implicitly blocks SIGCHLD, this ensures
 * that there are never concurrent calls to this routine.
 */
private Process *create_process(pid)
	int	pid;
{
	register Process	*p;

	p = process_alloc();

	p->p_id = nextprocid++;
	p->p_state = PENDING;
	p->p_pid = pid;
	/* p_reason */
	p->p_stopeventid = 0;
	p->p_signal = 0;
	/* p_regs */
	p->p_parent = 0;
	p->p_obj = 0;
	p->p_job = 0;

	p->p_sigmask = 0;
	p->p_flags = default_flags;
	p->p_tracing = false;
	p->p_stepi = false;
	p->p_tracei = false;
	p->p_quiet = false;

	p->p_cmd = 0;
	p->p_cmdstate = INIT;
	p->p_contmode = PT_CONTSIG;
	p->p_oldpc = 0;
	p->p_tempbp = 0;
	p->p_newstate = STOPPED;
	p->p_pending = false;
	p->p_trap_pending = false;
	p->p_stepover = false;


	p->p_corename = 0;
	bzero(p->p_tcache, sizeof(p->p_tcache));
	p->p_next = processlist;
	p->p_regschanged = false;
	/* no need to zero p_stack since we're resetting p_sp */
	p->p_sp = 0;

	processlist = p;
	if (debug_process) {
		printf("create_process(pid= %d) entered\n", pid);
		printf("    p->p_id= %d, p->p_parent= %d\n",
		    p->p_id, p->p_parent);
		fflush(stdout);
	}
	return(p);
}

/*
 * this routine is called to copy the process structure from parent to
 * child after a fork.  account for each field.
 *
 * If this is dbx then we want to ignore EXIT and SEGV in all descendents
 * of the process being debugged.  Also clear PF_SETFLAGS so we don't
 * reset the default flags.
 */
private copy_process(from, to)
	Process *from;
	Process *to;
{

	/* p_id:	unique to each process */
	/* p_state: 	set already by babysitter() */
	/* p_pid:	set already by babysitter() */
	/* p_reason:	set already by babysitter() */
	/* p_stopinfo:	in the process of being set by babysitter() */
	to->p_stopeventid = 0; /* will be set or not as appropriate */
	/* p_signal:	will by set by babysitter() */
	/* p_regs:	not valid yet, read in when needed */
	to->p_parent = from->p_id;
	to->p_obj = from->p_obj;
	if (to == wp)
		curobj = from->p_obj;
	to->p_job = from->p_job;	/* job stays the same */
	addjob(to->p_job);

	to->p_sigmask = from->p_sigmask;
	to->p_flags = from->p_flags;
	to->p_tracing = from->p_tracing;
	to->p_stepi = from->p_stepi;
	to->p_tracei = from->p_tracei;
	to->p_quiet = from->p_quiet;

	to->p_cmd = from->p_cmd;
	to->p_cmdstate = from->p_cmdstate;
	to->p_contmode = from->p_contmode;
	to->p_oldpc = from->p_oldpc;
	to->p_tempbp = from->p_tempbp;
	/* to->p_newstate: in the process of being set by babysitter() */
	to->p_pending = from->p_pending; /* should always be false! */
	/* p_trap_pending: not inherited */
	to->p_stepover = from->p_stepover;

	to->p_corename = 0;
	/* p_tcache is clear, don't bother to copy */
	/* p_next:	set already by create_process() */
	/* p_regschanged: already cleared */
	bcopy(from->p_stack, to->p_stack, from->p_sp);
	to->p_sp = from->p_sp;

	copy_globals(from->p_savearea, to->p_savearea);
	copy_bptsave(from->p_savearea, to->p_savearea);
	copy_bps(from->p_id);
	copy_trcmds(from->p_id);

	if (is_dbx) {
	    to->p_sigmask = IGNORESIGS;
	    to->p_quiet = true;
	    to->p_flags &= ~(PF_CATCH_FORK|PF_CATCH_EXEC|PF_CATCH_EXIT|PF_SETFLAGS);
	    to->p_flags |= (PF_DELETE_EXIT|PF_NO_BPTS);
	    xptrace(XPT_SETSIGMASK, to->p_pid, 0, to->p_sigmask);
	}

	if (debug_process) {
	    printf("copy_process: from: p_id= %d, p_pid= %d, p_parent=%d\n",
		from->p_id, from->p_pid, from->p_parent);
	    printf("copy_process:   to: p_id= %d, p_pid= %d\n",
		to->p_id, to->p_pid, to->p_parent);
	    fflush(stdout);
	}
}

/*
 * reset process structure due to exec.  account for each field.
 */
private reset_process()
{
	/* p_id:	unique to each process */
	/* p_state: 	set already by babysitter() */
	/* p_pid:	set already by babysitter() */
	/* p_reason:	set already by babysitter() */
	/* p_stopinfo:	in the process of being set by babysitter() */
	wp->p_stopeventid = 0; /* will be set or not as appropriate */
	/* p_signal:	will by set by babysitter() */
	/* p_regs:	not valid yet, read in when needed */
	/* p_parent:	stays the same */ 
	wp->p_obj = create_object(processname());
	/* p_job:	stays the same */

	/* p_sigmask:	stays the same */
	/* p_flags:	stays the same */
	/* p_tracing:	stays the same */
	/* p_stepi:	stays the same */
	/* p_tracei:	stays the same */
	/* p_quiet:	stays the same */

	/* p_cmd = from->p_cmd;
	/* p_cmdstate = from->p_cmdstate;
	/* p_contmode = from->p_contmode;
	wp->p_oldpc = 0;
	wp->p_tempbp = 0;
	/* p_newstate: in the process of being set by babysitter() */
	/* p_pending: should always be false! */
	/* p_trap_pending: not inherited */
	/* p_stepover */

	wp->p_corename = 0;
	bzero(wp->p_tcache, sizeof(wp->p_tcache));
	/* p_next:	stays the same */
	/* p_regschanged: already cleared */
	bzero(wp->p_stack, sizeof(wp->p_stack));
	wp->p_sp = 0;
	clear_globals(wp->p_savearea);

	if (debug_process)
	    printf("reset_process: wp->p_id= %d, wp->p_pid= %d, wp->p_parent= %d\n",
		wp->p_id, wp->p_pid, wp->p_parent);
}


/*
 * Get rid of process 'proctodelete'.
 */
private delete_process(proctodelete)
	Process *proctodelete;
{
	register Process *p, **ptr;

	if (debug_process)
	    printf("delete_process:  p_id= %d, p_pid= %d, p_parent= %d\n",
		proctodelete->p_id, proctodelete->p_pid,
		proctodelete->p_parent);

	pushproc(proctodelete);
	unsetallbps();
	popproc();
	jobexit(proctodelete->p_job);
	unbusy(proctodelete->p_job, proctodelete->p_id);
	ptr = &processlist;
	for (p = processlist; p; p = p->p_next) {
		if (p == proctodelete) {
			p->p_state = DELETED;
			*ptr = p->p_next;
		}
		ptr = &p->p_next;
	}
}


/*
 * the following six routines deal with the current process stack.
 */
#define PROCSTACKSIZE	20
private Process	*procstack[PROCSTACKSIZE];
private Process	**procsp = procstack;
private Process *botstack;		/* marks bottom of process stack */

private init_procstack()
{
	/*
	 * create a process to mark bottom of process stack, 
	 * and delete it out of the process list
	 */
	
	if (debug_procstack)
		printf("init_procstack: ENTERED\n");
	botstack = process_alloc();
	processlist = 0;
	botstack->p_next = 0;
	botstack->p_state = DELETED;

	curprocess = 0;
	curobj = primobj;
	resetproc();
	if (debug_procstack)
		printf("init_procstack: EXITED\n");
}


public popproc()
{
	int oldmask;

	if (debug_procstack) {
		printf("popproc: ENTERED, wp = 0x%x\n", wp);
		if (wp == botstack)
			printf("popproc: EXITED, wp = botstack\n");
	}
	if (wp == botstack)
		return;

	if (md_state == USERSIDE)
		oldmask = sigblock(mask(SIGCHLD) | mask(SIGINT));
	if (procsp == procstack)	/* shouldn't happen */
		panic("Process stack underflow");
	procsp--;
	wp = *procsp;
	get_globals(wp->p_savearea);
	curobj = wp->p_obj;

	if (md_state == USERSIDE)
		(void) sigsetmask(oldmask);
	if (debug_procstack)
		printf("popproc: EXITED, wp = 0x%x\n", wp);
}

public pushproc(p)
	Process	*p;
{
	int oldmask;

	if (debug_procstack)
		printf("pushproc: ENTERED, p = 0x%x\n", p);
	if (md_state == USERSIDE)
		oldmask = sigblock(mask(SIGCHLD) | mask(SIGINT));

	if (procsp == procstack + PROCSTACKSIZE)
		error("Process stack overflow");
	*procsp++ = wp;
	wp = p;
	get_globals(wp->p_savearea);
	curobj = wp->p_obj;

	if (md_state == USERSIDE)
		(void) sigsetmask(oldmask);
	if (debug_procstack)
		printf("pushproc: EXITED\n");
}

public resetproc()
{
	int oldmask;

	if (debug_procstack)
		printf("resetproc: ENTERED, wp = 0x%x\n", wp);
	oldmask = sigblock(mask(SIGCHLD)|mask(SIGINT));
	procsp = procstack;
	wp = botstack;
	if (curprocess)
		pushproc(curprocess);
	get_globals(wp->p_savearea);
	curobj = wp->p_obj;
	(void) sigsetmask(oldmask);
	if (debug_procstack)
		printf("resetproc: EXITED, wp = 0x%x\n", wp);
}

public setprocess(procid)
	int	procid;
{
	Process	*p;

	if (debug_procstack)
		printf("setprocess: ENTERED, procid = %d\n", procid);
	p = findpid(procid);
	curprocess = p;
	if (debug_procstack)
		printf("setprocess: EXITED, curprocess 0x%x\n", curprocess);
	resetproc();
}

public clearproc()
{
	if (debug_procstack)
		printf("clearproc: ENTERED\n");
	curprocess = 0;
	resetproc();
	if (debug_procstack)
		printf("clearproc: EXITED\n");
}

/*
 * manage pushes and pops to per-process stack
 */

public pcheck(n)
	int n;
{
	int val = wp->p_sp + n;
	if (val > PSSIZE)
		error("per-process stack overflow");
	if (val < 0)
		error("per-process stack underflow");
}

public ppush(length, what)
	int length;
	pointer what;
{
	pcheck(length);
	bcopy(what, &wp->p_stack[wp->p_sp], length);
	wp->p_sp += length;
}

public pointer ppop(length)
	int length;
{
	pcheck(-length);
	wp->p_sp -= length;
	return((pointer)&wp->p_stack[wp->p_sp]);
}


/*
 * make sure current process exists
 */
private check_exists(p)
	Process *p;
{
	if (p == 0 || p == botstack)
		error("No current process");
	switch (p->p_state) {
	case UNKNOWN:
		error("state unknown for %%%d", p->p_id);
	case DELETED:
		error("%%%d has been deleted", p->p_id);
	case IMAGE:
	case COREDUMP:
	case RUNNING:
	case STOPPED:
	case PENDING:
	case EXITED:
	case PROTO:
		break;
	default:
		panic("Bad process state (%d) for %%%d", p->p_state, p->p_id);
	}
}

/*
 * make sure it's ok to read from process 'p'.  only ok for stopped 
 * processes and core dumps.
 */
public check_read(p)
	Process *p;
{
	check_exists(p);
	switch(p->p_state) {
	case STOPPED:
	case COREDUMP:
	case IMAGE:
		break;
	case RUNNING:
	case PENDING:
	case EXITED:
	case DELETED:
		error("Process %%%d is not stopped", p->p_id);
	case PROTO:
		panic("attempt to read from prototype");
	default:
		panic("Bad process state (%d) for %%%d", p->p_state, p->p_id);
	}
}

public boolean read_ok()
{
	switch(wp->p_state) {
	case STOPPED:
	case COREDUMP:
	case PENDING:
	case IMAGE:
		return(true);
	default:
		return(false);
	}
}

/*
 * make sure it's ok to write to process 'p'.
 */
public check_write(p)
	Process *p;
{
	check_exists(p);
	switch(p->p_state) {
	case STOPPED:
	case IMAGE:
		/* ok to write on stopped process */
		break;
	case COREDUMP:
		error("Can't write on coredump (%%%d)", p->p_id);
	case EXITED:
		error("Process %%%d has exited", p->p_id);
	case PENDING:
	case RUNNING:
		error("Process %%%d is not stopped", p->p_id);
	case PROTO:
		panic("attempt to write on prototype");
	default:
		panic("Bad process state (%d) for %%%d", p->p_state, p->p_id);
	}
}

check_continue(p)
	Process *p;
{
	check_exists(p);
	switch(p->p_state) {
	case STOPPED:
		/* ok to continue stopped process */
		break;
	case COREDUMP:
		error("Can't continue coredump (%%%d)", p->p_id);
	case IMAGE:
		error("Can't continue program image (%%%d)", p->p_id);
	case EXITED:
		error("Process %%%d has exited", p->p_id);
	case PENDING:
	case RUNNING:
		error("Process %%%d is not stopped", p->p_id);
	case PROTO:
		panic("attempt to continue prototype");
	default:
		panic("Bad process state (%d) for %%%d", p->p_state, p->p_id);
	}
}

public boolean isprocess()
{
	if (wp == 0 || wp == botstack)
		return(false);
	switch(wp->p_state) {
	case RUNNING:
	case STOPPED:
	case PENDING:
		return(true);
	default:
		return(false);
	}
}

private char *signame[NSIG + 1] = {
	"signal 0", "SIGHUP", "SIGINT", "SIGQUIT", "SIGILL", "SIGTRAP",
	"SIGIOT", "SIGEMT", "SIGFPE", "SIGKILL", "SIGBUS", "SIGSEGV", "SIGSYS",
	"SIGPIPE", "SIGALRM", "SIGTERM", "SIGURG", "SIGSTOP", "SIGTSTP",
	"SIGCONT", "SIGCHLD", "SIGTTIN", "SIGTTOU", "SIGIO", "SIGXCPU",
	"SIGXFSZ", "SIGVTALRM", "SIGPROF", "SIGWINCH", "signal 29",
	"SIGUSR1", "SIGUSR2", "signal 32"
};

/*
 * print the state of working process
 */
public printstatus(is_ps)
	boolean is_ps;
{
	Process proc;
	int	oldmask;

	if (is_ps)
		printf(wp == curprocess ? "* " : "  ");
	else
		printf("  ");
	printprocid();
	if (multi_objs)
		printf("(%-*s) ", longest_objname, wp->p_obj->o_name);
	switch(wp->p_state) {
	case IMAGE:
		printf("Program Image\n");
		break;
	case COREDUMP:
		printf("Core in \"%s\" caused ", wp->p_corename);
		showstopped(is_ps);
		break;
	case STOPPED:
		printf("Stopped ");
		showstopped(is_ps);
		break;
	case PENDING:
		/*
		 * as far as the user is concerned, PENDING is RUNNING.
		 */
#ifdef DEBUG
		printf("Pending\n");
		break;
#endif
	case RUNNING:
		/*
		 * grab a consistent copy of the process
		 */
		oldmask = sigblock(mask(SIGCHLD));		/* lock */
		proc = *wp;
		(void) sigsetmask(oldmask);			/* unlock */
		/*
		 * print detailed info about what's running
		 */
		if (proc.p_state != RUNNING)
			printf("Running\n");
		else {
			printf("Running ");
			if (proc.p_cmd == pstep && proc.p_stepover)
				printf("next");
			else if (proc.p_cmd == pstep)
				printf("step");
			else if (proc.p_cmd == pcont)
				printf("cont");
			else
				printf("unknown command!");
			printf("\n");
		}
		break;
	case EXITED:
		printf("Exited\n");
		break;
	case DELETED:
		printf("Deleted\n");
		break;
	case UNKNOWN:
	default:
		printf("Unknown\n");
		break;
	}
	fflush(stdout);
}

private char *whystopped[] = {
	/* BY_EVENT */		"event",
	/* BY_BPT */		"stop",
	/* BY_STEP */		"step",
	/* BY_STOP */		"stop",
	/* BY_ERROR */		"error",
	/* BY_CALLRTN */	"call return",
	/* BY_RTERR */		"run-time error",
	/* BY_CMD */		"state machine",
	/* BY_STOPALL */	"stop all",
	/* BY_RETURN */		"return"
};

public char *whystop(cmd)
	Stopinfo cmd;
{
	return(whystopped[(int)cmd]);
}


/*
 * print the status of a stopped process
 *
 * Since this gets called before grabtty() is called to restore tty modes
 * we could be running without CRMOD.  So put out both \r\n instad of
 * just \n as a terminator.
 */
private showstopped(is_ps)
	boolean is_ps;
{
	switch(wp->p_reason) {
	case PTS_FORK:
		printf("by Fork");
		break;
	case PTS_EXEC:
		printf("by Exec");
		break;
	case PTS_EXIT:
		printf("by Exit");
		break;
	case SIGTRAP:
		switch(wp->p_stopinfo) {
		case BY_STOPALL:
		case BY_BPT:
			printf("at breakpoint");
			if (wp->p_stopeventid)
				printf(" %d", wp->p_stopeventid);
			break;
		case BY_ERROR:
		case BY_STEP:
		case BY_RTERR:
		case BY_STOP:
		case BY_CALLRTN:
		case BY_RETURN:
			printf("by %s", whystopped[(int)wp->p_stopinfo]);
			break;
		case BY_CMD:
			if (wp->p_cmd == pstep && wp->p_stepover)
				printf("after next");
			else if (wp->p_cmd == pstep)
				printf("after step");
			else if (wp->p_cmd == pcont)
				printf("after cont");
			else
				printf("after ????");
			break;
		case BY_EVENT:
		default:
			printf("by ");
			printsig(wp->p_reason);
			break;
		}
		break;
	default:
		printf("by ");
		printsig(wp->p_reason);
		break;
	}
	putchar(' ');
	pushproc(wp);
	if (ptrace_read_ok(reg(PROGCTR)) or is_ps) {
		printloc();
		putchar('\r');
		putchar('\n');
		if (!is_ps) {
			if (curline > 0 && !(wp->p_stepi || wp->p_tracei))
				printlines(curline, curline, false);
			else
				printinst(reg(PROGCTR), reg(PROGCTR));
		}
	} else {
		printf(" PC = 0x%x\r\n", reg(PROGCTR));
	}
	popproc();
}


public create(argv, infile, outfile)
	String argv[];
	String infile;
	String outfile;
{
	Process *p;
	String pargv[2];
	int pid;
	int oldmask;
	int i;

	if (debug_process)
		printf("create called\n");

	if (argv == nil) {
		argv = pargv;
		pargv[0] = primobj->o_pathname;
		pargv[1] = nil;
	} else {
		argv[argc] = nil;
	}
	fflush(stdout);
	fflush(stderr);

	/*
	 * fork and exec the process.  have to block sigchild because
	 * the child could stop before we get back from fork.  since the
	 * sigchild handler will create a process if there isn't one
	 * already, we might create to process structures for the same pid.
	 */

	oldmask = sigblock(mask(SIGCHLD) | mask(SIGINT));	/* lock */
	pid = vfork();
	if (pid == -1)
		error("can't fork");
	if (pid == 0) {						/* child */
		if (infile)
			infrom(infile);
		if (outfile)
			outto(outfile);
		for (i = 3; i < NOFILE; i++)
			(void) close(i);
		fixtty();
		(void) sigsetmask(0);
		execv(argv[0], argv);
		(void) write(1, "eek! bad exec!\n", 15);
		_exit(1);
		/*NOTREACHED*/
	}
	if (find_process(pid))
		panic("pid %d already used!\n", pid);
	p = create_process(pid);
	p->p_quiet = true;
	p->p_cmd = pnothing;
	p->p_obj = create_object(argv[0]);
	p->p_job = newjob(pid);
	setproc(p->p_id);
	(void) sigsetmask(oldmask);				/* unlock */
	waitfor(p);
	(void) xptrace(XPT_SETSIGMASK, p->p_pid, 0, p->p_sigmask);
	checkobj(p->p_obj);
	p->p_quiet = false;
}

/*
 * create a process whose only purpose is to supply symbol tables for
 * setting breakpoints.
 */
public protoproc(o)
	Object *o;
{
	Process *p;

	if (debug_process)
		printf("protoproc called\n");

	p = process_alloc();
	p->p_state = PROTO;
	p->p_obj = o;
	p->p_id = -1;
	p->p_pid = -1;
	o->o_proto = p;
}


/*
 * create a process context for resolving variables, files, etc,
 * during parse of a command.
 */
pushproto(proto)
	Node proto;
{
	int procid;
	char *objname;
	Process *old;
	Object *o;

	old = wp;
	if (proto) {
		switch(proto->op) {
		case O_LCON:
			procid = proto->value.lcon;
			if (procid == 0 || procid == ALLPROCS) {
				o = pathobj(curpath, cursource);
				if (o == nil)
					o = curobj;
			} else {
				pushpid(procid);
				return;
			}
			break;
		case O_SCON:
			objname = proto->value.scon;
			o = find_object(objname);
			if (o == nil)
				error("no program named %s", objname);
			break;
		default:
			panic("unexpected op in pushproto");
		}
	} else {
		pushproc(wp);
		return;
	}
	assert(o != 0);
	if (o == wp->p_obj) {
		pushproc(wp);
	} else if (o->o_proto == 0) {
		error("no symbols for program %s", o->o_name);
	} else {
		pushproc(o->o_proto);
		copy_globals(old->p_savearea, wp->p_savearea);
		get_globals(wp->p_savearea);
	}
}


/*
 * try to decide what object we're talking about on the basis of the
 * file name given.  returns whether a process was pushed.
 */
public boolean protofile(filename)
	String filename;
{
	Object *o;
	Process *old;

	o = pathobj(nil, filename);
	if (o == nil || o == curobj)
		return(false);
	old = wp;
	pushproc(o->o_proto);
	copy_globals(old->p_savearea, wp->p_savearea);
	get_globals(wp->p_savearea);
	return(true);
}


/*
 * state machine for the 'cont' function.  returns the next state
 * for the process.
 *
 *	if at a breakpoint
 *		step one instruction
 *	continue
 */
private Pstate pcont(state)
	Cmdstate state;
{
	if (debug_process) {
		printf("pcont(%s) entered, wp->p_signal= %d\n",
		    show_cmdstate(state), wp->p_signal);
		printf("   wp->p_id= %d, wp->p_pid= %d, wp->p_parent= %d\n",
		    wp->p_id, wp->p_pid, wp->p_parent);
		printf("   wp->p_tracing= %c, wp->p_tracei= '%c'",
		    wp->p_tracing ? 'T' : 'F', wp->p_tracei ? 'T' : 'F');
		printf(", wp->p_signal= '%c', wp->p_tempbp= %c\n",
		wp->p_signal ? 'T' : 'F', wp->p_tempbp ? 'T' : 'F');
		fflush(stdout);
	}

	if (wp->p_tempbp)
		unsetbp(wp->p_tempbp);

	wp->p_cmd = pcont;
	wp->p_cmdstate = CONT;

	if (wp->p_tracing || at_bpt()) {
		unsetallbps();
		if (wp->p_tracing && !wp->p_tracei && (wp->p_tempbp = shouldskip(reg(PROGCTR)))) {
			setbp(wp->p_tempbp);
			wp->p_contmode = PT_CONTSIG;
			if (debug_process)
			     printf("Pcont: setbp and continuing with PT_CONTSIG\n");
		} else {
			if (wp->p_signal) {	/* ptrace change 22-Jul-86 */
				setallbps();
				wp->p_contmode = PT_CONTSIG;
				if (debug_process)
				     printf("Pcont: setallbps and continuing with PT_CONTSIG\n");
			} else {
				wp->p_contmode = PT_SSTEP;
				if (debug_process)
				     printf("Pcont: continuing with PT_SSTEP\n");
			}
		}
	} else {
		setallbps();
		wp->p_contmode = PT_CONTSIG;
		if (debug_process)
		    printf("Pcont: setallbps and continuing with PT_CONTSIG\n");
	}

	/*
	 * if p_tempbp is set, we're skipping a function with no source.
	 * printnews() would be fooled into thinking we actually entered
	 * the function, since p_oldpc would point to a call.  so p_oldpc
	 * is cleared.
	 */
	if (wp->p_tempbp)
		wp->p_oldpc = 0;
	else
		wp->p_oldpc = reg(PROGCTR);

	if (debug_process) {
	    printf("pcont returning, wp->p_tempbp= 0x%x, wp->p_oldpc= 0x%x\n",
		wp->p_tempbp, wp->p_oldpc);
	    fflush(stdout);
	}
	return(RUNNING);	/* cont never causes process to stop */
}

/*
 * state machine for stepping.  returns the next state for the process.
 *
 *	repeat
 *
 *		if on procedure call instruction && doing 'next'
 *			step one instruction
 *			set a bpt at return address
 *			continue to bpt
 *		else
 *			step one instruction
 *
 *	until statement boundary (or instr bndry if stepping by insts)
 */
private Pstate pstep(state)
	Cmdstate state;
{
	boolean fall_into = false;

	switch(state) {

	case STEPOVER:
		/*
		 * instruction was a subroutine call instruction; the
		 * process has stepped into the subroutine.  set a break
		 * at the return address and continue.
		 * there is no way for the pc to be sitting on a breakpoint
		 * since breakpoints are out.
		 * have to be careful that this 'unofficial' breakpoint
		 * is deleted properly in fixbps().
		 */
		/* NEXT LINE IS MACHINE AND RUN-TIME DEPENDENT!!! */
		dread(&wp->p_tempbp, reg(STKP), sizeof(Address)); /* ret addr */
		if (debug_process) {
		    printf("pstep(STEPOVER), wp->p_tempbp= 0x%x\n",
			wp->p_tempbp);
		    fflush(stdout);
		}
		setbp(wp->p_tempbp);
		wp->p_cmdstate = CONT;
		wp->p_contmode = PT_CONTSIG;
		return(RUNNING);

	case CONT:
		/*
		 * hopefully continued to the temporary breakpoint set at the
		 * return address of the subroutine.  if we're not there yet,
		 * just try again.
		 */
		if (debug_process) {
		    printf("pstep(CONT), PC= 0x%x, wp->p_tempbp= 0x%x\n",
			reg(PROGCTR), wp->p_tempbp);
		    fflush(stdout);
		}
		if (reg(PROGCTR) != wp->p_tempbp)
			return(RUNNING);
		unsetbp(wp->p_tempbp);
		if (debug_process) {
		    printf("pstep(CONT) falls into STEP, wp->p_stepi= %c\n",
			wp->p_stepi ? 'T' : 'F');
		    fflush(stdout);
		    fall_into = true;
		}

		/* fall into...  */

	case STEP:
		/*
		 * just stepped over an instruction.
		 * stop the process if we've reached a source line
		 * boundary.  otherwise fall into the init case...
		 */
		if (debug_process and !fall_into) {
		    printf("pstep(STEP), wp->p_stepi= %c\n",
			wp->p_stepi ? 'T' : 'F');
		}
		if (wp->p_stepi || linelookup(reg(PROGCTR)) != 0) {
			wp->p_oldpc = 0;
			wp->p_stopinfo = BY_CMD;
			return(STOPPED);
		}
		if (debug_process) {
		    printf("pstep(STEP) falls into INIT");
		    printf(", wp->p_stepi= %c, wp->p_stepover= %c\n",
			wp->p_stepi ? 'T' : 'F', wp->p_stepover ? 'T' : 'F');
		    fall_into = true;
		    fflush(stdout);
		}

	case INIT:
		/*
		 * starting a step from the current location.  note 
		 * whether we're stepping into a subroutine.
		 */
		 if (debug_process and !fall_into) {
		    printf("pstep(INIT)");
		    printf(", wp->p_stepi= %c, wp->p_stepover= %c\n",
			wp->p_stepi ? 'T' : 'F', wp->p_stepover ? 'T' : 'F');
		 }
		unsetallbps();
		wp->p_cmd = pstep;
		if (wp->p_stepover && iscall(reg(PROGCTR))) {
			/* next command, and we're on a call instr. */
			/* step in, and prepare to continue out */
			wp->p_cmdstate = STEPOVER;
			wp->p_contmode = PT_SSTEP;
		} else if (!wp->p_stepi && (wp->p_tempbp = shouldskip(reg(PROGCTR)))) {
			/* step command on call, but skip anyway because */
			/* there's no source for the routine and we know how */
			/* to set bp after this instr */
			wp->p_cmdstate = CONT;
			wp->p_contmode = PT_CONTSIG;
			setbp(wp->p_tempbp);
		} else {
			/* normal instruction -- just step through it */
			wp->p_cmdstate = STEP;
			wp->p_contmode = PT_SSTEP;
		}
		wp->p_oldpc = reg(PROGCTR);
		return(RUNNING);

	default:
		panic("pstep: bad cmd state %d for process %%%d",
			wp->p_cmdstate, wp->p_id);
		/*NOTREACHED*/
	}
}

/*ARGSUSED*/
private Pstate pnothing(state)
	Cmdstate state;
{
	return(STOPPED);
}

/*
 * print some info about a process that's just been noticed
 * by babysitter().
 */
private printstop()
{
	printf("process %d stopped at 0x%x by ", wp->p_id, reg(PROGCTR));

	if (wp->p_reason < NSIG)
		printf("%s", signame[wp->p_reason]);
	else {
		char *s;
		switch (wp->p_reason) {
		case PTS_FORK:
			s = "PTS_FORK";
			break;
		case PTS_EXEC:
			s = "PTS_EXEC";
			break;
		case PTS_EXIT:
			s = "PTS_EXIT";
			break;
		default:
			s = "????";
			break;
		}
		printf("%s", s);
	}
	printf("\n");
	fflush(stdout);
}

/* 
 * this is the SIGCHLD handler.  the debugger is sent a SIGCHLD when any of
 * its descendants changes state.  the signal mask for this function
 * should block SIGCHLD and SIGINT.
 *
 * the basic job of this routine is to take a look at a process 
 * that has stopped and decide what to do with it.
 */

private struct proc proc;

private babysitter()
{
	struct pt_stop	status;
	Process		*p;
	Process		*parent;
	int		save_errno;

	md_state = INTSIDE;
	save_errno = errno;
	if (debug_process)
		printf("entering babysitter **\n");
	while(ptrace(XPT_STOPSTAT, 0, (int) &status, 0) != -1) {

		/*
		 * find the process that stopped and why
		 */
		p = find_process(status.ps_pid);
		if (p == 0)
			p = create_process(status.ps_pid);
		p->p_state = STOPPED;
		p->p_reason = status.ps_reason;
		p->p_stopinfo = BY_EVENT;
		(void) xptrace(XPT_RREGS, p->p_pid, (int) &p->p_regs, 0);
		pushproc(p);
		setjob(wp->p_job);
		wp->p_regschanged = false;
		retry = false;

		if (debug_process) {
			printstop();
			show_mask();
			show_flags();
		}

		/*
		 * based on the reason why the process stopped, decide
		 * what to do with it.
		 */
		switch(p->p_reason) {

		case PTS_FORK:
			/*
			 * In the case of a fork, we have a new process
			 * structure that must inherit flags from its parent.
			 * It should stop if the parent is supposed to have
			 * forks caught.  If we were invoked as dbx then we
			 * are going to give away this process.
			 */
#ifdef XPT_UNDEBUG
			if (is_dbx) {
			    (void) xptrace(XPT_UNDEBUG, p->p_pid, 1, 0);
			    delete_process(p);
			    unsetjob();
			    popproc();
			    continue;
			}
#endif XPT_UNDEBUG
			(void) xptrace(XPT_RPROC, p->p_pid, (int)&proc, 0);
			parent = find_process(proc.p_ppid);
			if (parent != 0) {
				copy_process(parent, p);
			        if (p->p_flags & PF_CATCH_FORK)
					p->p_newstate = STOPPED;
				else
					p->p_newstate = RUNNING;
			} else {
				printf("can't find parent for %%%d!\n", p->p_id);
				p->p_newstate = STOPPED;
			}
			p->p_signal = 0;
#define STEPFORKBUG
#ifdef STEPFORKBUG
			/*
			 * if parent was continued with PT_SSTEP,
			 * T bit is still on, so clear it.
			 * needless to say, this is ns32000 dependent.
			 */
			if (p->p_contmode == PT_SSTEP)
				setreg(PSR, reg(PSR) & ~PSR_TRACEBIT);
#endif STEPFORKBUG
			(void) bpact();
			break;

		case PTS_EXEC:
			p->p_signal = 0;
			reset_process();
			curobj = p->p_obj;
			if (p->p_flags & PF_CATCH_EXEC)
				p->p_newstate = STOPPED;
			else {
				p->p_newstate = RUNNING;
				p->p_cmdstate = INIT;
			}
			(void) bpact();
			break;

		case PTS_EXIT:
			p->p_signal = 0;
			if (p->p_flags & PF_CATCH_EXIT)
				p->p_newstate = STOPPED;
			else
				p->p_newstate = RUNNING;
			(void) bpact();
			break;

		case SIGTRAP:
			checktrap();
			break;

		default:
			/* 
			 * this means we got a signal.  currently any
			 * signals to be ignored are taken care of by
			 * the operating system so we never see them.
			 * any signal that gets here should stop the process.
			 */
			p->p_newstate = STOPPED;
			if (p->p_reason > NSIG)
				panic("p_reason > NSIG for %%%d", p->p_id);
			p->p_signal = p->p_reason;
			break;
		}

		/*
		 * make the appropriate state transition based on p_newstate.
		 */
		newstate();

		unsetjob();
		popproc();
	}
	
#ifdef NOTYET
	/*
	 * collect up exited processes and note why they exited.
	 * do i really want to do this since only direct children
	 * of dbx will be inherited?
	 */
#include <errno.h>
#include <sys/wait.h>
	(void) onsyserr(ECHILD, 0); /* ECHILD from wait3 just means no more */
	for(;;) {
		int pid;
		pid = wait3(0, WNOHANG, 0);
		if (pid < 1)
			break;
		printf("collected %d\n", pid);
	}
#endif NOTYET

	errno = save_errno;
	md_state = USERSIDE;
	if (debug_process) {
		printf("    exiting babysitter **\n");
		fflush(stdout);
	}
}


/*
 * decide what next state should be for a process that received a
 * SIGTRAP, and do the appropriate side effects like breakpoints.
 */
private checktrap()
{
	wp->p_newstate = RUNNING;
	wp->p_stopeventid = 0;

	/*
	 * if doing a "stop all" command, stop the process no matter what.
	 * this is a global because stopall() doesn't know for sure who's
	 * going to stop.  it just sends out a broadcast signal and waits
	 * until everyone's stopped.
	 */
	if (stopallmode) {
		wp->p_signal = 0;
		wp->p_stopinfo = BY_STOP;
		wp->p_newstate = STOPPED;
	}

	/*
	 * it's possible that there is a SIGTRAP pending to the process
	 * from a "stop all" command.  if the process stopped because
	 * of this, just continue it.
	 */
	else if (wp->p_trap_pending == true) {
		wp->p_trap_pending = false;
		wp->p_newstate = RUNNING;
		retry = true;
	}

	/*
	 * check if a breakpoint was hit.  this doesn't count the
	 * temporary breakpoints that the state machines might set.
	 */
	else if (bpact()) {
		/* everything is handled by bpact */
	}
	
	/*
	 * did the process stop because of a state machine?  if so,
	 * the state machine will be consulted in newstate() which will
	 * say whether to stop or not.
	 */
	else if (wp->p_contmode == PT_SSTEP || reg(PROGCTR) == wp->p_tempbp) {
		wp->p_signal = 0;
		wp->p_stopinfo = BY_STEP;
		wp->p_newstate = RUNNING;
	}

	/*
	 * otherwise must have stopped because of a SIGTRAP signal.
	 */
	else {
		wp->p_signal = SIGTRAP;
		wp->p_stopinfo = BY_EVENT;
		wp->p_newstate = STOPPED;
	}

	if (debug_process) {
	    printf("multi-p: checktrap exiting\n");
	    printf("    wp->p_signal=   %s\n", show_signal(wp->p_signal));
	    printf("    wp->p_stopinfo= %s\n", show_stopinfo(wp->p_stopinfo));
	    printf("    wp->p_newstate= %s\n", show_pstate(wp->p_newstate));
	    fflush(stdout);
	}
}


/* 
 * make the current process do the state transition in 'p_newstate'.
 * this is only place where the p_state changes (except for the
 * EXITED==>DELETED transition).
 */
public newstate()
{
	if (debug_process) {
		printf("newstate entered, wp->p_state= %s, wp->p_newstate= %s, ",
		    show_pstate(wp->p_state), show_pstate(wp->p_newstate));
		printf("p_pending= %c, ", wp->p_pending ? 'T' : 'F');
		printf("retry= %c\n", wp->p_pending ? 'T' : 'F');
		fflush(stdout);
	}

	if (wp->p_newstate == RUNNING && !wp->p_pending && !retry) {
		if (debug_process) {
		    printf("newstate calling %s(%s)\n", show_pcmd(wp->p_cmd),
			show_cmdstate(wp->p_cmdstate));
		}
		wp->p_newstate = (*wp->p_cmd)(wp->p_cmdstate);
		if (debug_process) {
		    printf("returned to newstate(), wp->p_newstate = %s\n",
			show_pstate(wp->p_newstate));
		}
	}

	if (wp->p_newstate == RUNNING && !wp->p_pending) {

		if (wp->p_reason == PTS_EXIT) {
			wp->p_state = EXITED;		/* never see it again */
			unbusy(wp->p_job, wp->p_id);
			if (wp->p_flags & PF_DELETE_EXIT)
				delete_process(wp);
		} else {
			wp->p_state = RUNNING;
			if (wp->p_regschanged)
				(void)xptrace(XPT_WREGS, wp->p_pid, (int)&wp->p_regs, 0);
			if (wp->p_sigmask != IGNORESIGS && wp->p_sigmask != default_sigmask) {
				wp->p_sigmask = default_sigmask;
				xptrace(XPT_SETSIGMASK, wp->p_pid, 0, wp->p_sigmask);
			}
			if (wp->p_flags & PF_SETFLAGS)
				wp->p_flags = default_flags;
		}

		if (debug_process) {
			printf("process %d %s with %s\n", wp->p_id,
			 wp->p_contmode == PT_SSTEP ? "stepping" : "continuing",
			 signame[wp->p_signal]);
			fflush(stdout);
		}

		(void) xptrace(wp->p_contmode, wp->p_pid, wp->p_regs.pr_PC, wp->p_signal);
		return;
	}
	
	else if (wp->p_newstate == STOPPED || wp->p_pending) {
		unsetallbps();
		if (wp->p_tempbp)
			unsetbp(wp->p_tempbp);

		if (debug_process) {
			printf("process %d %s\n", wp->p_id,
			 wp->p_pending ? "pending" : "stopping");
			fflush(stdout);
		}

		if (wp->p_pending) {
			wp->p_state = PENDING;
			wp->p_stopeventid = 0;
		} else /* wp->p_newstate == STOPPED */ {
			delbpts(wp->p_id, false);	/* delete temp bpts */
			if (wp->p_state != IMAGE)
				wp->p_state = STOPPED;
			setcurfunc(whatblock(reg(PROGCTR)));
			getsrcpos();
			unbusy(wp->p_job, wp->p_id);
			if (!wp->p_quiet && wp->p_stopinfo != BY_CALLRTN)
				printstatus(false);
			wp->p_stepi = false;
		}
	
	} else {
		panic("Bad state (%d) for %%%d in newstate() (%spending)\n", 
			wp->p_newstate, wp->p_id, wp->p_pending ? "" : "not ");
	}
	if (debug_process) {
		printf("newstate exiting\n");
		fflush(stdout);
	}
}

/*
 * print out the id of the current process
 */
public printprocid()
{
	printf(PROCIDFMT, wp->p_id);
}

/*
 * set the current process up so newstate will stop it, but if it was
 * stopped by exit leave it that way.  processes stopped by exit will
 * not report back once continued, and dbx needs to know not to wait
 * for them.
 */
public setstopped(why)
	Stopinfo why;
{
	wp->p_newstate = STOPPED;
	if (wp->p_reason != PTS_EXIT) {
		wp->p_stopinfo = why;
		wp->p_signal = 0;
		wp->p_reason = SIGTRAP;
		wp->p_stopeventid = cureventid;  /* see doqueue() */
	}
}

/*
 * are we really tracing the process?
 */
public boolean tracing()
{
	if (wp->p_tracing && wp->p_cmd == pcont)
		return(true);
	else
		return(false);
}


public process_init()
{
	struct sigvec v;

	if (debug_process)
	    if (wp)
		printf("process_init: wp->p_id=%d, wp->p_pid= %d\n",
		    wp->p_id, wp->p_pid);
	    else
	        printf("process_init called, wp= (nil)\n");

	allocate_processes();
	init_procstack();
	default_sigmask = mask(SIGHUP) | mask(SIGKILL) | mask(SIGALRM)
		| mask(SIGCONT) | mask(SIGCHLD);
	default_flags = PF_CATCH_EXEC | PF_CATCH_EXIT | PF_SETFLAGS;
	/* special actions of pdbx and dbx */
	if (has_mplib) {
		default_sigmask |= mask(SIGSEGV);
	}
	if (is_dbx) {
		default_flags &= ~PF_CATCH_EXEC;
	}
	v.sv_handler = babysitter;
	v.sv_mask = mask(SIGCHLD) | mask(SIGINT);
	v.sv_onstack = FALSE;
	(void) sigvec(SIGCHLD, &v, (struct sigvec *)0);
	(void) xptrace(XPT_MPDEBUGGER, 0, 0, 0);
	md_state = USERSIDE;
}


public boolean readcore(filename)
	String filename;
{
	Process *p;
	boolean coreok;

	p = create_process(-1);
	pushproc(p);
	wp->p_stopinfo = BY_EVENT;
	wp->p_state = STOPPED;
	coreok = (boolean) coredump_readin(filename);
	wp->p_state = COREDUMP;
	popproc();
	if (!coreok) {
		delete_process(p);
	} else {
		setproc(p->p_id);
		printf("[using memory image in %s]\n", wp->p_corename);
		set_default_source_file();
		setcurfunc(whatblock(reg(PROGCTR)));
		getsrcpos();
	}
	return(coreok);
}


/*
 * restore stuff in this file to a useful state
 */
public process_reset()
{
	synchronous = (boolean) not async_flag;
	resetproc();
}

private char *pfname[] = {
	"PT_CHILD", "PT_RTEXT", "PT_RDATA", "PT_RUSER", "PT_WTEXT", "PT_WDATA", 
	"PT_WUSER", "PT_CONTSIG", "PT_KILL", "PT_SSTEP", "XPT_MPDEBUGGER", 
	"XPT_SETSIGMASK", "XPT_RREGS", "XPT_WREGS", "XPT_RPROC", "XPT_SIGNAL", 
	"XPT_STOPSTAT", "XPT_DUPFD", "XPT_DEBUG", "XPT_OPENT",
	"XPT_WATCHPT_SET", "XPT_WATCHPT_CLEAR", "XPT_UNDEBG"
};

public xptrace(func, pid, address, data)
	int	func, pid, address, data;
{
	int val;
	int e;				/* save errno */

	if (debug_ptrace) {
		fflush(stderr);
		printf("calling ptrace(%s, %d, 0x%x, 0x%x)",
			pfname[func], pid, address, data);
		if (func == PT_CONTSIG or func == PT_SSTEP)
			printf("\n");
		fflush(stdout);
	}
	errno = 0;
	val = ptrace(func, pid, address, data);
	e = errno;
	if (e) {
		printf("error ret: ptrace(%s, %d, 0x%x, 0x%x) = 0x%x\n",
			pfname[func], pid, address, data, val);
		fflush(stdout);
	} else if (debug_ptrace and func != PT_CONTSIG and func != PT_SSTEP) {
		printf(" = 0x%x\n", val);
		fflush(stdout);
	}
	if (e) {
		errno = e;
		perror("ptrace");
		erecover();
	}
	return(val);
}

/*
 * Is a ptrace of "address" for the current process going to be ok?
 */

private boolean ok_to_ptrace(address, op)
	int address;
	int op;
{
	int val1, val2, loc, e, save_errno;
	INTFUNC *f;

	/* Read location. */

	save_errno = errno;
	f = onsyserr(EIO, ERR_IGNORE);
	errno = 0;
	loc = address & ~0x1;
	val1 = ptrace(PT_RTEXT, wp->p_pid, loc, 0);
	e = errno;
	(void) onsyserr(EIO, f);
	errno = save_errno;

	if (e)
		return (false);
	else if (op == PT_RTEXT)
		return (true);

	/* Now write it back if checking for write ok. */

	save_errno = errno;
	f = onsyserr(EIO, ERR_IGNORE);
	errno = 0;
	loc = address & ~0x1;
	val2 = ptrace(PT_WTEXT, wp->p_pid, loc, val1);
	e = errno;
	(void) onsyserr(EIO, f);
	errno = save_errno;

	if (e)
		return (false);
	else 
		return (true);
}

public boolean ptrace_read_ok(address)
	int address;
{
	boolean b;

	b = ok_to_ptrace(address, PT_RTEXT);
	return(b);
}

public boolean ptrace_write_ok(address)
	int address;
{
	boolean b;

	b = ok_to_ptrace(address, PT_WTEXT);
	return(b);
}

/*
 * continue current process with signal 'sig'.  this procedure gets the
 * process going, but does not wait for it to stop.  so when this procedure
 * returns, the process is probably running. 
 */
public cont(sig, withtty)
	int	sig;
	boolean	withtty;
{
	int oldmask;

	if (debug_process) {
		printf("cont(sig= %d, withtty= %s)\n",
		    sig, withtty ? "true" : "false");
		printf("    wp->p_id= %d, wp->p_pid= %d, wp->p_parent= %d\n",
		    wp->p_id, wp->p_pid, wp->p_parent);
	}

	if (sig < DEFSIG || sig > NSIG)
		error("signal number must be in 0..%d", NSIG);
	check_continue(wp);
	oldmask = sigblock(mask(SIGINT));
	if (synchronous && withtty)
		giveawaytty(wp->p_job, wp->p_id); /* now because can error() */
	wp->p_newstate = RUNNING;
	wp->p_cmd = pcont;
	wp->p_cmdstate = INIT;
	if (sig != DEFSIG)
		wp->p_signal = sig;
	newstate();
	(void) sigsetmask(oldmask);
	if (debug_process) {
		printf("cont exiting, wp->p_signal= %d\n",
		    wp->p_signal);
		fflush(stdout);
	}
}

/*
 * apply step, next, or cont to all stopped processes.  don't have
 * to lock process list because stopped processes stay stopped.
 */
public contall(n)
	Node	n;
{
	Process	*p;
	boolean	save_sync;

	if (debug_process)
		printf("contall called\n");

	if (wp->p_state == STOPPED && synchronous)
		giveawaytty(wp->p_job, wp->p_id); /* now because can error() */

	save_sync = synchronous;
	synchronous = false;
	
	/*
	 * continue all the processes in background 
	 */
	for (p = processlist; p != (Process *)0; p = p->p_next) {
		if (p->p_state == STOPPED) {
			pushproc(p);
			eval(n);
			popproc();
		}
	}

	synchronous = save_sync;
}


/*
 * 'kill all' command
 */
public kill_all()
{
	if (debug_process)
		printf("kill_all called\n");

	while(processlist) {
		pushproc(processlist);
		kill_process();
		popproc();
	}
}

/*
 * return whether the given process is being killed 
 */
private int killprocid;

public boolean killing(id)
	int id;
{
	if (id == killprocid)
		return(true);
	else
		return(false);
}

/*
 * 'kill' command
 */
public kill_process()
{
	int	oldmask;
	Pstate	oldstate;

	check_exists(wp);
	oldmask = sigblock(mask(SIGCHLD) | mask(SIGINT));
	killprocid = wp->p_id;
	delbpts(wp->p_id, true);	/* delete all breakpoints */
	wp->p_quiet = true;

	/*
	 * get the process to be about to exit.  this loop is necessary
	 * because the process may be making state transitions on its own.
	 */
	while(wp->p_state == RUNNING || wp->p_state == STOPPED) {
		oldstate = wp->p_state;
		if (oldstate == RUNNING)
			kill(wp->p_pid, SIGKILL);
		else if (oldstate == STOPPED)
			cont(SIGKILL, false);
		else {
			panic("kill_process: bad state %d for %%%d",
				wp->p_state, wp->p_id);
			/*NOTREACHED*/
		}
		while(wp->p_state == oldstate && wp->p_reason != PTS_EXIT)
			(void) sigpause(oldmask);
	}
	delete_process(wp);
	killprocid = 0;
	(void) sigsetmask(oldmask);
}


/*
 * the 'next' command
 */
public next(sig)
	int	sig;
{
	int oldmask;

	if (sig < DEFSIG || sig > NSIG)
		error("signal number must be in 0..%d", NSIG);
	check_continue(wp);
	oldmask = sigblock(mask(SIGINT));
	if (synchronous)
		giveawaytty(wp->p_job, wp->p_id); /* now because can error() */
	wp->p_stepover = true;
	wp->p_newstate = RUNNING;
	wp->p_cmd = pstep;
	wp->p_cmdstate = INIT;
	if (sig != DEFSIG)
		wp->p_signal = sig;
	newstate();
	(void) sigsetmask(oldmask);
}

/*
 * the 'ps' command.  it's dangerous just to loop through p_next because
 * the process list could be being updated at the same time.  however,
 * i don't think this is a problem because insertions are done at the
 * head of the list, and deletions are done so that p_next is always
 * pointing somewhere meaningful.
 */
public ps(id)
	int id;
{
	register Process *p;

	if (id) {
		pushpid(id);
		printstatus(true);
		if (isprocess()) {
			printf("  ");
			printprocid();
			printf("pid %d", wp->p_pid);
			printf(", parent ");
			if (wp->p_parent == 0)
				printf("[pdbx]");
			else
				printf("%%%d", wp->p_parent);
			printf(".\n");
		}
		popproc();
	} else if (processlist) {
		for (p = processlist; p; p = p->p_next) {
			pushproc(p);
			printstatus(true);
			popproc();
		}
	} else {
		printf("No active processes\n");
	}
}

/*
 * the 'run' command
 */
public run(background)
	boolean background;
{
	if (argc == 0)
		error("no program specified");
	fixbps();
	kill_all();
	create(argv, infile, outfile);
	wfixdefprocs(wp->p_id);
	cont(0, !background);
	erecover();
}

/*
 * the 'create' command
 */
createcmd()
{
	create(argv, infile, outfile);
	set_default_source_file();
	printstatus(false);
	erecover();
}


/*
 * the 'step' command
 */
public step(sig)
	int	sig;
{
	int oldmask;

	if (sig < DEFSIG || sig > NSIG)
		error("signal number must be in 0..%d", NSIG);
	check_continue(wp);
	oldmask = sigblock(mask(SIGINT));
	if (synchronous)
		giveawaytty(wp->p_job, wp->p_id); /* now because can error() */
	wp->p_stepover = false;
	wp->p_newstate = RUNNING;
	wp->p_cmd = pstep;
	wp->p_cmdstate = INIT;
	if (sig != DEFSIG)
		wp->p_signal = sig;
	newstate();
	(void) sigsetmask(oldmask);
}


/*
 * the 'stop all' command.  RUNNING processes are a real pain since
 * they do things on their own... i can tell a process to stop for
 * the "stop all" command, but there's no guaranteeing that
 * it won't stop for some other reason.  there is a period between
 * the time i notice a process is stopped and when i actually
 * stop it.  anything could happen during this time.
 *
 * so this routine sets 'stopallmode' which informs the SIGCHLD
 * handler that any process that stops with p_reason == SIGTRAP
 * has done so because of a "stop all" command.  this takes care
 * of processes that run into breakpoints during a stop all, as
 * well as processes that are created after i lock the process
 * list but still get the broadcast signal.
 */
public stopall()
{
	int oldmask;
	register Process *p;
	Process **waitfor;
	register unsigned int nrunning = 0;
	register int	i;

	if (debug_process)
		printf("stopall called\n");

	oldmask = sigblock(mask(SIGCHLD) | mask(SIGINT));
	stopallmode = true;

	/*
	 * count the running processes, and assure that any pending
	 * processes don't get restarted.  anything else is already
	 * stopped.
	 */
	for (p = processlist; p; p = p->p_next) {
		if (p->p_state == RUNNING) {
			nrunning++;
			if (debug_stopall)
				printf("stopall: %%%d running\n", p->p_id);
		} else if (p->p_state == PENDING || p->p_pending) {
			pushproc(p);
			setstopped(BY_STOP);
			popproc();
			if (debug_stopall)
				printf("stopall: %%%d pending\n", p->p_id);
		}
	}

	if (nrunning != 0) {

		/*
		 * send a SIGTRAP to all processes.  it's possible but unlikely
		 * that the signal will miss a process being created.
		 */
		(void) xptrace(XPT_SIGNAL, 0 /* to all */, 0, SIGTRAP);


		/*
		 * allocate a list of running processes.  ok to malloc 
		 * because this isn't called from the interupt side.
		 */
		waitfor = (Process **)malloc(nrunning * sizeof(Process *));
		i = 0;
		for (p = processlist; p; p = p->p_next)
			if (p->p_state == RUNNING)
				waitfor[i++] = p;

		/*
		 * wait until all processes stop.  it is possible though 
		 * unlikely that this loop will terminate before a process
		 * has stopped, if that process was just created and is
		 * taking a long time to do so.
		 */
		do {
			(void) sigpause(defsigmask|mask(SIGINT));
			for (i = 0; i < nrunning; i++)
				if (waitfor[i]->p_state == RUNNING)
					break;
		} while(i < nrunning);

		free((char *)waitfor);

		/*
		 * the broadcast SIGTRAP goes to stopped processes, too.  when
		 * they are continued, the pending SIGTRAP will be delivered.
		 * for each process, note whether there is a pending SIGTRAP.
		 * the SIGCHLD handler will ignore the first SIGTRAP from these
		 * processes.
		 */
		for (p = processlist; p; p = p->p_next) {
			if (p->p_state == STOPPED) {
				(void) xptrace(XPT_RPROC, p->p_pid, (int)&proc, 0);
				if (proc.p_sig & mask(SIGTRAP)) {
					p->p_trap_pending = true;
					if (debug_stopall)
						printf("stopall: %%%d trap_pending\n", p->p_id);
				}
			}
		}
	}
		
	stopallmode = false;
	(void) sigsetmask(oldmask);
}

/*
 * the 'stop' command.
 */
public pstop()
{
	int oldmask;

	if (debug_process)
		printf("pstop called\n");

	oldmask = sigblock(mask(SIGCHLD));
	assert(wp->p_state == RUNNING || wp->p_state == STOPPED);
	if (wp->p_state == RUNNING)
		(void) xptrace(XPT_SIGNAL, wp->p_pid, 0, SIGTRAP);
	(void) sigsetmask(oldmask);
	waitfor(wp);
	wp->p_stopinfo = BY_STOP;
	wp->p_signal = 0;
}

/*
 * send a signal to process 'procid' or to all processes if procid
 * is zero.  if sending to all processes, make sure that there are
 * some around.
 */
public sendsig(procid, sig)
	int procid;
	int sig;
{
	Process *p;
	int oldmask;

	if (procid) {
		p = findpid(procid);
		kill(p->p_pid, sig);
	} else {
		oldmask = sigblock(mask(SIGCHLD) | mask(SIGINT));
		for (p = processlist; p; p = p->p_next)
			if (p->p_state == STOPPED
			|| p->p_state == RUNNING
			|| p->p_state == PENDING)
				break;
		if (p)
			(void) xptrace(XPT_SIGNAL, 0, 0, sig);
		(void) sigsetmask(oldmask);
		if (p == 0)
			error("no processes to send signal to");
	}
}

/*
 * If process 'procid' is stopped, release control for it.
 */
public release(procid)
	int procid;
{
	Process *p;
	int oldmask;

	p = findpid(procid);
	if (p->p_state != STOPPED)
		error("process must be stopped to release control");
#ifndef XPT_UNDEBUG
	error("ptrace support for XPT_UNDEBUG not available in this kernel");
#else
	(void) xptrace(XPT_UNDEBUG, p->p_pid, 1, 0);
#endif XPT_UNDEBUG
	pushproc(p);
	delbpts(p->p_pid);
	popproc(p);
	delete_process(p);
}

/*
 * evaluate some actions for all processes.  used by translate to turn
 * on traces.
 */
public evalall(procid, s, actions)
	int procid;
	Symbol s;
	Cmdlist actions;
{
	register Process *p;
	int oldhandler;

	if (procid == ALLPROCS) {
		oldhandler = sigblock(mask(SIGCHLD));
		for (p = processlist; p; p = p->p_next) {
			if (p->p_state == STOPPED && p->p_obj == wp->p_obj) {
				pushproc(p);
				evalevent(s, actions);
				popproc();
			}
		}
		(void) sigsetmask(oldhandler);
	} else {
		assert(procid > 0);
		p = find_procid(procid);
		if (p == nil)
			panic("evalall: can't find %%%d", procid);
		if (p->p_state == STOPPED) {
			pushpid(procid);
			evalevent(s, actions);
			popproc();
		}
	}
}

/*
 * execute 'func' for each process.  func should be fast, because
 * child events are blocked.  it's not necessarily safe to call this
 * for any function that should be applied to all processes.
 */
public doallp(func)
	int (*func)();
{
	int oldmask;
	Process *p;
	oldmask = sigblock(mask(SIGCHLD)|mask(SIGINT));
	for (p = processlist; p; p = p->p_next) {
		pushproc(p);
		(*func)();
		popproc();
	}
	(void) sigsetmask(oldmask);
}


/*
 * this should only be called at exit because it turns off the SIGCHLD
 * handler.
 */
public killkids()
{
	(void) signal(SIGCHLD, SIG_IGN);
	(void) ptrace(XPT_SIGNAL, 0, 0, 9);
}

/*
 * Initialize the argument list.
 */

public arginit()
{
    infile = nil;
    outfile = nil;
    argc = 0;
    if (!Oflag) {
	argv[argc] = primobj->o_pathname;
	argc++;
    }
}

/*
 * Add an argument to the list for the debuggee.
 */

public newarg(arg)
String arg;
{
    String path;
    if (argc >= MAXNCMDARGS) {
	error("too many arguments");
    }
    if (argc == 0) {
	path = objpath(arg);
	if (path == nil) {
		error("no program named %s", arg);
	}
	arg = path;
    }
    argv[argc++] = arg;
}

/*
 * Set the standard input for the debuggee.
 */

public inarg(filename)
String filename;
{
    if (infile != nil) {
	error("multiple input redirects");
    }
    infile = filename;
}

/*
 * Set the standard output for the debuggee.
 * Probably should check to avoid overwriting an existing file.
 */

public outarg(filename)
String filename;
{
    if (outfile != nil) {
	error("multiple output redirect");
    }
    outfile = filename;
}

/*
 * Return from execution when the given signal is pending.
 */

public psigtrace(sig, sw)
	int sig;
	Boolean sw;
{
	if (sig < NSIG) {
		if (sw)
			default_sigmask &= ~mask(sig);
		else
			default_sigmask |= mask(sig);
	} else {
		unsigned int bit;
		switch(sig) {
		case PTS_FORK:
			bit = PF_CATCH_FORK;
			break;
		case PTS_EXEC:
			bit = PF_CATCH_EXEC;
			break;
		case PTS_EXIT:
			bit = PF_CATCH_EXIT;
			break;
		default:
			error("illegal signal number");
		}
		if (sw)
			default_flags |= bit;
		else
			default_flags &= ~bit;
	}
}

/*
 * Print all signals being ignored by the debugger.
 * These signals are automatically
 * passed on to the debugged process.
 */

public printsigsignored ()
{
    if ((default_flags & PF_CATCH_FORK) == 0)
	printf("FORK ");
    if ((default_flags & PF_CATCH_EXEC) == 0)
	printf("EXEC ");
    if ((default_flags & PF_CATCH_EXIT) == 0)
	printf("EXIT ");
    printsigs(default_sigmask);
}

/*
 * Print all signals being intercepted by the debugger
 */

public printsigscaught()
{
    if (default_flags & PF_CATCH_FORK)
	printf("FORK ");
    if (default_flags & PF_CATCH_EXEC)
	printf("EXEC ");
    if (default_flags & PF_CATCH_EXIT)
	printf("EXIT ");
    printsigs(~default_sigmask);
}

/*
 * Run a process that dbx won't interfere with.
 */
public runalone(name, argv, infile, outfile, bg)
	String name;	/* name of command to run */
	String argv[];	/* command args */
	String infile;	/* input file name, 0 ==> dbx's stdin */
	String outfile; /* output file name, 0 ==> dbx's stdout */
	boolean bg;	/* run command in background? */
{
	int savecurproc = 0;
	char **cp;

	assert(strcmp(name, argv[0]) == 0); /* XXX delete name arg later? */
	savecurproc = curprocess ? curprocess->p_id : 0;
	argc = 0;
	for(cp = argv; *cp; cp++)
		argc++;

	create(argv, infile, outfile);
	wp->p_quiet = true;
	wp->p_sigmask = IGNORESIGS;
	xptrace(XPT_SETSIGMASK, wp->p_pid, 0, wp->p_sigmask);
	wp->p_flags &= ~(PF_CATCH_FORK|PF_CATCH_EXEC|PF_CATCH_EXIT|PF_SETFLAGS);
	wp->p_flags |= (PF_DELETE_EXIT|PF_NO_BPTS);

	cont(0, !bg);

	if (savecurproc)
		setproc(savecurproc);
	return(0);
}

/*
 * make a process for the given object if there isn't one already.
 */
public makeprocess(obj)
	Object *obj;
{
	register Process *p;
	int oldmask;
	oldmask = sigblock(mask(SIGCHLD) | mask(SIGINT));
	for (p = processlist; p && p->p_obj != obj; p = p->p_next)
		;
	(void) sigsetmask(oldmask);
	if (p == 0) {
		arginit();
		newarg(obj->o_pathname);
		create(argv, 0, 0);
		wp->p_state = IMAGE;
		set_default_source_file();
	}
}


/*
 * change file descriptors 0, 1, and 2 in process with id 'id' to 
 * in, out, and err, respectively.  return id of process.  if id
 * is zero, return id of current process.
 */
public int moveio(id, in, out, err)
	int id;			/* id of process to move io for */
	int in, out, err;	/* new file descriptors */
{
	Process *p;

	if (wp == 0)
		error("No current process");
	if (id == 0)
		return(wp->p_id);
	p = findpid(id);
	if (p->p_state == STOPPED) {
		xptrace(XPT_DUPFD, p->p_pid, 0, in);
		xptrace(XPT_DUPFD, p->p_pid, 1, out);
		xptrace(XPT_DUPFD, p->p_pid, 2, err);
		p->p_job = dupjob(wp->p_job);
	}
	return(p->p_id);
}


/*
 * Some functions for testing the state of the process.
 */

public Boolean notstarted()
{
    return (Boolean) (wp->p_state == IMAGE);
}

public Boolean isfinished()
{
    return (Boolean) (wp->p_state == EXITED);
}

public Boolean intside()
{
	return (Boolean) (md_state == INTSIDE);
}


/*
 * save/restore "globals"
 */

/*
 * in each module that has globals, define a structure that contains them,
 * a pointer to the currently active such structure, and a global that contains
 * the size of the structure.  all references to globals are based on that
 * pointer.
 */
	
struct globaldefs {
	pointer *g_pointerloc;
	int	*g_sizeloc;
};

private struct globaldefs globals[] = {
	(pointer *) &events_globals_ptr,	&events_globals_size,
	(pointer *) &machine_globals_ptr,	&machine_globals_size,
	(pointer *) &runtime_globals_ptr,	&runtime_globals_size,
	(pointer *) &source_globals_ptr,	&source_globals_size,
	(pointer *) &symbols_globals_ptr,	&symbols_globals_size,
	0,					0
};
private int global_save_size;

/*
 * when multip_init() allocates the process structures, it computes
 * the size of a process structure plus the save area sizes.
 */

allocate_processes()
{
	register struct globaldefs *gp;
	register roundoff;

	global_save_size = 0;
	for (gp = globals; gp->g_sizeloc; gp++) {
		roundoff = *gp->g_sizeloc % sizeof(pointer);
		if (roundoff != 0)
			*gp->g_sizeloc += sizeof(pointer) - roundoff;
		global_save_size += *gp->g_sizeloc;
	}
	if (debug_globals)
	    printf("allocate_processes: computed global_save_size = %d\n",
		global_save_size);
	get_globals((pointer) calloc(1, global_save_size));
	/*
	processtab = (stuct process *) malloc (NPROCS * sizeof(Process));
	global_save_area = (pointer) malloc (NPROCS * global_save_size);
	*/
}

/*
 * when each process is allocated it gets a save area, too.
 */


private Process *process_alloc()
{
	Process *p;
#ifdef notyet
	/* n = index of a free process */
	p = &processtab[n];
	p->p_savearea = savarea + (n * savesize);
	return(p);
#endif
	p = (Process *) malloc (sizeof(Process));		/* XXX */
	p->p_savearea = (pointer) calloc(1, global_save_size);	/* XXX */
	if (debug_globals)
		printf("process_alloc: p->p_savearea = 0x%x\n",
		   p->p_savearea);
	fflush(stdout);
	return(p);
}

/*
 * when processes change, the globals are reloaded.
 */

get_globals(savearea)
	pointer savearea;
{
	register int offset = 0;
	register struct globaldefs *gp;

	if (debug_globals)
		printf("get_globals: copying from savearea at 0x%x\n",
			savearea);

	for (gp = globals; gp->g_sizeloc; gp++) {
		*gp->g_pointerloc = savearea + offset;
		offset += *gp->g_sizeloc;
	}
}

copy_globals(from, to)
	pointer from;
	pointer to;
{
    if (debug_globals)
	printf("copy_globals: copying %d from 0x%x, to savearea 0x%x\n",
	  global_save_size, from, to);
	
	bcopy(from, to, global_save_size);
}

clear_globals(p)
	pointer p;
{
	if (debug_globals)
		printf("clear_globals: clearing %d at 0x%x\n",
			global_save_size, p);
	bzero(p, global_save_size);
}

public String show_cmdstate(st)
	Cmdstate st;
{
	switch(st) {
	case INIT:
		return("INIT");
	case STEP:
		return("STEP");
	case STEPOVER:
		return("STEPOVER");
	case CONT:
		return("CONT");
	default:
		return("?Bad Cmdstate?");
	}
}

public String show_pstate(st)
	Pstate st;
{
	switch(st) {
	case UNKNOWN:
		return("UNKNOWN");
	case IMAGE:
		return("IMAGE");
	case COREDUMP:
		return("COREDUMP");
	case STOPPED:
		return("STOPPED");
	case RUNNING:
		return("RUNNING");
	case PENDING:
		return("PENDING");
	case EXITED:
		return("EXITED");
	case DELETED:
		return("DELETED");
	case PROTO:
		return("PROTO");
	default:
		return("?Bad Pstate?");
	}
}

public String show_pcmd(f)
	int (*f)();
{
	if ((int) f == (int) pcont)
		return("pcont");
	if ((int) f == (int) pstep)
		return("pstep");
	return("?unknown?");
}

public String show_stopinfo(st)
	Stopinfo st;
{
	switch(st) {
	case BY_EVENT:
		return("BY_EVENT");
	case BY_BPT:
		return("BY_BPT");
	case BY_STEP:
		return("BY_STEP");
	case BY_STOP:
		return("BY_STOP");
	case BY_ERROR:
		return("BY_ERROR");
	case BY_CALLRTN:
		return("BY_CALLRTN");
	case BY_RTERR:
		return("BY_RTERR");
	case BY_CMD:
		return("BY_CMD");
	case BY_STOPALL:
		return("BY_STOPALL");
	case BY_RETURN:
		return("BY_RETURN");
	default:
		return("?Bad Stopinfo?");
	}
	/*NOTREACHED*/
}


public String show_signal(n)
	int n;
{
	char buf[40];

	if (n < 0 or n > 32) {
		sprintf(buf, "signal %d", n);
		return(buf);
	} else
		return(signame[n]);
}

show_mask()
{
	register int i, n;
	n = wp->p_sigmask;

	printf("  mask= ");
	if (n == IGNORESIGS) {
		printf("IGNORESIGS\n");
		return;
	} else if (n == 0) {
		printf("0\n");
		return;
	}

	for (i=1; i < (NSIG + 1); i++)
		if (n & mask(i))
			printf("%d ", i);
		
	printf("\n");
}

show_flags()
{
	register unsigned n;

	n = wp->p_flags;

	printf("  flags= ");
	if (n == 0) {
		printf("0\n");
		return;
	}

	if (n & PF_CATCH_FORK)
		printf("C_FORK ");
	if (n & PF_CATCH_EXEC)
		printf("C_EXEC ");
	if (n & PF_CATCH_EXIT)
		printf("C_EXIT ");
	if (n & PF_DELETE_EXIT)
		printf("D_EXIT ");
	if (n & PF_SETFLAGS)
		printf("SETFLAGS ");
	if (n & PF_NO_BPTS)
		printf("NO_BPTS ");
	printf("\n");
}
