/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)names.c	1.5 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: names.c 2.6 87/10/08 $";
#endif

/*
 * Name are the internal representation for identifiers.
 *
 * A hash table is used to map identifiers to names.
 */

#include "defs.h"
#include "names.h"

#ifndef public
typedef struct Name *Name;

/*
 * Inline (for speed) function to return the identifier (string)
 * associated with a name.  Because C cannot support both inlines
 * and data hiding at the same time, the Name structure must be
 * publicly visible.  It is not used explicitly, however, outside of this file.
 */

struct Name {
    char *identifier;
    Name chain;
};

#define ident(n) ((n == nil) ? "(noname)" : n->identifier)
#endif

#define HASHTABLESIZE 2997
private Name nametable[HASHTABLESIZE];

/*
 * Names are allocated in large chunks to avoid calls to malloc
 * and to cluster names in memory so that tracing hash chains
 * doesn't cause many a page fault.
 * Care is taken to make the size of a Namepool close to a power of
 * two (but under!!) so malloc wastes as little space as possible.
 */

#define CHUNKSIZE POOLSIZE(2048, sizeof(struct Name))

typedef struct Namepool {
    struct Name name[CHUNKSIZE];
    struct Namepool *prevpool;
} *Namepool;


private Namepool namepool = nil;
private Integer nleft = 0;

/*
 * Given an identifier, convert it to a name.
 * If it's not in the hash table, then put it there.
 *
 * The second argument specifies whether the string should be copied
 * into newly allocated space if not found.
 *
 * Set "isallocated": 
 *   - true if string is permanently aloocated and need not be done here.
 *   - false if it must be saved by this routine.
 *
 * Pardon my use of goto's, but it seemed more efficient and less convoluted
 * than adding a collection of boolean variables.  This routine is time
 * critical when starting up the debugger on large programs.
 */

public Name identname(s, isallocated)
String s;
Boolean isallocated;
{
    register unsigned h;
    register Char *p, *q;
    register Name n;
    register Integer len;
    Namepool newpool;

    if (s == nil)
	error("internal error: nill name passed to identname");
    h = 0;
    for (p = s; *p != '\0'; p++) {
	h = (h << 1) ^ (*p);
    }
    h = h mod HASHTABLESIZE;
    len = p - s;
    n = nametable[h];
    while (n != nil) {
	p = s;
	q = n->identifier;
	for (;;) {
	    if (*p != *q) {
		goto nomatch;
	    } else if (*p == '\0') {
		goto match;
	    }
	    ++p;
	    ++q;
	}
    nomatch:
	n = n->chain;
    }

    /*
     * Now we know that name hasn't been found (otherwise we'd have jumped
     * down to match), so we allocate a name, store the identifier, and
     * enter it in the hash table.
     */
    if (nleft <= 0) {
	newpool = new(Namepool);
	bzero(newpool, sizeof(newpool));
	newpool->prevpool = namepool;
	namepool = newpool;
	nleft = CHUNKSIZE;
    }
    --nleft;
    n = &(namepool->name[nleft]);
    if (isallocated) {
	n->identifier = s;
    } else {
	n->identifier = (char *)stralloc(len + 1);
	p = s;
	q = n->identifier;
	while (*p != '\0') {
	    *q++ = *p++;
	}
	*q = '\0';
    }
    n->chain = nametable[h];
    nametable[h] = n;

    /*
     * The two possibilities (name known versus unknown) rejoin.
     */
match:
    return n;
}


/*
 * routine for storing short string variables with little space wasted.
 * there is no way to free a string out of this, unless the entire
 * string pool block is freed.
 */

#define STRINGSIZE ((2048 - sizeof(struct Stringpool *)) / sizeof(char))

struct stringpool {
	char chars[STRINGSIZE];
	struct stringpool *prevpool;
};
typedef struct stringpool Stringpool;
private Stringpool *s_pool, *newpool;
private nextchar = STRINGSIZE + 1;		/* force first alloc */

public String stralloc(n)
	int n;
{
	if (nextchar + n > STRINGSIZE) {
		if (n > STRINGSIZE)
			panic("stralloc: string too long!");
		newpool = (Stringpool *)malloc(sizeof(Stringpool));
		newpool->prevpool = s_pool;
		s_pool = newpool;
		nextchar = 0;
	}
	nextchar += n;
	return(&s_pool->chars[nextchar - n]);
}
