/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)object.c	1.16 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: object.c 2.40 88/03/16 $";
#endif

/*
 * Object code interface, mainly for extraction of symbolic information.
 */

#include "defs.h"
#include "library.h"
#include "object.h"
#include "stabstring.h"
#include "main.h"
#include "symbols.h"
#include "languages.h"
#include "mappings.h"
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <ctype.h>
#include <stab.h>

/*
 * Debugger sees some symbols twice -- stab symbols, rcsid, etc.
 * Second time it is assumed that value is already correct.  This was
 * wrong for "shared statics" at one time because the stab symbols
 * were not  being relocated.  Define "WARN_DUP_VALUE" to have procedure
 * "check_local()" make sure that value does not change the second time
 * a symbol is seen.
 */
#define	WARN_DUP_VALUE

#ifndef public

#include <a.out.h>
#include "languages.h"
#include "symbols.h"
#include "lists.h"
#include "mappings.h"
#include "names.h"

/* info about tables */
struct nlhdr {
    unsigned int stringsize;	/* size of the dumped string table */
    unsigned int nsyms;		/* number of symbols */
    unsigned int nfiles;	/* number of files */
    unsigned int nlines;	/* number of lines */
};

/* how to map address to a file */
typedef struct {
    Address begin;
    Address end;
    Address seekaddr;
} Map;

/* value for o_flags */
#define	OBJ_NONE	0	/* a.out not found */
#define	OBJ_STRIPPED	1	/* a.out is stripped */
#define OBJ_NODASHG	2	/* a.out not compiled with -g */
#define OBJ_HASDASHG	3	/* a.out compiled with -g */

struct object {

	String	o_name;		/* name of object file */
	String	o_pathname;	/* pathname of object file */
	unsigned o_flags;	/* what is known about this object */
	long	o_readtime;	/* when did we read object? */
	List	o_srcpath;	/* where to look for source */ 
	struct process *o_proto;/* process prototype for this object */

	/*
	 * tables
	 */

	struct nlhdr o_nlhdr;	/* sizes of tables */
	Symbol	*o_symtab;	/* e.g. for insert(), lookup(), "find" */
	Name	*o_nametab;	/* e.g. for identname() */
	Linetab *o_linetab;	/* e.g. for srcline(), objaddr(), etc */
	Filetab *o_filetab;	/* e.g. for srcfilename(), objaddr() */
	AddrOfFunc *o_functab;	/* e.g. for whatblock() */
	int	o_nfuncs;	/* #funcs in functab */

	/*
	 * per-object symbols (see symbols.c)
	 */
	Symbol	o_program;
	Symbol	o_dbsubc;
	Symbol	o_sigcode;

	/*
	 * breakpoints
	 */

	List	o_bplist;	/* breakpoints */
	List	o_eachinst;	/* per instruction traces */
	List	o_eachline;	/* per source line traces */

	/*
	 * coredump
	 */

	Map	o_datamap;	/* for coredump_readdata(), etc */
	Map	o_stkmap;	/* for coredump_readdata(), etc */
	struct exec o_hdr;	/* header from a.out */

	/*
	 * administration
	 */

	struct object *o_next;
};

#define nlhdr		(curobj->o_nlhdr)
#define objsize		(curobj->o_hdr.a_text)

typedef struct object Object;
Object *curobj;			/* current object file we're working with */


#endif

public List objectlist;

#ifndef N_MOD2
#    define N_MOD2 0x50
#endif

private char dbg_ext[] = ".dbg";

public Language curlang;
public Symbol curmodule;
public Symbol curparam;
public Symbol curcomm;
public Symbol commchain;
public int	longest_objname = 0;
public boolean multi_objs;		/* more than one a.out? */
public Object *primobj;			/* object to use by default */
public boolean readingsyms;		/* true while reading symbolic info */
public boolean has_mplib;		/* compiled with multi-proc lib */
public Boolean strip_ = false;		/* seen fortran, strip final underscore */
public Address pas_globals_addr;	/* address of //GLOBALS// (pascal) */

private char *stringtab;
private struct nlist *namelist;
private struct nlist *curnp;
private Boolean warned;

public String curfilename ()
{
    return ((filep-1)->filename);
}

/*
 * Blocks are figured out on the fly while reading the symbol table.
 */

#define MAXBLKDEPTH 25

public Symbol curblock;

private Symbol blkstack[MAXBLKDEPTH];
private integer curlevel;
private integer nesting;
private Address addrstk[MAXBLKDEPTH];

public pushBlock (b)
Symbol b;
{
    if (curlevel >= MAXBLKDEPTH) {
	fatal("nesting depth too large (%d)", curlevel);
    }
    blkstack[curlevel] = curblock;
    ++curlevel;
    curblock = b;
    if (traceblocks) {
	printf("entering block %s\n", symname(b));
    }
}

/*
 * Change the current block with saving the previous one,
 * since it is assumed that the symbol for the current one is to be deleted.
 */

public changeBlock (b)
Symbol b;
{
    curblock = b;
}

public enterblock (b)
Symbol b;
{
    if (curblock == nil) {
	b->level = 1;
    } else {
	b->level = curblock->level + 1;
    }
    b->block = curblock;
    pushBlock(b);
}

public exitblock ()
{
    if (curblock->class == FUNC or curblock->class == PROC) {
	curblock->symvalue.funcv.src = true;
    }
    if (curlevel <= 0) {
	panic("nesting depth underflow (%d)", curlevel);
    }
    --curlevel;
    if (traceblocks) {
	printf("exiting block %s\n", symname(curblock));
    }
    curblock = blkstack[curlevel];
}

/*
 * Enter a source line or file name reference into the appropriate table.
 * Expanded inline to reduce procedure calls.
 *
 * private enterline (linenumber, address)
 * Lineno linenumber;
 * Address address;
 *  ...
 */

#define enterline(linenumber, address) \
{ \
    register Linetab *lp; \
 \
    lp = linep - 1; \
    if (linenumber != lp->line) { \
	if (address != lp->addr) { \
	    ++lp; \
	} \
	lp->line = linenumber; \
	lp->addr = address; \
	linep = lp + 1; \
    } \
}

/*
 * Read in the namelist from the obj file.
 *
 * Reads and seeks are used instead of fread's and fseek's
 * for efficiency sake; there's a lot of data being read here.
 */

public readobj (file, ignore_stabs)
String file;
Boolean ignore_stabs;
{
    Fileid f;
    struct nlist nlist;
    char *dbg_file_name;
    Object *savecurobj, *o, *saveprimobj;
    struct exec hdr;
    struct stat s;
    Symbol sf;

    nosyserrs();

    if (access(file, (R_OK | X_OK)) == -1 || stat(file, &s) == -1) {
	syserrs();
	beginerrmsg();
	perror(file);
	erecover();
    }
    o = create_object(file);
    if (o->o_flags != OBJ_NONE && o->o_readtime > s.st_mtime) {
	syserrs();
	error("%s is already read in\n", file);
    }

    f = open(file, 0, 0);
    if (f < 0) {
	syserrs();
	beginerrmsg();
	perror(file);
	erecover();
    }

    syserrs();

    read(f, &hdr, sizeof(hdr));
    if (N_BADMAG(hdr)) {
	close(f);
	error("%s is not an object file", file);
    }
    if (hdr.a_magic == OMAGIC) {
	syserrs();
	close(f);
	error("%s is a .o file, cannot exec", file);
    }
    if (hdr.a_magic == SMAGIC) {
	syserrs();
	close(f);
	error("%s is a standalone file, cannot exec", file);
    }

    (void) time(&o->o_readtime);
    savecurobj = curobj;
    saveprimobj = primobj;
    curobj = o;
    if (primobj == nil)
	primobj = o;
    curobj->o_hdr = hdr;
    if (o->o_proto == nil)
	    protoproc(o);

    if (ignore_stabs)
	printf("reading object information ");
    else
        printf("reading symbolic information ");
    if (multi_objs)
	printf("for %s ", file);
    printf("... ");
    fflush(stdout);
    readingsyms = true;

    if (seterror()) {
	bzero(&curobj->o_nlhdr, sizeof(curobj->o_nlhdr));
	readingsyms = false;
	curobj = savecurobj;
	primobj = saveprimobj;
	o->o_flags = OBJ_NONE;
	close(f);
	if (!ignore_stabs)
	    dbg_close();
	erecover();
    }

    symbols_init();
    mappings_init();
    defregnames();

    objsize = hdr.a_text;
    nlhdr.nsyms = hdr.a_syms / sizeof(nlist);
    if (nlhdr.nsyms == 0) {
	warning("program \"%s\" is stripped -- no symbols available", curobj->o_name);
	curobj->o_flags = OBJ_STRIPPED;
	initsyms(hdr.a_entry);
    }
    if (nlhdr.nsyms != 0) {
  	lseek(f, (long) N_STROFF(hdr), 0);
  	read(f, &(nlhdr.stringsize), sizeof(nlhdr.stringsize));
	if (nlhdr.stringsize == 0) {
	    warning("cannot read string table for \"%s\" -- no symbols available",
		curobj->o_name);
	    curobj->o_flags = OBJ_STRIPPED;
	initsyms(hdr.a_entry);
	}
    }
    if (nlhdr.nsyms != 0 and nlhdr.stringsize != 0) {
  	nlhdr.stringsize -= 4;
  	stringtab = newarr(char, nlhdr.stringsize);
  	read(f, stringtab, nlhdr.stringsize);
	if (!ignore_stabs) {
	    dbg_file_name = (char *)malloc(strlen(file) + sizeof(dbg_ext));
	    strcpy(dbg_file_name, file);
	    strcat(dbg_file_name, dbg_ext);
	    dbg_open(dbg_file_name);
	}
	nlhdr.nlines = 8192;		/* 64k worth of line table */
	nlhdr.nfiles = 85;		/* 1k worth of file table */
	allocmaps(nlhdr.nfiles, nlhdr.nlines);
	lseek(f, (long) N_SYMOFF(hdr), 0);
	initsyms(hdr.a_entry);
	reallocmaps(nlhdr.nsyms);	/* make room for a.out lines & files */
	readsyms(f, ignore_stabs);
	if (!ignore_stabs)
	    dbg_close();
	dbsubc = lookup(identname("_dbsubc", true));	   /* see walkstack() */
	sigcode_sym = lookup(identname("_sigcode", true)); /* see nextframe() */
	if (!has_mplib)
		check_mplib();
	/*
	 * the following four calls must be done in this order!
	 */
	setnlines();	/* fixes up function table, too */
	ordfunctab();
	remdupfuncs();	/* whatblock() should work after this */
	setnfiles();
	curobj->o_flags = dbsubc ? OBJ_HASDASHG : OBJ_NODASHG;
	if (nlhdr.nlines == 0 && !ignore_stabs)
	    warning("no source files were compiled with `-g'");
	checkfiles(s.st_mtime);
    }

    bpinit();

    unseterror();
    curobj = savecurobj;
    close(f);

    printf("\n");
    readingsyms = false;
    fflush(stdout);
}

/*
 * Found the beginning of the externals in the object file
 * (signified by the "-lg" or find an external), close the
 * block for the last procedure.
 */

private foundglobals ()
{
    if (curblock != nil && curblock->class != PROG) {
	exitblock();
	if (curblock != nil && curblock->class != PROG) {
	    exitblock();
	}
    }
    if (linep > linetab)
	enterline(0, (linep-1)->addr + 1);
}

# ifdef DEBUG

private struct ntl {
	Name       n;
	struct ntl *next;
} *trap_name_list = nil;

private check_for_trap(name)
String name;
{
	struct ntl *np;
	Name this_one;

	if (name[0] == '_' and name[1] != '\0')
		this_one = identname(&name[1], false);
	else
		this_one = identname(&name[0], false);
	np = trap_name_list;
	while (np) {
		if (np->n == this_one) {
			trap_name();
			return;
		} else
			np = np->next;
	}

}

private read_trap_names()
{
	File f;
	char buf[80], *p;
	Name tn;
	struct ntl *x;

	if ((f = fopen("dbxtrapname", "r")) == NULL)
		return;
	while (fgets(buf, 80, f) != NULL) {
		p = buf;
		while (*p && *p != '\n')
			p++;
		*p = '\0';
		if (buf[0] == '\n')
			continue;
		tn = identname(buf, false);
		x = (struct ntl *) malloc(sizeof(struct ntl));
		x->n = tn;
		x->next = trap_name_list;
		trap_name_list = x;
	}
	fclose(f);
}

/* Routine that we can set a breakpoint on to trace specific symbol names. */

private trap_name()
{
}

# endif DEBUG

/*
 * Read in symbols from object file.
 */

private readsyms (f, ignore_stabs)
Fileid f;
register Boolean ignore_stabs;
{
    register struct nlist *np;
    register String name;
    register Boolean afterlg;
    register integer index;
    struct nlist *ub;
    char *lastchar;
    Boolean could_be_common;

# ifdef DEBUG
	read_trap_names();
# endif
    namelist = newarr(struct nlist, nlhdr.nsyms);
    read(f, namelist, nlhdr.nsyms * sizeof(struct nlist));
    afterlg = false;
    ub = &namelist[nlhdr.nsyms];
    curnp = &namelist[0];
    np = curnp;
    while (np < ub) {
	index = np->n_un.n_strx;
	if (index != 0) {
	    name = &stringtab[index - 4];
	    /*
             *  If the program contains any .f files a trailing _ is stripped
       	     *  from the name on the assumption it was added by the compiler.
	     *  This only affects names that follow the sdb N_SO entry with
             *  the .f name. 
	     *  We need to know if this could be a common name.  That is
	     *  determined by the final character.  Information is needed in
	     *  "check_global()" and "check_var()", but lost here after we
	     *  strip final char.  Fortran "save" variables are implemented
	     *  just like common, but the common name is take from the
	     *  module name with "=" on the front.
             */
	    could_be_common = false;
	    if (name[0] != '\0') {
# if DEBUG
		if (trap_name_list != nil)
		    check_for_trap(name);
# endif
		lastchar = &name[strlen(name) -1];
		if (*lastchar == '_' or *lastchar == '/' or name[0] == '=')
		    could_be_common = true;
		if (afterlg or ((np->n_type & N_EXT) == N_EXT) ) {
		    if (strip_ and not streq(name, "__") and not streq(name, "_ _"))
		        if (*lastchar == '_')
			    *lastchar = '\0';
		}
	    }
	} else {
	    name = nil;
	} 

	/*
	 * Assumptions:
	 *	not an N_STAB	==> name != nil
	 *	name[0] == '-'	==> name == "-lg"
	 *	name[0] != '_'	==> filename or invisible
	 *
	 * The "-lg" signals the beginning of global loader symbols.
         *
	 */
	if ((np->n_type&N_STAB) != 0) {
	    if (!ignore_stabs)
	        enter_nl(name, np);
	} else if (name[0] == '-') {
	    afterlg = true;
	    foundglobals();
#ifdef NOTDEF
	} else if ((np->n_type&N_FN) == N_FN) {
	    check_filename(name, np->n_value, ignore_stabs);
#endif NOTDEF
	} else if (name[0] == '$' && is_pascal_internal(name)) {
	    fix_pascal_internal(name, np->n_value);
	} else if (afterlg) {
	    check_global(name, np, could_be_common);
	} else if ((np->n_type&N_EXT) == N_EXT) {
	    afterlg = true;
	    foundglobals();
	    check_global(name, np, could_be_common);
	} else if (name[0] == '_') {
	    check_local(&name[1], np);
	} else if ((np->n_type&N_TEXT) == N_TEXT) {
	    check_filename(name, np->n_value, ignore_stabs);
	}
	++curnp;
	np = curnp;
    }
    dispose(namelist);
}

/*
 * Get a continuation entry from the name list.
 * Return the beginning of the name.
 */

public String getcont ()
{
    register integer index;
    register String name;

    ++curnp;
    index = curnp->n_un.n_strx;
    if (index == 0) {
	panic("continuation followed by empty stab");
    }
    name = &stringtab[index - 4];
    return name;
}

/*
 * Initialize symbol information.
 */

private initsyms (program_start_addr)
    Address program_start_addr;
{
    curblock = nil;
    curlevel = 0;
    nesting = 0;
    program = insert(identname("", true));
    program->class = PROG;
    program->language = primlang;
    program->symvalue.funcv.beginaddr = program_start_addr;
    program->symvalue.funcv.inline = false;
    newfunc(program, codeloc(program));	/* XXX int side malloc */
    findbeginning(program);
    enterblock(program);
    curmodule = program;
}


/*
 * Enter a namelist entry.
 */

private enter_nl (name, np)
String name;
register struct nlist *np;
{
    register Name n;

    switch (np->n_type) {
	/*
	 * Build a symbol for the FORTRAN common area.  All GSYMS that follow
	 * will be chained in a list with the head kept in common.offset, and
	 * the tail in common.chain.
	 */
	case N_BCOMM:
 	    if (curcomm) {
		curcomm->symvalue.common.chain = commchain;
	    }
	    n = identname(name, true);
	    curcomm = lookup(n);
	    if (curcomm == nil) {
		curcomm = insert(n);
		curcomm->class = COMMON;
		curcomm->block = curblock;
		curcomm->level = program->level;
		curcomm->symvalue.common.chain = nil;
	    }
	    commchain = curcomm->symvalue.common.chain;
	    break;

	case N_ECOMM:
	    if (curcomm) {
		curcomm->symvalue.common.chain = commchain;
		curcomm = nil;
	    }
	    break;

	case N_LBRAC:
	    ++nesting;
	    addrstk[nesting] = (linep - 1)->addr;
	    break;

	case N_RBRAC:
	    --nesting;
	    if (addrstk[nesting] == NOADDR) {
		exitblock();
		newfunc(curblock, (linep - 1)->addr);
		addrstk[nesting] = (linep - 1)->addr;
	    }
	    break;

	case N_SLINE:
	    enterline((Lineno) np->n_desc, (Address) np->n_value);
	    break;

	/*
	 * Source files.
	 */
	case N_SO:
	    n = identname(name, true);
	    enterSourceModule(n, (Address) np->n_value);
	    break;

	/*
	 * Textually included files.
	 */
	case N_SOL:
# ifdef DEBUG
	    if (name == nil)
		warning("nil name in N_SOL");
# endif DEBUG
	    if (name != nil)
	        enterfile(name, (Address) np->n_value);
	    break;

	/*
	 * These symbols are assumed to have non-nil names.
	 */
	case N_GSYM:
	case N_FUN:
	case N_STSYM:
	case N_SSTSYM:		/* shared static */
	case N_LCSYM:
	case N_SLCSYM:		/* shared lcomm */
	case N_RSYM:
	case N_PSYM:
	case N_LSYM:
	case N_SSYM:
	case N_LENG:
	    if (index(name, ':') == nil) {
		if (not warned) {
		    warned = true;
		    warning("old style symbol information found in \"%s\"",
			curfilename());
		}
	    } else {
		entersym(name, np);
	    }
	    break;

	case N_PC:
	case N_MOD2:
	    break;

	default:
	    printf("warning:  stab entry unrecognized: ");
	    if (name != nil) {
		printf("name %s,", name);
	    }
	    printf("ntype %2x, desc %x, value %x'\n",
		np->n_type, np->n_desc, np->n_value);
	    break;
    }
}


/*
 * fix up the address of a 'non-visible' pascal procedure.  these are
 * symbols that are nested deep enough not to be in the symbol table.
 * they are kept in the symbol table with a symbol that looks something
 * like "$1000000".  multiple sources may cause duplicates, so we must
 * ensure that we're updating the correct 'module'.
 */
private fix_pascal_internal(name, addr)
String name;
Address addr;
{
    register Symbol s;
    Name n;

    n = identname(name, true);
    do {
	find(s, n) where
	    (
		s->level == program->level and
		s->class == EXTREF and
		s->block == curmodule
	    )
	endfind(s);
	if (s != nil) {
	    s->symvalue.extref->symvalue.funcv.beginaddr = addr;
	    delete(s);
	}
    } while(s != nil);
}

private is_pascal_internal(name)
    char *name;
{
    register char *cp = name;
    register char c;

    cp++; 	/* skip '$' */
    for (; cp < &name[7]; cp++) {
	c = *cp;
	if (c < '0' || c > '9')
	    return(0);
    }
    return(1);
}

/*
 * Find a symbol with the given name and address;
 * This is used to see if we already have a symbol entered at the given
 * address.  This happens with Pascal procedures, and since the .dbg file
 * gives us the address we do a search here for them.  The goal is to prevent
 * creation of duplicate sumbols.  This would be done by the call to deffunc()
 * in routine check_global().
 */

private Symbol find_addr_sym(n, a)
Name n;
Address a;
{
	register Symbol s;

	find (s, n) where
	    (
	      (s->class == VAR or s->class == PROC or s->class == FUNC) and
	      (s->symvalue.funcv.beginaddr == a)
	    )
	endfind(s);

	return(s);
}

/*
 * Try to find the symbol that is referred to by the given name.  Since it's
 * an external, we need to follow a level or two of indirection.
 */

private Symbol findsym (n, var_isextref)
Name n;
boolean *var_isextref;
{
    register Symbol r, s;

    *var_isextref = false;
    find(s, n) where
	s->level == program->level and s->class == EXTREF
    endfind(s);
    if (s == nil) {
        find(s, n) where
	    (
	        s->level == program->level and (
		    s->class == VAR or s->class == PROC or s->class == FUNC
	        )
	    ) or (
	        s->block == program and s->class == MODULE
	    )
        endfind(s);
    }
    if (s == nil) {
	r = nil;
    } else if (s->class == EXTREF) {
	*var_isextref = true;
	r = s->symvalue.extref;
	delete(s);

	/*
	 * Now check for another level of indirection that could come from
	 * a forward reference in procedure nesting information.  In this case
	 * the symbol has already been deleted.
	 */
	if (r != nil and r->class == EXTREF) {
	    r = r->symvalue.extref;
	}
/*
    } else if (s->class == MODULE) {
	s->class = FUNC;
	s->level = program->level;
	r = s;
 */
    } else {
	r = s;
    }
    return r;
}

/*
 * Create a symbol for a text symbol with no source information.
 * We treat it as an assembly language function.
 */

private Symbol deffunc (n)
Name n;
{
    Symbol f;

    f = insert(n);
    f->language = findlanguage(".s");
    f->class = FUNC;
    f->type = t_int;
    f->block = curblock;
    f->level = program->level;
    f->symvalue.funcv.src = false;
    f->symvalue.funcv.inline = false;
    return f;
}

/*
 * Create a symbol for a data or bss symbol with no source information.
 * We treat it as an assembly language variable.
 */

private Symbol defvar (n)
Name n;
{
    Symbol v;

    v = insert(n);
    v->language = findlanguage(".s");
    v->class = VAR;
    v->type = t_int;
    v->level = program->level;
    v->block = curblock;
    return v;
}

/*
 * Update a symbol entry with a text address.
 */

private updateTextSym (s, name, addr)
Symbol s;
char *name;
Address addr;
{
    if (s->class == VAR) {
	if (s->language != findlanguage(".f") and
	    s->language != findlanguage(".p")     ) {
		if (s->symvalue.offset != 0 and s->symvalue.offset != addr) {
		    warning("re-updating variable %s, old offset 0x%x, new 0x%x",
		       symname(s), s->symvalue.offset, addr);
		}
		s->symvalue.offset = addr;
	}
    } else {
	if (s->symvalue.funcv.beginaddr != 0 and s->symvalue.funcv.beginaddr != addr) {
	    warning("re-updating function %s, old beginaddr 0x%x, new 0x%x",
		symname(s), s->symvalue.funcv.beginaddr, addr);
	}
	s->symvalue.funcv.beginaddr = addr;
	if ((name[0] == '_' || name[0] == '%')
	&& (s->symvalue.funcv.isentry == false || s->chain->symvalue.entryv.entryid == 0)) {
	    newfunc(s, codeloc(s));
	    findbeginning(s);
	}
    }
}

/*
 * Check to see if a global _name is already in the symbol table,
 * if not then insert it.
 */

private check_global (name, np, could_be_common)
String name;
register struct nlist *np;
Boolean could_be_common;
{
    register Name n;
    register Symbol t, u;
    boolean isextref;
    integer count;

    if (not streq(name, "_end")) {
	if (name[0] == '_') {
	    n = identname(&name[1], true);
	} else {
	    n = identname(name, true);
	    /*
	    if (lookup(n) != nil && name[0] != '$') {
		sprintf(buf, "$%s", name);
		n = identname(buf, false);
	    }
	    */
	}
	if ((np->n_type&N_TYPE) == N_TEXT) {
	    count = 0;
	    t = findsym(n, &isextref);
	    while (isextref) {
		++count;
		updateTextSym(t, name, np->n_value);
		t = findsym(n, &isextref);
	    }
	    if (count == 0) {
		if (t == nil) {
		    t = find_addr_sym(n, np->n_value);
		    if ( t == nil)
		        t = deffunc(n);
		    updateTextSym(t, name, np->n_value);
		    if (tracesyms) {
			printdecl(t);
		    }
		} else {
		    u = find_addr_sym(n, np->n_value);
		    if (t->class == MODULE and u == nil) {
			u = t;
			t = deffunc(n);
			t->block = u;
			if (tracesyms) {
			    printdecl(t);
			}
		        updateTextSym(t, name, np->n_value);
		    } else if (u != nil) {
		        updateTextSym(u, name, np->n_value);
		    } else {
		        updateTextSym(t, name, np->n_value);
		    }
		}
	    }
	} else if ((np->n_type&N_TYPE&(~N_SHARED)) == N_BSS) {
#ifdef i386
	    if (streq(name, "//GLOBALS//"))
		pas_globals_addr = np->n_value;
#endif i386
	    t = nil;
	    if (could_be_common) {
	        find(t, n) where
		    t->class == COMMON
	        endfind(t);
	    }
	    if (t != nil) {
		if (t->symvalue.common.fixed_addr) {
		    fprintf(stderr, "\nwarning: variable name conflict for common '%s'\n",
		        ident(n));
		    fprintf(stderr, "unable to resolve addresses for variable and common.\n");
		} else
		    t->symvalue.common.fixed_addr = true;
		fixcommonvars(t, np->n_value);
            } else {
		check_var(np, n, could_be_common);
	    }
        } else {
	    check_var(np, n, could_be_common);
	}
    }
}

/*
 * in the symbol tables, the address of a variable in a common is 
 * represented only by its offset in the common.  now that we know
 * the address of the common, add its address to the
 * offset of each of its variables.
 */
private fixcommonvars(t, common_addr)
	Symbol t;
	Address common_addr;
{
	register Symbol u;

	u = (Symbol) t->symvalue.common.offset;
	while (u != nil) {
	    u->symvalue.offset = u->symvalue.common.offset + common_addr;
# ifdef notyet
	    if (u->symvalue.common.fixed_addr) {
		fprintf(stderr, "internal error: fixcommonvars(), duplicate symbol update\n");
		fprintf(stderr, "symbol '%s', common '%s'\n", symname(u), symname(t));
	    }
# endif notyet
	    u->symvalue.common.fixed_addr = true;
	    u = u->symvalue.common.chain;
	}
}

/*
 * Check to see if a namelist entry refers to a variable.
 * If not, create a variable for the entry.  In any case,
 * set the offset of the variable according to the value field
 * in the entry.
 *
 * If the external name has been referred to by several other symbols,
 * we must update each of them.
 *
 * 26-Oct-87 DEM: In the code again, fix for mailbug #03752.  Note the use
 * of variable "conflict".  If we find a module BEFORE we find a var (last
 * if) then we set conflict to nil and don't define a variable.  But if
 * we find the module AFTER finding the variable then we rewrite "conflict"
 * and will create a duplicate variable.  This is a wrong.  Fixed with the
 * introduction of flag "found_one" to prevent creating a duplicate variable.
 * Uncertain if this should be done for the other matches.
 */

private check_var (np, n, could_be_common)
struct nlist *np;
register Name n;
Boolean could_be_common;
{
    register Symbol t, u, next;
    Symbol conflict;
    Boolean found_one;

    t = lookup(n);
    if (t == nil) {
	t = defvar(n);
	t->symvalue.offset = np->n_value;
	if (tracesyms) {
	    printdecl(t);
	}
    } else {
	conflict = nil;
	found_one = false;
	do {
	    next = t->next_sym;
	    if (t->name == n) {
		if (t->class == MODULE and t->block == program) {
		    conflict = t;
		} else if (t->class == EXTREF and t->level == program->level) {
		    u = t->symvalue.extref;
		    while (u != nil and u->class == EXTREF) {
			u = u->symvalue.extref;
		    }
		    if (u->class == COMMON and could_be_common)
			fixcommonvars(u, np->n_value);
		    else
			u->symvalue.offset = np->n_value;
		    delete(t);
		    found_one = true;
		/*
		 * Added next else 1-Mar-87   DEM
		 * Have to update common variable addresses for variables
		 * created in docommon.  If extname and internal name match
		 * the symbols have class COMMON instead of EXTREF.
		 */
		} else if (t->class == COMMON and could_be_common) {
		    fixcommonvars(t, np->n_value);
		    found_one = true;
		/* End of added statement.  DEM */
		} else if (t->level == program->level and
		    (t->class == VAR or t->class == PROC or t->class == FUNC)
		) {
		    conflict = nil;
		    t->symvalue.offset = np->n_value;
		    found_one = true;
		}
	    }
	    t = next;
	} while (t != nil);
	if (conflict != nil and not found_one) {
	    u = defvar(n);
	    u->block = conflict;
	    u->symvalue.offset = np->n_value;
	}
    }
}

/*
 * Check to see if a local _name is known in the current scope.
 * If not then enter it.  Note that .dbg files may cause local
 * procedures to be defined before we find them in the a.out.  So
 * before entering them here we check for a match of name and address.
 */

private check_local (name, np)
String name;
register struct nlist *np;
{
    register Name n;
    register Symbol t, cur;

    n = identname(name, true);

    /* Pascal local procedures have been defined in the .dbg files. */
    t = find_addr_sym(n, np->n_value);
    if (t != nil)
	return;

    cur = ((np->n_type&N_TYPE) == N_TEXT) ? curmodule : curblock;
    find(t, n) where t->block == cur endfind(t);
    if (t == nil) {
	t = insert(n);
	t->language = findlanguage(".s");
	t->type = t_int;
	t->block = cur;
	t->level = cur->level;
	if ((np->n_type&N_TYPE) == N_TEXT) {
	    t->class = FUNC;
	    t->symvalue.funcv.src = false;
	    t->symvalue.funcv.inline = false;
	    t->symvalue.funcv.beginaddr = np->n_value;
	    newfunc(t, codeloc(t));
	    findbeginning(t);
	} else if ((np->n_type&N_TYPE) == N_BSS) {
	    t->class = VAR;
	    t->symvalue.offset = np->n_value;
	    t->block = curblock;
	    t->level = program->level;
	} else {
	    t->class = VAR;
	    t->symvalue.offset = np->n_value;
	}
    }
#ifdef WARN_DUP_VALUE    
      else if ( (t->symvalue.offset != np->n_value) && (t->class == VAR) &&
		((np->n_type&N_FN) != N_FN) && (np->n_value != 0) ) {
	    if (name != nil)
		printf("Warning: updating address of symbol '%s'\n", name);
	    else
		printf("Warning: updating address of symbol - no name\n");
	    printf("old address = 0x%x, new addr = 0x%x\n",
		t->symvalue.offset, np->n_value);
	    t->symvalue.offset = np->n_value;
    }
#endif WARN_DUP_VALUE
}

/*
 * Check to see if a symbol corresponds to a object file name.
 * For some reason these are listed as in the text segment.
 */

private check_filename (name, mod_offset, ignore_stabs)
String name;
int mod_offset;
Boolean ignore_stabs;
{
    register String mname;
    register integer i;
    Name n;
    Symbol s;

    mname = strdup(name);
    i = strlen(mname) - 2;
    if (i >= 0 and mname[i] == '.' and mname[i+1] == 'o') {
	mname[i] = '\0';
	--i;
	while (mname[i] != '/' and i >= 0) {
	    --i;
	}
	n = identname(&mname[i+1], true);
	find(s, n) where s->block == program and s->class == MODULE endfind(s);
	if (s == nil) {
	    s = insert(n);
	    s->language = findlanguage(".s");
	    s->class = MODULE;
	    s->symvalue.funcv.beginaddr = 0;
	    findbeginning(s);
	}
	if (curblock->class != PROG) {
	    exitblock();
	    if (curblock->class != PROG) {
		exitblock();
	    }
	}
	enterblock(s);
	curmodule = s;

	if (!ignore_stabs)
	    dbg_read(n, mod_offset);
    }
}

/*
 * Check to see if a symbol is about to be defined within an unnamed block.
 * If this happens, we create a procedure for the unnamed block, make it
 * "inline" so that tracebacks don't associate an activation record with it,
 * and enter it into the function table so that it will be detected
 * by "whatblock".
 */

public chkUnnamedBlock ()
{
    register Symbol s;
    static int bnum = 0;
    char buf[100];
    Address startaddr;

    if (nesting > 0 and addrstk[nesting] != NOADDR) {
	startaddr = (linep - 1)->addr;
	++bnum;
	sprintf(buf, "$b%d", bnum);
	s = insert(identname(buf, false));
	s->language = curlang;
	s->class = PROC;
	s->symvalue.funcv.src = false;
	s->symvalue.funcv.inline = true;
	s->symvalue.funcv.beginaddr = startaddr;
	enterblock(s);
	newfunc(s, startaddr);
	addrstk[nesting] = NOADDR;
    }
}

/*
 * Compilation unit.  C associates scope with filenames
 * so we treat them as "modules".  The filename without
 * the suffix is used for the module name.
 *
 * Because there is no explicit "end-of-block" mark in
 * the object file, we must exit blocks for the current
 * procedure and module.
 */

public enterSourceModule (n, addr)
Name n;
Address addr;
{
    register Symbol s;
    Name nn;
    String mname, suffix;

    mname = strdup(ident(n));
    if (rindex(mname, '/') != nil) {
	mname = rindex(mname, '/') + 1;
    }
    suffix = rindex(mname, '.');
    curlang = findlanguage(suffix);
    if (curlang == findlanguage(".f")) {
	strip_ = true;
    } 
    if (suffix != nil) {
	*suffix = '\0';
    }
    if (not (*language_op(curlang, L_HASMODULES))()) {
	if (curblock->class != PROG) {
	    exitblock();
	    if (curblock->class != PROG) {
		exitblock();
	    }
	}
	nn = identname(mname, true);
	if (curmodule == nil or curmodule->name != nn) {
	    s = insert(nn);
	    s->class = MODULE;
	    s->symvalue.funcv.beginaddr = codeloc(curmodule);
	} else {
	    s = curmodule;
	    s->symvalue.funcv.beginaddr = addr;
	    findbeginning(s);
	}
	s->language = curlang;
	enterblock(s);
	curmodule = s;
    }
    if (program->language == nil) {
	program->language = curlang;
    }
    warned = false;
    enterfile(ident(n), addr);
    initTypeTable();
}

/*
 * Allocate file and line tables and initialize indices.
 */

private allocmaps (nf, nl)
integer nf, nl;
{
    filetab = newarr(Filetab, nf);
    linetab = newarr(Linetab, nl);
    filep = filetab;
    linep = linetab;
    endfiletab = &filetab[nf];
    endlinetab = &linetab[nl];
}

private reallocmaps(howmanymore)
	long howmanymore;
{
	int index;

	if ((linep - linetab) + howmanymore > nlhdr.nlines) {
	    index = linep - linetab;
	    nlhdr.nlines += howmanymore;
	    linetab = (Linetab *)realloc(linetab, nlhdr.nlines * sizeof(Linetab));
	    linep = &linetab[index];
	}
	if ((filep - filetab) + howmanymore > nlhdr.nfiles) {
	    index = filep - filetab;
	    nlhdr.nfiles += howmanymore;
	    filetab = (Filetab *)realloc(filetab, nlhdr.nfiles * sizeof(Filetab));
	    filep = &filetab[index];
	}
}

/*
 * Add a file to the file table.
 *
 * If the new address is the same as the previous file address
 * this routine used to not enter the file, but this caused some
 * problems so it has been removed.  It's not clear that this in
 * turn may not also cause a problem.
 */

static String unknownfile = "?file?";

public enterfile (filename, addr)
String filename;
Address addr;
{
    if (filep == endfiletab)
	realloc_filetab();
    filep->addr = addr;
    if (filename == nil) {
	warning("enterfile: null file name");
	filep->filename = unknownfile;
    } else
        filep->filename = filename;
    filep->lineindex = linep - linetab;
    ++filep;
}

private int badlineindex;

/*
 * Since we only estimated the number of lines (and it was a poor
 * estimation) and since we need to know the exact number of lines
 * to do a binary search, we set it when we're done.
 */

private setnlines ()
{
    register Linetab *lp, *llp;
#ifdef STATS
    if (not testing)
	printf("Linetab: %d alloc, %d used, %d%%.\n",
	    nlhdr.nlines, linep - linetab, (linep - linetab) * 100 / nlhdr.nlines);
#endif
    nlhdr.nlines = linep - linetab;
    fixlines();
    if (linep[-1].line == 0)
	linep[-1].addr = linep[-2].addr + 1;

    badlineindex = 0;
    llp = linetab;
    for (lp = &linetab[1]; lp < &linetab[nlhdr.nlines - 1]; lp++, llp++) {
	if (lp->addr <= llp->addr) {
	    badlineindex = lp - linetab;
	    break;
	}
    }
}

/*
 * Similarly for nfiles ...
 */

private setnfiles ()
{
    register Filetab *f;
    register Symbol s;
    register Language lang;

#ifdef STATS
    if (not testing)
	printf("Filetab: %d alloc, %d used, %d%%.\n",
	    nlhdr.nfiles, filep - filetab, (filep - filetab) * 100 / nlhdr.nfiles);
#endif

    nlhdr.nfiles = filep - filetab;

    /*
     * if the line table is out of order (see setnlines), find the file
     * and linenumber where it went bad, and issue a warning.
     */

    if (badlineindex != 0) {
	Filetab *best = filetab;
	int span;
        for (f = filetab; f < &filetab[nlhdr.nfiles]; f++) {
	    span = badlineindex - f->lineindex;
	    if (span > 0 && span < badlineindex - best->lineindex)
		best = f;
	}
	warning("line table out of order at line %d in file \"%s\"",
		linetab[badlineindex].line,
		best->filename);
    }

}

public makeprocs()
{
    register Object *op;
    foreach(Object *, op, objectlist)
	if (op->o_flags != OBJ_NONE)
	    makeprocess(op);
    endfor
}

/*
 * find the object structure for the object file given by the
 * pathname.  if there isn't one, create it.  for newly created
 * objects, add the path prefix to the source path search list.
 */
public Object *create_object(pathname)
	String pathname;		/* pathname of object */
{
	register Object *op;
	register String objname;	/* file name of object */
	register String pathcopy;
	String path;
	Object *save;
	int len;
	boolean found = false;

	pathcopy = strdup(pathname);
	objname = rindex(pathcopy, '/');
	if (objname) {
		*objname = '\0';
		objname++;
	} else {
		objname = pathcopy;
	}

	len = strlen(objname);
	if (len > longest_objname)
		longest_objname = len;

	if (objectlist == nil)
		objectlist = list_alloc();

	foreach(Object *, op, objectlist)
		if (streq(op->o_name, objname)) {
			found = true;
			break;
		}
	endfor

	if (not found) {
		save = curobj;
		op = (Object *)calloc(1, sizeof(Object));     /* XXX int side */
		curobj = op;
		op->o_name = objname;
		op->o_pathname = strdup(pathname);
		op->o_flags = OBJ_NONE;
		op->o_srcpath = list_alloc();
	        if (objname != pathcopy)
			sourcepath_append(pathcopy);
		foreach(String, path, genericsrcpath)
			sourcepath_append(path);
			list_delete(list_curitem(genericsrcpath), genericsrcpath);
		endfor
		sourcepath_append(".");
		op->o_program = newSymbol(identname("?", true), 0, PROG, 0, 0);
		list_append(list_item(op), nil, objectlist);
		curobj = save;
	}

	if (list_size(objectlist) > 1)
		multi_objs = true;

	return(op);
}

/*
 * find the object with name 'name', or return nil.
 */
public Object *find_object(name)
	String name;
{
	register String objname;
	register Object *op;

	if (name == nil)
		return(nil);
	objname = rindex(name, '/');
	objname = objname ? objname + 1 : name; 
	foreach(Object *, op, objectlist)
		if (streq(op->o_name, objname))
			return(op);
	endfor
	return(nil);
}

/*
 * find the object referred to by name.  search for the entire path
 * name first, then just the file name.
 */
public String objpath(name)
    String name;
{
    register Object *op;

    foreach(Object *, op, objectlist)
	if (streq(op->o_pathname, name))
	    return(op->o_pathname);
    endfor

    foreach(Object *, op, objectlist)
	if (streq(op->o_name, name))
	    return(op->o_pathname);
    endfor

    return(nil);
}

/*
 * return the object with containing source file 'filename', and path
 * 'pathname', if given.  check current object first.
 */
public Object *pathobj(pathname, filename)
String pathname;
String filename;
{
    register Object *o;
    register String path;

    if (pathname == nil) {
	foreach(String, path, curobj->o_srcpath)
	    if (havefile(filename, curobj))
		return(curobj);
	endfor
    }

    foreach(Object *, o, objectlist)
	foreach(String, path, o->o_srcpath)
	    if ((pathname == nil or streq(pathname, path))
	    and havefile(filename, o))
		return(o);
	endfor
    endfor
    return(nil);
}

/*
 * We want to make sure that nobody has fussed with the program since
 * it was compiled.  One problem that comes up is on networked systems
 * where the times can get out of sync.  We give a little grace,
 * more when testing.  Assumption is that by the time this code sees
 * heavy use Berkeley and Sun will have arrived at some common scheme
 * for keeping the times close and we will have adopted some method.
 */

public checkobj(o)
    Object *o;
{
	struct stat s;
	long delta;

	if (testing) {
		if (varIsSet("$notimedelta"))
			delta = 0;
		else
			delta = 60L;
	} else
		delta = 10L;

	if (o->o_flags == OBJ_HASDASHG || o->o_flags == OBJ_NODASHG) {
		stat(o->o_pathname, &s);
		if (o->o_readtime < (s.st_mtime - delta) )
			warning("program \"%s\" has been modified -- symbolic information may be out of date", o->o_name);
	}
}

/*
 * See if we were invoked as pdbx (not dbx) and were linked with the
 * multi-proc library.  We assume this library was used if we can find
 * one of the symbols "_m_fork" or "m_fork".
 */

/*
 * This code being backed out for the 3.0 Alpha release due to some problems
 * discovered.  It appears that "check_mplib()" is called AFTER
 * "process_init()" where the flag "has_mplib" is used.  Also, the flag
 * is set "false" below, assume it should be set "true".
 *
 * Modified for V3.0.9 to ignore SEGV when m_fork is found.  Also
 * issues a message that this is being done.
 */

static boolean have_warned = false;

private check_mplib()
{
	register Symbol mp_proc;
	Boolean found_it = false;

	if (! is_pdbx)
		return;
	if (curobj == nil)
		return;
	
	mp_proc = lookup(identname("_m_fork", true));	/* C lib */

	if (mp_proc)
		found_it = true;
	else {
		mp_proc = lookup(identname("m_fork", true));	/* Pascal lib */
		if (mp_proc)
			found_it = true;
	}

	if (found_it and not have_warned) {
	    printf("\n[found symbol m_fork, assuming parallel library used, ignoring SEGV]\n");
	    have_warned = true;
	    psigtrace(SIGSEGV, false);
	}
}

#ifdef NOTDEF
/*
 * issue a warning message if the module 's' wasn't compiled with -g.
 * it would be nice to do this, but it's hard to decide when to call it.
 */
checkdashgmod(s)
Symbol s;
{
    if (s != nil and s->class == MODULE and s->language == findlanguage(".s")
    and not streq("crt0", symname(s))) {
	warning("\"%s.o\" was not compiled with `-g'", symname(s));
    }
}
#endif NOTDEF


#ifdef DUMP_SYM_INFO

/*
 * show_sym(lg, s_name, val, type)
 *
 * This prints information about a symbol.
 */

private show_sym(lg, s_name, val, type)
boolean lg;		/* after "-lg" loader global symbols */
char *s_name;		/* symbol name */
unsigned val;		/* symbol value */
int type;		/* symbol type */
{
	int ty;

	if (lg)
		printf("* ");
	else
		printf("  ");
	if (s_name == nil)
		s_name = "{none}";
	printf("Name= '%s', value= 0x%x, type= 0x%x = ", s_name, val, type);
	ty = type & N_TYPE;
	if (ty              == N_UNDF)   printf("N_UNDF ");
	if ((ty & N_ABS)    == N_ABS)    printf("N_ABS ");
	if ((ty & N_TEXT)   == N_TEXT)   printf("N_TEXT ");
	if ((ty & N_DATA)   == N_DATA)   printf("N_DATA ");
	if ((ty & N_BSS)    == N_BSS)    printf("N_BSS ");
	if ((ty & N_COMM)   == N_COMM)   printf("N_COMM ");
	if ((ty & N_FN)     == N_FN)     printf("N_FN ");
	if ((ty & N_SHARED) == N_SHARED) printf("N_SHARED ");
	if ((ty & N_SHUNDF) == N_SHUNDF) printf("N_SHUNDF ");
	if ((ty & N_SHDATA) == N_SHDATA) printf("N_SHDATA ");
	if ((ty & N_SHBSS)  == N_SHBSS)  printf("N_SHBSS ");
	if ((ty & N_SHCOMM) == N_SHCOMM) printf("N_SHCOMM ");
	if (type & N_EXT)                printf("N_EXT ");
	if (type & N_STAB)               printf("N_STAB ");
	printf("\n");
}

#endif DUMP_SYM_INFO
