/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)printsym.c	1.14 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: printsym.c 2.23 87/11/04 $";
#endif

/*
 * Printing of symbolic information.
 */

#include <setjmp.h>
#include <signal.h>
#include "defs.h"
#include "symbols.h"
#include "object.h"
#include "languages.h"
#include "printsym.h"
#include "tree.h"
#include "eval.h"
#include "mappings.h"
#include "runtime.h"
#include "machine.h"
#include "names.h"
#include "keywords.h"
#include "main.h"

#ifndef public
#endif

/*
 * Maximum number of arguments to a function.
 * This is used as a check for the possibility that the stack has been
 * overwritten and therefore a saved argument pointer might indicate
 * to an absurdly large number of arguments.
 */

#define MAXARGSPASSED 20

/*
 * Return a pointer to the string for the name of the class that
 * the given symbol belongs to.
 */

private String clname[] = {
    "bad use", "constant", "type", "variable", "array", "@dynarray",
    "@subarray", "fileptr", "record", "field",
    "procedure", "function", "funcvar",
    "ref", "pointer", "file", "set", "range", "label", "withptr",
    "scalar", "string", "program", "improper", "variant",
    "procparam", "funcparam", "module", "tag", "common", "extref", "typeref"
};

public String classname(s)
Symbol s;
{
    return clname[ord(s->class)];
}

public String showclass(cl)
Symclass cl;
{
    return clname[ord(cl)];
}

/*
 * Note the entry of the given block, unless it's the main program.
 */

public printentry(s)
Symbol s;
{
    if (s != program) {
	printf("\n");
	printprocid();
	printf("entering %s ", classname(s));
	printname(stdout, s);
	printf("\n");
    }
}

/*
 * Note the exit of the given block
 */

public printexit(s)
Symbol s;
{
    if (s != program) {
	printprocid();
	printf("leaving %s ", classname(s));
	printname(stdout, s);
	printf("\n\n");
    }
}

/*
 * Note the call of s from t.
 */

public printcall(s, t)
Symbol s, t;
{
    printf("calling ");
    s = checkentry(s, nil);
    printname(stdout, s);
    printparams(s, nil);
    printf(" from %s ", classname(t));
    printname(stdout, t);
    printf("\n");
}

/*
 * Note the return from s.  If s is a function, print the value
 * it is returning.  This is somewhat painful, since the function
 * has actually just returned.
 * It would be nice to be able to tell what entry has returned,
 * but since the return has already been done, the marker on the
 * stack is gone.
 */

public printrtn(s)
Symbol s;
{
    register int len;

    printf("returning ");
    if (s->class == FUNC && (!istypename(s->type,"void"))) {
	len = size(s->type);
	if (canpush(len)) {
	    pushretval(s);
	    printval(s->type);
	    putchar(' ');
	} else {
	    printf("(value too large) ");
	}
    }
    printf("from ");
    printname(stdout, s);
    printf("\n");
}

/*
 * Print the values of the parameters of the given procedure or function.
 * The frame distinguishes recursive instances of a procedure.
 *
 * If the procedure or function is internal, the argument count is
 * not valid so we ignore it.
 *
 * Note that Fortran used to have the callee pop args, but in 2.9.1 release
 * (SVS 2.7) the caller pops args like C.  This means that there is an
 * adjust stack instruction to tell debugger how many bytes were pushed.
 * Debugger used to just follow symbol->param chain to print args, but now
 * checks for extra words after listing args.  This means we have to be
 * careful that psize() returns correct number of words used by each argument
 * type.  For instance, character strings for Fortran and Pascal use two
 * words on stack, string address and string lenght.  Had to fix up psize()
 * to check for this type.
 */

public printparams(f, frame)
Symbol f;
Frame frame;
{
    Symbol param;
    Symbol procty;
    int n, m, s;

    /*
     * In some cases we won't have a frame pointer and will pick up
     * incorrect args.  For instance stopped in "kill()" above "main()"
     * on the 386 where system calls don't save a frame pointer.
     * If desired to not display args, the following code can be included.
     * Things usually work out ok though, especially when dealing with
     * C sub-blocks.  So code is commented out for now.
     *
     *
     *# ifdef i386
     *   if ( (frame != nil)           and
     *        (frame->entered == true) and
     *        (frame->no_frameptr == true) ) {
     *           printf("()");
     *           return;
     *   }
     *#endif i386
     */

    n = nargspassed(frame);	/* really number of WORDS passed */
    if (isinternal(f)) {
	n = 0;
    }
    printf("(");
    param = f->chain;
    if (param != nil or n > 0) {
	m = n;
	if (param != nil) {
	    for (;;) {
		s = psize(param) div sizeof(Word);
		if (s == 0) {
		    s = 1;
		}
		m -= s;
		if (showaggrs) {
		    printv(param, frame);
		} else {
		    printparamv(param, frame);
		}
		param = param->chain;
	    if (param == nil) break;
		printf(", ");
	    }
	}
#ifdef ns32000
	/*
	 * The NS32000 compiler pushes an extra argument after the
	 * parameters for procedures that return structures.  We have
	 * to subtract a word to take this into account.
	 */
	if (not f->type == nil) {
	    procty = rtype(f->type);
	    if (procty->class == RECORD or procty->class == VARNT)
	        m -= 1;		/* subtract one word */
	}
#endif ns32000
	if (m > 0) {
	    if (m > MAXARGSPASSED) {
		m = MAXARGSPASSED;
	    }
	    if (f->chain != nil) {
		printf(", ");
	    }
	    for (;;) {
		printf("0x%x", argn(n - m, frame));
		--m;
	    if (m <= 0) break;
		printf(", ");
	    }
	}
    }
    printf(")");
}

/*
 * Test if a symbol should be printed.  We don't print files,
 * for example, simply because there's no good way to do it.
 * The symbol must be within the given function.
 */

public Boolean should_print(s)
Symbol s;
{
    Boolean b;
    register Symbol t;

    switch (s->class) {
	case VAR:
	case FVAR:
	    if (isparam(s)) {
		b = false;
	    } else {
		t = rtype(s->type);
		if (t == nil) {
		    b = false;
		} else {
		    switch (t->class) {
			case FILET:
			case SET:
			case BADUSE:
			    b = false;
			    break;

			default:
			    b = true;
			    break;
		    }
		}
	    }
	    break;

	default:
	    b = false;
	    break;
    }
    return b;
}

/*
 * Print out a parameter value.
 *
 * Since this is intended to be printed on a single line with other information
 * aggregate values are not printed.
 */

public printparamv (p, frame)
Symbol p;
Frame frame;
{
    Symbol t;

    t = rtype(p->type);
    switch (t->class) {
	case ARRAY:
	case DYNARRAY:
	case SUBARRAY:
	    t = rtype(t->type);
	    if (compatible(t, t_char)) {
		printv(p, frame);
	    } else {
		printf("%s = (...)", symname(p));
	    }
	    break;

	case RECORD:
	    printf("%s = (...)", symname(p));
	    break;

	default:
	    printv(p, frame);
	    break;
    }
}

/*
 * Print the name and value of a variable.
 */

public printv(s, frame)
Symbol s;
Frame frame;
{
    Address addr;
    int len;

    if (isambiguous(s) and ismodule(container(s))) {
	printname(stdout, s);
	printf(" = ");
    } else {
	printf("%s = ", symname(s));
    }
    /* Can't print assumed size array -- see fortran_printarray(). */
    if (trap_assumed_size_array(s)) {
	printf("(...)");
	return;
    }
    if (isvarparam(s) and not isopenarray(s)) {
	rpush(address(s, frame), sizeof(Address));
	addr = pop(Address);
    } else if (not isfreg8(s)) {
	addr = address(s, frame);
    }
    len = size(s);
    if (not canpush(len)) {
	printf("*** expression too large ***");
    } else if (isreg(s)) {
	if (isfreg8(s)) {
	    pushregdouble(s);
	} else {
	    push(Address, addr);
	}
	if (isparam(s) and not isentered(frame)) {
	    addr = pop(Address);
	    rpush(addr, len);
	}
	printval(s->type);
    } else {
	rpush(addr, len);
	printval(s->type);
    }
}

/*
 * Print out the name of a symbol.
 */

public printname(f, s)
File f;
Symbol s;
{
    if (s == nil) {
	fprintf(f, "(noname)");
    } else if (s == program) {
	fprintf(f, ".");
    } else if (isredirected() or isambiguous(s)) {
	printwhich(f, s);
    } else {
	fprintf(f, "%s", symname(s));
    }
}

/*
 * Print the fully specified variable that is described by the given identifer.
 */

public printwhich(f, s)
File f;
Symbol s;
{
    printouter(f, container(s));
    fprintf(f, "%s", symname(s));
}

/*
 * Print the fully qualified name of each symbol that has the same name
 * as the given symbol.
 */

public printwhereis(f, s)
File f;
Symbol s;
{
    register Name n;
    register Symbol t;

    checkref(s);
    n = s->name;
    t = lookup(n);
    printwhich(f, t);
    t = t->next_sym;
    while (t != nil) {
	if (t->name == n) {
	    putc(' ', f);
	    printwhich(f, t);
	}
	t = t->next_sym;
    }
    putc('\n', f);
}

private printouter(f, s)
File f;
Symbol s;
{
    Symbol outer;

    if (s != nil) {
	outer = container(s);
	if (outer != nil and outer != program) {
	    printouter(f, outer);
	}
	fprintf(f, "%s.", symname(s));
    }
}

public printdecl(s)
Symbol s;
{
    Language lang;

    checkref(s);
    if (s->language == nil or s->language == primlang) {
	lang = findlanguage(".s");
    } else {
	lang = s->language;
    }
    (*language_op(lang, L_PRINTDECL))(s);
}

/*
 * Show symbols with given name.
 */

showsym(c)
String c;
{
	Name n;
	Symbol s;
	int i = 0;

	n = identname(c, false);
	s = lookup(n);

	if (s == nil)
		printf("No symbols with name '%s'\n", c);
	else {
		while (s != nil) {
			if (s->name == n) {
				i++;
				psym(s);
			}
			s = s->next_sym;
		}
		if (i > 1)
			printf("%d symbols with name '%s\n", i, c);
	}
}

/*
 * Straight dump of symbol information.
 */

public psym(s)
Symbol s;
{
    if (s == nil) {
	printf("NIL\n");
	fflush(stdout);
	return;
    }

    printf("name=%s ", symname(s));

    if (s->language == nil)
	printf("lang=NIL ");
    else
        printf("lang=%s ", language_name(s->language));

    printf("level=%d ", s->level);
    printf("class=%s ", classname(s));
    printf("type=");
    if (s->type != nil and s->type->name != nil) {
	printf("(%s) ", symname(s->type));
    } else {
	printf("%#x ", s->type);
    }
    printf("chain=");
    if (s->chain != nil and s->chain->name != nil) {
	printf("(%s) ", symname(s->chain));
    } else {
	printf("%#x ", s->chain);
    }

    printf("block=");
    if (s->block != nil and s->block->name != nil) {
	 putchar('(');
	 printname(stdout, s->block);
	 printf(") ");
    } else {
	printf("%#x ", s->block);
    }
    printf("module=");
    if (s->module != nil and s->module->name != nil) {
	 putchar('(');
	 printname(stdout, s->module);
	 printf(") ");
    } else if (s->module == nil) {
	printf("(nil)");
    } else {
	printf("%#x ", s->module);
    }
    switch (s->class) {
	case TYPE:
	    printf("size=%d ", size(s));
	    break;

	case VAR:
	case REF:
	    if (s->level == 1 or s->level >= 3) {
		printf("address=0x%x ", s->symvalue.offset);
	    } else {
		printf("offset=%d ", s->symvalue.offset);
	    }
	    printf("size=%d ", size(s));
	    break;

	case RECORD:
	case VARNT:
	    printf("size=%d ", s->symvalue.offset);
	    break;

	case STR:
	    printf("length=%d ", s->symvalue.offset);
	    break;

	case FIELD:
	    printf("offset=%d ", s->symvalue.field.offset);
	    printf("size=%d ", s->symvalue.field.length);
	    break;

	case PROG:
	case PROC:
	case FUNC:
	    printf("address=0x%x ", s->symvalue.funcv.beginaddr);
	    if (isinline(s)) {
		printf("inline ");
	    }
	    if (nosource(s)) {
		printf("no");
	    }
	    printf("source ");
	    break;

	case RANGE:
	    prangetype(s->symvalue.rangev.lowertype);
	    printf(" lower=%d ", s->symvalue.rangev.lower);
	    prangetype(s->symvalue.rangev.uppertype);
	    printf(" upper=%d ", s->symvalue.rangev.upper);
	    break;

	case COMMON:
	    printf(" common: offset= 0x%x, chain= 0x%x",
		s->symvalue.common.offset, s->symvalue.common.chain);
	    if (s->symvalue.common.chain != nil)
		printf(" (%s)", symname(s->symvalue.common.chain));
	    if (s->symvalue.common.fixed_addr)
		printf(", fixed_addr = T");
	    else
		printf(", fixed_addr = F");
	    break;

	default:
	    /* do nothing */
	    break;
    }
    printf("\n");
    fflush(stdout);
}

private prangetype(r)
Rangetype r;
{
    switch (r) {
	case R_CONST:
	    printf("CONST");
	    break;

	case R_ARG:
	    printf("ARG");
	    break;

	case R_TEMP:
	    printf("TEMP");
	    break;

	case R_ADJUST:
	    printf("ADJUST");
	    break;
    }
}

/*
 * Print out the value on top of the stack according to the given type.
 */

public printval(t)
Symbol t;
{
    Symbol s;

    checkref(t);
    if (t->class == TYPEREF) {
	resolveRef(t);
    }
    switch (t->class) {
	case PROC:
	case FUNC:
	case MODULE:
	    s = pop(Symbol);
	    printf("%s", symname(s));
	    break;

	default:
	    if (t->language == nil or t->language == primlang) {
		(*language_op(findlanguage(".c"), L_PRINTVAL))(t);
	    } else {
		(*language_op(t->language, L_PRINTVAL))(t);
	    }
	    break;
    }
}

/*
 * Print out the value of a record, field by field.
 */

public printrecord(s)
Symbol s;
{
    Symbol f;

    if (s->chain == nil) {
	error("record has no fields");
    }
    printf("(");
    sp -= size(s);
    f = s->chain;
    if (f != nil) {
	for (;;) {
	    printfield(f);
	    f = f->chain;
	if (f == nil) break;
	    printf(", ");
	}
    }
    printf(")");
}

/*
 * Print out a field.
 */

private printfield(f)
Symbol f;
{
    Stack *savesp;
    register int off, len;

    printf("%s = ", symname(f));
    savesp = sp;
    off = f->symvalue.field.offset;
    len = f->symvalue.field.length;
    sp += ((off + len + BITSPERBYTE - 1) div BITSPERBYTE);
    printval(f);
    sp = savesp;
}

/*
 * Print out the contents of an array.
 * Haven't quite figured out what the best format is.
 *
 * This is rather inefficient.
 *
 * The "2*elsize" is there since "printval" drops the stack by elsize.
 */

public printarray(a)
Symbol a;
{
    Stack *savesp, *newsp;
    Symbol eltype;
    long elsize;
    String sep;
    Symbol indexedby;
    integer nels;

    savesp = sp;
    sp -= (size(a));
    newsp = sp;
    eltype = rtype(a->type);
    elsize = size(eltype);

    /*
     * compensate for svs aligning odd-sized arrays..
     */
    if (elsize == 1 && a->language == lang_pascal) {
	indexedby = rtype(a->chain);
	nels = indexedby->symvalue.rangev.upper - indexedby->symvalue.rangev.lower + 1;
	if (nels & 1)
	    savesp--;
    }

    printf("(");
    if (eltype->class == RECORD or eltype->class == ARRAY or
      eltype->class == VARNT) {
	sep = "\n";
	putchar('\n');
    } else {
	sep = ", ";
    }
    for (sp += elsize; sp <= savesp; sp += 2*elsize) {
	if (sp - elsize != newsp) {
	    fputs(sep, stdout);
	}
	printval(eltype);
    }
    sp = newsp;
    if (streq(sep, "\n")) {
	putchar('\n');
    }
    printf(")");
}

/*
 * Print out the value of a real number in Pascal notation.
 * This is, unfortunately, different than what one gets
 * from "%g" in printf.
 */

private jmp_buf	illegal_float;

private catch_illegal_float()
{
    longjmp(illegal_float, 1);
}

public prtreal(r)
double r;
{
    extern char *index();
    char buf[256];
    int catch_illegal_float();
    int (*old_fpe_handler)();
    int code;

    old_fpe_handler = signal(SIGFPE, catch_illegal_float);
    code = setjmp(illegal_float);

    if (code == 0) {
	sprintf(buf, "%g", r);
	if (buf[0] == '.') {
	    printf("0%s", buf);
	} else if (buf[0] == '-' and buf[1] == '.') {
	    printf("-0%s", &buf[1]);
	} else {
	    printf("%s", buf);
	}
	if (index(buf, '.') == nil) {
	    printf(".0");
	}
    } else {
	printf("Illegal floating point value");
    }

    signal(SIGFPE, old_fpe_handler);
}

/*
 * Print out a character using ^? notation for unprintables.
 */

public printchar(c)
char c;
{
    if (c == 0) {
	putchar('\\');
	putchar('0');
    } else if (c == '\n') {
	putchar('\\');
	putchar('n');
    } else if (c > 0 and c < ' ') {
	putchar('^');
	putchar(c - 1 + 'A');
    } else if (c >= ' ' && c <= '~') {
	putchar(c);
    } else {
	printf("\\0%o",c);
    }
}

/*
 * Print out a value for a range type (integer, char, or boolean).
 */

public printRangeVal (val, t)
int val;
Symbol t;
{
    if (t == t_boolean->type or istypename(t->type, "boolean")) {
	if ((boolean) val) {
	    printf("true");
	} else {
	    printf("false");
	}
    } else if (t == t_char->type or istypename(t->type, "char")) {
	if (varIsSet("$hexchars")) {
	    printf("0x%lx", val);
	} else {
	    putchar('\'');
	    printchar(val);
	    putchar('\'');
	}
    } else if (varIsSet("$hexints")) {
	printf("0x%lx", val);
    } else if (t->symvalue.rangev.lower >= 0) {
	printf("%lu", val);
    } else {
	printf("%ld", val);
    }
}

/*
 * Print out an enumerated value by finding the corresponding
 * name in the enumeration list.
 */

public printEnum (i, t)
integer i;
Symbol t;
{
    register Symbol e;

    e = t->chain;
    while (e != nil and e->symvalue.constval->value.lcon != i) {
	e = e->chain;
    }
    if (e != nil) {
	printf("%s", symname(e));
    } else {
	printf("%d", i);
    }
}

/*
 * Print out a null-terminated string (pointer to char)
 * starting at the given address.
 */

public printString (addr, quotes)
Address addr;
boolean quotes;
{
    register Address a;
    register integer i;
    register boolean endofstring;
    union {
	char ch[sizeof(Word)];
	int word;
    } u;

    if (varIsSet("$hexstrings")) {
	printf("0x%x", addr);
    } else {
	if (quotes) {
	    putchar('"');
	}
	a = addr;
	endofstring = false;
	while (not endofstring) {
	    dread(&u, a, sizeof(u));
	    i = 0;
	    do {
		if (u.ch[i] == '\0') {
		    endofstring = true;
		} else {
		    printchar(u.ch[i]);
		}
		++i;
	    } while (i < sizeof(Word) and not endofstring);
	    a += sizeof(Word);
	}
	if (quotes) {
	    putchar('"');
	}
    }
}
