/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)source.c	1.12 (Berkeley) 3/1/85";
static char rcsid[] = "$Header: source.c 2.9 88/03/17 $";
#endif

/*
 * Source file management.
 */

#include "defs.h"
#include "source.h"
#include "mappings.h"
#include "machine.h"
#include "keywords.h"
#include "tree.h"
#include "eval.h"
#include "main.h"
#include <sys/file.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef public
typedef int Lineno;

struct source_globals {
	String source_cursource;
	String source_curpath;
	Lineno source_curline;
	Lineno source_cursrcline;
} *source_globals_ptr;
#define	cursource	(source_globals_ptr->source_cursource)
#define	curpath		(source_globals_ptr->source_curpath)
#define	curline		(source_globals_ptr->source_curline)
#define	cursrcline	(source_globals_ptr->source_cursrcline)

#define LASTLINE 0		/* recognized by printlines */

#include "lists.h"
List genericsrcpath;

#include "object.h"
#define	sourcepath	(curobj->o_srcpath)

#endif

public int source_globals_size = sizeof(struct source_globals);

extern char *re_comp();

private int lastlinenum = 0;
private String open_source = nil;
private String open_path = nil;

/*
 * Data structure for indexing source seek addresses by line number.
 *
 * The constraints are:
 *
 *  we want an array so indexing is fast and easy
 *  we don't want to waste space for small files
 *  we don't want an upper bound on # of lines in a file  (but we have one!)
 *  we don't know how many lines there are
 *
 * The solution is a "dirty" hash table.  We have NSLOTS pointers to
 * arrays of NLINESPERSLOT addresses.  To find the source address of
 * a particular line we find the slot, allocate space if necessary,
 * and then find its location within the pointed to array.
 */

typedef long Seekaddr;

#define MAXLINESPERFILE	100000
#define NLINESPERSLOT	500
#define NSLOTS 		MAXLINESPERFILE / NLINESPERSLOT

#define slotno(line)    ((line) div NLINESPERSLOT)
#define index(line)	((line) mod NLINESPERSLOT)
#define slot_alloc()    newarr(Seekaddr, NLINESPERSLOT)
#define srcaddr(line)	seektab[slotno(line)][index(line)]

private File srcfp;
private Seekaddr *seektab[NSLOTS];

/*
 * Determine if the current source file is available.
 */

public boolean canReadSource ()
{
    boolean b;

    if (cursource == nil) {
	b = false;
    } else if (curpath != open_path or cursource != open_source or lastlinenum == 0) {
	skimsource();
	b = (boolean) (lastlinenum != 0);
    } else {
	b = true;
    }
    return b;
}

public checksource()
{
    if (not canReadSource())
	error("no source file");
}

/*
 * Print out the given lines from the source.
 *
 * Note that we may be called from a breakpoint, before grabtty() has
 * been called.  So there is a possibility that CRMOD mapping is off.
 * This is why this routine puts out \r\n instead of just \n.
 */

public printlines(l1, l2, listi)
Lineno l1, l2;
boolean listi;
{
    register int c;
    register Lineno i, lb, ub;
    register File f;

    if (cursource == nil) {
	beginerrmsg();
	fprintf(stderr, "no source file\r\n");
    } else {
	if (curpath != open_path or cursource != open_source or lastlinenum == 0) {
	    skimsource();
	}
	if (lastlinenum == 0) {
	    beginerrmsg();
	    fprintf(stderr, "couldn't read \"%s\"\r\n", cursource);
	} else {
	    lb = (l1 == LASTLINE) ? lastlinenum : l1;
	    ub = (l2 == LASTLINE) ? lastlinenum : l2;
	    if (lb < 1) {
		beginerrmsg();
		fprintf(stderr, "line number must be positive\r\n");
	    } else if (lb > lastlinenum) {
		beginerrmsg();
		if (lastlinenum == 1) {
		    fprintf(stderr, "\"%s\" has only 1 line\r\n", cursource);
		} else {
		    fprintf(stderr, "\"%s\" has only %d lines\r\n",
			cursource, lastlinenum);
		}
	    } else if (ub < lb) {
		beginerrmsg();
		fprintf(stderr, "second number must be greater than first\r\n");
	    } else {
		if (ub > lastlinenum) {
		    ub = lastlinenum;
		}
		f = srcfp;
		fseek(f, srcaddr(lb), 0);
		for (i = lb; i <= ub; i++) {
		    printf("%5d   ", i);
		    while ((c = getc(f)) != '\n') {
			putchar(c);
		    }
		    putchar('\r');
		    putchar('\n');
		    if (listi) {
			Address lowaddr = objaddr(i, cursource);
			Address highaddr = nextline(lowaddr);
			if (lowaddr != NOADDR && highaddr != NOADDR)
				printinst(lowaddr, highaddr - 1);
		    }
		}
		cursrcline = ub + 1;
	    }
	}
    }
}

/*
 * set the source file search path list to 'pathlist' or print
 * the current list.  the first element of 'pathlist' may be the
 * name of the object to set the path search list for.  this is 
 * done here because it's beyond the ability of the parser.
 */
public setsourcepath(pathlist)
List pathlist;
{
    register String dir;
    register Object *o;
    boolean havetopop = false;
    String objname;

    if (list_size(pathlist) == 0) {
	foreach(Object *, o, objectlist)
	    if (list_size(objectlist) > 1)
	        printf("%%%s: ", o->o_name);
	    foreach (String, dir, o->o_srcpath)
	        printf("%s ", dir);
	    endfor
	    printf("\n");
	endfor
    } else {
	objname = list_element(String, list_head(pathlist));
	if (objname[0] == '%') {
	    objname++;
	    o = find_object(objname);
	    if (o == nil)
	        error("no program named %s", objname);
	    pushproc(o->o_proto);
	    havetopop = true;
	    list_delete(list_head(pathlist), pathlist);
	    if (list_size(pathlist) == 0) {
	        foreach (String, dir, curobj->o_srcpath)
	    	    printf("%s ", dir);
	        endfor
	        printf("\n");
	        goto out;
	    }
	}
	foreach (String, dir, sourcepath)
	    list_delete(list_curitem(sourcepath), sourcepath);
	endfor
	sourcepath = pathlist;
	skimsource();
out:
	if (havetopop)
		popproc();
    }
}

public sourcepath_append(pathname)
String pathname;
{
     list_append(list_item(pathname), nil, sourcepath);
}

/*
 * Search the sourcepath for a file.  if 'path' isn't nil, just try to
 * find path/filename.  otherwise search source paths, trying current object
 * first, then all the rest.
 * Side effect: sets foundPath to the search pathname used when file found.
 */

static char fileNameBuf[1024];
private String foundPath;

public String findsource(path, filename)
String path;
String filename;
{
    register String src, dir;
    register Object *op;

    if (filename == nil) {
	src = nil;
    } else if (filename[0] == '/') {
	src = filename;
    } else {

	if (path) {
	    sprintf(fileNameBuf, "%s/%s", path, filename);
	    if (access(fileNameBuf, R_OK) == 0) {
		foundPath = path;
		return(fileNameBuf);
	    }
	    return(nil);
	}

	src = nil;
	foreach (String, dir, sourcepath)
	    sprintf(fileNameBuf, "%s/%s", dir, filename);
	    if (access(fileNameBuf, R_OK) == 0) {
		src = fileNameBuf;
		foundPath = dir;
		goto found;
	    }
	endfor

        foreach(Object *, op, objectlist)
	    foreach (String, dir, op->o_srcpath)
	        sprintf(fileNameBuf, "%s/%s", dir, filename);
	        if (access(fileNameBuf, R_OK) == 0) {
		    src = fileNameBuf;
		    foundPath = dir;
		    goto found;
	        }
	    endfor
        endfor

    }
found:
    return src;
}

/*
 * Open a source file looking in the appropriate places.
 */

private File opensource(path, filename)
String path;
String filename;
{
    String s;
    File f;

    s = findsource(path, filename);
    if (s == nil) {
	f = nil;
    } else {
	f = fopen(s, "r");
    }
    return f;
}

/*
 * Set the current source file.
 */

public setsource(filename)
String filename;
{
    if (filename != nil) {
	cursource = filename;
	cursrcline = 1;
	curpath = nil;
    }
}

public setpath(pathname)
String pathname;
{
    if (pathname != nil) {
	curpath = pathname;
	cursrcline = 1;
    }
}

/*
 * Read the source file getting seek pointers for each line.
 */

private skimsource()
{
    register int c;
    register Seekaddr count;
    register File f;
    register Lineno linenum;
    register Seekaddr lastaddr;
    register int slot;

    f = opensource(curpath, cursource);
    if (f == nil) {
	lastlinenum = 0;
    } else {
	if (open_source != nil) {
	    free_seektab();
	    if (srcfp != nil) {
		fclose(srcfp);
	    }
	}
	open_source = cursource;
	open_path = curpath = foundPath;
	linenum = 0;
	count = 0;
	lastaddr = 0;
	while ((c = getc(f)) != EOF) {
	    ++count;
	    if (c == '\n') {
		slot = slotno(++linenum);
		if (slot >= NSLOTS) {
		    panic("skimsource: too many lines");
		}
		if (seektab[slot] == nil) {
		    seektab[slot] = slot_alloc();
		}
		seektab[slot][index(linenum)] = lastaddr;
		lastaddr = count;
	    }
	}
	lastlinenum = linenum;
	srcfp = f;
    }
}

/*
 * Erase information and release space in the current seektab.
 * This is in preparation for reading in seek pointers for a
 * new file.  It is possible that seek pointers for all files
 * should be kept around, but the current concern is space.
 */

private free_seektab()
{
    register int slot;

    for (slot = 0; slot < NSLOTS; slot++) {
	if (seektab[slot] != nil) {
	    dispose(seektab[slot]);
	}
    }
}

/*
 * Figure out current source position.
 */

public getsrcpos()
{
    String filename;

    curline = srcline(reg(PROGCTR));
    filename = srcfilename(reg(PROGCTR));
    setsource(filename);
    if (curline != 0) {
	cursrcline = curline;
    }
}

public set_default_source_file()
{
    if (cursource == nil and nlhdr.nfiles != 0)
	setsource(filetab[0].filename);
}

/*
 * Print out the current source position.
 */

public printsrcpos()
{
    printf("at line %d", curline);
    if (nlhdr.nfiles > 1) {
	printf(" in file \"%s\"", cursource);
    }
}

/*
 * Print out the current location in the debuggee.
 */

public printloc()
{
	int line;
	char *file;

	printf("in ");
	printname(stdout, curfunc);
	putchar(' ');
	line = srcline(reg(PROGCTR));
	if (line > 0) {
		printf("at line %d", curline);
		file = srcfilename(reg(PROGCTR));
		if (nlhdr.nfiles > 1) {
			printf(" in file \"%s\"", file);
		}
	} else {
		curline = 0;
		if (testing)
			printf("at {0x%x}", reg(PROGCTR));
		else
			printf("at 0x%x", reg(PROGCTR));
	}
}

private boolean isprog(path, prog)
String path;
String prog;
{
    String tail;
    if (streq(path, prog))
	return(true);
    if ((tail = rindex(path, '/')) && streq(tail + 1, prog))
	return(true);
    return(false);
}

#define DEF_EDITOR  "/usr/ucb/vi"

/*
 * Invoke an editor on the given file.  Which editor to use might change
 * installation to installation.  For now, we use "vi".  In any event,
 * the environment variable "EDITOR" overrides any default.
 */

public edit(filename)
String filename;
{
    extern String getenv();
    String ed, src, s;
    Symbol f;
    Address addr;
    char lineno[10];

    ed = getenv("EDITOR");
    if (ed == nil) {
	ed = DEF_EDITOR;
    }
    if (filename == nil)
        src = findsource(curpath, cursource);
    else
        src = findsource(nil, filename);
    if (src == nil) {
	if (filename == nil)
	    error("no source file");
	f = which(identname(filename, true));
	if (not isblock(f)) {
	    error("can't read \"%s\"", filename);
	}
	addr = firstline(f);
	if (addr == NOADDR) {
	    error("no source for \"%s\"", filename);
	}
	src = srcfilename(addr);
	s = findsource(nil, src);
	if (s != nil) {
	    src = s;
	}
	sprintf(lineno, "+%d", srcline(addr));
    } else {
	sprintf(lineno, "+1");
    }
    if (isprog(ed, "vi") or isprog(ed, "ex")) {
	call(ed, 0, 0, lineno, src, nil);
    } else {
	call(ed, 0, 0, src, nil);
    }
}

/*
 * Strip away portions of a given pattern not part of the regular expression.
 */

private String getpattern (pattern)
String pattern;
{
    register char *p, *r;

    p = pattern;
    while (*p == ' ' or *p == '\t') {
	++p;
    }
    r = p;
    while (*p != '\0') {
	++p;
    }
    --p;
    if (*p == '\n') {
	*p = '\0';
	--p;
    }
    if (*p == *r) {
	*p = '\0';
	--p;
    }
    return r + 1;
}

/*
 * Search the current file for a regular expression.
 */

public search (direction, pattern)
char direction;
String pattern;
{
    register String p;
    register File f;
    String re, err;
    Lineno line;
    boolean matched;
    char buf[512];
    
    if (cursource == nil) {
	beginerrmsg();
	fprintf(stderr, "no source file\n");
    } else {
	if (cursource != open_source or lastlinenum == 0) {
	    skimsource();
	}
	if (lastlinenum == 0) {
	    beginerrmsg();
	    fprintf(stderr, "couldn't read \"%s\"\n", cursource);
	} else {
	    re = getpattern(pattern);
	    /* circf = 0; */
	    if (re != nil and *re != '\0') {
		err = re_comp(re);
		if (err != nil) {
		    error(err);
		}
	    }
	    matched = false;
	    f = srcfp;
	    line = cursrcline;
	    do {
		if (direction == '/') {
		    ++line;
		    if (line > lastlinenum) {
			line = 1;
		    }
		} else {
		    --line;
		    if (line < 1) {
			line = lastlinenum;
		    }
		}
		fseek(f, srcaddr(line), L_SET);
		p = buf;
		*p = getc(f);
		while ((*p != '\n') and (*p != EOF)) {
		    ++p;
		    *p = getc(f);
		}
		*p = '\0';
		matched = (boolean) re_exec(buf);
	    } while (not matched and line != cursrcline);
	    if (not matched) {
		beginerrmsg();
		fprintf(stderr, "no match\n");
	    } else {
		printlines(line, line, false);
		cursrcline = line;
	    }
	}
    }
}

/*
 * Compute a small window around the given line.
 */

public getsrcwindow (line, l1, l2)
Lineno line, *l1, *l2;
{
    Node s;
    integer size;

    s = findvar(identname("$listwindow", true));
    if (s == nil) {
	size = 10;
    } else {
	eval(s);
	size = pop(integer);
    }
    *l1 = line - (size div 2);
    if (*l1 < 1) {
	*l1 = 1;
    }
    *l2 = *l1 + size;
}


/*
 * the 'file' command.  argument 'l' is a list with the first item possibly
 * being the name of the image to change (or report) the file in.  the
 * other item is the name of the file to change to.  sorry this is so ugly.
 * really should be done by the parser, but it isn't up to it.
 */
public changefile(l)
List l;
{
    String objname = nil, filename, path;
    Object *o;
    File file;
     
    /*
     * check to see if an object name was explicitly specified, e.g.
     * "file %prog joe.c".
     */
    if (list_size(l) > 1) {
	objname = list_element(String, list_head(l));
	if (objname[0] == '%') {
	    objname++;
	    o = find_object(objname);
	    if (o == nil)
	        error("no program named %s", objname);
	    list_delete(list_head(l), l);
	}
    }

    /*
     * print current file or make sure the one given is available.
     */
    if (list_size(l) == 0) {
	checksource();
	printf("%s\n", cursource);
    } else {
	filename = list_element(String, list_head(l));
	if (objname) {
	    foreach(String, path, o->o_srcpath)
	        file = opensource(path, filename);
		if (file != nil)
		    break;
	    endfor
	} else  {
	    file = opensource(nil, filename);
	}
	if (file == nil) {
	    error("can't read \"%s\"", filename);
	} else {
	    fclose(file);
	    setsource(filename);
	    setpath(foundPath);
	}
    }
}

/*
 * check that files are up to date.
 */
public checkfiles(objtime)
int objtime;
{
    struct stat s;
    Filetab *fp;
    String dir;

    for (fp = filetab; fp < &filetab[nlhdr.nfiles]; fp++) {
	foreach (String, dir, sourcepath)
	    sprintf(fileNameBuf, "%s/%s", dir, fp->filename);
	    if (stat(fileNameBuf, &s) != -1) {
		if (s.st_mtime > objtime)
			warning("file \"%s\" newer than program \"%s\"",
				fileNameBuf, curobj->o_name);
			break;
	    }
	endfor
    }
}
