/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* Copyright (c) 1982 Regents of the University of California */
#ifndef lint
static	char sccsid[] = "@(#)stabstring.c	1.2 (Berkeley) 3/1/85"; /* from 1.4 84/03/27 10:24:04 linton Exp */
static char rcsid[] = "$Header: stabstring.c 2.9 87/11/09 $";
#endif


/*
 * String information interpretation
 *
 * The string part of a stab entry is broken up into name and type information.
 */

#include "defs.h"
#include "object.h"
#include "main.h"
#include "symbols.h"
#include "printsym.h"
#include "stabstring.h"
#include "names.h"
#include "languages.h"
#include "tree.h"
#include <ctype.h>

#ifdef NSC_OBJ
#	include <nlist.h>
#endif

#ifndef public
/*
 * Table of types indexed by per-file unique identification number.
 */

#define NTYPES 1000

Symbol typetable[NTYPES];

#endif

boolean	debug_stab = false;	/* debug reading stab info */

/*
 * Special characters in symbol table information.
 */

#define CONSTNAME 'c'
#define TYPENAME 't'
#define TAGNAME 'T'
#define MODULEBEGIN 'm'
#define EXTPROCEDURE 'P'
#define PRIVPROCEDURE 'Q'
#define INTPROCEDURE 'I'
#define EXTFUNCTION 'F'
#define PRIVFUNCTION 'f'
#define INTFUNCTION 'J'
#define EXTVAR 'G'
#define MODULEVAR 'S'
#define OWNVAR 'V'
#define REGVAR 'r'
#define VALUEPARAM 'p'
#define VARIABLEPARAM 'v'
#define LOCALVAR /* default */

/*
 * Type information special characters.
 */

#define T_SUBRANGE 'r'
#define T_ARRAY 'a'
#define T_OLDOPENARRAY 'A'
#define T_OPENARRAY 'O'
#define T_DYNARRAY 'D'
#define T_SUBARRAY 'E'
#define T_RECORD 's'
#define T_UNION 'u'
#define T_ENUM 'e'
#define T_PTR '*'
#define T_FUNCVAR 'f'
#define T_PROCVAR 'p'
#define T_IMPORTED 'i'
#define T_SET 'S'
#define T_OPAQUE 'o'
#define T_FILE 'd'

# ifdef STABDEBUG
private String xsymname(s)
Symbol s;
{
    if (s == nil)
        return("nil");
    else
        return(symname(s));
}

private String showspecial1(c)
char c;
{
    switch (c) {
    case CONSTNAME:
	return("CONTSNAME");
    case TYPENAME:
	return("TYPENAME");
    case TAGNAME:
	return("TAGNAME");
    case MODULEBEGIN:
	return("MODULEBEGIN");
    case EXTPROCEDURE:
	return("EXTPROCEDURE");
    case PRIVPROCEDURE:
	return("PRIVPROCEDURE");
    case INTPROCEDURE:
	return("INTPROCEDURE");
    case EXTFUNCTION:
	return("EXTFUNCTION");
    case PRIVFUNCTION:
	return("PRIVFUNCTION");
    case INTFUNCTION:
	return("INTFUNCTION");
    case EXTVAR:
	return("EXTVAR");
    case MODULEVAR:
	return("MODULEVAR");
    case OWNVAR:
	return("OWNVAR");
    case REGVAR:
	return("REGVAR");
    case VALUEPARAM:
	return("VALUEPARAM");
    case VARIABLEPARAM:
	return("VARIABLEPARAM");
    default:
	return("devault: LOCALVAR");
    }
}

private String showspecial2(c)
char c;
{
    switch (c) {
    case T_SUBRANGE:
	return("T_SUBRANGE");
    case T_ARRAY:
	return("T_ARRAY");
    case T_OLDOPENARRAY:
	return("T_OLDOPENARRAY");
    case T_DYNARRAY:
	return("T_DYNARRAY");
    case T_SUBARRAY:
	return("T_SUBARRAY");
    case T_RECORD:
	return("T_RECORD");
    case T_UNION:
	return("T_UNION");
    case T_ENUM:
	return("T_ENUM");
    case T_PTR:
	return("T_PRT");
    case T_FUNCVAR:
	return("T_FUNCVAR");
    case T_PROCVAR:
	return("T_PROCVAR");
    case T_IMPORTED:
	return("T_IMPORTED");
    case T_SET:
	return("T_SET");
    case T_OPAQUE:
	return("T_OPAQUE");
    case T_FILE:
	return("T_FILE");
    default:
	return("?unknown type?");
    }
}
# endif

public initTypeTable ()
{
    bzero(typetable, sizeof(typetable));
    (*language_op(curlang, L_MODINIT))(typetable);
# ifdef STABDEBUG
  if (debug_stab) {
    printf("\ninitTypeTable: cleared and initialized type table\n");
  }
# endif
}

/*
 * Put an nlist entry into the symbol table.
 * If it's already there just add the associated information.
 *
 * Type information is encoded in the name following a ":".
 */

private Symbol constype();
private Char *curchar;

#define skipchar(ptr, ch) \
{ \
    if (*ptr != ch) { \
	panic("expected char '%c', found '%s'", ch, ptr); \
    } \
    ++ptr; \
}

#define optchar(ptr, ch) \
{ \
    if (*ptr == ch) { \
	++ptr; \
    } \
}

#define chkcont(ptr) \
{ \
    if (*ptr == '?') { \
	ptr = getcont(); \
    } \
}

#define newSym(s, n) \
{ \
    s = insert(n); \
    s->level = curblock->level + 1; \
    s->language = curlang; \
    s->block = curblock; \
}

#define makeVariable(s, n, off) \
{ \
    newSym(s, n); \
    s->class = VAR; \
    s->symvalue.offset = off; \
    getType(s); \
}

#define makeParameter(s, n, cl, off) \
{ \
    newSym(s, n); \
    s->class = cl; \
    s->symvalue.offset = off; \
    curparam->chain = s; \
    curparam = s; \
    getType(s); \
}

private makeRegParameter(s, n, cl, value)
Symbol s;
Name n;
Symclass cl;
int value;
{
    int reg, off;

# ifdef STABDEBUG
  if (debug_stab) {
    printf("makeRegParameter(s='%s', n='%s', cl='%s', value=0x%x)\n",
	xsymname(s), ident(n), showclass(cl), value);
  }
# endif
    assert(isreg(s));
    reg = s->symvalue.offset;
    off = value;
    s->class = cl;
    curparam->chain = s;
    curparam = s;
    s->symvalue.regparam.paramoff = off;
    s->symvalue.regparam.regnum = reg;
    getType(s);
}

public entersym (name, np)
String name;
struct nlist *np;
{
    Symbol s;
    Symbol ds;
    char *p;
    register Name n;
    char c;

# ifdef STABDEBUG
  if (debug_stab) {
    printf("entersym(name='%s', np)\n", name);
  }
# endif

    p = index(name, ':');
    *p = '\0';
    c = *(p+1);
    /*
     * If we are linking with fortran, AND this is an external procedure,
     * AND it has a trailing underscore, we must delete that underscore.
     *
     * But in readsyms the code only checks "strip_" and if true does the
     * removal of a trailing underscore.  Now working on some variations
     * of this policy.
     */

/*    if (strip_ and c == EXTFUNCTION) {
 *	if ( *(p-1) == '_')
 *	    *(p-1) = '\0';
 *    }
*/
    if (strip_) {
	if ( *(p-1) == '_')
	    *(p-1) = '\0';
    }
    n = identname(name, true);
    chkUnnamedBlock();
    curchar = p + 2;
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  entersym: n='%s', c='%c' = %s\n", ident(n), c, showspecial1(c));
  }
# endif
    switch (c) {
	case CONSTNAME:
	    newSym(s, n);
	    constName(s);
	    break;

	case TYPENAME:
	    newSym(s, n);
	    typeName(s);
	    break;

	case TAGNAME:
	    s = symbol_alloc();
	    s->name = n;
	    s->level = curblock->level + 1;
	    s->language = curlang;
	    s->block = curblock;
	    tagName(s);
	    break;

	case MODULEBEGIN:
	    publicRoutine(&s, n, MODULE, np->n_value, false);
	    curmodule = s;
	    break;

	case EXTPROCEDURE:
	    publicRoutine(&s, n, PROC, np->n_value, false);
	    break;

	case PRIVPROCEDURE:
	    privateRoutine(&s, n, PROC, np->n_value);
	    break;

	case INTPROCEDURE:
	    publicRoutine(&s, n, PROC, np->n_value, true);
	    break;

	case EXTFUNCTION:
	    publicRoutine(&s, n, FUNC, np->n_value, false);
	    break;

	case PRIVFUNCTION:
	    privateRoutine(&s, n, FUNC, np->n_value);
	    break;

	case INTFUNCTION:
	    publicRoutine(&s, n, FUNC, np->n_value, true);
	    break;

	case EXTVAR:
	    extVar(&s, n, np->n_value);
	    break;

	case MODULEVAR:
	    if (curblock->class != MODULE) {
		exitblock();
	    }
	    makeVariable(s, n, np->n_value);
	    s->level = program->level;
	    s->block = curmodule;
# ifdef NOTYET
	    /*
	     * We may have already encountered this symbol and not correctly
	     * defined it.  Now that we have the correct info delete the
	     * bogus symbol.
	     */
	    find (ds, n) where
		(
		  (ds->class == VAR and ds->level == s->level and
		   ds->block == s->block and
		   ds->symvalue.offset == s->symvalue.offset)
		)
	    endfind(ds);
	    if (ds != nil)
		delete(ds);
# endif NOTYET
	    getExtRef(s);
	    break;

	case OWNVAR:
	    makeVariable(s, n, np->n_value);
	    ownVariable(s, np->n_value);
	    getExtRef(s);
	    break;

	case REGVAR:
	    makeVariable(s, n, fixregnumber(np->n_value));
	    s->level = -(s->level);
	    break;

	case VALUEPARAM:
	    s = lookup(n);
	    if (s != nil and s->block == curblock) {
		makeRegParameter(s, n, VAR, np->n_value);
	    } else {
		makeParameter(s, n, VAR, np->n_value);
	    }
	    break;

	case VARIABLEPARAM:
	    makeParameter(s, n, REF, np->n_value);
	    break;

	default:	/* local variable */
	    --curchar;
	    makeVariable(s, n, np->n_value);
	    break;
    }
    if (tracesyms) {
	printdecl(s);
	fflush(stdout);
    }
}

/*
 * Enter a named constant.
 */

private constName (s)
Symbol s;
{
    double d;
    char *p, buf[1000];

# ifdef STABDEBUG
  if (debug_stab) {
    printf("constName(s='%s')  [enter a named constant, s->class = CONST]\n",
      xsymname(s));
  }
# endif
    s->class = CONST;
    skipchar(curchar, '=');
    p = curchar;
    ++curchar;
    switch (*p) {
	case 'b':
	    s->type = t_boolean;
	    s->symvalue.constval = build(O_LCON, getint());
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  constname: 'b', type= t_boolean, contsval = O_LCON\n");
  }
# endif
	    break;

	case 'c':
	    s->type = t_char;
	    s->symvalue.constval = build(O_LCON, getint());
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  constname: 'c', type= t_char, contsval = O_LCON\n");
  }
# endif
	    break;

	case 'i':
	    s->type = t_int;
	    s->symvalue.constval = build(O_LCON, getint());
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  constname: 'i', type= t_int, contsval = O_LCON\n");
  }
# endif
	    break;

	case 'r':
	    sscanf(curchar, "%lf", &d);
	    while (*curchar != '\0' and *curchar != ';') {
		++curchar;
	    }
	    --curchar;
	    s->type = t_real;
	    s->symvalue.constval = build(O_FCON, d);
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  constname: 'r', type= t_real, contsval = O_FCON\n");
  }
# endif
	    break;

	case 's':
	    p = &buf[0];
	    skipchar(curchar, '\'');
	    while (*curchar != '\'') {
		*p = *curchar;
		++p;
		++curchar;
	    }
	    *p = '\0';
	    s->symvalue.constval = build(O_SCON, strdup(buf));
	    s->type = s->symvalue.constval->nodetype;
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  constname: 's', type= <array of char>, contsval = O_SCON\n");
  }
# endif
	    break;

	case 'e':
	    getType(s);
	    skipchar(curchar, ',');
	    s->symvalue.constval = build(O_LCON, getint());
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  constname: 'e', type= <constype(nil)>, contsval = O_LCON\n");
  }
# endif
	    break;

	case 'S':
	    getType(s);
	    skipchar(curchar, ',');
	    (void) getint(); /* set size */
	    skipchar(curchar, ',');
	    (void) getint(); /* number of bits in constant */
	    s->symvalue.constval = build(O_LCON, 0);
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  constname: 'S', type= <constype(nil)>, contsval = O_LCON\n");
  }
# endif
	    break;

	default:
	    s->type = t_int;
	    s->symvalue.constval = build(O_LCON, 0);
	    printf("[internal error: unknown constant type '%c']", *p);
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  constname: ?, type= t, int, contsval = O_LCON\n");
  }
# endif
	    break;
    }
    s->symvalue.constval->nodetype = s->type;
}

/*
 * Enter a type name.
 */

private typeName (s)
Symbol s;
{
    register integer i;

    s->class = TYPE;
    s->language = curlang;
    s->block = curblock;
    s->level = curblock->level + 1;
    i = getint();
# ifdef STABDEBUG
  if (debug_stab) {
    printf("typeName(s='%s')\n", xsymname(s));
    printf("  typeName: class= TYPE, block='%s', level= %d, i= %d\n",
      xsymname(s->block), s->level, i);
  }
# endif
    if (i == 0) {
	panic("bad input on type \"%s\" at \"%s\"", symname(s), curchar);
    } else if (i >= NTYPES) {
	panic("too many types in file \"%s\"", curfilename());
    }
    /*
     * A hack for C typedefs that don't create new types,
     * e.g. typedef unsigned int Hashvalue;
     *  or  typedef struct blah BLAH;
     */
    if (*curchar != '=') {
	s->type = typetable[i];
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  typeName: *curchar != '=', type = typetable[%d]\n", i);
    if (s->type == nil)
      printf("            nil entry, allocating sym, setting recursive type\n");
  }
# endif
	if (s->type == nil) {
	    s->type = symbol_alloc();
	    typetable[i] = s->type;
	}
    } else {
# ifdef STABDEBUG
  if (debug_stab) {
    if (typetable[i] != nil) {
      printf("  typeName: *curchar == '=', typetable[%d] was not nil\n", i);
      printf("            set typetable[%d]->type to symbol '%s'\n", i, xsymname(s));
    } else {
      printf("  typeName: *curchar == '=', typetable[%d] was nil\n", i);
      printf("            set typetable[%d] to symbol '%s'\n", i, xsymname(s));
    }
  }
# endif
	if (typetable[i] != nil) {
	    typetable[i]->language = curlang;
	    typetable[i]->class = TYPE;
	    typetable[i]->type = s;
	} else {
	    typetable[i] = s;
	}
	skipchar(curchar, '=');
	getType(s);
    }
}

/*
 * Enter a tag name.
 */

private tagName (s)
Symbol s;
{
    register integer i;

    s->class = TAG;
    i = getint();
    if (i == 0) {
	panic("bad input on tag \"%s\" at \"%s\"", symname(s), curchar);
    } else if (i >= NTYPES) {
	panic("too many types in file \"%s\"", curfilename());
    }
# ifdef STABDEBUG
  if (debug_stab) {
    printf("tagName(s='%s'), s->class = TAG, type # %d\n",
      xsymname(s), i);
  }
# endif
    if (typetable[i] != nil) {
	typetable[i]->language = curlang;
	typetable[i]->class = TYPE;
	typetable[i]->type = s;
    } else {
	typetable[i] = s;
    }
    skipchar(curchar, '=');
    getType(s);
}

/*
 * Setup a symbol entry for a public procedure or function.
 *
 * If it contains nested procedures, then it may already be defined
 * in the current block as a MODULE.
 */

/*ARGSUSED*/
private publicRoutine (s, n, class, addr, isinternal)
Symbol *s;
Name n;
Symclass class;
Address addr;
boolean isinternal;
{
    Symbol nt, t;

# ifdef STABDEBUG
  if (debug_stab) {
    printf("publicRoutine(s, n='%s', class=%s, addr=0x%x, isinternal='%s)\n",
      ident(n), showclass(class), addr, isinternal ? "T" : "F");
  }
# endif
    newSym(nt, n);
    if (isinternal) {
	markInternal(nt);
    }
    enterRoutine(nt, class);
    find(t, n) where
	t != nt and t->class == MODULE and t->block == nt->block
    endfind(t);
    if (t == nil) {
	t = nt;
    } else {
	t->language = nt->language;
	t->class = nt->class;
	t->type = nt->type;
	t->chain = nt->chain;
	t->symvalue = nt->symvalue;
	nt->class = EXTREF;
	nt->symvalue.extref = t;
	delete(nt);
	curparam = t;
	changeBlock(t);
    }
    if (t->block == program) {
	t->level = program->level;
    } else if (t->class == MODULE) {
	t->level = t->block->level;
    } else if (t->block->class == MODULE) {
	t->level = t->block->block->level;
    } else {
	t->level = t->block->level + 1;
    }
    *s = t;
}

/*
 * Setup a symbol entry for a private procedure or function.
 */

private privateRoutine (s, n, class, addr)
Symbol *s;
Name n;
Symclass class;
Address addr;
{
    Symbol t;
    boolean isnew;

# ifdef STABDEBUG
  if (debug_stab) {
    printf("privateRoutine(s, n='%s', class='%s', addr=0x%x)\n",
      ident(n), showclass(class), addr);
  }
# endif
    find(t, n) where
	t->level == curmodule->level and t->class == class
    endfind(t);
    if (t == nil) {
	isnew = true;
	t = insert(n);
    } else {
	isnew = false;
    }
    t->language = curlang;
    enterRoutine(t, class);
    if (isnew) {
	t->symvalue.funcv.src = false;
	t->symvalue.funcv.inline = false;
	t->symvalue.funcv.beginaddr = addr;
	newfunc(t, codeloc(t));
	findbeginning(t);
    }
    *s = t;
}

/*
 * Set up for beginning a new procedure, function, or module.
 * If it's a function, then read the type.
 *
 * If the next character is a ",", then read the name of the enclosing block.
 * Otherwise assume the previous function, if any, is over, and the current
 * routine is at the same level.
 */

private enterRoutine (s, class)
Symbol s;
Symclass class;
{
# ifdef STABDEBUG
  if (debug_stab) {
    printf("enterRoutine(s='%s', class='%s')\n", xsymname(s), showclass(class));
  }
# endif
    s->class = class;
    if (class == FUNC) {
	getType(s);
    }
    if (s->class != MODULE) {
	getExtRef(s);
    } else if (*curchar == ',') {
	++curchar;
    }
    if (*curchar != '\0') {
	exitblock();
	enterNestedBlock(s);
    } else {
	if (curblock->class == FUNC or curblock->class == PROC) {
	    exitblock();
	}
	if (class == MODULE) {
	    exitblock();
	}
	enterblock(s);
    }
    curparam = s;
}

/*
 * Handling an external variable is tricky, since we might already
 * know it but need to define it's type for other type information
 * in the file.  So just in case we read the type information anyway.
 */

private extVar (symp, n, off)
Symbol *symp;
Name n;
integer off;
{
    Symbol s;

# ifdef STABDEBUG
  if (debug_stab) {
    printf("extVar(symp=, n='%s', off= %d)\n", ident(n), off);
  }
# endif
    find(s, n) where
	s->level == program->level and s->class == VAR
    endfind(s);
    if (s == nil) {
	makeVariable(s, n, off);
	s->level = program->level;
	s->block = curmodule;
	getExtRef(s);
    } else {
	(void) constype(nil);
    }
    *symp = s;
}

/*
 * Check to see if the stab string contains the name of the external
 * reference.  If so, we create a symbol with that name and class EXTREF, and
 * connect it to the given symbol.  This link is created so that when
 * we see the linker symbol we can resolve it to the given symbol.
 */

private getExtRef (s)
Symbol s;
{
    char *p;
    Name n;
    Symbol t;

# ifdef STABDEBUG
  if (debug_stab) {
    printf("getExtRef(s='%s')\n", symname(s));
  }
# endif
    if (*curchar == ',' and *(curchar + 1) != '\0') {
	p = index(curchar + 1, ',');
	*curchar = '\0';
	if (p != nil) {
	    *p = '\0';
	    n = identname(curchar + 1, false);
	    curchar = p + 1;
	} else {
	    n = identname(curchar + 1, true);
	}
	t = insert(n);
	t->language = s->language;
	t->class = EXTREF;
	t->block = program;
	t->level = program->level;
	t->symvalue.extref = s;
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  getExtRef: created EXTREF sym '%s'\n", xsymname(t));
  }
# endif
    }
}

/*
 * Find a block with the given identifier in the given outer block.
 * If not there, then create it.
 */

private Symbol findBlock (id, m)
String id;
Symbol m;
{
    Name n;
    Symbol s;

# ifdef STABDEBUG
  if (debug_stab) {
    printf("findBlock(id='%s', m='%s')\n", id, xsymname(m));
  }
# endif
    n = identname(id, true);
    find(s, n) where s->block == m and isblock(s) endfind(s);
    if (s == nil) {
	s = insert(n);
	s->block = m;
	s->language = curlang;
	s->class = MODULE;
	s->level = m->level + 1;
    }
    return s;
}

/*
 * Enter a nested block.
 * The block within which it is nested is described
 * by "module{:module}[:proc]".
 */

private enterNestedBlock (b)
Symbol b;
{
    register char *p, *q;
    Symbol m;

# ifdef STABDEBUG
  if (debug_stab) {
    printf("enterNestedBlock(b='%s')\n", xsymname(b));
  }
# endif
    q = curchar;
    p = index(q, ':');
    m = program;
    while (p != nil) {
	*p = '\0';
	m = findBlock(q, m);
	q = p + 1;
	p = index(q, ':');
    }
    if (*q != '\0') {
	m = findBlock(q, m);
    }
    b->level = m->level + 1;
    b->block = m;
    pushBlock(b);
}

/*
 * Enter a statically-allocated variable defined within a routine.
 *
 * Global BSS variables are chained together so we can resolve them
 * when the start of common is determined.  The list is kept in order
 * so that f77 can display all vars in a COMMON.
 */

private ownVariable (s, addr)
Symbol s;
Address addr;
{
# ifdef STABDEBUG
  if (debug_stab) {
    printf("ownVariable(s='%s', addr=0x%x)\n", xsymname(s), addr);
  }
# endif
    s->level = 1;
    if (curcomm) {
	if (commchain != nil) {
	    commchain->symvalue.common.chain = s;
	} else {
	    curcomm->symvalue.common.offset = (integer) s;
	}			  
	commchain = s;
	s->symvalue.common.offset = addr;
	s->symvalue.common.chain = nil;
    }
}

/*
 * Get a type from the current stab string for the given symbol.
 */

private getType (s)
Symbol s;
{
# ifdef STABDEBUG
  if (debug_stab) {
    printf("getType(s='%s')  [s->class = %s]\n",
      xsymname(s), showclass(s->class));
  }
# endif
    s->type = constype(nil);
    if (s->class == TAG) {
	addtag(s);
    }
}

/*
 * Construct a type out of a string encoding.
 */

private Rangetype getRangeBoundType();

private Symbol constype (type)
Symbol type;
{
    register Symbol t;
    register integer n;
    char class;
    char *p;

# ifdef STABDEBUG
  if (debug_stab) {
    printf("constype(type='%s')\n", xsymname(type));
  }
# endif
    while (*curchar == '@') {
	p = index(curchar, ';');
	if (p == nil) {
	    fflush(stdout);
	    fprintf(stderr, "missing ';' after type attributes");
	} else {
	    curchar = p + 1;
	}
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  constype: curchar was '@', advanced past ';'\n");
  }
# endif
    }
    if (isdigit(*curchar)) {
	n = getint();
	if (n >= NTYPES) {
	    panic("too many types in file \"%s\"", curfilename());
	}
# ifdef STABDEBUG
  if (debug_stab) {
    printf("  constype: curchar was digit, type # = %d\n", n);
    if (*curchar == '=')
      printf(" constype: curchar == '='\n");
    else
      printf(" constype: curchar != '='\n");
    if (typetable[n] != nil)
      printf("  constype: typetable[%d] not nil, set t to typetable entry\n", n);
    else
      printf("  constype: typetable[%d] was nil, set to t=symbol_alloc\n", n);
  }
# endif
	if (*curchar == '=') {
	    if (typetable[n] != nil) {
		t = typetable[n];
	    } else {
		t = symbol_alloc();
		typetable[n] = t;
	    }
	    ++curchar;
	    constype(t);
	} else {
	    t = typetable[n];
	    if (t == nil) {
		t = symbol_alloc();
		typetable[n] = t;
	    }
	}
    } else {
	if (type == nil) {
	    t = symbol_alloc();
	} else {
	    t = type;
	}
	t->language = curlang;
	t->level = curblock->level + 1;
	t->block = curblock;
	class = *curchar++;
# ifdef STABDEBUG
  if (debug_stab) {
    if (type == nil)
      printf("  constype: curchar was not integer, type nil so t = symbol_alloc\n");
    else
      printf("  constype: curchar was not integer, type not nil so t = type\n");
    printf("  constype: t->level= %d, t->block= '%s'\n",
      t->level, xsymname(t->block));
    printf("  constype: curchar = '%c' = class '%s'\n", class, showspecial2(class));
  }
# endif
	switch (class) {
	    case T_SUBRANGE:
		consSubrange(t);
		break;

	    case T_ARRAY:
		t->class = ARRAY;
		t->chain = constype(nil);
		skipchar(curchar, ';');
		chkcont(curchar);
		t->type = constype(nil);
		break;

	    case T_OLDOPENARRAY:
		t->class = DYNARRAY;
		t->symvalue.ndims = 1;
		t->type = constype(nil);
		t->chain = t_int;
		break;

	    case T_OPENARRAY:
	    case T_DYNARRAY:
		consDynarray(t);
		break;

	    case T_SUBARRAY:
		t->class = SUBARRAY;
		t->symvalue.ndims = getint();
		skipchar(curchar, ',');
		t->type = constype(nil);
		t->chain = t_int;
		break;

	    case T_RECORD:
		consRecord(t, RECORD);
		break;

	    case T_UNION:
		consRecord(t, VARNT);
		break;

	    case T_ENUM:
		consEnum(t);
		break;

	    case T_PTR:
		t->class = PTR;
		t->type = constype(nil);
		break;

	    /*
	     * C function variables are different from Modula-2's.
	     */
	    case T_FUNCVAR:
		t->class = FFUNC;
		t->type = constype(nil);
		if (not streq(language_name(curlang), "c")) {
		    skipchar(curchar, ',');
		    consParamlist(t);
		}
		break;

	    case T_PROCVAR:
		t->class = FPROC;
		consParamlist(t);
		break;

	    case T_IMPORTED:
		consImpType(t);
		break;

	    case T_SET:
		t->class = SET;
		t->type = constype(nil);
		break;

	    case T_OPAQUE:
		consOpaqType(t);
		break;

	    case T_FILE:
		t->class = FILET;
		t->type = constype(nil);
		break;

	    default:
		badcaseval(class);
	}
    }
    return t;
}

/*
 * Construct a subrange type.
 */

private consSubrange (t)
Symbol t;
{
    t->class = RANGE;
    t->type = constype(nil);
    skipchar(curchar, ';');
    chkcont(curchar);
    t->symvalue.rangev.lowertype = getRangeBoundType();
    t->symvalue.rangev.lower = getint();
    skipchar(curchar, ';');
    chkcont(curchar);
    t->symvalue.rangev.uppertype = getRangeBoundType();
    t->symvalue.rangev.upper = getint();
}

/*
 * Figure out the bound type of a range.
 *
 * Some letters indicate a dynamic bound, ie what follows
 * is the offset from the fp which contains the bound; this will
 * need a different encoding when pc a['A'..'Z'] is
 * added; J is a special flag to handle fortran a(*) bounds
 */

private Rangetype getRangeBoundType ()
{
    Rangetype r;

    switch (*curchar) {
	case 'A':
	    r = R_ARG;
	    curchar++;
	    break;

	case 'T':
	    r = R_TEMP;
	    curchar++;
	    break;

	case 'J': 
	    r = R_ADJUST;
	    curchar++;
	    break;

	default:
	    r = R_CONST;
	    break;
    }
    return r;
}

/*
 * Construct a dynamic array descriptor.
 */

private consDynarray (t)
register Symbol t;
{
    t->class = DYNARRAY;
    t->symvalue.ndims = getint();
    skipchar(curchar, ',');
    t->type = constype(nil);
    t->chain = t_int;
}

/*
 * Construct a record or union type.
 */

private consRecord (t, class)
Symbol t;
Symclass class;
{
    register Symbol u;
    register char *cur, *p;
    Name name;
    integer d;

    t->class = class;
    t->symvalue.offset = getint();
    d = curblock->level + 1;
    u = t;
    cur = curchar;
    while (*cur != ';' and *cur != '\0') {
	p = index(cur, ':');
	if (p == nil) {
	    panic("index(\"%s\", ':') failed", curchar);
	}
	*p = '\0';
	name = identname(cur, true);
	u->chain = newSymbol(name, d, FIELD, nil, nil);
	cur = p + 1;
	u = u->chain;
	u->language = curlang;
	curchar = cur;
	u->type = constype(nil);
	skipchar(curchar, ',');
	u->symvalue.field.offset = getint();
	skipchar(curchar, ',');
	u->symvalue.field.length = getint();
	skipchar(curchar, ';');
	chkcont(curchar);
	cur = curchar;
    }
    if (*cur == ';') {
	++cur;
    }
    curchar = cur;
}

/*
 * Construct an enumeration type.
 */

private consEnum (t)
Symbol t;
{
    register Symbol u;
    register char *p;
    register integer count;

    t->class = SCAL;
    count = 0;
    u = t;
    while (*curchar != ';' and *curchar != '\0') {
	p = index(curchar, ':');
	assert(p != nil);
	*p = '\0';
	u->chain = insert(identname(curchar, true));
	curchar = p + 1;
	u = u->chain;
	u->language = curlang;
	u->class = CONST;
	u->level = curblock->level + 1;
	u->block = curblock;
	u->type = t;
	u->symvalue.constval = build(O_LCON, (long) getint());
	++count;
	skipchar(curchar, ',');
	chkcont(curchar);
    }
    if (*curchar == ';') {
	++curchar;
    }
    t->symvalue.iconval = count;
}

/*
 * Construct a parameter list for a function or procedure variable.
 */

private consParamlist (t)
Symbol t;
{
    Symbol p;
    integer i, d, n, paramclass;

    n = getint();
    skipchar(curchar, ';');
    p = t;
    d = curblock->level + 1;
    for (i = 0; i < n; i++) {
	p->chain = newSymbol(nil, d, VAR, nil, nil);
	p = p->chain;
	p->type = constype(nil);
	skipchar(curchar, ',');
	paramclass = getint();
	if (paramclass == 0) {
	    p->class = REF;
	}
	skipchar(curchar, ';');
	chkcont(curchar);
    }
}

/*
 * Construct an imported type.
 * Add it to a list of symbols to get fixed up.
 */

private consImpType (t)
Symbol t;
{
    register char *p;

    p = curchar;
    while (*p != ',' and *p != ';' and *p != '\0') {
	++p;
    }
    if (*p == '\0') {
	panic("bad import symbol entry '%s'", curchar);
    }
    t->class = TYPEREF;
    t->symvalue.typeref = curchar;
    if (*p == ',') {
	curchar = p + 1;
	(void) constype(nil);
    } else {
	curchar = p;
    }
    skipchar(curchar, ';');
    *p = '\0';
}

/*
 * Construct an opaque type entry.
 */

private consOpaqType (t)
Symbol t;
{
    register char *p;
    register Symbol s;
    register Name n;
    boolean def;

    p = curchar;
    while (*p != ';' and *p != ',') {
	if (*p == '\0') {
	    panic("bad opaque symbol entry '%s'", curchar);
	}
	++p;
    }
    def = (Boolean) (*p == ',');
    *p = '\0';
    n = identname(curchar, true);
    find(s, n) where s->class == TYPEREF endfind(s);
    if (s == nil) {
	s = insert(n);
	s->class = TYPEREF;
	s->type = nil;
    }
    curchar = p + 1;
    if (def) {
	s->type = constype(nil);
	skipchar(curchar, ';');
    }
    t->class = TYPE;
    t->type = s;
}

/*
 * Read an integer from the current position in the type string.
 */

private integer getint ()
{
    register integer n;
    register char *p;
    register Boolean isneg;

    n = 0;
    p = curchar;
    if (*p == '-') {
	isneg = true;
	++p;
    } else {
	isneg = false;
    }
    while (isdigit(*p)) {
	n = 10*n + (*p - '0');
	++p;
    }
    curchar = p;
    return isneg ? (-n) : n;
}

/*
 * Add a tag name.  This is a kludge to be able to refer
 * to tags that have the same name as some other symbol
 * in the same block.
 */

private addtag (s)
register Symbol s;
{
    register Symbol t;
    char buf[100];

    sprintf(buf, "$$%.90s", ident(s->name));
    t = insert(identname(buf, false));
    t->language = s->language;
    t->class = TAG;
    t->type = s->type;
    t->block = s->block;
}
