/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */
#ifndef lint
static char rcsid[] = "$Header: tty.c 1.4 87/03/10 $";
#endif

/*
 * Manage ttys.  A "window" is really a tty to accept commands from and
 * direct output to.
 */

#include "defs.h"
#include "tty.h"
#include "main.h"
#include "library.h"
#include <signal.h>
#include <sgtty.h>
#include <fcntl.h>
#include <errno.h>

typedef struct {
    struct sgttyb sg;		/* standard sgttyb structure */
    struct tchars tc;		/* terminal characters */
    struct ltchars ltc;		/* local special characters */
    integer ldisc;		/* line discipline */
    integer local;		/* TIOCLGET */
    integer fcflags;		/* fcntl(2) F_GETFL, F_SETFL */
} Ttyinfo;

#define check(x)	((x) == -1 ? perror("x") : 0)
#define mask(b)		(1 << ((b) - 1))	/* for signals */

#define TTPATH	"/dev/tty"  	/* path to try for unrooted tty names */

public boolean notty = false;	/* don't actually mess with ttys */
public int defsigmask;		/* normal sigmask while running */

#ifndef public
/*
 * window i/o defines.  stdio is redirected to the appropriate window.
 */
#define printf	wprintf
#undef putchar
#define	putchar	wputchar
#define perror	wperror
File	win;			/* std window input; same as cw->w_in */ 
File	wout;			/* std window output; same as cw->w_out */
File	werr;			/* std window error; same as cw->w_err */
#undef stdin
#undef stdout
#undef stderr
#define stdin	win
#define	stdout	wout
#define stderr	werr
#endif public

typedef struct Window Window;
struct Window {
	char	*w_name;		/* name of device supporting window */
	File	w_in, w_out, w_err;	/* file descriptors */
	int	w_defid;		/* default process id for window */
	boolean	w_istty;		/* is window a tty? */
	boolean w_hastty;		/* does dbx have control of window? */
	boolean w_prompted;		/* have we printed a prompt already? */
	boolean	w_busy;			/* is a job using the window? */
	struct Job *w_job;		/* ptr to job using window */
	File	w_sourcing;		/* file we're sourcing; zero if none */
	Ttyinfo	w_dbxttyinfo;		/* dbx's idea of tty state */
	Ttyinfo w_origttyinfo;		/* original state of tty */
	int	w_origpgrp;		/* original pgrp of tty */
	Window	*w_next;		/* for winlist */
};

typedef struct Job Job;
struct Job {
	Window	*j_window;		/* window job is running in */
	Ttyinfo	j_ttyinfo;		/* job's idea of tty state */
	int	j_pgrp;			/* job's process group */
	int	j_proc;			/* id of process in foreground */
	Job	*j_next;		/* for joblist */
};


private Window	*winlist;		/* active windows */
private Window	*cw;			/* current window */
private Window	*cons;			/* window of 'console', e.g. stdin */
private int	oldpgrp;		/* pgrp initially */
private int	dbxpgrp;		/* process group of dbx */
private Job	*joblist;		/* list of active jobs */


/*
 * call this just before exiting dbx.
 */
public wquit()
{
	register Window *w;
	for (w = winlist; w; w = w->w_next)
		wclose(w);
}

/*
 * called by "quit" command.  if we're not in the original window,
 * just close the window, rather than quitting.
 */
closewindow()
{
	if (cw == cons)
		quit(0);
	else {
		wclose(cw);
		erecover();
	}
}


private iohandler()
{
}


#ifdef NOTDEF
/*
 * get a SIGHUP when a window closes.  close down the window.
 * doesn't work -- retval is always zero.
 */
private huphandler()
{
	register Window *w;
	int ready, retval;

	for (w = winlist; w; w = w->w_next) {
		retval = ioctl(fileno(w->w_in), FIONREAD, &ready);
		if (retval == -1) {
			if (w == cons)
				quit(1);
			else
				wclose(w);
		}
	}
}
#endif


/*
 * set up a window for the current stdin, stdout, and stderr.  make
 * it the current window.  this has to be called before any io is done.
 */
public winit()
{
	nosyserrs();

	defsigmask |= mask(SIGIO);
	sigsetmask(defsigmask);
	signal(SIGIO, iohandler);
	signal(SIGHUP, SIG_IGN);
	oldpgrp = getpgrp(0);
	dbxpgrp = getpid();

	cons = (Window *)calloc(sizeof(Window), 1);
	cons->w_in = &_iob[0];
	cons->w_out = &_iob[1];
	cons->w_err = &_iob[2];
	cons->w_istty = (boolean) isatty(fileno(cons->w_in));

	savetty(cons->w_in, &cons->w_origttyinfo);
	cons->w_dbxttyinfo = cons->w_origttyinfo;
	if (cons->w_istty) {
		cons->w_name = strdup(ttyname(fileno(cons->w_in)));
		cons->w_dbxttyinfo.fcflags = FASYNC;
		ioctl(fileno(cons->w_in), TIOCGPGRP, &cons->w_origpgrp);
	} else {
		cons->w_name = "stdio";
	}

	cons->w_next = winlist;
	winlist = cons;
	setwin(cons);

	syserrs();
}


/*
 * wait around until input is available for some window, and set the
 * current window.  "input available" means there is input and dbx
 * is ready for it.  while waiting, deal with the possibility of 
 * process events occuring.
 */
public wwait()
{
	register Window *w;
	int ready;
	int oldmask;
	extern char *sys_errlist[];
	extern int errno;

	oldmask = sigblock(mask(SIGCHLD) | mask(SIGINT));

	for (;;) {	/* until input available in a window */

		/*
		 * take care of any debugee transitions, if any.
		 */
		doqueue();

		/*
		 * for all 'unbusy' windows, print a prompt if necessary
		 * and check for input.  unbusy means no process is running
		 * in the window in foreground.
		 */ 
		for (w = winlist; w; w = w->w_next) {
			if (w->w_busy)
				continue;
			if (!w->w_hastty)
				grabtty(w);
			if (!w->w_prompted) {
				if (w->w_istty || interactive)
					fprintf(w->w_out, "(%s) ", cmdname);
				fflush(w->w_out);
				w->w_prompted = true;
			}
			if (!w->w_istty) {	/* e.g. sourcing */
				goto found;
			} else {		/* chars ready on tty? */
				ready = 0;
				check(ioctl(fileno(w->w_in), FIONREAD, &ready));
				if (ready)
					goto found;
			}
		}

		/*
		 * wait for SIGIO (input available in a window), or SIGCHLD
		 * (debugee process state changed).
		 */
		(void) sigpause(0);

	};

found:
	/*
	 * found input in window 'w'...
	 */
	(void) sigsetmask(oldmask);
	assert(w->w_busy == false);
	w->w_prompted = false;
	setwin(w);
}


/*
 * select window 'w' as the current window.
 */
private setwin(w)
	Window *w;
{
	if (w->w_sourcing)
		win = w->w_sourcing;
	else
		win = w->w_in;
	wout = w->w_out;
	werr = w->w_err;
	cw = w;
}

/*
 * push default process for current window.
 */
public setdefproc()
{
	if (find_procid(cw->w_defid))
		setprocess(cw->w_defid);
	else {
		clearproc();
		resetproc();
	}
}


/*
 * re-assign a default process for each window whose default process
 * no longer exists.
 */
wfixdefprocs(id)
	int id;
{
	register Window *w;

	if (find_procid(id) == 0)
		id = firstprocid();
	for (w = winlist; w; w = w->w_next) {
		if (find_procid(w->w_defid) == 0) {
			w->w_defid = id; 
			fprintf(w->w_out, "(Current process now %%%d) ", id);
			fflush(w->w_out);
		}
	}
}

/*
 * change default process for window.
 */
public setproc(id)
{
	setprocess(id);
	cw->w_defid = id;
}


/*
 * select and unselect the window for the current job.
 */

#define WSTACKSIZE	2
private int windex;
private Window *wstack[WSTACKSIZE];

public setjob(j)
	Job *j;
{
	if (windex < WSTACKSIZE) {
		wstack[windex++] = cw;
		if (j && j->j_window)
			setwin(j->j_window);
	} else {
		panic("window stack overflow");
	}
}

public unsetjob()
{
	if (windex > 0)
		setwin(wstack[--windex]);
	else
		panic("window stack underflow");
}


/*
 * create a window structure for the device name in 'ttyname' and add
 * it to the list of known windows.  if the name doesn't begin with '/',
 * try prepending TTPATH to it.  this alows names like "h9" for "/dev/ttyh9".
 * make default process for window 'procid'.
 */
public wopen(ttyname, procid)
	char *ttyname;
	int procid;
{
	Window *w;

	if (!cons->w_istty)
		error("can't create window because stdin is not a terminal");
	nosyserrs();
	w = (Window *)calloc(sizeof(Window), 1);
	w->w_name = ttyname;
	if (*w->w_name != '/') {
		w->w_name = (char *)stralloc(strlen(ttyname)+sizeof(TTPATH)+1);
		strcpy(w->w_name, TTPATH);
		strcat(w->w_name, ttyname);
	}
	w->w_in = fopen(w->w_name, "r");
	if (w->w_in == NULL) {
		beginerrmsg();
		perror(w->w_name);
		goto errormsg;
	}
	if (!isatty(fileno(w->w_in))) {
		beginerrmsg();
		fprintf(stderr, "%s: not a tty\n", w->w_name);
		goto errormsg;
	}
	w->w_istty = true;
	w->w_out = fopen(w->w_name, "w");
	if (w->w_out == NULL) {
		beginerrmsg();
		perror(w->w_name);
		goto errormsg;
	}

	w->w_err = fdopen(fileno(w->w_out), "w");
	if (w->w_err == NULL) {
		beginerrmsg();
		perror(w->w_name);
		goto errormsg;
	}
	setbuf(w->w_err, NULL);

	savetty(w->w_in, &w->w_origttyinfo);
	w->w_dbxttyinfo = cons->w_dbxttyinfo;
	check(ioctl(fileno(w->w_in), TIOCGPGRP, &w->w_origpgrp));
	check(ioctl(fileno(w->w_in), TIOCSPGRP, &dbxpgrp));
	restoretty(w->w_in, &w->w_dbxttyinfo);
	w->w_hastty = true;

	check(ioctl(fileno(w->w_in), TIOCEXCL, 0));

	if (wmoveproc(procid, w) == 0)
		goto error;

	w->w_next = winlist;
	winlist = w;
	syserrs();
	return;

errormsg:
	enderrmsg();
error:
	wclose(w);
	syserrs();
}


/*
 * move process with id 'id' to window 'w'.
 */
wmoveproc(id, w)
	int id;
	Window *w;
{
	Window *savecw = cw;

	if (seterror()) {
		cw = savecw;
		return(0);
	}

	setwin(w);
	w->w_defid = moveio(id, fileno(w->w_in), fileno(w->w_out),
		fileno(w->w_err));

	unseterror();
	return(1);
}


/*
 * set the window input for sourcing.  notice restoring the normal
 * window input.
 */
wsource(f)
	File f;
{
	if (f == cw->w_in)
		cw->w_sourcing = 0;
	else
		cw->w_sourcing = f;
	cw->w_istty = (boolean) isatty(fileno(f));
	win = f;
}


/*
 * close a window.
 */
private wclose(w)
	register Window *w;
{
	Window *this;
	Window **last_p;

	if (w) {
		nosyserrs();

		/*
		 * close file descriptors for the window
		 */
		if (w->w_in) {
			if (w->w_istty) {
				restoretty(w->w_in, &w->w_origttyinfo);
				ioctl(fileno(w->w_in), TIOCSPGRP, &w->w_origpgrp);
				if (w != cons)
					check(ioctl(fileno(w->w_in), TIOCNXCL, 0));
			}
			check(fclose(w->w_in));
		}

		if (w->w_out) {
			if (w != cons)
				fprintf(w->w_out, "Closing window %s.\n", w->w_name);
			check(fclose(w->w_out));
		}

		/* don't worry about w_err, it's a copy of w_out or stdout */

		/*
		 * delete the window out of the window list
		 */
		this = winlist;
		last_p = &winlist;
		while(this && this != w) {
			last_p = &this->w_next;
			this = this->w_next;
		}
		if (this)
			*last_p = this->w_next;

		/*
		 * clear it
		 */
		free(w);

		syserrs();
	}
}


/*
 * the following replace the normal 'printf', 'putchar', and 'perror',
 * substituting window in/out.
 */

/*VARARGS1*/
public wprintf(fmt, args)
{
	_doprnt(fmt, &args, wout);
	return(ferror(wout) ? EOF : 0);
}

public wputchar(c)
	char c;
{
	return(putc(c, wout));
}

wperror(s)
	char *s;
{
	fprintf(werr, "%s: %s\n", s, sys_errlist[errno]);
}



/*
 * deal with stdin tty.  should switch to new tty driver, but just
 * assumes it's in use.
 */

public setuptty()
{
    int termpgrp;
    Window *w;

    if (cons->w_istty && !notty) {
	/*
	 * wait until dbx is in foreground.
	 */
	ioctl(fileno(cons->w_in), TIOCGPGRP, &termpgrp);
	while(oldpgrp != termpgrp) {	/* dbx is in background */
	    signal(SIGTTIN, SIG_DFL);
	    kill(0, SIGTTIN);		/* stop self  */
	    ioctl(fileno(cons->w_in), TIOCGPGRP, &termpgrp);
	}
	/*
	 * ignore terminal signals so dbx can change terminals pgrps
	 * without changing into that pgrp first.  otherwise dbx
	 * would get suspended.
	 */
	signal(SIGTTIN, SIG_IGN);
	signal(SIGTTOU, SIG_IGN);
	signal(SIGTSTP, SIG_IGN);
	/*
	 * dbx needs its own process group so that SIGIOs don't
	 * get sent to parent or child.
	 */
	setpgrp(0, dbxpgrp);
	/*
	 * get control of windows and put them in known state.
	 */
	for(w = winlist; w; w = w->w_next) {
		ioctl(fileno(w->w_in), TIOCSPGRP, &dbxpgrp);
		restoretty(w->w_in, &w->w_dbxttyinfo);
	}
    }
}

/*
 * put dbx back in initial process group and restore terminals
 * to their original states and process groups.
 */
public untty()
{
	register Window *w;

	for (w = winlist; w; w = w->w_next) {
		if (w->w_istty) {
			restoretty(w->w_in, &w->w_origttyinfo);
			ioctl(fileno(w->w_in), TIOCSPGRP, &w->w_origpgrp);
		}
	}
	if (oldpgrp)
		setpgrp(0, oldpgrp);
}

/*
 * this is called from child processes that dbx forks directly.  undoes
 * all the mess that was made to handle the tty.
 */
public fixtty()
{
	if (!notty) {
		signal(SIGTTIN, SIG_DFL);
		signal(SIGTTOU, SIG_DFL);
		signal(SIGTSTP, SIG_DFL);
		signal(SIGHUP, SIG_DFL);
		setpgrp(0, getpid());
	}
}

/*
 * get the tty for dbx's use
 */
public grabtty(w)
	Window *w;
{
	Job *job;

	job = w->w_job;
	assert(w->w_busy == false);
	if (job) {
		if (w->w_istty && !w->w_hastty && !notty) {
			ioctl(fileno(w->w_in), TIOCGPGRP, &job->j_pgrp);
			if (job->j_pgrp != dbxpgrp) {
				ioctl(fileno(w->w_in), TIOCSPGRP, &dbxpgrp);
				savetty(w->w_in, &job->j_ttyinfo);
				restoretty(w->w_in, &w->w_dbxttyinfo);
			}
		}
		job->j_proc = 0;
    	}
	w->w_hastty = true;
	w->w_job = 0;
}

/*
 * let job 'job' have the tty
 */
public giveawaytty(job, id)
	register Job *job;
	int id;
{
	register Window *w;

	if (job->j_proc)
		error("%%%d is already running in foreground", job->j_proc);
	job->j_proc = id;
	w = job->j_window;
	if (w->w_istty && !notty) {
		restoretty(w->w_in, &job->j_ttyinfo);
		ioctl(fileno(w->w_in), TIOCSPGRP, &job->j_pgrp);
	}
	assert(w->w_busy == false);
	w->w_job = job;
	w->w_busy = true;
	w->w_hastty = false;
}

/*
 * suspend dbx, just like csh.  let the processes fend for themselves.
 */
public suspend()
{
	int (*old)();

	untty();
	old = signal(SIGTSTP, SIG_DFL);
	kill(0, SIGTSTP);
	signal(SIGTSTP, old);
	setuptty();
}

/*
 * Save/restore the state of a tty.
 */

private savetty(f, t)
File f;
Ttyinfo *t;
{
    if (!notty) {
	ioctl(fileno(f), TIOCGETP, &(t->sg));
	ioctl(fileno(f), TIOCGETC, &(t->tc));
	ioctl(fileno(f), TIOCGLTC, &(t->ltc));
	ioctl(fileno(f), TIOCGETD, &(t->ldisc));
	ioctl(fileno(f), TIOCLGET, &(t->local));
    }
    t->fcflags = fcntl(fileno(f), F_GETFL, 0);
}

private restoretty(f, t)
File f;
Ttyinfo *t;
{
    if (!notty) {
	ioctl(fileno(f), TIOCSETN, &(t->sg));
	ioctl(fileno(f), TIOCSETC, &(t->tc));
	ioctl(fileno(f), TIOCSLTC, &(t->ltc));
	ioctl(fileno(f), TIOCSETD, &(t->ldisc));
	ioctl(fileno(f), TIOCLSET, &(t->local));
    }
    (void) fcntl(fileno(f), F_SETFL, t->fcflags);
}

public boolean isterm(file)
	File file;
{
	return((boolean)(interactive or isatty(fileno(file))));
}


/*
 * Return whether we are currently reading from standard input.
 */
public boolean notsourcing()
{
    return (boolean) (cw->w_sourcing == 0);
}

/*
 * create a new job.
 */
public struct Job *newjob(pgrp)
{
	register Job *j;

	j = (Job *)calloc(sizeof(Job), 1);
	j->j_pgrp = pgrp;
	j->j_window = cw;
	j->j_ttyinfo = cw->w_dbxttyinfo;
	j->j_ttyinfo.fcflags = 0;
	j->j_next = joblist;
	joblist = j;
	return(j);
}

public struct Job *dupjob(from)
	register Job *from;
{
	register Job *to;

	to = (Job *)calloc(sizeof(Job), 1);
	to->j_pgrp = from->j_pgrp;
	to->j_window = cw;
	to->j_ttyinfo = from->j_ttyinfo;
	to->j_ttyinfo.fcflags = from->j_ttyinfo.fcflags;
	to->j_next = joblist;
	joblist = to;
	return(to);
}


/*
 * note a new process in a job.
 */
public addjob(j)
	Job *j;
{
}

/*
 * note the exit of a process from a job.
 */
public jobexit(j)
	Job *j;
{
}

/*
 * note that job has relinquished the tty.  this is equivalent to
 * command completion for commands that continue processes.
 */
public unbusy(job, id)
	Job *job;
	int id;
{
	if (job && job->j_proc == id) {
		job->j_window->w_busy = false;
	}
}

/*
 * dump info about windows.  should be a "debug" function.
 */
dumpttys()
{
	Window *w;
	Job *j;
	int pgrp;

	printf("dbxpgrp %d oldpgrp %d now %d\n", dbxpgrp, oldpgrp, getpgrp(0));
	for (w = winlist; w; w = w->w_next) {
		check(ioctl(fileno(w->w_in), TIOCGPGRP, &pgrp));
		printf("%s: in %d out %d pgrp %d", w->w_name,
			fileno(w->w_in), fileno(w->w_out), pgrp);
		if (w == cons) printf(" cons");
		if (w->w_istty) printf(" istty");
		if (w->w_hastty) printf(" hastty");
		if (w->w_busy) printf(" busy");
		if (w->w_prompted) printf(" prompted");
		printf("\n");
	}

	for (j = joblist; j; j = j->j_next) {
		printf("job: pgrp %d", j->j_pgrp);
		if (j->j_proc)
			printf(" %%%d", j->j_proc);
		printf("\n");
	}
	fflush(stdout);
}
