/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef lint
static char rcsid[] = "$Header: machine.c 1.34 88/03/15 $";
#endif

/*
 * Target machine dependent stuff for the Intel 80386.
 */

#include "defs.h"
#include "main.h"
#include "machine.h"
#include "multi-p.h"
#include "runtime.h"
#include "events.h"
#include "main.h"
#include "symbols.h"
#include "source.h"
#include "mappings.h"
#include "object.h"
#include "keywords.h"
#include <signal.h>
#include <math.h>

#ifndef public
typedef unsigned int Address;
typedef unsigned char Byte;
typedef unsigned int Word;

/*
 * register numbers:
 *
 * 0..2		temporary registers (%eax, %edx, %ecx)
 * 3..4		temporary floating registers (%st(0), %st(1))
 * 5..7		user int/pointer registers (%ebx, %esi, %edi)
 * 8..13	user floating registers (%st(2), ..., %st(7))
 * 14		stack pointer (%esp)
 * 15		frame pointer (%ebp)
 * 16		program counter (%eip)
 * 17		flags (%eflags)
 * 18..48	Weitek FPA registers (%fp1..%fp31)
 */

#define FPA_BASE	18		/* subtract to use as bitmap index */
#define	REG_MASK(reg)	(1<<(reg))
#define FREG_MASK(reg)	(1<<( (reg) - FPA_BASE) )
#define FPA_INDEX(reg)	( (reg) - FPA_BASE)

	/* Temps int/pointer */
#define REG_EAX		0
#define REG_EDX		1
#define REG_ECX		2

	/* Temps floating point */
#define REG_FP0		3
#define REG_FP1		4

	/* User Regs int/pointer */
#define REG_EBX		5
#define REG_ESI		6
#define REG_EDI		7

	/* User Regs floating point */
#define REG_FP2		8
#define REG_FP3		9
#define REG_FP4		10
#define REG_FP5		11
#define REG_FP6		12
#define REG_FP7		13

	/* Stack Pointer/Frame Pointer */
#define REG_ESP		14
#define REG_EBP		15

	/* Program Counter/Flags */
#define REG_EIP		16
#define REG_EFLAGS	17

	/* Weitek FPA registers */
#define REG_FPA1	18
#define REG_FPA2	19
#define REG_FPA3	20
#define REG_FPA4	21
#define REG_FPA5	22
#define REG_FPA6	23
#define REG_FPA7	24
#define REG_FPA8	25
#define REG_FPA9	26
#define REG_FPA10	27
#define REG_FPA11	28
#define REG_FPA12	29
#define REG_FPA13	30
#define REG_FPA14	31
#define REG_FPA15	32
#define REG_FPA16	33
#define REG_FPA17	34
#define REG_FPA18	35
#define REG_FPA19	36
#define REG_FPA20	37
#define REG_FPA21	38
#define REG_FPA22	39
#define REG_FPA23	40
#define REG_FPA24	41
#define REG_FPA25	42
#define REG_FPA26	43
#define REG_FPA27	44
#define REG_FPA28	45
#define REG_FPA29	46
#define REG_FPA30	47
#define REG_FPA31	48

#define NREG		49

/* for backwards compatibility */
#define FRP		REG_EBP
#define STKP		REG_ESP
#define	PROGCTR		REG_EIP
#define PSR		REG_EFLAGS

#define	PSR_TRACEBIT	0x100

#define BITSPERBYTE 8
#define BITSPERWORD (BITSPERBYTE * sizeof(Word))

#include "source.h"
#include "symbols.h"

Address prtaddr;

#define	pr_PC	pr_eip

#endif

public String regname[] = {
    "eax", "edx", "ecx", "st0", "st1", "ebx", "esi", "edi", 
    "st2", "st3", "st4", "st5", "st6", "st7",
    "esp", "ebp", "eip", "eflags",
    "fp1",  "fp2",  "fp3",  "fp4",  "fp5",  "fp6",  "fp7",  "fp8",
    "fp9",  "fp10", "fp11", "fp12", "fp13", "fp14", "fp15", "fp16",
    "fp17", "fp18", "fp19", "fp20", "fp21", "fp22", "fp23", "fp24",
    "fp25", "fp26", "fp27", "fp28", "fp29", "fp30", "fp31"
};

/*
 * return a byte for the disassembler
 */

private unsigned char disbyte(addr)
    unsigned int addr;
{
    unsigned char c;

    iread(&c, addr, sizeof(c));
    return(c);
}

#define DISNAMESIZE	128

private char *disname(addr)
    unsigned int addr;
{
    Symbol b;

    b = whatblock(addr);
    if (b != nil and codeloc(b) == addr)
	return(symname(b));
    else
	return((char *)0);
}

/*
 * initialize machine-dependent stuff.
 */

public machine_init()
{
    dis386_init(0, disbyte, disname, 0);
}

/*
 * define symbols for the machine's registers.
 */

public defregnames()
{
    defregname(identname("$eax", false), REG_EAX);
    defregname(identname("$edx", false), REG_EDX);
    defregname(identname("$ecx", false), REG_ECX);
    defregname(identname("$st0", false), REG_FP0);
    defregname(identname("$st1", false), REG_FP1);
    defregname(identname("$ebx", false), REG_EBX);
    defregname(identname("$esi", false), REG_ESI);
    defregname(identname("$edi", false), REG_EDI);
    defregname(identname("$st2", false), REG_FP2);
    defregname(identname("$st3", false), REG_FP3);
    defregname(identname("$st4", false), REG_FP4);
    defregname(identname("$st5", false), REG_FP5); 
    defregname(identname("$st6", false), REG_FP6);
    defregname(identname("$st7", false), REG_FP7);
    defregname(identname("$esp", true), REG_ESP);
    defregname(identname("$sp", true), REG_ESP);	/* alias */
    defregname(identname("$ebp", true), REG_EBP);
    defregname(identname("$fp", true), REG_EBP);	/* alias */
    defregname(identname("$eip", true), REG_EIP);
    defregname(identname("$pc", true), REG_EIP);	/* alias */
    defregname(identname("$eflags", true), REG_EFLAGS);
    defregname(identname("$fp1", true), REG_FPA1);
    defregname(identname("$fp2", true), REG_FPA2);
    defregname(identname("$fp3", true), REG_FPA3);
    defregname(identname("$fp4", true), REG_FPA4);
    defregname(identname("$fp5", true), REG_FPA5);
    defregname(identname("$fp6", true), REG_FPA6);
    defregname(identname("$fp7", true), REG_FPA7);
    defregname(identname("$fp8", true), REG_FPA8);
    defregname(identname("$fp9", true), REG_FPA9);
    defregname(identname("$fp10", true), REG_FPA10);
    defregname(identname("$fp11", true), REG_FPA11);
    defregname(identname("$fp12", true), REG_FPA12);
    defregname(identname("$fp13", true), REG_FPA13);
    defregname(identname("$fp14", true), REG_FPA14);
    defregname(identname("$fp15", true), REG_FPA15);
    defregname(identname("$fp16", true), REG_FPA16);
    defregname(identname("$fp17", true), REG_FPA17);
    defregname(identname("$fp18", true), REG_FPA18);
    defregname(identname("$fp19", true), REG_FPA19);
    defregname(identname("$fp20", true), REG_FPA20);
    defregname(identname("$fp21", true), REG_FPA21);
    defregname(identname("$fp22", true), REG_FPA22);
    defregname(identname("$fp23", true), REG_FPA23);
    defregname(identname("$fp24", true), REG_FPA24);
    defregname(identname("$fp25", true), REG_FPA25);
    defregname(identname("$fp26", true), REG_FPA26);
    defregname(identname("$fp27", true), REG_FPA27);
    defregname(identname("$fp28", true), REG_FPA28);
    defregname(identname("$fp29", true), REG_FPA29);
    defregname(identname("$fp30", true), REG_FPA30);
    defregname(identname("$fp31", true), REG_FPA31);
}

/*
 * translate the register number given by the compiler to the
 * number used internally by dbx.
 */

public fixregnumber( n )
	int n;
{
	if ((unsigned)n >= NREG)
		error("Internal error: bad register number %d", n);

	return n;
}

/*
 * convert a pair of register values into a double.
 */

public double makedouble(lo, hi)
	unsigned int lo;
	unsigned int hi;
{
	union {
		double	d;
		struct {
			unsigned int	low;
			unsigned int	high;
		} word;
	} dub;

	dub.word.low = lo;
	dub.word.high = hi;
	return dub.d;
}

private Address printop();

/*
 * Decode and print the instructions within the given address range.
 */

public printinst(lowaddr, highaddr)
Address lowaddr;
Address highaddr;
{
    register Address addr;
    Word value;
    int length;
    char *mnem;
    extern char *dis386();

    for (addr = lowaddr; addr <= highaddr; addr += length) {
	printf("%08x  ", addr);
	mnem = dis386(addr, &length);
	if (mnem == (char *)0) {
	    iread(&value, addr, sizeof(value));
	    length = sizeof(value);
	    printf("??? (%08x)\n", value);
	} else
	    printf("%s\n", mnem);
    }
    prtaddr = addr;
}

/*
 * Another approach:  print n instructions starting at the given address.
 */

public printninst(count, addr)
int count;
Address addr;
{
    register Integer i;
    register Address newaddr;
    int value, length;
    char *mnem;
    extern char *dis386();

    if (count <= 0) {
	error("non-positive repetition count");
    } else {
	newaddr = addr;
	for (i = 0; i < count; i++) {
	    printf("%08x  ", newaddr);
	    mnem = dis386(newaddr, &length);
	    if (mnem == (char *)0) {
	        iread(&value, newaddr, sizeof(value));
	        length = sizeof(value);
	        printf("??? (%08x)\n", value);
	    } else
	        printf("%s\n", mnem);
	    newaddr += length;
	}
	prtaddr = newaddr;
    }
}


/*
 * Print the contents of the addresses within the given range
 * according to the given format.
 */

typedef struct {
    String name;
    String printfstring;
    int length;
} Format;

private Format fmt[] = {
    { "d", " %d", sizeof(short) },
    { "D", " %ld", sizeof(long) },
    { "o", " %o", sizeof(short) },
    { "O", " %lo", sizeof(long) },
    { "x", " %04x", sizeof(short) },
    { "X", " %08x", sizeof(long) },
    { "b", " \\%o", sizeof(char) },
    { "c", " '%c'", sizeof(char) },
    { "s", "%c", sizeof(char) },
    { "f", " %14.7e", sizeof(float) },
    { "g", " %23.15e", sizeof(double) },
    { nil, nil, 0 }
};

private Format *findformat(s)
String s;
{
    register Format *f;

    f = &fmt[0];
    while (f->name != nil and not streq(f->name, s)) {
	++f;
    }
    if (f->name == nil) {
	error("bad print format \"%s\"", s);
    }
    return f;
}

public Address printdata(lowaddr, highaddr, format)
Address lowaddr;
Address highaddr;
String format;
{
    register int n;
    register Address addr;
    register Format *f;
    union {
	char charv;
	short shortv;
	int intv;
	float floatv;
	double doublev;
    } value;

    if (lowaddr > highaddr) {
	error("first address larger than second");
    }
    f = findformat(format);
    n = 0;
    value.intv = 0;
    for (addr = lowaddr; addr <= highaddr; addr += f->length) {
	if (n == 0) {
	    printf("%08x: ", addr);
	}
	dread(&value, addr, f->length);
	if (streq(f->name, "f"))
	    value.doublev = value.floatv;
	printf(f->printfstring, value);
	++n;
	if (n >= (16 div f->length)) {
	    putchar('\n');
	    n = 0;
	}
    }
    if (n != 0) {
	putchar('\n');
    }
    prtaddr = addr;
    return addr;
}

/*
 * The other approach is to print n items starting with a given address.
 */

public printndata(count, startaddr, format)
int count;
Address startaddr;
String format;
{
    register int i, n;
    register Address addr;
    register Format *f;
    register Boolean isstring;
    char c;
    union {
	char charv;
	short shortv;
	int intv;
	float floatv;
	double doublev;
    } value;

    if (count <= 0) {
	error("non-positive repetition count");
    }
    f = findformat(format);
    isstring = (Boolean) streq(f->name, "s");
    n = 0;
    addr = startaddr;
    value.intv = 0;
    for (i = 0; i < count; i++) {
	if (n == 0) {
	    printf("%08x: ", addr);
	}
	if (isstring) {
	    putchar('"');
	    dread(&c, addr, sizeof(char));
	    while (c != '\0') {
		printchar(c);
		++addr;
		dread(&c, addr, sizeof(char));
	    }
	    putchar('"');
	    putchar('\n');
	    n = 0;
	    addr += sizeof(String);
	} else {
	    dread(&value, addr, f->length);
	    if (streq(f->name, "f"))
		value.doublev = value.floatv;
	    printf(f->printfstring, value);
	    ++n;
	    if (n >= (16 div f->length)) {
		putchar('\n');
		n = 0;
	    }
	    addr += f->length;
	}
    }
    if (n != 0) {
	putchar('\n');
    }
    prtaddr = addr;
}

/*
 * Print out a value according to the given format.
 */

public printvalue(v, format)
long v;
String format;
{
    Format *f;
    char *p, *q;

    f = findformat(format);
    if (streq(f->name, "s")) {
	putchar('"');
	p = (char *) &v;
	q = p + sizeof(v);
	while (p < q) {
	    printchar(*p);
	    ++p;
	}
	putchar('"');
    } else {
	printf(f->printfstring, v);
    }
    putchar('\n');
}

/*
 * Print out a signal.
 */

private String siglist[] = {
	/* SIG0 */	"Signal 0",
	/* SIGHUP */	"Hangup",
	/* SIGINT */	"Interrupt",
	/* SIGQUIT */	"Quit",
	/* SIGILL */	"Illegal instruction",
	/* SIGTRAP */	"Trace/BPT trap",
	/* SIGIOT */	"IOT trap",
	/* SIGEMT */	"EMT trap",
	/* SIGFPE */	"Floating point exception",
	/* SIGKILL */	"Kill",
	/* SIGBUS */	"Bus error",
	/* SIGSEGV */	"Segmentation fault",
	/* SIGSYS */	"Bad system call",
	/* SIGPIPE */	"Broken pipe",
	/* SIGALRM */	"Alarm clock",
	/* SIGTERM */	"Software termination signal",
	/* SIGURG */	"Urgent I/O condition",
	/* SIGSTOP */	"Stop signal (SIGSTOP)",
	/* SIGTSTP */	"Stop signal from terminal",
	/* SIGCONT */	"Continue",
	/* SIGCHLD */	"Child exit (SIGCHLD)",
	/* SIGTTIN */	"Tty input",
	/* SIGTTOU */	"Tty output",
	/* SIGIO */	"I/O possible",
	/* SIGXCPU */	"Cputime limit exceeded",
	/* SIGXFSZ */	"Filesize limit exceeded",
	/* SIGVTALRM */	"Virtual timer",
	/* SIGPROF */	"Profiling timer",
	/* SIGWINCH */	"Window size change",
	/* SIG29 */	"Signal 29",
	/* SIGUSR1 */	"User defined signal 1",
	/* SIGUSR2 */	"User defined signal 2"
};

private struct illinames {
    int illi_code;
    String illi_name;
} illinames[] = {
{   ILL_PRIVIN_FAULT, "priviliged instruction fault",		},
{   ILL_RESOP_FAULT,  "reserved operand fault"			},
{   NULL, NULL 							},
};

private struct fpenames {
    int	fpe_code;
    String fpe_name;
} fpenames[] = {
{   FPE_FLTINV_TRAP,  "floating invalid operation"		},
{   FPE_FLTDEN_TRAP,  "floating denormal"			},
{   FPE_FLTDIV_TRAP,  "floating divide by zero"			},
{   FPE_FLTOVF_TRAP,  "floating overflow"			},
{   FPE_FLTUND_TRAP,  "floating underflow"			},
{   FPE_FLTPRE_TRAP,  "floating precision"			},
{   FPE_FLTSTK_TRAP,  "floating stack overflow/underflow"	},
{   FPE_INTDIV_TRAP,  "integer divide by zero"			},
{   FPE_INTOVF_TRAP,  "integer overflow"			}, 
{   FPE_NOFPA_AVAIL,  "no FPA available"			},
{   NULL, NULL, 						},
};

public printsig (signo)
integer signo;
{
    unsigned code;
    register int i;

    if (signo < 0 or signo > NSIG) {
	printf("[signal %d]", signo);
    } else {
	printf("%s", siglist[signo]);
    }
    code = errcode();
    if (signo == SIGILL) {
	for (i=0; illinames[i].illi_name != NULL; i++)
	    if (code == illinames[i].illi_code) {
	        printf(" (%s)", illinames[i].illi_name);
		break;
	}
    } else if (signo == SIGFPE) {
	for (i=0; fpenames[i].fpe_name != NULL; i++)
	    if (code == fpenames[i].fpe_code) {
	        printf(" (%s)", fpenames[i].fpe_name);
		break;
	    }
    }
}

/*
 * Determine if it's ok to skip function f entered by instruction ins.
 * If so, we're going to compute the return address and step to it.
 */

private boolean skipfunc (f)
Symbol f;
{
    return ((boolean) (
	not (wp->p_stepi or wp->p_tracei) and nlhdr.nlines != 0 and
	nosource(f) and canskip(f)
    ));
}

#define	CALL	0xe8		/* 386 'call'  opcode */
#define LCALL	0x9a		/* 386 'lcall' opcode */
#define LEAVE	0xc9		/* 386 'leave' opcode */

/*
 * is the instruction at 'addr' a subroutine call instruction?
 */
public boolean iscall(addr)
	Address addr;
{
	Word i;
	unsigned char op0, op1;

	if (addr == 0)
		return(false);
	iread(&i, addr, sizeof(i));
	op0 = (i & 0xff);
	op1 = (i >> 8) & 0xff;
	/* 3 forms of the call instruction */
	if ( op0 == CALL				/* 32bit displacement */
	or ( op0 == 0xff and ((op1 & 0x38) == 0x10) )	/* register */
	or   op0 == LCALL )				/* selector */
		return(true);
	else
		return(false);
}

/*
 * For the commonest kind of subroutine call instruction,
 * check if the instruction at 'addr' is a call to a routine with
 * no source.  If so, return the address of the next instruction.
 * Routines with no source should be skipped for performance.
 */
public Address shouldskip(addr)
	register Address addr;	/* address at which to look for skippable call*/
{
	Byte opcode;		/* opcode of instruction */
	Word disp;		/* displacement of CALL instr */
	Symbol target;		/* call target of CALL instr */

	iread(&opcode, addr, sizeof(opcode));
	if (opcode == CALL) {
		iread(&disp, addr + sizeof(opcode), sizeof(disp));
		target = whatblock(addr + (int) disp +
			sizeof(opcode) + sizeof(disp) );
		if (nosource(target) and canskip(target))
			return(addr + sizeof(opcode) + sizeof(disp));
	}
	return((Address) 0);
}

/*
 * get_enter_args looks for an subroutine prologue at startaddr, and stuffs
 * the mask and local storage args found into the locations in mask_p
 * and locals_p.  startaddr is assumed to be the first instruction of a
 * procedure.  the mask is 16 bits long, with a bit set for each register
 * saved.  
 *
 *    bits 0..2  	eax, edx, ecx
 *    bits 3..4  	st(0), st(1)
 *    bits 5..7  	ebx, esi, edi
 *    bits 8..13 	st(2), ..., st(7)
 *
 * This is easy on other machines with an enter instruction and register
 * bit mask.  On the 386, we have to disassemble the instructions and keep
 * around a state machine to know when to stop.  Here is the set of 
 * instructions in the order expected:
 *
 *	pushl	%ebp		-- subroutine frame pointer linkage
 *	movl	%esp,%ebp
 *
 *	pushl	%eax		-- allocate local stack storage (=4 bytes)
 *	subl	$#,%esp		-- allocate local stack storage (>4 bytes)
 *
 *	subl	$8,%esp		-- save a floating point register (#=0..7)
 *	fxch	%st(#)
 *	fstl	(%esp)
 *
 *	anything else		- stops state machine
 */

# define SHOW_M	m == NULL ? "null" : m

public get_enter_args(startaddr, mask_p, wmask_p, locals_p, stringmove, no_frame)
    Address startaddr;	/* where to start looking */
    int *mask_p;	/* where to put mask arg we found */
    int *wmask_p;	/* wytek register mask */
    int *locals_p;	/* where to put locals arg we found */
    Boolean *stringmove; /* was string move used to save wytek regs */
    Boolean *no_frame;	/* does not push frame ptr on stack */
{
    Word instr;		/* instruction to look at */
    register char *m;
    register Address addr = startaddr;
    Address newaddr;
    int length;
    int locals, lmask, wmask, i;
    Boolean smove = false;
    Boolean no_frame_save = true;
    extern char *dis386();

    m = dis386(addr, &length);
    if (debug_regsave)
	printf("get_enter_args: addr= 0x%x, m= '%s'\n", addr, SHOW_M);
    if (m == NULL) {
	*mask_p = *locals_p = *wmask_p = 0;
	*stringmove = false;
	*no_frame = false;
	return;
    } 

    /*
     * in non-optimized code, the subroutine prologue is at the end
     * of the procedure.  first thing the code does is jump there.
     */
    locals = lmask = wmask = 0;
    if (strncmp(m, "jmp", 3) == 0 and sscanf(m, "%*[^0]0x%x", &newaddr) == 1) {
	    addr = newaddr;
	    m = dis386(addr, &length);
	    if (debug_regsave)
		printf("following jump to addr 0x%x\n", addr);
	    if (m == NULL)
	        error("internal error: bad bfcode\n");
    }

    /*
     * Loop through instructions watching stack adjust instructions
     * and saving of local registers and floating registers.  We create
     * a local mask of saved registers and remember the number of bytes
     * on the local stack
     */
    for(;;) {
	if (debug_regsave)
	    printf("get_enter_args: addr= 0x%x, m= '%s'\n", addr, SHOW_M);
	if (m == NULL or strncmp("jmp", m, 3) ==  0) {
	    if (debug_regsave)
		printf("disassembly stopped by jump\n");
	    break;	/* no disassembly or a jump instruction stops us */
	}
	if (strncmp("pushl", m, 5) == 0) {
	    if (substr(",", m)) {
	        if (debug_regsave)
		    printf("disassembly stopped by 'pushl' with two args\n");
		break;	/* pushl with two operands stops us */
	    }
	    if (substr("%eax", m)) 
		if (locals == 0)	/* 4 bytes of local storage */
			locals = sizeof(int);
		    else
			lmask |= REG_MASK(REG_EAX);
	    else if (substr("%edx", m)) lmask |= REG_MASK(REG_EDX);
	    else if (substr("%ecx", m)) lmask |= REG_MASK(REG_ECX);
	    else if (substr("%ebx", m)) lmask |= REG_MASK(REG_EBX);
	    else if (substr("%esi", m)) lmask |= REG_MASK(REG_ESI);
	    else if (substr("%edi", m)) lmask |= REG_MASK(REG_EDI);
	    else if (substr("%ebp", m)) no_frame_save = false;
	    else if (substr("%fp", m)) {	/* weitek save */
	        if (sscanf(m, "%*[^%]%%fp%d", &i) == 1) {
		    if (i < 1 or i > 31)
		       error("internal error: bad fpa reg %d store", i); 
		    else
		        wmask |= REG_MASK(i - 1);
		}
	    } else {
		if (debug_regsave)
		    printf("unknown 'pushl' stops disassembly\n");
		break;	/* any other 'pushl' stops us */
	    }
	} else if (strncmp("wstors", m, 6) == 0) {
	    if (sscanf(m, "%*[^%]%%fp%d", &i) == 1) {
		if (i < 1 or i > 31)
		   error("internal error: bad fpa reg %d store", i); 
		else
		    wmask |= REG_MASK(i - 1);
	    } else {
		if (debug_regsave)
		    printf("unknown 'wstores' stops disassembly\n");
		break;  /* any other "wstors" stops us */
	    }
	} else if (strncmp("movl", m, 4) == 0) {
	    if (substr("%esp,%ebp", m)) /* void */;
	    else if (substr("%esp,%edi", m)) /* void */;
	    else if (substr("movl", m) and substr("%ecx", m)) /* void */;
	    else if (substr("$0xffc0", m)) { /* weitek smovl register save */
		if (sscanf(m, "%*[^$]$0x%x,%%esi", &i) == 1) {
		    wmask = decode_w_smovl(i);
		    smove = true;
		} else {
		    if (debug_regsave)
			printf("unknown 'movl' stops disassembly\n");
		    break;
		}
		if (debug_regsave)
		    printf("disassembly stopped by weitek smovl reg save\n");
		break;
	    }
	    else {
		if (debug_regsave)
		    printf("unknown 'movl' stops disassembly\n");
		break;	/* any other 'movl' stops us */
	    }
	} else if (strncmp("subl", m, 4) == 0) {
	    if (locals != 0) /* void */;
	    /* explicit pushes for floating variables are ignored here */
	    else if (substr("$", m) and substr("%esp", m)) {
		if (substr("$-", m))
			error("internal error: neg stack adjust, '%s'\n", m);
		if (sscanf(m, "%*[^$]$0x%x,%%esp", &i) != 1 and
		    sscanf(m, "%*[^$]$%x,%%esp", &i) != 1)
			error("internal error: stack subtract?, '%s'\n", m);
		locals += i;
	    } else {
		if (debug_regsave)
		    printf("unknown 'subl' stops disassembly\n");
		break;	/* any other 'subl' stops us */
	    }
	} else if (strncmp("fxch", m, 4) == 0) {
	    if (sscanf(m, "%*[^%]%%st(%d)", &i) == 1) {
		switch (i) {
		case 0:  lmask |= REG_MASK(REG_FP0); break;
		case 1:  lmask |= REG_MASK(REG_FP1); break;
		case 2:  lmask |= REG_MASK(REG_FP2); break;
		case 3:  lmask |= REG_MASK(REG_FP3); break;
		case 4:  lmask |= REG_MASK(REG_FP4); break;
		case 5:  lmask |= REG_MASK(REG_FP5); break;
		case 6:  lmask |= REG_MASK(REG_FP6); break;
		case 7:  lmask |= REG_MASK(REG_FP7); break;
		default:
		       error("internal error: bad floating reg, '%s'\n", m);
		       break;
		}
	    } else {
		if (debug_regsave)
		    printf("unknown 'fxch' stops disassembly\n");
		break;  /* any other 'fxch' stops us */
	    }
	} else if (strncmp("fstl", m, 4) == 0) {
		/* void */;
	} else {
		if (debug_regsave)
		    printf("disassembly stopped by '%s'\n", SHOW_M);
		break;  /* anything else stops us */
	}
next:
	addr += length;
	m = dis386(addr, &length);
    }
    *mask_p = lmask;
    *wmask_p = wmask;
    *locals_p = locals;
    *stringmove = smove;
    *no_frame = no_frame_save;
    if (debug_regsave) {
	printf("get_enter_args: regs= 0x%x, weitek= 0x%x, locals= %d",
	    lmask, wmask, locals);
	if (smove)
	    printf(" smove= T,");
	else
	    printf(" smove= F,");
	if (no_frame_save)
	    printf(" no_frame= T\n");
	else
	    printf(" no_frame= F\n");
    }
}

/*
 * Decode the "movl $0xffc00nn,%esi" instruction setting up for a
 * string save of weitek registers.  Bits 2 -> 6 give the lowest
 * register used for the save.
 */

private decode_w_smovl(n)
    int n;
{
    int i, mask;

    mask = 0;
    n &= 0x7f;
    n >>= 2;

    for (i = n; i <= 31; i++)
    	mask |= REG_MASK(i-1);
	
    if (debug_regsave)
    	printf("decode_w_smove: weitek registers %%fp%0d ... %%fp31 saved\n", n);

    return(mask);
}

/*
 * Is s1 a substring of s2?
 */
private substr(s1, s2)
	register char *s1, *s2;
{
	register n = strlen(s1);
	
	if (n == 0) 
		return 0;
	while (s2 and *s2) {
		if (*s1 == *s2 and strncmp(s1, s2, n) == 0)
			return 1;
		++s2;
	}
	return 0;
}

#define POP_ECX		  0x59	/* opcode for 386 'popl %ecx' instruction */
#define ADDLIMM8	0xc483	/* opcode for 386 'addl $imm8' instruction */
#define	ADDLIMM32	0xc481	/* opcode for 386 'addl $imm32' instruction */
/*
 * given the return pc address of a procedure, decide how many bytes
 * were passed to it by looking at the stack adjust instruction.  return the
 * number of bytes pushed.
 */
public int nbytespushed( pc )
	Address pc;
{
	Word instr = 0;
	int offset;

	iread(&instr, pc, 4);
	if ((instr & 0xff) == POP_ECX) 
		return 4;
	else if ((instr & 0xffff) == ADDLIMM8)
		return ((unsigned)(instr & 0xff0000)) >> 16;
	else if ((instr & 0xffff) == ADDLIMM32) {
		iread(&offset, pc+2, 4);
		return offset;
	} else
		return 0;
}

/*
 * return true when a subroutine is about to return, but before
 * its stack frame has gone away (so we can say something useful
 * about it).
 */
public boolean isreturn(addr)
	Address addr;
{
	Word i;

	iread(&i, addr, 1);
	if ((i&0xff) == LEAVE)
		return(true);
	else
		return(false);
}

/*
 * check if we're leaving a subroutine and note it before
 * we're out so the name (etc) is reported correctly.
 */
dobefore()
{
	Word instr;

	iread(&instr, reg(REG_EIP), 1);
	if ((instr&0xff) == LEAVE)
		callnews(false);
}

/*
 * check if we've just entered and note it now
 * so the name (etc) is reported correctly.
 */
doafter()
{
	Word instr;

	if (iscall(wp->p_oldpc))
		callnews(true);
}


#define BP_OP       0xcc        /* breakpoint trap (int3) */
#define BP_ERRNO    SIGTRAP     /* signal received at a breakpoint */

/*
 * Setting a breakpoint at a location consists of saving
 * the word at the location and poking a BP_OP there.
 *
 * We save the locations and words on a list for use in unsetting.
 */

#define NSAVE 100
typedef struct Savelist *Savelist;

struct Savelist {
    Address location;
    Byte save;
    Byte refcount;
    Savelist link;
};

struct machine_globals {
	Savelist machine_savelist;
	struct Savelist machine_bpsavearray[NSAVE];
};
public struct machine_globals *machine_globals_ptr;
public int machine_globals_size = sizeof(struct machine_globals);
#define	savelist	(machine_globals_ptr->machine_savelist)
#define	bpsavearray	(machine_globals_ptr->machine_bpsavearray)

/*
 * return a new Savelist entry.  a Savelist is free if its refcount is zero.
 * can't malloc() because this is called from the interrupt side.
 */
private Savelist save_alloc()
{
	register Savelist sp = bpsavearray;

	for (sp = bpsavearray; sp < bpsavearray + NSAVE; sp++)
		if (sp->refcount == 0)
			return(sp);
	fatal("out of breakpoints!");
}

/*
 * 'copy' breakpoints from one process to another.  the entries themselves
 * got copied in copy_globals().  just have to rearrange the pointers.
 * sorry about the horrible casts, especially since 'from' and 'to'
 * point to the base of the save area and not to the breakpoint save
 * array.
 */
public copy_bptsave(from, to)
	pointer from, to;
{
	pointer *spp;

	spp = (pointer *) &savelist;
	while(*spp) {
		*spp = (*spp - from) + to;
		spp = (pointer *) &((Savelist)*spp)->link;
	}
}

/*
 * Set a breakpoint at the given address.  Only save the word there
 * if it's not already a breakpoint.
 */

public setbp(addr)
Address addr;
{
    Byte w;
    Byte save;
    register Savelist newsave, s;

    for (s = savelist; s != nil; s = s->link) {
	if (s->location == addr) {
	    s->refcount++;
	    return;
	}
    }
    iread(&save, addr, sizeof(save));
    newsave = save_alloc();
    newsave->location = addr;
    newsave->save = save;
    newsave->refcount = 1;
    newsave->link = savelist;
    savelist = newsave;
    w = BP_OP;
    iwrite(&w, addr, sizeof(w));
}

/*
 * Unset a breakpoint; unfortunately we have to search the SAVELIST
 * to find the saved value.  The assumption is that the SAVELIST will
 * usually be quite small.
 */

public unsetbp(addr)
Address addr;
{
    register Savelist s, prev;

    prev = nil;
    for (s = savelist; s != nil; s = s->link) {
	if (s->location == addr) {
	    iwrite(&s->save, addr, sizeof(s->save));
	    s->refcount--;
	    if (s->refcount == 0) {	/* this means 's' is free, too */
		if (prev == nil) {
		    savelist = s->link;
		} else {
		    prev->link = s->link;
		}
	    }
	    return;
	}
	prev = s;
    }
    /*
     * breakpoint must be out
     */
}

public printbpsavelist()
{
    register Savelist s;

    for (s = savelist; s != nil; s = s->link) {
	printf("savelist 0x%x loc=0x%x save=0x%x refcount=%d\n",
		s, s->location, s->save, s->refcount);
    }
    fflush(stdout);
}

/*
 * clear out the breakpoint save list.  This should be done whenever
 * a new process is started up.
 */

public clearbps()
{
	register Savelist s;

	for (s = savelist; s != nil; s = s->link)
		s->refcount = 0;
	savelist = nil;
}

/*
 * Enter a procedure by simulating a procedure call to it, leaving the
 * process stopped at the first instruction of the procedure.  This is
 * done as if the procedure were called from 'dbsubc'.  '__dbsubc' and
 * '__dbsubn' come from libg.a (thus no call command unless linked with
 * -lg).  They are arranged such that dbsubc is the address of a procedure
 * call instruction, and dbsubn is the address of the following
 * instruction.  dbsubn is thus the address the callee will return to.
 * dbx sets an event there to handle the callee returning (see
 * callproc()).  walkstack() will check to see if the caller is dbsubc.
 * if so, it will stop the traceback there.
 * 
 * libg.a is set up so that the instruction at dbsubc is an indirect
 * subroutine call using the address at '__dbdesc'.  so beginproc could be
 * implemented by writing the address of the procedure at dbdesc, setting
 * the pc to dbsubc, and continuing to the beginning of the procedure.
 */

public beginproc(p)
Symbol p;
{
    Symbol dbsubn;
    Address retaddr;

    /*
     * find return address
     */
    dbsubn = lookup(identname("_dbsubn", true));
    if (dbsubn == nil)
	error("program not linked with `-g'");
    retaddr = codeloc(dbsubn);

    /*
     * simulate a call to 'p' from 'retaddr'
     */
    setreg(REG_ESP, reg(REG_ESP) - sizeof(retaddr));
    iwrite(&retaddr, reg(REG_ESP), sizeof(retaddr));
    setreg(REG_EIP, codeloc(p));
}


/*
 * return the length in bytes of the machine instruction at addr.
 */
public int instr_len(addr)
	Address addr;
{
	extern char *dis386();
	int length;

	(void) dis386(addr, &length);
	return(length);
}

/*
 * This function supports subroutines with multiple entry points in SVS
 * fortran.  When an entry point is called, it writes a unique 'entry id'
 * onto some location on the stack to identify which entrypoint is being
 * executed.  The address at which the entry id is stored varies.  It's
 * normally at -2(fp), but if the entrypoint has alternate returns it
 * could be somewhere else.  This routine returns the displacement off
 * the frame pointer to use to find the entry id.
 *
 * The heuristic the routine uses is based on the observation that the
 * instruction that stores the entryid is the last instruction of the
 * prologue.  The basic idea is to find the last instruction of the 
 * prologue, check that the entryid we're looking for is being stored
 * relative to the frame pointer, and extract the displacement from the 
 * frame pointer dest operand.
 */
public Word entryid_offset(entryid, prologue_addr, code_addr)
	Word	entryid;	/* entry id # for entry starting at code_addr */
	Address	prologue_addr;	/* address of beginning of entry prologue code*/
	Address	code_addr;	/* address of start of actual code for entry */
{
	Address a;		/* address of instruction we're disassembling */
	int length = 0;		/* length of instruction we last disassembled */
	char *mnem = 0;		/* mnemonic of instr we just disassembled */
	int id;			/* the id the instr we're looking for stores */
	int off;		/* the offset we're looking for */

	/*
	 * find the last instruction of the prologue
	 */
	for (a = prologue_addr; a < code_addr; a += length)
		mnem = dis386(a, &length);
	if (mnem == 0)
		return(0);

	/*
	 * We're expecting something like "movw    $2,-2(%ebp)", where
	 * the "$2" is the entryid.  We want to return the "-2".
	 * Might also see "movw    $2,-0x12(%ebp)",
	 * or "movw    $-0x13,-0x12(ebp)".
	 */
	 if (substr("movw", mnem)) {
	     if ( (sscanf(mnem, "%*[^$]$0x%x,-0x%x(%ebp)", &id, &off) == 2)
	          and (id == entryid) )
			return(-off);
	      else if ( (sscanf(mnem, "%*[^$]$0x%x,0x%x(%ebp)", &id, &off) == 2)
	          and (id == entryid) )
			return(off);
	      else if ( (sscanf(mnem, "%*[^$]$%d,-0x%x(%ebp)", &id, &off) == 2)
	          and (id == entryid) )
			return(-off);
	      else if ( (sscanf(mnem, "%*[^$]$%d,0x%x(%ebp)", &id, &off) == 2)
	          and (id == entryid) )
			return(off);
	      else if ( (sscanf(mnem, "%*[^$]$%d,%d(%ebp)", &id, &off) == 2)
	          and (id == entryid) )
			return(off);
	       else
			return(0);
	} else
		return(0);
}

/*
 * print n words from child's stack (meant to be called from debugger)
 */
xxstack( n )
	int n;
{
	Address sp;
	Word val;
	int i;

	sp = reg(REG_ESP);
	for (i = 0; i < n; i++) {
		dread(&val, sp, sizeof(Word));
		printf("%08x: %8x\n", sp, val);
		sp += sizeof(Word);
	}
	fflush(stdout);
}

public Address reg(regnum)
	int	regnum;
{
	Address i;
	check_read(wp);

	/* take care of fpa regs */
	if (regnum >= REG_FPA1 and regnum <= REG_FPA31) {
		i = (Address) wp->p_regs.pr_fpa.fpa_regs[FPA_INDEX(regnum)];
		return i;
	} 

	switch(regnum) {
	case REG_EAX:	 i = (Address) wp->p_regs.pr_eax; break;
	case REG_EDX:	 i = (Address) wp->p_regs.pr_edx; break;
	case REG_ECX:	 i = (Address) wp->p_regs.pr_ecx; break;
	case REG_EBX:	 i = (Address) wp->p_regs.pr_ebx; break;
	case REG_ESI:	 i = (Address) wp->p_regs.pr_esi; break;
	case REG_EDI:	 i = (Address) wp->p_regs.pr_edi; break;
	case REG_ESP:	 i = (Address) wp->p_regs.pr_esp; break;
	case REG_EBP:	 i = (Address) wp->p_regs.pr_ebp; break;
	case REG_EIP:	 i = (Address) wp->p_regs.pr_eip; break;
	case REG_EFLAGS: i = (Address) wp->p_regs.pr_flags; break;
	default:
		error("Internal error: bad register number %d", regnum);
	}
	return i;
}

public double freg(regnum)
	int	regnum;
{
	Word i;
	extern double t_to_d();

	check_read(wp);
	switch(regnum) {
	case REG_FP0:	i = 0; break;
	case REG_FP1:	i = 1; break;
	case REG_FP2:	i = 2; break;
	case REG_FP3:	i = 3; break;
	case REG_FP4:	i = 4; break;
	case REG_FP5:	i = 5; break;
	case REG_FP6:	i = 6; break;
	case REG_FP7:	i = 7; break;
	default:
		error("Internal error: bad register number %d", regnum);
	}
	return t_to_d(wp->p_regs.pr_fpu.fpu_stack[i]);
}

public setreg(regnum, value)
	int	regnum;
	Word	value;
{
	check_write(wp);

	/* take care of fpa regs */
	if (regnum >= REG_FPA1 and regnum <= REG_FPA31) {
		wp->p_regs.pr_fpa.fpa_regs[FPA_INDEX(regnum)] = value;
		wp->p_regschanged = true;
		return;
	} 

	switch(regnum) {
	case REG_EAX:	 wp->p_regs.pr_eax   = value; break;
	case REG_EDX:	 wp->p_regs.pr_edx   = value; break;
	case REG_ECX:	 wp->p_regs.pr_ecx   = value; break;
	case REG_EBX:	 wp->p_regs.pr_ebx   = value; break;
	case REG_ESI:	 wp->p_regs.pr_esi   = value; break;
	case REG_EDI:	 wp->p_regs.pr_edi   = value; break;
	case REG_ESP:	 wp->p_regs.pr_esp   = value; break;
	case REG_EBP:	 wp->p_regs.pr_ebp   = value; break;
	case REG_EIP:	 wp->p_regs.pr_eip   = value; break;
	case REG_EFLAGS: wp->p_regs.pr_flags = value; break;
	default:
		error("Internal error: bad register number %d", regnum);
	}
	wp->p_regschanged = true;
}

/*
 * setregfp(regnum, value)
 *
 * Like setreg, but sets floating point registers.
 */

public setregfp(regnum, value)
	int	regnum;
	double	value;
{
	int i;
	int rh;		/* high bit address */
	int rl;		/* low bit address */

	union {
		Word	di[2];
		double	dd;
	} d;

	union {
		Word	fi;
		float	ff;
	} f;

	check_write(wp);

	switch(regnum) { 
	case REG_FP0:  i = 0; break;
	case REG_FP1:  i = 1; break;
	case REG_FP2:  i = 2; break;
	case REG_FP3:  i = 3; break;
	case REG_FP4:  i = 4; break;
	case REG_FP5:  i = 5; break;
	case REG_FP6:  i = 6; break;
	case REG_FP7:  i = 7; break;
	case REG_FPA2:    case REG_FPA4:    case REG_FPA6:    case REG_FPA8:
	case REG_FPA10:   case REG_FPA12:   case REG_FPA14:  case REG_FPA16:
	case REG_FPA18:   case REG_FPA20:   case REG_FPA22:   case REG_FPA24:
	case REG_FPA26:   case REG_FPA28:   case REG_FPA30:
	    rh = regnum;		/* low 4 bytes */
	    rl = regnum + 1;		/* high 4 bytes */
	    if (varIsSet("$fpasingle")) {
		f.ff = value;
		wp->p_regs.pr_fpa.fpa_regs[FPA_INDEX(rh)] = (long) f.fi;
	    } else {
		d.dd = value;
		wp->p_regs.pr_fpa.fpa_regs[FPA_INDEX(rh)] = (long) d.di[1];
		wp->p_regs.pr_fpa.fpa_regs[FPA_INDEX(rl)] = (long) d.di[0];
		wp->p_regschanged = true;
	    }
		return;
	case REG_FPA1:    case REG_FPA3:    case REG_FPA5:    case REG_FPA7:
	case REG_FPA9:    case REG_FPA11:   case REG_FPA13:   case REG_FPA15:
	case REG_FPA17:   case REG_FPA19:   case REG_FPA21:   case REG_FPA23:
	case REG_FPA25:   case REG_FPA27:   case REG_FPA29:   case REG_FPA31:
	    rh = regnum;
	    f.ff = value;
	    wp->p_regs.pr_fpa.fpa_regs[FPA_INDEX(rh)] = (long) f.fi;
	    wp->p_regschanged = true;
	    return;
	default:
		error("Internal error: bad register number %d", regnum);
	}
	wp->p_regschanged = true;
	d_to_t(value, wp->p_regs.pr_fpu.fpu_stack[i]);
}

/* Don't change value of T387_OK.  See later note. */

#define T387_OK		0 	/*  Normal or zero, ok to convert.  */
#define T387_PNAN	1	/*  + NaN (not a number)  */
#define T387_PINF	2	/*  + infinity  */
#define T387_PUNNORM	3	/*  + unnormals  */
#define T387_PDNORM	4	/*  + denormals  */
#define T387_NUNNORM	5	/*  - unnormals  */
#define T387_NDNORM	6	/*  - denormals  */
#define T387_NINF	7	/*  - infinity  */
#define T387_NNAN	8	/*  - NaN (not a number)  */
#define T387_OVERF	9	/*   overflow   (if converted to 64 bit fp val) */
#define T387_UNDERF	10	/*   underflow  (if converted to 64 bit fp val) */

/*
 * These routines are used to display the floating point stack for
 * the 80387 floating point coprocessor.
 *
 * Figuring out the state of the registers is a little tricky.  The
 * tag word provides the state of each register.  We have to index into
 * it relative to top of stack.  Even though the registers are dumpped
 * from memory the same way regardless of tos, and the last value is
 * always st0, the tag information for the registers shifts depending
 * on the top of stack value.
 */

typedef enum {		/* 80386 tag word format codes */
	TAG_VALID,
	TAG_ZERO,
	TAG_INVAL,
	TAG_EMPTY
} Tagcode;

private Tagcode tag_387(tag_word, tag_pos, tag_string)
	int tag_word, tag_pos;
	char **tag_string;
{
	int n;

	n = tag_word >> (tag_pos * 2);
	n &= 0x3;
	switch (n) {
	case 0x0:				/* 00 */
		*tag_string = "valid";
		return(TAG_VALID);
	case 0x1:				/* 01 */
		*tag_string = "zero ";
		return(TAG_ZERO);
	case 0x2:				/* 10 */
		*tag_string = "inval";
		return(TAG_INVAL);
	case 0x3:				/* 11 */
		*tag_string = "empty";
		return(TAG_EMPTY);
	}
	/*NOTREACHED*/
}

public printfpustack()
{
	register int i, j;
	int tos_ptr, r;
	Tagcode tag;
	char *tag_string;
	double d;
	extern double t_to_d();

    /* See previous comment about tos pointer calc in tag word */
    tos_ptr = (wp->p_regs.pr_fpu.fpu_status >> 11) & 0x7;
    r = tos_ptr - 1;
    if (r < 0)
	r = 7;
    for (i=7; i >= 0; i--) {
	tag = tag_387(wp->p_regs.pr_fpu.fpu_tag, r, &tag_string);
	if (varIsSet("$fpustackall") or (tag != TAG_EMPTY)) {
	    printf(" st(%d)= ", i);
	    for (j=4; j >= 0; j--)
		printf(" %04x", wp->p_regs.pr_fpu.fpu_stack[i][j]);
	    printf(" [%s]", tag_string);
	    if ( (j = temp_387_code(wp->p_regs.pr_fpu.fpu_stack[i])) == 0 ) { 
	        d = t_to_d(wp->p_regs.pr_fpu.fpu_stack[i]);
	        printf(" = %-18.18g\n", d);
	    } else {
		if (j == T387_OVERF or j == T387_UNDERF)
		    printf(" <%s>, %s\n", show_387_code(j),
			conv_387_num(wp->p_regs.pr_fpu.fpu_stack[i]));
		else
		    printf(" <%s>\n", show_387_code(j)); 
	    }
	}
	r -= 1;
	if (r < 0)
	    r = 7;
    }
    printf("control word= 0x%04x  ", wp->p_regs.pr_fpu.fpu_control);
    printf("status= 0x%04x  ", wp->p_regs.pr_fpu.fpu_status);
    printf("tag= 0x%04x  ", wp->p_regs.pr_fpu.fpu_tag);
    printf("ip= 0x%08x ", wp->p_regs.pr_fpu.fpu_ip);
    printf("tos= %d\n", tos_ptr);
}

/*
 * Care must be used when converting from an 80 bit temporary to a floating
 * point number.  If the 80 bit temporary is not a "normalized" number an
 * exception may occur.  Code allows testing this.
 *
 * This routine returns an int which is a code for the type of 80 bit
 * temporary.  NOTE:  Code returned by this routine is checked in assem
 * routine "t_to_d()", see "fpu387.s".  Zero must remain "ok" return.
 */

public int temp_387_code(temp_reg)
    u_short temp_reg[5];
{
	union {
		struct {
			unsigned significand_low  : 32;
			unsigned significand_high : 31;
			unsigned significand_top  :  1;
			unsigned exponent         : 15;
			unsigned sign_bit         :  1;
		} tmp_80_bits;
		unsigned short tmp_80_array[5];
	} tmp_387;

	int i, n;

	n = 0;
	for (i=0; i < 5; i++) {
		tmp_387.tmp_80_array[i] = temp_reg[i];
		if (temp_reg[i] != 0x0)
			n++;
	}

	if (n == 0)			/* zero */
		return(T387_OK);
	
	/* Test for + NaNs		return code 1 */

	if ( (tmp_387.tmp_80_bits.sign_bit         == 0x0     and
	      tmp_387.tmp_80_bits.exponent         == 0x7fff  and
	      tmp_387.tmp_80_bits.significand_top  == 0x1         ) and 
	     (tmp_387.tmp_80_bits.significand_high != 0x0   or
	      tmp_387.tmp_80_bits.significand_low  != 0x0 ) )
			return(T387_PNAN);

	/* Test for + infinity		return code 2 */

	if ( tmp_387.tmp_80_bits.sign_bit         == 0x0     and
	     tmp_387.tmp_80_bits.exponent         == 0x7fff  and
	     tmp_387.tmp_80_bits.significand_top  == 0x1     and
	     tmp_387.tmp_80_bits.significand_high == 0x0     and
	     tmp_387.tmp_80_bits.significand_low  == 0x0         )
			return(T387_PINF);

	/* Test for + unnormals		return code 3 */

	if   (tmp_387.tmp_80_bits.sign_bit         == 0x0  and
	      tmp_387.tmp_80_bits.exponent         != 0x0  and
	      tmp_387.tmp_80_bits.significand_top  == 0x0      )
			return(T387_PUNNORM);

	/* Test for + denormals		return code 4 */

	if ( (tmp_387.tmp_80_bits.sign_bit         == 0x0     and
	      tmp_387.tmp_80_bits.exponent         == 0x0     and
	      tmp_387.tmp_80_bits.significand_top  == 0x0         ) and 
	     (tmp_387.tmp_80_bits.significand_high != 0x0   or
	      tmp_387.tmp_80_bits.significand_low  != 0x0 ) )
			return(T387_PDNORM);

	/* Test for - unnormals		return code 5 */

	if   (tmp_387.tmp_80_bits.sign_bit         == 0x1     and
	      tmp_387.tmp_80_bits.exponent         != 0x0     and
	      tmp_387.tmp_80_bits.significand_top  == 0x0 )
			return(T387_NUNNORM);

	/* Test for - denormals		return code 6 */

	if ( (tmp_387.tmp_80_bits.sign_bit         == 0x1     and
	      tmp_387.tmp_80_bits.exponent         == 0x0     and
	      tmp_387.tmp_80_bits.significand_top  == 0x0         ) and 
	     (tmp_387.tmp_80_bits.significand_high != 0x0   or
	      tmp_387.tmp_80_bits.significand_low  != 0x0 ) )
			return(T387_NDNORM);

	/* Test for - infinity		return code 7 */

	if ( (tmp_387.tmp_80_bits.sign_bit         == 0x0     and
	      tmp_387.tmp_80_bits.exponent         == 0x7fff  and
	      tmp_387.tmp_80_bits.significand_top  == 0x1         ) and 
	     (tmp_387.tmp_80_bits.significand_high != 0x0   or
	      tmp_387.tmp_80_bits.significand_low  != 0x0 ) )
			return(T387_NINF);

	/* Test for - NaNs		return code 8 */

	if ( (tmp_387.tmp_80_bits.sign_bit         == 0x0     and
	      tmp_387.tmp_80_bits.exponent         == 0x7fff  and
	      tmp_387.tmp_80_bits.significand_top  == 0x1         ) and 
	     (tmp_387.tmp_80_bits.significand_high != 0x0   or
	      tmp_387.tmp_80_bits.significand_low  != 0x0 ) )
			return(T387_NNAN);

	/* Check for zero exponent, this is ok. */

	if (tmp_387.tmp_80_bits.exponent == 0x0)
		return(T387_OK);

	/* Test for overflow/underflow		return code 9/10 */

	n = (int) tmp_387.tmp_80_bits.exponent;
	n -= 16383;	/* subtract bias for 80 bit floating point num */
	n += 1023;	/* add bias for 64 bit floating point num */
	if (n >= 0x7ff)	/* will it fit in smaller field of 64 bit num? */
		return(T387_OVERF);
	if (n < 0x0)
		return(T387_UNDERF);

	return(T387_OK);
}

/*
 * Convert 80 bit temp to a floating point representation.
 */

public String conv_387_num(temp_reg)
    u_short temp_reg[5];
{
	union {
		struct {
			unsigned significand_low  : 32;
			unsigned significand_high : 31;
			unsigned significand_top  :  1;
			unsigned exponent         : 15;
			unsigned sign_bit         :  1;
		} tmp_80_bits;
		unsigned short tmp_80_array[5];
	} tmp_387;

	int i, n, s, x;
	double e, f, a;
	double l = 0.301029996;			/* log 2 10 */
	extern double t_to_d();
	char buf[1024], *p;

	/* copy array into struct for bit fiddling */
	n = 0;
	for (i=0; i < 5; i++) {
		tmp_387.tmp_80_array[i] = temp_reg[i];
		if (temp_reg[i] != 0x0)
			n++;
	}

	if (n == 0)
		return("0.0 E+0");
	
	/* save sign of number */
	if (tmp_387.tmp_80_bits.sign_bit == 1)
		s = 1;		/* negative number */
	else
		s = 0;		/* positive number */
	tmp_387.tmp_80_bits.sign_bit = 0;

	/* subtract normalizaiton from exponent */
	x = (int) tmp_387.tmp_80_bits.exponent - 16383;

	/* convert exponent to powers of 10 as we wish to display	*/
	/* multiply exponent by log 2 10 to convert to base pwr of 10	*/
	/*   x = exponent						*/
	/*   f = fraction of exponent multipled with a			*/
	f = (double) x * l;
	x = f;
	f = f - (double) x;
	f = pow(10.0, f);

	/* replace exponent in number with exponent for 1.0	*/
	/* a = base floating point value displayed */
	tmp_387.tmp_80_bits.exponent = 0x3fff;
	a = t_to_d(tmp_387.tmp_80_array);
	a *= f;

	/* now generate string for display */
	p = buf;
	if (s == 1)
		*p++ = '-';
	sprintf(p, "%.8f", a);
	while (*p)
		p++;
	*p++ = ' ';
	*p++ = 'E';
	if (x >= 0)
		*p++ = '+';
	sprintf(p, "%d", x);
	return(buf);
}


/*
 * Display the codes used to indicate type of 80 bit temporary.
 */

public String show_387_code(n)
{
	char buf[100];

	switch (n) {
	case T387_OK:
		return("normal number");
	case T387_PNAN:
		return("positive nan (not a number)");
	case T387_PINF:
		return("positive infinity");
	case T387_PUNNORM:
		return("positive unnormailzed number");
	case T387_PDNORM:
		return("positive denormalized number");
	case T387_NUNNORM:
		return("negative unnormalized  number");
	case T387_NDNORM:
		return("negative denormalized number");
	case T387_NINF:
		return("negative infinity");
	case T387_NNAN:
		return("negative nan (not a number)");
	case T387_OVERF:
		return("overflow");
	case T387_UNDERF:
		return("underflow");
	default:
		sprintf(buf, "unknown 387 code %d", n);
		return(buf);
	}
}

/*
 * This routine is called when an floating point exception occurs.
 */

/*ARGSUSED*/
public catchfpe(sig, code)
    int sig, code;
{
    int i;

    if (code & FPE_FLTINV_TRAP) {
        printf("%s", fpenames[i].fpe_name);
	code &= ~FPE_FLTINV_TRAP;
	if (code == 0) {
	    printf("\n");
	    fflush(stdout);
	    erecover();
	} else
	    printf(" & ");
    }
    switch (code) {
	case FPE_FLTINV_TRAP:	i = 0; break;
	case FPE_FLTDEN_TRAP:	i = 1; break;
	case FPE_FLTDIV_TRAP:	i = 2; break;
	case FPE_FLTOVF_TRAP:	i = 3; break;
	case FPE_FLTUND_TRAP:	i = 4; break;
	case FPE_FLTPRE_TRAP:	i = 5; break;
	case FPE_FLTSTK_TRAP:	i = 6; break;
	case FPE_INTDIV_TRAP:	i = 7; break;
	case FPE_INTOVF_TRAP:	i = 8; break;
	case FPE_NOFPA_AVAIL:	i = 9; break;
	default:
	    printf("unknown fpe trap (0x%x)", code);
	    i = -1;
    }
    if (i >= 0)
        printf("%s\n", fpenames[i].fpe_name);
    fflush(stdout);
    erecover();
}

/*
 * isfreg8(s)
 *   Symbol s;
 *
 * Is symbol an 8 byte floating point value held in a register?
 */

public boolean isfreg8(s)
	Symbol s;
{
	int r;

	if (isreg(s)) {
		r = regnum(s);
		/* take care of fpa regs */
		if (r >= REG_FPA1 and r <= REG_FPA31) {
			if (size(s) == 8)
				return true;
			else
				return false;
		} 
		switch(r) {
		case REG_FP0:
		case REG_FP1:
		case REG_FP2:
		case REG_FP3:
		case REG_FP4:
		case REG_FP5:
		case REG_FP6:
		case REG_FP7:
			return true;
		}
	}
	return false;
}

/*
 * Special print for floating point registers.
 * Routines decode to a register # the strings $st0 .. $st7 and $fp1 .. $fp31.
 *
 * Returned register number is the MACHINE # (ie st0 returns REG_FP0).
 *
 * Weitek registers print double precision (if even numbered) unless the
 * variable $fpasingle is set.  Normal display is floating point.  Can be
 * changed to hex by setting $hexfloats.
 */

private int get_387_reg(reg_name)
	String reg_name;
{
	int n, i;
	i = sscanf(reg_name, "$st%d", &n);
	if (i == 1)
		switch (n) {
		case 0:  return(REG_FP0);
		case 1:  return(REG_FP1);
		case 2:  return(REG_FP2);
		case 3:  return(REG_FP3);
		case 4:  return(REG_FP4);
		case 5:  return(REG_FP5);
		case 6:  return(REG_FP6);
		case 7:  return(REG_FP7);
		default:
			return(-1);
		}
	else
		return(-1);
}

private int get_1167_reg(reg_name)
	String reg_name;
{
	int n, i;
	i = sscanf(reg_name, "$fp%d", &n);
	if (i == 1 and n >= 1 and n <= 31)
		return(n + FPA_BASE - 1);
	else
		return(-1);
}

/*
 * print_fpu_reg(name)
 *
 * Pass this routine a string which may be a floating point regiser name.
 * This returns true if it was, and will print the value of the register.
 * Returns false if the string didn't match a floating point register name.  
 *
 * This is called from eval() for case O_PRINT.  In eval, it has been
 * detected that the user has called print with one argument which is a
 * symbol.  This the case of a single floating point register and indicates
 * that this routine has done the necessary display by returning "true".
 * If it returns "false" then the normal code in eval will be used.
 */

public boolean print_fpu_reg(reg_name)
	String reg_name;
{
	int n;

	n = get_387_reg(reg_name);
	if (n != -1) {
		print_387_reg(n);
		return(true);
	}

	n = get_1167_reg(reg_name);
	if (n != -1) {
		print_1167_reg(n);
		return(true);
	}

	return(false);
}

/*
 * The 387 registers are stored in 80 bit temporary format.  No problem
 * distinguishing between those used for "float" or "double" since we are
 * able to pick up the temporary value.  We want to check that we can
 * correctly convert from 80 bit to 64 bit though.
 */

private print_387_reg(reg_num)
	int reg_num;
{
	register int i, j, t_code;
	union {
		Address di[2];
		u_short ds[4];
		double  dd;
	} dbl;

	check_read(wp);
	switch (reg_num) {
	case REG_FP0:	i = 0; break;
	case REG_FP1:	i = 1; break;
	case REG_FP2:	i = 2; break;
	case REG_FP3:	i = 3; break;
	case REG_FP4:	i = 4; break;
	case REG_FP5:	i = 5; break;
	case REG_FP6:	i = 6; break;
	case REG_FP7:	i = 7; break;
	default:
		error("Internal error: bad register number %d", reg_num);
	}

	t_code = temp_387_code(wp->p_regs.pr_fpu.fpu_stack[i]);

	if (t_code != T387_OK) {
	    printf("st(%d)= ", i);
	    for (j=4; j >= 0; j--)
		printf(" %04x", wp->p_regs.pr_fpu.fpu_stack[i][j]);
		if (t_code == T387_OVERF or t_code == T387_UNDERF)
		    printf(" <%s>, %s", show_387_code(t_code),
			conv_387_num(wp->p_regs.pr_fpu.fpu_stack[i]));
		else
		    printf(" <%s>", show_387_code(j)); 
	} else {
	    dbl.dd = freg(reg_num);
	    if (varIsSet("$hexfloats")) {
		for (i=3; i >= 0; i--)
		    printf("%04x ", dbl.ds[i]);
	    } else {
		printf("%-.18g", dbl.dd);
	    }
	}
}

/*
 * Display of the 1167 registers is made difficult becuase we don't really 
 * know how they are being used.  Single registers can hold "floats" and
 * two register pairs can hold "doubles".  If user specifies an odd
 * register (ie. $fp3, $fp5, etc) then we display as single precision.
 * Even registers (ie. $fp2, $fp4, etc) default to double precision unless
 * the variable $fpasingle is set.
 *
 * Note that double precision numbers are held such that the low order bits
 * are in the higher numbered register.  So:
 *       $fp2 -- high order bits
 *       $fp2 -- low order bits
 *
 * The register number passed in is the number used to locate the register in
 * in the save area.  So $fp2 == REG_FPA2 == 19.  This means that an odd value
 * for "reg_num" is an even fpa register. 
 */

private print_1167_reg(reg_num)
	int reg_num;
{
	register int i;
	Address a;
	union {
		Address di[2];
		u_short ds[4];
		double  dd;
	} dbl;
	union {
		Address fi;
		u_short fs[2];
		float   ff;
	} flt;

	if (varIsSet("$fpasingle") or ((FPA_INDEX(reg_num) & 0x1) == 0)) {
		flt.fi = reg(reg_num);
		if (varIsSet("$hexfloats")) {
			printf("%04x %04x", flt.fs[1], flt.fs[0]);
		} else {
			printf("%-.18g", flt.ff);
		}
	} else {
		dbl.di[1] = reg(reg_num);
		dbl.di[0] = reg(reg_num + 1);
		if (varIsSet("$hexfloats")) {
			for (i=3 ; i >= 0; i--)
				printf("%04x ", dbl.ds[i]);
		} else {
			printf("%-.18g", dbl.dd);
		}
	}
}

/*
 * This looks a the name of a symbol, returns true if it is one of
 * the floating point registers.
 */

public boolean isfpreg(s)
Symbol s;
{
	int i;

	if (s->name == nil)
		return(false);

	if (get_1167_reg(ident(s->name)) != -1)
		return(true);

	if (get_387_reg(ident(s->name)) != -1)
		return(true);
	
	return(false);
}

/*
 * This returns a double, the value of a floating point register.
 * If $fpasingle is set then this will pick up a single precision
 * number when possible (still returns double).
 */

public double get_fpu_reg(reg_num)
	int reg_num;
{
	union {
		Address di[2];
		u_short ds[4];
		double  dd;
	} dbl;
	union {
		Address fi;
		u_short fs[2];
		float   ff;
	} flt;

	/* Return 1167 registers. */

	if (reg_num >= REG_FPA1 and reg_num <= REG_FPA31) {
		if (varIsSet("$fpasingle") or ((FPA_INDEX(reg_num) & 0x1) == 0)) {
			flt.fi = reg(reg_num);
			return( (double) flt.ff);
		} else {
			dbl.di[1] = reg(reg_num);
			dbl.di[0] = reg(reg_num + 1);
			return(dbl.dd);
	    	}
	}

	/* Return 387 registers -- this does error checking. */

	return(freg(reg_num));
}

/*
 * See if number is one of the internal fpu registers.
 */

public boolean is_fpu_reg_num(n)
	int n;
{
	if (n >= REG_FPA1 and n <= REG_FPA31)
		return(true);
	
	switch (n) {
	case REG_FP0: case REG_FP1:
	case REG_FP2: case REG_FP3:
	case REG_FP4: case REG_FP5:
	case REG_FP6: case REG_FP7:
		return (true);
	default:
		return(false);
	}
}

/*
 * Return a string for a register name.
 */

String bad_reg = "??";

public String get_reg_name(r)
	int r;
{
	if (r < REG_EAX or r >= NREG)
		return(bad_reg);
	else
		return(regname[r]);
}

/*
 * Indicate if symbol is a reg name.
 */

public boolean isregname(s)
	Symbol s;
{
	register int i;
	char *c;

	c = symname(s);

	if (*c != '$')
		return(false);
	
	c++;
	if (*c == '\0')
		return(false);
	
	/* take care of the "aliases" */

	if (streq(c, "pc") or streq(c, "sp") or streq(c, "fp"))
		return(true);
	
	for (i=0; i < NREG; i++)
		if (streq(c, regname[i]))
			return(true);
	
	return(false);
}

