/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef lint
static char rcsid[] = "$Header: runtime.c 1.27 87/11/03 $";
#endif

/*
 * Runtime organization dependent routines, mostly dealing with
 * activation records.
 */

#include "defs.h"
#include "main.h"
#include "runtime.h"
#include "multi-p.h"
#include "machine.h"
#include "events.h"
#include "mappings.h"
#include "symbols.h"
#include "tree.h"
#include "eval.h"
#include "operators.h"
#include "object.h"
#include "library.h"
#include <signal.h>

#ifndef public
typedef struct Frame *Frame;

#include "machine.h"

#define NSAVEREG  (REG_FPA31 + 1)	/* 0 .. edi, FPA1 .. FPA31 */
#define NSAVEFREG 14		/* 0 .. st(7) */
#define NSAVEWREG 31		/* fp1 .. fp31 */
#define ARGOFF	  8		/* distance from fp to first arg */

struct Frame {
    Address save_fp;		/* frame pointer */
    Address save_pc;		/* program counter */
    Word save_reg[NSAVEREG];	/* not necessarily there */
    double save_freg[NSAVEFREG];/* 387 floating point registers */
    Integer r_mask;		/* register save mask, gen and 387 fp regs */
    Integer w_mask;		/* Wytek regiser save mask */
    Integer nlocals;		/* size of local area */
    Symbol block;		/* block this frame is associated with */
    boolean entered;		/* true if entry prologue executed */
    boolean iscaller;		/* frame doing subroutine call */
    boolean stringsave;		/* is string move used to save weitek regs */
    boolean no_frameptr;	/* does not push frame ptr */
};

#endif

public boolean debug_regsave;
public boolean debug_frame;

struct runtime_globals {
	Frame runtime_curframe;
	struct Frame runtime_curframerec;
};
public struct runtime_globals *runtime_globals_ptr;
public int runtime_globals_size = sizeof(struct runtime_globals);
#define	curframe	(runtime_globals_ptr->runtime_curframe)
#define	curframerec	(runtime_globals_ptr->runtime_curframerec)

private Boolean walkingstack = false;

#define frameeq(f1, f2) ((f1)->save_fp == (f2)->save_fp)

private boolean callisfunc;	/* is procedure being called a function?
				   save/restored by pushenv/popenv */

/*
 * this means 4-byte floating point type
 */
#define isfloat(s) 					\
	       ( (s)->symvalue.rangev.upper == 0 and	\
	         (s)->symvalue.rangev.lower == 4    )

/*
 * return whether we've entered the function for the frame yet
 */
public boolean isentered(f)
Frame f;
{
    if (f != nil)
        return(f->entered);
    else
	return(false);
}

/*
 * See if we are in the setup code.  This is where we have the PC in a
 * sub but haven't set up the frame yet.  See code illustration in
 * procedure "getcurframe()".  Return true if in entry code.
 */

private boolean is_in_entry(frp, func)
register Frame frp;
register Symbol func;
{
	Word	w;
	Address	addr;

	/*
	 * If we are on the first instruction of the routine
	 * we assume we are in entry and no set up yet.
	 */

	if (entryloc(func) == frp->save_pc)
		return(true);
	
	/*
	 * Now check for PC sitting in code sequence:
	 *
	 *     pushl   %ebp             0x55
	 *     movl    %esp,%ebp        0x8bec
	 */

#define	PUSH_SETUP	0x55		/* push  %ebp */
#define MOVL_SETUP	0xec8b		/* movl  %esp,%ebp */
	
	addr = frp->save_pc;
	dread(&w, addr, sizeof(Word));

	if ( (w & 0xff)  == PUSH_SETUP) {
		addr += sizeof(Byte);
		dread(&w, addr, sizeof(Word));
		if ( (w & 0xffff) == MOVL_SETUP)
			return(true);
		else
			return(false);
	} else if ( (w & 0xffff) == MOVL_SETUP) {
		addr -= sizeof(Byte);
		dread(&w, addr, sizeof(Word));
		if ( (w & 0xff)  == PUSH_SETUP)
			return(true);
		else
			return(false);
	} else
		return(false);		/* not in setup code */
}

/*
 * Set a frame to the current activation record.
 */

private getcurframe(frp)
register Frame frp;
{
    register int i;
    register Symbol func;
    Boolean string_move;
    Boolean no_frame_setup;

    checkref(frp);
    frp->save_pc = reg(REG_EIP);
    frp->save_reg[REG_EAX] = reg(REG_EAX);
    frp->save_reg[REG_EDX] = reg(REG_EDX);
    frp->save_reg[REG_ECX] = reg(REG_ECX);
    frp->save_reg[REG_EBX] = reg(REG_EBX);
    frp->save_reg[REG_ESI] = reg(REG_ESI);
    frp->save_reg[REG_EDI] = reg(REG_EDI);
    frp->save_reg[REG_FPA1] = reg(REG_FPA1);
    frp->save_reg[REG_FPA2] = reg(REG_FPA2);
    frp->save_reg[REG_FPA3] = reg(REG_FPA3);
    frp->save_reg[REG_FPA4] = reg(REG_FPA4);
    frp->save_reg[REG_FPA5] = reg(REG_FPA5);
    frp->save_reg[REG_FPA6] = reg(REG_FPA6);
    frp->save_reg[REG_FPA7] = reg(REG_FPA7);
    frp->save_reg[REG_FPA8] = reg(REG_FPA8);
    frp->save_reg[REG_FPA9] = reg(REG_FPA9);
    frp->save_reg[REG_FPA10] = reg(REG_FPA10);
    frp->save_reg[REG_FPA11] = reg(REG_FPA11);
    frp->save_reg[REG_FPA12] = reg(REG_FPA12);
    frp->save_reg[REG_FPA13] = reg(REG_FPA13);
    frp->save_reg[REG_FPA14] = reg(REG_FPA14);
    frp->save_reg[REG_FPA15] = reg(REG_FPA15);
    frp->save_reg[REG_FPA16] = reg(REG_FPA16);
    frp->save_reg[REG_FPA17] = reg(REG_FPA17);
    frp->save_reg[REG_FPA18] = reg(REG_FPA18);
    frp->save_reg[REG_FPA19] = reg(REG_FPA19);
    frp->save_reg[REG_FPA20] = reg(REG_FPA20);
    frp->save_reg[REG_FPA21] = reg(REG_FPA21);
    frp->save_reg[REG_FPA22] = reg(REG_FPA22);
    frp->save_reg[REG_FPA23] = reg(REG_FPA23);
    frp->save_reg[REG_FPA24] = reg(REG_FPA24);
    frp->save_reg[REG_FPA25] = reg(REG_FPA25);
    frp->save_reg[REG_FPA26] = reg(REG_FPA26);
    frp->save_reg[REG_FPA27] = reg(REG_FPA27);
    frp->save_reg[REG_FPA28] = reg(REG_FPA28);
    frp->save_reg[REG_FPA29] = reg(REG_FPA29);
    frp->save_reg[REG_FPA30] = reg(REG_FPA30);
    frp->save_reg[REG_FPA31] = reg(REG_FPA31);
    frp->save_freg[REG_FP0] = freg(REG_FP0);
    frp->save_freg[REG_FP1] = freg(REG_FP1);
    frp->save_freg[REG_FP2] = freg(REG_FP2);
    frp->save_freg[REG_FP3] = freg(REG_FP3);
    frp->save_freg[REG_FP4] = freg(REG_FP4);
    frp->save_freg[REG_FP5] = freg(REG_FP5);
    frp->save_freg[REG_FP6] = freg(REG_FP6);
    frp->save_freg[REG_FP7] = freg(REG_FP7);
    frp->block = func = whatblock(frp->save_pc);

    /*
     * two possible cases: either the frame is set up already, 
     * or we fake it.  If pc appears to be in entry code we fake it.
     * Two entry sequences we allow:
     *
     *   sub:    jmp   Lxx2
     *   Lxx1:   ...
     *           ....
     *   Lxxx:   pushl   %ebp
     *           movl    %esp,%ebp
     *           subl    $nn,%esp
     *           jmp     Lxx1
     * 
     *   sub:     pushl   %ebp
     *            movl    %esp,ebp
     *            ...
     */

    if ( not is_in_entry(frp, func) or isinline(func)) {
	frp->entered = true;
	frp->save_fp = reg(REG_EBP);
	if (debug_regsave)
	    printf("getcurframe: entered, save_fp= 0x%x\n", frp->save_fp);
	if (func->symvalue.funcv.have_rmasks) {
	    frp->r_mask = func->symvalue.funcv.r_mask;
	    frp->w_mask = func->symvalue.funcv.w_mask;
	    frp->nlocals = func->symvalue.funcv.nlocals;
	    frp->stringsave = func->symvalue.funcv.stringsave;
	    frp->no_frameptr = func->symvalue.funcv.no_frameptr;
		if (debug_regsave) {
		    printf("getcurframe: had reg masks, r_mask= 0x%x, w_mask= 0x%x\n",
			frp->r_mask, frp->w_mask);
		    printf("getcurframe: nlocals= %d, stringsave= %s, no_fp= %s\n",
		        frp->nlocals, frp->stringsave ? "T" : "F",
			frp->no_frameptr ? "T" : "F");
		}
	} else {
	    if (debug_regsave)
		printf("getcurframe: calling get_enter_args\n");
	    get_enter_args(codeloc(func), &frp->r_mask, &frp->w_mask,
	   	 &frp->nlocals, &string_move, &no_frame_setup);
	    func->symvalue.funcv.r_mask = frp->r_mask;
	    func->symvalue.funcv.w_mask = frp->w_mask;
	    func->symvalue.funcv.nlocals = frp->nlocals;
	    func->symvalue.funcv.have_rmasks = true;
	    func->symvalue.funcv.stringsave = string_move;
	    func->symvalue.funcv.no_frameptr = no_frame_setup;
	    frp->stringsave = string_move;
	    frp->no_frameptr = no_frame_setup;
	}
    } else {
	frp->entered = false;
	frp->save_fp = reg(REG_ESP) - sizeof(Address);	/* where it would be */
	frp->no_frameptr = false;
	frp->stringsave = false;
	frp->r_mask = 0;				/* nothing saved yet */
	frp->w_mask = 0;
	frp->nlocals = 0;
	if (debug_regsave)
	    printf("getcurframe: not entered, save_fp= 0x%x\n", frp->save_fp);
    }

    frp->iscaller = false;
}

/*
 * Return a pointer to the next activation record up the stack.
 * Return nil if there is none.
 * Writes over space pointed to by given argument.
 *
 *					high memory
 *        |-----------|
 *        | ret addr  |  4(ebq)
 *        |-----------|
 *        | old fp    |  0(ebp)   (fp)
 *        |-----------|
 *        |  locals   |  -4(ebp)
 *        |--- ... ---|			Space for local variables placed
 *        |  locals   |			on the stack.
 *        |-----------|
 *        |   fp7     |			Floating point registers placed on the
 *        |-----------|			stack, FP7...FP0.
 *        |   fp6     |
 *        |-----------|
 *        |   edi     |			Registers EDI, ESI, and EBX placed
 *        |-----------|			on the stack (if used).
 *        |   esi     |
 *        |-----------|
 *        |   ebx     |
 *        |-----------|
 *        |           | <-  esp   (sp)
 *						low memory
 *
 * Note that if Weitek FPA is in use the FPA registers will be on the stack
 * where the 387 floating point registers are.  It is very unlikely that
 * there could be both, but the code in nextframe() will handle that as long
 * as they are on the stack "on top of" the 387 registers (higher memory).
 */

private Frame nextframe(frp)
    register Frame frp;
{
    Address regaddr;		/* address of saved registers in next frame */
    register int i, r;
    struct sigcontext *scp;	/* signal context pointer to old pc */
    register Symbol func;
    int fp_bytes = 0;
    Boolean string_move;
    Boolean no_frame_setup;

    struct stackframe {		/* what a stack frame looks like */
	Address	return_fp;
	Address	return_pc;
	Word	arg[1];
    } *fp;			/* frame pointer to next frame */

    /*
     * Need to know when to stop a stack walk back.  For C and Fortran
     * this is indicated when the frame pointer (epb) points to a zero.
     * The main routine in Pascal saves ebp on the stack and loads it
     * with the value of //GLOBALS//.  So we have to check for this
     * value as a possible terminating condition.  We save the value in
     * "check_global" since it will be deleted in "check_var".
     */

    if (debug_frame)
	printf("\nnextframe: frp->save_fp= 0x%x\n", frp->save_fp);
    if (frp->save_fp == 0 or frp->block == program
      or frp->save_fp == pas_globals_addr) {
	if (debug_frame) {
	    printf("nextframe: stack walk back stopped by ");
	    if (frp->save_fp == 0)
		printf("frame pointer at 0\n");
	    else if (frp->block == program)
		printf("frame block in 'program'\n");
	    else
		printf("frame pointer at 'pas_globals_addr'\n");
	}
		return nil;
    }

    /*
     * Update registers saved in current frame.
     *
     * Note that value of "frp->nlocals" is calculated by the space the
     * prologue code allocates on the stack.  If floating point registers
     * are saved this value includes storage for those registers, since
     * we put them on the stack with fstl (floating store) instead of
     * pushes as is done for other registers. 
     *
     * We are trying to calculate "regaddr" which will be the start of the
     * saved registers.  This is going to be an offset from the saved frame
     * pointer.  After calculation of this location we can pick up the
     * registers as if they were all saved with pushes on the stack.
     *
     * Calcualte value by subtracting number of bytes subtraced by prologue
     * code then add back 8 bytes per saved floating point register.
     */

#define bis(b, n)		(((b) & (1 << (n))) != 0)
#define fbis(b, n)		(((b) & (1 << ((n)-FPA_BASE))) != 0)

  if (frp->r_mask != 0) {
    if (bis(frp->r_mask, REG_FP7)) fp_bytes += 8;
    if (bis(frp->r_mask, REG_FP6)) fp_bytes += 8;
    if (bis(frp->r_mask, REG_FP5)) fp_bytes += 8;
    if (bis(frp->r_mask, REG_FP4)) fp_bytes += 8;
    if (bis(frp->r_mask, REG_FP3)) fp_bytes += 8;
    if (bis(frp->r_mask, REG_FP2)) fp_bytes += 8;
    if (bis(frp->r_mask, REG_FP1)) fp_bytes += 8;
    if (bis(frp->r_mask, REG_FP0)) fp_bytes += 8;
  }

  if (frp->w_mask != 0) {
    if (fbis(frp->w_mask, REG_FPA1)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA2)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA3)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA4)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA5)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA6)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA7)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA8)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA9)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA10)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA11)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA12)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA13)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA14)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA15)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA16)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA17)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA18)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA19)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA20)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA21)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA22)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA23)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA24)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA25)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA26)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA27)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA28)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA29)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA30)) fp_bytes += 4;
    if (fbis(frp->w_mask, REG_FPA31)) fp_bytes += 4;
  }

    if (debug_regsave)
	printf("nextframe: r_mask= 0x%x, w_mask= 0x%x, nlocals= %d, fp_bytes= %d\n",
	    frp->r_mask, frp->w_mask, frp->nlocals, fp_bytes);

    /*
     * Change to method of saving floating point registers
     * so we no longer subtract "fp_bytes" from register
     * address.  Old code was:
     *
     *   regaddr = frp->save_fp - frp->nlocals + fp_bytes;
     */

    regaddr = frp->save_fp - frp->nlocals;
    if (debug_regsave or debug_frame)
	printf("nextframe: regaddr= 0x%x, save_fp= 0x%x\n",
	    regaddr, frp->save_fp);

    /*
     * Now pick up registers, decrementing "regaddr" (sp) by the size of the
     * variable pulled off the stack and put into a frame variable.
     */

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA31)) dread(&frp->save_reg[REG_FPA31], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA30)) dread(&frp->save_reg[REG_FPA30], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA30)) dread(&frp->save_reg[REG_FPA30], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA31)) dread(&frp->save_reg[REG_FPA31], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA29)) dread(&frp->save_reg[REG_FPA29], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA28)) dread(&frp->save_reg[REG_FPA28], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA28)) dread(&frp->save_reg[REG_FPA28], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA29)) dread(&frp->save_reg[REG_FPA29], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA27)) dread(&frp->save_reg[REG_FPA27], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA26)) dread(&frp->save_reg[REG_FPA26], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA26)) dread(&frp->save_reg[REG_FPA26], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA27)) dread(&frp->save_reg[REG_FPA27], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA25)) dread(&frp->save_reg[REG_FPA25], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA24)) dread(&frp->save_reg[REG_FPA24], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA24)) dread(&frp->save_reg[REG_FPA24], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA25)) dread(&frp->save_reg[REG_FPA25], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA23)) dread(&frp->save_reg[REG_FPA23], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA22)) dread(&frp->save_reg[REG_FPA22], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA22)) dread(&frp->save_reg[REG_FPA22], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA23)) dread(&frp->save_reg[REG_FPA23], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA21)) dread(&frp->save_reg[REG_FPA21], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA20)) dread(&frp->save_reg[REG_FPA20], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA20)) dread(&frp->save_reg[REG_FPA20], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA21)) dread(&frp->save_reg[REG_FPA21], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA19)) dread(&frp->save_reg[REG_FPA19], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA18)) dread(&frp->save_reg[REG_FPA18], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA18)) dread(&frp->save_reg[REG_FPA18], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA19)) dread(&frp->save_reg[REG_FPA19], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA17)) dread(&frp->save_reg[REG_FPA17], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA16)) dread(&frp->save_reg[REG_FPA16], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA16)) dread(&frp->save_reg[REG_FPA16], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA17)) dread(&frp->save_reg[REG_FPA17], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA15)) dread(&frp->save_reg[REG_FPA16], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA14)) dread(&frp->save_reg[REG_FPA14], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA14)) dread(&frp->save_reg[REG_FPA14], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA15)) dread(&frp->save_reg[REG_FPA16], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA13)) dread(&frp->save_reg[REG_FPA13], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA12)) dread(&frp->save_reg[REG_FPA12], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA12)) dread(&frp->save_reg[REG_FPA12], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA13)) dread(&frp->save_reg[REG_FPA13], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA11)) dread(&frp->save_reg[REG_FPA11], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA10)) dread(&frp->save_reg[REG_FPA10], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA10)) dread(&frp->save_reg[REG_FPA10], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA11)) dread(&frp->save_reg[REG_FPA11], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA9)) dread(&frp->save_reg[REG_FPA9], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA8)) dread(&frp->save_reg[REG_FPA8], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA8)) dread(&frp->save_reg[REG_FPA8], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA9)) dread(&frp->save_reg[REG_FPA9], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA7)) dread(&frp->save_reg[REG_FPA7], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA6)) dread(&frp->save_reg[REG_FPA6], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA6)) dread(&frp->save_reg[REG_FPA6], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA7)) dread(&frp->save_reg[REG_FPA7], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA5)) dread(&frp->save_reg[REG_FPA5], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA4)) dread(&frp->save_reg[REG_FPA4], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA4)) dread(&frp->save_reg[REG_FPA4], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA5)) dread(&frp->save_reg[REG_FPA5], regaddr -= 4, 4);
   }

   if (frp->stringsave) {
    if (fbis(frp->w_mask, REG_FPA3)) dread(&frp->save_reg[REG_FPA3], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA2)) dread(&frp->save_reg[REG_FPA2], regaddr -= 4, 4);
   } else {
    if (fbis(frp->w_mask, REG_FPA2)) dread(&frp->save_reg[REG_FPA2], regaddr -= 4, 4);
    if (fbis(frp->w_mask, REG_FPA3)) dread(&frp->save_reg[REG_FPA3], regaddr -= 4, 4);
   }

    if (fbis(frp->w_mask, REG_FPA1)) dread(&frp->save_reg[REG_FPA1], regaddr -= 4, 4);

    if (bis(frp->r_mask, REG_FP7)) dread(&frp->save_freg[REG_FP7], regaddr -= 8, 8);
    if (bis(frp->r_mask, REG_FP6)) dread(&frp->save_freg[REG_FP6], regaddr -= 8, 8);
    if (bis(frp->r_mask, REG_FP5)) dread(&frp->save_freg[REG_FP5], regaddr -= 8, 8);
    if (bis(frp->r_mask, REG_FP4)) dread(&frp->save_freg[REG_FP4], regaddr -= 8, 8);
    if (bis(frp->r_mask, REG_FP3)) dread(&frp->save_freg[REG_FP3], regaddr -= 8, 8);
    if (bis(frp->r_mask, REG_FP2)) dread(&frp->save_freg[REG_FP2], regaddr -= 8, 8);
    if (bis(frp->r_mask, REG_FP1)) dread(&frp->save_freg[REG_FP1], regaddr -= 8, 8);
    if (bis(frp->r_mask, REG_FP0)) dread(&frp->save_freg[REG_FP0], regaddr -= 8, 8);
    if (bis(frp->r_mask, REG_EDI)) dread(&frp->save_reg[REG_EDI], regaddr -= 4, 4);
    if (bis(frp->r_mask, REG_ESI)) dread(&frp->save_reg[REG_ESI], regaddr -= 4, 4);
    if (bis(frp->r_mask, REG_EBX)) dread(&frp->save_reg[REG_EBX], regaddr -= 4, 4);

    /*
     * Go to next frame.  in the normal case, when frp->entered == true,
     * the next frame pointer is pointed to by the current frame pointer.
     * if frp->entered == false, the current procedure hasn't executed
     * its frame linkage code yet, so the frame pointer is still in the
     * register.
     */

    fp = (struct stackframe *)frp->save_fp;
    if (frp->entered) {
	dread(&frp->save_fp, (Address) &fp->return_fp, sizeof(Address));
        if (debug_regsave or debug_frame) {
	    printf("nextframe: frp->entered true, had frame ptr, read new frp->save_fp = 0x%x\n",
	        frp->save_fp);
	    printf("  read from previous stack return_fp of 0x%x\n", &fp->return_fp);
	}
    } else {
	frp->save_fp = reg(REG_EBP);
	frp->entered = true;
	if (debug_regsave or debug_frame) {
	    printf("nextframe: frp->entered was false, frp->save_fp = 0x%x\n",
		frp->save_fp);
	    printf("    read from EBP register\n");
	}
    }

    dread(&frp->save_pc, (Address) &fp->return_pc, sizeof(Address));
    frp->block = func = whatblock(frp->save_pc - 1);

    if (debug_regsave or debug_frame)
	printf("nextframe: read frp->save_pc = 0x%x from frame ptr offset\n",
	    frp->save_pc);

    /*
     * if the pc for the new frame frp->save_pc is in 'sigcode', we've
     * encountered a signal context.  we want to return the frame of
     * the function that got hit by the signal, not the frame of sigcode.
     * since the sigcode trampoline doesn't do ENTER's, the frame ptr as set
     * above is ok.  the problem is to find the pc location.  it is
     * pointed to by the scp pushed on the signal stack.  this is the
     * third argument of the signal handler.  see the picture in
     * src/lib/libc/topaz/sys/sigvec.c.
     */
    if (func == sigcode_sym) {
	dread(&scp, (Address) &fp->arg[2], sizeof(Address));
	dread(&frp->save_pc, (Address) &scp->sc_pc, sizeof(Address));
	frp->block = func = whatblock(frp->save_pc);
	frp->iscaller = false;
        if (debug_regsave or debug_frame)
	    printf("nextframe: func == sigcode_sym\n");
    } else {
	frp->iscaller = true;
    }

    /*
     * find register save mask and locals in new frame.
     */
    if (func->symvalue.funcv.have_rmasks) {
	frp->r_mask = func->symvalue.funcv.r_mask;
	frp->w_mask = func->symvalue.funcv.w_mask;
	frp->nlocals = func->symvalue.funcv.nlocals;
	frp->stringsave = func->symvalue.funcv.stringsave;
	frp->no_frameptr = func->symvalue.funcv.no_frameptr;
	if (debug_regsave) {
	    printf("nextframe: had reg masks, r_mask= 0x%x, w_mask= 0x%x,\n",
	      frp->r_mask, frp->w_mask);
	    printf("nextframe: nlocals= %d, stringsave= %s, no_fptr= %s\n",
	    frp->nlocals, frp->stringsave ? "T" : "F",
	    frp->no_frameptr ? "T" : "F" );
	}
    } else {
	if (debug_regsave)
	    printf("nextframe calling get_enter_args\n");
	get_enter_args(codeloc(func), &frp->r_mask, &frp->w_mask,
	    &frp->nlocals, &string_move, &no_frame_setup);
	func->symvalue.funcv.r_mask = frp->r_mask;
	func->symvalue.funcv.w_mask = frp->w_mask;
	func->symvalue.funcv.nlocals = frp->nlocals;
	func->symvalue.funcv.have_rmasks = true;
	func->symvalue.funcv.stringsave = string_move;
	func->symvalue.funcv.no_frameptr = no_frame_setup;
	frp->stringsave = string_move;
	frp->no_frameptr = no_frame_setup;
    }

    return frp;
}

/*
 * Get the current frame information in the given Frame and store the
 * associated function in the given value-result parameter.
 */

private getcurfunc (frp, fp)
Frame frp;
Symbol *fp;
{
    getcurframe(frp);
    *fp = frp->block;
}

/*
 * Return the frame associated with the next function up the call stack, or
 * nil if there is none.  The function is returned in a value-result parameter.
 * For "inline" functions the statically outer function and same frame
 * are returned.
 */

public Frame nextfunc (frp, fp)
Frame frp;
Symbol *fp;
{
    Symbol t;
    Frame nfrp;

    t = *fp;
    checkref(t);
    if (isinline(t)) {
	t = container(t);
	nfrp = frp;
    } else {
	nfrp = nextframe(frp);
	if (nfrp == nil) {
	    t = nil;
	} else {
	    t = nfrp->block;
	}
    }
    *fp = t;
    return nfrp;
}


/* 
 * Return the frame associated with the given symbol
 */
public Frame symframe(s, frame)
Symbol s;
Frame frame;
{
    register Frame frp;
    register Symbol cur;

    frp = frame;
    if (frp == nil) {
        cur = s->block;
        while (cur != nil and cur->class == MODULE) {
	    cur = cur->block;
        }
        if (cur == nil) {
	    frp = nil;
        } else {
	    frp = findframe(cur);
	    if (frp == nil) {
	        error("[internal error: unexpected nil frame for \"%s\"]",
		    symname(s)
	        );
	    }
        }
    }
    return(frp);
}


/*
 * Return the frame associated with the given function.
 * If the function is nil, return the most recently activated frame.
 *
 * Static allocation for the frame.
 */

public Frame findframe(f)
Symbol f;
{
    Frame frp;
    static struct Frame frame;
    Symbol p;
    Boolean done;

    frp = &frame;
    getcurframe(frp);
    if (f != nil) {
	if (f == curfunc and curframe != nil) {
	    *frp = *curframe;
	} else {
	    done = false;
	    p = frp->block;
	    do {
		if (p == f) {
		    done = true;
		} else if (p == program) {
		    done = true;
		    frp = nil;
		} else {
		    frp = nextfunc(frp, &p);
		    if (frp == nil) {
			done = true;
		    }
		}
	    } while (not done);
	}
    }
    return frp;
}


/*
 * Find the return address of the given function, or if none given,
 * the current procedure/function.
 */

public Address return_addr(s)
    Symbol s;
{
    register Frame frp;
    Address addr;
    struct Frame frame;

    frp = &frame;
    getcurframe(frp);

    do {
	frp = nextframe(frp);
    } while (s and frp and frp->block != s);

    if (frp == nil) {
	addr = 0;
    } else {
	addr = frp->save_pc;
    }
    return addr;
}

/*
 * See if this object (curobj) has the symbol for an 1167 FPA (Weitek).
 */

private boolean has_1167()
{
	Symbol s;

	s = lookup(identname("_1167_flt", true));
	if (s)
		return(true);
	s = lookup(identname("1167_flt", true));
	if (s)
		return(true);
	return(false);
}

/*
 * Push the value associated with function f.  Relies on the compiler
 * to leave the address of returned structure in eax.  If procedure returns
 * a float or double we pick that up from 387 or Weitek (whichever is used).
 * Not that even if something is declared "float" C evaluates expressions as
 * "double" so we pick up a two word floating point.  Because the later
 * routine that wants to pop something off the stack will check size we
 * convert a "float" to a one word value and push on stack.
 */

public pushretval(f)
    Symbol f;
{
    Word eax;
    int		len;		/* length of what f returns */
    Symbol	t;		/* type of what f returns */

    float ff;
    double dd;
    union {
	float u_f;
	Word  u_w;
    } u;

    eax = reg(REG_EAX);
    len = size(f->type);
    if (is_void(f->type))
        return;
    t = rtype(f->type);
    if (t->class == RECORD or t->class == VARNT) {
	rpush((Address) eax, len);
    } else {
	switch (len) {
	    case sizeof(char):
		push(char, eax);
		break;

	    case sizeof(short):
		push(short, eax);
		break;

	    default:
		if (istypename(f->type, "float")) {
		    if (has_1167())
			dd = makedouble(reg(REG_FPA3), reg(REG_FPA2));
		    else
		        dd = freg(REG_FP0);
		    ff = dd;
		    push(float, ff);
		} else if (istypename(f->type, "double")) {
		    if (has_1167())
			dd = makedouble(reg(REG_FPA3), reg(REG_FPA2));
		    else
		        dd = freg(REG_FP0);
		    push(double, dd);
		} else if (len == sizeof(Word)) {
		    push(Word, eax);
		} else {
		    error("[internal error: bad size %d in pushretval]", len);
		}
		break;
	}
    }
}

/*
 * Return the base address for locals in the given frame.
 */

public Address locals_base(frp)
Frame frp;
{
    return (frp == nil) ? reg(REG_EBP) : frp->save_fp;
}

/*
 * Return the base address for arguments in the given frame.
 */

public Address args_base(frp)
Frame frp;
{
    return (frp == nil) ? reg(REG_EBP) : frp->save_fp;
}

/*
 * Return saved register n from the given frame.
 */

public Word savereg(n, frp)
integer n;
Frame frp;
{
    Word w;

    if (frp == nil) {
	w = reg(n);
    } else {
	switch (n) {

	    case REG_EBP:
		w = frp->save_fp;
		break;

	    case REG_ESP:
		w = reg(REG_ESP);
		break;

	    case REG_EIP:
		w = frp->save_pc;
		break;
		
	    case REG_EAX:
	    case REG_EDX:
	    case REG_ECX:
	    case REG_EBX:
	    case REG_ESI:
	    case REG_EDI:
	    case REG_FPA1:  case REG_FPA2:  case REG_FPA3:  case REG_FPA4:
	    case REG_FPA5:  case REG_FPA6:  case REG_FPA7:  case REG_FPA8:
	    case REG_FPA9:  case REG_FPA10: case REG_FPA11: case REG_FPA12:
	    case REG_FPA13: case REG_FPA14: case REG_FPA15: case REG_FPA16:
	    case REG_FPA17: case REG_FPA18: case REG_FPA19: case REG_FPA20:
	    case REG_FPA21: case REG_FPA22: case REG_FPA23: case REG_FPA24:
	    case REG_FPA25: case REG_FPA26: case REG_FPA27: case REG_FPA28:
	    case REG_FPA29: case REG_FPA30: case REG_FPA31:
		w = frp->save_reg[n];
		break;

	    default:
		error("internal error: reg #%d is bad reg", n);
		break;
	}
    }
    return w;
}

/*
 * Return saved double float register n from the given frame.
 */

public double savefreg(n, frp)
integer n;
Frame frp;
{

    if (frp == nil) {
	return(freg(n));
    } else {
	switch (n) {
	    case REG_FP0:
	    case REG_FP1:
	    case REG_FP2:
	    case REG_FP3:
	    case REG_FP4:
	    case REG_FP5:
	    case REG_FP6:
	    case REG_FP7:
		return(frp->save_freg[n]);
		break;
	    case REG_FPA2:
	    case REG_FPA4:
	    case REG_FPA6:
	    case REG_FPA8:
	    case REG_FPA10:
	    case REG_FPA12:
	    case REG_FPA14:
	    case REG_FPA16:
	    case REG_FPA18:
	    case REG_FPA20:
	    case REG_FPA22:
	    case REG_FPA24:
	    case REG_FPA26:
	    case REG_FPA28:
	    case REG_FPA30:
	    	return(makedouble(frp->save_reg[n+1], frp->save_reg[n]));
		break;

	    default:
		error("internal error: reg #%d is bad reg", n);
		break;
	}
    }
}

/*
 * Return the nth argument to the current procedure.
 */

public Word argn(n, frp)
integer n;
Frame frp;
{
    Word w;

    dread(&w, args_base(frp) + ARGOFF + (n * sizeof(Word)), sizeof(w));
    return w;
}

/*
 * return number of arguments passed to given frame.  have to make
 * a local copy of the frame structure pointed to by 'frp' in 'frame'
 * so as not to mess it up with nextframe();
 */

public int nargspassed(frp)
    Frame frp;
{
    struct Frame frame;

    if (frp == nil)
	return( 0 );
    else {
        frame = *frp;
        frp = nextframe(&frame);
	if (frp == nil)
		return(0);
	else
		return( nbytespushed(frp->save_pc) / sizeof(Word) );
    }
}

/*
 * Print a list of currently active blocks starting with most recent.
 */

public wherecmd()
{
    walkstack(false);
}

/*
 * Print the variables in the given frame or the current one if nil.
 */

public dump (func)
Symbol func;
{
    Symbol f;
    Frame frp;

    if (func == nil) {
	f = curfunc;
	if (curframe != nil) {
	    frp = curframe;
	} else {
	    frp = findframe(f);
	}
    } else {
	f = func;
	frp = findframe(f);
    }
    showaggrs = true;
    if (frp != nil)
	    printcallinfo(f, frp);
    dumpvars(f, frp);
}

/*
 * Dump all values.
 */

public dumpall ()
{
    walkstack(true);
}

/*
 * Walk the stack of active procedures printing information
 * about each active procedure.
 */

private walkstack(dumpvariables)
Boolean dumpvariables;
{
    Frame frp;
    boolean save;
    Symbol f;
    struct Frame frame;
    boolean no_frame = false;
    int frame_count = 0;

    if (notstarted() or isfinished()) {
	error("program is not active");
    } else {
	save = walkingstack;
	walkingstack = true;
	showaggrs = dumpvariables;
	frp = &frame;
	getcurfunc(frp, &f);
	for (;;) {
	    if (no_frame and frame_count == 1) {
		printf("unknown calling procedure, frame pointer not saved\n");
		no_frame = false;
	    }
	    printcallinfo(f, frp);
	    if (dumpvariables) {
		dumpvars(f, frp);
		putchar('\n');
	    }
	    if (frp->no_frameptr and not isinline(f))
		no_frame = true;
	    frp = nextfunc(frp, &f);
	    frame_count++;
	    if (frp == nil or f == program or f == dbsubc) {
		break;
	    }
	}
	if (dumpvariables) {
	    printf("in \"%s\":\n", symname(program));
	    dumpvars(program, nil);
	    putchar('\n');
	}
	walkingstack = save;
    }
}

/*
 * return the symbol that has the parameter chain.  for normal procedure
 * symbols, this is just the symbol itself.  if the procedure has multiple
 * entry points (each of which could have different parameter lists), the
 * PROC symbol's chain field points to a cycle of ENTRY symbols.  Each
 * ENTRY symbol's chain field points to the parameter list for the
 * corresponding entry point.  this procedure returns the ENTRY symbol for
 * the entry point active in frame 'f' if PROC symbol 'p' has multiple
 * entry points.
 */
public Symbol checkentry(p, f)
    Symbol p;			/* the class==PROC symbol */
    Frame f;			/* its frame */
{
    register Symbol e;		/* class=ENTRY symbol for entry points */
    short eid;		 	/* entry id actually on stack */
    register Symbol first;	/* marks where we started ENTRY loop */
    Symbol proc;		/* proc corresponding to this entry */
    Name n;

    if ((p->class == PROC || p->class == FUNC) && p->symvalue.funcv.isentry) {

	/*
	 * get the class==ENTRY symbol corresponding to p.
	 */
	e = p->chain;		

	if (f && f->entered) {

	    /*
	     * decide where the entry id is located on the stack if we
	     * don't know already.
	     */

	    if (e->symvalue.entryv.eid_offset == 0) {

		/*
		 * since none of the entry points are in the function 
		 * table, 'p' is the PROC symbol for the subroutine itself.
		 * the prologue for the subroutine is different than the
		 * prologues for the entry points, and entryid_offset()
		 * won't work for it.  the entry id offset has to be the
		 * same for all entrypoints, so find it out from the
		 * first entry.
		 */

		e = e->symvalue.entryv.nextentry;	/* go to first entry */

		/*
		 * find the PROC symbol that corresponds to this entry
		 * so we can find its address.
		 */

		n = e->name;
		find(proc, n) where
			proc->class == PROC or proc->class == FUNC
		endfind(proc);

		/*
		 * find out the entry id location.
		 */

	    	e->symvalue.entryv.eid_offset = 
			entryid_offset(e->symvalue.entryv.entryid, 
				entryloc(proc), codeloc(proc));

		/*
		 * save it for next time.
		 */

		 p->chain->symvalue.entryv.eid_offset =
			e->symvalue.entryv.eid_offset;

	    }

	    /*
	     * grab the entry id off the stack.
	     */

	    dread(&eid, f->save_fp + e->symvalue.entryv.eid_offset, 2);

	    /*
	     * loop through the cycle of entry points looking for the
	     * one who's id matches the one found on the stack.
	     */

	    first = nil;
	    while(e->symvalue.entryv.entryid != eid && e != first) {
		if (first == nil)
		    first = e;
		e = e->symvalue.entryv.nextentry;
	    }

	}

	/*
	 * return the ENTRY symbol that has been found.
	 */

	p = e;

    }
    return(p);
}

/*
 * Print out the information about a call, i.e.,
 * routine name, parameter values, and source location.
 */

private printcallinfo (f, frp)
Symbol f;
Frame frp;
{
    Lineno line;
    Address srclineaddr;

    /*
     * if this frame is doing a subroutine call with no arguments,
     * the call instruction might compile to a single instruction.
     * in this case save_pc would point to the source line *after*
     * the call (since it is just the return pc off the stack).
     * in order to get tracebacks to show the source line
     * of the call statement, decrement the pc.  it's safe to decrement
     * the pc for any call, since there are no 1-byte procedure call
     * instructions.
     */
    if (frp->iscaller)
	srclineaddr = frp->save_pc - 1;
    else
	srclineaddr = frp->save_pc;

    f = checkentry(f, frp);
    printname(stdout, f);
    if (not isinline(f)) {
	printparams(f, frp);
    }
    line = srcline(srclineaddr);
    if (line != 0) {
	printf(", line %d", line);
	printf(" in \"%s\"\n", srcfilename(srclineaddr));
    } else {
	printf(" at 0x%x\n", frp->save_pc);
    }
}


/*
 * Set the current function to the given symbol.
 * We must adjust "curframe" so that subsequent operations are
 * not confused; for simplicity we simply clear it.
 */

public setcurfunc (f)
Symbol f;
{
    curfunc = f;
    curframe = nil;
}

/*
 * Return the frame for the current function.
 * The space for the frame is allocated statically.
 */

public Frame curfuncframe ()
{
    static struct Frame frame;
    Frame frp;

    if (curframe == nil) {
	frp = findframe(curfunc);
	if (frp == nil)
	    return(nil);
	curframe = &curframerec;
	*curframe = *frp;
    } else {
	frp = &frame;
	*frp = *curframe;
    }
    return frp;
}

/*
 * Set curfunc to be N up/down the stack from its current value.
 */

public up (n)
integer n;
{
    integer i;
    Symbol f;
    Frame frp;
    boolean done;

    if (not isactive(program)) {
	error("program is not active");
    } else if (curfunc == nil) {
	error("no current function");
    } else {
	i = 0;
	f = curfunc;
	frp = curfuncframe();
	done = false;
	do {
	    if (frp == nil) {
		done = true;
		error("not that many levels");
	    } else if (i >= n) {
		done = true;
		curfunc = f;
		curframe = &curframerec;
		*curframe = *frp;
		showaggrs = false;
		printcallinfo(curfunc, curframe);
	    } else if (f == program) {
		done = true;
		error("not that many levels");
	    } else {
		frp = nextfunc(frp, &f);
	    }
	    ++i;
	} while (not done);
    }
}

public down (n)
integer n;
{
    integer i, depth;
    Frame frp, curfrp;
    Symbol f;
    struct Frame frame;

    if (not isactive(program)) {
	error("program is not active");
    } else if (curfunc == nil) {
	error("no current function");
    } else {
	depth = 0;
	frp = &frame;
	getcurfunc(frp, &f);
	if (curframe == nil) {
	    curfrp = findframe(curfunc);
	    curframe = &curframerec;
	    *curframe = *curfrp;
	}
	while ((f != curfunc or !frameeq(frp, curframe)) and f != nil) {
	    frp = nextfunc(frp, &f);
	    ++depth;
	}
	if (f == nil or n > depth) {
	    error("not that many levels");
	} else {
	    depth -= n;
	    frp = &frame;
	    getcurfunc(frp, &f);
	    for (i = 0; i < depth; i++) {
		frp = nextfunc(frp, &f);
		assert(frp != nil);
	    }
	    curfunc = f;
	    *curframe = *frp;
	    showaggrs = false;
	    printcallinfo(curfunc, curframe);
	}
    }
}

/*
 * the 'return' command.
 */
public rtnfunc(f)
	Symbol f;
{
    Address addr;
    Node cond;

    addr = return_addr(f);
    if (addr == 0) {
	if (f == 0)
	    error("nowhere to return to");
	else
	    error("\"%s\" is not active", symname(f));
    } else {
	cond = build(O_EQ, build(O_SYM, pcsym), build(O_LCON, addr));
	inherit(event_temp(THISPROC, cond, buildcmdlist(build(O_DBXSTOP, BY_RETURN))));
	cont(0, true);
    }
}

/*
 * Find the entry point of a procedure or function.
 */

/*ARGSUSED*/
public findbeginning (f)
Symbol f;
{
}

/*
 * Return the address corresponding to the first line in a function.
 */

public Address firstline(f)
Symbol f;
{
    Address addr;

    addr = codeloc(f);
    while (linelookup(addr) == 0 and addr < objsize) {
	++addr;
    }
    if (addr == objsize) {
	addr = -1;
    }
    return addr;
}

/*
 * Return the address corresponding to the end of the program.
 *
 * We look for the entry to "exit".
 */

public Address lastaddr()
{
    Symbol s;

    s = lookup(identname("exit", true));
    if (s == nil) {
	panic("can't find exit");
    }
    return codeloc(s);
}

/*
 * Decide if the given function is currently active.
 *
 * We avoid calls to "findframe" during a stack trace for efficiency.
 * Presumably information evaluated while walking the stack is active.
 */

public Boolean isactive(f)
Symbol f;
{
    Boolean b;

    if (walkingstack)
	b = true;
    else if (isfinished())
	b = false;
    else if (f == program)
	b = true;
    else if (f and (f->class == COMMON or (ismodule(f) and isactive(container(f)))))
	b = true;
    else
	b = (Boolean) (findframe(f) != nil);
    return b;
}

/*
 * Evaluate a call to a procedure.
 */

#define DBARGS_SIZE	512	/* how much room in _dbargs for return struct*/

public callproc(exprnode, isfunc)
Node exprnode;
boolean isfunc;
{
    Node procnode, arglist, idnode;
    Symbol proc;
    Symbol procty;		/* type of what 'proc' returns */
    Event e = 0;
    Address retsp;		/* sp just before call */

    procnode = exprnode->value.arg[0];
    arglist = exprnode->value.arg[1];
    if (procnode->op != O_SYM) {
	beginerrmsg();
	fprintf(stderr, "can't call \"");
	prtree(stderr, procnode);
	fprintf(stderr, "\"");
	enderrmsg();
    }
    assert(procnode->op == O_SYM);
    proc = procnode->value.sym;
    if (not isblock(proc)) {
	error("\"%s\" is not a procedure or function", symname(proc));
    }
    if (proc->language == lang_pascal or proc->language == lang_fortran) {
	error("'call' command not implemented for %s",
		language_name(proc->language));
    }
    if (wp->p_state != STOPPED and wp->p_state != IMAGE) 
	error("Process is not stopped");
    callisfunc = isfunc;
    pushenv();
    wp->p_state = STOPPED;
    if (seterror()) {
	popenv();
	if (e)
		delevent(event_id(e), true);
	erecover();
    }
    (void) pushargs(proc, arglist);
    if (not is_void(proc->type))
        procty = rtype(proc->type);
    else
	procty = nil;
    if (not procty == nil and (procty->class == RECORD or procty->class == VARNT)) {
	/*
	 * deal with functions that return structures.  Move the address
	 * of the return structure into the EAX register.
	 * HIGHLY COMPILER DEPENDENT!
	 *
	 * NOTE: Two lines commented out push return address on stack.
	 *       This was formerly done, (as well as having in EAX).
	 *       Removed by compiler change.
	 */
	Symbol dbargs;
	Address saddr;		/* place to store returned structure */

	if (size(procty) > DBARGS_SIZE)
	    error("return structure too large");
	dbargs = lookup(identname("_dbargs", true));
	if (dbargs == nil)
	    error("program not linked with `-g'");
/*	setreg(STKP, reg(STKP) - sizeof(Address));  XXX */
	saddr = address(dbargs, nil); 
	setreg(REG_EAX, saddr); 
/*	dwrite(&saddr, reg(STKP), sizeof(Address));  XXX */
        retsp = reg(REG_ESP) + 0; /* REG_ESP == retsp in procreturn */
    } else
        retsp = reg(REG_ESP);     /* REG_ESP == retsp in procreturn  */
    /*
     * Note:  Comment below is previous to compiler change.  The " + 0; "
     * above was previously " + 4; ".  Stack is now the same for calls
     * whether or not they return structures.
     *
     * Note the different return stack depending on wheather or not this
     * procedure returns a structure.  This is because when a struct is going
     * to be returned the callee pops 4 bytes off and the caller will pop
     * the rest.  This means that after call the stack will be different
     * depending on wheather or not this returns a struct.
     * Code before standard prologue of procedure will look something like:
     *      popl       %eax
     *      xchgl      (%esp),%eax
     * This leaves stack 4 bytes different (higher address) than would be
     * expected for typical procedure.
     */

    beginproc(proc);

    /*
     * create an event for when the procedure finishes.  have to 
     * deal with multiple "call" commands to the same procedure,
     * so we use the stack pointer to disambiguate.  also as a
     * result have to make the event delete itself, so pass eventid.
     */
    idnode = build(O_LCON, 0);
    e = dbxevent(THISPROC,
	build(O_EQ, build(O_SYM, pcsym), build(O_SYM, retaddrsym)),
	buildcmdlist(build(O_PROCRTN, build(O_SYM, proc),
		build(O_LCON, retsp), idnode))
    );
    idnode->value.lcon = event_id(e);

    /*
     * get the procedure going.  the bpact is in case there's a breakpoint
     * at the first instruction. 
     */
    unseterror();
    wp->p_newstate = RUNNING;
    if (bpact())
	erecover();

    /*
     * if the procedure is being called from inside an expression (e.g. by
     * O_CALL instead of O_CALLPROC), we don't give it the tty, and we
     * wait for it explicitly.  if the procedure stops for some other
     * reason than returning (like a breakpoint), longjmp out.
     */
    cont(0, not isfunc);
    if (isfunc) {
	waitfor(wp);
        if (wp->p_stopinfo != BY_CALLRTN)
	    erecover();
    }
}

/*
 * Push the arguments on the process' stack.  We do this by first
 * evaluating them on the "eval" stack, then copying into the process'
 * space.
 */

private integer pushargs(proc, arglist)
Symbol proc;
Node arglist;
{
    Stack *savesp;
    int argc, args_size;

    savesp = sp;
    argc = evalargs(proc, arglist);
    args_size = sp - savesp;
    setreg(REG_ESP, reg(REG_ESP) - args_size);
    dwrite(savesp, reg(REG_ESP), args_size);
    sp = savesp;
    return argc;
}

/*
 * Check to see if an expression is correct for a given parameter.
 * If the given parameter is false, don't worry about type inconsistencies.
 *
 * Return whether or not it is ok.
 */

private boolean chkparam (actual, formal, chk)
Node actual;
Symbol formal;
boolean chk;
{
    boolean b;

    b = true;
    if (chk) {
	if (formal == nil) {
	    beginerrmsg();
	    fprintf(stderr, "too many parameters");
	    b = false;
	} else if (not compatible(formal->type, actual->nodetype)) {
	    beginerrmsg();
	    fprintf(stderr, "type mismatch for %s", symname(formal));
	    b = false;
	}
    }
    if (b and formal != nil and
	isvarparam(formal) and not isopenarray(formal->type) and
	not (
	    actual->op == O_RVAL or actual->nodetype == t_addr or
	    (
		actual->op == O_TYPERENAME and
		(
		    actual->value.arg[0]->op == O_RVAL or
		    actual->value.arg[0]->nodetype == t_addr
		)
	    )
	)
    ) {
	beginerrmsg();
	fprintf(stderr, "expected variable, found \"");
	prtree(stderr, actual);
	fprintf(stderr, "\"");
	b = false;
    }
    return b;
}

/*
 * Pass an expression to a particular parameter.
 *
 * Normally we pass either the address or value, but in some cases
 * (such as C strings) we want to copy the value onto the stack and
 * pass its address.
 *
 * Another special case raised by strings is the possibility that
 * the actual parameter will be larger than the formal, even with
 * appropriate type-checking.  This occurs because we assume during
 * evaluation that strings are null-terminated, whereas some languages,
 * notably Pascal, do not work under that assumption.
 */

private passparam (actual, formal)
Node actual;
Symbol formal;
{
    Address addr;
    Stack *savesp;
    integer actsize;

    if (formal != nil and isvarparam(formal) and
	(not isopenarray(formal->type))
    ) {
	addr = lval(actual->value.arg[0]);
	push(Address, addr);
    } else if (passaddr(formal, actual->nodetype)) {
	savesp = sp;
	eval(actual);
	actsize = sp - savesp;
	setreg(REG_ESP,
	    reg(REG_ESP) - ((actsize + sizeof(Word) - 1) & ~(sizeof(Word) - 1))
	);
	dwrite(savesp, reg(REG_ESP), actsize);
	sp = savesp;
	push(Address, reg(REG_ESP));
	if (formal != nil and isopenarray(formal->type)) {
	    push(integer, actsize div size(formal->type->type));
	}
    } else if (isfloat(rtype(actual->nodetype))) {	/* XXX 3 calls! */
	/*
	 * deal with passing floats correctly -- should be passed as double.
	 */
	float f;
	double d;

	eval(actual);
	f = pop(float);
	d = f;
	push(double, d); 
    } else {
	/*
	 * evaluate argument, push on stack, and word align the stack.
	 * not quite right for things that ought to be sign extended.
	 */
	savesp = sp;
	eval(actual);
	actsize = sp - savesp;
	if ((actsize % sizeof(Word)) != 0) {
	    *((int *) sp) = 0;
	    sp += sizeof(Word) - (actsize % sizeof(Word));
	}
    }
}

/*
 * Evaluate an argument list left-to-right.
 */

private integer evalargs(proc, arglist)
Symbol proc;
Node arglist;
{
    Node p, actual;
    Symbol formal;
    Stack *savesp;
    integer count;
    boolean chk;

    savesp = sp;
    count = 0;
    formal = proc->chain;
    chk = (boolean) (not nosource(proc));
    for (p = arglist; p != nil; p = p->value.arg[1]) {
	assert(p->op == O_COMMA);
	actual = p->value.arg[0];
	if (not chkparam(actual, formal, chk)) {
	    fprintf(stderr, " in call to %s", symname(proc));
	    sp = savesp;
	    enderrmsg();
	}
	passparam(actual, formal);
	if (formal != nil) {
	    formal = formal->chain;
	}
	++count;
    }
    if (chk) {
	if (formal != nil) {
	    sp = savesp;
	    error("not enough parameters to %s", symname(proc));
	}
    }
    return count;
}

public procreturn(f, stkp, eventid)
Symbol f;
Address stkp;
integer eventid;
{
    integer retvalsize;
    Node tmp;
    char *copy;
    Symbol p, iob;
    Stack *savesp;
    int iobsize;		/* TEMPORARY XXX */

    /*
     * make sure that we're returning from the correct subroutine.
     * if we're in the right one, delete the event.
     */
    if (reg(REG_ESP) != stkp)
	return;
    else
	if (delevent(eventid, false) != true)
		fatal("can't delete event %d in procreturn!", eventid);

    /*
     * get the return value off of the stack, if possible.
     */

   /*
    * The 32000 version places a popenv()/pushenv() around the block of
    * code which may do the pushretval().  Unclear why, need to see what
    * happens on 32K version when this is removed.  This causes problems
    * on the 386 version because register contents may be needed in
    * pushretval() and the pop will loose them.
    */

/*  popenv(); */
    if (callisfunc) {
	retvalsize = size(f->type);
	pushretval(f);
    }

    /*
     * Arrange for output to be flushed, if possible.
     * This is VERY dependent on the use of stdio.
     *
     * A problem surfaced when debugging a program compiled in the att
     * universe.  We tried to call fflush passing the address of stdout.
     * This was calculated by finding the base address of "_iob" and adding
     * the size of an "_iobuf" stuct.  This is 20 for ucb, but only 16
     * for att.  This clobbered the IO structures in att and caused a
     * segmentation fault the next time printf was used.
     *
     * Fix is to include a library routine "_fflushstdout()" which will
     * take no arguments and simply flush stdout.
     */

    /* pushenv(); */
    p = lookup(identname("_fflushstdout", true));
    while (p != nil and not isblock(p)) {
	p = p->next_sym;
    }
    if (p != nil) {
	savesp = sp;
	dwrite(savesp, reg(REG_ESP), sizeof(long));
	sp = savesp;
	beginproc(p);
	event_once(THISPROC,
	    build(O_EQ, build(O_SYM, pcsym), build(O_SYM, retaddrsym)),
	    buildcmdlist(build(O_PROCFLUSHED, f))
	);
	return;
    }

    /*
     * If that didn't work, we will temporarily try and do the old
     * flush technique.  When we remove remainder of code enable the
     * call to procdone here.
     */

    /* procdone(); */
    /* end of routine */

    iob = lookup(identname("_bufsync", true));
    if (iob == nil)
	iobsize = 20;
    else
	iobsize = 16;

/*  pushenv(); */
    p = lookup(identname("fflush", true));
    while (p != nil and not isblock(p)) {
	p = p->next_sym;
    }

    if (p != nil) {
	iob = lookup(identname("_iob", true));
	if (iob != nil) {
	    savesp = sp;
	    push(long, address(iob, nil) + iobsize);
	    setreg(REG_ESP, reg(REG_ESP) - sizeof(long));
	    dwrite(savesp, reg(REG_ESP), sizeof(long));
	    sp = savesp;
	    beginproc(p);
	    event_once(THISPROC,
		build(O_EQ, build(O_SYM, pcsym), build(O_SYM, retaddrsym)),
		buildcmdlist(build(O_PROCFLUSHED, f))
	    );
	    return;
	}
    }


    /*
     * if output can't be flushed, finish now.
     */

    procdone(f);
}

public procdone(f)
Symbol f;
{
    popenv();
    setstopped(BY_CALLRTN);
    if (not callisfunc) {
	putchar('\n');
	printname(stdout, f);
	printf(" returns successfully\n", symname(f));
    }
}

/*
 * Push the current environment.
 */

private pushenv()
{
    Word r;

    pcheck(sizeof(Lineno) + sizeof(String) + sizeof(Symbol) + sizeof(Frame) + 
        sizeof(struct Frame) + sizeof(boolean) + sizeof(Word) + sizeof(Word) +
	sizeof(Word) + sizeof(Pstate) + sizeof(Word) +
	sizeof(Word) + sizeof(Word));
    ppush(sizeof(Lineno), (pointer)&curline);
    ppush(sizeof(String), (pointer)&cursource);
    ppush(sizeof(Symbol), (pointer)&curfunc);
    ppush(sizeof(Frame), (pointer)&curframe);
    ppush(sizeof(struct Frame), (pointer)&curframerec);
    ppush(sizeof(boolean), (boolean)&callisfunc);
    r = reg(REG_EIP);
    ppush(sizeof(Word), (pointer)&r);
    r = reg(REG_ESP);
    ppush(sizeof(Word), (pointer)&r);
    r = reg(REG_EAX);
    ppush(sizeof(Word), (pointer)&r);
    ppush(sizeof(Pstate), (pointer)&wp->p_state);
    r = (Word) wp->p_regs.pr_fpu.fpu_tag;
    ppush(sizeof(Word), (pointer)&r);
    r = (Word) wp->p_regs.pr_fpu.fpu_control;
    ppush(sizeof(Word), (pointer)&r);
    r = (Word) wp->p_regs.pr_fpu.fpu_status;
    ppush(sizeof(Word), (pointer)&r);
}

/*
 * Pop back to the real world.
 */

public popenv()
{
    String filename;
    Word r;

    r = *(Word*)ppop(sizeof(Word));
    wp->p_regs.pr_fpu.fpu_status = r;
    r = *(Word*)ppop(sizeof(Word));
    wp->p_regs.pr_fpu.fpu_control = r;
    r = *(Word*)ppop(sizeof(Word));
    wp->p_regs.pr_fpu.fpu_tag = r;
    wp->p_state = *(Pstate*)ppop(sizeof(Pstate));
    setreg(REG_EAX, *(Word*)ppop(sizeof(Word)));
    setreg(REG_ESP, *(Word*)ppop(sizeof(Word)));
    setreg(REG_EIP, *(Word*)ppop(sizeof(Word)));
    callisfunc = *(boolean*)ppop(sizeof(boolean));
    curframerec = *(struct Frame*)ppop(sizeof(struct Frame));
    curframe = *(Frame*)ppop(sizeof(Frame));
    curfunc = *(Symbol*)ppop(sizeof(Symbol));
    filename = *(String*)ppop(sizeof(String));
    curline = *(Lineno*)ppop(sizeof(Lineno));
    setsource(filename);
}

/*
 * Deal with Pascal run time range error.  Pascal handles them by
 * calling %_RGERR or %_SRERR.  Breakpoints are set in these routines
 * by catch_range_errors().  Hitting the breakpoints causes range_err to
 * be called. What we do is "return" from %_RGERR without executing all of it,
 * and print a message.  Execution can continue.
 */

private char *range_errmsg[] = {
    "Range error ",
    "String range error "
};

private char *handler_name[] = {
    "%_RGERR",
    "%_SRERR"
};

public range_err(which)
	int which;
{
    Address retaddr;

    /*
     * pop stack into pc.  THIS IS MACHINE AND RUNTIME DEPENDENT!!!
     */
    dread(&retaddr, reg(REG_ESP), sizeof(Address));
    setreg(REG_EIP, retaddr);
    setreg(REG_ESP, reg(REG_ESP) - sizeof(Address));

    /*
     * print informative message and cause process to stop.
     */
    printprocid();
    printf("%s\n", range_errmsg[which]);
    setstopped(BY_RTERR);
}

public catch_range_error(which)
    int which;
{
    Symbol s;
    Node cond;

    s = lookup(identname(handler_name[which], true)); 
    if (s != nil) {
        cond = build(O_EQ, build(O_SYM, pcsym), build(O_LCON, codeloc(s)));
        dbxevent(ALLPROCS, cond, buildcmdlist(build(which ? O_SRERR : O_RGERR)));
    }
}
