
#include "crash.h"
#include <sdb.h>
#include <setjmp.h>

#define error(s) longjmp(jmp, s)
static jmp_buf jmp;
char *err_atoi;

static char *str;
static int nbkt, nbrc;
static struct val term(), lval(), qqqq(), factor(), _expr(), getnum(), getsym();
struct val dot;

#define	ID(c)	(isalpha(c) || (c) == '_' || isdigit(c))

expr()
{
	register char *s;
	int n;

	s = token();
	if (s == NULL)
		s = ".";
	n = atoi(s);
	if ( err_atoi )
		printf("'%s', %s?\n", s, err_atoi);
	else {
		printf("= %#x\n", n);
		if (dot.v_sdb.sdb_name)
			printf("(sdb info) name '%s', type '%#x', desc '%#x', offset '%#x', size '%#x'\n", 
			  dot.v_sdb.sdb_name, dot.v_sdb.sdb_type, dot.v_sdb.sdb_desc, dot.v_sdb.sdb_soff, dot.v_sdb.sdb_size);
	}
}

set()
{
	register char *s, *p;
	register n;
	register struct set_val *v;

	s = token();
	if (s == NULL) {
		for (v = &set_val; v->sv_name; v = v->sv_next)
			printf("%s\tvalue '%#x', type '%#x', desc '%#x', offset '%#x', size '%#x'\n",
				v->sv_name, v->sv_val.v_value, 
				v->sv_val.v_sdb.sdb_type, v->sv_val.v_sdb.sdb_desc,
				v->sv_val.v_sdb.sdb_soff, v->sv_val.v_sdb.sdb_size);
		return;
	}
	if ((p=index(s, '=')) == NULL) {
		printf("'%s', missing '='\n", s);
		return;
	}
	*p = 0;
	if (isalpha(*s) == 0 && *s != '_') {
		printf("'%s', illegal lhs for a token\n", s);
		return;
	}
	if (searchbynam(s) || searchsdb(s)) {
		printf("'%s' already exists\n", s);
		return;
	}
	for (p=s+1; *p; p++) {
		if ( ID(*p) == 0 ) {
			printf("'%s', illegal lhs for token\n", s);
			return;
		}
	}
	(void) atoi(p+1);	/* void cause value stored in dot */
	if ( err_atoi ) {
		printf("'%s', %s?\n", p+1, err_atoi);
		return;
	}
	add_value(s);
}

struct set_val set_val;

add_value(s)
	char *s;
{
	register struct set_val *v;

	v = searchsetval(s);
	if (v) {  /* already exists */
		v->sv_val = dot;
		return;
	}
	/*
	 * Name not in list, so add it
	 */
	for (v = &set_val; v->sv_name; v = v->sv_next)
		/* void */;
	v->sv_next = (struct set_val *) malloc(sizeof (struct set_val));
	bzero(v->sv_next, sizeof (struct set_val));
	v->sv_name = malloc( strlen(s) + 1 );
	strcpy(v->sv_name, s);
	v->sv_val = dot;
	return;
}

struct set_val *
searchsetval(s)
	register char *s;
{
	register struct set_val *v;

	for (v = &set_val; v->sv_name; v = v->sv_next)
		if (strcmp(s, v->sv_name) == 0)
			return v;
	return NULL;
}

atoi(s)
char *s;
{
	char *ret;
	struct val v;

	str = s;
	nbkt = nbrc = 0;
	err_atoi = 0;
	if((ret = (char *)setjmp(jmp)) == 0) {
		v = _expr();
		if( nbkt == 0 && nbrc == 0) {
			dot = v;
			return v.v_value;	/* normal return */
		}
		ret = "unbalanced parens";
		/* and fall into ... */
	}
	err_atoi = ret;
	return 0;	/* error */
}

static struct val _expr()
{ 
	struct val lhs, v;

	lhs = factor();
	for(;;) switch (*str++) {
		case ' ':
		case '\t':
			continue;
		case '+':
			v = factor(); 
			lhs.v_value += v.v_value;
			continue;
		case '-':
			v = factor(); 
			lhs.v_value -= v.v_value;
			continue;
		case '&':
			v = factor(); 
			lhs.v_value &= v.v_value;
			continue;
		case '|':
			v = factor();
			lhs.v_value |= v.v_value;
			continue;
		default:
			str--;
			return lhs;
	}
}

static struct val factor()
{
	struct val lhs, v;

	lhs = term();
	for(;;) switch( *str++ ) {
		case ' ':
		case '\t':
			continue;
		case '*':
			v = term();
			lhs.v_value *= v.v_value;
			continue;
		case '/':
			v = term();
			lhs.v_value /= v.v_value;
			continue;
		case '%':
			v = term();
			lhs.v_value %= v.v_value;
			continue;
		default:
			str--;
			return lhs;
	}
}

static struct val term()
{
	struct val v;

	for(;;) switch ( *str++ ) {
		case ' ':
		case '\t':
			continue;
		case '.':
			return dot;
		case '*':
			v = term();
			readv(v.v_value, &v.v_value, sizeof (int *));
			return v;
		case '-':
			v = term();
			v.v_value = -v.v_value;
			return v;
		case '~':
			v = term();
			v.v_value = ~v.v_value;
			return v;
		case '(':
			nbkt++;
			v = _expr();
			if (*str!=')')
				error("unbalanced parens");
			else { 
				str++;
				nbkt--;
				return v;
			}
		default:
			str--;
			if (0) ;	/* assembler bug workaround! */
			return qqqq();
	}
}

static struct val qqqq()
{
	struct val lhs, v;

	lhs = lval();
	for (;;) switch( *str++ ) {
		 case ' ':
		 case '\t':
			continue;
		case '[':
			nbrc++;
			v = _expr();
			if (*str!=']')
				error("unbalanced brackets");
			else {
				str++;
				nbrc--;
				lhs.v_value += lhs.v_sdb.sdb_size * v.v_value;
				continue;
			}
		 case '.':
			v = lval();
			v.v_value = lhs.v_value + v.v_sdb.sdb_soff;
			lhs = v;
			continue;
		case '-':
			if( *str == '>' ) {
				str++;
				v = lval();
				v.v_value = lhs.v_value + v.v_sdb.sdb_soff;
				lhs = v;
				continue;
			}
		default:
			str--;
			return lhs;
	}
}

static struct val lval()
{
	if( isdigit(*str) )
		return getnum();
	if( isalpha(*str) || *str == '_')
		return getsym();
	error("badly formed expression");
}


static struct val getsym()
{
	char symbol[ MAXSTRLEN ], *p = symbol;
	struct val v;

	while (ID(*str) && p < &symbol[ MAXSTRLEN -1])
		*p++ = *str++;
	*p = '\0';
	v.v_value = search(symbol);
	if ( err_search )
		error(symbol);
	v.v_sdb = sdbinfo;
	if (debug)  printf("\n..getsym('%s'): value '%#x', size '%#x', type '%#x', desc '%#x', soff '%#x'..\n",
		    symbol, v.v_value, v.v_sdb.sdb_size, v.v_sdb.sdb_type, v.v_sdb.sdb_desc, v.v_sdb.sdb_soff);
	/* 
	 * if basic type is a pointer, must de-reference it now
	 */
	if (desc(&v.v_sdb.sdb_desc) == PCC_Q_POINTER && searchsetval(symbol) == NULL) {
		readv(v.v_value, &v.v_value, sizeof (int *));
		if (debug) printf("de-reference pointer to '%#x'\n", v.v_value);
	}
	return v;
}

static struct val getnum()
{
	register int c;
	register unsigned int n = 0;
	struct val v;

	c = *str++;
	if( (c=='0') && ( (*str=='x') || (*str=='X') ) ) {
		str++;
		while( isxdigit( c = *str++ ) ) {
			if( isdigit(c) )
				c = c - '0';
			else if( isupper(c) )
				c = c - 'A' + 10;
			else
				c = c - 'a' + 10;
			n = n * 16 + c;
		}
	} else for (n =  c - '0'; isdigit(c = *str++); ) {
		n = n * 10 + c - '0';
	}
	str--;
	v.v_value = n;
	bzero(&v.v_sdb, sizeof (struct sdb));
	v.v_sdb.sdb_size = sizeof (int);
	return v;
}
