#ifndef lint
static char rcsid[] = "$Header: main.c 2.1 86/07/23 $";
#endif

#include "crash.h"
#include "define.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <signal.h>
#include <setjmp.h>
#include <errno.h>

char	*dynix = NULL;
int	dynixfd;
char	*vmcore = NULL;
int	vmcorefd;
char	*swap = NULL;
int	swapfd;
int	live = 0;
int	end_locore;
int	start_locore;

static	jmp_buf jmp;
struct pte *Sysmap;
FILE *filter = NULL;
FILE *input = NULL;
FILE *popen();
char *index();

int tok = 0;		/* no token queue'd in token() */
int source = FALSE;

main(argc, argv)
int	argc;
char	**argv;
{
	int	sigint(), exit(), ouch(), onpipe(), Disread();
	register  struct  tsw	*tp;
	register  char *c;
	struct initfuns *ip;
	struct stat sbuf;

	signal(SIGINT, exit);

	if (strcmp(argv[1], "-d") == 0)
		debug = 10, ++argv, --argc;

	switch(argc) {
	default:
		fatal("usage: crash [ dynix vmcore ]");
	case 3:
		if (a_out(argv[2]))
			dynix = argv[2];
		else
			vmcore = argv[2];
	case 2:
		if (a_out(argv[1])) {
			if (dynix) 
				fatal("dynix specified twice!");
			dynix = argv[1];
		} else {
			if (vmcore) 
				fatal("vmcore specified twice!");
			vmcore = argv[1];
		}
	case 1:
		if (dynix == NULL)
			dynix = "/dynix";
		if (vmcore == NULL)
			vmcore = "/dev/mem";
		if (a_out(vmcore))
			fatal("vmcore is an a.out?");
		if (!a_out(dynix))
			fatal("dynix is not an a.out?");
	}

	printf("Version %s of %s\n", Version, Date); 
	fflush(stdout);

	if ((dynixfd = open(dynix, 0)) < 0)
		fatal("cannot open dynix file");
	if ((vmcorefd = open(vmcore, 0)) < 0)
		fatal("cannot open vmcore file");

	/* Is it live or Memorex? */
	if (strcmp(vmcore, "/dev/mem") == 0 ||
	    fstat(vmcorefd, &sbuf) == 0 && (sbuf.st_mode & S_IFMT) == S_IFCHR) {
		printf("Warning: running on a LIVE system.\n");
		if (swap == NULL)
			swap = "/dev/drum";
		if ((swapfd = open(swap, 0)) < 0)
			fatal("cannot open swap file");
		live = 1;
	} else {
	}

	if (Etext == 0 || (unsigned)Etext > 600*1024)
		printf("Warning: etext == %#x, must be HUGE.\n", Etext);

	for(ip = initfuns; ip->i_fun ; ip++) (*ip->i_fun)();
	input = stdin;

	signal(SIGINT, sigint);
	signal(SIGSEGV, ouch);
	signal(SIGPIPE, onpipe);

#ifdef	ns32000
	dis32000_init(0, Disread, addr_str, 1);
#endif
#ifdef	i386
	dis386_init(0, Disread, addr_str, 1);
#endif

	if (!live) {
		time_t t;
		readv(search("time"), &t, sizeof (time_t));
		printf("System crashed at: %.24s\n", ctime(&t));
	}

	(void) setjmp(jmp);
	for(;;) {
		c = token();
		if (c == NULL)
			continue;
		for (tp = t; tp->t_snm; tp++)
			if (strncmp(tp->t_snm, c, strlen(tp->t_snm)) == 0)
				break;
		if ( tp->t_snm && tp->t_fun )
			(*tp->t_fun)();
		else
			printf("Huh?\n");
		tok = 0;
	}
}

sigint()
{
	tok = 0;
	printf("\nInterrupt!\n");
	fflush(stdout);
	if (source == TRUE) {
		fclose(input);
		input = stdin;
		source = FALSE;
	}
	longjmp(jmp, 1);
}

ouch()
{
	tok = 0;
	printf("Ouch!, take off hoser.\n");
	fflush(stdout);
	longjmp(jmp, 1);
}

onpipe()
{
	tok = 0;
	longjmp(jmp, 1);
}

char *
token()
{
	char  *cp;
	static char line[512];
	static char pline[512];
	static char *p;
	char *pp;

	while( tok == 0 ) {
		if (filter != NULL) {
			(void) pclose(filter);
			filter = NULL;
		}
		if (source == FALSE) {
			printf("> "); 
			fflush(stdout);
		}
		if(fgets(p = line, sizeof line, input) == NULL) {
			if (source == TRUE) {
				fclose(input);
				input = stdin;
				source = FALSE;
				continue;
			}
			return("quit");
		}
		if (debug && source == TRUE)
			fprintf(stderr, "+ %s", p);
		if (strncmp(p, "!!", 2) == 0) {
			if (strlen(p)+strlen(pline) >= sizeof line) {
				printf("line too long\n");
				continue;
			}
			strcpy(pline+strlen(pline)-1, p+2);
			strcpy(p, pline);
			fprintf(stderr, "%s", p);
		} else 
			if (p[0] != '\n')
				strcpy(pline, p);	/* salt away current command */
		
		if (line[0] == '!') {		/* shell escape */
			system(&line[1]);
		} else if (line[0] == '.') {	/* source file */
			if (source == TRUE) {
				fprintf(stderr, "Sorry, nested source not allowed\n");
				continue;
			}
			++p;
			if (!isspace(*p)) {
				fprintf(stderr,"'.' command must be followed by white space\n");
				continue;
			}
			while (*p != '\n' && isspace(*p)) ++p;
			if (*p == '\n') {
				fprintf(stderr, "source what?\n");
				continue;
			}
			pp = p;
			while (*p != '\n' && !isspace(*p)) ++p;
			if (index(p, '|') != NULL)
				fprintf(stderr, "warning: extra chars after '. <file>' ignored\n");
			*p = '\0';
			source = TRUE;
			if ((input = fopen(pp, "r")) == NULL) {
				fprintf(stderr, "can't source %s\n", pp);
				input = stdin;
				source = FALSE;
				continue;
			}
		} else {
			tok = 1;
			/*
			 * Check for "|" into a shell command line
			 */
			pp = p;
			while ((pp = index(pp, '|')) != NULL) {
				if (!isspace(pp[-1])) {
					++pp; continue;
				}
				*pp = 0;
				filter = popen(pp+1, "w");
				if (filter == NULL) {
					printf("crash: can't pipe to %s\n", pp+1);
					fflush(stdout);
					tok = 0;
				}
			}
		}
	}
	cp = NULL;
	for(;;) {
		switch(*p) {
		case '\n':
			*p = '\0';
		case '\0':
			if (cp == NULL)
				tok = 0;
			break;

		case ' ':
		case '\t':
			*p++ = '\0';
			if (cp == NULL)
				continue;
			break;

		default:
			if (cp == NULL)
				cp = p;
			++p;
			continue;
		}
		return(cp);
	}
}

fatal(str)
{
	printf("crash: %s\n", str);
	fflush(stderr); fflush(stdout);
	exit(1);
}

int paniced;
char *panicstr;
char *version;

heuristics()
{
	char buf1[512];
	char buf2[512];
	register i,j;
	register char *p;
	char c;

	/*
	 * Locate kernel version in kernel
	 */
	bzero(buf1, sizeof buf1);
	bzero(buf2, sizeof buf2);
	version = (char *)search("version");
	if (version == NULL)
		printf("Warning: No 'version' symbol in symbol table!\n");
	else {
		lseek(dynixfd, (long) version, 0);
		if (read(dynixfd, buf1, sizeof (buf1)-1) != sizeof (buf1)-1)
			printf("Warning: read error on dynix version string\n");
		lseek(vmcorefd, (long) version, 0);
		if (read(vmcorefd, buf2, sizeof (buf2)-1) != sizeof (buf2)-1) 
			printf("Warning: read error on vmcore version string\n");
		buf1[strlen(buf1)-1] = '\0';
		buf2[strlen(buf2)-1] = '\0';
		if (buf1[0] == '\0' || buf2[0] == '\0')
			printf("Warning: version string NULL, memory probably zero'ed or dump taken incorrectly\n");
		else if (strcmp(buf1, buf2) != 0) {
			printf("Warning: version string mismatch:\n");
			printf("\t dynix: %s\n", buf1);
			printf("\tvmcore: %s\n", buf2);
		} else
			printf("Version: %s\n", buf1);
	}

	/*
	 * Locate panic string in kernel
	 */
	bzero(buf1, sizeof buf1);
	if (!live && readv(search("panicstr"), &panicstr, sizeof panicstr) == sizeof panicstr) {
		if (panicstr == NULL)
			printf("Warning: panicstr is NULL\n");
		else {
			paniced = 1;
			for (j=i=0; j < sizeof buf1; i++,j += strlen(p)) {
				if (readv(panicstr+i, &c, 1) != 1 || c == 0)
					break;
				strcat(&buf1[j], p = nice_char(c));
			}
			if (j < sizeof buf1)
				printf("Panicstr: %s\n", buf1);
		}
	}

	get_locore();
}
