
#include "crash.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <a.out.h>
#include <stab.h>

#define MIN(a,b)	((a)>(b)?(b):(a))

#define	tprintf		if (debug > 2) printf

/* Tune these for speed */
#define NUM_STE	 (8*BUFSIZ/sizeof (struct nlist))	/* 8k reads */
#define NUM_SYM	 5000
#define NUM_SDB	 5000

char *bsearch();
static	struct	sym *symtable, **symtablepv, **symtablepn, *searchbyval();
static	struct	sdb *sdbtable, **sdbtablepn;
static	int cmpsymbyval(), cmpsymbynam(), cmpsdbbynam();
static	char *stringtable;


static	struct exec exec;
static	int total_symbols, glb_symbols, sdb_symbols;

struct	sdb sdbinfo;

getsymboltable()
{
	register ret;
	struct stat sbuf;
	int stringtblsize;

	ret = fstat(dynixfd, &sbuf);
	SANITY(ret == -1, "bad stat on", dynix);

	ret = lseek(dynixfd, 0L, 0);
	SANITY(ret == -1, "bad lseek on", dynix);

	ret = read(dynixfd, &exec, sizeof (struct exec));
	SANITY(ret != sizeof (struct exec), "can't read a.out header for", dynix);

	SANITY(N_BADMAG(exec), "bad magic number for", dynix);
	total_symbols = exec.a_syms / sizeof (struct nlist);

	ret = lseek(dynixfd, (long) N_TXTOFF(exec), 0);
	SANITY(ret == -1, "N_TXTOFF lseek failed on", dynix);

	stringtblsize = sbuf.st_size - N_STROFF(exec);
	stringtable = malloc( (unsigned)stringtblsize );
	SANITY(stringtable == NULL, "malloc of string table failed", "");

	ret = lseek(dynixfd, (long) N_STROFF(exec), 0);
	SANITY(ret == -1, "N_STROFF lseek failed", "");

	ret = read(dynixfd, stringtable, stringtblsize);
	SANITY(ret != stringtblsize, "can't read string table for", dynix);

	process();
}

process()
{
	struct nlist lnlist[ NUM_STE ];
	register struct nlist *s;
	register n, ret, type;
	register char *cp;
	
	ret = lseek(dynixfd, (long) N_SYMOFF(exec), 0);
	SANITY(ret == -1, "N_SYMOFF lseek failed", "");

	for (n = total_symbols; n > 0; n -= ret) {
		ret = read(dynixfd, (char *)lnlist, sizeof (struct nlist) * MIN(NUM_STE, n));
		SANITY(ret <= 0, "bad read of symbol table", "");
		ret /= sizeof (struct nlist);
		for (s=lnlist; s < &lnlist[ret]; s++) {
			if (s->n_un.n_strx == 0) {
				tprintf("%d: skipped			<<<<\n", (total_symbols - n) + (s - lnlist));
				continue;
			}
			if (s->n_type & N_STAB) {
				s->n_un.n_strx += (unsigned long) stringtable;
				tprintf("%d: sdb_add: ", (total_symbols - n) + (s - lnlist));
				sdb_add(s);
				continue;
			}
			type = (s->n_type & N_TYPE);
			tprintf("%d: n_type,type == %#x,%#x: ", (total_symbols - n) + (s - lnlist), s->n_type, type);
			if (type == N_TEXT || type == N_DATA || type == N_BSS) {
				s->n_un.n_strx += (unsigned long) stringtable;
				cp = (char *) s->n_un.n_strx;
				if (   (   cp[0] == 'L'
					&& cp[1] >= '0' 
					&& cp[1] <= '9'
				        )
				  ||   (   cp[0] == '.'
				        && cp[1] == 'L'
					)
				  ||    strcmp(cp, "_rcsid") == 0
				  ) { /* skip these symbols */
					tprintf("'%s', value=%#x   SKIPPED <<<<<<<\n", cp, s->n_value);
					continue;
				}
				sym_add(s);
				tprintf("'%s', value=%#x   REG symbol\n", (char *) s->n_un.n_strx, s->n_value);
				continue;
			}
			tprintf("\n");
		}
	}

	/* allocate memory for for arrays of pointers */
	symtablepn = (struct sym **) malloc ( (unsigned) glb_symbols * sizeof (struct sym *) );
	SANITY(symtablepn == NULL, "malloc failed", "");
	symtablepv = (struct sym **) malloc ( (unsigned) glb_symbols * sizeof (struct sym *) );
	SANITY(symtablepv == NULL, "malloc failed", "");
	sdbtablepn = (struct sdb **) malloc ( (unsigned) sdb_symbols * sizeof (struct sdb *) );
	SANITY(sdbtablepn == NULL, "malloc failed", "");

	/* set all pointers before sorting them */
	for (n=0; n < glb_symbols; n++)
		symtablepn[n] = symtablepv[n] = &symtable[n];
	for (n=0; n < sdb_symbols; n++)
		sdbtablepn[n] = &sdbtable[n];

	/* sort global symbols by name and value, sdb by name only */
	qsort((char *)symtablepn, glb_symbols, sizeof (struct sym *), cmpsymbynam);
	qsort((char *)symtablepv, glb_symbols, sizeof (struct sym *), cmpsymbyval);
	qsort((char *)sdbtablepn, sdb_symbols, sizeof (struct sdb *), cmpsdbbynam);

#ifdef	VERBOSE
	printf("%d total symbols (%d=global, %d=sdb) %d%% useful\n", 
		total_symbols, glb_symbols, sdb_symbols, 
		((glb_symbols+sdb_symbols)*100)/total_symbols);
#endif
}

static int
cmpsymbyval(a, b)
	struct sym **a, **b;
{

	return  (*a)->sym_value - (*b)->sym_value;
}

static int
cmpsymbynam(a, b)
	struct sym **a, **b;
{
	register char *s1, *s2;

	/* expanded strcmp() */
	s1 = (*a)->sym_name;
	s2 = (*b)->sym_name;
	if (debug > 5) printf("cmpsymbynam: '%s' .. '%s'\n", s1, s2);
	while (*s1 == *s2++)
		if (*s1++ == '\0')
			return 0;
	return (*s1 - *--s2);
}

static int
cmpsdbbynam(a, b)
	struct sdb **a, **b;
{
	register char *s1, *s2;

	/* expanded strcmp() */
	s1 = (*a)->sdb_name;
	s2 = (*b)->sdb_name;
	if (debug > 5) printf("cmpsdbbynam: '%s' .. '%s'\n", s1, s2);
	while (*s1 == *s2++)
		if (*s1++ == '\0')
			return 0;
	return (*s1 - *--s2);
}

sym_add(s)
	struct nlist *s;
{
	static struct sym *syme, *symp;

	if (symp == NULL || symp >= syme) {
		unsigned size = (unsigned)syme - (unsigned)symtable + (NUM_SYM * sizeof (struct sym));
		if (symp == NULL)
			symtable = (struct sym *) malloc ( size );
		else
			symtable = (struct sym *) realloc( (char *) symtable, size );
		SANITY(symtable == NULL, "malloc for ste symbol table failed", "");
		size /= sizeof (struct sym);
		symp = &symtable[ size - NUM_SYM ];
		syme = &symtable[ size ];
	}
	symp->sym_name = (char *)s->n_un.n_strx;
	symp->sym_value = s->n_value;
	++glb_symbols;
	++symp;
}

sdb_add(s)
	register struct nlist *s;
{
	static struct sdb *sdbe, *sdbp;

	if (sdbp == NULL || sdbp >= sdbe) {
		unsigned size = (unsigned)sdbe - (unsigned)sdbtable + (NUM_SDB * sizeof (struct sdb));
		if (sdbp == NULL)
			sdbtable = (struct sdb *) malloc ( size );
		else
			sdbtable = (struct sdb *) realloc( (char *)sdbtable, size );
		tprintf("*** MALLOC ***\n");
		SANITY(sdbtable == NULL, "malloc for sdb symbol table failed", "");
		size /= sizeof (struct sdb);
		sdbp = &sdbtable[ size - NUM_SDB ];
		sdbe = &sdbtable[ size ];
	}

	if (s->n_type == N_LENG) {
		tprintf("n_type == N_LENG, n_strx = %#x, n_other = %#x, n_desc = %#x, *n_value = %#x*\n",
			s->n_un.n_strx, s->n_other, s->n_desc, s->n_value);
		sdbp[-1].sdb_size = s->n_value;
		return;
	}

	/* Only keep global symbols (N_GSYM) and structure elements (N_SSYM) for now. */
	if (s->n_type != N_GSYM && s->n_type != N_SSYM) {
		tprintf("(n_type=%#x) != N_GSYM or N_SSYM => FLUSH\n", s->n_type);
		return;
	}

	sdbp->sdb_name = (char *)(s->n_un.n_strx);
	sdbp->sdb_type = s->n_type;
	sdbp->sdb_desc = s->n_desc;
	sdbp->sdb_soff = s->n_value;
	sdbp->sdb_size = 1;		/* may be adjusted later by N_LENG stab */
	tprintf("kept '%s', type=%#x, desc=%#x, soff=%#x\n", sdbp->sdb_name, sdbp->sdb_type, sdbp->sdb_desc, sdbp->sdb_soff);
	++sdb_symbols;
	++sdbp;
}


/*
 * Search routines
 */

/* given a value, looks up the name */
static struct sym *
searchbyval(value, offset)
	unsigned value, *offset;
{
	struct sym lsym, *p = &lsym;
	register struct sym **pp, **ppend;

	*offset = 0;
	p->sym_value = value;
	pp = (struct sym **) bsearch((char *)&p, (char *)symtablepv, glb_symbols, sizeof (struct sym *), cmpsymbyval);
	if (pp == NULL) {
		/*
		 * Binary search for exact value failed, so try again with linear
		 * search to find value just before so result is "value+offset".
		 */
		for( pp = symtablepv, ppend = &symtablepv[glb_symbols]; pp != ppend; pp++) {
			if ( (*pp)->sym_value < value ) 
				continue;
			--pp;
			*offset = value - (*pp)->sym_value;
			return *pp;
		}
		return (struct sym *)NULL;
	}
	return *pp;
}

/* given a name, look up a value */
struct sym *
searchbynam(s)
	char *s;
{
	struct sym lsym, *p = &lsym, **pp;
	char buf[ MAXSTRLEN + 1 ];

	p->sym_name = s;
	pp = (struct sym **) bsearch((char *)&p, (char *)symtablepn, glb_symbols, sizeof (struct sym *), cmpsymbynam);
	if (pp == NULL)  {	/* try '_'name */
		buf [ 0 ] = '_';
		strcpy(&buf[1], s);
		p->sym_name = buf;
		pp = (struct sym **) bsearch((char *)&p, (char *)symtablepn, glb_symbols, sizeof (struct sym *), cmpsymbynam);
		if (pp == NULL)
			return (struct sym *)NULL;
	}
	return *pp;
}

/* given a name, return any sdb info */
struct sdb *
searchsdb(s)
	char *s;
{
	struct sdb lsdb, *p = &lsdb, **pp;

	p->sdb_name = s;
	pp = (struct sdb **) bsearch((char *)&p, (char *)sdbtablepn, sdb_symbols, sizeof (struct sdb *), cmpsdbbynam);
	if (pp == NULL)
		return (struct sdb *)NULL;
	return *pp;
}
	
/* Global interfaces */
char * 
lookbyval(addr, offsetp)
	unsigned addr, *offsetp;
{
	struct sym *p;

	if (addr < symtablepv[0]->sym_value || addr > symtablepv[glb_symbols-1]->sym_value)
		return NULL;
	p = searchbyval(addr, offsetp);
	if (p == NULL)
		return NULL;
	return p->sym_name;
}

unsigned int
lookbysym(symbol, value)
	char *symbol;
	unsigned *value;
{
	struct sym *p;
	struct sdb *pp;
	struct set_val *ppp;
	
	if (symbol == 0 || *symbol == '\0')
		return FALSE;
	*value = 0;
	bzero(&sdbinfo,  sizeof sdbinfo);
	p = searchbynam(symbol);
	if ( p ) *value = p->sym_value;
	pp = searchsdb(symbol);
	if ( pp ) sdbinfo = *pp;
	ppp = searchsetval(symbol);
	if ( ppp ) {
		*value = ppp->sv_val.v_value;
		sdbinfo = ppp->sv_val.v_sdb;
	}
	if ( p || pp || ppp )
		return TRUE;
	return FALSE;
}

/* return the first descripter in *x */
/* see struct desc in crash.h */
desc(x)
	register short *x;
{
	register i;

	for (i = *x &~ 0xf; i; i <<= 2) {
		if ((i & 0xc000) == 0)
			continue;
		return  (i & 0xc000) >> 14;
	}
	return 0;
}
