/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */


/*
 * ident	"$Header: zdverify.c 1.3 89/09/13 $"
 * Disk verification routines
 */

/* $Log:	zdverify.c,v $
 */

#include <sys/types.h>
#include <sys/param.h>
#include <stdio.h>
#include <errno.h>
#ifdef BSD
#include <sys/fs.h>
#include <sys/ioctl.h>
#include <zdc/zdc.h>
#include <zdc/zdbad.h>
#include <zdc/zdioctl.h>
#else
#include <sys/ufsfilsys.h>
#include <sys/zdc.h>
#include <sys/zdbad.h>
#include <sys/zdioctl.h>
#endif
#include "zdformat.h"
#include "format.h"

#define	MIN(a,b) (((a)<(b))?(a):(b))

extern caddr_t malloc();
extern int	sigints;

#define NBEATS		16	/* no. times to beat up a track after error */

extern int	fullpasses;	/* no. of full passes (all cylinders) */
extern int	defectpasses;	/* no. of defect passes (tracks with spots) */
extern bool_t	checkdata;	/* check data? */

caddr_t	vwbp;			/* Verify write buffer */
caddr_t	vrbp;			/* Verify read buffer */
struct cnterr {
	int	ecc;		/* EECC */
	int	hdrecc;		/* EHDRECC */
	int	bse;		/* EBSE */
	int	so;		/* ESO  */
	int	her;		/* EHER */
	int	misc;		/* Miscellaneaous */
	int	all;		/* Sum of above errors */
} cnterr;

/*
 * Purdue/EE severe burn-in patterns.
 * 	Modified by Sequent, Inc. - Feb 86
 *	Swabbed for ZDC.
 */
unsigned ppat[] = {
	0x4CDBF6CC,	0xB66DDBB6,	0xAAAAAAAA,	0x55555555,
	0x33333333,	0x711CC771,	0xDBB66DDB,	0x8EE38EE3,
	0xC7711CC7,	0x8EE3388E,	0x1CC7711C,	0x388EE338,
	0x49922449,	0x24499224,	0x92244992,	0x6DDBB66D,
	0xB66DDBB6,	0x55555555,	0x00000000,	0xAAAAAAAA,
	0xFFFFFFFF,	0x4CDBF6CC,	0xAAAAAAAA,	0x4CDBF6CC,
	0x55555555,	0x4CDBF6CC,	0xAAAAAAAA,	0x4CDBF6CC,
	0x55555555,	0x4CDBF6CC,	0xAAAAAAAA,	0x4CDBF6CC,
	0x55555555,	0x599AA559,	0xAAAAAAAA,	0x599AA559,
	0x55555555,	0x599AA559,	0xAAAAAAAA,	0x599AA559,
	0x55555555,	0xAAAAAAAA,	0x55555555,	0xAAAAAAAA,
	0x55555555,	0xAAAAAAAA,	0x55555555,	0xAAAAAAAA,
};

/*
 * verify
 *	Verify the disk between startcyl and lastcyl. Typically, the
 *	whole disk is verified. 
 * WARNING: This will scribble on the disk.
 */
doverify()
{
	extern int sigints;
	register int pass;
	int	npat;			/* Pattern number */
	int	tracksize;		/* bytes in a track */

	npat = 0;
	if (verbose)
		printf("Starting verify passes.\n");
	/*
	 * Set for severe burn-in.
	 */
	if (ioctl(fd, ZIOSEVERE, (char *)NULL) < 0) {
		perror("ioctl ZIOSEVERE error");
		fprintf(stderr, "Unable to set device for severe burn-in\n");
		fprintf(stderr, "...exiting\n");
	}
	tracksize = chancfg->zdd_sectors << DEV_BSHIFT;
	vwbp = MALLOC_ALIGN(tracksize, DEV_BSIZE);
	vrbp = MALLOC_ALIGN(tracksize, DEV_BSIZE);
	if (!vwbp || !vrbp) {
		fprintf(stderr, "doaddbad: malloc error\n");
		fprintf(stderr, "...exiting\n");
		exit(1);	
	}
	bzero((caddr_t)vwbp, tracksize);
	bzero((caddr_t)vrbp, tracksize);

	/*
	 * Do full passes.
	 */
	sigints = 0;
	for (pass = 0; pass < fullpasses; pass++) {
		if (sigints)
			break;
		if (verbose)
			printf("Begin Full verify pass %d.\n", pass + 1);
		if (dofullpass(tracksize, npat) == FAIL)
			fprintf(stderr,
			"Warning: Defect(s) found during full pass %d\n",
				pass + 1);
		if (++npat > (sizeof(ppat) / sizeof(ppat[0])))
			npat = 0;
	}

	/*
	 * Do defect passes. If error, add another pass.
	 */
	sigints = 0;
	for (pass = 0; pass < defectpasses; pass++) {
		if (sigints)
			break;
		printf("Begin Defect verify pass %d.\n", pass + 1);
		if (dodefectpass(tracksize, npat) == FAIL) {
			defectpasses++;
			fprintf(stderr,
			"Error found - Adding an additional Defect verify pass.\n");
		}
		if (++npat > (sizeof(ppat) / sizeof(ppat[0])))
			npat = 0;
	}

	/*
	 * Report Errors.
	 */
	printf("Errors found during verify:\n"); 
	printf("%d ECC errors.\n", cnterr.ecc);
	printf("%d HDR ECC errors.\n", cnterr.hdrecc);
	printf("%d BSE errors.\n", cnterr.bse);
	printf("%d SO errors.\n", cnterr.so);
	printf("%d HARD errors.\n", cnterr.her);
	printf("%d Misc. errors.\n", cnterr.misc);
	printf("%d Total errors.\n", cnterr.all);

	/*
	 * Update various lists and data on disk since verify may have
	 * either found and added new bad blocks or clobbered the cylinder
	 * where the data resides.
	 */
	if (ioctl(fd, ZIONSEVERE, (char *)NULL) < 0)
		perror("ioctl ZIONSEVERE error");
	return;
}

/*
 * dofullpass
 *	Do full verify pass across entire disk.
 *
 * Return:	SUCCESS - No failures found.
 *		Fail	- At least 1 failure found during verify.
 */
static
dofullpass(tracksize, npat)
	int	tracksize;	/* tracksize in bytes */
	int	npat;		/* pattern (index into ppat array) */
{
	register int track;
	register int cyl;
	int	sector;
	int	trkadj;		/* delta for short tracks */
	int	ndefect;	/* number of defects - spare on short track */
	bool_t	badblk;

	badblk = FALSE;
	/*
	 * Initialize buffer with pattern.
	 */
	bufinit(vwbp, npat, tracksize);

	/*
	 * Begin verify, for each track:
	 * 	- Write data in buffer
	 *	- Read data. Hardware checks header and data ECC.
	 *	- If error, add to bad block list and format appropriatly.
	 */
	sigints = 0;
	for (cyl = startcyl; cyl <= lastcyl; cyl++) {
		track = 0;
		if (sigints)
			break;
		while (track < chancfg->zdd_tracks) {
			if (cyl == ZDD_DDCYL && track == 0) {
				/*
				 * Special case: Do not try to write/read
				 * Special Sectors (zdcdd).
				 */
				testwrite(ZDD_NDDSECTORS, vwbp,
					tracksize - (ZDD_NDDSECTORS*DEV_BSIZE));
				if (testread(ZDD_NDDSECTORS, vrbp,
						tracksize - (ZDD_NDDSECTORS*DEV_BSIZE),
						npat) == FAIL) {
					badblk = TRUE;
					continue;	/* redo track */
				}
			} else {
				trkadj = 0;
				if (cyl == ZDD_DDCYL
				|| cyl >= (chancfg->zdd_cyls - ZDD_NDGNCYL)) {
					/*
					 * Only write good part of short tracks
					 */
					ndefect = badsectontrk(cyl, track);
					ndefect -= chancfg->zdd_spare;
					if (ndefect > 0)
						trkadj = ndefect << DEV_BSHIFT;
				}
				sector = (cyl * chancfg->zdd_tracks * chancfg->zdd_sectors)
					 + (track * chancfg->zdd_sectors);
				testwrite(sector, vwbp, tracksize - trkadj);
				if (testread(sector, vrbp, tracksize - trkadj, npat) == FAIL) {
					badblk = TRUE;
					continue;	/* Redo track */
				}
			}
			++track;
		}
		/*
		 * test spares until no more problem
		 */
		while (testspares(cyl, vwbp, vrbp, npat) == FAIL)
			badblk = TRUE;
	}
	return((badblk == TRUE) ? FAIL : SUCCESS);
}

/*
 * dodefectpass
 *	Do verify pass on only tracks with defects.
 * This is done by simultaneously walking down both the bad block list
 * and the mfg defect list. The mfg defect list is needed so that tracks
 * with defects in the gap or head scatter are tested.
 *
 * Return:	SUCCESS - No failures found.
 *		Fail	- At least 1 failure found during verify.
 */
static
dodefectpass(tracksize, npat)
	int	tracksize;	/* tracksize in bytes */
	int	npat;		/* pattern (index into ppat array) */
{
	register struct bz_bad *bzp;
	register struct bm_mfgbad *bmp;
	register int	trk;		/* track to test */
	register bool_t	doanother;	/* telescope test pattern */
	register int	cyl;
	int	btrack;			/* base track */
	int	bcyl;			/* base cylinder */
	struct	cyl_hdr *chp;
	bool_t	foundbad;
	bool_t	badblk;
	struct	bz_bad	*find_trk();

	foundbad = FALSE;
	badblk = FALSE;
	bzp = bbl->bz_bad;
	bmp = mfg->bm_mfgbad;
	/*
	 * Initialize buffer with pattern.
	 */
	bufinit(vwbp, npat, tracksize);

	/*
	 * Skip to start cylinder.
	 */
	while (bzp < &bbl->bz_bad[bbl->bz_nelem] && bzp->bz_cyl < startcyl)
		++bzp;
	while (bmp < &mfg->bm_mfgbad[mfg->bm_nelem] && bmp->bm_cyl < startcyl)
		++bmp;
	/*
	 * Verify those tracks with defects
	 */
	for(;;) {
		/*
		 * Check if we are done.
		 * If not, find base (bcyl, btrack).
		 */
		if (bzp == &bbl->bz_bad[bbl->bz_nelem]) {
			if (bmp == &mfg->bm_mfgbad[mfg->bm_nelem])
				break;		/* DONE */
			if (bmp->bm_cyl > lastcyl)
				break;		/* DONE */
			bcyl = bmp->bm_cyl;
			btrack = bmp->bm_head;
		} else if (bmp == &mfg->bm_mfgbad[mfg->bm_nelem]) {
			if (bzp->bz_cyl > lastcyl)
				break;		/* DONE */
			bcyl = bzp->bz_cyl;
			btrack = bzp->bz_head;
		} else if (bzp->bz_cyl < bmp->bm_cyl) {
			bcyl = bzp->bz_cyl;
			btrack = bzp->bz_head;
		} else if (bmp->bm_cyl < bzp->bz_cyl) {
			bcyl = bmp->bm_cyl;
			btrack = bmp->bm_head;
		} else {
			bcyl = bzp->bz_cyl;
			btrack = MIN(bzp->bz_head, bmp->bm_head);
		}

		/*
		 * Test base (bcyl, btrack)
		 * remap_cyl so testtrack can find spares.
		 */
		chp = &cyls[bcyl & 1];
		remap_cyl(bcyl, chp);
		while (testtrack(bcyl, btrack, tracksize, npat, chp) == FAIL)
			foundbad = TRUE;

		/*
		 * Test tracks around (bcyl, btrack).
		 * First before then after.
		 */
		trk = btrack;
		do {
			trk = (trk-1+chancfg->zdd_tracks) % chancfg->zdd_tracks;
			doanother = FALSE;
			while (testtrack(bcyl,trk,tracksize,npat,chp) == FAIL) {
				doanother = TRUE;
				foundbad = TRUE;
			}
		} while (doanother == TRUE);

		trk = btrack;
		do {
			trk = (trk + 1) % chancfg->zdd_tracks;
			doanother = FALSE;
			while (testtrack(bcyl,trk,tracksize,npat,chp) == FAIL) {
				doanother = TRUE;
				foundbad = TRUE;
			}
		} while (doanother == TRUE);

		/*
		 * Test same track on adjacent cylinders.
		 * Do not go out of "startcyl, lastcyl" bounds.
		 */
		for (cyl = bcyl - 1; cyl >= startcyl; cyl--) {
			doanother = FALSE;
			chp = &cyls[cyl & 1];
			remap_cyl(cyl, chp);
			while (testtrack(cyl,btrack,tracksize,npat,chp) == FAIL)
				doanother = TRUE;
			if (doanother == FALSE)
				break;		/* done */
			foundbad = TRUE;
		}

		for (cyl = bcyl + 1; cyl <= lastcyl; cyl++) {
			doanother = FALSE;
			chp = &cyls[cyl & 1];
			remap_cyl(cyl, chp);
			while (testtrack(cyl,btrack,tracksize,npat,chp) == FAIL)
				doanother = TRUE;
			if (doanother == FALSE)
				break;		/* done */
			foundbad = TRUE;
		}

		/*
		 * Skip to next track or cylinder.
		 * If the bad block list has changed, then must find
		 * this track in the new bad block list.
		 */
		if (foundbad == TRUE) {
			badblk = TRUE;
			bzp = find_trk(bcyl, btrack);
			foundbad = FALSE;
		}
		for (; bzp < &bbl->bz_bad[bbl->bz_nelem]; bzp++) {
			if (bzp->bz_cyl == bcyl && bzp->bz_head == btrack)
				continue;
			break;
		}
		for (; bmp < &mfg->bm_mfgbad[mfg->bm_nelem]; bmp++) {
			if (bmp->bm_cyl == bcyl && bmp->bm_head == btrack)
				continue;
			break;
		}
	}
	return((badblk == TRUE) ? FAIL : SUCCESS);
}

/*
 * find_trk
 *	find entry in bad block list for given cyl and trk.
 * return pointer to first bad block entry on track. If there are no
 * entries for this track, return pointer to next highest entry or end of list.
 */
static struct bz_bad *
find_trk(cyl, trk)
	register int cyl;
	int trk;
{
	register struct bz_bad *bzp;

	for (bzp = bbl->bz_bad; bzp < &bbl->bz_bad[bbl->bz_nelem]; bzp++) {
		if (bzp->bz_cyl < cyl)
			continue;
		if (bzp->bz_cyl > cyl)
			break;
		if (bzp->bz_head < trk)
			continue;
		if (bzp->bz_head >= trk)
			break;
	}
	return(bzp);
}

/*
 * badsectontrk
 *	how many bad sectors on this track
 */
static int
badsectontrk(cyl, trk)
	register int cyl;
	register int trk;
{
	register struct bz_bad *bzp;
	int ndefect;

	ndefect = 0;
	for (bzp = bbl->bz_bad; bzp < &bbl->bz_bad[bbl->bz_nelem]; bzp++) {
		if (bzp->bz_cyl < cyl)
			continue;
		if (bzp->bz_cyl > cyl)
			break;
		if (bzp->bz_head < trk)
			continue;
		if (bzp->bz_head > trk)
			break;
		/* Found one */
		++ndefect;
	}
	return(ndefect);
}

/*
 * testtrack
 *	write/read track to verify disk integrity.
 *	Also test unused spare(s) on track.
 * Return: FAIL if error found; else SUCCESS.
 */
static
testtrack(cyl, track, tracksize, npat, chp)
	register int	cyl;
	int	track;
	int	tracksize;	/* tracksize in bytes */
	int	npat;		/* pattern (index into ppat array) */
	struct	cyl_hdr *chp;
{
	register struct hdr	*hp;
	register struct track_hdr *tp;
	register int	trkadj;	/* delta for short tracks */
	int	ndefect;	/* number of defects - spare on short track */
	int	sector;
	bool_t	retval;

	retval = SUCCESS;
	if (cyl == ZDD_DDCYL && track == 0) {
		/*
		 * Special case: Do not try to write/read
		 * Special Sectors (zdcdd).
		 */
		testwrite(ZDD_NDDSECTORS, vwbp,
				tracksize - (ZDD_NDDSECTORS*DEV_BSIZE));
		if (testread(ZDD_NDDSECTORS, vrbp,
				tracksize - (ZDD_NDDSECTORS*DEV_BSIZE), npat)
				== FAIL) {
			retval = FAIL;
			remap_cyl(cyl, chp);
		}
	} else {
		trkadj = 0;
		if (cyl == ZDD_DDCYL ||
		    cyl >= (chancfg->zdd_cyls - ZDD_NDGNCYL)) {
			/*
			 * Only write good part of short tracks
			 */
			ndefect = badsectontrk(cyl, track);
			ndefect -= chancfg->zdd_spare;
			if (ndefect > 0)
				trkadj = ndefect << DEV_BSHIFT;
		}
		sector = (cyl * chancfg->zdd_tracks * chancfg->zdd_sectors)
			 + (track * chancfg->zdd_sectors);
		testwrite(sector, vwbp, tracksize - trkadj);
		if (testread(sector, vrbp, tracksize - trkadj, npat) == FAIL) {
			retval = FAIL;
			/*
			 * If there can be a spare, then we must remap
			 * the cylinder so that appropriate spare(s) can
			 * be tested.
			 */
			if (trkadj == 0)
				remap_cyl(cyl, chp);
		}
	}

	/*
	 * Test any spares on this track.
	 */
	tp = &chp->c_trk[track];
	for (hp = tp->t_hdr; hp < &tp->t_hdr[totspt]; hp++) {
		if (hp->h_type == ZD_GOODSPARE) {
			if (rwspare(hp, vwbp, vrbp, npat) == FAIL) {
				retval = FAIL;
				/* short cut remap_cyl() */
				hp->h_type = ZD_BADUNUSED;
			}
		}
	}
	return(retval);
}

/*
 * testwrite
 *	write test pattern to disk.
 */
static
testwrite(sector, cbp, length)
	register int sector;
	caddr_t	cbp;
	int length;
{
	register long sn;

	lseek(fd, sector << DEV_BSHIFT, 0);
	if (write(fd, cbp, length) == length)
		return;

	for (sn = sector; sn < sector + (length >> DEV_BSHIFT); sn++) {
		lseek(fd, sn << DEV_BSHIFT, 0);
		(void) write(fd, cbp, DEV_BSIZE);
	}
}

/*
 * testread
 *	read back previously written track. If checkdata, confirm data
 *	matches that written. Otherwise, rely on hardware to verify via
 *	ECC errors.
 *
 * Return:	SUCCESS - No failures found.
 *		Fail	- At least 1 failure found during verify.
 */
static
testread(sector, cbp, length, npat)
	register int sector;
	caddr_t	cbp;
	int	length;
	int	npat;			/* Test pattern to check */
{
	int cc;
	register int i;
	register int sn;
	int cyl;
	int trk;
	int nspc;
	int ok;
	int errcode;

	ok = SUCCESS;
	nspc = chancfg->zdd_tracks * chancfg->zdd_sectors;
	cyl = sector / nspc;
	trk = (sector % nspc) / chancfg->zdd_sectors;
	lseek(fd, sector << DEV_BSHIFT, 0);
	if (checkdata)
		bzero((caddr_t)cbp, (uint)length);
	cc = read(fd, cbp, length);
	if (cc == length) {
		/*
		 * Read all data. Check if valid.
		 */
		if (checkdata && checkbuf(cbp, &cc, npat)) {
			fprintf(stderr, "Data bad at (%d, %d, %d+%d)\n",
				cyl, trk, sector + (cc >> DEV_BSHIFT), cc);
			recorderror(sector + (cc >> DEV_BSHIFT),
					0, (struct hdr *)NULL);
			return(FAIL);
		}
		return(SUCCESS);
	}

	/*
	 * Find the problem sector(s).
	 */
	for (sn = sector; sn < sector + (length >> DEV_BSHIFT); sn++) {
		for (i = NBEATS; i > 0; i--) {
			lseek(fd, sn << DEV_BSHIFT, 0);
			if (checkdata)
				bzero(cbp, DEV_BSIZE);
			cc = read(fd, cbp, DEV_BSIZE);
			if (cc == DEV_BSIZE) {
				if (checkdata && checkbuf(cbp, &cc, npat)) {
					fprintf(stderr,
					"Data bad at (%d, %d, %d+%d)\n",
						cyl, trk, sn, cc);
					recorderror(sn, 0,
							(struct hdr *)NULL);
					ok = FAIL;
					break;
				}
			} else {
				errcode = 0;
				if (cc < 0) 	/* get error */
					if (ioctl(fd, ZIOGERR, &errcode) < 0)
						perror("ioctl ZIOGERR failed");
				fprintf(stderr, "Sector %d, read error\n", sn);
				recorderror(sn, errcode, (struct hdr *)NULL);
				ok = FAIL;
				break;
			}
		}
	}
	return(ok);
}

/*
 * rwspare
 *	If checkdata, confirm data matches that written.
 *	Otherwise, rely on hardware to verify via ECC errors.
 *
 * Return:	SUCCESS - No failures found.
 *		Fail	- At least 1 failure found during verify.
 */
static
rwspare(hp, wcbp, rcbp, npat)
	register struct hdr	*hp;
	caddr_t	wcbp;			/* buffer with pattern to write */
	caddr_t	rcbp;			/* buffer with pattern to read  */
	int	npat;			/* Test pattern to check */
{
	int	cc;
	int	errcode;		/* copy of errno */
	struct cb lcb;			/* cb for ioctl */

	bzero((caddr_t)&lcb, sizeof(struct cb));
	lcb.cb_cmd = ZDC_WRITE;
	lcb.cb_cyl = hp->h_cyl;
	lcb.cb_head = hp->h_head;
	lcb.cb_sect = hp->h_sect;
	lcb.cb_addr = (u_long)wcbp;
	lcb.cb_count = DEV_BSIZE;
	lcb.cb_mod = 0;
	lcb.cb_iovec = 0;
	/* ignore write errors */
	(void)ioctl(fd, ZIOCBCMD, (char *)&lcb);

	/*
	 * Now read back the written data.
	 */
	if (checkdata)
		bzero(rcbp, DEV_BSIZE);
	lcb.cb_cmd = ZDC_READ;
	lcb.cb_cyl = hp->h_cyl;
	lcb.cb_head = hp->h_head;
	lcb.cb_sect = hp->h_sect;
	lcb.cb_addr = (u_long)rcbp;
	lcb.cb_count = DEV_BSIZE;
	lcb.cb_mod = 0;
	lcb.cb_iovec = 0;
	errcode = 0;
	if (ioctl(fd, ZIOCBCMD, (char *)&lcb) < 0) {
		if (ioctl(fd, ZIOGERR, &errcode) < 0)
			perror("ioctl ZIOGERR failed");
		fprintf(stderr, "Read error at (%d, %d, %d)\n", hp->h_cyl,
			hp->h_head, hp->h_sect);
		recorderror((int)hp->h_sect, errcode, hp);
		return(FAIL);
	}

	cc = DEV_BSIZE;

	if (checkdata && checkbuf(rcbp, &cc, npat)) {
		fprintf(stderr,
		"Data Bad at (%d, %d, %d+%d).\n", hp->h_cyl, hp->h_head,
			hp->h_sect, cc);
		recorderror((int)hp->h_sect, 0, hp);
		return(FAIL);
	}
	return(SUCCESS);
}

/*
 * testspares
 *	Find and write/read the spare sectors in a cylinder.
 *	rwspare does actual dirty work.
 *
 * Return:	SUCCESS - No failures found.
 *		Fail	- At least 1 failure found during verify.
 */
static
testspares(cyl, wcbp, rcbp, npat)
	int	cyl;
	caddr_t	wcbp;		/* buffer with pattern to write */
	caddr_t	rcbp;		/* buffer with pattern to read  */
	int	npat;			/* Test pattern to check */
{
	register struct hdr	*hp;
	register struct track_hdr *tp;
	struct	cyl_hdr *chp;
	int	retval;

	chp = &cyls[cyl & 1];
	retval = SUCCESS;

	remap_cyl(cyl, chp);
	for (tp = chp->c_trk; tp < &chp->c_trk[chancfg->zdd_tracks]; tp++) {
		for (hp = tp->t_hdr; hp < &tp->t_hdr[totspt]; hp++) {
			if (hp->h_type == ZD_GOODSPARE) {
				if (rwspare(hp, wcbp, rcbp, npat) == FAIL)
					retval = FAIL;
			}
		}
	}
	return(retval);
}

/*
 * recorderror
 *	record the error (from errno) and add sector to bad block list.
 *
 * if spare sector is in error, hp points to hdr structure where cyl, head
 * and sector are extracted. If USED sector hp == NULL.
 */
static
recorderror(sn, errcode, hp)
	register int	sn;		/* sector number */
	int	errcode;
	struct	hdr	*hp;
{
	register struct bz_bad *bzp;
	register int nspc;
	int cyl;
	int trk;

	(void)ioctl(fd, ZIONSEVERE, (char *)NULL);

	fprintf(stderr, "Error %d on read.\n", errcode);
	cnterr.all++;

	switch (errcode) {

	case ZDC_ECC:
		cnterr.ecc++;
		break;

	case ZDC_HDR_ECC:
		cnterr.hdrecc++;
		break;

	case ZDC_SNF:
		cnterr.bse++;
		break;

	case ZDC_SO:
		cnterr.so++;
		break;

	case ZDC_ILLCHS:
	case ZDC_BADDRV:
	case ZDC_ACCERR:
		cnterr.misc++;
		break;

	default:			/* hard error */
		cnterr.her++;
		break;

	}

	/*
	 * Don't mess around - add bad spot to bad block list.
	 */
	if (errcode == ZDC_HDR_ECC || errcode == ZDC_SNF || errcode == ZDC_SO) 
		addlist[0].al_type = BZ_BADHEAD;
	else
		addlist[0].al_type = BZ_BADDATA;
	if (hp != (struct hdr *)NULL) {
		/*
		 * addbad the spare sector.
		 */
		addlist[0].al_addr.da_cyl  = hp->h_cyl;
		addlist[0].al_addr.da_head = hp->h_head;
		addlist[0].al_addr.da_sect = hp->h_sect;
		doaddbad(1, Z_VERIFY);
		(void)ioctl(fd, ZIOSEVERE, (char *)NULL);
		return;
	}
	nspc = chancfg->zdd_tracks * chancfg->zdd_sectors;
	cyl = sn / nspc;
	sn %= nspc;
	trk = sn / chancfg->zdd_sectors;
	sn %= chancfg->zdd_sectors;

	/*
	 * Search bad block list to determine if error is really in
	 * a replacement sector.
	 */
	bzp = find_bbl_entry(bbl, cyl, (u_char)trk, (u_char)sn);
	/*
	 * If not found in list enter disk address.
	 * If AUTOREVECT sector header now is bad, AUTOREVECT can
	 * no longer be done.
	 */
	if (bzp == (struct bz_bad *)NULL
	||  (bzp->bz_rtype == BZ_AUTOREVECT && errcode == ZDC_HDR_ECC)) {
		addlist[0].al_addr.da_cyl = cyl;
		addlist[0].al_addr.da_head = trk;
		addlist[0].al_addr.da_sect = sn;
	} else if (bzp->bz_rtype == BZ_AUTOREVECT) {
		/*
		 * Replacement for BZ_AUTOREVECT entry is bad.
		 */
		addlist[0].al_addr.da_cyl = cyl;
		addlist[0].al_addr.da_head = trk;
		addlist[0].al_addr.da_sect = sn | ZD_AUTOBIT;
	} else {
		/*
		 * Replacement of BZ_SNF type entry is bad.
		 */
		addlist[0].al_addr = bzp->bz_rpladdr;
	}
	doaddbad(1, Z_VERIFY);
	(void)ioctl(fd, ZIOSEVERE, (char *)NULL);
}

/*
 * bufinit
 *	Initialize the buffer with the requested pattern.
 */
static
bufinit(bp, npat, size)
	char *bp;
	int npat;
	int size;
{
	register long *lp;
	register long *end;
	register ptrn;

	size = (size + sizeof(long) -1) / sizeof(long);
	lp = (long *)bp;
	end = lp + size;
	ptrn = ppat[npat];
	if (verbose)
		printf("Write pattern 0x%x\n", ptrn);
	
	while (lp < end)
		*lp++ = ptrn;
}

/*
 * checkbuf
 *	compare buffer against expected pattern.
 *
 * Return: number of mismatches found. Size is modified to return character
 *	   position of first mismatch.
 */
static
checkbuf(bp, size, npat)
	char	*bp;
	int	*size;
	int	npat;
{
	register nlongs;
	register long *lp;
	register long p;
	register nerrs = 0;

	nlongs = (*size + sizeof(long) -1) / sizeof(long);
	lp = (long *)bp;
	p = ppat[npat];

	while (nlongs--) {
		if (*lp != p) {
			fprintf(stderr, "Data error 0x%x != 0x%x\n", *lp, p);
			if (nerrs == 0)
				*size = (char *)lp - bp;
			nerrs++;
		}
		lp++;
	}
	return(nerrs);
}
