/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifdef	KXX
/* $Header: atan.s 1.4 87/06/19 $
 *
 *		ARC TANGENT FUNCTION
 *
 * Functional Description:
 *
 *	This routine implements the arctangent function for the C language.
 *	Both atan and atan2 are computed in this routine.
 *	It uses the implicit fpatan instruction of the 287 NPX.
 *
 * Input:	Argument is located at 4(%esp)
 *		For ATAN2, second argument is located at 12(%esp)
 *
 * Output:	Double Precision value is returned in ST(0)
 *
 * Method:
 *		Uses the FPATAN instruction to calculate atan & atan2.
 *		Check for errors where the arguments are zero.
 *
 *		where fpatan = arctan(st(1)/st)
 *		let x = st  & y = st(1)
 *		sign(y)	   sign(x)   |y| < |x|		final result
 *	       ________  __________  ____________      __________________
 *		  +	     +	       yes		 atan(y/x)
 *		  +	     +	       no		 Pi/2-atan(x/y)
 *		  +	     -	       yes		 Pi-atan(y/-x)
 *		  +	     -	       no		 Pi/2+atan(x/y)
 *		  -	     +	       yes		 atan(-y/x)
 *		  -	     +	       no		 -Pi/2+atan(x/-y)
 *		  -	     +	       yes		 -Pi+atan(-y/-x)
 *		  -	     +	       no		 -Pi/2-atan(-x/-y)
 *		
 *
 * Created:
 *	05/28/86	MF - Quantitative Technology Corporation for
 *				Sequent Computer Systems
 *	09/04/86	Phil Hochstetler - converted to C library conventions
 *
 *	05/22/87	Gary Tracy - converted to stack model C interface
 *			  under this model, the caller is responsible for
 *			  saving and restoring any required 387 state.
 *			  this need not be done if:
 *				1.  the callee does no calls.
 *				2.  the sum of 387 registers used by the caller
 *				    and callee does not exceed 8.
 *
 */

#include "PROF.h"

	.set maskC3,0x4000
	.set maskC0,0x0100

ENTRY(atan)
	movw	$0,%dx			#set flag for ATAN
	jmp	do_atan			#go perform atan function
ENTRY(atan2)
	movw	$1,%dx			#set flag for ATAN2
do_atan:
	fldl	4(%esp)			#put argument onto stack
	ftst				#compare arg against zero
	fstsw	%ax			#store status of arg1
	movw	%ax,%cx			# into cx
	fabs				#form ABS(arg1)
	testw	$1,%dx			#ATAN2?
	jne	atan_load_arg2		#if so, load the second argument
	fld1				#put 1.0 onto stack
	movw	$0,%dx			#condition codes for 1.0 are 0
	jmp	atan_check_arg		#go check the argument

atan_load_arg2:
	fldl	12(%esp)		#put second argument onto stack
	ftst				#compare arg against zero
	fstsw	%ax			#store status of arg2
	fabs				#form ABS(arg2)
	movw	%ax,%dx			# into dx

atan_check_arg:
	testw	$maskC3,%edx		#arg2 = 0?
	je	atan_comp_args		#if not, go compare args
	testw	$maskC3,%ecx		#arg1 = 0?
	je	atan_arg2_zero		#if not, go check sign of arg1

/*
 *    Error return is here for both arg1 and arg2 = 0
 */

/* no error, just return +/- PI */

/*
     Arg2 = 0; Check sign of Arg1
 */

atan_arg2_zero:
	fstp	%st(0)			#pop arg2 from stack
	fldl	libm_half_pi		#load PI/2 onto stack
	testw	$maskC0,%ecx		#arg1 > 0?
	je	atan_return		#if so, return PI/2
	fchs				#if not, return -PI/2
	jmp	atan_return		#go return

atan_comp_args:
	fcom				#compare arg1 and arg2
	fstsw	%ax			#get status word
	testw	$maskC3,%ax		#arg1 = arg2?
	jne	atan_ret_piover4	#if so, return PI/4
	testw	$maskC0,%ax		#arg2 > arg1?
	je	atan_no_switch		#everything OK. go perform atan
	fxch				#if not, reverse arguments for divide
atan_no_switch:
	fpatan				#compute result = atan(st(1)/st(0))
	testw	$maskC0,%ax		#were arguments reversed?
	je	atan_no_reduction	#if not, go check signs of args
	fldl	libm_half_pi		#if so, load PI/2
	testw	$maskC0,%ecx		#arg1 < 0?
	jne	atan_negarg1		#if so, go check arg2
	
/*
 *    ARG1 > 0 and arguments were reduced
 */
	testw	$maskC0,%edx		#arg2 < 0?
	jne	atan_posarg1_negarg2	#if so, go form result

/*
 *    ARG1 > 0, ARG2 > 0, and arguments were reduced
 */
	fsub	%st(1),%st(0)		#result = PI/2 - result
	jmp	atan_return		#go return
/*
 *    ARG1 > 0, ARG2 < 0, and arguments were reduced
 */
atan_posarg1_negarg2:
	fadd	%st(1),%st(0)		#result = PI/2 + result
	jmp	atan_return		#go return
/*
 *    ARG1 < 0 and arguments were reduced
 */
atan_negarg1:
	fchs				#form -PI/2
	testw	$maskC0,%edx		#arg2 < 0?
	jne	atan_negarg2		#if so, go form final result
/*
 *    ARG1 < 0, ARG2 > 0, and arguments were reduced
 */
	fadd	%st(1),%st(0)		#result = -PI/2 + result
	jmp	atan_return		#go return
/*
 *    ARG1 < 0, ARG2 < 0, and arguments were reduced
 */
atan_negarg2:
	fsub	%st(1),%st(0)		#result =  -PI/2 - result
	jmp	atan_return		#go return

atan_ret_piover4:
	fstp	%st(0)			#pop both arguments
	fstp	%st(0)			# off the stack
	fldl	libm_quarter_pi		#load PI/4 onto stack

atan_no_reduction:
	testw	$maskC0,%edx		#arg2 < 0?
	jne	atan_load_pi		#if so, go load PI
/*
 *    ARG2 > 0 and arguments were not reduced
 */
	fld	%st(0)			#make a copy of st(0)
	testw	$maskC0,%ecx		#arg1 > 0?
	je	atan_return		#if so, go return
	fchs				#if not, result = -result
	jmp	atan_return		#go return
/*
 *    ARG2 < 0 and arguments were not reduced
*/
atan_load_pi:
	fldpi				#load PI onto stack
	testw	$maskC0,%ecx		#arg1 < 0?
	jne	atan_ch_pi_sign		#if so, go form -PI
/*
 *    ARG2 > 0, ARG1 > 0, and arguments were not reduced
 */
	fsub	%st(1),%st(0)		#if not, result = PI - result
	jmp	atan_return		#go return
/*
 *    ARG2 > 0, ARG1 < 0, and arguments were not reduced
 */
atan_ch_pi_sign:
	fchs				#form -PI
	fadd	%st(1),%st(0)		#result = -PI + result

atan_return:
	fstp	%st(1)		#stack model conversion requires empty stack
	ret
#else

#include "PROF.h"

ENTRY(atan)
	fldl	4(%esp)
	fld1
	fpatan
	ret

ENTRY(atan2)
	fldl	4(%esp)
	fldl	12(%esp)
	fpatan
	ret

#endif
