/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifdef	KXX
/* $Header: exp.s 1.5 87/08/13 $
 *
 *		EXPONENTIAL FUNCTION
 *
 * Functional Description:
 *
 *	This routine implements the exponenetial function for the C language.
 *	It uses the implicit f2xm1 instruction of the 287 NPX. No error checking
 *	is performed since it is assumed that the 287 will correctly handle
 *	an illegal argument.
 *
 * Input:	Argument is located at 4(%esp)
 *
 * Output:	Double Precision value is returned in ST(0)
 *
 * Method:
 *		k = NINT(arg/ln2)
 *		arg = |arg|
 *		IF (arg > 0) 
 *		   exp(arg) = 2 ** (log2(e) * arg - (k * ln2))
 *		ELSE
 *		   exp(arg) = 1/(2 ** (log2(e) * arg - (k * ln2)))
 *		ENDIF
 *		   
 *
 * Created:
 *	05/23/86	MF - Quantitative Technology Corporation for
 *				Sequent Computer Systems
 *	09/04/86	Phil Hochstetler - converted to C library conventions
 *
 *	05/22/87	Gary Tracy - converted to stack model C interface
 *			  under this model, the caller is responsible for
 *			  saving and restoring any required 387 state.
 *			  this need not be done if:
 *				1.  the callee does no calls.
 *				2.  the sum of 387 registers used by the caller
 *				    and callee does not exceed 8.
 *
 *
 */

#include "PROF.h"

	.set	maskC0,0x0100

ENTRY(exp)
	fldln2				#put LN2 onto stack
	fldl	4(%esp)			#put argument onto stack
	subl	$4, %esp		#local storage

	fldl	libm_maxexp		#load in maximum range
	fcomp	%st(1)			#Compare maximum with argument
	fstsw	%ax			#Save the result of the test
	sahf				#Set the flags
	jc	exp_big_err		#if argument > maximum, issue error

	fldl	libm_minexp		#load in minimum range
	fcomp	%st(1)			#Compare minimum with argument
	fstsw	%ax			#Save the result of the test
	sahf				#Set the flags
	jc	exp_cont_calc		#if argument > minimum, continue
	jz	exp_cont_calc		#if argument = minimum, continue
	fstp	%st(0)			#point to st(2)
	fstp	%st(0)			#point to st(3)
	fstp	%st(0)			#point to st(4)
	fldz				#return a 0.0 in st(3)
	jmp	exp_return		#go return

exp_big_err:
	pushl	$ecode
	leal	4+4(%esp), %eax
	pushl	%eax
	fstp	%st(0)
	fstp	%st(0)
	call	_MATHERR
	addl	$4+8, %esp		#clear stack -- local storage
	ret

ecode:	.long	6

exp_cont_calc:
	# Change the precision to 64 bits and rounding to nearest
	fstcw	(%esp)			#get control word
	fwait				#wait until it is memory
	movw	(%esp),%ax		#put it into ax and
	andw	$0xF3FF,%ax		# set rounding mode to nearest
	orw	$0x0300,%ax		# set precision mode to 64 bits
	movw	%ax,2(%esp)		#move new control word to memory and
	fldcw	2(%esp)			# store it back
	fwait				#wait until control word is stored
	
	fst	%st(2)			#save arg in st2
	fdiv	%st(1),%st		#form arg/ln2
	frndint				#form k = NINT(arg/ln2)
	fxch	%st(1)			#switch k and ln2
	fmul	%st(1),%st		#form k * ln2
	fsubr	%st(2),%st		#form x - (k * ln2)
	fldl2e				#load LOG2e
	fmul				#form LOG2e * (x - (k*ln2))
	ftst				#check if value < 0
	fstsw	%ax			#get status word from test
	fabs				#form absolute value
	f2xm1				#form (2 ** (LOG2e * x - (k*ln2))) - 1.0
	fld1				#put 1.0 onto stack
	fxch	%st(1)			#switch 1.0 and result of exp
	fadd	%st(1),%st		#add back in the 1.0 to form Y
	testw	$maskC0,%ax		#was value < 0?
	je	exp_no_invert		#if not, don't invert result
	fdivr	%st(1),%st(0)		#if so, invert result (1.0/result)
exp_no_invert:
	fxch	%st(2)			#switch k and Y
	fxch	%st(1)			#switch 1.0 and k
	fscale				#scale 1.0 by k
	fstp	%st(1)			#pop 2 ** k over k
	fmul				#form result = Y * 2 ** k

exp_return: 
	fldcw	(%esp)			#restore control to original value
	fwait				#wait until store is complete
exp_return_2:
	fstp	%st(1)			#stack model requires empty %st(1)
	addl	$4, %esp		#clear stack -- local storage
	ret

#else

#include "PROF.h"

ONED0:	.long	0, 0x3ff00000		# 1.0d0

ENTRY(exp)
	fldl	4(%esp)			#load argument
	fldl	libm_maxexp		#load in maximum range
	fcomp	%st(1)			#Compare max with arg
	fstsw	%ax			#Save the result of the test
	sahf				#Set the flags
	jc	exp_big_err		#if arg > maximum, error
	fldl2e				# Load log2(e)
	fmulp				# Multiply by value
	fstl	4(%esp)			# Store modified argument
	movl	8(%esp), %eax		# Load modified argument
	andl	$ 0x7fffffff, %eax	# Mask off sign
	cmpl	$ 0x3fe00000, %eax	# Is |X'| > 0.5?
	jg	exp_big			# Yes.  The hard way
	f2xm1				# No. Compute EXP(X)-1.0
	faddl	ONED0			# Now EXP(X)
	ret
exp_big:
	fld	%st(0)			# Duplicate X'
	frndint				# Get nearest integer
	fxch	%st(1)			# Swap st(1) with st(0)
	fsub	%st(1), %st		# Compute -0.5 <= X'' <= 0.5
	f2xm1				# Compute EXP(X'')-1.0
	faddl	ONED0			# Now EXP(X'')
	fscale				# Now EXP(X)
err_return:
	fstp	%st(1)			# pop stack
	ret

exp_big_err:
	pushl	$ecode
	leal	4(%esp), %eax
	pushl	%eax
	fstp	%st(0)
	call	_MATHERR
	addl	$8, %esp
	ret

ecode:	.long	6

#endif
