  # Double Precision Power Function
  #
  # 
  # Functional Description:
  # 	This routine implements the double precision power (BASE ** POWER)
  # 	function, where BASE and POWER are both real numbers. The algorithm
  # 	used was obtained from the original PASCAL library that this replaces.
  # 
  # 	The algorithm for the power function is as follows:
  # 
  # 	IF (base . EQ. 0)
  # 		IF (power .LE. 0)
  # 			return ERROR
  # 		ELSE
  # 			result = 0
  # 		ENDIF
  # 	ELSE
  # 		ipower = trunc(power)
  # 		IF (ipower .NE. power)
  # 			IF (base .GT. 0)
  # 				result = exp(ln(base) * power)
  # 			ELSE
  # 				return ERROR
  # 			ENDIF
  # 		ELSE
  # 			result = __powi (base, ipower)
  # 		ENDIF
  # 	ENDIF
  # 
  # Created:
  # 	11/25/86 	BS - Quantitative Technology Corporation for
  # 				Sequent Computer Systems
  #
  #
	.data
  #
  # DOMERR	11			 error code for illegal value
  # OVFERR	12			 error code for overflow
  #
  #ipower	edx			 integer value of power 
  #base	fp2			 holds base 
  #power	fp4			 holds power 
  #tempf	fp6			 temporary register 
  #
one:	.double	0Dx3ff0000000000000	#fp one
inf:	.double 0Dx7ff0000000000000	#fp infinity
  #ctxtrnc 0x013a0000			#context register bits for truncate
  #ctxrnd  0x073a0000			#context register bits for rounding
  #
  #
#include "PROF.h"

ENTRY(pow)
	wloadl	4(%esp),%fp2		#get base
	wloadl	12(%esp),%fp4		#get power
	wtstl	%fp2			#test base = 0
	wstctx  %eax			#transfer contents of control reg
	sahf				#receive to flags
	jne	chk_pwr			#if not, go check value of power
	wtstl	%fp4			#test power > 0
	wstctx  %eax			#transfer contents of control reg
	sahf				#receive to flags
	ja	return			#if so, return a value of 0 (jlt)

	pushl	$ecode_one
	leal	4(%esp), %eax
	pushl	%eax
	call	_MATHERR
	addl	$8, %esp
	ret

ecode_one:	.long 12

  #
  # base <> 0
  #

chk_pwr:
	pushl	%ebx			#preserve non-scratch register
	wstctx	%ebx			#save old value of context work
	movl	$0x073a0000,%ecx	#truncate context word
	wldctx	%ecx			#place 1167 in truncate mode
	wfixl	%fp4,%fp6		#get INT(power)
	wstors	%fp6,%edx		#
	wfloatl	%fp6,%fp6		#
	wcmpl	%fp4,%fp6		#is power an integer value?
	wstctx  %eax			#transfer contents of control reg
	wldctx	%ebx			#restore rounding mode
	popl	%ebx			#restore contents of %ebx
	sahf				#receive to flags
	je	up_int			#power is an integer, go compute value
	wtstl	%fp2			# base > 0?
	wstctx  %eax			#transfer contents of control reg
	sahf				#receive to flags
	jae	ln_exp			#if so, go compute via ln and exp(jge)

  ###############################################################
  # Error return for base < 0 and power is not an integer value #
  ###############################################################

	pushl	$ecode_two
	leal	4(%esp), %eax
	pushl	%eax
	call	_MATHERR
	addl	$8, %esp
	ret

ecode_two:	.long 11

  #
  # base > 0 and power is not an integer value 
  #
ln_exp:
	subl	$8,%esp
	wstorl	%fp2,(%esp)		# get value of base
	call	_log			# compute ln(base)
	wmull	20(%esp),%fp2		# compute ln(base) * power
	wstorl	%fp2,(%esp)		# get ln(base) * power
	call	_exp			# compute exp(ln(base) * power)
	addl	$8,%esp
	ret

  #
  # power is an integer value	
  #

up_int:
	wloadl	%fp2,%fp4		#move base

  # Double Precision Integer Power Function
  #
  # 
  # Functional Description:
  # 	This routine implements the double precision integer power
  # 	(BASE ** IPOWER) function, where BASE is real and IPOWER is integer.
  # 	The algorithm used was obtained from the original PASCAL library that
  # 	this replaces. (NOTE: This routine is called by the flt. pt. power
  # 	function to compute the result when the BASE is less than 0 or when
  # 	the power is an integer value.)
  # 
  # 	The algorithm for the integer power function is as follows:
  # 
  # 	result = 1
  # 	invert = ipower .LT. 0
  # 	IF (ipower .NE. 0)
  # 		ipower = abs(ipower)
  # 		IF (odd(ipower))
  # 			result = base
  # 		ENDIF
  # 		bit = 2
  # 		DO WHILE (ipower .GE. bit)
  # 			base = base * base
  # 			IF ((ipower .AND. bit) .NE. 0)
  # 				result = result * base
  # 			ENDIF
  # 			bit = bit + bit
  # 		ENDDO
  # 		IF (invert)
  # 			result = 1.0 / result
  # 		ENDIF
  # 	ENDIF
  # 
  # Created:
  # 	11/25/86 	BS - Quantitative Technology Corporation for
  # 				Sequent Computer Systems
  #
  # ipower	edx	 integer value of power 
  # invert	al	 flag for inverting(negative power) 
  # bitval	ecx	 integer value of bit mask 
  # result	fp2	 holds result 
  # base		fp4	 holds base 
  #
	wloadl	one,%fp2		#init result to 1.0
	cmpl	$0,%edx			#power < 0?
	setg	%al			#if so, set flag
	jz	return			#if power = 0, return a 1.0
	jg	nabs			#if positive don't perform absolute val
	negl	%edx			#ipower = abs(ipower)
nabs:
	testb	$0x01,%edx		#is ipower even?
	jz	cont			#if so, continue
	wloadl	%fp4,%fp2		#if not, init result to base
cont:
	movl	$2,%ecx			#initialize bit tester value
loopf:
	cmpl	%ecx,%edx		#power < bitval?
	jb	chkinv			#if so, thru. Go check on invert
	wmull	%fp4,%fp4		#base = base * base
	testl	%ecx,%edx		#bit set in power?
	jz	cont1			#if not, don't update result
	wmull	%fp4,%fp2		#if so, update result
cont1:
	addl	%ecx,%ecx		#update bit tester value
	jmp	loopf			#continue with computation
					#
chkinv:
	cmpb	$0,%al			#was power < 0?
	jnz	return			#is not, go return
	wdivl	one,%fp2		#result = 1.0/result

  #
  # finish and clean up stack
  #

return:
	wloadl	inf,%fp4		#check for infinity return (overflow)
	wcmpl	%fp4,%fp2
	wstctx  %eax			#transfer contents of control reg
	sahf				#receive to flags
	je	overflow
	ret

overflow:
	pushl	$ecode_three
	leal	4(%esp), %eax
	pushl	%eax
	call	_MATHERR
	addl	$8, %esp
	ret

	.align 2
ecode_three:	.long 6
