/*LINTLIBRARY*/

/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: microtask.c 1.26 87/09/24 $";
#endif

/*
 * microtask.c
 *	Simple microtasking support for shared-memory C and Fortran programs.
 *
 * TODO:
 *	Profile execution and optimize.
 *	Handle reentrant call to m_fork().
 *	look over protocol for modifying _single_stack and mystate variables
 *		(these should be shared vars, modified by MASTER only?)
 *	deal with SIGSTOP, SIGTTIN, etc in a nice way
 *	do a binary tree fork, for speed
 *
 * Assumes shared-memory environment is already set up.
 */

/* $Log:	microtask.c,v $
 */

#include <signal.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <parallel/parallel.h>
#include <parallel/microtask.h>
#include "parc.h"

/*
 * Use this variable to tell ppinit to set up a shared parent stack
 */
private int _do_shared_stack = 1;
private static int m_nested = 0;		/* m_fork nesting level */

/*
 * Internal variables.
 */

shared	static	long	*argv;			/* argument dispatching */
shared	static	bool_t	active;			/* active multi in progress */
shared	static	void	(*exec_func)();		/* function to execute */
shared	static	int	argbc;			/* argument byte-count */

shared	static	slock_t	lock_mutex;		/* m_lock(), m_unlock */
shared	static	long	Snext_val;		/* next counter value */

shared	static	slock_t	next_mutex;
private	static	long	*next_val = &Snext_val;

#define	EMPTY	0
shared	int	_procid[MAXPROCS];	/* pids */
shared	int	m_numprocs = -1;		/* total # processes working */
private		int	m_myid;			/* procid[] idx for each proc */

shared	static	sbarrier_t	syncb;		/* for m_sync() and m_join() */

shared	static	sbarrier_t	singleb;	/* for m_single() */
shared	static	unsigned long	multi_ret;	/* return address */

static int child_handler();
static int sig_handler();

static  private struct sigvec DFL_vec = {SIG_DFL, 0, 0};
static  private struct sigvec CHLD_vec = {child_handler, 0, 1};
/*
 * Want to make sure that cleanup procedures for a signal don't cause
 * SIGCHLD to be invoked.
 */
static  private struct sigvec HAND_vec = {sig_handler, sigmask(SIGCHLD), 1};
static private int (*sig_array[NSIG + 1])();
static private int first_dump;

/*
 * process state information
 */

private static int mystate;

#define S_SINGLE	0
#define S_WAIT		1
#define S_PARKED	2
#define S_FORKED	3
#define S_MSINGLE	4

#ifdef PROF
/*
 * pointer to monitor name file.
 */
extern char *_mon_file;
static  char mon_name[20];
#endif PROF

/*
 * m_set_procs()
 *	Start processes, making them ready to help with the computation.
 *
 * Forks kids, remembers pid's.
 *
 * Returns 0 for all ok, else -1 (and errno).
 */

m_set_procs(nprocs)				/* C interface */
	register int	nprocs;
{
	register int maxprocs;

	maxprocs = cpus_online() - 1;

	if ((mystate != S_SINGLE) || (nprocs > MAXPROCS) || nprocs > maxprocs) {
		errno = EINVAL;
		return(-1);
	}


	m_numprocs = (nprocs > 0)?nprocs:get_default();
	return(0);

}

#ifndef lint
static
#endif 
m_fset_procs(nprocsp)				/* FORTRAN interface */
	int	*nprocsp;
{
	return(m_set_procs(*nprocsp));
}

/*
 * Initialization routine called from m_fork(), when there are not
 * process waiting to run in parallel.
 */

/*
 * THESE MUST be a private, non-automatic variable
 */
static private int mfork_pid;
static private int save_sp;
static private int mfork_i;

_m_fork_init()
{
	int	err = errno;
	extern int *_priv_sp;

	if (mystate == S_SINGLE) {

		if (m_numprocs == -1) {
			m_numprocs = get_default();
		}
		S_INIT_BARRIER(&syncb, m_numprocs);
		S_INIT_BARRIER(&singleb, m_numprocs);
	
		/*
		 * Check nprocs <= MAXPROCS?  Check not reentered?
		 */
	
		(void) sigvec(SIGCHLD, &CHLD_vec, (struct sigvec *)0);
		for (mfork_i = 1; mfork_i < m_numprocs; mfork_i++) {
#ifdef lint
			save_sp = 0;
#endif lint
#ifdef ns32000
			;asm("	sprd	sp, _save_sp");
			;asm("	lprd	sp, __priv_sp");
#endif 
#ifdef i386
			;asm("	movl	%esp, _save_sp");
			;asm("	movl	__priv_sp, %esp");
#endif i386
			mfork_pid = fork();
			if (mfork_pid < 0) {
#ifdef ns32000
				;asm("	lprd	sp, _save_sp");
#endif ns32000
#ifdef i386
				;asm("	movl	_save_sp, %esp");
#endif i386
				(void) sigvec(SIGCHLD, &DFL_vec, (struct sigvec *)0);
				(void) sigvec(SIGCHLD, &DFL_vec, (struct sigvec *)0);
				while (--mfork_i > 1) {
					(void) kill(_procid[mfork_i], SIGKILL);
					/* could wait for processes */
				}
				errno = err;
				return(-1);
			}
	
			if (mfork_pid == 0) {
				/*
				 * Child.
				 */
				m_myid = mfork_i;
				mystate = S_WAIT;
#ifdef PROF
				(void) sprintf(mon_name, "%s.%d", _mon_file, m_myid);
				_mon_file = mon_name;
#endif PROF
				m_worker();		/* do this forever */
				/*NOTREACHED*/
			}
	
			/*
			 * Parent.  Record kids pid.
			 */
#ifdef ns32000
			;asm("	lprd	sp, _save_sp");
#endif ns32000
#ifdef i386
			;asm("	movl	_save_sp, %esp");
#endif i386
			_procid[mfork_i] = mfork_pid;
			/* maybe SIGSTOP child? */
		}
	
		/*
		 * Only parent gets here.
		 */
	
		init_sigvec();
		m_myid = MASTER;
		_procid[0] = getpid();
		mystate = S_WAIT;
	
	}

	/*
	 * Handle reentrancy -- if already active, have nested m_fork() call.
	 * Note that if !active, caller must be master.
	 */

	/*
	 * If they're parked, let'em go.
	 */

	if (mystate == S_PARKED)
		(void)m_rele_procs();

	if (mystate != S_WAIT) {
		errno = EINVAL;
		return(-1);
	}
	return(0);
}

/*
 * m_fork(func [,args]), m_ffork(args)
 *	Start kids executing "func" with given args.
 *
 * Need C and Fortran interface since args passed in different order.
 */

/*VARARGS*/
m_fork(func, args)				/* C interface */
	void	(*func)();
	int	args;
{
	do {
		if (mystate == S_WAIT)
			return(m_startwork(func, (long *) &args, nargs()-1));
		if ((mystate == S_FORKED) || (mystate == S_MSINGLE))
			return(m_nestwork(func, (long *) &args, nargs()-1));
			
	} while (_m_fork_init() == 0);
	return(-1);
}

/*VARARGS*/
#ifndef lint
static
#endif 
m_ffork(args)					/* FORTRAN interface */
	int	args;
{
	int	argc = nargs() - 1;

	do {
		if (mystate == S_WAIT)
			return(m_startwork(*((void (**)()) (&args + argc)),
			 (long *) &args, argc));
		if ((mystate == S_FORKED) || (mystate == S_MSINGLE))
			return(m_nestwork(*((void (**)()) (&args + argc)),
			 (long *) &args, argc));
	} while (_m_fork_init() == 0);
	return(-1);
}

/*
 * _m_fork_(link, func [,args])		V2.9.1 Fortran
 *
 * _m_pfork(link, func [,args])	 	V2.9.1 Pascal, via ENTRY macro
 *
 *	Start kids executing "func" with given args.
 *
 * This is the V2.9.1 compatible calling interface for Fortran and Pascal.
 * flink is the static link for the function argument. In Fortran, this link
 * is always null.  In Pascal, it may or may not be null.
 *
 * The new Fortran naming convention "_m_fork_" is taken care of by
 * naming this "m_fork_" in C.  
 *
 *   Stack on ENTRY		stack for calling func
 *
 *	argn				argn
 *	...				...
 *	arg0				arg0
 *	func				{static link}
 *	flink				ret_pc
 *	ret_pc	
 *
 * if flink is non-null, the {static link} is a copy of flink.  Otherwise,
 * {static link} is absent from the calling stack entirely.
 *
 */

/*
 * F interface (_m_fork_) and P interface (_m_pfork) for V2.9.1
 */

/*VARARGS*/
m_fork_(flink, func, args)				
	long	flink;
	void	(*func)();
	long	args;
{
	void	(*Sfunc)() = func;
	long	*aargs = &args;
	int argc = nargs() - 2;

	if (argc < 0)
		return(-1);

	if (flink != 0) {
		aargs--;
		argc++;
		*aargs = flink;	/* copy flink over func argument */
	}
			
	do {
		if (mystate == S_WAIT)
			return(m_startwork(Sfunc, aargs, argc));
		if ((mystate == S_FORKED) || (mystate == S_MSINGLE))
			return(m_nestwork(Sfunc, aargs, argc));
			
	} while (_m_fork_init() == 0);
	return(-1);
}

/*
 * m_pfork(args)		-e Pascal, mapped to _m_pfork1 by ENTRY macro
 *
 *	Start kids executing "func" with given args.
 *	
 * This supports the -e flag calling convtion for Pascal
 *
 *	stack on entry			stack for called fntn
 *	 func					arg0
 *	 flink					...
 *	 arg0					argn
 *	 ...					{static link}
 *	 argn					ret-pc
 *	 ret-pc
 *
 * if flink is null, then {static link} is absent from called fntn stack.
 *
 */

/*VARARGS*/
#ifndef lint
static
#endif 
m_pfork1(args)					/* PASCAL -e interface */
	long	args;
{
	register long *aargs = &args;
	int	argc = nargs() - 2; /* knock off func, flink */
	long	save_pc;	/* place to save pc */
	long 	tmp;

		
#define FUNC	(void (*)())aargs[argc + 1]	/* function address */
#define	FLINK	aargs[argc]			/* function's static link */
#define	ARGN	aargs[0]			/* the n'th argument */
#define RET_PC	aargs[-1]			/* saved return-pc */

	/*
	 * there must be at least a function ptr, AND static link.
	 */
	if (argc < 0)
		return(-1);

	/*
	 * We need the save_pc as a temporary.
	 * 	This is BAD.  It'll do strange things to stack traces!!
	 */

	do {
		if (FLINK != 0) {
			if (mystate == S_WAIT) {
				/*
				 * COPY the static link to the "right place"
				 *	(over-write the return pc)
				 * pass static link to FUNC
				 */
				save_pc = RET_PC;
				RET_PC = FLINK;
				tmp = m_startwork( FUNC, &RET_PC, argc + 1);
				RET_PC = save_pc;	/* restore pc */
				return(tmp);
			}
			if ((mystate == S_FORKED) || (mystate == S_MSINGLE)) {
				/*
				 * COPY the static link to the "right place"
				 *	(over-write the return pc)
				 * pass static link to FUNC
				 */
				tmp = m_nestwork( FUNC, &RET_PC, argc + 1);
				RET_PC = save_pc;	/* restore pc */
				return(tmp);
			}
		} else {
			/*
			 * discard the static link.
			 */
			if (mystate == S_WAIT)
				return(m_startwork( FUNC, &ARGN, argc));
			if ((mystate == S_FORKED) || (mystate == S_MSINGLE))
				return(m_startwork( FUNC, &ARGN, argc));
		}
	} while (_m_fork_init() == 0);
	return(-1);

#undef FUNC
#undef	FLINK
#undef	ARGN
#undef RET_PC
#undef SLINK
}

/*
 * m_single()
 *	Only single thread of code returns until m_multi() is called.
 *
 * Barrier code cloned from s_wait_barrier().
 *
 * RET_PC() accesses PC given 1st argument to procedure; VERY MACHINE DEPENDENT!
 */

#define	RET_PC(firstarg)	*(((unsigned long *) &(firstarg)) - 1)

m_single(x)
	int	x;
{
	register sbarrier_t *b = &singleb;
	unsigned char	ent_use = b->b_useno;

	if (mystate != S_FORKED) {
		errno = EINVAL;
		return(-1);
	}
	if (m_nested)
		return(0);

	/*
	 * Check self "in".  If last to check-in, return -- only one thread.
	 */

	S_LOCK(&b->b_mutex);

#ifdef ALM
	++b->b_count;
#else
	--b->b_count;
#endif ALM

	S_UNLOCK(&b->b_mutex);

	/*
	 * Wait for m_multi() to be called, and return to
	 * where m_multi() caller returns to.
	 */

	while (ent_use == b->b_useno) {
#ifdef ALM
		if ((m_myid == MASTER) && (b->b_count == b->b_limit)) {
#else
		if ((m_myid == MASTER) && (b->b_count == 0)) {
#endif ALM
			mystate = S_MSINGLE;
			return(0);
		}
		continue;
	}

	RET_PC(x) = multi_ret;			/* alter return PC */
	return(0);
}

m_multi(x)
	int	x;
{
	/*
	 * Check in "single" state?
	 */

	if (mystate != S_MSINGLE) {
		errno = EINVAL;
		return(-1);
	}

	*next_val = 0;				/* re-init trip counter */

	if (m_nested) 
		return(0);

	/*
	 * Set up new return address (all threads appear to return from
	 * m_multi()), clean up barrier, and release waiters.
	 */

	multi_ret = RET_PC(x);			/* return PC */
#ifdef ALM
	singleb.b_count = 0;			/* ready for next time */
#else
	singleb.b_count = singleb.b_limit;
#endif
	++singleb.b_useno;			/* let'em go! */
	mystate = S_FORKED;
	return(0);
}

/*
 * m_park_procs()
 *	Park slave processes (kids) in non-spin-loop wait.
 *
 * Used by application when it knows there are long periods of non-use
 * of slaves.
 */

m_park_procs()
{
	register int i;

	/*
	 * Only allow park if not currently active.
	 * Note that if !active, only master thread is calling.
	 */

	if (mystate == S_SINGLE)
		return(0);

	if (mystate != S_WAIT) {
		errno = EINVAL;
		return(-1);
	}

	/*
	 * Stop worker processes to get them off processors.
	 */

	for (i = 1; i < m_numprocs; i++) {
		(void) kill(_procid[i], SIGSTOP);
	}
	mystate = S_PARKED;
	return(0);
}

/*
 * m_rele_procs()
 *	Relese processes from "parked" state -- ready to compute again.
 */

m_rele_procs()
{
	register int i;

	if (mystate == S_SINGLE)
		return(0);

	if (mystate != S_PARKED) {
		errno = EINVAL;
		return(-1);
	}
	for (i = 1; i < m_numprocs; i++)
		(void) kill(_procid[i], SIGCONT);
	mystate = S_WAIT;
	return(0);
}

/*
 * m_sync()
 *	Barrier for participating processes.  Also clears next_val.
 */

m_sync()
{
	/* Check active, parked, killed? */
	if (mystate != S_FORKED) {
		errno = EINVAL;
		return(-1);
	}

	if (m_nested == 0)
		_M_JOIN(&syncb, next_val);
	else
		*next_val = 0;
	return(0);
}

/*
 * m_lock(), m_unlock()
 *	Single spin-lock, simple interface.
 *
 * For better performance, application can have multiple locks, use
 * S_LOCK(), S_UNLOCK(), etc.
 */

m_lock()
{
	register slock_t *l = &lock_mutex;		/* -O -i has holes */

	if (m_nested == 0)
		S_LOCK(l);
}

m_unlock()
{
	if (m_nested == 0)
		S_UNLOCK(&lock_mutex);
}

/*
 * m_next()
 *	Returns next value of counter, atomically.
 *
 * Can be used for self-scheduled DO-loops.
 */

m_next()
{
	register slock_t *l = &next_mutex;		/* -O -i has holes */
	int	val;

	if (m_nested == 0) {
		S_LOCK(l);

		*next_val += 1;
		val = *next_val;		/* Ahhh, for fetch-and-add... */

		S_UNLOCK(&next_mutex);
	} else {
		*next_val += 1;
		val = *next_val;
	}

	return(val);
}


/*
 * m_kill_procs()
 *	Release slave processes to system, no longer needed.
 *
 * Also, wait for them to terminate (note this could get other non-worker
 * children).
 */

static shared long zero_word = 0;

m_kill_procs()
{
	register int togo = m_numprocs-1;
	register int i;
	register int pid;
	void	exit();

	/*
	 * If slaves are "parked", let'em wakeup so they can die.
	 */

	if (mystate == S_SINGLE)
		return(0);

	if (mystate == S_PARKED)
		(void)m_rele_procs();

	/*
	 * Only allow kill if not currently active.
	 * Note that if !active, only master thread is calling.
	 */

	if (mystate != S_WAIT) {
		errno = EINVAL;
		return(-1);
	}

	/*
	 * Sneaky -- fool them into calling exit!
	 */

	restore_sigvec();
	(void) sigvec(SIGCHLD, &DFL_vec, (struct sigvec *)0);
	exec_func = exit;			/* arrange exit(0) */
	argv = &zero_word;
	argbc = 1 * sizeof(long);
	active = 1;				/* off they go to death! */

	/*
	 * Wait for the processes to exit, to collect their run-time
	 * statistics.
	 */

	mystate = S_SINGLE;
	while (togo != 0) {
		pid = wait((union wait *)0);
		for (i = 1; i < m_numprocs; i++) {
			if (pid == _procid[i]) {
				--togo;
				_procid[i] = EMPTY;
				break;
			}
		}
	}
	active = 0;
	return(0);
}

/*
 * Internal procedures.
 */

/*
 * m_startwork()
 *	Common code (C, Fortran) to start workers in given function
 *	with given arguments.
 */

static
m_startwork(func, argp, argc)
	void	(*func)();
	long	*argp;
	int	argc;
{

	/*
	 * Set up func pointer, then args.
	 */

	exec_func = func;
	argbc = argc * sizeof(long);
	argv = argp;

	/*
	 * Release eager worker processes and join in the fun!
	 *
	 * Note: if callee cleared stack, could fuss it and just "return"
	 * to func; but can't since caller pops all args, including "func".
	 */

	Snext_val = 0;				/* clear for next stuff */
	active = 1;				/* GO! */
#ifdef ns32000
	;asm("	sprd	sp, _save_sp");
	;asm("	lprd	sp, __priv_sp");
#endif ns32000
#ifdef i386
	;asm("	movl	%esp, _save_sp");
	;asm("	movl	__priv_sp, %esp");
#endif i386
	m_work();
#ifdef ns32000
	;asm("	lprd	sp, _save_sp");
#endif ns32000
#ifdef i386
	;asm("	movl	_save_sp, %esp");
#endif i386
	return(0);
}

/*
 * m_nestwork()
 *	a process has nested m_fork call.  Arrange an environment
 *	like for a real m_fork, but dispatch it only to the current
 * 	process.  Likewise, set up so m_synch, m_lock, etc. behave
 *	properly.
 *
 * MACHINE DEPENDENT!!  Relies on compiler allocating myargs[]
 * as lowest address auto.  Also, can't have register variables in this
 * procedure since they come after locals.
 */
static
m_nestwork(func, argp, argc)
	void	(*func)();
	long	*argp;
	int	argc;
{
	long	Nnext_val = 0;
	long	*old_next_val;
	long	myargv[MAXARGS];
	extern int _single_stack;

	/*
	 * If there were args, copy into place in local args.
	 */

	m_nested++;
	old_next_val = next_val;
	next_val = &Nnext_val;

	if (argc > 0)
		bcopy((char *) argp, (char *) myargv, 
					(unsigned)argc*sizeof(long));

	/* 
	 * Call the appropriate procedure.  Correctness assumes
	 * myargv[] appears as "args" to called procedure.
	 */

	(*func)();
	--m_nested;
	next_val = old_next_val;
	return(0);
}

/*
 * m_worker()
 *	Kids execute here, waiting for work to do, then do it.
 *
 * There is no return from here.
 */

static
m_worker()
{
	register bool_t	*actp = &active;

	for(;;) {
		/*
		 * Wait for work (master to call m_fork()).
		 */

		while(!*actp)			/* volitile would be nice... */
			continue;

		/*
		 * Dispatch the work.
		 */

		m_work();
	}
}

/*
 * m_work()
 *	Called to dispatch micro-task "process".  Put args
 *	on stack and call function.
 *
 * MACHINE DEPENDENT!!  Relies on compiler allocating myargs[]
 * as lowest address auto.  Also, can't have register variables in this
 * procedure since they come after locals.
 */

static
m_work()
{
	long	myargv[MAXARGS];
	extern int _single_stack;

	/*
	 * If there were args, copy into place in local args.
	 */

	mystate = S_FORKED;
	_single_stack = 0;

	if (argbc > 0)
		bcopy((char *) argv, (char *) myargv, (unsigned)argbc);

	/* 
	 * Call the appropriate procedure.  Correctness assumes
	 * myargv[] appears as "args" to called procedure.
	 */

	(*exec_func)();

	/*
	 * Sync with others.  Last caller of m_join() clears
	 * active (atomically).
	 */

	_M_JOIN(&syncb, &active);
	_single_stack = 1;
	mystate = S_WAIT;
}

#ifdef ALM
/*
 * m_join()
 *	Barrier with atomic clear of argument.
 *
 * Barrier code cloned from s_wait_barrier().
 */

static
_m_join(b, clearit)
	register sbarrier_t *b;
	bool_t	*clearit;
{
	unsigned char	ent_use = b->b_useno;
	/*
	 * Check self "in".  If last to check-in, clear *clearit,
	 * set up for next usage and release others.
	 * Bumping b_useno "opens" the barrier.
	 */
	S_LOCK(&b->b_mutex);
	if (++b->b_count == b->b_limit) {
		b->b_count = 0;
		*clearit = 0;				/* "atomic" clear */
		++b->b_useno;
	}
	S_UNLOCK(&b->b_mutex);

	/*
	 * Wait for last process to check-in.
	 */

	while (ent_use == b->b_useno)
		continue;
}
#endif ALM

static
get_default()
{
	return((cpus_online()+1)/2);
}

m_get_myid()
{
	if (m_nested)
		return(0);
	return(m_myid);
}

m_get_numprocs()
{
	if (m_nested)
		return(1);
	if (m_numprocs == -1) {
		m_numprocs = get_default();
	}
	return(m_numprocs);
}

/*
 * called only by MASTER from m_fork()
 * Make sure that signals declared SIG_DFL point to the microtasker's handler.
 */
static 
init_sigvec()
{
	int sig, oldmask;
	struct sigvec oldvec;

	oldmask = sigblock(0xffffffff);
	for (sig = 1; sig <= NSIG; sig++) {
		switch (sig) {
		case	SIGSTOP:
		case	SIGTSTP:
		case	SIGTTIN:
		case	SIGTTOU:
		case	SIGURG:
		case	SIGCONT:
		case	SIGIO:
		case	SIGCHLD:
		case	SIGWINCH:
			continue;

		default:
			if (sig_array[sig] != SIG_DFL)
				continue;
			(void) _sigvec(sig, &HAND_vec, &oldvec);
			if (oldvec.sv_handler == SIG_IGN)
				sig_array[sig] = SIG_IGN;
		}
	}
	sigsetmask(oldmask);
}

/*
 * Called only by MASTER from m_kill_procs.
 * Return unused signals to the SIG_DFL state.
 */
static 
restore_sigvec()
{
	int sig;

	for (sig = 1; sig <= NSIG; sig++) {
		switch (sig) {
		case	SIGSTOP:
		case	SIGTSTP:
		case	SIGTTIN:
		case	SIGTTOU:
		case	SIGURG:
		case	SIGCONT:
		case	SIGIO:
		case	SIGCHLD:
		case	SIGWINCH:
			continue;

		default:
			if (sig_array[sig] != SIG_DFL)
				continue;
			(void) _sigvec(sig, &DFL_vec, (struct sigvec *)0);
		}
	}
}


/*
 * for MASTER only, make sure that all SIG_DFL's point at microtasker's 
 * handler.
 */

static int sig_handler();

int
_lib_sigvec(sig, vec, ovec)
int sig;
struct sigvec *vec, *ovec;
{
	register int ret_val;

	switch (sig) {
	case	SIGSTOP:
	case	SIGTSTP:
	case	SIGTTIN:
	case	SIGTTOU:
	case	SIGURG:
	case	SIGCONT:
	case	SIGIO:
	case	SIGCHLD:
	case	SIGWINCH:
		ret_val = _sigvec(sig, vec, ovec);
		break;

	default:
		ret_val = _sigvec(sig, vec, ovec);
		if ((m_myid == MASTER) && (ret_val == 0)) {
			sig_array[sig] = vec->sv_handler;
			if (sig_array[sig] == SIG_DFL)
				(void) _sigvec(sig, &HAND_vec, (struct sigvec *)0);
		}
	}
	return(ret_val);
}

/*
 * All SIG_DFL handlers for MASTER(only) reach here
 */
/*ARGSUSED*/
static
sig_handler(sig, code, scp)
int sig, code;
struct sigcontext *scp;
{

	if (sig_array[sig] == SIG_IGN)
		return;
	(void) _sigvec(SIGCHLD, &DFL_vec, (struct sigvec *)0);
	_clean_house(sig);

	/*
	 * all the kids are dead and gone.  
	 * Time to blow myself away.
	 */

	(void) _sigvec(sig, &DFL_vec, (struct sigvec *)0);
	(void) kill(getpid(), sig);
	/* Not reached */
	return;
}

_clean_house(sig)
int sig;
{
	register int i;
	int pid, tpid;
	union wait status;

	if ((mystate == S_WAIT) || (mystate == S_PARKED)) {
		(void)m_kill_procs();
		return;
	}

	for (i = 1; i < m_numprocs; i++) {
		if (_procid[i] == EMPTY) 
			continue;

		/*
		 * gotta make sure he's dead.  
		 *
		 * For example, if this is a SIGSEGV, and there
		 * just happened to be an unmapped part of the shared
		 * data region, the shsbrk() code will eat the first signal.
		 * In fact, it'll keep eating them until a signal gets
		 * in and there's no growth to do.  Conceivable, this race
		 * could go on forever.  But, that's unlikely.
		 */
		while (kill(_procid[i], sig) == 0) sleep(1);
		/*
		 * clean up any dead kids.  
		 *
		 * Try to rename core files if possible.  Also deal with 
		 * the possibility that other kids may have died spontaneously.
		 * This, of course makes the core file renaming a little 
		 * unreliable.
		 */
	
		pid = _procid[i];
		do {
			tpid = wait(&status);
			(void) clean_kid(i, tpid, status.w_coredump);
		} while (tpid != pid);
	}
}


static
clean_kid(i, tpid, coredump)
register int i, tpid; 
unsigned short coredump;
{
	register int j;
	char buf[16];

	for (j = i; j < m_numprocs; j++) {
		if (_procid[j] != tpid)
			continue;
		_procid[j] = EMPTY;
		if (coredump) {
			if (first_dump) 
				(void) sprintf(buf, "core.%d.First", j);
			 else
				(void) sprintf(buf, "core.%d", j);
			/*
			 * I really don't care if this fails.
			 */
			(void) rename("core", buf);
		}
		first_dump = 0;
		return(0);
	}
	return(-1);
}

static
child_handler()
{
	register int tpid, sig;
	union wait status;
	
	
	if (m_myid != MASTER)
		return;


	tpid = wait3(&status, WNOHANG|WUNTRACED, (struct rusage *)0);
	if (tpid <= 0) 
		return;
	if (status.w_stopval == WSTOPPED) {
		return;
	}

	first_dump = 1;
	if (clean_kid(1, tpid, status.w_coredump))
		return;

	sig = status.w_termsig;
	if (sig == 0) {
		_clean_house(9);
		exit((int)status.w_retcode);
	} else {
		_clean_house(sig);
		(void) _sigvec(SIGCHLD, &DFL_vec, (struct sigvec *)0);
		(void) _sigvec(sig, &DFL_vec, (struct sigvec *)0);
		(void) kill(getpid(), sig);
	}
}

_lib_exit(status)
int status;
{
	
	if ((m_myid != MASTER) || (mystate == S_SINGLE)) 
		__exit(status);

	(void) _sigvec(SIGCHLD, &DFL_vec, (struct sigvec *)0);
	_clean_house(SIGKILL);
	__exit(status);
}

/* Generate Fortran and Pascal -e compatability entry points */
ENTRY(m_set_procs, _m_fset_procs);	
ENTRY(m_fork, _m_ffork);
ENTRY(m_pfork, _m_pfork1);
ENTRY(m_single, _m_single);
ENTRY(m_multi, _m_multi);
ENTRY(m_park_procs, _m_park_procs);
ENTRY(m_rele_procs, _m_rele_procs);
ENTRY(m_sync, _m_sync);
ENTRY(m_lock, _m_lock);
ENTRY(m_unlock, _m_unlock);
ENTRY(m_next, _m_next);
ENTRY(m_kill_procs, _m_kill_procs);
ENTRY(m_get_myid, _m_get_myid);
ENTRY(m_get_numprocs, _m_get_numprocs);

/*
 * Fortran and Pascal V2.9.1 interfaces 
 * 	Fortran _m_fork_ function is defined entirely in C
 */
ENTRY(_m_set_procs_, _m_fset_procs);
ENTRY(_m_pfork, _m_fork_);		/* Pascal */
ENTRY(_m_single_, _m_single);
ENTRY(_m_multi_, _m_multi);
ENTRY(_m_park_procs_, _m_park_procs);
ENTRY(_m_rele_procs_, _m_rele_procs);
ENTRY(_m_sync_, _m_sync);
ENTRY(_m_lock_, _m_lock);
ENTRY(_m_unlock_, _m_unlock);
ENTRY(_m_next_, _m_next);
ENTRY(_m_kill_procs_, _m_kill_procs);
ENTRY(_m_get_myid_, _m_get_myid);
ENTRY(_m_get_numprocs_, _m_get_numprocs);
