/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* $Header: parallel.h 1.15 87/11/03 $
 *
 * parallel.h
 *	Definitions for use in parallel, shared-memory programs.
 */

/* $Log:	parallel.h,v $
 */

/*
 * A "lock" is a byte of memory, initialized to zero (unlocked).
 */

typedef unsigned char	slock_t;		/* 's' for "spin"-lock */

#define	L_UNLOCKED	0
#define	L_LOCKED	1

/*
 * Was a conditional lock request granted (L_SUCCESS) or denied (L_FAILED)
 */
#define L_FAILED	0
#define L_SUCCESS	1

/*
 * A "barrier" allows multiple processes to synchronize by having
 * all of them exit the barrier construct simultaneously.
 *
 * This version assumes <= 255 processes, fits in one 4-byte integer,
 * and is based on spin-locks.
 */

typedef struct	{
	slock_t		b_mutex;	/* mutual exclusion */
	unsigned char	b_limit;	/* number participants */
	unsigned char	b_count;	/* state counter */
	unsigned char	b_useno;	/* current use # (state flag) */
} sbarrier_t;				/* 's' for "spin"-barrier */

/*
 * Other useful declarations.
 */

extern	char	*sbrk(), *shsbrk();
extern	char	*shmalloc();

#ifdef ALM
/*
 * Put implementation stuff here.
 *
 * ALM_HASH() is used to hash an address to an ALM offset.
 */

extern	char	*_alm_base;		/* virt addr of mapped ALM's */

#define	ALM_HASH(x)	((int)(&(x)) & (0xFF << 2))
#define	NBALM		4		/* number bytes per ALM */
#define	ALM_UNLOCKED	0
#define	ALM_LOCKED	1


/*
 * S_LOCK() and S_UNLOCK() provide in-line access to locks for C-programs;
 * these can be used in time-criticial situations, at a cost in code size.
 *
 * Need to use the -i flag with the optimizer when this inline is
 * used. (cc -O -i).   This is to prevent the optimizer from removing
 * "redundant" moves (which aren't really redundant in this case)
 * in the spin loop.
 */

#define	S_LOCK(lp)	{ \
	register char	*lock_alm = &_alm_base[ALM_HASH(*(lp))]; \
	for (;;) { \
		/* Wait for lock to be available */ \
		while (*(lp) == L_LOCKED) \
			continue; \
		/* Grab ALM gate for atomic access to lock */ \
		while (*lock_alm & ALM_LOCKED) \
			continue; \
		/* Can race with others trying to get the lock */ \
		if (*(lp) == L_UNLOCKED) { \
			/* No race (or won it) -- grab the lock */ \
			*(lp) = L_LOCKED; \
			*lock_alm = ALM_UNLOCKED; \
			break; \
		} \
		/* Lost race, try again */ \
		*lock_alm = ALM_UNLOCKED; \
	} \
}

#define	S_UNLOCK(lp)	(*(lp) = L_UNLOCKED)
#define S_INIT_LOCK(lp)	S_UNLOCK(lp)
/*
 * there's no inline implementation of these functions with alms.
 * provide these macros for compatability.
 */
#define S_CLOCK(lp)	s_clock(lp)
#define S_INIT_BARRIER(bp, l)	s_init_barrier(bp, l)
#define S_WAIT_BARRIER(bp)	s_wait_barrier(bp)
#define _M_JOIN(b, clearl)	_m_join((b), (clearl))

#ifdef ns32000
#define	NALMDEVS	64		/* # different ALM's to try for */
#define	ADDR_RND	0x800		/* boundary round (text/data/end) */
#endif ns32000

#ifdef i386
#define	NALMDEVS	16		/* # different ALM's to try for */
#define	ADDR_RND	4096		/* boundary round (text/data/end) */
#endif i386

#else ALM

#ifdef i386
/*
 * S_LOCK() S_UNLOCK(), S_CLOCK(), S_INIT_LOCK, and S_WAIT_BARRIER functions.
 * These are implemented as "asm functions", and will produce very good code.
 *
 * Need to use the -i flag with the optimizer when these inline functions are
 * used. (cc -O -i).   This is to prevent the optimizer from removing
 * "redundant" moves (which aren't really redundant in this case)
 * in the spin loops.
 */

asm void S_LOCK(laddr)
{
%reg laddr; lab loop, spin, done;
/PEEPOFF
loop:	movb	$L_LOCKED, %al
	xchgb	%al, (laddr)		/* an atomic "test and set" */
	cmpb	$L_UNLOCKED, %al
	je	done			/* if equal, we got the lock */
spin:	cmpb	$L_UNLOCKED, (laddr)	/* spin in cache until unlocked */
	je	loop			/* then, try get lock again */
	jmp	spin
done:
/PEEPON
%mem laddr; lab loop, spin, done;
/PEEPOFF
loop:	movb	$L_LOCKED, %al
	movl	laddr, %ecx
	xchgb	%al, (%ecx)		/* an atomic "test and set" */
	cmpb	$L_UNLOCKED, %al
	je	done			/* if equal, we got the lock */
spin:	cmpb	$L_UNLOCKED, (%ecx)	/* spin in cache until unlocked */
	je	loop			/* then, try get lock again */
	jmp	spin
done:
/PEEPON
}

asm void S_UNLOCK(laddr)
{
%reg laddr;
	movb	$L_UNLOCKED, %al
	xchgb	%al, (laddr)		/* clear lock, "atomically" */
%mem laddr;
	movb	$L_UNLOCKED, %al
	movl	laddr, %ecx
	xchgb	%al, (%ecx)		/* clear lock, "atomically" */
}

asm S_CLOCK(laddr)
{
%reg laddr; lab done;
	movb	$L_LOCKED, %dl
	movl	$L_SUCCESS, %eax
	xchgb	%dl, (laddr)		/* test and set lock */
	cmpb	$L_UNLOCKED, %dl	/* if we got it, return success */
	je	done
	movl	$L_FAILED, %eax		/* else return failure */
done:
%mem laddr; lab done;
	movb	$L_LOCKED, %dl
	movl	laddr, %ecx
	movl	$L_SUCCESS, %eax
	xchgb	%dl, (%ecx)		/* test and set lock */
	cmpb	$L_UNLOCKED, %dl	/* if we got it, return success */
	je	done
	movl	$L_FAILED, %eax		/* else return failure */
done:
}

#define S_INIT_LOCK(lp)	S_UNLOCK(lp)

/*
 * In line barrier initialization macro
 *
 * Note, there's no need to initialize either b_useno (due to its manner
 * of useage, or b_mutex (due to its non-usage).
 */

#define S_INIT_BARRIER(b, l)	((b)->b_limit = (b)->b_count = (l))

/*
 * In line barrier wait procedure.
 *
 * Define offsets of fields in barrier, for use in assembly language.
 */

#define MUTEX	0
#define COUNT	1
#define LIMIT	2
#define USENO	3

asm void S_WAIT_BARRIER(baddr)
{
%reg baddr; lab spin;
/PEEPOFF
	movb	USENO(baddr), %al	/* save old barrier useno value */
	lock				/* assert lock on next instruction */
	decb	COUNT(baddr)		/* atomic decrement of barrier count */
	jne	spin			/* if not all checked in, go spin */
	movb	LIMIT(baddr), %dl
	movb	%dl, COUNT(baddr) 	/* all checked in! re-init barrier */
	incb	USENO(baddr)		/* and free all the spinners */
spin:	cmpb	USENO(baddr), %al	/* spin here until all are checked in */
	je	spin
/PEEPON
%mem baddr; lab spin;
/PEEPOFF
	movl	baddr, %ecx		/* load barrier pointer into register */
	movb	USENO(%ecx), %al	/* save old barrier useno value */
	lock				/* assert lock on next instruction */
	decb	COUNT(%ecx)		/* atomic decrement of barrier count */
	jne 	spin			/* if not all checked in, go spin */
	movb	LIMIT(%ecx), %dl
	movb	%dl, COUNT(%ecx)	/* all checked in! re-init barrier */
	incb	USENO(%ecx)		/* and free all the spinners */
spin:	cmpb	USENO(%ecx), %al	/* spin here until all are checked in */
	je	spin
/PEEPON
}

/*
 * This is for internal use by parallel library functions only
 */
asm void _M_JOIN(baddr, clearl)
{
%mem baddr; mem  clearl; lab spin;
/PEEPOFF
	movl	baddr, %ecx		/* load barrier pointer into register */
	movb	USENO(%ecx), %al	/* save old barrier useno value */
	lock				/* assert lock on next instruction */
	decb	COUNT(%ecx)		/* atomic decrement of barrier count */
	jne 	spin			/* if not all checked in, go spin */
	movb	LIMIT(%ecx), %dl
	movb	%dl, COUNT(%ecx)	/* all checked in! re-init barrier */
	movl	clearl, %edx
	movl	$0, (%edx)
	incb	USENO(%ecx)		/* and free all the spinners */
spin:	cmpb	USENO(%ecx), %al	/* spin here until all are checked in */
	je	spin
/PEEPON
}

#endif i386

#ifdef ns32000
/*
 * There IS no non-alm implementation for ns32000
 */
#define S_LOCK(lp)	printf("S_LOCK unimplemented\n");
#define S_UNLOCK(lp)	printf("S_UNLOCK unimplemented\n");
#define S_CLOCK(lp)	printf("S_CLOCK unimplemented\n");
#define S_INIT_LOCK(lp)	printf("S_INIT_LOCK unimplemented\n");
#define S_INIT_BARRIER(bp) printf("S_INIT_BARRIER unimplemented\n");
#define S_WAIT_BARRIER(bp) printf("S_WAIT_BARRIER unimplemented\n");
#define _M_JOIN(b, clearl)	_m_join((b), (clearl))
#endif ns32000

#endif ALM


/*
 * Convenience definitions.
 */

#ifndef	NULL
#define	NULL		0
#endif

/*
 * Various globally used data.
 */

extern	int	errno;

extern	int	_shm_fd;		/* fd for shared data mapped file */
extern	int	_pgoff;			/* getpagesize() - 1 */

/*
 * PGRND() rounds up a value to next page boundary.
 */

#define	PGRND(x)	(char *) (((int)(x) + _pgoff) & ~_pgoff)
