/*LINTLIBRARY*/

/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef	lint
static	char	rcsid[] = "$Header: ppinit.c 1.23 87/05/28 $";
#endif


/*
 * ppinit.c
 *	Parallel program run-time environment initialization.
 *
 * -DOPENT ==> system supports "get RO text fd" primitive.
 *
 * TODO:
 *	Allow user to somehow specify where to create shared-data file,
 *		and not unlink it (for later rendezvous).
 *	Worry about taking signal in sbrk()/shmbrk() (use process local
 *		signal stack; only really issue for multi-tasker).
 */

/* $Log:	ppinit.c,v $
 */

#include <a.out.h>
#include <strings.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <machine/pmap.h>
#include <machine/param.h>
#include <machine/vmparam.h>

#include <parallel/parallel.h>
#include "parc.h"

extern	char	*index();
extern	char	*getenv();
extern  char	*_shared_end;
extern int _priv_stack_limit;
/*
 * Note require that stack_limit and STACK_HOLE values all be multiples
 * of STACK_INCR
 */
#define	STACK_INCR	(64*1024)		/* stack growth granularity */
int _stack_incr;
#define STACK_HOLE	(0)		/* gap between stacks */
int _stack_hole;
int _stack_limit;

/*
 * Define "_curbrk" to satisfy mmap()'s ref without
 * requiring libc's sbrk().
 */

char	*_curbrk = NULL;

#ifndef	lint
/*
 * Must insure we pull in run-time's version of sbrk().  If program
 * doesn't directly reference, will get libc's sbrk() due to malloc().
 *
 * Could just include the module, but that's less modular.
 */
extern	char	*sbrk();
static	char	*(*sbrkp)() = sbrk;
#endif	lint

/*
 * _ppinit()
 *	Parallel startup for C programs.
 *
 * Called by crt0.o before main().  Arranges shared parts of address space
 * to be shared and initialize shared-memory allocator.
 *
 * Leaves shared-data file-descriptor in _shm_fd, ALM fd in _alm_fd.
 */

extern int _do_shared_stack;

int	_shm_fd;				/* fd for shared data */
int	_alm_fd;				/* fd for mapped ALMs */
int	_stk_fd;				/* fd used for parent stack */
int	_pgoff;

static private int *_new_sp;
private int *_priv_sp;
private int _single_stack = 1;

static	char	shm_name[] = " shared data XXXXXX";
static	char	stk_name[] = " stack data XXXXXX";

#ifdef ALM
static	char	alm_name[] = "/dev/alm/almXX";
static	char	alm_map[ADDR_RND+ADDR_RND-1];	/* place to map alm's */
char		*_alm_base;			/* rounded base address */
#endif ALM

/*ARGSUSED*/
_ppinit(argc, argv)
	int	argc;
	char	**argv;

{
	register int	szshared;		/* sizeof all shared */
	register int	szinitsh;		/* sizeof init'd shared */
	extern	shared char _shstart_, _shedata_, _shend_; /* shared bounds */
	extern int *_init_shstk();
	struct rlimit priv_stack_lim;
	register struct exec *ex = (struct exec *)(LOWPAGES*NBPG);

	/*
	 * If no shared stuff, nothing to do.
	 */

	if (&_shstart_ == NULL)
		return;

	_pgoff = getpagesize() - 1;

	/*
	 * Create unnamed "shared" file to map.
	 */

	_shm_fd = _mkuntmp(shm_name);
	if (_shm_fd < 0)
		bad_init("error creating unnamed temp file", errno);

	/*
	 * Map file in at appropriate place.
	 */

	szshared = ((int)PGRND(ex->a_shdata) + (int)PGRND(ex->a_shbss));
	/*szshared = (int) PGRND(&_shend_ - &_shstart_); */
	if (MMAP(_shm_fd, &_shstart_, szshared, 0) < 0)  {
		errno = ENOEXEC;
		bad_init("shared region at invalid address", ENOEXEC);
	}
	_shared_end = &_shstart_ + szshared;

	/*
	 * If there is initialized shared data, find and read it.
	 */

	szinitsh = &_shedata_ - &_shstart_;
	if (szinitsh != 0) {
		int	tfd;

		if ((tfd = get_ro_text(argv[0])) < 0) {
#ifdef DEBUG
			printf("get_ro_text\n");
#endif DEBUG
			bad_init(argv[0], errno);
		}

		if (lseek(tfd, (off_t)(N_SHDATAOFF(*ex)), L_SET) == (off_t)-1) {
#ifdef DEBUG
			printf("lseek %x\n", N_SHDATAOFF(*ex));
#endif DEBUG
			bad_init(argv[0], ENOEXEC);
		}
		if (read(tfd, &_shstart_, szinitsh) != szinitsh) {
#ifdef DEBUG
			printf("read in shared data %x\n", szinitsh);
#endif DEBUG
			bad_init(argv[0], ENOEXEC);
		}

		(void) close(tfd);
	}

#ifdef ALM
	/*
	 * Open (exclusively) and map an ALM device.
	 */

	_alm_base = PGRND(alm_map);
	if ((_alm_fd = open_alm(alm_name, _alm_base)) < 0)
		bad_init(alm_name, errno);
#endif ALM

	/*
	 * get shared and private stack configuration information,
	 * do any rounding that's needed.
	 */

	if (_stack_incr == 0)
		_stack_incr = STACK_INCR;
	if (_stack_hole == 0)
		_stack_hole = STACK_HOLE;
	(void) getrlimit(RLIMIT_STACK, &priv_stack_lim);
	if (_stack_limit == 0)
		_stack_limit = priv_stack_lim.rlim_cur;
	_stack_limit = (_stack_limit + _stack_incr) & (~_stack_incr);

	if (_priv_stack_limit == 0)
		_priv_stack_limit = priv_stack_lim.rlim_cur;
	else {
		priv_stack_lim.rlim_cur = _priv_stack_limit;
		(void) setrlimit(RLIMIT_STACK, &priv_stack_lim);
	}

	if ( _do_shared_stack == 1 ) {
		_stk_fd = _mkuntmp(stk_name);
		if (_stk_fd < 0)
			bad_init("failure creating unnamed temp", errno);
#ifdef DEBUG
		printf("shared stack fd is %d\n", _stk_fd);
#endif DEBUG
		/*
		 * allocate and initialize the shared stack segment.
		 * new_sp points to the place in the shared stack which
		 * corresponds to our current stack pointer.
		 * only the portion of the stack from &argc
		 * 
		 * Note, that we don't copy the environment and argument
		 * strings and string pointer arrays to the
		 * shared stack.  We rely on the private copies of these
		 * things to be at the same address in every child's
		 * private stack.
		 *
		 * we only copy the stack environment necessary to put
		 * allow us to exit this routine sanely back into crt0
		 * after having switched the stack pointer register to the
		 * shared stack.
		 */

 		_new_sp = _init_shstk((int *)(&argc - 2), (int *)(&argv + 1));
		if ((int)_new_sp < 0) 
			bad_init(argv[0], errno);
#ifdef DEBUG
		printf("new sp %x\n", _new_sp);
#endif DEBUG
#ifdef ns32000
		;asm("	sprd	fp, __priv_sp");
		;asm("	lprd	fp, __new_sp");
#endif ns32000
#ifdef i386
		;asm("	movl	%ebp, __priv_sp");
		;asm("	movl	__new_sp,%ebp");
#endif i386
	}
	/*
	 * Init basic shared-memory allocator.
	 */

	if (_init_shbrk() == -1)
		bad_init("failure initializing shared heap", errno);
}

/*
 * _mkuntmp()
 *	Make an unnamed temporary file, insure we are only reference.
 */

_mkuntmp(as)
	char	*as;
{
	register char *s, *temp;
	register unsigned pid;
	register int i, j;
	register int fd;
	int tmp_sigmask;
	char name_buffer[1024];

	pid = getpid();
	for (j = 0; j < 5; j++) {

		switch(j) {
		case 0:	
			temp = getenv("TMPPATH");
			if (temp == 0)
				continue;
			(void) strcpy(name_buffer, temp);
			(void) strcat(name_buffer, "/");
			break;
		
		case 1:	
			name_buffer[0] = 0;
			break;
		case 2:
			temp = getenv("HOME");
			if (temp == 0)
				continue;
			(void) strcpy(name_buffer, temp);
			(void) strcat(name_buffer, "/");
			break;
		case 3:
			(void) strcpy(name_buffer, "/usr/tmp/");
			break;
		case 4:
			(void) strcpy(name_buffer, "/tmp/");
			break;
		}
		(void) strcat(name_buffer,as);
		s = name_buffer;
		while (*s++)
			;
		s--;
		while (*--s == 'X') {
			*s = (pid%10) + '0';
			pid /= 10;
		}
		s++;
		i = 'a';
		do {
			if (i=='z')
				break;
			*s = i++;
			tmp_sigmask = sigblock((int)0xffffffff);
			fd = open(name_buffer, O_CREAT|O_EXCL|O_RDWR, 0666);
			if (fd > 0) 
				(void) unlink(name_buffer);
			(void)sigsetmask(tmp_sigmask);
		} while((fd < 0) && (errno == EEXIST));

		if (fd >= 0) {
			break;
		}
	}

	(void)fcntl(fd, F_SETFD, 1);
	return(fd);
}

#ifdef ALM
/*
 * open_alm(dev_name)
 *	Find a free ALM device to use.  Open it and assert exclusive
 *	access.
 *
 * "dev_name" is a pointer to a device name template that looks like:
 *
 *		/dev/alm/almXX
 *
 * where "XX" will be replaced by decimal digits corresponding to
 * a free device.
 *
 * Returns file-descriptor or -1 (and errno holds reason).
 */

static
open_alm(xx_name, vaddr)
	char		*xx_name;
	register char	*vaddr;
{
	register char	*numstr;
	register int	i, fd;
	register int	alm_size = getpagesize();
	char	dev_name[80];

	strcpy(dev_name, xx_name);

	/*
	 * Search for a device to use.
	 */

	if ((numstr = index(dev_name, 'X')) == 0) {
		errno = EBUSY;
		return(-1);
	}

	for(i = 0; i < NALMDEVS; i++) {

		/*
		 * Insert the device number into the template.
		 */

		numstr[0] = '0' + (i / 10);
		numstr[1] = '0' + (i % 10);

		/*
		 * Open the ALM device, then try to exclusive
		 * lock (will fail if somebody beat us to it).
		 *
		 * Return ENOENT error if any not found (eg,
		 * ALM devices don't exist).
		 */

		if ((fd = open(dev_name, O_RDWR)) >= 0) {
			(void)fcntl(fd, F_SETFD, 1);
			if (ioctl(fd, PMAPIOCEXCL, (char *)0) >= 0) {
				if (MMAP(fd, vaddr, alm_size, 0) < 0) {
					(void) close(fd);
					return(-1);
				}
				/*
				 * Make sure all ALM gates start out unlocked.
				 */
				for (i = 0; i < alm_size; i += NBALM)
					 vaddr[i] = ALM_UNLOCKED;
				return(fd);
			}
			(void) close(fd);
		} else if (errno == ENOENT)
			return(-1);
	}

	errno = ENFILE;
	return(-1);
}
#endif ALM

/*
 * get_ro_text()
 *	Get a RO reference to executing a.out.
 *
 * Use syscall if it exists, else determine from program name and environment.
 */

#ifdef	OPENT

static
get_ro_text() { return(approp syscall); }

#else

static
get_ro_text(prog)
	char	*prog;
{
	char	name[1024];
	register char *path;
	register char *colon;
	register int len;
	register int fd;

	if (index(prog, '/'))
		return(chk_text(prog));

	if ((path = getenv("PATH")) == NULL)
		path = ":/bin:/usr/bin";

	for (colon = path; colon != NULL; path = colon + 1) {
		if (colon = index(path, ':')) {
			len = colon - path;
			(void) strncpy(name, path, len);
			name[len] = '\0';
		} else
			(void) strcpy(name, path);
		if (name[0])
			(void) strcat(name, "/");
		(void) strcat(name, prog);
		if ((fd = chk_text(name)) > 0) {
			return(fd);
		}
	}

	errno = ENOENT;
	return(-1);
}

/*
 * try to check that this is at least an executable a.out format file!!
 *	return an open file descriptor of true.
 *	otherwise, return -1, with errno set to an appropriate error code.
 */

static
chk_text(name)
char *name;
{
	int fd;
	struct exec exec_buf;
	

	if (access(name, R_OK|X_OK))
		return(-1);

	if ((fd = open(name, O_RDONLY)) < 0) 
		return(-1);

	if (read(fd, (char *)&exec_buf, sizeof(exec_buf)) < sizeof(exec_buf))  {
		errno = ENOEXEC;
		return(-1);
	}

	/*
	 * compare the a.out file header read from the file with
	 * what's in memory.  If they match, must be the same file.
	 * otherwise, not
	 *
	 * 2048 should be a defined constant in a header file some place.
	 * It's not yet, so I'm leaving it as an ugly number so every one will
	 * barf when they see it and want to fix it.
	 */
	if (bcmp((char *)(LOWPAGES*NBPG), (char *)&exec_buf, sizeof(exec_buf)))
	{
		errno = ENOEXEC;
		return(-1);
	}

	return(fd);
}

#endif	OPENT

/*
 * bad_init()
 *	For some reason, couldn't init -- complain and exit with error status.
 */

static
bad_init(msg, err)
	char	*msg;
	int	err;
{
	perror(msg);
	_exit(err);
}


_exit(status)
int status;
{
	_lib_exit(status);
}


int
sigvec(sig, vec, ovec)
int sig;
struct sigvec *vec, *ovec;
{
		return(_lib_sigvec(sig, vec, ovec));
}
