/*LINTLIBRARY*/

/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988, 1989, 1990 
 * Sequent Computer Systems, Inc.   All rights reserved.
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

#ifndef lint
static char rcsid[] = "$Header: slock.c 1.11 87/09/28 $";
#endif

/*
 * lock.c
 *	Implementation of spin-locks.
 * 
 * A "lock" is a byte of memory, initialized to zero (unlocked).
 * To atomically change value to "locked" an ALM is used (determined
 * by address of the lock).
 *
 * A lock may be released by just clearing it (uses bus-watch cache).
 */

/* $Log:	slock.c,v $
 */

#include <parallel/parallel.h>
#include "parc.h"

/*
 * s_init_lock()
 *	Initialize the lock by simply writing it with L_UNLOCKED.
 *
 * Note that locks are implicitly initialized (unlocked) by placing them
 * in newly created space (shsbrk()) or shared bss.  Locks may also be
 * explicitly initialized to L_UNLOCKED.
 */

s_init_lock(lock)
	slock_t	*lock;
{
	S_UNLOCK(lock);
}


/*
 * s_lock()
 *	Acquire a lock by waiting for it to be unlocked, then
 *	atomically setting it to locked.
 *
 * The ALM "gate" to use is determined from the address of the lock.
 */

s_lock(lock)
	register slock_t *lock;
{
	S_LOCK(lock);
}


/*
 * s_unlock()
 *	Release a lock.
 */

s_unlock(lock)
	slock_t	*lock;
{
	S_UNLOCK(lock);
}


/*
 * s_clock()
 *
 * Conditional "spin" lock.
 *
 * If the lock is available, grab it and return L_SUCCESS
 * Otherwise DON'T spin on the lock.  Just return L_FAILED
 */

#ifdef ALM
	/*
	 * This is the alm-based implementation of s_clock.
	 * Turns out, this is hard to implement in-line.
	 */
s_clock(lp)	
register slock_t *lp;
{ 
	register char	*lock_alm = &_alm_base[ALM_HASH(*(lp))]; 

		if (*(lp) == L_LOCKED) 
			return(L_FAILED); 

		/* Grab ALM gate for atomic access to lock */ 
		if (*lock_alm & ALM_LOCKED) 
			return(L_FAILED);

		/* Can race with others trying to get the lock */ 
		if (*(lp) == L_UNLOCKED) { 
			/* No race (or won it) -- grab the lock */ 
			*(lp) = L_LOCKED; 
			*lock_alm = ALM_UNLOCKED; 
			return(L_SUCCESS);
		} 
		/* Lost race, try again */ 
		*lock_alm = ALM_UNLOCKED; 
		return(L_FAILED);
}
#else
	/*
	 * Use a machine-dependent implementation, extracted from parallel.h
	 */
s_clock(lp)	
slock_t *lp;
{
	return(S_CLOCK(lp));
}
#endif ALM

/* Generate Fortran and Pascal -e entry points */
ENTRY(s_init_lock, _s_init_lock);
ENTRY(s_lock, _s_lock);
ENTRY(s_unlock, _s_unlock);
ENTRY(s_clock, _s_clock);

/* Generate Fortran and Pascal V2.9.1 entry points */
ENTRY(_s_init_lock_, _s_init_lock);
ENTRY(_s_lock_, _s_lock);
ENTRY(_s_unlock_, _s_unlock);
ENTRY(_s_clock_, _s_clock);
