/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */

/* type semantic routines */
#define isconst(z) (z->producer->x.op == OPLOAD && z->producer->inputs->intype == INCONST)
#define MAXINT 0x7fffffff
#include <stdio.h>
#include "symbol.h"
#include "pool.h"
#include "ic.h"
#include "lex.h"
#include "spec.h"
extern struct pool progtables, tempic, pupass1, stmts;
extern struct tentry *oldsym(), *newsym(), *unitname;
extern struct symindex *externals, *thisunit;
extern struct stseg *specinsert;
extern struct doacross *dotab;

#define NELEMTYPES 13
struct roottype {
	char *name;
	int btype, size;
	struct tentry *t;
} elemtype[NELEMTYPES] = {
	{"INTEGER",TYINT,4,0}, {"LOGICAL",TYLOG,4,0}, 
	{"LOGICAL*1",TYLOG,1,0}, {"LOGICAL*2",TYLOG,2,0}, 
	{"REAL",TYREAL,4,0}, {"DOUBLE PRECISION",TYDOUBLE,8,0}, 
	{"COMPLEX",TYCOMPLEX,8,0}, {"COMPLEX*16",TYCOMPLEX,16,0}, 
	{"CHARACTER",TYUSCHAR,1,0}, {"VOID",TYVOID,0,0},
	{"INTEGER*1",TYCHAR,1,0}, {"INTEGER*2",TYSHORT,2,0},
	{"STMT LABEL",TYLABEL,0,0}
};
/* internal procedure to look up types */
static struct tentry *findtype(t,p,s)
char *s;
int t, p;
{
	struct tentry *tab;
	tab = oldsym(externals,s);
	while (tab && (tab->class != CLTYPE ||
		tab->x.type.elemtype !=  t ||
		tab->x.type.size != p*8)) tab = tab->next;
	if (!tab) {
		tab = newsym(&progtables,externals,s,0);
		tab->class = CLTYPE;
		tab->x.type.elemtype = t;
		tab->x.type.size = p*8;
	}
	return(tab);
}
/* return type table entry for elementary type */
struct tentry *maketype(t,p)
int t,p;
{
	struct tentry *tab;
	int i;
	char buf[80];
#ifdef debugmaketype
	printf("maketype(%d,%d)\n",t,p);
#endif
	if (t == TYDOUBLE && p == 4) t = TYREAL;
	else if (t == TYREAL && p == 8) t = TYDOUBLE;
	else if (t == TYINT && p == 1) t = TYCHAR;
	else if (t == TYINT && p == 2) t = TYSHORT;
	else if (t == TYLOG && p == 1) t = TYUSCHAR;
	else if (t == TYLOG && p == 2) t = TYUSHORT;
	for (i = 0; i < NELEMTYPES; i++) {
		if (elemtype[i].btype == t && 
			elemtype[i].size == p) {
			if ((tab = elemtype[i].t) == 0) {
				elemtype[i].t = tab = 
					findtype(t,p,elemtype[i].name);
			}
#ifdef debugmaketype
			printf("maketype returns....\n");
			dumptentry(tab);
#endif
			return(tab);
		}
	}
	/* make character*n entries */
	if (t == TYCHAR || t == TYUSCHAR) {
		if (p) {
			sprintf(buf,"CHARACTER*%d", p);
		} else {
			sprintf(buf,"CHARACTER*(*)");
		}
		tab = oldsym(thisunit,buf);
		if (tab) return(tab);
		tab = newsym(&pupass1,thisunit,buf,0);
		tab->class = CLTYPE;
		tab->x.type.elemtype = t;
		tab->x.type.size = p*8;
		return(tab);
	}
	yyerror("unsupported type and precision");
	return(findtype(t,p,"invalid"));
}

int intcon(e)
struct output *e;
{
	if (isconst(e) && e->t->x.type.elemtype == TYINT)
		return(e->producer->inputs->x.table->x.const.intval);
	else yyerror("integer constant expected");
}


/* there were 26 letters until systems programming
	came along-- they had to add '%' (26), and '_' (27). */
struct tentry *lettertype[28], *inttype, *realtype, *ptrundeftype, *logtype,
	*doubletype, *complextype, *voidtype, *chartype, *labeltype;

/* initialize implicit types */
initimplicit()
{
	int i;
	inttype = maketype(TYINT,4);
	realtype = maketype(TYREAL,4);
	logtype = maketype(TYLOG,4);
	doubletype = maketype(TYDOUBLE,8);
	complextype = maketype(TYCOMPLEX,8);
	ptrundeftype = newptrtype(0);
	chartype = maketype(TYUSCHAR,1);
	voidtype = maketype(TYVOID,0);
	labeltype = maketype(TYLABEL,0);
	for (i = 0; i < 'i'-'a'; ) lettertype[i++] = realtype;
	for (i = 'i'-'a'; i < 'o'-'a'; ) lettertype[i++] = inttype;
	for (i = 'o'-'a'; i < 26;) lettertype[i++] = realtype;
	lettertype[26] = lettertype[27] = inttype;
}

/* process implicit semantics */
int lettercode(c)
register char c;
{
	return ((c >= 'a' && c <= 'z') ? c - 'a' : 
		((c >= 'A' && c <= 'Z') ? c - 'A' : 
		((c == '%' ? 26 : 27))));
}
/* return bitmask for set of letters */
int letterbound(low,high)
struct tentry *low, *high;
{
	int lv, hv;
	lv = lettercode(low->symbol[0]); 
	hv = lettercode(high->symbol[0]); 
	return( /* all letters <= high */	((1<<(hv+1))-1) & 
		/* all letters >= low */	~((1<<lv)-1) );
}

/* process implicit entries */
impset(t,lset)
int lset;
struct tentry *t;
{
	int i, j;
	for (i=0, j=1; lset; i++,j+=j)
		if (lset & j) {
			lettertype[i] = t;
			lset -= j;
		}
}
int strlenp1(s)
char *s;
{
	char *r;
	r = s;
	while (*r++);
	return(r-s);
}
/* expression action routines */
struct tentry *makestring(i)
char *i;
{
	struct tentry *t;
	char buf[YYLMAX];
	sprintf(buf,"'%x",i);
	t = oldsym(thisunit,buf);
#ifdef debugmakestring
	if (t) {
		printf("makestring %s returns old symbol...\n", i);
		dumptentry(t);
	}
#endif
	if (t) return(t);
	t = newsym(&pupass1,thisunit,buf,0);
	t->t = maketype(TYUSCHAR,strlen(i));
	t->class = CLCONST;
	t->x.const.charval = i;
#ifdef debugmakestring
	printf("makestring %s returns...\n",i);
	dumptentry(t);
#endif
	return(t);
}
struct tentry *makelog(i)
int i;
{
	char buf[80];
	struct tentry *t;
	sprintf(buf,"%s",i ? ".true." : ".false.");
	t = oldsym(thisunit,buf);
#ifdef debugmakelog
	if (t) {
		printf("makelog %d returns old symbol...\n", i);
		dumptentry(t);
	}
#endif
	if (t) return(t);
	t = newsym(&pupass1,thisunit,buf,0);
	t->t = logtype;
	t->class = CLCONST;
	t->x.const.intval = i;	/* just use intval variant */
#ifdef debugmakelog
	printf("makelog %d returns...\n",i);
	dumptentry(t);
#endif
	return(t);
}
struct tentry *makeint(i)
int i;
{
	char buf[80];
	struct tentry *t;
	sprintf(buf,"%d",i);
	t = oldsym(thisunit,buf);
#ifdef debugmakeint
	if (t) {
		printf("makeint %d returns old symbol...\n", i);
		dumptentry(t);
	}
#endif
	if (t) return(t);
	t = newsym(&pupass1,thisunit,buf,0);
	t->t = inttype;
	t->class = CLCONST;
	t->x.const.intval = i;
#ifdef debugmakeint
	printf("makeint %d returns...\n",i);
	dumptentry(t);
#endif
	return(t);
}
struct tentry *makereal(i)
double *i;
{
	char buf[80];
	struct tentry *t;
	sprintf(buf,"0f%8x%8x",*i);
	t = oldsym(thisunit,buf);
#ifdef debugmakereal
	if (t) {
		printf("makereal %f returns old symbol...\n", *i);
		dumptentry(t);
	}
#endif
	if (t) return(t);
	t = newsym(&pupass1,thisunit,buf,0);
	t->t = realtype;
	t->class = CLCONST;
	t->x.const.realval = *i;
#ifdef debugmakereal
	printf("makereal %f returns...\n",*i);
	dumptentry(t);
#endif
	return(t);
}
struct tentry *makedbl(i)
double *i;
{
	char buf[80];
	struct tentry *t;
	sprintf(buf,"0d%8x%8x",*i,((int *)i)[1]);
	t = oldsym(thisunit,buf);
#ifdef debugmakedbl
	if (t) {
		printf("makedbl %f returns old symbol...\n", *i);
		dumptentry(t);
	}
#endif
	if (t) return(t);
	t = newsym(&pupass1,thisunit,buf,0);
	t->t = doubletype;
	t->class = CLCONST;
	t->x.const.realval = *i;
#ifdef debugmakedbl
	printf("makedbl %f returns...\n",*i);
	dumptentry(t);
#endif
	return(t);
}
struct tentry *makecomplex(i,j)
double *i, *j;
{
	char buf[80];
	struct tentry *t;
	sprintf(buf,"(0x%6x%6x,0x%6x%6x)",*((int *)i),((int *)i)[1],
		*((int *)j),((int *)j)[1]);
	t = oldsym(thisunit,buf);
#ifdef debugmakecomplex
	if (t) {
		printf("makecomplex (%f,%f) returns old symbol...\n", *i,*j);
		dumptentry(t);
	}
#endif
	if (t) return(t);
	t = newsym(&pupass1,thisunit,buf,0);
	t->t = complextype;
	t->class = CLCONST;
	t->x.const.cmpxval.real = *i;
	t->x.const.cmpxval.imag = *j;
#ifdef debugmakecomplex
	printf("makecomplex (%f,%f) returns new symbol...\n", *i,*j);
	dumptentry(t);
#endif
	return(t);
}
double *getdbladdr(c)
struct tentry *c;
{
	int t;
	double *dp;
	struct output *o;
	t = c->t->x.type.elemtype;
	if (t == TYREAL || t == TYDOUBLE) return(&c->x.const.realval);
	if (t != TYINT && t != TYSHORT) yyerror("invalid complex constant");
	dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
	*dp = c->x.const.intval;
	return(dp);
}
struct tentry *makecmpx(i, j)
struct tentry *i, *j;
{
	double *r, *imag;
	r = getdbladdr(i);
	imag = getdbladdr(j);
	return(makecomplex(r, imag));
}

struct output *unop(opcode,inty,invalue)
int inty, opcode;
union inptr invalue;
{
	struct input *i;
	struct output *o;
	struct operation *op;
	extern struct output *fold();

	if (inty == INRESULT && (o = fold(opcode,invalue,invalue)))
		return(o);
	o = NEW(&tempic,output);
	o->next = 0;
	o->consumers = 0;
	o->producer = op = NEW(&tempic,operation);
	switch (opcode) {
	case OPCNVINT:
			o->t = inttype;
			break;
	case OPCNVREAL:
			o->t = realtype;
			break;
	case OPCNVDOUBLE:
			o->t = doubletype;
			break;
	default:
		o->t = (inty != INRESULT) ? 
			(basetype(invalue.table)) : invalue.result->t;
		break;
	}
	o->reg = 0;
	o->regclass = 0;
	o->outnumb = 1;
	op->optext = 0;
	op->type = OPTYOP;
	op->outputs = o;
	op->inputs = i = NEW(&tempic,input);
	op->pred = op->succ = 0;
	op->x.op = opcode;
	i->next = i->nextconsumer = 0;
	i->nextconsumer = 0;
	i->intype = inty;
	i->innumb = 1;
	i->x = invalue;
	return(o);
}

struct output *expid(id,db)
struct tentry *id;
struct text *db;
{
	if (basetype(id) == 0) id->t = addbasetype(id->t, imptype(id));
	doacrossid(id,db);
	/* check storage classes as well as type */
	if (id->class == CLCONST)
		return(unop(OPLOAD,INCONST,id));
	else	return(unop(OPLOAD,INADDR,id));
}
struct tentry *addfun(targ,t)
struct tentry *t;
{
	switch (t->x.type.elemtype) {
	case TYPTR:
		return(newptrtype(addfun(targ,t->x.type.subtype.ptr.to)));
	default:
		return(newfuntype(targ,t));
	}
}
struct output *expidfun(id,arg)
struct tentry *id;
struct output *arg;
{
	struct output *o;
	struct tentry *ty;
	struct stseg *seg;
	char *t;
	char *specname;
	extern struct tentry *generic2specific(), *isintrinsic();
	struct input *in;
	ty = id->t;
	while (ty && ty->x.type.elemtype == TYPTR)
		ty = ty->x.type.subtype.ptr.to;
	if (ty->x.type.elemtype != TYFUNC && ty->x.type.elemtype != TYARRAY) {
		 id->t = addfun(arg->t,id->t);
		/* add external statement if in doacross loop -- the call may disappear */
		if (dotab && id->t->x.type.elemtype == TYPTR) {
			seg = NEW(&stmts, stseg);
			seg->newstmt = 1;
			seg->f = 0;
			seg->lineno = 0;
			seg->prev = specinsert->prev;
			seg->next = specinsert;
			if (seg->prev) seg->prev->next = seg;
			seg->next->prev = seg;
			sprintf(seg->text, "      EXTERNAL %s", id->symbol);
			for (t = seg->text; *t;) t++;
			seg->stlen = t - seg->text;
	}	}
	if (ty->x.type.elemtype != TYARRAY) {
		if (arg) {
			/* convert generic name to specific name */
			id = generic2specific(id, arg->t);
#ifdef debugexpidfun
			dumptentry(id);
#endif
			if (id->class == CLCONST && id->t && 
				id->t->x.type.elemtype == TYFUNC && 
				id->x.func.funtype == INTRINSICFUNC) {
				/* return operator for intrinsic function */
				/* temporary implementation until statement functions */
				if (arg->producer->x.op == OPCOMPOSE) {
					arg->producer->x.op = (int) id->x.func.intop;
					for (in = arg->producer->inputs; in; 
						in = in->next) 
						in->innumb -= 1;
				} else {
					arg = unop((int)id->x.func.textseg, 
						INRESULT, arg);
				}
#ifdef debugexpidfun
				fprintf(stderr,"expidfun recognizes generic function:\n");
				dumpoutput(arg);
#endif
				return(arg);
			}
			if (basetype(id) == 0) {
				ty = isintrinsic(id->symbol,arg->t);
				id->t = addbasetype(id->t, ty ? ty : imptype(id));
			}
		}
		if (id->class == 0) {
			id->class = CLCONST;
			id->x.func.fg = 0;
			id->x.func.funtype = EXTFUNC;
		}
	} 
	if (basetype(id) == 0) {
		id->t = addbasetype(id->t, (imptype(id)));
	}
	doacrossid(id,0);
	/* check storage classes as well as type */
	o = (id->class == CLCONST) ? unop(OPLOAD,INCONST,id)
		: unop(OPLOAD,INADDR,id);
	o->t = basetype(id);
	return(o);
}
/* check for algebraic identities */
struct output *algiden(opcode,aout,bout)
struct output *aout, *bout;
int opcode;
{
	register struct tentry *a, *b;
	extern struct output *expint(), *binop();
	a = isconst(aout) ? aout->producer->inputs->x.table : 0;
	b = isconst(bout) ? bout->producer->inputs->x.table : 0;
	if (a && a->t->x.type.elemtype != TYINT) return(0);
	if (b && b->t->x.type.elemtype != TYINT) return(0);
	switch (opcode) {
	case OPADD:
			if (a && a->x.const.intval == 0) return(bout);
			if (b && b->x.const.intval == 0) return(aout);
			break;
	case OPSUB:
			if (a && a->x.const.intval == 0) return(bout);
			if (b && b->x.const.intval == 0) return(aout);
			if (a && b && a == b) return(expint(0));
			break;
	case OPMULT:
			if (a && a->x.const.intval == 1) return(bout);
			if (b && b->x.const.intval == 1) return(aout);
			if (a && a->x.const.intval == 0) return(expint(0));
			if (b && b->x.const.intval == 0) return(expint(0));
			if (a && a->x.const.intval == 2) 
				return(binop(OPADD,bout,bout));
			if (b && b->x.const.intval == 2) 
				return(binop(OPADD,aout,aout));
			break;
	case OPDIV:
			if (b && b->x.const.intval == 1) return(aout);
			if (a && a->x.const.intval == 0) return(expint(0));
			if (a && b && a == b) return(expint(1));
			break;
	case OPPWR:
			if (b && b->x.const.intval == 1) return(aout);
			if (b && b->x.const.intval == 0) return(expint(1));
			if (b && b->x.const.intval == 2) 
				return(binop(OPMULT,aout,aout));
			if (b && b->x.const.intval == -1) 
				return(binop(OPDIV,expint(1),aout));
			break;
	}
	return(0);
}
/* constant folding */
struct output *fold(opcode,aout,bout)
struct output *aout, *bout;
int opcode;
{
	register struct tentry *a, *b;
	register int ty;
	double d, *dp, *dp2;
	int i, ibase, iexp, itmp;
	char *zeroptr = "\0\0\0\0\0\0\0\0";
	/* check for both constants */
	if (aout->producer->x.op != OPLOAD || 
		bout->producer->x.op != OPLOAD || 
		aout->producer->inputs->intype != INCONST ||
		bout->producer->inputs->intype != INCONST) 
			return(algiden(opcode,aout,bout));
	a = aout->producer->inputs->x.table;
	b = bout->producer->inputs->x.table;
	/* check for both same type */
	ty = a->t->x.type.elemtype;
	if (ty != b->t->x.type.elemtype) return(0);
	switch (opcode) {
	case OPCNVINT:	
		switch (ty) {
		case TYINT:
			return(aout);
		case TYCOMPLEX:	
			if ((d = a->x.const.cmpxval.real) > 0) {
				if (d > MAXINT) return(0);
			} else 	if (d < -MAXINT-1) return(0);
			return(unop(OPLOAD,INCONST,makeint((int) d)));
		case TYREAL:
		case TYDOUBLE:
			if ((d = a->x.const.realval) > 0) {
				if (d > MAXINT) return(0);
			} else if (d < -MAXINT-1) return(0);
			return(unop(OPLOAD,INCONST,makeint((int) d)));
		};
		break;
	case OPCNVREAL:
		dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
		switch (ty) {
		case TYINT:	*dp = a->x.const.intval;
			return(unop(OPLOAD,INCONST,makereal(dp)));
		case TYDOUBLE:	*dp = a->x.const.realval;
			return(unop(OPLOAD,INCONST,makereal(dp)));
		case TYCOMPLEX:	*dp = a->x.const.cmpxval.real;
			return(unop(OPLOAD,INCONST,makereal(dp)));
		case TYREAL:
			return(aout);
		};
		break;
	case OPCNVDOUBLE:
		dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
		switch (ty) {
		case TYINT:	*dp = a->x.const.intval;
			return(unop(OPLOAD,INCONST,makedbl(dp)));
		case TYREAL:	*dp = a->x.const.realval;
			return(unop(OPLOAD,INCONST,makedbl(dp)));
		case TYDOUBLE:
			return(aout);
		case TYCOMPLEX:	*dp = a->x.const.cmpxval.real;
			return(unop(OPLOAD,INCONST,makedbl(dp)));
		};
		break;
	case OPCNVCOMPLEX:
		dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
		switch (ty) {
		case TYINT:	*dp = a->x.const.intval;
			return(unop(OPLOAD,INCONST,makecomplex(dp,zeroptr)));
		case TYDOUBLE:
		case TYREAL:	*dp = a->x.const.realval;
			return(unop(OPLOAD,INCONST,makecomplex(dp,zeroptr)));
		case TYCOMPLEX:	return(aout);
		};
		break;
	case OPNEG:	
		switch (ty) {
		case TYINT:
			return(unop(OPLOAD,INCONST,makeint(
				 - a->x.const.intval)));
		case TYREAL:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = - a->x.const.realval;
			return(unop(OPLOAD,INCONST,makereal(dp)));
		case TYDOUBLE:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = - a->x.const.realval;
			return(unop(OPLOAD,INCONST,makedbl(dp)));
		case TYCOMPLEX:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = - a->x.const.cmpxval.real;
		  dp2 = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp2 = - a->x.const.cmpxval.imag;
			return(unop(OPLOAD,INCONST,makecomplex(dp,dp2)));
		};
		break;
	case OPNOT:	
		switch (ty) {
		case TYINT:
			return(unop(OPLOAD,INCONST,makeint(~a->x.const.intval)));
		case TYLOG:
			return(unop(OPLOAD,INCONST,makelog(~a->x.const.intval)));
		};
		break;
	case OPAND:
		switch (ty) {
		case TYLOG:
			return(unop(OPLOAD,INCONST,makelog(
				a->x.const.intval & b->x.const.intval)));
		case TYINT:
			return(unop(OPLOAD,INCONST,makeint(
				a->x.const.intval & b->x.const.intval)));
		};
		break;
	case OPOR:
		switch (ty) {
		case TYLOG:
			return(unop(OPLOAD,INCONST,makelog(
				a->x.const.intval | b->x.const.intval)));
		case TYINT:
			return(unop(OPLOAD,INCONST,makeint(
				a->x.const.intval | b->x.const.intval)));
		};
		break;
	case OPADD:	
		switch (ty) {
		case TYINT:
			return(unop(OPLOAD,INCONST,makeint(
				a->x.const.intval + b->x.const.intval)));
		case TYREAL:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.realval + b->x.const.realval;
			return(unop(OPLOAD,INCONST,makereal(dp)));
		case TYDOUBLE:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.realval + b->x.const.realval;
			return(unop(OPLOAD,INCONST,makedbl(dp)));
		case TYCOMPLEX:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.cmpxval.real + b->x.const.cmpxval.real;
		  dp2 = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp2 = a->x.const.cmpxval.imag + b->x.const.cmpxval.imag;
			return(unop(OPLOAD,INCONST,makecomplex(dp,dp2)));
		};
		break;
	case OPSUB:	
		switch (ty) {
		case TYINT:
			return(unop(OPLOAD,INCONST,makeint(
				a->x.const.intval - b->x.const.intval)));
		case TYREAL:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.realval - b->x.const.realval;
			return(unop(OPLOAD,INCONST,makereal(dp)));
		case TYDOUBLE:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.realval - b->x.const.realval;
			return(unop(OPLOAD,INCONST,makedbl(dp)));
		case TYCOMPLEX:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.cmpxval.real - b->x.const.cmpxval.real;
		  dp2 = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp2 = a->x.const.cmpxval.imag - b->x.const.cmpxval.imag;
			return(unop(OPLOAD,INCONST,makecomplex(dp,dp2)));
		};
		break;
	case OPMULT:	
		switch (ty) {
		case TYINT:
			return(unop(OPLOAD,INCONST,makeint(
				a->x.const.intval * b->x.const.intval)));
		case TYREAL:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.realval * b->x.const.realval;
			return(unop(OPLOAD,INCONST,makereal(dp)));
		case TYDOUBLE:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.realval * b->x.const.realval;
			return(unop(OPLOAD,INCONST,makedbl(dp)));
		case TYCOMPLEX:
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.cmpxval.real * b->x.const.cmpxval.real -
			a->x.const.cmpxval.imag * b->x.const.cmpxval.imag;
		  dp2 = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp2 = a->x.const.cmpxval.imag * b->x.const.cmpxval.real +
			a->x.const.cmpxval.real * b->x.const.cmpxval.imag;
			return(unop(OPLOAD,INCONST,makecomplex(dp,dp2)));
		};
		break;
	case OPDIV:	
		switch (ty) {
		case TYINT:
			if (b->x.const.intval == 0) return(0);
			return(unop(OPLOAD,INCONST,makeint(
				a->x.const.intval / b->x.const.intval)));
		case TYREAL:
			if (b->x.const.realval == 0) return(0);
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.realval / b->x.const.realval;
			return(unop(OPLOAD,INCONST,makereal(dp)));
		case TYDOUBLE:
			if (b->x.const.realval == 0) return(0);
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = a->x.const.realval / b->x.const.realval;
			return(unop(OPLOAD,INCONST,makedbl(dp)));
		case TYCOMPLEX:
		   d = b->x.const.cmpxval.real * b->x.const.cmpxval.real + 
		   	b->x.const.cmpxval.imag * b->x.const.cmpxval.imag;
			if (d == 0) return(0);
		  dp = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp = (a->x.const.cmpxval.real * b->x.const.cmpxval.real +
			a->x.const.cmpxval.imag * b->x.const.cmpxval.imag) / d;
		  dp2 = (double *) (*(tempic.alloc))(&tempic,sizeof(double));
			*dp2 = (a->x.const.cmpxval.imag * b->x.const.cmpxval.real -
			a->x.const.cmpxval.real * b->x.const.cmpxval.imag) / d;
			return(unop(OPLOAD,INCONST,makecomplex(dp,dp2)));
		};
		break;
	case OPPWR:
		switch (ty) {
		case TYINT:
			itmp = 1;
			ibase = a->x.const.intval;
			iexp = b->x.const.intval;
			if (iexp < 0) itmp = 0;
			for (i = 0; i < iexp; i++)
				itmp *= ibase;
			return(unop(OPLOAD,INCONST,makeint(itmp)));
		};
		break;
	}
	return(0);	/* can't do it */
}
/* binary operations */
/* precedence of conversion */
int cnvfirst[TYINC] = {
	0,	/* undefined is done */
	0,	/* void is done */
	0,	/* arrays are problems */
	1,	/* subranges go to integers quickly */
	0,	/* pointers are a problem */
	0,	/* records are a problem */
	0,	/* fields are a problem */
	0,	/* logical can't be converted in FORTRAN */
	2,	/* integer goes to real in FORTRAN */
	4,	/* real goes to double in FORTRAN */
	5,	/* double goes to complex in FORTRAN */
	6,	/* complex is the end of the line */
	3,	/* unsigned integers(?) go to real */
	1,	/* unsigned characters go to unsigned integers */
	0,	/* functions can't be converted */
	1,	/* parameterized ranges go to integer */
	0,	/* labels can't be converted */
	1,	/* signed shorts go to integers */
	1	/* unsigned shorts go to unsigned integers */
	};
/* what to convert to */
int cnvto[TYINC] = {
	0,		/* undefined is in trouble */
	0,		/* void is trouble */
	0,		/* arrays are trouble */
	TYINT,		/* subranges go to integer */
	0,		/* pointers are a problem */
	0,		/* records are a problem */
	0,		/* fields are a problem */
	0,		/* logical can't be converted */
	TYREAL,		/* integers go to reals in FORTRAN */
	TYDOUBLE,	/* reals go to double in FORTRAN */
	TYCOMPLEX,	/* doubles go to complex in FORTRAN */
	0,		/* complex is the end of the line */
	TYINT,		/* char goes to integer in FORTRAN */
	TYREAL,		/* unsigned int goes to real */
	TYUSINT,	/* unsigned char goes to unsigned int */
	0,		/* functions are problems */
	TYINT,		/* parameterized ranges go to integer */
	0,		/* labels can't be converted */
	TYINT,		/* short goes to int */
	TYUSINT		/* unsigned short goes to unsigned int */
	};
/* sizes of target */
int cnvsize[TYINC] = {
	0,0,0,4,0,0,0,0,4,8,8,0,4,4,4,0,4,0,4,4
};

struct output *binop(opcode,expa,expb)
int opcode;
struct output *expa, *expb;
{
	struct input *ia, *ib;
	struct output *o;
	struct operation *op;
	int otya, otyb, tya, tyb, sizea, sizeb;
	extern struct output *cnvtype();
	char buf[100], *opsym();


	tya = expa->t->x.type.elemtype;
	tyb = expb->t->x.type.elemtype;
	sizea = expa->t->x.type.size;
	sizeb = expb->t->x.type.size;
	otya = tya;
	otyb = tyb;
#ifdef debugbinop
		printf("operator %d (type %d, type %d)\n", opcode, tya, tyb);
#endif
	while (tya != tyb) {
		if (cnvfirst[tya] < cnvfirst[tyb]) {
			if (cnvsize[tya]) sizea = cnvsize[tya];
			tya = cnvto[tya];
		} else{
			if (cnvsize[tyb]) sizeb = cnvsize[tyb];
			tyb = cnvto[tyb];
		};
		if (!tya || !tyb) {
			opsym(opcode, buf);
				printf("incompatible operands %s and %s\n",
				expa->t->symbol, expb->t->symbol);
			yyerror(buf);
#ifdef debugbinop2
			dumpoutput(expa);
			dumpoutput(expb);
#endif
			return(expa);
		}
#ifdef debugbinop
		printf("operator %d (type %d, type %d)\n", opcode, tya, tyb);
#endif
	}
	if (otya != tya) expa = cnvtype(expa, maketype(tya,sizea));
	if (otyb != tyb) expb = cnvtype(expb, maketype(tyb,sizeb));
	if (o = fold(opcode,expa,expb)) return(o);
	o = NEW(&tempic,output);
	o->next = 0;
	o->consumers = 0;
	o->producer = op = NEW(&tempic,operation);
	switch (opcode) {
	case OPNE:
	case OPEQ:
	case OPGT:
	case OPGE:
	case OPLT:
	case OPLE:
			o->t = logtype;
			break;
	default:	o->t = expa->t;
			break;
	}
	o->reg = 0;
	o->regclass = RGCLINT;
	o->outnumb = 1;
	op->optext = 0;
	op->type = OPTYOP+o->t->x.type.elemtype;
	op->outputs = o;
	op->inputs = ia = NEW(&tempic,input);
	op->pred = op->succ = 0;
	op->x.op = opcode+tya;
	ia->next = ib = NEW(&tempic, input);
	ia->nextconsumer = 0;
	ia->intype = INRESULT;
	ia->innumb = 1;
	ia->x.result = expa;
	ib->next = ib->nextconsumer = 0;
	ib->intype = INRESULT;
	ib->innumb = 2;
	ib->x.result = expb;
	return(o);
}
struct output *cnvtype(cn,newtype)
struct tentry *newtype;
struct output *cn;
{
	struct output *o;
	o = unop(OPCNV+(newtype->x.type.elemtype<<TYSIZE),INRESULT,cn);
	return(o);
}
parmlist(id,exp)
struct tentry *id;
struct output *exp;
{
	struct tentry *cn;
	struct input *in;
#ifdef debugparmlist
	char buf[80], *exprsym();
	printf("parameter constant ...\n");
	dumptentry(exp->producer->inputs->x.table);
#endif
	newid(id);
	id->class = CLCONST;
	if (!id->t) id->t = imptype(id);
	if (id->t != exp->t) exp = cnvtype(exp,id->t);
	in = exp->producer->inputs;
	if (in->intype != INCONST || (cn = in->x.table)->class != CLCONST) {
		id->class = CLVAR;
		id->x.var.sb = 0;
		id->x.var.offset = 0;
		id->x.var.redop = 0;
		id->x.var.dim_copied = 0;
		id->x.var.sclass = 0;
		switch(id->t->x.type.elemtype) {
		case TYINT:
		case TYREAL:
		case TYLOG:
		case TYDOUBLE:
		case TYCOMPLEX:
		case TYSHORT:
			yyerror("expression cannot be evaluated");
			break;
		default:
			break;
		}
#ifdef debugparmlist
		*exprsym(exp, buf, buf + 72) = 0;
		fprintf(stderr,"%s is not a constant: %s\n", id->symbol, buf);
#endif
		return;
	}
	id->class = CLCONST;
	id->x.const = cn->x.const;
#ifdef debugparmlist
	printf("parameter definition...\n");
	dumptentry(id);
#endif
}
newid(id)
struct tentry *id;
{
	char c[80];
	if (id->class) {
		sprintf(c,"identifier %s is already defined",id->symbol);
		yyerror(c);
	}
}
dumptentry(t)
struct tentry *t;
{
	if (!t) {
		printf("dump of symbol <nil>\n");
		return;
	}
	printf("dump of symbol %s ",t->symbol);
	if (t->parent) printf("in scope %s ",t->parent->symbol);
	printf(" of class ");
	switch (t->class) {
	case CLVAR:	printf("VAR\n"); 
			break;
	case CLTYPE:	printf("TYPE\n"); 
			break;
	case CLCONST:	printf("CONST "); 
			switch (t->t->x.type.elemtype) {
			case TYLOG:	printf("LOGICAL*%d %d\n", 
						t->t->x.type.size,
						t->x.const.intval);
					break;
			case TYUSINT:	printf("UNSIGNED ");
			case TYINT:	printf("INTEGER*%d %d\n", 
						t->t->x.type.size,
						t->x.const.intval);
					break;
			case TYDOUBLE:	printf("DOUBLE*%d %f\n", 
						t->t->x.type.size,
						t->x.const.realval);
					break;
			case TYREAL:	printf("REAL*%d %f\n", 
						t->t->x.type.size,
						t->x.const.realval);
					break;
			case TYUSCHAR:	printf("UNSIGNED ");
			case TYCHAR:	printf("CHAR*%d %s\n",
						t->t->x.type.size,
						t->x.const.charval);
					break;
			case TYCOMPLEX:	printf("COMPLEX %f %f\n", 
						t->x.const.cmpxval.real,
						t->x.const.cmpxval.imag);
					break;
			default:	printf("UNKNOWN TYPE\n");
					break;
			}
			break;
	default:	printf("UNKNOWN\n");
			break;
	}
}
dumpoutput(o)
struct output *o;
{
	struct input *in;
	char *t, buf[80];
	extern char *exprsym();
	if (!o) {
		printf("dump of output <nil>\n");
		return;
	}
	printf("dump of output at 0x%x. next: 0x%x\n", o, o->next);
	printf("	producer: 0x%x consumers: ", o->producer);
	for (in = o->consumers; in; in = in->nextconsumer)
		printf(" 0x%x ", in);
	printf("\n");
	printf("type 0x%x %s \n", o->t, o->t ? o->t->symbol : "<nil>");
	t = exprsym(o, buf);
	*t = 0;
	printf("expression: %s\n", buf);
}

int genvarindex = 0;
/* generate a temporary of type t */
struct tentry *genvar(t)
struct tentry *t;
{
	char buf[80];
	struct tentry *newt;
	sprintf(buf,"t&%d\0",genvarindex++);
	newt = newsym(&pupass1,thisunit,newt,0);
	newt->t = t;
	newt->class = CLVAR;
	newt->x.var.sclass = 0;
	newt->x.var.sb = 0;
	newt->x.var.offset = 0;
	newt->x.var.redop = 0;
	newt->x.var.dim_copied = 0;
	return(newt);
}
/* output a string for a given input */
char *inputsym(in,which,p)
struct input *in;
int which;
char *p;
{
	extern char *exprsym();
	char *t, *s;
	t = p;
	while (in && in->innumb != which) in = in->next;
	if (in) switch (in->intype) {
	case INADDR:
		for (s = in->x.table->symbol; *t++ = *s++;);
		t--;
		return(t);
	case INCONST:
		switch (in->x.table->t->x.type.elemtype) {
		case TYDOUBLE:
			sprintf(t, "%g", in->x.table->x.const.realval);
			for (; *t; t++) {
				if (*t == '.') break;
			}
			if (*t != '.') {
				*t++ = '.'; *t = 0;
			}
			for (; *t; t++) {
				if (*t == 'e') {
					*t = 'd';
					break;
			}	}
			if (*t != 'd') {
				*t++ = 'd'; *t++ = '0'; *t = 0;
			}
			while (*t) t++;
		case TYREAL:
			sprintf(t, "%g", in->x.table->x.const.realval);
			for (; *t; t++) {
				if (*t == '.') break;
			}
			if (*t != '.') {
				*t++ = '.'; *t = 0;
			}
			while (*t) t++;
		default:
			*t++ = '(';
			for (s = in->x.table->symbol; *t++ = *s++;);
			t[-1] = ')';
			*t = 0;
			return(t);
		}
	case INRESULT:
			return(exprsym(in->x.result,p));
	default:	
			break;
	} 
	*t++ = '?';
	return(t);
}
char *opsym(op,p)
int op;
char *p;
{
#define its(x) for(s=x;*t++ = *s++;);t--;break
#define itc(x) *t++ = x;break
	char *t, *s;
	t = p;
	switch (op & (-TYINC)) {
	case OPLOAD:	itc('*');
	case OPSTORE:	itc('=');
	case OPCNVINT:	its("int(");
	case OPCNVREAL:	its("real(");
	case OPCNVDOUBLE: its("dble(");
	case OPADD:	itc('+');
	case OPNEG:
	case OPSUB:	itc('-');
	case OPMULT:	itc('*');
	case OPDIV:	itc('/');
	case OPNOT:	its(".not.");
	case OPLT:	its(".lt.");
	case OPGT:	its(".gt.");
	case OPGE:	its(".ge.");
	case OPEQ:	its(".eq.");
	case OPNE:	its(".ne.");
	case OPLE:	its(".le.");
	case OPPWR:	its("**");
	case OPAND:	its(".and.");
	case OPOR:	its(".or.");
	case OPXOR:	its(".neqv.");
	case OPCOMPOSE:	itc(',');
	case OPABS:	its("abs");
	case OPSIN:	its("sin");
	case OPCOS:	its("cos");
	case OPMIN:	its("min");
	case OPMAX:	its("max");
	case OPTAN:	its("tan");
	case OPSQRT:	its("sqrt");
	case OPEXP:	its("exp");
	case OPLOG:	its("log");
	case OPSINH:	its("sinh");
	case OPCOSH:	its("cosh");
	case OPTANH:	its("tanh");
	case OPATAN:	its("atan");
	case OPATAN2:	its("atan2");
	case OPASIN:	its("asin");
	default:	itc('?');
	}
	*t = 0;
	return(t);
}
struct output *expint(i)
int i;
{
	return(unop(OPLOAD,INCONST,makeint(i)));
}
/* name the unit */
newunit(id,ty)
struct tentry *id, *ty;
{
	unitname = id;
	id->t = ty;
}
/* output a string for the given expression */
char *exprsym(e,p)
struct output *e;
char *p;
{
	int i;
	struct input *in;
	char *t;
	struct tentry *tab;
	/* is it binary or unary? */
	t = p;
	switch (e->producer->x.op & (-TYINC)) {
	case OPSTORE:
	case OPADD:
	case OPSUB:
	case OPMULT:
	case OPDIV:
	case OPPWR:
	case OPCOPY:
	case OPAND:
	case OPOR:
	case OPCAT:
	case OPEQ:
	case OPNE:
	case OPGT:
	case OPGE:
	case OPLT:
	case OPLE:
		*t++ = '(';
		t = inputsym(e->producer->inputs,1,t);
		t = opsym(e->producer->x.op,t);
		t = inputsym(e->producer->inputs,2,t);
		*t++ = ')';
		return(t);
	case OPSTKALLOC:
	case OPSTKRELEASE:
	case OPABS:
	case OPSIN:
	case OPCOS:
	case OPTAN:
	case OPSINH:
	case OPCOSH:
	case OPATAN:
	case OPATAN2:
	case OPASIN:
	case OPMIN:
	case OPMAX:
	case OPSQRT:
	case OPCNVINT:
	case OPCNVREAL:
	case OPCNVDOUBLE:
	case OPRECV:
	case OPSPINNZ:
	case OPWRFMTLIST:
	case OPSTARTWRFMT:
	case OPENDWRFMT:
	case OPSIZEOF:
			*t++ = '(';
			t = opsym(e->producer->x.op,t);
			*t++ = '(';
			i = 1;
			do {
				for (in = e->producer->inputs; in; in = in->next) {
					if (in->innumb == i) {
						if (i > 1) *t++ = ',';
						t = inputsym(in,i,t);
						break;
					}
				}
				i++;
			} while (in);
			*t++ = ')';
			*t++ = ')';
			return(t);
	
	case OPCOMPOSE:
			*t++ = '(';
			i = 1;
			do {
				for (in = e->producer->inputs; in; in = in->next) {
					if (in->innumb == i) {
						if (i > 1) *t++ = ',';
						t = inputsym(in,i,t);
						break;
					}
				}
				i++;
			} while (in);
			*t++ = ')';
			return(t);
	case OPCALL:
			*t++ = '(';
			t = inputsym(e->producer->inputs,1,t);
			*t++ = '(';
			i = 2;
			do {
				for (in = e->producer->inputs; in; in = in->next) {
					if (in->innumb == i) {
						if (i > 2) *t++ = ',';
						t = inputsym(in,i,t);
						break;
					}
				}
				i++;
			} while (in);
			*t++ = ')';
			*t++ = ')';
			return(t);
	case OPLOAD:
		if ((in = e->producer->inputs)->intype != INRESULT ) {
			return(inputsym(e->producer->inputs,1,t));
		}
	case OPNEG:
	case OPNOT:
	case OPLDADDR:
	default:	/* note fall thru from OPLOAD */
			*t++ = '(';
			t = opsym(e->producer->x.op,t);
			t = inputsym(e->producer->inputs,1,t);
			*t++ = ')';
			return(t);
	}
}
