/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */

#include <stdio.h>
#include "pool.h"
#include "symbol.h"
#include "lex.h"
extern struct text *yydbval;	/* I lied--its really YYTYPE */
extern struct pool pupass1, tempic;
extern struct lexcontext *lexbase;
/* lexical action routines */
/* return unsigned integer value from octal string */
unsigned int octal(s0)
char *s0;
{
	unsigned int i;
	char *s;
	while (*s0 == '0')s0++;	
	for (s=s0,i = 0; *s >= '0' && *s <= '7'; ) i = (i<<3) + (*s++ - '0'); 
	if ((s-s0 > 11) || (s-s0 == 10 && *s0 > '3'))
		yyerror("octal constant too large \n");
	return(i);
}
/* return unsigned integer value from binary string */
unsigned int binary(s0)
char *s0;
{
	unsigned int i;
	char *s;
	while (*s0 == '0')s0++;	
	for (s=s0,i = 0; *s >= '0' && *s <= '1'; ) i = i + i + (*s++ - '0'); 
	if (s-s0 > 32)
		yyerror("binary constant too large \n");
	return(i);
}
/* return unsigned integer value from decimal string */
unsigned int decimal(s0)
char *s0;
{
	int i;
	char *s;
#ifdef debugdecimal
	printf("begin decimal conversion of %s\n",s0);
#endif
	while (*s0 == '0')s0++;	
	for (s=s0,i = 0; *s >= '0' && *s <= '9'; ) i = i*10 + (*s++ - '0'); 
	if ((s-s0 > 10) || (s-s0 == 9 && i < 100000000) )
		yyerror("decimal constant too large \n");
#ifdef debugdecimal
		printf("decimal constant %d\n",i);
#endif
	yydbval = NEW(&tempic,text);
	yydbval->next = 0;
	yydbval->loopseg = lexbase->tokenseg;
	yydbval->offset = lexbase->tokenchar - lexbase->tokenseg->text;
	return(i);
}
/* return unsigned integer value from hexidecimal string */
unsigned int hexidecimal(s0)
char *s0;
{
	int i;
	char *s;
	while (*s0 == '0')s0++;	
	for(s=s0, i = 0;;) {
		if (*s >= '0' && *s <= '9')i = (i<<4) + (*s++ - '0'); 
		else if (*s >= 'a' && *s <= 'f')i = (i<<4) + (*s++ - 'a'+10); 
		else if (*s >= 'A' && *s <= 'F')i = (i<<4) + (*s++ - 'A'+10); 
		else break;
	}
	if (s-s0 > 8)
		yyerror("hexidecimal constant too large \n");
	return(i);
}
/* return pointer to string/hollerith constant value in table area */
char *string(s0,l)
int l;
char *s0;
{
	register char *t, c, *s;
	s = s0;
	if (*s == '"' || *s == '\''){
		s[l-2] = 0;
		s++;
		l -= 2;
	} else {
		while ((c = *s++) != 'h' && c !=  'H');
		l -= s - s0;
	}
	t = (*(pupass1.alloc))(&pupass1,l);
	while (*t++ = *s++);
	return(t);
}	
/* collect floating point constant */
/* The host format is the same as the target */
#define MAXMANT 18
#define EXPBITS 12
#define DOUBLESIZE 8
double tenpower[] = {
	1.0e1, 1.0e2, 1.0e4, 1.0e8, 1.0e16, 
	1.0e32, 1.0e64, 1.0e128, 1.0e256
};
double *floatcons(s)
char *s;
{
	int scale, nd, i;
	double mant, *buf;
#ifdef debugfloatcons
	printf("begin floatcons(%s)\n",s);
#endif
	while (*s == '0')s++;
	nd = scale = 0;
	mant = 0.0;
	while (*s >= '0' && *s <= '9') 
		if (nd < MAXMANT) {
			nd++;
			mant = mant * 10 + (*s++ - '0');
		} else  {
			s++;
			scale++;
		}
	if (*s == '.') {
		s++;
		while (*s >= '0' && *s <= '9') {
			if (nd < MAXMANT) {
				nd++;
				mant = mant * 10 + (*s++ - '0');
				scale--;
			} else s++;
		}
	}
	if (*s++) {
		if (*s == '+') scale += decimal(++s);
		else if (*s == '-') scale -= decimal(++s);
		else scale += decimal(s);
	}
	if (scale > 0) {
		for (i = 0, nd = 1; i < EXPBITS; i++, nd += nd) {
			if (scale & nd) {
				scale -= nd;
				mant *= tenpower[i];
			}
		}
	} else if (scale < 0) {
		scale = - scale;
		for (i = 0, nd = 1; i < EXPBITS; i++, nd += nd) {
			if (scale & nd) {
				scale -= nd;
				mant /= tenpower[i];
			}
		}
	}
	if (scale != 0) yyerror("exponent magnitude too large");
	buf = (double *)(*(tempic.alloc))(&tempic,DOUBLESIZE);
#ifdef debugfloatcons
	printf("floatcons returns value %f\n",mant);
#endif
	*buf = mant;
	return(buf);
}
