/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */

/* some action routines for the doacross processor (and some others) */
#include <stdio.h>
#include "symbol.h"
#include "pool.h"
#include "ic.h"
#include "action.h"
#include "lex.h"
#include "spec.h"
#include "y.tab.h"
#define isconst(z) (z->producer->x.op == OPLOAD && z->producer->inputs->intype == INCONST)
extern struct tentry *lettertype[28];
extern struct pool progtables, tempic, pupass1, stmts;
extern struct tentry *oldsym(), *newsym(), *unitname;
extern struct symindex *externals, *thisunit, *newtable();
extern struct doacross *dotab;
extern struct lexcontext *lexbase;
extern struct stseg *specinsert, *execinsert;
extern char *exprsym();
extern int reentrant;
extern int reversed_args;

static enddoflag = 0;
struct dotbl *dostack = 0;
struct stseg *firstdosub = 0, *lastdosub = 0;
struct idlist *statictemplist = 0;
int tempidno = 0;
int needsdynalloc = 0;	/* uses m_dynalloc */
int didexecutable = 0;	/* did we find an executable statement */
struct tentry *stfunid = 0;
/* the following variables are used to construct new statements from the
	input text. Currently, this is done only for statements with 
	function arguments which are expressions. */
int newstmtltext = 0;	/* length of function argument substitute text */
char *newstmtptrs[YYLMAX];	/* pointers to text */

struct chgs {
	struct chgs *next;
	struct stseg *seg;
	int offset;
	struct looplbls *label;
};
extern struct tentry *genlbl();
static	int iodd;
static	char pnamebuf[80];
static	int dynamicarrays, numargs;
static	struct stseg *enddobodyseg, *dobodyseg, *newstart, 
		*newend, *oldseg;
static 	struct tentry *schedlbl, *chunklbl, *exitlbl, *tripid;

/* find/create a variable s of type t */
struct tentry *makevar(s,t)
char *s;
struct tentry *t;
{
	struct tentry *label;
	struct idlist *l;
	label = oldsym(thisunit,s);
	while (label && (label->class != CLVAR || 
		strcmp(label->symbol,s) != 0))
		label = label->next;
	if (label && label->t != t) yyerror("makevar type conflict");
	if (label) return (label);
	label = newsym(&pupass1,thisunit,s,unitname);
	label->class = CLVAR;
	label->t = t;
	label->x.var.sb = 0;
	label->x.var.offset = 0;
	label->x.var.sclass = 0;
	label->x.var.dim_copied = 0;
	label->x.var.redop = 0;
	l = NEW(&tempic,idlist);
	l->id = label;
	l->next = statictemplist;
	statictemplist = l;
	return(label);
}
/* change a number into a statement label */
struct tentry *makelabel(i)
int i;
{
	char buf[20];
	struct tentry *label;
	if (i == 0) return (0);
	sprintf(buf,"l#%d\0",i);
	label = oldsym(thisunit,buf);
	while (label && (label->class != CLCONST || 
		strcmp(label->symbol,buf) != 0 ||
		label->t->x.type.elemtype != TYLABEL))
		label = label->next;
	if (label) return (label);
	label = newsym(&pupass1,thisunit,buf,unitname);
	label->class = CLCONST;
	label->t= labeltype;
	return(label);
}
struct tentry *dolabel(i,db,def)
int i, def;
struct text *db;
{
	struct tentry *lbl;
	struct looplbls *l;
	lbl = makelabel(i);
	if (dotab) {
		/* mark as defined in loop */
		for (l = dotab->labels; l; l = l->next)
			if (l->table == lbl) {
				if (def) l->inloop = 1;
				break;
			}
		if (l == 0) {
			l = NEW(&tempic,looplbls);
			l->next = dotab->labels;
			dotab->labels = l;
			l->table = lbl;
			l->ltext = 0;
			l->inloop = def;
		}
		if (db) {
			db->next = l->ltext;
			l->ltext = db;
		}
	}
	return(lbl);
}
struct tentry *deflabel(i,db)
int i;
struct text *db;
{
#ifdef debugdeflabel
	printf("deflabel %d ", i);
	if (db) printf("at %s", db->loopseg->text + db->offset);
	printf("\n");
#endif
	return(dolabel(i,db,1));
}
struct tentry *reflabel(i,db)
int i;
struct text *db;
{
	return(dolabel(i,db,0));
}

struct idlist *idcons(id, rest)
struct idlist *rest;
struct tentry *id;
{
	struct idlist *l;
	l = NEW(&tempic,idlist);
	l->id = id;
	l->next = rest;
#ifdef debugidcons
	printf("idcons adds %s to list\n",id->symbol);
#endif
	return(l);
}
struct stseg *newsegappend(newend)
struct stseg *newend;
{
	struct stseg *seg;
	seg = NEW(&stmts,stseg);
	seg->newstmt = 1;
	seg->f = 0;
	seg->lineno = 0;
	newend->next = seg;
	seg->prev = newend;
	seg->next = 0;
	return(seg);
}

struct stseg *chkseglen(seg,s,t)
char *s, *t;
struct stseg *seg;
{	
	seg->stlen = t - seg->text;
	if (seg->stlen + strlen(s) > 64) {
#ifdef debugendexec
		printf("endexec: %s\n", seg->text);
#endif
		seg = newsegappend(seg);
		seg->newstmt = 0;
	} 
	return(seg);
}
subst(seg,off,s1,s2)
struct stseg *seg;
int off;
char *s1, *s2;
{
	int ls;
	char *t, *s, *ct, *cs;
	t = seg->text + off;
#ifdef debugsubst
	seg->text[seg->stlen] = 0;
	printf("enter subst %s to %s in %s\n",s1,s2,t);
#endif
	s = t;
	while (*s1)  {
		while (*s == ' ')s++;
#ifdef debugsubst
		if (*s != *s1 && *s != *s1 - 'a' + 'A') 
			printf("subst %c should be %c\n",s[-1],s1[-1])
#endif
		s++;
		s1++;
	}
	/* compute extra blanks to fill in */
	ls = (s - t) - strlen(s2);
	/* if negative, must move rest of line */
	if (ls < 0) {
		seg->stlen -= ls;
		ct = seg->text + seg->stlen -1;
		cs = ct + ls;
		while (cs >= s) *ct-- = *cs--;
		ls = 0;
	}
	while (*s2) *t++ = *s2++;
	while (ls-- > 0) *t++ = ' ';
#ifdef debugsubst
	t = seg->text + off;
	seg->text[seg->stlen] = 0;
	printf("exit subst %s\n",t);
#endif
}
struct chgs *buildsubst(chglist, t, lbl)
struct chgs *chglist;
struct text *t;
struct looplbls *lbl;
{
	struct chgs *c, *cl;
	c = NEW(&tempic,chgs);
	c->next = chglist;
	c->seg = t->loopseg;
	c->offset = t->offset;
	c->label = lbl;
	if (chglist && c->offset < chglist->offset) {
		for (cl = chglist; cl->next && cl->next->offset > c->offset;
			cl = cl->next);
		c->next = cl->next;
		cl->next = c;
		return(chglist);
	} else {
		return(c);
	}
}

endexec(label)
struct tentry *label;
{
	char buf[MAXLINE];
	struct slicelist *lr;
	struct idlist *l, *l2;
	struct tentry *t;
	/* reset temporary ids */
	if (!stfunid) {
		didexecutable = 1;
		if (!execinsert) {
			execinsert = lexbase->stsegptr->prev;
			while (!execinsert->newstmt) execinsert = execinsert->prev;
	}	}
	tempidno = 0;
	stfunid = 0;
#ifdef debugendexec
	printf("endexec: checking for stmt %s\n%s\n",
				dostack ? dostack->label->symbol+2 : 0,
				lexbase->stsegptr->prev->text);
#endif
	while (enddoflag || (label && dostack && dostack->label == label)) {
		enddoflag = 0;
		/* end of do loop processing */
		if (dotab && dotab->doindex == dostack->doindex) {
			/* check for unreferenced variables */
			for (l2 = 0, l = dotab->lastlocals; l; l = l->next) {
				if (l->id->referenced) {
				    l2 = l;
				    if (l->id->t->x.type.elemtype == TYPTR &&
	(t = l->id->t->x.type.subtype.ptr.to)->x.type.elemtype == TYARRAY &&
	(t = t->x.type.subtype.array.index)->x.type.elemtype == TYPRANGE &&
					!t->x.type.subtype.parmrange.highexp){
					sprintf(buf,
	"lastlocal variable %s has unknown dimension--doacross ignored", 
					l->id->symbol);
					yywarn(buf);
					dotab = 0;
					dostack = dostack->next;
					return;
				    }
				} else {
					sprintf(buf,
					"%s appears in doacross clause but not in loop", 
						l->id->symbol);
					yywarn(buf);
					if (l2) l2->next = l->next;
					else dotab->lastlocals = l->next;
			}	}
			for (l2 = 0, l = dotab->locals; l; l = l->next) {
				if (l->id->referenced) {
				    l2 = l;
				    if (l->id->t->x.type.elemtype == TYPTR &&
	(t = l->id->t->x.type.subtype.ptr.to)->x.type.elemtype == TYARRAY &&
	(t = t->x.type.subtype.array.index)->x.type.elemtype == TYPRANGE &&
					!t->x.type.subtype.parmrange.highexp){
					sprintf(buf,
	"local variable %s has unknown dimension--doacross ignored", 
					l->id->symbol);
					yywarn(buf);
					dotab = 0;
					dostack = dostack->next;
					return;
				    }
				} else {
					sprintf(buf,
					"%s appears in doacross clause but not in loop", 
						l->id->symbol);
					yywarn(buf);
					if (l2) l2->next = l->next;
					else dotab->locals = l->next;
			}	}
			for (l2 = 0, l = dotab->share; l; l = l->next) {
				if (l->id->referenced) l2 = l;
				else {
					/* Since KAP puts dimension variables in the
						loop to avoid bozo warnings from a
						early version of this preprocessor,
						we won't complain about extra shared
						dimension variables */
#ifdef COMPLAINSHARED
					if (l->id->class != CLVAR ||
						l->id->x.var.dim_copied == 0) {
					  sprintf(buf,
					  "%s appears in doacross clause but not in loop",
						l->id->symbol);
					  yywarn(buf);
					}
#endif COMPLAINSHARED
					if (l2) l2->next = l->next;
					else dotab->share = l->next;
			}	}
			for (l2 = 0, l = dotab->order; l; l = l->next) {
				if (l->id->referenced) l2 = l;
				else {
					sprintf(buf,
					"%s appears in doacross clause but not in loop", 
						l->id->symbol);
					yywarn(buf);
					if (l2) l2->next = l->next;
					else dotab->order = l->next;
			}	}
			for (l2 = 0, l = dotab->locks; l; l = l->next) {
				if (l->id->referenced) l2 = l;
				else {
					sprintf(buf,
					"%s appears in doacross clause but not in loop", 
						l->id->symbol);
					yywarn(buf);
					if (l2) l2->next = l->next;
					else dotab->locks = l->next;
			}	}
			/* check for reduction operations found */
			for (lr = dotab->reduce; lr; lr = lr->next) {
				if (lr->id->t->x.type.elemtype == TYPTR &&
	(t = lr->id->t->x.type.subtype.ptr.to)->x.type.elemtype == TYARRAY &&
	(t = t->x.type.subtype.array.index)->x.type.elemtype == TYPRANGE &&
					!t->x.type.subtype.parmrange.highexp){
					sprintf(buf,
	"reduction variable %s has unknown dimension--doacross ignored", 
					lr->id->symbol);
					yywarn(buf);
					dotab = 0;
				}
				if (lr->id->x.var.redop == 0) {
	sprintf(buf,"did not find reduction operator for %s--doacross ignored", 
					lr->id->symbol);
					yywarn(buf);
					dotab = 0;
		}	}	}
		if (dotab && dotab->doindex == dostack->doindex) {
			do_doacross();
		}
		dostack = dostack->next;
}	}
/* make all chunk variables shared if not already */
share_chunk_vars()
{
	struct idlist *l, *l2;
	if (dotab->chunk) doexpid(dotab->chunk);
	for (l = dotab->stfuns; l; l = l->next) {
#ifdef debugchkstfun
		printf("endexec sees stfun %s: %s\n",
			l->id->symbol, l->id->x.func.textseg->text);
#endif
		for (l2 = l->id->x.func.lhsids; l2; l2 = l2->next)
			idlocalsearch(l2->id,0);
	}
}
/* remove m_get_numprocs and m_get_myid from local list */
remove_m_intrinsics()
{
	struct idlist *l, *l2;
	for (l2 = 0,l = dotab->locals;l;l = l->next) {
		if (strcmp(l->id->symbol,"m_get_myid") == 0 ||
			strcmp(l->id->symbol,"m_get_numprocs")==0) {
			if (l2) {
				l2->next = l->next;
			} else {
				dotab->locals = l->next;
			}
		} else {
			l2 = l;
}	}	}
/* end of doacross processing */
do_doacross()
{
	char buf[MAXLINE];
	struct text *dotxt;
	struct looplbls *lbls;
	struct idlist *l;
	struct stseg *seg;
	share_chunk_vars();
	remove_m_intrinsics();
#ifdef debugendexec
	printf("endexec: begin end of do processing for stmt\n%s\n",
		lexbase->stsegptr->prev->text);
	printf("locals...");
	for (l = dotab->locals;l;l = l->next)
		printf("%s ",l->id->symbol);
	printf("\nshared...");
	for (l = dotab->share;l;l = l->next)
		printf("%s ",l->id->symbol);
	printf("\nlabels...");
	for (lbls = dotab->labels;lbls;lbls=lbls->next) {
		printf("%s ",lbls->table->symbol);
		for (dotxt = lbls->ltext; dotxt; dotxt = dotxt->next)
			printf("(%d,%d),",
				dotxt->loopseg->lineno, 
				dotxt->offset);
	}
	printf("\nloop index text...");
	for (dotxt = dotab->itext; dotxt; dotxt = dotxt->next)
		printf("(%d,%d),", 
			dotxt->loopseg->lineno, dotxt->offset);
	printf("\n");
#endif
	for (lbls = dotab->labels;lbls;lbls=lbls->next) {
		if (lbls->inloop == 0) {
			sprintf(buf,"label %s is a loop exit",
				lbls->table->symbol+2);
			yyerror(buf);
		}
	}
	schedlbl = genlbl();
	exitlbl = genlbl();
	/* free up seg links */
	newstart = newend = seg = dotab->dostmt;
	enddobodyseg = lexbase->stsegptr->prev;
	dobodyseg = seg->next;
	seg->next = 0;
	dobodyseg->prev = 0;
	enddobodyseg->next = 0;
	lexbase->stsegptr->prev = 0;
#ifdef debugendexec
	printf("endexec: %s\n",seg->text);
#endif
	build_call_stmt();
	build_external_stmt();
	build_subroutine_stmt();
	build_type_stmts();
	build_parameter_stmt();
	build_dim_stmt();
	output_stmt_funcs();
	setup_chunk_control();
	init_reduction_vars();
	build_done_test();
	insert_original_loop();
	insert_end_do();
	finish_lastlocals();
	finish_reductions();
	insert_end_stmt();
	append_subroutine();
}
/* build call statement */
build_call_stmt()
{
	register char *t, *s;
	char buf[MAXLINE];
	struct stseg *seg;
	struct idlist *l;
	struct slicelist *lr;
	struct tentry *ty, *ixty;
	int i;

	seg = newend;
	/* now delete the continuation cards */
	while (seg->next && !seg->next->newstmt) {
		seg->next = seg->next->next;
		if (seg->next) seg->next->prev = seg;
	}
	seg->stlen = 6;
	for (t=pnamebuf,s = dostack->label->symbol; *t++ = *s++;);
	pnamebuf[0]='d';
	pnamebuf[1]='o';
	t[-1] = '%';
	if (unitname) {
		for (s = unitname->symbol; *t++ = *s++;);
	} else 	*t = 0;
	makevar("m_fork",inttype);
	dynamicarrays = 0;
	for (l = dotab->locals; l; l = l->next) {
	    if (l->id->class == CLVAR) {
		ty = l->id->t;
		while (ty->x.type.elemtype == TYPTR) 
			ty = ty->x.type.subtype.ptr.to;
		if (ty->x.type.size == 0 &&
			ty->x.type.elemtype != TYFUNC) {
			/* character * (*) argument */
			dynamicarrays += 2;
		} else {
		    while (ty->x.type.elemtype == TYARRAY) {
			ixty = ty->x.type.subtype.array.index;
			if(ixty->x.type.elemtype == TYPRANGE){
			    dynamicarrays += 
				basetype(l->id)->x.type.elemtype == TYUSCHAR ? 2 : 1;
			    break;
			}
			ty = ty->x.type.subtype.array.base;
	}   }	}   }
	for (lr = dotab->reduce; lr; lr = lr->next) {
	    if (lr->id->class == CLVAR) {
		ty = lr->id->t;
		while (ty->x.type.elemtype == TYPTR) 
			ty = ty->x.type.subtype.ptr.to;
		if (ty->x.type.size == 0 &&
			ty->x.type.elemtype != TYFUNC) {
			/* character * (*) argument */
			dynamicarrays += 2;
		} else {
		    while (ty->x.type.elemtype == TYARRAY) {
			ixty = ty->x.type.subtype.array.index;
			if(ixty->x.type.elemtype == TYPRANGE){
			    dynamicarrays += 
				basetype(lr->id)->x.type.elemtype == TYUSCHAR ? 2 : 1;
			    break;
			}
			ty = ty->x.type.subtype.array.base;
	}   }	}   }
	t = seg->text+seg->stlen;
	numargs = 0;		/* fix for watson mailbug */
	if (dostack->do1->class != CLCONST) numargs++;
	if (dostack->do2->class != CLCONST) numargs++;
	if (dostack->do3->class != CLCONST) numargs++;
	for (l = dotab->share; l; l = l->next) {
		ty = l->id->t;
		if (ty->x.type.elemtype == TYPTR &&
		  ty->x.type.subtype.ptr.to->x.type.elemtype == TYFUNC) {
			numargs += 2;
		} else {
			i = basetype(l->id)->x.type.elemtype;
			numargs += 
				(i == TYUSCHAR || i == TYCHAR) ? 2 : 1;
	}	}
	for (l = dotab->lastlocals; l; l = l->next) {
		ty = l->id->t;
		if (ty->x.type.elemtype == TYPTR &&
		  ty->x.type.subtype.ptr.to->x.type.elemtype == TYFUNC) {
			numargs += 2;
		} else {
			i = basetype(l->id)->x.type.elemtype;
			numargs += 
				(i == TYUSCHAR || i == TYCHAR) ? 2 : 1;
	}	}
	for (l = dotab->locks; l; l = l->next)
		numargs += 1;
	for (l = dotab->order; l; l = l->next)
		numargs += 1;
	for (lr = dotab->reduce; lr; lr = lr->next)
		numargs += basetype(lr->id)->x.type.elemtype == TYUSCHAR ? 2 : 1;
	sprintf(t," if (m_fork(%s", dynamicarrays ? "m_dynalloc" : pnamebuf);
	while (*t++);
	t--;
	if (!reversed_args && dynamicarrays) {
		*t++ = ',';
	  	s = pnamebuf;
	  	newend = seg = chkseglen(seg,s,t);
	  	t = seg->text + seg->stlen;
	  	while (*t++ = *s++);
	  	t--;
		sprintf(t, ",%d,%d", numargs, dynamicarrays);
		while (*t++);
		t--;
	}
	if (dostack->do1->class != CLCONST) {
		sprintf(t,",%s", dostack->do1->symbol);
		while (*t)t++;
	}
	if (dostack->do2->class != CLCONST) {
		sprintf(t,",%s", dostack->do2->symbol);
		while (*t)t++;
	}
	if (dostack->do3->class != CLCONST) {
		sprintf(t,",%s", dostack->do3->symbol);
		while (*t)t++;
	}
	for (l = dotab->share; l; l = l->next) {
#ifdef debugendexec
		dumptentry(l->id);
#endif
		*t++ = ',';
		s = l->id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
	}
	for (l = dotab->lastlocals; l; l = l->next) {
#ifdef debugendexec
		dumptentry(l->id);
#endif
		*t++ = ',';
		s = l->id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
	}
	for (l = dotab->locks; l; l = l->next) {
#ifdef debugendexec
		dumptentry(l->id);
#endif
		*t++ = ',';
		s = l->id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
	}
	for (l = dotab->order; l; l = l->next) {
#ifdef debugendexec
		dumptentry(l->id);
#endif
		*t++ = ',';
		s = l->id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
	}
	for (lr = dotab->reduce; lr; lr = lr->next) {
#ifdef debugendexec
		dumptentry(lr->id);
#endif
		*t++ = ',';
		s = lr->id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
	}
	seg->stlen = t - seg->text;
	/* append sizes of dynamic arrays */
	if (dynamicarrays) {
	  	needsdynalloc = 1;
	  	for (l = dotab->locals; l; l = l->next) {
			append_dynamic_size(l->id);
		}
	  	for (lr = dotab->reduce; lr; lr = lr->next) {
			append_dynamic_size(lr->id);
		}
		seg = newend;
		t = seg->text + seg->stlen;
		/* output number of dynamic arrays */
		if (reversed_args) {
	  		s = buf;
	  		sprintf(s,",%d,%d,", dynamicarrays, numargs);
	  		newend = seg = chkseglen(seg,s,t);
	  		t = seg->text + seg->stlen;
	  		while (*t++ = *s++);
	  		t--;
	  		s = pnamebuf;
	  		newend = seg = chkseglen(seg,s,t);
	  		t = seg->text + seg->stlen;
	  		while (*t++ = *s++);
	  		t--;
	}	}
	sprintf(buf,").ne.0) stop 'do %s %s'", dostack->label->symbol+2,
		unitname ? unitname->symbol : " ");
	s = buf;
	seg = newend;
	t = seg->text + seg->stlen;
	newend = seg = chkseglen(seg,s,t);
	t = seg->text + seg->stlen;
	while (*t++ = *s++);
	t--;
	seg->stlen = t - seg->text;
#ifdef debugendexec
	printf("endexec: %s\n",seg->text);
#endif
	/* connect call statement to next statement */
	lexbase->stsegptr->prev = seg;
	seg->next = lexbase->stsegptr;
}
/* construct external statement */
build_external_stmt()
{
	struct stseg *seg;
	char *t;
	seg = NEW(&stmts,stseg);
	seg->newstmt = 1;
	if (specinsert) {
		seg->f = 0;
		seg->lineno = 0;
		seg->prev = specinsert->prev;
		if (seg->prev) seg->prev->next = seg;
		specinsert->prev = seg;
		seg->next = specinsert;
	}
	t = seg->text;
	sprintf(t,"      external do%s%%%s",
		dostack->label->symbol+2, unitname ? unitname->symbol : 0);
	while (*t++);
	t--;
	seg->stlen = t - seg->text;
#ifdef debugendexec
	printf("endexec: %s\n",seg->text);
#endif
}
build_subroutine_stmt()
{
	struct stseg *seg;
	char *t, *s, lp_or_comma;
	struct idlist *l;
	struct slicelist *lr;
	/* construct subroutine statement */
	newstart = newend = seg = NEW(&stmts,stseg);
	seg->newstmt = 1;
	seg->f = 0;
	seg->lineno = 0;
	t = seg->text;
	for (s = "      subroutine "; *t++ = *s++;);
	t--;
	for (s = pnamebuf; *t++ = *s++;);
	t--;
	lp_or_comma = '(';
	if (dostack->do1->class != CLCONST) {
		sprintf(t,"%ci%%1", lp_or_comma);
		lp_or_comma = ',';
		while (*t)t++;
	}
	if (dostack->do2->class != CLCONST) {
		sprintf(t,"%ci%%2", lp_or_comma);
		lp_or_comma = ',';
		while (*t)t++;
	}
	if (dostack->do3->class != CLCONST) {
		sprintf(t,"%ci%%3", lp_or_comma);
		lp_or_comma = ',';
		while (*t)t++;
	}
	for (l = dotab->share; l; l = l->next) {
		*t++ = lp_or_comma;
		lp_or_comma = ',';
		s = l->id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
	}
	for (l = dotab->lastlocals; l; l = l->next) {
		*t++ = lp_or_comma;
		lp_or_comma = ',';
		s = l->id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t[-1] = '%';
	}
	for (l = dotab->locks; l; l = l->next) {
		*t++ = lp_or_comma;
		lp_or_comma = ',';
		s = l->id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
	}
	for (l = dotab->order; l; l = l->next) {
		*t++ = lp_or_comma;
		lp_or_comma = ',';
		s = l->id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
	}
	for (lr = dotab->reduce; lr; lr = lr->next) {
		*t++ = lp_or_comma;
		lp_or_comma = ',';
		s = lr->id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t[-1] = '%';
	}
	seg->stlen = t - seg->text;
	/* append names of dynamic arrays */
	if (dynamicarrays) {
		for (l = dotab->locals; l; l = l->next) {
			append_dynamic_name(l->id);
		}
		for (lr = dotab->reduce; lr; lr = lr->next) {
			append_dynamic_name(lr->id);
	}	}
	seg = newend;
	t = seg->text + seg->stlen;
	*t++ = ')';
	*t = 0;
	seg->stlen = t - seg->text;
#ifdef debugendexec
	printf("endexec: %s\n",seg->text);
#endif
	/* start with integer i%1, i%2, i%3, etc */
	newend = seg = newsegappend(newend);
	t = seg->text;
	for(s = "\tinteger i%1,i%2,i%3,i%chunk,m_get_myid,m_get_numprocs,i%end";
		*t++ = *s++;);
	t--;
	seg->stlen = t - seg->text;
#ifdef debugendexec
	printf("endexec: %s\n",seg->text);
#endif
	if (dotab->chunk) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		for(s = "\tinteger m_next, i%step";
			*t++ = *s++;);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
	}
}
/* output type declarations */
build_type_stmts()
{
	char *t, *s;
	struct idlist *l;
	struct slicelist *lr;
	int i, lasti;
	struct stseg *seg;
	struct tentry *id;
	if (dotab->locks || dotab->order || dotab->reduce || 
		(dotab->lastlocals && dotab->chunk)) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"      integer ");
		while (*t++);
		t--;
		for (l = dotab->locks; l; l = l->next) {
			s = l->id->symbol;
			newend = seg = chkseglen(seg,s,t);
			t = seg->text + seg->stlen;
			while (*t++ = *s++);
			t[-1] = ',';
		}
		for (l = dotab->order; l; l = l->next) {
			s = l->id->symbol;
			newend = seg = chkseglen(seg,s,t);
			t = seg->text + seg->stlen;
			while (*t++ = *s++);
			t[-1] = ',';
		}
		if (dotab->lastlocals && dotab->chunk) {
			s = "last%i";
			newend = seg = chkseglen(seg,s,t);
			t = seg->text + seg->stlen;
			while (*t++ = *s++);
			t[-1] = ',';
		}
		if (dotab->reduce) {
	s = "is%1,is%2,is%3,is%4,is%5,is%6,is%7";
			newend = seg = chkseglen(seg,s,t);
			t = seg->text + seg->stlen;
			while (*t++ = *s++);
			t[-1] = ',';
		}
		t--;
		*t = 0;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
	}
	/* output type declarations */
	lasti = 0;
	for (i = 2; i < TYINC; i++) {
		for (l = dotab->lastlocals; l; l = l->next) {
			if ((id = basetype(l->id)) &&
				id->x.type.elemtype == i) {
				if (lasti != i) 
					start_new_type_stmt(id);
				lasti = i;
				add_type_var(i, l->id, "");
		}	}
		for (l = dotab->lastlocals; l; l = l->next) {
			if ((id = basetype(l->id)) &&
				id->x.type.elemtype == i) {
				add_type_var(i, l->id, "%");
		}	}
		for (l = dotab->locals; l; l = l->next) {
			if ((id = basetype(l->id)) &&
				id->x.type.elemtype == i) {
				if (lasti != i) 
					start_new_type_stmt(id);
				lasti = i;
				add_type_var(i, l->id, "");
		}	}
		for (l = dotab->share; l; l = l->next) {
			if ((id = basetype(l->id)) &&
				id->x.type.elemtype == i) {
				if (lasti != i) 
					start_new_type_stmt(id);
				lasti = i;
				add_type_var(i, l->id, "");
		}	}
		for (lr = dotab->reduce; lr; lr = lr->next) {
			if ((id = basetype(lr->id)) &&
				id->x.type.elemtype == i) {
				if (lasti != i) 
			start_new_type_stmt(id);
				lasti = i;
				add_type_var(i, lr->id, "");
		}	}
		for (lr = dotab->reduce; lr; lr = lr->next) {
			if ((id = basetype(lr->id)) &&
				id->x.type.elemtype == i) {
				add_type_var(i, lr->id, "%");
		}	}
		if (lasti) remove_last_comma(newend);
}	}
/* output parameter statements */
build_parameter_stmt()
{
	struct idlist *l;
	struct stseg *seg;
	char *s, *t, buf[MAXLINE];
	if (dostack->do1->class == CLCONST) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t, "      parameter (i%%1 = %d)", 
			dostack->do1->x.const.intval);
		while (*t) t++;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
	}
	if (dostack->do2->class == CLCONST) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t, "      parameter (i%%2 = %d)", 
			dostack->do2->x.const.intval);
		while (*t) t++;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
	}
	if (dostack->do3->class == CLCONST) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t, "      parameter (i%%3 = %d)", 
			dostack->do3->x.const.intval);
		while (*t) t++;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
	}
	for (l = dotab->locals; l; l = l->next) {
		if (l->id->class == CLCONST &&
		  l->id->t->x.type.elemtype != TYFUNC) 
			break;
	}
	if (l) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		for (s = "      parameter ("; *t++ = *s++;);
		t[-1] = ' ';
		for (; l; l = l->next) 
		  if (l->id->class == CLCONST &&
			l->id->t->x.type.elemtype != TYFUNC) {
			s = l->id->symbol;
			newend = seg = chkseglen(seg,s,t);
			t = seg->text + seg->stlen;
			while (*t++ = *s++);
			t[-1] = '=';
			switch (l->id->t->x.type.elemtype) {
			case TYINT:
				sprintf(buf,"%d", l->id->x.const.intval);
				break;
			case TYREAL:
			case TYDOUBLE:
				sprintf(buf,"%20.16g", l->id->x.const.realval);
				break;
			case TYUSCHAR:
				sprintf(buf,"%s", l->id->x.const.charval);
				break;
			case TYCOMPLEX:
				sprintf(buf,"(%20.16g,%20.16g)",
				  l->id->x.const.cmpxval.real,
				  l->id->x.const.cmpxval.imag);
				break;
			default:
				buf[0] = '?';
				buf[1] = 0;
				break;
			}
			s = buf;
			newend = seg = chkseglen(seg,s,t);
			t = seg->text + seg->stlen;
			while (*t++ = *s++);
			t[-1] = ',';
		}
		t[-1] = ')';
		*t = 0;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
}	}
build_dim_stmt()
{
	int lasti;
	struct stseg *seg;
	struct idlist *l;
	struct slicelist *lr;
	/* output dimension statement */
	lasti = 0;
	for (l = dotab->lastlocals; l; l = l->next) {
		lasti = add_dim_var(lasti, l->id, "");
	}
	for (l = dotab->lastlocals; l; l = l->next) {
		lasti = add_dim_var(lasti, l->id, "%");
	}
	for (l = dotab->locals; l; l = l->next) {
		lasti = add_dim_var(lasti, l->id, "");
	}
	for (l = dotab->share; l; l = l->next) {
		lasti = add_dim_var(lasti, l->id, "");
	}
	for (lr = dotab->reduce; lr; lr = lr->next) {
		lasti = add_dim_var(lasti, lr->id, "");
	}
	for (lr = dotab->reduce; lr; lr = lr->next) {
		lasti = add_dim_var(lasti, lr->id, "%");
	}
	if (lasti) {
		remove_last_comma(newend);
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
}	}
/* output statement functions referenced in loop */
output_stmt_funcs()
{
	struct idlist *l;
	struct stseg *seg, *oldseg;
	char *t, *s;
	for (l = dotab->stfuns; l; l = l->next) {
		for (oldseg = l->id->x.func.textseg; oldseg; 
			oldseg = oldseg->next) {
			newend = seg = newsegappend(newend);
			t = seg->text;
			s = oldseg->text;
			s[oldseg->stlen] = 0;
			while (*t++ = *s++);
			seg->stlen = oldseg->stlen;
#ifdef debugchkstfun
			printf("endexec stfun: %s\n",seg->text);
#endif
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
			if (oldseg->next && 
				oldseg->next->newstmt) 
				break;
}	}	}
/* set up chunk control and i*/
setup_chunk_control()
{
	char *t;
	struct stseg *seg;
	if (dotab->chunk) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\t\ti%%chunk = i%%3 ");
		while (*t++);
		t--;
		if (dotab->unroll > 1) {
			sprintf(t,"* %d", dotab->unroll);
			while (*t++);
			t--;
		}
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\ti%%step = max(1,");
		while (*t++);
		t--;
		t = exprsym(dotab->chunk, t);
		sprintf(t,") * i%%chunk");
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\t%s = m_get_myid() * i%%step + i%%1",
			dotab->doindex->symbol);
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
	} else {
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\ti%%end = i%%2");
		while (*t++);
		t--;
		if (dotab->unroll > 1) {
			sprintf(t," - %d * i%%3", dotab->unroll - 1);
			while (*t++);
			t--;
		}
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\t%s = m_get_numprocs()", dotab->doindex->symbol);
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\tif (%s .gt. 1) then", dotab->doindex->symbol);
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\t\ti%%chunk = i%%3 * %s", dotab->doindex->symbol);
		while (*t++);
		t--;
		if (dotab->unroll > 1) {
			sprintf(t,"* %d", dotab->unroll);
			while (*t++);
			t--;
		}
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		/* set up i */
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\t\t%s = i%%1 + i%%3 * m_get_myid()", 
			dotab->doindex->symbol);
		while (*t++);
		t--;
		if (dotab->unroll > 1) {
			sprintf(t,"* %d", dotab->unroll);
			while (*t++);
			t--;
		}
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\telse");
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\t\ti%%chunk = i%%3");
		while (*t++);
		t--;
		if (dotab->unroll > 1) {
			sprintf(t,"* %d", dotab->unroll);
			while (*t++);
			t--;
		}
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\t\t%s = i%%1", dotab->doindex->symbol);
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\tendif");
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
	}
}
/* initialize reduction variables */
init_reduction_vars()
{
	struct slicelist *lr;
	int i, mustlock;
	struct triple *trip;
	struct tentry *id;
	char buf[MAXLINE], buf2[MAXLINE], *t, *s;
	struct stseg *seg;
#ifdef INDEX_REDUCTION
	/* check for lock needed before picking up index reduction values */
	mustlock = 0;
	for (lr = dotab->reduce; lr; lr = lr->next) {
		switch (lr->id->x.var.redop & -TYINC) {
		case OPMINX:
		case OPMINXL:
		case OPMAXX:
		case OPMAXXL:
			mustlock = 1;
			break;
		default:	break;
	}	}
	if (mustlock) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\tcall m_lock");
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
	}
#endif
	for (lr = dotab->reduce; lr; lr = lr->next) {
		buf[0] = 0;
		id = lr->id->t;
		while (id->x.type.elemtype == TYPTR)
			id = id->x.type.subtype.ptr.to;
		if (id->x.type.elemtype == TYARRAY) {
		    exitlbl = genlbl();
		    /* single element in one dimensions */
		    for (i = 1, trip = lr->triples[i-1]; i <= 7; i++, trip++) {
			trip = lr->triples[i-1];
			if (trip->lower && !trip->upper) {
				newend = seg = newsegappend(newend);
				t = seg->text;
				sprintf(t,"\tis%%%d = ", i);
				while (*t) t++;
				t = exprsym(trip->lower, t);
		        	seg->stlen = t - seg->text;
#ifdef debugendexec
				printf("endexec: %s\n",seg->text);
#endif
		    }   }
		    for (i = 1; id->x.type.elemtype == TYARRAY;
			    i++, id = id->x.type.subtype.array.base) {
			trip = lr->triples[i-1];
			if (!trip->lower) {
			    newend = seg = newsegappend(newend);
			    t = seg->text;
			    substdim(buf2, id->x.type.subtype.array.index->symbol);
		            sprintf(t,"\tdo %s is%%%d = %s",
				exitlbl->symbol+2, i, buf2);
		            for (; *t;t++) 
			      if (*t == ':') *t = ',';
		            seg->stlen = t - seg->text;
#ifdef debugendexec
			    printf("endexec: %s\n",seg->text);
#endif
			} else if (trip->upper) { /* two or three do parameters */
			    newend = seg = newsegappend(newend);
			    t = seg->text;
		            sprintf(t,"\tdo %s is%%%d = ",
				exitlbl->symbol+2, i);
			    while (*t) t++;
			    t = exprsym(trip->lower, t);
			    *t++ = ',';
			    t = exprsym(trip->upper, t);
			    if (trip->inc) {
				*t++ = ',';
				t = exprsym(trip->inc, t);
			    }
		            seg->stlen = t - seg->text;
#ifdef debugendexec
			    printf("endexec: %s\n",seg->text);
#endif
		    }	}
		    s = buf;
		    *s++ = '(';
		    while (--i > 0) {
		        sprintf(s,"is%%%d,",i);
		        while (*s)s++;
		    }
		    s[-1] = ')';
		}
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"%s      %s%s = ",
			exitlbl ? exitlbl->symbol+2 : 0,
			lr->id->symbol,buf);
		while (*t++);
		t--;
		switch (lr->id->x.var.redop & (-TYINC)) {
		case OPMULT:
			*t++ = '1';
			break;
		case OPAND:
			for (s = ".true."; *s;)*t++ = *s++;
			break;
		case OPOR:
		case OPXOR:
			for (s = ".false."; *s;)*t++ = *s++;
			break;
		case OPMINX:
		case OPMINXL:
		case OPMAXX:
		case OPMAXXL:
		case OPMIN:
		case OPMAX:
			for (s = lr->id->symbol; *s;) *t++ = *s++;
			*t++ = '%';
			for (s = buf; *s;) *t++ = *s++;
			break;
		case OPADD:
		case OPSUB:
		default:
			*t++ = '0';
			break;
		}
		*t = 0;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		exitlbl = genlbl();
	}
#ifdef INDEX_REDUCTION
	if (mustlock) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\tcall m_unlock");
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
	}
#endif
}
/* now for the if test */
build_done_test()
{
	char *t;
	struct stseg *seg;
	if (dotab->chunk) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		if (dostack->do3->class == CLCONST) {
			sprintf(t,"%s\tif(%s .%s. i%%2) goto %s",
				schedlbl->symbol+2,
				dotab->doindex->symbol, 
				(dostack->do3->x.const.intval > 0) ?
					"gt" : "lt",
				exitlbl->symbol+2);
		} else {
			sprintf(t,"%s\tif((i%%3.gt.0.and.%s.gt.i%%2).or.",
				schedlbl->symbol+2,
				dotab->doindex->symbol);
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
			newend = seg = newsegappend(newend);
			seg->newstmt = 0;
			t = seg->text;
			sprintf(t,"(i%%3.lt.0.and.%s.lt.i%%2)) goto %s",
				dotab->doindex->symbol,
				exitlbl->symbol+2);
		}
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		/* now for the if test */
		newend = seg = newsegappend(newend);
		t = seg->text;
		if (dostack->do3->class == CLCONST) {
			sprintf(t,"\t\ti%%end = %s(i%%2",
			  (dostack->do3->x.const.intval > 0) ? "min" : "max");
			while (*t++);
			t--;
			if (dotab->unroll > 1) {
				sprintf(t," - %d * i%%3", dotab->unroll - 1);
				while (*t++);
				t--;
			}
			sprintf(t,",%s + i%%step - i%%chunk)",
				dotab->doindex->symbol);
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
		} else {
			sprintf(t, "\tif(i%%3.gt.0)then");
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
			newend = seg = newsegappend(newend);
			t = seg->text;
			sprintf(t,"\t\ti%%end = min(i%%2");
			while (*t++);
			t--;
			if (dotab->unroll > 1) {
				sprintf(t," - %d * i%%3", dotab->unroll - 1);
				while (*t++);
				t--;
			}
			sprintf(t,",%s + i%%step - i%%chunk)", dotab->doindex->symbol);
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
			newend = seg = newsegappend(newend);
			t = seg->text;
			sprintf(t,"\telse");
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
			newend = seg = newsegappend(newend);
			t = seg->text;
			sprintf(t,"\t\ti%%end = max(i%%2");
			while (*t++);
			t--;
			if (dotab->unroll > 1) {
				sprintf(t," - %d * i%%3", dotab->unroll - 1);
				while (*t++);
				t--;
			}
			sprintf(t,",%s + i%%step - i%%chunk)", dotab->doindex->symbol);
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
			newend = seg = newsegappend(newend);
			t = seg->text;
			sprintf(t,"\tendif");
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
}	}	}
/* put in DO */
insert_original_loop()
{
	struct stseg *seg;
	char *t, *s, buf[MAXLINE];
	int i, iroll;
	struct looplbls *lbls;
	struct text *dotxt;
	struct output *exp;
	struct chgs *chglist, *chgitem;

	newend = seg = newsegappend(newend);
	t = seg->text;
	chunklbl = genlbl();
	sprintf(t,"      do %s, %s = %s, i%%end, i%%chunk",
		chunklbl->symbol+2, dotab->doindex->symbol,
		dotab->doindex->symbol);
	while (*t++);
	t--;
	seg->stlen = t - seg->text;
#ifdef debugendexec 
	printf("endexec: %s\n",seg->text);
#endif 
	/* insert do body */
	/* i means we have "odd" iterations */
	i = dotab->unroll > 1;
	if (i) {
		if (dostack->do1->class == CLCONST &&
			dostack->do2->class == CLCONST &&
			dostack->do3->class == CLCONST &&
			dostack->do3->x.const.intval) {
			i = (dostack->do2->x.const.intval -
				dostack->do1->x.const.intval +
				dostack->do3->x.const.intval)/
				(dostack->do3->x.const.intval*
				dotab->unroll);
			i = dostack->do1->x.const.intval + i * 
				dotab->unroll *
				dostack->do3->x.const.intval -
				dostack->do2->x.const.intval;
			if (dostack->do3->x.const.intval < 0)
				i = -i;
			i = i <= 0;
		}
	}
	iodd = i;
	if (i) {
		tripid = makevar("tc%", inttype);
		/* generate trip count assignment */
		seg = NEW(&stmts,stseg);
		seg->newstmt = 1;
		seg->f = 0;
		seg->lineno = 0;
		seg->prev = lexbase->stsegptr->prev;
		seg->prev->next = seg;
		seg->next = lexbase->stsegptr;
		lexbase->stsegptr->prev = seg;
		t = seg->text;
		exp = binop(OPDIV, binop(OPADD, binop(OPSUB,
	unop(OPLOAD,dostack->do3->class == CLCONST ? INCONST : INADDR,dostack->do3),
	unop(OPLOAD,dostack->do1->class == CLCONST ? INCONST : INADDR,dostack->do1)),
	unop(OPLOAD,dostack->do2->class == CLCONST ? INCONST : INADDR,dostack->do2)),
		  binop(OPMULT,
	unop(OPLOAD,dostack->do3->class == CLCONST ? INCONST : INADDR,dostack->do3),
		    expint(dotab->unroll)) );
		s = exprsym(exp,buf);
		*s = 0;
		sprintf(t,"\ttc%% = max(0,%s)", buf);
		while(*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		/* generate do to finish off loop */
		seg = NEW(&stmts,stseg);
		seg->newstmt = 1;
		seg->f = 0;
		seg->lineno = 0;
		seg->prev = lexbase->stsegptr->prev;
		seg->prev->next = seg;
		seg->next = lexbase->stsegptr;
		lexbase->stsegptr->prev = seg;
		t = seg->text;
		exp = binop(OPMULT, unop(OPLOAD,INCONST,dostack->do3),
			expint(dotab->unroll));
		s = exprsym(exp,buf);
		*s = 0;
		sprintf(t,"\tdo %s, %s = %s+tc%%*%s,%s,%s",
			dostack->label->symbol+2,
			dotab->doindex->symbol,
			dostack->do1->symbol,
			buf,
			dostack->do2->symbol,
			dostack->do3->symbol);
		while(*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		/* restore original do body */
		dobodyseg->prev = lexbase->stsegptr->prev;
		lexbase->stsegptr->prev->next = dobodyseg;
		enddobodyseg->next = lexbase->stsegptr;
		lexbase->stsegptr->prev = enddobodyseg;
	}
	if (i || dotab->unroll > 1) {
	  /* duplicate do body unroll times */
	  for (iroll = 0; iroll < dotab->unroll; iroll++){
	    if (iroll) {
		sprintf(buf,"(%s+(%d))",
			dotab->doindex->symbol,iroll);
		for (lbls = dotab->labels; lbls; 
			lbls = lbls->next) {
			lbls->newlabel = genlbl();
		}
	    } else {
		chglist = 0;
		for (dotxt = dotab->itext; dotxt;
			dotxt = dotxt->next)
			chglist = buildsubst(chglist,dotxt,0);
		for (lbls = dotab->labels; lbls; 
			lbls = lbls->next) {
		  for (dotxt = lbls->ltext; dotxt;
			dotxt = dotxt->next) 
			chglist =buildsubst(chglist,dotxt,lbls);
		}
	    }
	    for (oldseg = dobodyseg; oldseg != enddobodyseg->next; 
		oldseg = oldseg->next) {
	      seg = NEW(&stmts,stseg);
	      *seg = *oldseg;
	      seg->f = 0;
	      seg->lineno = 0;
	      seg->prev = newend;
	      newend->next = seg;
	      seg->next = 0;
	      newend = seg;
	      if (iroll) {
		for (chgitem = chglist; chgitem; 
			chgitem=chgitem->next) {
		  if (chgitem->seg == oldseg) {
		    if (chgitem->label)
		      subst(seg, chgitem->offset,
			chgitem->label->table->symbol+2,
			chgitem->label->newlabel->symbol+2);
		    else
		      subst(seg, chgitem->offset,
			dotab->doindex->symbol, buf);
	      } } }
#ifdef debugendexec
	printf("endexec unroll: %s\n",seg->text);
#endif
	    }
	  }
	} else {
		/* insert orginial do body */
		newend->next = dobodyseg;
		dobodyseg->prev = newend;
		newend = enddobodyseg;
}	}
/* insert end of do continue statement */
insert_end_do()
{
	struct stseg *seg;
	char *t, *s;
	newend = seg = newsegappend(newend);
	t = seg->text;
	sprintf(t,"%s\tcontinue", chunklbl->symbol+2);
	while(*t++);
	t--;
	seg->stlen = t - seg->text;
#ifdef debugendexec
	printf("endexec: %s\n",seg->text);
#endif
	if (dotab->chunk) {
		if (dotab->lastlocals) {
			newend = seg = newsegappend(newend);
			t = seg->text;
			sprintf(t,"\tlast%%i = %s",
				dotab->doindex->symbol);
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
		}
		/* now for i = m_next()*i%chunk+(i%1-i%chunk) */
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\t%s = (m_next() + m_get_numprocs()-1)",
			dotab->doindex->symbol);
		while (*t++);
		t--;
		sprintf(t,"* i%%step + i%%1");
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		/* insert goto statement */
		newend = seg = newsegappend(newend);
		t = seg->text;
		for (s = "       goto "; *t++ = *s++;);
		t--;
		for (s = schedlbl->symbol+2; *t++ = *s++;);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
}	}
/* assign last values to lastlocal variables */
finish_lastlocals()
{
	struct stseg *seg;
	char *t;
	struct idlist *l;
	if (dotab->lastlocals) {
		if (dotab->chunk) {
			newend = seg = newsegappend(newend);
			t = seg->text;
			sprintf(t,"%s\t%s = last%%i", 
				exitlbl ? exitlbl->symbol+2 : " ",
				dotab->doindex->symbol);
			exitlbl = 0;	
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
		}
		newend = seg = newsegappend(newend);
		t = seg->text;
		if (dostack->do3->class == CLCONST) {
	sprintf(t,"%s\tif (%s-i%%chunk+i%%3 .%s. i%%2) then",
				exitlbl ? exitlbl->symbol+2 : " ",
				dostack->doindex->symbol,
				dostack->do3->x.const.intval > 0 ? "gt" : "lt");
		} else {
sprintf(t,"%s\tif ((i%%3 .gt. 0 .and. %s-i%%chunk+i%%3 .gt. i%%2) .or.",
				exitlbl ? exitlbl->symbol+2 : " ",
				dostack->doindex->symbol);
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
			newend = seg = newsegappend(newend);
			t = seg->text;
			seg->newstmt = 0;
sprintf(t,"(i%%3 .lt. 0 .and. %s-i%%chunk+i%%3 .lt. i%%2))then",
				dostack->doindex->symbol);
		}
		exitlbl = 0;	
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		for (l = dotab->lastlocals; l; l = l->next) {
			newend = seg = newsegappend(newend);
			t = seg->text;
			sprintf(t,"\t    %s%% = %s  %s", 
				l->id->symbol, l->id->symbol,
				l->id == dotab->doindex ? "-i%chunk+i%3" : " ");
			while (*t++);
			t--;
			seg->stlen = t - seg->text;
#ifdef debugendexec
			printf("endexec: %s\n",seg->text);
#endif
		}
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"\tendif");
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
}	}
/* add local reduction variables to shared variables */
finish_reductions()
{
	struct stseg *seg;
	struct slicelist *lr;
	struct triple *trip;
	char buf[MAXLINE], buf2[MAXLINE], *t, *s, *opsym();
	struct tentry *id;
	int i;
	if (dotab->reduce) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"%s\tcall m_lock", exitlbl ? exitlbl->symbol+2 : " ");
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
		/* we already used exit label */
		exitlbl = 0;	
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
		/* for all reduction variables */
		for (lr = dotab->reduce; lr; lr = lr->next) {
			buf[0] = 0;
			id = lr->id->t;
			while (id->x.type.elemtype == TYPTR)
				id = id->x.type.subtype.ptr.to;
			if (id->x.type.elemtype == TYARRAY) {
			    exitlbl = genlbl();
			    /* single element in one dimensions */
			    for (i = 1, trip = lr->triples[i-1]; i <= 7; i++, trip++) {
				trip = lr->triples[i-1];
				if (trip->lower && !trip->upper) {
					newend = seg = newsegappend(newend);
					t = seg->text;
					sprintf(t,"\tis%%%d = ", i);
					while (*t) t++;
					t = exprsym(trip->lower, t);
			        	seg->stlen = t - seg->text;
#ifdef debugendexec
					printf("endexec: %s\n",seg->text);
#endif
			    }   }
			    for (i = 1; id->x.type.elemtype == TYARRAY;
				    i++, id = id->x.type.subtype.array.base) {
				trip = lr->triples[i-1];
				if (!trip->lower) {
				    newend = seg = newsegappend(newend);
				    t = seg->text;
				    substdim(buf2, 
					id->x.type.subtype.array.index->symbol);
			            sprintf(t,"\tdo %s is%%%d = %s",
					exitlbl->symbol+2, i, buf2);
			            for (; *t;t++) 
				      if (*t == ':') *t = ',';
			            seg->stlen = t - seg->text;
#ifdef debugendexec
				    printf("endexec: %s\n",seg->text);
#endif
				} else if (trip->upper) { /* two or three do parameters */
				    newend = seg = newsegappend(newend);
				    t = seg->text;
			            sprintf(t,"\tdo %s is%%%d = ",
					exitlbl->symbol+2, i);
				    while (*t) t++;
				    t = exprsym(trip->lower, t);
				    *t++ = ',';
				    t = exprsym(trip->upper, t);
				    if (trip->inc) {
					*t++ = ',';
					t = exprsym(trip->inc, t);
				    }
			            seg->stlen = t - seg->text;
#ifdef debugendexec
				    printf("endexec: %s\n",seg->text);
#endif
			    }	}
			    s = buf;
			    *s++ = '(';
			    while (--i > 0) {
			        sprintf(s,"is%%%d,",i);
			        while (*s)s++;
			    }
			    s[-1] = ')';
			}
			newend = seg = newsegappend(newend);
			t = seg->text;
			sprintf(t,"%s      %s%%%s = ",
				exitlbl ? exitlbl->symbol+2 : 0,
				lr->id->symbol,buf);
			while (*t) t++;
			switch (lr->id->x.var.redop & -TYINC) {
#ifdef INDEX_REDUCTION
			case OPMAXX:
				finish_index(lr, 0, 0);
				break;
			case OPMAXXL:
				finish_index(lr, 1, 1);
				break;
			case OPMINX:
				finish_index(lr, 1, 0);
				break;
			case OPMINXL:
				finish_index(lr, 0, 1);
				break;
#endif
			case OPMAX:
			case OPMIN:
				t = opsym(lr->id->x.var.redop, t);
				sprintf(t, "(%s%%%s, %s%s)",
					lr->id->symbol,buf, lr->id->symbol,buf);
				while (*t) t++;
				seg->stlen = t - seg->text;
#ifdef debugendexec
				printf("endexec: %s\n",seg->text);
#endif
				break;
			default:
				newend = seg = newsegappend(newend);
				t = seg->text;
				sprintf(t,"%s      %s%%%s = ",
					exitlbl ? exitlbl->symbol+2 : 0,
					lr->id->symbol,buf);
				while (*t) t++;
				sprintf(t,"%s%%%s ", lr->id->symbol,buf);
				while (*t) t++;
				t = opsym(lr->id->x.var.redop, t);
				sprintf(t," %s%s", lr->id->symbol,buf);
				while (*t) t++;
				seg->stlen = t - seg->text;
#ifdef debugendexec
				printf("endexec: %s\n",seg->text);
#endif
				break;
			}
			/* we already used exit label */
			exitlbl = 0;	
		}
		/* clear lock */
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"      call m_unlock");
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
		newend = seg;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
}	}
#ifdef INDEX_REDUCTION
/* finish minindex, maxindex reductions */
finish_index(lr, flip, last)
struct slicelist *lr;
int flip, last;
{
	struct stseg *seg, *s1, *s2, *s3;
	char *t, *s, *e1, *e2, *e3, e1buf[MAXLINE], e2buf[MAXLINE];
	struct tentry *l1, *l2, *l3, *v1, *v2, *v3, *v4, *valt, *vred;
	struct text *db;
	struct input *in;
	struct slicelist *rs;
	int piggyredop;
	/* implement "piggyback reduction" scheme or 
		"function of opposing local variable" scheme */
	for (in = lr->id->x.var.redexp->inputs; in; in = in->next) {
		db = in->x.result->producer->optext;
		switch (in->innumb) {
		case 0:	s1 = db->loopseg;
			e1 = s1->text + (db->offset - s1->offset);
			v1 = in->x.result->producer->inputs->x.table;
			break;
		case 1:	s2 = db->loopseg;
			e2 = s2->text + (db->offset - s2->offset);
			v2 = in->x.result->producer->inputs->x.table;
			break;
		case 2:	s3 = db->loopseg;
			e3 = s3->text + (db->offset - s3->offset);
			v3 = in->x.result->producer->inputs->x.table;
			break;
		case 3:
			v4 = in->x.result->producer->inputs->x.table;
			break;
		default:	break;
	}	}
	if (v3 == lr->id) {
		valt = v4;
		vred = v1;
	} else {
		valt = v3;
		vred = v2;
	}
	piggyredop = (flip ^ last) ? OPMIN : OPMAX;
	for (rs = dotab->reduce; rs; rs = rs->next) {
		if (rs->id == vred && vred->x.var.redop == piggyredop) break;
	}
	if (rs) {
		sprintf(e1buf, "%s", vred->symbol);
		sprintf(e2buf, "%s", vred->symbol);
		t = last ? e1buf : e2buf;
		while (*t) t++;
		*t++ = '%';
		*t++ = 0;
		seg = newend;
		t = seg->text + 6;
		sprintf(t,"continue");
		while (*t) t++;
		seg->stlen = t - seg->text;
	} else {
		/* assign opposing local variable to shared value of reduction vqriable */
		seg = newend;
		t = seg->text + 6;
		sprintf(t,"%s = %s%%", valt->symbol, lr->id->symbol);
		while (*t) t++;
		seg->stlen = t - seg->text;
		if (!e1 || !e2 || !e3) {
			yyerror("internal error generating reduction code");
			return;
		}
		/* collect argument 1 */
		e1++;
		for (t = e1buf; e1 != e2; ) {
			if (e1 == s1->text + s1->stlen) {
				s1 = s1->next;
				e1 = s1->text;
			}
			*t++ = *e1++;
		}
		*t++ = 0;
		/* collect argument 2 */
		e2++;
		for (t = e2buf; e2 != e3; ) {
			if (e2 == s2->text + s2->stlen) {
				s2 = s2->next;
				e2 = s2->text;
			}
			*t++ = *e2++;
		}
		*t++ = 0;
	}
	l1 = genlbl();
	l2 = genlbl();
	l3 = genlbl();
	newend = seg = newsegappend(newend);
	t = seg->text;
	sprintf(t,"      if(%s - %s) %s, %s, %s",
		e1buf, e2buf, flip ? l1->symbol+2 : l3->symbol+2, l2->symbol+2, 
		flip ? l3->symbol+2 : l1->symbol+2);
	while (*t++);
	seg->stlen = t - seg->text;
	if (dostack->do3->class != CLCONST) {
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"%s        if (i%%3 .gt. 0) then", l2->symbol+2);
		while (*t++);
		seg->stlen = t - seg->text;
		l2 = 0;
	} else {
		if (dostack->do3->x.const.intval < 0) last = !last;
	}
	newend = seg = newsegappend(newend);
	t = seg->text;
	sprintf(t,"%s      if(%s%% %s %s) goto %s",
		l2 ? l2->symbol + 2 : " ",
		lr->id->symbol, 
		last ? ".ge." : ".le.",
		lr->id->symbol, 
		l3->symbol+2);
	while (*t++);
	seg->stlen = t - seg->text;
	if (dostack->do3->class != CLCONST) {
		last = !last;
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"        else");
		while (*t++);
		seg->stlen = t - seg->text;
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"      if(%s%% %s %s) goto %s",
			lr->id->symbol, 
			last ? ".le." : ".ge.",
			lr->id->symbol, 
			l3->symbol+2);
		while (*t++);
		seg->stlen = t - seg->text;
		newend = seg = newsegappend(newend);
		t = seg->text;
		sprintf(t,"        endif");
		while (*t++);
		seg->stlen = t - seg->text;
	}
	newend = seg = newsegappend(newend);
	t = seg->text;
	sprintf(t,"%s        %s%% = %s", l1->symbol+2, 
		lr->id->symbol, lr->id->symbol);
	while (*t++);
	seg->stlen = t - seg->text;
	newend = seg = newsegappend(newend);
	t = seg->text;
	sprintf(t,"%s        continue", l3->symbol+2);
	while (*t++);
	seg->stlen = t - seg->text;
}
#endif
/* insert end statement */
insert_end_stmt()
{
	struct stseg *seg;
	char *t;
	newend = seg = newsegappend(newend);
	t = seg->text;
	sprintf(t,"%s      end", exitlbl ? exitlbl->symbol+2 :0);
	while (*t++);
	t--;
	seg->stlen = t - seg->text;
#ifdef debugendexec
	for (seg = newstart; seg; seg = seg->next)
		printf("endexec: %s\n",seg->text);
#endif
}
/* append to subroutines to be output */
append_subroutine()
{
	struct stseg *seg;
	char *t;
	if (lastdosub)
		lastdosub->next = newstart;
	else	firstdosub = newstart;
	newstart->prev = lastdosub;
	lastdosub = newend;
	/* all done */
	dotab = 0;
	/* fix up outer do with same label */
	if (!iodd && dostack->next && 
		dostack->next->label == dostack->label) {
		seg = NEW(&stmts,stseg);
		seg->newstmt = 1;
		seg->f = 0;
		seg->lineno = 0;
		seg->prev = lexbase->stsegptr->prev;
		seg->next = lexbase->stsegptr;
		if (lexbase->stsegptr->prev) 
			lexbase->stsegptr->prev->next = seg;
		lexbase->stsegptr->prev = seg;
		t = seg->text;
		sprintf(t,"%s\tcontinue", dostack->label->symbol+2);
		while (*t++);
		seg->stlen = t - seg->text;
#ifdef debugendexec
		printf("endexec: %s\n",seg->text);
#endif
}	}
/* start new type statement */
start_new_type_stmt(id)
struct tentry *id;
{
	struct stseg *seg;
	char *t, *s;
#ifdef debugendexec
	printf("begin new type...\n");
	dumptentry(id);
#endif
	newend = seg = newsegappend(newend);
	t = seg->text;
	for (s = "      "; *t++ = *s++;);
	t[-1] = ' ';
	for (s = id->symbol; *t++ = *s++;);
	t[-1] = ' ';
	seg->stlen = t - seg->text;
}
add_type_var(i, id, suffix)
struct tentry *id;
char *suffix;
int i;
{
	struct stseg *seg;
	struct tentry *ty;
	char *t, *s;
	seg = newend;
	t = seg->text + seg->stlen;
	/* begin fix for mailbug 4537 */
	if ((ty = basetype(id)) && ty->x.type.elemtype == i) {
		/* end mailbug 4537 fix */
		s = id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
		for (s = suffix; *t++ = *s++;);
		t--;
		/* begin fix for mailbug 4537 */
		if (i == TYCHAR || i == TYUSCHAR || i == TYLOG) {
			if (ty->x.type.size > 0) {
				sprintf(t,"*%d", ty->x.type.size >>3);
				while (*t++);
			} else {
				for(s = "*(*)"; *t++ = *s++;);
			}
			t--;
		}
		/* end mailbug 4537 fix */
		*t++ = ',';
	}
	seg->stlen = t - seg->text;
#ifdef debugendexec
	printf("endexec: %s\n",seg->text);
#endif
}
start_dim_stmt()
{
	struct stseg *seg;
	char *t, *s;
#ifdef debugendexec
	printf("begin dimension...\n");
#endif
	/* start dimension statement */
	newend = seg = newsegappend(newend);
	t = seg->text;
	for (s = "      dimension "; *t++ = *s++;);
	t--;
	seg->stlen = t - seg->text;
}
int add_dim_var(lasti, id, suffix)
int lasti;
struct tentry *id;
char *suffix;
{
	struct stseg *seg;
	char *t, *s;
	struct tentry *idtype;
#ifdef debugendexec
	dumptentry(id);
	dumptentry(id->t);
#endif
	idtype = id->t;
	while (idtype->x.type.elemtype == TYPTR)
		idtype = idtype->x.type.subtype.ptr.to;
	while (idtype->x.type.elemtype == TYFUNC)
		idtype = idtype->x.type.subtype.func.resulttype;
	if (idtype->x.type.elemtype == TYARRAY) {
		if (!lasti) start_dim_stmt();
		seg = newend;
		s = id->symbol;
		t = seg->text + seg->stlen;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
		for (s = suffix; *t++ = *s++;);
		t[-1] = ' ';
		seg->stlen = t - seg->text;
		outputdim(idtype);
		seg = newend;
		t = seg->text + seg->stlen;
		t[-1] = ',';
		seg->stlen = t - seg->text;
		return(1);
	}
	return(lasti);
}
remove_last_comma(seg)
struct stseg *seg;
{
	char *t;
	t = seg->text + seg->stlen - 1;
	if (*t == ',') *t = ' ';
}
append_dynamic_name(id)
struct tentry *id;
{
	struct stseg *seg;
	struct tentry *ty, *ixty;
	int dynamic;
	char *s, *t;
	seg = newend;
	ty = id->t;
	while (ty->x.type.elemtype == TYPTR) 
		ty = ty->x.type.subtype.ptr.to;
	dynamic = ty->x.type.size == 0 && 
	ty->x.type.elemtype != TYFUNC;
	while (ty->x.type.elemtype == TYARRAY) {
		ixty = ty->x.type.subtype.array.index;
		if (ixty->x.type.elemtype == TYPRANGE){
			dynamic = 1;
			break;
		}
		ty = ty->x.type.subtype.array.base;
	}
	if (dynamic) {
		t = seg->text + seg->stlen;
		*t++ = ',';
		s = id->symbol;
		newend = seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t--;
		seg->stlen = t - seg->text;
}	}
append_dynamic_size(id)
struct tentry *id;
{
	struct stseg *seg;
	struct tentry *ty, *ty2, *ixty;
	int dynamic, i;
	char buf[MAXLINE], *t, *s, *s2, buf2[MAXLINE];
	if (id->class == CLVAR) {
		seg = newend;
		ty = id->t;
		while (ty->x.type.elemtype == TYPTR) 
			ty = ty->x.type.subtype.ptr.to;
		ty2 = ty;
		dynamic = ty->x.type.size == 0 && ty->x.type.elemtype != TYFUNC;
		while (ty->x.type.elemtype == TYARRAY) {
			ixty = ty->x.type.subtype.array.index;
			if(ixty->x.type.elemtype == TYPRANGE){
				dynamic = 1;
				break;
			}
			ty = ty->x.type.subtype.array.base;
		}
		if (dynamic) {
#ifdef debugendexec
			printf("endexec dynamic array...");
			dumptentry(id);
#endif
			s = buf;
			*s++ = ',';
			if (!reversed_args) {
				/* output minus length if character array */
				ty = basetype(id);
				i = ty->x.type.elemtype;
				if (i == TYCHAR || i == TYUSCHAR) {
					if (ty->x.type.size > 0) {
						sprintf(s, "-%d,", ty->x.type.size);
					} else {
						sprintf(s,"-len(%s),", id->symbol);
			        	}
					while (*s) s++;
			}	}
			ty = ty2;
			while (ty->x.type.elemtype == TYARRAY) {
				ixty = ty->x.type.subtype.array.index;
				switch (ixty->x.type.elemtype) {
				case TYPRANGE:
					s2 = buf2;
					*s2++ = '(';
					s2 = exprsym(ixty->x.type.subtype.parmrange.highexp,s2);
					if (isconst(ixty->x.type.subtype.parmrange.lowexp)) {
	i = 1 - ixty->x.type.subtype.parmrange.lowexp->producer->inputs->x.table->x.const.intval;
						sprintf(s2, (i < 0) ? "%d" : "+%d", i);
						while (*s2) s2++;
					} else {
						*s2++ = '-';
				s2 = exprsym(ixty->x.type.subtype.parmrange.lowexp,s2);
						*s2++ = '+';
						*s2++ = '1';
					}
					*s2++ = ')';
					*s2++ = '*';
					*s2 = 0;
					substdim(s, buf2);
					while (*s) s++;
					break;
				case TYRANGE:
					sprintf(s,"%d*", ixty->x.type.subtype.subrange.high -
				 		ixty->x.type.subtype.subrange.low + 1);
					while (*s++);
					break;
				default:
					break;
				}
				ty = ty->x.type.subtype.array.base;
			}
			if (ty->x.type.size > 0) {
				sprintf(s,"%d", ty->x.type.size>>3);
			} else {
				sprintf(s, "len(%s)", id->symbol);
			}
			s = buf;
			t = seg->text + seg->stlen;
			newend = seg = chkseglen(seg,s,t);
			t = seg->text + seg->stlen;
			while (*t++ = *s++);
			t--;
			s = buf;	
			if (reversed_args) {
				/* output minus length if character array */
				ty = basetype(id);
				i = ty->x.type.elemtype;
				if (i == TYCHAR || i == TYUSCHAR) {
					if (ty->x.type.size > 0) {
						sprintf(s, ",-%d", ty->x.type.size);
					} else {
						sprintf(s,",-len(%s)", id->symbol);
					}
					s = buf;
					newend = seg = chkseglen(seg,s,t);
					t = seg->text + seg->stlen;
					while (*t++ = *s++);
					t--;
			}	}
			seg->stlen = t - seg->text;
}	}	}
outputdim(type)
struct tentry *type;
{
	struct stseg *seg;
	char *s, *t, buf[MAXLINE];
#ifdef debugendexec
	printf("outputdim...\n");
	dumptentry(type);
#endif
	seg = newend;
	t = seg->text + seg->stlen;
	while (type->x.type.elemtype == TYPTR) 
		type = type->x.type.subtype.ptr.to;
	while (type->x.type.elemtype == TYFUNC) 
		type = type->x.type.subtype.func.resulttype;
	if (type->x.type.elemtype == TYARRAY) {
		seg->stlen = t - seg->text;
		outputdim(type->x.type.subtype.array.base,seg);
		seg = newend;
		t = seg->text + seg->stlen;
		t--;
		if (t[-1] == ')') t[-1] = ',';
		else  *t++ = '(';
		/* substitute copies for variables in dimension expression */
		substdim(buf, type->x.type.subtype.array.index->symbol);
		s = buf;
		/* tack it on */
		seg = chkseglen(seg,s,t);
		t = seg->text + seg->stlen;
		while (*t++ = *s++);
		t[-1] = ')';
		*t++ = 0;
	}
	seg->stlen = t - seg->text;
	newend = seg;
}
/* substitute copies of dimension parameters in copying string */
substdim(buf, s)
char buf[MAXLINE], *s;
{
	struct tentry *id;
	char *p, c, idbuf[MAXLINE], *pid, *sid;
	for (p = buf; c = *s;) {
		if (c == '_' || c == '%' || c == '$' || (c >= 'a' && c <= 'z')) {
			for (pid = idbuf, sid = "dim%"; *sid;) *pid++ = *sid++;
			while (c == '_' || c == '%' || c == '$' || 
			    (c >= 'a' && c <= 'z') || (c >= '0' && c <= '9')) {
				*pid++ = c; 
				s++;
				c = *s;
			}
			*pid = 0;
			if ((id = oldsym(thisunit, idbuf)) && 
				id->class == CLVAR && id->x.var.dim_copied) {
				pid = idbuf;
			} else {
				pid = idbuf + 4;
			}
			while (*pid) *p++ = *pid++;
		} else {
			*p++ = c; s++;
	}	}
	*p = 0;
}
/* store do parameter expressions in static variables for now.
	This insures that the expression will be in shared memory. */
struct tentry *storedoparm(i, e, doseg, dl, dr)
int i;
char *dl, *dr;	/* left and right delimiters (dr may be nil) */
struct output *e;
struct stseg *doseg;
{
	struct tentry *temp;
	struct stseg *seg, *dlseg, *drseg, *newseg;
	char tempname[10], *t;

	if (isconst(e) && e->producer->inputs->x.table->t->x.type.elemtype != TYFUNC) {
		return(e->producer->inputs->x.table);
	}
	sprintf(tempname,"i%%%d",i);
	temp = makevar(tempname, inttype);
	seg = NEW(&stmts,stseg);
	seg->newstmt = 1;
	seg->f = 0;
	seg->lineno = 0;
	seg->prev = doseg->prev;
	seg->prev->next = seg;
	doseg->prev = seg;
	seg->next = doseg;
	t = seg->text;
	sprintf(t,"       %s = ", tempname);
	while (*t++);
	t--;
	dl++;
	if (!dr) {
		for (drseg = doseg; drseg && drseg->next && !drseg->next->newstmt;
			drseg = drseg->next);
		dr = drseg->text + drseg->stlen;
	}
	for (dlseg = doseg; dlseg &&
		&dlseg->text[MAXLINE] < dl || dlseg->text > dl;
		dlseg = dlseg->next);
	for (drseg = doseg; drseg &&
		&drseg->text[MAXLINE] < dr || drseg->text > dr;
		drseg = drseg->next);
	if (!dlseg || !drseg) {
		yyerror("INTERNAL ERROR: do parameter delimiter not found");
		t = exprsym(e, t);
	} else {
		for (; dlseg && dlseg != drseg->next;
			dlseg = dlseg->next, dl = dlseg->text) {
			*t = 0;
			dlseg->text[dlseg->stlen] = 0;
			seg->stlen = t - seg->text;
			if (seg->stlen + strlen(dl) > 64) {
				newseg = NEW(&stmts, stseg);
				newseg->newstmt = 0;
				newseg->f = 0;
				newseg->lineno = 0;
				newseg->prev = seg;
				newseg->next = seg->next;
				seg->next = newseg;
				if (newseg->next) newseg->next->prev = newseg;
				seg = newseg;
			}
			t = seg->text + seg->stlen;
			while (*dl && (dlseg != drseg || dl < dr))
				*t++ = *dl++;
	}	}
	*t = 0;
	seg->stlen = t - seg->text;
#ifdef debugstoredoparm
	printf("storedoparm: %s\n",seg->text);
#endif
	return(temp);
}
/* add doacross clause */
addclause(head, l2)
struct idlist **head, *l2;
{
	struct idlist *l;
	if (dotab->dostmt == 0) {
		if (!*head) *head = l2;
		else {
			for (l = *head; l->next; l = l->next);
			l->next = l2;
}	}	}
/* add doacross reduction clause */
addreduceclause(head, l2)
struct slicelist **head, *l2;
{
	struct slicelist *l;
	if (dotab->dostmt == 0) {
		if (!*head) *head = l2;
		else {
			for (l = *head; l->next; l = l->next);
			l->next = l2;
}	}	}

/* start processing after reading do statement */
doacrosshead(ldo,dh,e1,e2,e3,d1,d2,d3)
struct output *e1, *e2, *e3;
char *d1, *d2, *d3;
struct dotbl *dh;
struct tentry *ldo;
{
	extern struct stseg *stsegpool;
	struct stseg *seg, *callseg; 
	char *t, *s, buf[MAXLINE];
	struct idlist *l, *l2;
	struct slicelist *lr;
	int didshared;
	struct tentry *ty;

	/* watch out for nested doacrosses--ignore inner */
	if (dotab->doindex) return;
	dotab->labels = 0;
	dotab->doindex = dh->doindex;
	seg = lexbase->stsegptr->prev;
	while (!seg->newstmt) seg = seg->prev;
	dotab->dostmt = callseg = seg;
	seg = seg->prev;
	while (!seg->newstmt) {
		seg->text[0] = 'c';
		seg->text[1] = '$';
		seg->text[2] = '&';
		seg->newstmt = 1;
		seg = seg->prev;
	}
	/* we may need to insert statements */
	if (!specinsert) specinsert = seg;
	if (!execinsert) execinsert = seg;
	seg = callseg;
	if (e3->t != inttype) e3 = unop(OPCNVINT, INRESULT, e3, 0);
	if (e2->t != inttype) e2 = unop(OPCNVINT, INRESULT, e2, 0);
	if (e1->t != inttype) e1 = unop(OPCNVINT, INRESULT, e1, 0);
	dh->do3 = storedoparm(3,e3,seg,d3,0);
	dh->do2 = storedoparm(2,e2,seg,d2,d3);
	dh->do1 = storedoparm(1,e1,seg,d1,d2);
	if (!dh->label) dh->label = genlbl();
	if (basetype(dh->doindex) == 0) 
		dh->doindex->t = addbasetype(dh->doindex->t, imptype(dh->doindex));
	ty = dh->doindex->t;
	while (ty->x.type.elemtype == TYPTR) ty = ty->x.type.subtype.ptr.to;
	if (ty != inttype) {
		yywarn("do index is not integer--doacross ignored");
		dotab = 0;
		return;
	}
	/* check for errors in doacross clauses -- clear out fields */
	for (l = dotab->locals; l; l = l->next) {
		l->id->doacrossclass = 0;
		l->id->referenced = 0;
	}
	for (l = dotab->share; l; l = l->next) {
		l->id->doacrossclass = 0;
		l->id->referenced = 0;
	}
	for (l = dotab->lastlocals; l; l = l->next) {
		l->id->doacrossclass = 0;
		l->id->referenced = 0;
	}
	for (l = dotab->locks; l; l = l->next) {
		l->id->doacrossclass = 0;
		l->id->referenced = 0;
	}
	for (l = dotab->order; l; l = l->next) {
		l->id->doacrossclass = 0;
		l->id->referenced = 0;
	}
	for (lr = dotab->reduce; lr; lr = lr->next) {
		lr->id->doacrossclass = 0;
		lr->id->referenced = 0;
	}
	/* check duplicate entries */
	for (l2 = 0, l = dotab->lastlocals; l; l = l->next) {
		if (!l->id->doacrossclass) {
			l->id->doacrossclass = DOLASTLOCAL;
			l2 = l;
		} else if (l->id->doacrossclass == DOLASTLOCAL) {
			sprintf(buf,"%s appears twice in lastlocal clause", 
				l->id->symbol);
			yywarn(buf);
			if (l2) l2->next = l->next;
			else dotab->lastlocals = l->next;
		} else {
			sprintf(buf,"%s appears in multiple clauses--doacross ignored",
				l->id->symbol);
			yyerror(buf);
			dotab = 0;
			return;
	}	}
	for (l2 = 0, l = dotab->locals; l; l = l->next) {
		if (!l->id->doacrossclass) {
			l->id->doacrossclass = DOLOCAL;
			l2 = l;
		} else if (l->id->doacrossclass == DOLOCAL ||
			l->id->doacrossclass == DOLASTLOCAL) {
			sprintf(buf,"%s appears twice in local/lastlocal clause", 
				l->id->symbol);
			yywarn(buf);
			if (l2) l2->next = l->next;
			else dotab->locals = l->next;
		} else {
			sprintf(buf,"%s appears in multiple clauses--doacross ignored",
				l->id->symbol);
			yyerror(buf);
			dotab = 0;
			return;
	}	}
	for (l2 = 0, l = dotab->share; l; l = l->next) {
		if (!l->id->doacrossclass) {
			l->id->doacrossclass = DOSHARE;
			l2 = l;
		} else if (l->id->doacrossclass == DOSHARE) {
			sprintf(buf,"%s appears twice in share clause", l->id->symbol);
			yywarn(buf);
			if (l2) l2->next = l->next;
			else dotab->share = l->next;
		} else {
			sprintf(buf,"%s appears in multiple clauses--doacross ignored",
				l->id->symbol);
			yyerror(buf);
			dotab = 0;
			return;
	}	}
	for (l2 = 0, l = dotab->locks; l; l = l->next) {
		if (!l->id->doacrossclass) {
			l->id->doacrossclass = DOLOCK;
			l2 = l;
		} else if (l->id->doacrossclass == DOLOCK) {
			sprintf(buf,"%s appears twice in locks clause", l->id->symbol);
			yywarn(buf);
			if (l2) l2->next = l->next;
			else dotab->locks = l->next;
		} else {
			sprintf(buf,"%s appears in multiple clauses--doacross ignored",
				l->id->symbol);
			yyerror(buf);
			dotab = 0;
			return;
	}	}
	for (l2 = 0, l = dotab->order; l; l = l->next) {
		if (!l->id->doacrossclass) {
			l->id->doacrossclass = DOORDER;
			l2 = l;
		} else if (l->id->doacrossclass == DOORDER) {
			sprintf(buf,"%s appears twice in order clause", l->id->symbol);
			yywarn(buf);
			if (l2) l2->next = l->next;
			else dotab->order = l->next;
		} else {
			sprintf(buf,"%s appears in multiple clauses--doacross ignored",
				l->id->symbol);
			yyerror(buf);
			dotab = 0;
			return;
	}	}
	for (lr = dotab->reduce; lr; lr = lr->next) {
		if (!lr->id->doacrossclass) {
			lr->id->doacrossclass = DOREDUCTION;
			if (!lr->id->class) {
				lr->id->class = CLVAR;
				lr->id->x.var.offset = 0;
				lr->id->x.var.sb = 0;
				lr->id->x.var.sclass = 0;
				lr->id->x.var.dim_copied = 0;
				lr->id->x.var.redop = 0;
				lr->id->x.var.redexp = 0;
			} else if (lr->id->class != CLVAR) {
				sprintf(buf,"%s is not a variable--doacross ignored\n",
					lr->id->symbol);
				yyerror(buf);
				dotab = 0;
				return;
			}
		} else {
			if (lr->id->doacrossclass == DOREDUCTION) {
				sprintf(buf,
				"%s appears twice in reduction clause--doacross ignored", 
					lr->id->symbol);
			} else {
				sprintf(buf,
				"%s appears in multiple clauses--doacross ignored",
				lr->id->symbol);
			}
			yyerror(buf);
			dotab = 0;
			return;
	}	}
	dotab->doindex->referenced = 1;
	for (l = dotab->order; l; l = l->next) {
		if (l->id->t == 0) break;
		if (l->id->t != inttype)
			yyerror("section name is already used");
	}
	if (!l) {
		for (l = dotab->locks; l; l = l->next) {
			if (l->id->t == 0) break;
			if (l->id->t != inttype)
				yyerror("section name is already used");
	}	}
	if (l) {
		seg = NEW(&stmts,stseg);
		seg->newstmt = 1;
		seg->f = 0;
		seg->lineno = 0;
		seg->next = 0;
		seg->prev = specinsert->prev;
		seg->next = specinsert;
		if (seg->prev) seg->prev->next = seg;
		specinsert->prev = seg;
		t = seg->text;
		sprintf(t,"      integer ");
		while (*t++);
		t--;
		for (; l; l = l->next) {
			if (l->id->t == 0) {
				for (s = l->id->symbol; *t++ = *s++;);
				t[-1] = ',';
				l->id->t = inttype;
			} else if (l->id->t != inttype)
				yyerror("section name is already used");
		}
		for (l = dotab->locks; l; l = l->next) {
			if (l->id->t == 0) {
				for (s = l->id->symbol; *t++ = *s++;);
				t[-1] = ',';
				l->id->t = inttype;
			} else if (l->id->t != inttype)
				yyerror("section name is already used");
		}
		t--;
		*t = 0;
		seg->stlen = t - seg->text;
#ifdef debugdoacrosshead
		printf("doacrosshead: %s\n", seg->text);
#endif
	}
	/* initialize order parameters */
	for (l = dotab->order; l; l = l->next) {
		seg = NEW(&stmts,stseg);
		seg->newstmt = 1;
		seg->f = 0;
		seg->lineno = 0;
		seg->next = callseg;
		seg->prev = callseg->prev;
		if (seg->prev) seg->prev->next = seg;
		callseg->prev = seg;
		t = seg->text;
		sprintf(t,"      %s = %s", 
			l->id->symbol, dh->do1->symbol);
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugdoacrosshead
		printf("doacrosshead: %s\n", seg->text);
#endif
	}
	/* initialize locks */
	for (l = dotab->locks; l; l = l->next) {
		seg = NEW(&stmts,stseg);
		seg->newstmt = 1;
		seg->f = 0;
		seg->lineno = 0;
		seg->next = callseg;
		seg->prev = callseg->prev;
		if (seg->prev) seg->prev->next = seg;
		callseg->prev = seg;
		t = seg->text;
		sprintf(t,"      %s = 0", l->id->symbol);
		while (*t++);
		t--;
		seg->stlen = t - seg->text;
#ifdef debugdoacrosshead
		printf("doacrosshead: %s\n", seg->text);
#endif
	}
#ifdef debugdoacrosshead
	printf("lastlocals...");
	for (l = dotab->lastlocals;l;l = l->next)
		printf("%s ",l->id->symbol);
	printf("\nlocals...");
	for (l = dotab->locals;l;l = l->next)
		printf("%s ",l->id->symbol);
	printf("\nshared...");
	for (l = dotab->share;l;l = l->next)
		printf("%s ",l->id->symbol);
	printf("\norder...");
	for (l = dotab->order;l;l = l->next)
		printf("%s ",l->id->symbol);
	printf("\nlocks...");
	for (l = dotab->locks;l;l = l->next)
		printf("%s ",l->id->symbol);
	printf("\nreduction...");
	for (lr = dotab->reduce;lr;lr = lr->next)
		printf("%s ",lr->id->symbol);
	printf("\n");
#endif
}

/* create a new doacross structure so we can store into it */
struct doacross *newdoacross()
{
	struct doacross *n;
	if (dotab) {
		yywarn("nested doacross loops--inner loop ignored");
		return(dotab);
	}
	n = NEW(&pupass1,doacross);
	n->chunk = 0;		/* default is static scheduling */
	n->unroll = 1;
	n->share = n->locals = n->order = n->lastlocals = n->locks = 0;
	n->reduce = 0;
	n->doindex = 0;
	n->dostmt = 0;
	n->labels = 0;
	n->itext = 0;
	n->stfuns = 0;
	return(n);
}

enddo()
{
	char *s;
	struct stseg *seg;
	seg = lexbase->stsegptr;
	while (!seg->newstmt) {
		seg->prev->next = seg->next;
		seg->next->prev = seg->prev;
		seg = seg->prev;
	}
	sprintf(seg->text+6, "continue");
	seg->stlen = 14;
	enddoflag = 1;
}
do_while(l, e)
int l;
struct output *e;
{
	struct dotbl *t;
	t = NEW(&pupass1,dotbl);
	t->label = makelabel(l);
	t->doindex = 0;
	t->do1 = t->do2 = t->do3 = t->tripctr = 0;
	t->next = dostack;
	dostack = t;
}

struct dotbl *dohead(i,id)
int i;
struct tentry *id;
{
	struct dotbl *t;
	t = NEW(&pupass1,dotbl);
	t->label = makelabel(i);
	t->doindex = id;
	if (id->t == 0) id->t = imptype(id);
	if (dotab) idlocalsearch(id,dotab->dostmt);
	t->do1 = t->do2 = t->do3 = t->tripctr = 0;
	t->next = dostack;
	dostack = t;
	return(t);
}
idsearch(id, warn)
struct tentry *id;
int warn;
{
	struct idlist *l;
	struct slicelist *lr;
	char buf[80];

	id->referenced = 1;
	for (l = dotab->lastlocals; l; l = l->next)
		if (l->id == id) return;
	for (l = dotab->locals; l; l = l->next)
		if (l->id == id) return;
	for (l = dotab->share; l; l = l->next)
		if (l->id == id) return;
	for (lr = dotab->reduce; lr; lr = lr->next)
		if (lr->id == id) return;
	dotab->share = idcons(id,dotab->share);
	id->doacrossclass = DOSHARE;
	if (warn) {
		sprintf(buf,"assuming %s is shared", id->symbol);
		yywarn(buf);
}	}
idlocalsearch(id,warn)
struct tentry *id;
{
	struct idlist *l;
	struct slicelist *lr;
	char buf[80];

	id->referenced = 1;
	for (l = dotab->lastlocals; l; l = l->next)
		if (l->id == id) return;
	for (l = dotab->locals; l; l = l->next)
		if (l->id == id) return;
	for (l = dotab->share; l; l = l->next)
		if (l->id == id) return;
	for (lr = dotab->reduce; lr; lr = lr->next)
		if (lr->id == id) return;
	if (id->class != CLCONST) {
		dotab->lastlocals = idcons(id,dotab->lastlocals);
		id->doacrossclass = DOLASTLOCAL;
		if (warn) {
			sprintf(buf,"assuming %s is local (with last value saved)", 
				id->symbol);
			yywarn(buf);
		}
	} else {
		dotab->locals = idcons(id,dotab->locals);
		id->doacrossclass = DOLOCAL;
		if (warn) {
			sprintf(buf,"assuming %s is local ", id->symbol);
			yywarn(buf);
}	}	}
/* search for id in doacross lists, if not found--shared */
doacrossid(id,db)
struct tentry *id;
struct text *db;
{
	struct tentry *ty, *indexty;
	struct idlist *l;
	struct text *dt;
	char c;
	c = id->symbol[0];
	if ((c < 'a' || c > 'z') && c != '_' && c != '%') return;
	id->referenced = 1;
	if (dotab && dotab->dostmt) {
		ty = id->t;
		/* note: pointers to functions should be shared */
		if (ty && ty->x.type.elemtype == TYFUNC ||
			id->class == CLCONST) {
			idlocalsearch(id,0);
			if (id->x.func.funtype == STFUNC) {
#ifdef debugchkstfun
				printf("doacrossid sees stfun %s\n",
					id->symbol);
#endif
				for (l = dotab->stfuns; l; l = l->next)
					if (l->id == id) break;	/* mailbug 4536 */
				if (!l) {
					l = NEW(&tempic,idlist);
					l->next = dotab->stfuns;
					l->id = id;
					dotab->stfuns = l;
				}
				for (l = id->x.func.rhsids; l; l = l->next)
					doacrossid(l->id, 0);
			}
			return;
		}
		if (id == dotab->doindex) {
			if (db) {
				for (dt = dotab->itext; dt; dt = dt->next)
					if (dt == db) break;
				if (dt == 0) {
					db->next = dotab->itext;
					dotab->itext = db;
				}
			}
		} else {
			idsearch(id, 1);
		}
		ty = id->t;
		while (ty && ty->x.type.elemtype == TYPTR)
			ty = ty->x.type.subtype.ptr.to;
		while(ty && ty->x.type.elemtype == TYARRAY) {
			indexty = ty->x.type.subtype.array.index;
			if (indexty->x.type.elemtype == TYPRANGE) {
			  dodimid(indexty->x.type.subtype.parmrange.lowexp);
			  dodimid(indexty->x.type.subtype.parmrange.highexp);
			}
			ty = ty->x.type.subtype.array.base;
		}
	} else if (stfunid) {
		for (l = stfunid->x.func.lhsids; l; l = l->next)
			if (l->id == id) break;
		if (!l) {
			l = NEW(&tempic,idlist);
			l->next = stfunid->x.func.rhsids;
			l->id = id;
			stfunid->x.func.rhsids = l;
#ifdef debugchkstfun
			printf("stfun %s rhs id is %s\n", 
				stfunid->symbol,id->symbol);
#endif
}	}	}
/* find all variables in dimension expression and copy them */
dodimid(e)
struct output *e;
{
	struct input *in;
	struct tentry *id;
	struct stseg *seg;
	char buf[MAXLINE], *t;
	for (in = e->producer->inputs; in; in = in->next) {
		switch (in->intype) {
		case INCONST:
		case INADDR: 
			id = in->x.table;
			if (!id->class) id->class = CLVAR;
			if (id->class == CLVAR) {
				sprintf(buf,"dim%%%s", id->symbol);
				id = makevar(buf, id->t);
				if (!id->x.var.dim_copied) {
					seg = NEW(&stmts,stseg);
					seg->newstmt = 1;
					seg->f = 0;
					seg->lineno = 0;
					if (!execinsert) execinsert = dotab->dostmt;
					seg->next = execinsert;
					seg->prev = execinsert->prev;
					execinsert->prev = seg;
					seg->prev->next = seg;
					if (execinsert == specinsert) specinsert = seg;
					t = seg->text;
					sprintf(t,"\t%s = %s", buf, buf + 4);
					while (*t) t++;
					seg->stlen = t - seg->text;
					id->x.var.dim_copied = 1;
				}
				idsearch(id, 0);
			}
			break;
		case INRESULT:
			doexpid(in->x.result);
			break;
		default:
			break;
}	}	}
/* call doacrossid for all variables in an expression */
doexpid(e)
struct output *e;
{
	struct input *in;
	for (in = e->producer->inputs; in; in = in->next) {
		switch (in->intype) {
		case INCONST:
		case INADDR: 
			doacrossid(in->x.table,e->producer->optext);
			break;
		case INRESULT:
			doexpid(in->x.result);
			break;
		default:
			break;
}	}	}
/* generate a label we haven't seen yet */
struct tentry *genlbl()
{
	int i;
	char buf[20];
	struct tentry *label;

	for (i = 99999; i; i--) {
		sprintf(buf,"l#%d\0",i);
		label = oldsym(thisunit,buf);
		while (label && (label->class != CLCONST || 
			strcmp(label->symbol,buf) != 0 ||
			label->t->x.type.elemtype != TYLABEL))
			label = label->next;
		if (!label) {
			label = newsym(&pupass1,thisunit,buf,unitname);
			label->class = CLCONST;
			label->t= labeltype;
#ifdef debuggenlbl
			printf("genlbl returns %s\n",label->symbol);
#endif
			return(label);
		}
	}
	yyerror("all 99999 labels taken");
	return(0);
}
/* do order directives */
doorder(id)
struct tentry *id;
{
	char buf[80], *t;
	struct tentry *label;
	struct stseg *seg;
	struct idlist *ids;

	id->referenced = 1;
	ids = 0;
	if (dotab) for (ids = dotab->order; ids; ids = ids->next)
		if (ids->id == id) break;
	if (!ids) {
		sprintf(buf,"doacross order section name %s undefined",
			id->symbol);
		yyerror(buf);
		return;
	}
	seg = lexbase->stsegptr;
	t = seg->text;
	label = genlbl();
	sprintf(t,"%s\tif (%s .ne. %s) goto %s",
		label->symbol+2,
		dotab->doindex->symbol,
		id->symbol,
		label->symbol+2);
	while (*t++);
	t--;
	seg->stlen = t - seg->text;
}
doendorder(id)
struct tentry *id;
{
	char buf[80], *t;
	struct stseg *seg;
	struct idlist *ids;
	struct dotbl *ds;

	id->referenced = 1;
	ids = 0;
	if (dotab) for (ids = dotab->order; ids; ids = ids->next)
		if (ids->id == id) break;
	if (!ids) {
		sprintf(buf,"doacross order section name %s undefined",
			id->symbol);
		yyerror(buf);
		return;
	}
	seg = lexbase->stsegptr;
	t = seg->text;
	for (ds = dostack; ds && ds->doindex != dotab->doindex; ds = ds->next);
	if (ds != dostack) 
		yywarn("c$endorder must be executed exactly once in parallel loop");
	sprintf(t,"\t%s = %s + (%s)", id->symbol, id->symbol,
		(ds->do3->class == CLCONST) ? ds->do3->symbol : "i%3");
	while (*t++);
	t--;
	seg->stlen = t - seg->text;
}
/* do lock/unlock directives */
dolock(id)
struct tentry *id;
{
	char buf[80], *t;
	struct stseg *seg;
	struct idlist *ids;

	id->referenced = 1;
	ids = 0;
	if (dotab) for (ids = dotab->locks; ids; ids = ids->next)
		if (ids->id == id) break;
	if (!ids) {
		sprintf(buf,"doacross locked section name %s undefined",
			id->symbol);
		yyerror(buf);
		return;
	}
	seg = lexbase->stsegptr;
	t = seg->text;
	sprintf(t,"\tcall s_lock(%s)", id->symbol);
	while (*t++);
	t--;
	seg->stlen = t - seg->text;
}
dounlock(id)
struct tentry *id;
{
	char buf[80], *t;
	struct stseg *seg;
	struct idlist *ids;

	id->referenced = 1;
	ids = 0;
	if (dotab) for (ids = dotab->locks; ids; ids = ids->next)
		if (ids->id == id) break;
	if (!ids) {
		sprintf(buf,"doacross locked section name %s undefined",
			id->symbol);
		yyerror(buf);
		return;
	}
	seg = lexbase->stsegptr;
	t = seg->text;
	sprintf(t,"\tcall s_unlock(%s)", id->symbol);
	while (*t++);
	t--;
	seg->stlen = t - seg->text;
}
int countopn(op, id, exp)
struct output *exp;
struct tentry *id;
int op;
{
	struct input *in;
	int i, expop;
	i = 0;
	expop = exp->producer->x.op & (-TYINC);
	in = exp->producer->inputs; 
	if (op == expop || (op == OPADD && expop == OPSUB) ||
		(op == OPMULT && expop == OPDIV) ) {
		for (; in; in = in->next) {
			if (in->intype == INRESULT) {
				if (op == expop || in->innumb == 1) {
					i += countopn(op, id, in->x.result);
				} else { /* don't allow it elsewhere */
					i += 4 * countopn(op, id, in->x.result);
		}	}	}
	} else if (expop == OPLOAD && 
		in->intype != INRESULT &&
		in->x.table == id) 
		i = 1;
	return(i);
}

reductionop(lval,exp)
struct output *exp;
struct tentry *lval;
{
	int i, op;
	struct slicelist *l;
	struct input *in;
	
	if (!lval->class) {
		lval->class = CLVAR;
		lval->x.var.sb = 0;
		lval->x.var.offset = 0;
		lval->x.var.sclass = 0;
		lval->x.var.redop = 0;
		lval->x.var.dim_copied = 0;
	}
	if (basetype(lval) == 0) lval->t = addbasetype(lval->t, imptype(lval));
	if (dotab) {
		for (l = dotab->reduce; l; l = l->next) 
			if (l->id == lval) break;
		if (l) {
			/* lval found in reduction variable list */
			op =  exp->producer->x.op & (-TYINC);
			if (op == OPDIV) {
				op = OPMULT;
			} else if (op == OPSUB) {
				op = OPADD;
			} 
#ifdef INDEX_REDUCTION
			if (op == OPMINX || op == OPMAXX) {
				for (in = exp->producer->inputs; in; in = in->next) {
					if (in->innumb == 2 && in->intype == INRESULT &&
		in->x.result->producer->inputs->x.table == lval) break;
					if (in->innumb == 3 && in->intype == INRESULT &&
		in->x.result->producer->inputs->x.table == lval) {
						op = op == OPMINX ? OPMINXL : OPMAXXL;
						break;
				}	}
				i = in != 0;
			} else {
				i = countopn(op, lval, exp);
			}
#else INDEX_REDUCTION
				i = countopn(op, lval, exp);
#endif INDEX_REDUCTION
			if (i != 1) 
			  yyerror("assignment is ineligible for reduction");
			if (lval->x.var.redop == 0) {
				lval->x.var.redop = op;
				lval->x.var.redexp = exp->producer;
			}
			if (lval->x.var.redop != op) 
			  yyerror("inconsistent reduction operator");
		} else 	idlocalsearch(lval,1);
	}
}
/* make debug mapping of operators to source text, if requested */
struct text *maketext(s)
register char *s;
{
	register struct text *t;
	register struct stseg *seg;

	t = 0;
	if (s) {
		for (seg = lexbase->stsegptr; seg; seg = seg->prev)
			if (seg->text <= s && s < (seg->text + seg->stlen)) break;
		if (seg) {
			t = NEW(&pupass1,text);
			t->next = 0;
			t->offset = seg->offset + (s - seg->text);
			t->loopseg = seg;
	} 	}
	return(t);
}
/* make sure actual argument scalar expression is shared (static) */
/* build up argument expression list and type */
struct output *argtemp(id,arglist,startpun,arg,endpun)
struct output *arg, *arglist;
char *startpun, *endpun;
struct tentry *id;
{
	extern char *yydbv;
	extern struct stseg *readst();
	struct tentry *temp;
	struct stseg *seg, *stmtseg;
	char *t, *s, **tp, **stp, **endp, **sp;
	char buf[80];
	int i, label;
	struct input *in;
	/* argument list debug info is ending punctuation debug info */
	yydbv = endpun;
	/* statement label--return (TEMP) */
	if (!arg) return(arglist);
	arg->producer->optext = maketext(startpun);
	if (arglist) {
		if (arglist->producer->x.op != OPCOMPOSE) {
			in = NEW(&tempic,input);
			in->next = 0;
			in->innumb = 1;
			in->intype = INRESULT;
			in->x.result = arglist;
			arglist = NEW(&tempic,output);
			arglist->next = 0;
			arglist->consumers = 0;
			arglist->t = arg->t;
			arglist->producer = NEW(&tempic,operation);
			arglist->producer->inputs = in;
			arglist->producer->outputs = 0;
			arglist->producer->pred = arglist->producer->succ = 0;
			arglist->producer->x.op = OPCOMPOSE;
		}
		in = NEW(&tempic,input);
		in->next = arglist->producer->inputs;
		arglist->producer->inputs = in;
		/* arglist->t = 0; */
		in->innumb = in->next->innumb + 1;
		in->intype = INRESULT;
		in->x.result = arg;
	} else	arglist = arg;
#ifdef debugargtemp
	dumpoutput(arglist);
#endif
	return(arglist);
}
/* make field types */
struct tentry *makefields(inlist,rec)
struct tentry *rec;
struct input *inlist;
{
	struct tentry *field;
	if (!inlist) return(0);
	field = NEW(&pupass1,tentry);
	field->next = 0;
	field->class = CLTYPE;
	field->parent = rec;
	field->x.type.elemtype = TYFIELD;
	field->t = inlist->intype == INRESULT ? inlist->x.result->t :
		inlist->x.table->t;
	field->symbol = field->t->symbol;
	field->x.type.subtype.field.prev = makefields(inlist->next,rec);
	return(field);
}
/* see how long a record type name will be */
int recnamelen(in)
struct input *in;
{
	int r;
	char *s;
	if (!in->next) r = 0;
	else	r = recnamelen(in->next)+1;
	s = in->intype == INRESULT ? in->x.result->t->symbol :
		in->x.table->t->symbol;
#ifdef debugchkstfun
	printf("recnamelen string: %s\n",s);
#endif
	r += strlen(s);
	return(r);
}
/* build a record type name from fields */
char *rectypename(in, t)
char *t;
struct input *in;
{
	char *r, *s;
	if (in->next) {
		r = rectypename(in->next, t);
		*r++ = ',';
	} else r = t;
	for (s = in->intype == INRESULT ? in->x.result->t->symbol :
		in->x.table->t->symbol; *s; ) *r++ = *s++;
	return(r);
}
struct tentry *makeargtype(arg)
struct output *arg;
{
	char *t, *newsymbol;
	struct tentry *newt;
	int newlen;
	if (!arg) return(voidtype);
	if (arg->producer->x.op == OPCOMPOSE) {
		newlen = recnamelen(arg->producer->inputs) + 3;
		newsymbol = t = (*(tempic.alloc))(&tempic,newlen);
		*t++ = '(';
		t = rectypename(arg->producer->inputs, t);
		*t++ = ')';
		*t = 0;
#ifdef debugchkstfun
		printf("stfun arg symbol: %s\n",newsymbol);
#endif
		newt = oldsym(thisunit,newsymbol);
		while (newt && (newt->class != CLTYPE || 
			newt->x.type.elemtype != TYREC ||
			strcmp(newsymbol,newt->symbol) != 0))
				newt = newt->next;
		if (!newt) {
			newt = newsym(&pupass1,thisunit,newsymbol,0);
			newt->class = CLTYPE;
			newt->x.type.elemtype = TYREC;
			newt->x.type.size = 0;
			newt->x.type.subtype.rec.lastfield = 
				makefields(arg->producer->inputs, newt);
		}
#ifdef debugchkstfun
		dumptentry(newt);
#endif
		arg->t = newt;
		return(newt);
	} else	return(arg->t);
}
struct tentry *chkstfun(id,db,arg)
struct tentry *id;
struct text *db;
struct output *arg;
{
	struct stseg *seg;
	struct tentry *a, *b;
	struct idlist *l;
	struct input *in;
	/* identify statement functions */
	a = id->t;
	while (a && a->x.type.elemtype == TYPTR) a = a->x.type.subtype.ptr.to;
	if (!didexecutable && 
		(!a || a->x.type.elemtype != TYARRAY) ) {
		lexbase->sttype = kFUNCTION;
		id->class = CLCONST;
		id->x.func.funtype = STFUNC;
		seg = lexbase->stsegptr;
		while (!seg->newstmt) seg = seg->prev;
		id->x.func.textseg = seg;
		b = id->t;
		if (!b) b = imptype(id);
#ifdef debugchkstfun
		printf("stfun processing for %s. arg: \n", id->symbol);
		dumpoutput(arg);
#endif
		a = makeargtype(arg);
#ifdef debugchkstfun
		printf("stfun arg type: %s\n", a->symbol);
		printf("stfun result type: %s\n", b->symbol);
#endif
		id->t = newfuntype(a,b);
		stfunid = id;
		id->x.func.lhsids = 0;
		id->x.func.rhsids = 0;
		if (arg) {
			if (arg->producer->x.op == OPCOMPOSE) {
				for (in = arg->producer->inputs;
					in; in = in->next) {
					l = NEW(&pupass1,idlist);
			l->id = in->x.result->producer->inputs->x.table;
					l->next = id->x.func.lhsids;
					id->x.func.lhsids = l;
#ifdef debugchkstfun
					printf("stfun lhsid: %s\n",
						l->id->symbol);
#endif
				}
			} else {
				in = arg->producer->inputs;	/* mailbug 4535 */
				l = NEW(&pupass1,idlist);
				l->id = in->x.table;
				l->next = id->x.func.lhsids;
				id->x.func.lhsids = l;
#ifdef debugchkstfun
					printf("stfun lhsid: %s\n",
						l->id->symbol);
#endif
			}
		}
#ifdef debugchkstfun
		printf("stfun type: %s\n", id->t->symbol);
		printf("stfun text: %s\n", seg->text);
#endif
	} else {
		if (!a || a->x.type.elemtype != TYARRAY) 
			yyerror("missing dimension declaration");
		if (basetype(id) == 0) id->t = addbasetype(id->t, imptype(id));
	 	doacrossid(id,db);
	}
}
/* type id as function returning void */
subrcall(id, db)
struct tentry *id;
char *db;
{
	struct tentry *ty;
	struct stseg *seg;
	char *t;
	ty = id->t;
	if (!id->class) {
		if (ty->x.type.elemtype == TYPTR)
			id->class = CLVAR;
		else 	id->class = CLCONST;
	}
	while (ty->x.type.elemtype == TYPTR) ty = ty->x.type.subtype.ptr.to;
	if (ty->x.type.elemtype != TYFUNC) {
		id->t = addfun(0, id->t);
		id->x.func.funtype = EXTFUNC;
		/* add external statement if in doacross loop -- the call may disappear */
		if (dotab) {
			seg = NEW(&stmts, stseg);
			seg->newstmt = 1;
			seg->f = 0;
			seg->lineno = 0;
			seg->prev = specinsert->prev;
			seg->next = specinsert;
			if(seg->prev)
				seg->prev->next = seg;
			seg->next->prev = seg;
			sprintf(seg->text, "      EXTERNAL %s", id->symbol);
			for (t = seg->text; *t;) t++;
			seg->stlen = t - seg->text;
	}	} 
	if (basetype(id) == 0) {
		id->t = addbasetype(id->t, voidtype);
	}
	doacrossid(id, db);
#ifdef debugsubrcall
	dumptentry(id);
	dumptentry(id->t);
#endif
}
struct slicelist *slice_list_cons(a, b)
struct slicelist *a, *b;
{
	a->next = b;
	return(a);
}
struct slicelist  *slice_whole_id(id)
struct tentry *id;
{
	int i;
	struct slicelist *s;
	s = NEW(&pupass1, slicelist);
	s->next = 0;
	s->id = id;
	for (i = 0; i < 7; i++) s->triples[i] = 0;
	id->x.var.redop = 0;
	id->x.var.redexp = 0;
	return(s);
}
struct slicelist  *slice_part_id(id, s)
struct tentry *id;
struct slicelist *s;
{
	s->id = id;
	id->x.var.redop = 0;
	id->x.var.redexp = 0;
	return(s);
}
struct slicelist *start_triple(t)
struct triple *t;
{
	int i;
	struct slicelist *s;
	s = NEW(&pupass1, slicelist);
	s->next = 0;
	s->id = 0;
	for (i = 0; i < 7; i++) s->triples[i] = 0;
	s->triples[0] = t;
	return(s);
}
struct slicelist *append_triple(s, t)
struct slicelist *s;
struct triple *t;
{
	int i;
	for (i = 0; s->triples[i];) {
		if (i >= 7) {
			yyerror("too many dimensions");
			return(s);
		}
		i++;
	}
	s->triples[i] = t;
	return(s);
}
struct triple *make_triple(low, hi, inc)
struct output *low, *hi, *inc;
{
	struct triple *t;
	t = NEW(&pupass1, triple);
	t->lower = low;
	t->upper = hi;
	t->inc = inc;
	return(t);
}
