/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */

#define NINTRINSICS 105
#include <stdio.h>
#include "symbol.h"
#include "action.h"
#include "ic.h"

struct intrinsic {
	int nargs;	/* 0,1,2,4 are actual, 3 means two or more */
	struct tentry **argtype, **restype;
	char *generic, *specific;
	int opcode;	/* intermediate code opcode, if > 0 */
};

extern struct pool progtables, tempic, pupass1;
extern struct tentry *oldsym(), *newsym(), *export(), *unitname;
extern struct symindex *externals, *thisunit;

struct intrinsic intrintable[NINTRINSICS] = {
	{1,&realtype,&inttype,"int","int",OPCNVINT},
	{1,&realtype,&inttype,"int","ifix",OPCNVINT},
	{1,&doubletype,&inttype,"int","idfix",OPCNVINT},
	{1,&inttype,&realtype,"real","float",OPCNVREAL},
	{1,&doubletype,&realtype,"real","sngl",OPCNVREAL},
	{1,&complextype,&realtype,"real","_cmplx2real",OPCNVREAL},
	{1,&inttype,&realtype,"dble","_int2dble",OPCNVDOUBLE},
	{1,&realtype,&doubletype,"dble","_real2dble",OPCNVDOUBLE},
	{1,&doubletype,&doubletype,"dble","_real2dble",OPCNVDOUBLE},
	{1,&complextype,&doubletype,"dble","_cmplx2dble",OPCNVDOUBLE},
	{1,&inttype,&complextype,"cmplx","_int2cmplx",0},
	{2,&inttype,&complextype,"cmplx","_int2cmplx",0},
	{1,&realtype,&complextype,"cmplx","_real2cmplx",0},
	{2,&realtype,&complextype,"cmplx","_real2cmplx",0},
	{1,&doubletype,&complextype,"cmplx","_dble2cmplx",0},
	{2,&doubletype,&complextype,"cmplx","_dble2cmplx",0},
	{1,&complextype,&complextype,"cmplx","",0},
	{1,&chartype,&complextype,"ichar","ichar",OPCNVINT},
	{1,&inttype,&chartype,"char","char",0},
	{1,&realtype,&realtype,"aint","aint",0},
	{1,&doubletype,&doubletype,"aint","dint",0},
	{1,&realtype,&realtype,"anint","anint",0},
	{1,&doubletype,&doubletype,"anint","dnint",0},
	{1,&realtype,&inttype,"nint","nint",0},
	{1,&doubletype,&inttype,"nint","idnint",0},
	{1,&inttype,&inttype,"abs","iabs",OPABS},
	{1,&realtype,&realtype,"abs","abs",OPABS},
	{1,&doubletype,&doubletype,"abs","dabs",OPABS},
	{1,&complextype,&realtype,"abs","cabs",OPABS},
	{1,&inttype,&inttype,"mod","mod",0},
	{1,&realtype,&realtype,"mod","amod",0},
	{1,&doubletype,&doubletype,"mod","dmod",0},
	{1,&inttype,&inttype,"sign","isign",0},
	{1,&realtype,&realtype,"sign","sign",0},
	{1,&doubletype,&doubletype,"sign","dsign",0},
	{1,&inttype,&inttype,"dim","idim",0},
	{1,&realtype,&realtype,"dim","dim",0},
	{1,&doubletype,&doubletype,"dim","ddim",0},
	{2,&realtype,&doubletype,"dprod","dprod",0},
	{3,&inttype,&inttype,"max","max0",OPMAX},
	{3,&realtype,&realtype,"max","amax1",OPMAX},
	{3,&doubletype,&doubletype,"max","dmax1",OPMAX},
	{3,&inttype,&realtype,"amax0","amax0",0},
	{3,&realtype,&inttype,"max1","max1",0},
	{3,&inttype,&inttype,"min","min0",OPMIN},
	{3,&realtype,&realtype,"min","amin1",OPMIN},
	{3,&doubletype,&doubletype,"min","dmin1",OPMIN},
	{3,&inttype,&realtype,"amin0","amin0",0},
	{3,&realtype,&inttype,"min1","min1",0},
	{1,&chartype,&inttype,"len","len",0},
	{2,&chartype,&inttype,"index","index",0},
	{1,&complextype,&realtype,"aimag","aimag",0},
	{1,&complextype,&complextype,"conjg","conjg",0},
	{1,&realtype,&realtype,"sqrt","sqrt",OPSQRT},
	{1,&doubletype,&doubletype,"sqrt","dsqrt",OPSQRT},
	{1,&complextype,&complextype,"sqrt","csqrt",OPSQRT},
	{1,&realtype,&realtype,"exp","exp",OPEXP},
	{1,&doubletype,&doubletype,"exp","dexp",OPEXP},
	{1,&complextype,&complextype,"exp","cexp",OPEXP},
	{1,&realtype,&realtype,"exp","exp",OPEXP},
	{1,&doubletype,&doubletype,"exp","dexp",OPEXP},
	{1,&complextype,&complextype,"exp","cexp",OPEXP},
	{1,&realtype,&realtype,"log","alog",OPEXP},
	{1,&doubletype,&doubletype,"log","dlog",OPEXP},
	{1,&complextype,&complextype,"log","clog",OPEXP},
	{1,&realtype,&realtype,"log10","alog10",0},
	{1,&doubletype,&doubletype,"log10","dlog10",0},
	{1,&realtype,&realtype,"sin","sin",OPSIN},
	{1,&doubletype,&doubletype,"sin","dsin",OPSIN},
	{1,&complextype,&complextype,"sin","csin",OPSIN},
	{1,&realtype,&realtype,"cos","cos",OPCOS},
	{1,&doubletype,&doubletype,"cos","dcos",OPCOS},
	{1,&complextype,&complextype,"cos","ccos",OPCOS},
	{1,&realtype,&realtype,"tan","tan",OPTAN},
	{1,&doubletype,&doubletype,"tan","dtan",OPTAN},
	{1,&realtype,&realtype,"asin","asin",OPASIN},
	{1,&doubletype,&doubletype,"asin","dasin",OPASIN},
	{1,&realtype,&realtype,"acos","acos",OPACOS},
	{1,&doubletype,&doubletype,"acos","dacos",OPACOS},
	{1,&realtype,&realtype,"atan","atan",OPATAN},
	{1,&doubletype,&doubletype,"atan","datan",OPATAN},
	{2,&realtype,&realtype,"atan2","atan2",OPATAN2},
	{2,&doubletype,&doubletype,"atan2","datan2",OPATAN2},
	{1,&realtype,&realtype,"sinh","sinh",OPSINH},
	{1,&doubletype,&doubletype,"sinh","dsinh",OPSINH},
	{1,&realtype,&realtype,"cosh","cosh",OPCOSH},
	{1,&doubletype,&doubletype,"cosh","dcosh",OPCOSH},
	{1,&realtype,&realtype,"tanh","tanh",OPTANH},
	{1,&doubletype,&doubletype,"tanh","dtanh",OPTANH},
	{2,&chartype,&logtype,"lge","lge",0},
	{2,&chartype,&logtype,"lgt","lgt",0},
	{2,&chartype,&logtype,"lle","lle",0},
	{2,&chartype,&logtype,"llt","llt",0},
	{2,&inttype,&inttype,"ior","ior",0},
	{2,&inttype,&inttype,"iand","iand",0},
	{1,&inttype,&inttype,"not","not",0},
	{1,&inttype,&inttype,"inot","inot",0},
	{2,&inttype,&inttype,"ieor","ieor",0},
	{2,&inttype,&inttype,"ishft","ishft",0},
	{3,&inttype,&inttype,"ishftc","ishftc",0},
	{3,&inttype,&inttype,"ibits","ibits",0},
	{2,&inttype,&inttype,"btest","btest",0},
	{2,&inttype,&inttype,"ibset","ibset",0},
	{2,&inttype,&inttype,"ibclr","ibclr",0}
};

/* return result type iff the string s is an intrinsic function name */
struct tentry *isintrinsic(s,t)
char *s;
struct tentry *t;
{
	struct intrinsic *p;
	int i;
#ifdef debuggeneric
		fprintf(stderr,"isintrinsic search for intrinsic name %s of type %s\n", 
			s, t->symbol);
#endif
	for (p = intrintable; p < intrintable + NINTRINSICS; p++) {
		if (*(p->argtype) == t) {
			if (strcmp(p->generic, s) == 0) {
#ifdef debuggeneric
			fprintf(stderr,"isintrinsic found generic name %s of type %s\n", 
					p->specific, (*(p->restype))->symbol);
#endif
				return(*(p->restype));
			}
			if (strcmp(p->specific, s) == 0) {
#ifdef debuggeneric
			fprintf(stderr,"isintrinsic found specific name %s of type %s\n", 
					p->specific, (*(p->restype))->symbol);
#endif
				return(*(p->restype));
	}	}	}
	return(0);
}
struct tentry *generic2specific(id,t)
struct tentry *t, *id;
{
	struct intrinsic *p;
	char *name;
	name = id->symbol;
#ifdef debuggeneric
	fprintf(stderr,"generic2specific search for generic name %s of type %s\n", 
			name, t->symbol);
#endif
	for (p = intrintable; p < intrintable + NINTRINSICS; p++) {
#ifdef debuggeneric
		fprintf(stderr,"generic2specific trying specific name %s of type %s\n", 
				p->specific, (*(p->argtype))->symbol);
#endif
		if ((*(p->argtype))->x.type.elemtype == t->x.type.elemtype && 
			(strcmp(p->generic, name) == 0) || 
				strcmp(p->specific, name) == 0){
#ifdef debuggeneric
		fprintf(stderr,"generic2specific found specific name %s of type %s\n", 
				p->specific, (*(p->restype))->symbol);
#endif
			id = oldsym(thisunit,p->specific);
			if (id == 0) {
				id =newsym(&pupass1,thisunit,p->specific,0);
			}
			if (!id->class) {
				id->t = addfun(t,*(p->restype));
				id->class = CLCONST;
				id->x.func.funtype = INTRINSICFUNC;
				id->x.func.intop = p->opcode + 
					(*(p->restype))->x.type.elemtype;
			}
			return(id);
		}
	}
	return(id);
}
