/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */

# line 1 keychk.h
/*	assignment statement scan for keyword disambiguation */
/* strategy: whenever a keyword is recognized by the lex actions,
	it calls keychk to determine if the remainder of the
	statement is a valid assignment statement. If so, the
	keyword is NOT a keyword, and should be rejected.
	Keywords should be disabled until re-enabled by
		1) end of the statement or 
		2) a logical IF statement or
		3) the keyword ASSIGN(we want to pick up "TO").
		4) the implicit statement
	In addition, it is used for typed function statements to
	distinguish them from type statements at the beginning of
	the text and after END statements.

	The grammar used is the same one used in parsing with no changes.
	The action routines are not signalled by the scan. Lexical
	semantic actions do not make table entries during the scan.
*/
#define YYNEWLINE 10
extern struct lexcontext *lexbase, parsercontext, keychkcontext;

funscan()
{
	int result;
	char *kw;
	/* if already in lexical scan, return the keyword */
	if (lexbase == &keychkcontext) {
		lexbase->funok = 0;
		return(0);
	}

	/* turn keywords (except function) off */
#ifdef debugkeychk
	printf("begin funscan at line %d with keyword %s\n",
		lexbase->stsegptr->lineno, lexbase->yytext);
#endif
	lexbase->keyok = 0;
	lexbase = &keychkcontext;
	lexbase->yysptr = lexbase->yysbuf;
	lexbase->yyprevious = YYNEWLINE;
	lexbase->hconok = 1;
	lexbase->funok = 1;
	lexbase->keyok = 0;
	lexbase->assignscan = 1;
	lexbase->stsegptr = parsercontext.stsegptr;
	lexbase->inptr = parsercontext.inptr;
	lexbase->endptr = parsercontext.endptr;

	/* unput keyword */
	for (kw = parsercontext.yytext + parsercontext.yyleng;
		kw > parsercontext.yytext; ) unput(*--kw); 
	result = stidscan();
	lexbase->funok = 0; /* no more functions until END keyword */
	return(result);
}

keychk()
{
	/* if keywords aren't ok already, this isn't one */
	if (lexbase->keyok == 0) return(0);
	/* turn keywords off */
#ifdef debugkeychk
	printf("begin keychk at line %d with keyword %s\n",
		lexbase->stsegptr->lineno, lexbase->yytext);
#endif
	lexbase->keyok = 0;
	lexbase->funok = 0;
	lexbase = &keychkcontext;
	lexbase->yysptr = lexbase->yysbuf;
	lexbase->yyprevious = YYNEWLINE;
	lexbase->hconok = 1;
	lexbase->keyok = 0;
	lexbase->funok = 0;
	lexbase->assignscan = 1;
	lexbase->stsegptr = parsercontext.stsegptr;
	lexbase->inptr = parsercontext.inptr;
	lexbase->endptr = parsercontext.endptr;

	/* just unput one alphabetic character to start identifer */
	unput(parsercontext.yytext[parsercontext.yyleng-1]);
	return(stidscan());
}

stidscan() {

YYSTYPE yyv[YYMAXDEPTH]; /* where the values are stored */
int yychar; /* current input token number */
int yynerrs;  /* number of errors */
short yyerrflag;  /* error recovery flag */
#ifdef YYDEBUG
int yydebug;
#endif

	short yys[YYMAXDEPTH];
	short yyj, yym;
	register YYSTYPE *yypvt;
	register short yystate, *yyps, yyn;
	register YYSTYPE *yypv;
	register short *yyxi;

#ifdef YYDEBUG
	yydebug = 1;
#endif

	yystate = 0;
	yychar = -1;
	yynerrs = 0;
	yyerrflag = 0;
	yyps= &yys[-1];
	yypv= &yyv[-1];

 yystack:    /* put a state and value onto the stack */

#ifdef YYDEBUG
	if( yydebug  ) printf( "stidscan state %d, char 0%o\n", yystate, yychar );
#endif
		if( ++yyps> &yys[YYMAXDEPTH] ) { yyerror( "yacc stack overflow" ); return(1); }
		*yyps = yystate;
		++yypv;
		*yypv = yyval;

 yynewstate:

	yyn = yypact[yystate];

	if( yyn<= YYFLAG ) goto yydefault; /* simple state */

	if( yychar<0 ) if( (yychar=yylex())<0 ) yychar=0;
	if( (yyn += yychar)<0 || yyn >= YYLAST ) goto yydefault;

	if( yychk[ yyn=yyact[ yyn ] ] == yychar ){ /* valid shift */
		yychar = -1;
		yyval = yylval;
		yystate = yyn;
		if( yyerrflag > 0 ) --yyerrflag;
		goto yystack;
		}

 yydefault:
	/* default state action */

	if( (yyn=yydef[yystate]) == -2 ) {
		if( yychar<0 ) if( (yychar=yylex())<0 ) yychar = 0;
		/* look through exception table */

		for( yyxi=yyexca; (*yyxi!= (-1)) || (yyxi[1]!=yystate) ; yyxi += 2 ) ; /* VOID */

		while( *(yyxi+=2) >= 0 ){
			if( *yyxi == yychar ) break;
			}
		if( (yyn = yyxi[1]) < 0 ) {
#ifdef debugkeychk
			printf("stidscan finds desired statement\n");
#endif
			/* the rest of the statement was a valid assignment
				statement. 
			/* reset input position */
			lexbase = &parsercontext;
			return(0);   /* accept */
			}
		}

	if( yyn == 0 ){ /* error */
		/* error ... attempt to resume parsing */

		switch( yyerrflag ){

		case 0:   /* brand new error */

			/* don't report error--just return status */
		yyerrlab:
			++yynerrs;

		case 1:
		case 2: /* incompletely recovered error ... try again */

			yyerrflag = 3;

			/* find a state where "error" is a legal shift action */

			while ( yyps >= yys ) {
			   yyn = yypact[*yyps] + YYERRCODE;
			   if( yyn>= 0 && yyn < YYLAST && yychk[yyact[yyn]] == YYERRCODE ){
			      yystate = yyact[yyn];  /* simulate a shift of "error" */
			      goto yystack;
			      }
			   yyn = yypact[*yyps];

			   /* the current yyps has no shift onn "error", pop stack */

#ifdef YYDEBUG
			   if( yydebug ) printf( "stidscan error recovery pops state %d, uncovers %d\n", *yyps, yyps[-1] );
#endif
			   --yyps;
			   --yypv;
			   }

			/* there is no state on the stack with an error shift ... abort */

	yyabort:
#ifdef debugkeychk
			printf("stidscan not function or assignment\n");
#endif
			/* reset input position */
			lexbase = &parsercontext;
			return(1);


		case 3:  /* no shift yet; clobber input char */

#ifdef YYDEBUG
			if( yydebug ) printf( "stidscan error recovery discards char %d\n", yychar );
#endif

			if( yychar == 0 ) goto yyabort; /* don't discard EOF, quit */
			yychar = -1;
			goto yynewstate;   /* try again in the same state */

			}

		}

	/* reduction by production yyn */

#ifdef YYDEBUG
		if( yydebug ) printf("stidscan reduce %d\n",yyn);
#endif
		yyps -= yyr2[yyn];
		yypvt = yypv;
		yypv -= yyr2[yyn];
		yyval = yypv[1];
		yym=yyn;
			/* consult goto table to find next state */
		yyn = yyr1[yyn];
		yyj = yypgo[yyn] + *yyps + 1;
		if( yyj>=YYLAST || yychk[ yystate = yyact[yyj] ] != -yyn ) yystate = yyact[yypgo[yyn]];
		goto yystack;  /* stack new state and value */

	}
