/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987, 1988 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 *
 */

#include <stdio.h>
#include "pool.h"
#include "lex.h"
#include "y.tab.h"
#include "symbol.h"
/* read in a Fortran statement and parse it */
struct filetab *currentfile = 0, *newfile, outputfiletab;
struct stseg *stsegpool = 0;
extern int yylval;
extern struct symindex *thisunit;
extern struct tentry *unitname;
extern struct pool stmts, pupass1, tempic, progtables;
extern int col1d, col72, systemids;
extern struct lexcontext *lexbase, parsercontext, keychkcontext;
extern FILE *outfile;
int lastlineno = 0;
char *lastfilename = 0;
extern struct doacross *dotab;
extern struct text *yydbval;	/* really YYSTYPE */

/* preprocessor output routine */
ppout(seg)
struct stseg *seg;
{
#ifdef debugreader
	fprintf(stderr,"seg->lineno %d lastlineno %d\n",seg->lineno,lastlineno);
#endif
	if (seg->f == 0) {
		seg->lineno = outputfiletab.lineno+1;
		seg->f = &outputfiletab;
	} 
	if ((seg->lineno > (lastlineno+10)) || 
		seg->lineno <= lastlineno ||
		(seg->f->filename != lastfilename)) {
#ifdef debugreader
		fprintf(stderr,"$line %d %s\n", seg->lineno, seg->f->filename);
#endif
		if (seg->f == &outputfiletab) seg->lineno++;
		fprintf(outfile,"$line %d %s\n", 
			seg->lineno, seg->f->filename);
		outputfiletab.lineno++;
		lastlineno = seg->lineno-1;
		lastfilename = seg->f->filename;
	}
	while (seg->lineno > ++lastlineno) {
#ifdef debugreader
		fputc('\n',stdout);
#endif
		fputc('\n',outfile);
		outputfiletab.lineno++;
	}
	/* change c$doacross and c$& stuff into comments */
	if (seg->text[1] == '$' && (seg->text[0] == 'c' || seg->text[0] == 'C'))
		seg->text[1] = ' ';
#ifdef debugreader
	if (!seg->newstmt) fputs("     *",stdout);
	seg->text[seg->stlen] = 0;
	fputs(seg->text,stdout);
	fputc('\n',stdout);
#endif
	if (!seg->newstmt) fputs("     *",outfile);
	seg->text[seg->stlen] = 0;
	fputs(seg->text,outfile);
	fputc('\n',outfile);
	outputfiletab.lineno++;
}
/* recycle statement buffers */
recyclestbuf()
{
	struct stseg *p, *pp;
	extern struct doacross *dotab;
	/* recycle statement lifetime storage */
	if (!dotab) {
		(*(tempic.reset))(&tempic);
		pp = parsercontext.stsegptr->prev;
		while (pp && !(pp->newstmt)) pp = pp->prev;
		if (pp && (p = pp->prev)) {
			pp->prev = 0;
			p->next = stsegpool;
			pp = stsegpool;
			while (p->prev) {
#ifdef debugreader
			fprintf(stderr,"recycling buffer at 0x%x %s\n", p, p->text);
#endif
				p = p->prev;
			}
#ifdef debugreader
			fprintf(stderr, "recycling buffer at 0x%x %s\n", p, p->text);
#endif
			stsegpool = p;
			while (p != pp) {
#ifdef debugreader
			fprintf(stderr, "recycling buffer at 0x%x %s\n", p, p->text);
#endif
				ppout(p);
				p = p->next;
			}
		}
	}
}

int instring(c,s)
int c;
char *s;
{
	while (*s) if (c == *s++) return(1);
	return(0);
}

int span(s,c,t)
char *s, *t;
int c;
{
	while (instring(c,s)) {
		*t++ = c;
		c = getc(currentfile->file);
		currentfile->offset++;
	}
	*t = 0;
	return(c);
}

lowercase(s)
char *s;
{
	while (*s)  {
		if (*s >= 'A' && *s <= 'Z') *s += 'a' - 'A';
		s++;
	}
}

int sbreak(s,c,t)
char *s, *t;
int c;
{
	while (!instring(c,s)) {
		*t++ = c;
		c = getc(currentfile->file);
		currentfile->offset++;
	}
	*t = 0;
	return(c);
}

int strcmp(s1,s2)
char *s1, *s2;
{
	for (;*s1 && *s1 == *s2; s1++,s2++);
	return(*s1 - *s2);
}

/* read statement segment */
struct stseg *readst()
{
    register struct stseg *seg;
    register int c;
    char  *tp, *op, *key, keybuf[MAXLINE], dlim;
    int i,j,k;

#ifdef debugreader
	fprintf(stderr, "enter readst\n");
#endif
	/* allocate a statement segment */
    if (stsegpool) {
	seg = stsegpool;
	stsegpool = seg->next;
    } else 	seg = NEW(&stmts,stseg);
#ifdef debugreader
	fprintf(stderr, "stseg allocated at 0x%x\n", seg);
#endif
    do {	/* repeat if blank line */
	seg->next =  seg->prev = 0;
	seg->stlen = 0;
	/* skip comment lines */
	while ((c = getc(currentfile->file)) == 'c' || c=='C' || c=='\n' ||
		c == '$' || c == EOF ||
		c == '*' || (!col1d && (c == 'd' || c == 'D')) ) {
		if (c == EOF) {
#ifdef debugreader
			fprintf(stderr, "end of file on %s\n", 
				currentfile->filename);
#endif
			if (currentfile->prev) {
				c = '\n';
				fclose(currentfile->file);
				currentfile = currentfile->prev;
			} else break;
		}
		/* collect processor directives */
		if (c == '$') {
			tp = seg->text;
			c = sbreak(" \t\n",c,tp);
			lowercase(tp);
#ifdef debugreader
			fprintf(stderr, "processing compiler directive %s\n", tp);
#endif
			if (strcmp(tp,"$include") == 0) {
				c = span(" \t\"",c,tp);
				newfile = NEW(&progtables, filetab);
				newfile->prev = currentfile;
				c = sbreak("\" \t\n",c,newfile->filename);
				c = sbreak("\n",c,tp);
				currentfile->lineno++;
				currentfile = newfile;
				while (newfile=newfile->prev) {
				  if (strcmp(newfile->filename,
					currentfile->filename)==0) {
				    fprintf(stderr, "already including file %s \n",
					currentfile->filename);
				    for (newfile=currentfile->prev; newfile;
					newfile=newfile->prev){
					fprintf(stderr, "file %s line %d \n",
					    newfile->filename,
					    newfile->lineno-1);
					fclose(newfile->file);
				    }
				    /* forget new file */
				    exit(1);
				  }
				}
#ifdef debugreader
				fprintf(stderr, "opening include file %s\n",
					currentfile->filename);
#endif
				if ((currentfile->file = 
				  fopen(currentfile->filename,"r")) ==
				  NULL) {
				  fprintf(stderr, "can't open file %s\n",
					currentfile->filename);
				  exit(1);
				}
				currentfile->offset = 0;
				currentfile->lineno = 0;
				c = '\n';
			} else if (strcmp(tp,"$line")==0) {
				newfile = NEW(&progtables, filetab);
				newfile->prev = currentfile->prev;
				newfile->file = currentfile->file;
				c = span(" \t",c,tp);
				c = sbreak(" \t\n",c,newfile->filename);
				sscanf(newfile->filename,"%d",&newfile->lineno);
				newfile->lineno--;
				if (c != '\n') {
					c = span(" \t\"",c,tp);
					c = sbreak("\" \t\n",c,
						newfile->filename);
					c = sbreak("\n",c,tp);
				} else {
			        	for (tp = newfile->filename,
						op = currentfile->filename;
						*tp++ = *op++;) {
						if (!strcmp(".cmp",op)) {
							op = ".f";
				}	}	}
				currentfile = newfile;
			} else if (strcmp(tp,"$col72")==0) {
				col72 = 1;
			} else if (strcmp(tp,"$system")==0) {
				systemids = 1;
			} else { 
				/* copy the directive out */
				fputs(tp,outfile);
				c = sbreak("\n",c,tp);
				fputs(tp,outfile);
				fputc('\n',outfile);
				outputfiletab.lineno++;
		}	}
		/* check for c$doacross directive */
		if (c == 'C' || c == 'c') {
			seg->offset = currentfile->offset;
			tp = seg->text;
			c = sbreak(" \t\n0123456789",c,tp);
			lowercase(tp);
			if (strcmp(tp,"c$doacross") == 0 ||
				strcmp(tp,"c$") == 0 ||
				strcmp(tp,"c$&") == 0 ||
				strcmp(tp,"c$order") == 0 ||
				strcmp(tp,"c$endorder") == 0 ||
				strcmp(tp,"c$lock") == 0 ||
				strcmp(tp,"c$unlock") == 0 ||
				strcmp(tp,"c$reentrant") == 0) {
				seg->newstmt = 1;
				seg->lineno = currentfile->lineno;
				seg->f = currentfile;
				if (tp[2] == '&') {
					seg->newstmt = 0;
					tp[0] = ' ';
					tp[1] = ' ';
					tp[2] = ' ';
				}
				if(strcmp(tp,"c$") == 0) {
					tp[0] = ' ';
					tp[1] = ' ';
				}
				while (*tp++);
				tp--;
				i = tp - seg->text;
				while (c != '\n') {
					*tp++ = c;
					i++;
					if (c == '\t') {
						tp[-1] = ' ';
						for (k = i | 7; 
							i <= k; i++) {
							*tp++ = ' ';
						}
					}
					c = getc(currentfile->file);
					currentfile->offset++;
				}
				currentfile->lineno++;
				*tp = 0;
				seg->stlen = tp - seg->text;
#ifdef debugreader
	fprintf(stderr, "readst returns line %d of %s newstmt %d, length %d\n%s\n",
		seg->lineno, seg->f->filename, seg->newstmt, seg->stlen, 
		seg->text);
#endif
				return(seg);
		}	}
		if (c != '\n')
			while ((c = getc(currentfile->file)) !='\n') 
				currentfile->offset++;
			currentfile->lineno++;
	}
	currentfile->offset++;
	/* decide if new statement or continuation */
	tp = seg->text;
	op = seg->stoffset;
	i = seg->stlen;
	j = 0;
	seg->newstmt = 1;
	seg->offset = currentfile->offset;
	seg->lineno = currentfile->lineno;
	seg->f = currentfile;
	if (c == '&')  {
		seg->newstmt = 0;
		c = getc(currentfile->file);
		j++;
	} else if (c != '\t') {
		/* if we got here with 'D' or 'd'--debug line */
		if (c == 'D' || c == 'd') c = ' ';
		for (; j < 5; ){
			*op++ = j++;
			*tp++ = c;
			i++;
			c = getc(currentfile->file);
			if (c != ' ' && (c < '0' || c > '9')) break;
		}
		/* check for continuation in column 6 */
		if (j == 5 && c != '\n' && c != '\t') {
			if (c == '0') c = ' ';
			if (c != ' ') {
				seg->newstmt = 0;
				i = seg->stlen;
				tp = seg->text;
				op = seg->stoffset;
				c = getc(currentfile->file);
				j++;
	}	}	}
	/* special handling of tab before column 6 */
	if (j < 6 && c == '\t') {
		/* tab to column 6 */
		for (;i < 5;i++) {
			*tp++ = ' ';
			*op++ = j;
		}
		c = getc(currentfile->file);
		j++;
		if (c >= '0' && c <= '9') {
			if (c == '0') {
				c = ' ';
			} else {	/* continuation line */
				seg->newstmt = 0;
				i = seg->stlen;
				tp = seg->text;
				op = seg->stoffset;
				c = getc(currentfile->file);
				j++;
			}
		} else {	/* tab to column 9 */
			for (;i < 8; i++) {
				*tp++ = ' ';
				*op++ = j;
	}	}	}
	seg->f = currentfile;
	seg->offset = currentfile->offset;
	while (c != '\n') {
		if (c == EOF) c = 0;
		if (j < (col72 ? 72 : MAXLINE)) {
			*tp++ = c;
			*op++ = j;
			i++;
			if (c == '\t') {
				tp[-1] = ' ';
				for (k = i | 7; 
					i <= k; i++) {
					*tp++ = ' ';
					*op++ = j;
		}	}	}
		j++;
		if (!c) break;
		c = getc(currentfile->file);
	} 
	*tp = 0;
	currentfile->lineno++;
	currentfile->offset += j;
	seg->stlen = i;
	if (seg->newstmt) {
	    /* check for blank line--comment */
	    for (tp = seg->text, j = 0; j < i; j++) 
		if (*tp++ != ' ') break;
	    if (j < i) {
		/* check for mil-std-1753 include statement */
		tp = seg->text; 
		op = tp + seg->stlen;
		/* skip leading digits and blanks */
		while (tp < op && *tp == ' ' || (*tp >= '0' && *tp <= '9')) tp++;
		if (*tp == 'i' || *tp == 'I') {
		    for (key = keybuf; tp < op &&
				((*tp >= 'a' && *tp <= 'z') ||
				(*tp >= 'A' && *tp <= 'Z'));) {
			*key++ = *tp++;
			while (tp < op && *tp == ' ') tp++;
		    }
		    *key++ = 0;
		    lowercase(keybuf);
		    if (strcmp(keybuf, "include") == 0 && 
			(*tp == '\'' || *tp == '"' || *tp == '(')) {
			dlim = *tp;
			if (dlim == '(') dlim = ')';
			tp++;
			for (key = keybuf; tp < op && *tp != dlim;)
				*key++ = *tp++;
			if (*tp == dlim) {
			    tp++;
			    while (tp < op && *tp == ' ') tp++;
			    if (tp == op || *tp == '!') {
				/* its a real include statement */
				*key = 0;
				currentfile->lineno--;
				newfile = NEW(&progtables, filetab);
				newfile->prev = currentfile;
				for (tp = newfile->filename, key = keybuf;
					*tp++ = *key++;);
				currentfile = newfile;
				while (newfile=newfile->prev) {
				  if (strcmp(newfile->filename,
					currentfile->filename)==0) {
				    fprintf(stderr, "already including file %s \n",
					currentfile->filename);
				    for (newfile=currentfile->prev; newfile;
					newfile=newfile->prev){
					fprintf(stderr, "file %s line %d \n",
					    newfile->filename,
					    newfile->lineno-1);
					fclose(newfile->file);
				    }
				    /* forget new file */
				    exit(1);
				} }
#ifdef debugreader
				fprintf(stderr, "opening include file %s\n",
					currentfile->filename);
#endif
				if ((currentfile->file = 
				  fopen(currentfile->filename,"r")) ==
				  NULL) {
				  fprintf(stderr, "can't open file %s\n",
					currentfile->filename);
				  exit(1);
				}
				currentfile->offset = 0;
				currentfile->lineno = 1;
				j = i;	/* throw include statement out */
	}   }	}   }	}   }
    } while (seg->newstmt && j >= i);
#ifdef debugreader
	fprintf(stderr, "readst returns line %d of %s newstmt %d, length %d\n%s\n",
		seg->lineno, seg->f->filename, seg->newstmt, seg->stlen, 
		seg->text);
#endif
    return(seg);
}
/* check $line directives for duplicate lines */
remove_dup_lineno(seg)
struct stseg *seg;
{
	int l;
	char *f;
	l = seg->lineno;
	f = seg->f->filename;
	if (seg->prev && seg->prev->lineno + 1 != seg->lineno) {
		for (seg = seg->prev; seg && 
			!strcmp(seg->f->filename, f) && seg->lineno >= l; 
			seg = seg->prev) {
#ifdef debug_remove_dup_lineno
	fprintf(stderr,"line %d file %s lineno regression: %s\n", 
		seg->lineno, seg->f->filename, seg->text);
#endif
			seg->lineno = 0;
			seg->f = 0;
	}	}
}
/* get a non-blank, non-tab character */
char input()
{
	register struct stseg *p;
	char c;
#ifdef debuginput
	if (lexbase == &keychkcontext)
		fprintf(stderr, "stidscan ");
#endif
	if (lexbase->unputsize > 0) {
		c = lexbase->unputbuffer[--lexbase->unputsize];
#ifdef debuginput
		fprintf(stderr, "from unput buffer[%d]", lexbase->unputsize);
#endif
	} else {
	    do {
		while (lexbase->inptr == lexbase->endptr) {
			if ((p = lexbase->stsegptr->next) == 0) {
				if (p = readst()) {
					p->prev = lexbase->stsegptr;
					lexbase->stsegptr->next = p;
					remove_dup_lineno(p);
				} else	return(0);
			}
			lexbase->stsegptr = p;
			lexbase->inptr = p->text;
			lexbase->endptr = p->text + p->stlen;
			if (p->newstmt) {
#ifdef debuginput
				fprintf(stderr, "input returns '\\n'\n");
#endif
				return('\n');
			}
		} 
	    } while ((c = *(lexbase->inptr++)) == ' ' || c == '\t');
	}
#ifdef debuginput
	fprintf(stderr, "input returns '%c'\n",c);
#endif
	return(c);
}

unput(c)
char c;
{
	register struct stseg *p;
	char cu, *cp;
#ifdef debuginput
	if (lexbase == &keychkcontext)
		fprintf(stderr, "stidscan ");
    if (c != '\n') fprintf(stderr, "unput '%c'\n",c);
    else fprintf(stderr, "unput '\\n'\n");
#endif
    for(;;){
	if (lexbase->inptr == lexbase->stsegptr->text) {
		p = lexbase->stsegptr->prev;
		if (!p) {
			if (lexbase->unputsize < MAXLINE) {
				lexbase->unputbuffer[lexbase->unputsize++] = c;
				break;
			} else {
				yyerror("unput past buffer size ");
				exit(1);
		}	}
		lexbase->stsegptr = p;
		lexbase->inptr = lexbase->endptr = p->text + p->stlen;
		if (c == '\n' && p->next->newstmt) return(0);
	}
	if (c == (cu = *(--lexbase->inptr))) 
		break;
	else if (cu != ' ') {
		if (lexbase->unputsize < MAXLINE) {
			lexbase->unputbuffer[lexbase->unputsize++] = c;
			break;
		} else {
			yyerror("unput past buffer size ");
			exit(1);
}   }	}	}

/* advance input "n" characters for hollerith constant collection */
hconskip()
{
	register struct lexcontext *lb;
	int n, nleft, i, oldyyleng;
	char c, *tp, *sp, *oldinptr;
	struct stseg *oldstsegptr, *p;

	lb = lexbase;
	oldinptr = lb->inptr;
	oldstsegptr = lb->stsegptr;
	oldyyleng = lb->yyleng;
	for (n=0, tp = lb->yytext; *tp >= '0' && *tp <= '9';)
		n = n * 10 + *tp++ - '0';
	tp = lb->yytext + lb->yyleng;
	lb->yyleng += n;
#ifdef debughconskip
	fprintf(stderr, "hconskip skipping %d characters \n", n);
#endif
	/* consume lines until satisfied */
	while ((nleft = lb->endptr - lb->inptr) < n) {
		/* copy rest of line */
		for (sp = lb->inptr; sp < lb->endptr; ) *tp++ = *sp++;
		lb->inptr = sp;
		n -= nleft;
		/* if col72 make sure line is fully extended with blanks */
		if (col72) {
			nleft = (lb->stsegptr->newstmt ? 72 : 66) - 
				lb->stsegptr->stlen;
			if (nleft > n) nleft = n;
			for (i = 0;i < nleft; i++) *tp++ = ' ';
			/* fix source to have trailing blanks for KAI/SVS interaction */
			if (!lexbase->assignscan) {
				for (i = 0;i < nleft; i++) *sp++ = ' ';
				lb->stsegptr->stlen += nleft;
			}
			n -= nleft;
		}
		if (n == 0) break;
		/* go to next statement segment */
		if ((p = lb->stsegptr->next) || (p = readst())) {
			p->prev = lb->stsegptr;
			lb->stsegptr->next = p;
			lb->stsegptr = p;
			lb->inptr = p->text;
			lb->endptr = p->text + p->stlen;
			remove_dup_lineno(p);
		}
		/* give up if not a continuation */
		if (!p || p->newstmt){
#ifdef debughconskip
			fprintf(stderr, "hollerith > statement by %d characters\n",
				n);
#endif
			/* forget we ever did anything */
			lb->stsegptr = oldstsegptr;
			lb->inptr = oldinptr;
			lb->yyleng = oldyyleng;
			(lb->yytext)[oldyyleng] = 0;
			return(0);	/* indicate error */
		}
	}
	/* copy rest out of current statement segment */
	sp = lb->inptr; 
	for (i = 0; i < n; i++) *tp++ = *sp++;
	lb->inptr = sp;
	*tp = 0;	/* terminate string */
#ifdef debughconskip
	fprintf(stderr, "hconskip returns length %d text %s\n", 
		lb->yyleng, lb->yytext);
#endif
	return(1);	/* indicate sucess */
}
initstrd()
{
	lexbase->stsegptr = readst();
	lexbase->inptr = lexbase->stsegptr->text;
	lexbase->endptr = lexbase->inptr + lexbase->stsegptr->stlen;
}
idgobble()
{
	char *cp, c;
	if (lexbase->sttype != kFORMAT) {
		cp = lexbase->yytext;
		while (c = *cp++) 
			if (c >= 'A' && c <= 'Z') cp[-1] += 'a' - 'A';
		cp--;
		while (((c = input()) >= 'a' && c <= 'z') ||
			(c >= 'A' && c <= 'Z' && c += 'a' -'A') || 
			(c >= '0' && c <= '9') ||
			(systemids && (c == '%' || c == '_' || c == '$'))) {
			*cp++ = c;
			lexbase->yyleng++;
		}
		*cp = 0;
		unput(c);

		if (!lexbase->assignscan) {
			yylval = (int)oldsym(thisunit, lexbase->yytext);
			if (!yylval) yylval = (int)newsym(&pupass1,thisunit,
				lexbase->yytext,unitname);
			yydbval = NEW(&tempic,text);
			yydbval->next = 0;
			yydbval->loopseg = lexbase->tokenseg;
			yydbval->offset = lexbase->tokenchar - 
				lexbase->tokenseg->text;
		}
	}
}
